/*! \addtogroup gl
 * @{
 */

/*! \file
 * \brief Miscellaneous utilities for OpenGL
 */

#include <glee/GLee.h>
#include <GL/gl.h>
#include <algorithm>
#include "geom/coords.h"
#include "alodgl/glmisc.h"
#include <glib.h>
using namespace std;

GLuint alodgl_load_texture(const MeshTexture &image, GLenum internal_format,
                           bool mipmap, GError **err)
{
    bool alpha = (image.channels == 4);

    glPushClientAttrib(GL_CLIENT_PIXEL_STORE_BIT);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    if (!alpha)
    {
        switch (internal_format)
        {
        case GL_RGBA: internal_format = GL_RGB; break;
        case GL_RGBA8: internal_format = GL_RGB8; break;
        }
    }
    glTexImage2D(GL_PROXY_TEXTURE_2D, 0, internal_format,
                 image.width,
                 image.height,
                 0, alpha ? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE,
                 image.data());
    GLint width;
    glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &width);
    if (width == 0)
    {
        g_set_error(err,
                    g_quark_from_static_string("glmisc"),
                    0,
                    "Image dimensions (%ldx%ld) are unsuitable for OpenGL",
                    (long) image.width,
                    (long) image.height);
        glPopClientAttrib();
        return 0;
    }

    GLuint texid;
    glGenTextures(1, &texid);
    glBindTexture(GL_TEXTURE_2D, texid);
    if (mipmap)
        glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
    if (image.sign && GLEE_NV_texture_expand_normal)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_UNSIGNED_REMAP_MODE_NV, GL_EXPAND_NORMAL_NV);
    glTexImage2D(GL_TEXTURE_2D, 0, internal_format,
                 image.width,
                 image.height,
                 0, alpha ? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE,
                 image.data());

    glPopClientAttrib();
    return texid;
}

Matrix4 alodgl_get_matrix(GLenum name)
{
    GLdouble cols[16];

    glGetDoublev(name, cols);
    return Matrix4(cols[0], cols[4], cols[8], cols[12],
                   cols[1], cols[5], cols[9], cols[13],
                   cols[2], cols[6], cols[10], cols[14],
                   cols[3], cols[7], cols[11], cols[15]);
}

void alodgl_load_matrix(const Matrix4 &mat)
{
    GLdouble cols[16] =
    {
        mat[0][0], mat[1][0], mat[2][0], mat[3][0],
        mat[0][1], mat[1][1], mat[2][1], mat[3][1],
        mat[0][2], mat[1][2], mat[2][2], mat[3][2],
        mat[0][3], mat[1][3], mat[2][3], mat[3][3]
    };

    glLoadMatrixd(cols);
}

void alodgl_mult_matrix(const Matrix4 &mat)
{
    const GLdouble cols[16] =
    {
        mat[0][0], mat[1][0], mat[2][0], mat[3][0],
        mat[0][1], mat[1][1], mat[2][1], mat[3][1],
        mat[0][2], mat[1][2], mat[2][2], mat[3][2],
        mat[0][3], mat[1][3], mat[2][3], mat[3][3]
    };

    glMultMatrixd(cols);
}

void alodgl_uniform_matrix(GLint location, const Matrix4 &mat)
{
    const GLfloat cols[16] =
    {
        mat[0][0], mat[1][0], mat[2][0], mat[3][0],
        mat[0][1], mat[1][1], mat[2][1], mat[3][1],
        mat[0][2], mat[1][2], mat[2][2], mat[3][2],
        mat[0][3], mat[1][3], mat[2][3], mat[3][3]
    };
    glUniformMatrix4fvARB(location, 1, false, cols);
}

#define CHECK_FRAMEBUFFER_CASE(name) case name: g_error(#name); break
void alodgl_check_framebuffer_status()
{
    GLenum status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    switch (status)
    {
    case GL_FRAMEBUFFER_COMPLETE_EXT: return;
        CHECK_FRAMEBUFFER_CASE(GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT);
        CHECK_FRAMEBUFFER_CASE(GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT);
        CHECK_FRAMEBUFFER_CASE(GL_FRAMEBUFFER_INCOMPLETE_DUPLICATE_ATTACHMENT_EXT);
        CHECK_FRAMEBUFFER_CASE(GL_FRAMEBUFFER_INCOMPLETE_DIMENSIONS_EXT);
        CHECK_FRAMEBUFFER_CASE(GL_FRAMEBUFFER_INCOMPLETE_FORMATS_EXT);
        CHECK_FRAMEBUFFER_CASE(GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT);
        CHECK_FRAMEBUFFER_CASE(GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT);
        CHECK_FRAMEBUFFER_CASE(GL_FRAMEBUFFER_UNSUPPORTED_EXT);
    default:
        g_error("Unknown return %#04x from glCheckFrameBufferStatusEXT", status);
    }
}

/*! @} */
