/*! \addtogroup gl
 * @{
 */

/*! \file
 * \brief Shader management for meshrender
 */

#include <glee/GLee.h>
#include <GL/gl.h>
#include <string>
#include <map>
#include <list>
#include <sstream>
#include "common/paths.h"
#include "alodgl/shaders.h"

using namespace std;

static map<string, string> shader_flags;

void alodgl_set_shader_flag(const string &key, const string &value)
{
    shader_flags[key] = value;
}

void alodgl_set_shader_flag(const string &key, int value)
{
    ostringstream stream;
    stream << value;
    shader_flags[key] = stream.str();
}

void alodgl_clear_shader_flags()
{
    shader_flags = map<string, string>();
}

void alodgl_clear_shader_flag(const string &key)
{
    shader_flags.erase(key);
}

bool alodgl_has_shader_flag(const string &key)
{
    return shader_flags.count(key);
}

string alodgl_get_shader_flag(const string &key)
{
    map<string, string>::const_iterator i = shader_flags.find(key);
    if (i != shader_flags.end()) return i->second;
    else return "";
}

GLhandleARB alodgl_load_hl_shader(GLenum type, const string &source,
                                  const string &origin) throw (bad_shader)
{
    GLhandleARB shader;
    GLint log_length, status;
    GLcharARB *log;
    string msg;
    const GLcharARB *source_ptr = source.c_str();
    GLint length = source.length();

    shader = glCreateShaderObjectARB(type);
    glShaderSourceARB(shader, 1, &source_ptr, &length);
    glCompileShaderARB(shader);
    glGetObjectParameterivARB(shader, GL_OBJECT_COMPILE_STATUS_ARB, &status);
    if (!status)
    {
        glGetObjectParameterivARB(shader, GL_OBJECT_INFO_LOG_LENGTH_ARB, &log_length);
        log = new GLcharARB[log_length];
        glGetInfoLogARB(shader, log_length, NULL, log);
        msg = "Failed to load shader " + origin + ":\n" + log;
        delete[] log;
        throw bad_shader(msg);
    }

    return shader;
}

void alodgl_load_ll_shader(GLenum target, const string &source,
                           const string &origin) throw (bad_shader)
{
    while (glGetError() != GL_NO_ERROR)
        g_warning("Error present at glProgramStringARB");
    glProgramStringARB(target, GL_PROGRAM_FORMAT_ASCII_ARB, source.size(), source.c_str());
    if (glGetError() != GL_NO_ERROR)
    {
        throw bad_shader("Failed to load shader " + origin + ":\n" +
                         (const char *) glGetString(GL_PROGRAM_ERROR_STRING_ARB));
    }
}

string alodgl_get_shader_file(const string &shader_name) throw (bad_shader)
{
    string file = resource_path("ALOD_SHADERS", PKGDATADIR, shader_name);
    gsize length;
    gchar *source = NULL;
    GError *error = NULL;
    string ans;

    if (!g_file_get_contents(file.c_str(), &source, &length, &error))
    {
        string msg = error->message;
        g_error_free(error);
        throw bad_shader(msg);
    }

    ans = source;
    g_free(source);
    return ans;
}

string alodgl_get_shader_file_m4(const string &shader_name) throw (bad_shader)
{
    string file = resource_path("ALOD_SHADERS", PKGDATADIR, shader_name);

    gchar *argv[4 + 2 * shader_flags.size()];
    gchar *std_out = NULL, *std_err = NULL;
    gint status = 0;
    GError *error = NULL;
    int argc = 0;
    list<string> storage; /* Backing data for argv */
    gchar *dir_path;

    argv[argc++] = "m4";
    dir_path = g_path_get_dirname(file.c_str());
    storage.push_back(string("-I") + dir_path);
    g_free(dir_path);
    argv[argc++] = (gchar *) storage.back().c_str();

    for (map<string, string>::iterator i = shader_flags.begin();
         i != shader_flags.end(); i++)
    {
        argv[argc++] = "-D";
        if (i->second == "")
            argv[argc++] = (gchar *) i->first.c_str();
        else
        {
            storage.push_back(i->first + "=" + i->second);
            argv[argc++] = (gchar *) storage.back().c_str();
        }
    }
    argv[argc++] = (gchar *) file.c_str();
    argv[argc] = NULL;

    if (!g_spawn_sync(NULL,          /* working dir */
                      argv,
                      NULL,          /* environment */
                      G_SPAWN_SEARCH_PATH,
                      NULL,          /* child_setup */
                      NULL,          /* user_data */
                      &std_out,
                      &std_err,
                      &status,
                      &error))
    {
        string msg = error->message;
        g_error_free(error);
        throw bad_shader(msg);
    }
    if (status != 0)
    {
        string msg = std_err;
        g_free(std_out);
        g_free(std_err);
        throw bad_shader(msg);
    }

    string text = std_out;
    g_free(std_out);
    g_free(std_err);
    return text;
}

void alodgl_link_program(GLhandleARB program) throw (bad_shader)
{
    string msg;
    GLcharARB *log;
    GLint log_length, status;

    glLinkProgramARB(program);
    glGetObjectParameterivARB(program, GL_OBJECT_LINK_STATUS_ARB, &status);
    if (!status)
    {
        glGetObjectParameterivARB(program, GL_OBJECT_INFO_LOG_LENGTH_ARB, &log_length);
        log = new GLcharARB[log_length];
        glGetInfoLogARB(program, log_length, NULL, log);
        msg = string("Failed to link program:\n") + log;
        delete[] log;
        throw bad_shader(msg);
    }
}

/*! @} */
