/*! \file maputils.h
 * \brief Defines classes to wrap aspects of \c map and \c set.
 *
 * If the SGI/GNU extension classes \c hash_map and \c hash_set
 * are available, the wrappers map to them. Otherwise they map
 * to \c map and \c set. In addition, wrappers are provided for
 * performing these maps on iterators to objects with \c id fields,
 * performing the map on the ID. This is used extensively by the mesh
 * code.
 */

#ifndef ALOD_COMMON_MAPUTILS_H
#define ALOD_COMMON_MAPUTILS_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <utility>

//! Compares two iterators to objects with \c id fields by the \c id.
template<typename _T> class equal_id
{
public:
    bool operator ()(const _T &a, const _T &b) const
    { return a->id == b->id; }
};

//! Compares two pairs of iterators to objects with \c id fields by \c id.
template<typename _T1, typename _T2> class equal_id<std::pair<_T1, _T2> >
{
private:
    equal_id<_T1> _equal1;
    equal_id<_T2> _equal2;
public:
    bool operator ()(const std::pair<_T1, _T2> &a,
                     const std::pair<_T1, _T2> &b) const
    { return _equal1(a.first, b.first) && _equal2(a.second, b.second); }
};

//! Compares two iterators to objects with \c id fields by the \c id.
template<typename _T> class less_id
{
public:
    bool operator ()(const _T &a, const _T &b) const
    { return a->id < b->id; }
};

//! Compares two pairs of iterators to objects with \c id fields by \c id.
template<typename _T1, typename _T2> class less_id<std::pair<_T1, _T2> >
{
private:
    less_id<_T1> _c1;
    less_id<_T2> _c2;
public:
    bool operator ()(const std::pair<_T1, _T2> &a, const std::pair<_T1, _T2> &b) const
    {
        if (_c1(a.first, b.first)) return true;
        if (_c1(b.first, a.first)) return false;
        return _c2(a.second, b.second);
    }
};

#if HAVE_HASH_CLASSES
#if HAVE_EXT_HASH_MAP
# include <ext/hash_map>
#endif
#if HAVE_EXT_HASH_SET
# include <ext/hash_set>
#endif
#if HAVE_HASH_MAP
# include <hash_map>
#endif
#if HAVE_HASH_SET
# include <hash_set>
#endif

//! Shuffle the bits of a hash value so that it can be added to another
//! without excessive collisions.
static inline std::size_t _hash_swizzle(std::size_t x)
{
    return ((x << (sizeof(std::size_t) / 2)) | (x >> (sizeof(std::size_t) / 2))) & 0xaaaaaaaa;
}

//! Similar to \c less_id and \c equal_id, but computes a hash function
template<typename _T> class hash_id
{
private:
    HASH_NAMESPACE::hash<std::size_t> h;
public:
    std::size_t operator() (const _T &p) const
    { return h(p->id); }
};

//! Similar to \c less_id and \c equal_id, but computes a hash function
template<typename _T1, typename _T2> class hash_id<std::pair<_T1, _T2> >
{
private:
    hash_id<_T1> h1;
    hash_id<_T2> h2;
public:
    std::size_t operator() (const std::pair<_T1, _T2> &p) const
    { return _hash_swizzle(h1(p.first)) ^ h2(p.second); }
};

/* HASH_NAMESPACE is either std or __gnu_cxx, depending on where
 * gcc puts the hash classes.
 */
namespace HASH_NAMESPACE
{
    template<typename _T1, typename _T2>
    class hash<std::pair<_T1, _T2> >
    {
    private:
        hash<_T1> h1;
        hash<_T2> h2;
    public:
        std::size_t operator() (const std::pair<_T1, _T2> &p) const
        { return _hash_swizzle(h1(p.first)) ^ h2(p.second); }
    };
}

template<typename _K, typename _V>
class fast_map
{
public:
    typedef typename HASH_NAMESPACE::hash_map<_K, _V> type;
};

template<typename _K, typename _V>
class id_map
{
public:
    typedef typename HASH_NAMESPACE::hash_map<_K, _V, hash_id<_K>, equal_id<_K> > type;
};

template<typename _K>
class fast_set
{
public:
    typedef typename HASH_NAMESPACE::hash_set<_K> type;
};

template<typename _K>
class id_set
{
public:
    typedef typename HASH_NAMESPACE::hash_set<_K, hash_id<_K>, equal_id<_K> > type;
};

#else /* !HAVE_HASH_CLASSES */
#include <map>
#include <set>

//! A wrapper around either \c map or \c hash_map
template<typename _K, typename _V>
class fast_map
{
public:
    //! The actual type to use i.e. use \c fast_map::type.
    typedef typename std::map<_K, _V, std::less<_K> > type;
};

//! A wrapper around \c map or \c hash_map that maps on \c obj->id
template<typename _K, typename _V>
class id_map
{
public:
    //! The actual type to use i.e. use \c id_map::type.
    typedef typename std::map<_K, _V, less_id<_K> > type;
};

//! A wrapper around either \c set or \c hash_set
template<typename _K>
class fast_set
{
public:
    //! The actual type to use i.e. use \c fast_set::type.
    typedef typename std::set<_K> type;
};

//! A wrapper around \c set or \c hash_set that maps on \c obj->id
template<typename _K>
class id_set
{
public:
    //! The actual type to use i.e. use \c id_set::type.
    typedef typename std::set<_K, less_id<_K> > type;
};

#endif /* !HAVE_HASH_CLASSES */

#endif /* ALOD_COMMON_MAPUTILS_H */
