#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <unistd.h>
#include <cstdio>
#if HAVE_GETOPT_H
# ifndef _GNU_SOURCE
#  define _GNU_SOURCE
# endif
# include <getopt.h>
# define GNU_OPTIONS 1
#else
# define GNU_OPTIONS 0
#endif
#include <iostream>
#include <sstream>
#include <string>
#include <time.h>
#include "options.h"

using namespace std;

AlodProgramOptionInfo::AlodProgramOptionInfo()
    : data(NULL)
{
}

AlodProgramOptionInfo::AlodProgramOptionInfo(char short_opt,
                                             const std::string &long_opt,
                                             const std::string &help,
                                             int has_arg,
                                             AlodProgramArgumentType arg_type,
                                             AlodProgramOption *data)
    : short_opt(short_opt), long_opt(long_opt), help(help), has_arg(has_arg),
    arg_type(arg_type), data(data)
{
}

void alod_show_help(const AlodProgramOptionInfo *options)
{
    string help = "";
    for (; options->data != NULL; options++)
    {
        string line, arg;
        int has_arg;

#if !GNU_OPTIONS
        if (!options->short_opt) return;
#endif

        if (options->short_opt)
            line = string("  -") + options->short_opt;
        has_arg = options->has_arg;
#if GNU_OPTIONS
        if (!options->long_opt.empty())
        {
            if (options->short_opt) line += ", ";
            else line += "      ";
            line += "--" + options->long_opt;
        }
#else
        if (has_arg == 2) has_arg = 0;
#endif
        switch (options->arg_type)
        {
        case PROGRAM_ARGUMENT_STRING: arg = "<str>"; break;
        case PROGRAM_ARGUMENT_DOUBLE: arg = "<num>"; break;
        case PROGRAM_ARGUMENT_INT: arg = "<int>"; break;
        }
        switch (options->has_arg)
        {
        case 1:
#if GNU_OPTIONS
            if (!options->long_opt.empty())
                line += "=" + arg;
            else
#endif
            {
                line += " " + arg;
            }
            break;
        case 2:
#if GNU_OPTIONS
            if (!options->long_opt.empty())
                line += "[=" + arg + "]";
            else
#endif
            {
                line += " [" + arg + "]";
            }
            break;
        }
        while (line.length() < 25) line += " ";
        line += "   " + options->help;
        help += line + "\n";
    }
    help += "  -h, --help                this help text\n";
    help += "  -V, --version             show version information\n";
    fprintf(stderr, "%s", help.c_str());
}

int alod_process_options(int argc, char * const argv[],
                         const AlodProgramOptionInfo *options)
{
    string optstring = "hV";
    int opt;
    int opt_count;

    for (opt_count = 0; options[opt_count].data; opt_count++);

#if GNU_OPTIONS
    struct option long_opts[opt_count + 3];
    memset(long_opts, 0, sizeof(long_opts));
    long_opts[0].name = "help";
    long_opts[0].val = 'h';
    long_opts[1].name = "version";
    long_opts[1].val = 'V';
#endif

    for (int i = 0; i < opt_count; i++)
    {
        if (options[i].short_opt)
        {
            optstring += options[i].short_opt;
            if (options[i].has_arg >= 1) optstring += ":";
#if GNU_OPTIONS
            if (options[i].has_arg >= 2) optstring += ":";
#endif
        }
#if GNU_OPTIONS
        if (!options[i].long_opt.empty())
        {
            long_opts[i + 2].name = options[i].long_opt.c_str();
            long_opts[i + 2].has_arg = options[i].has_arg;
            if (options[i].short_opt)
                long_opts[i + 2].val = options[i].short_opt;
            else
                long_opts[i + 2].val = 256 + i;
        }
#endif
        options[i].data->used = false;
        options[i].data->arg = false;
        options[i].data->string_value = "";
        options[i].data->double_value = 0.0;
        options[i].data->int_value = 0;
    }

#if GNU_OPTIONS
    opt = getopt_long(argc, argv, optstring.c_str(), long_opts, NULL);
#else
    opt = getopt(argc, argv, optstring.c_str());
#endif
    while (opt != -1)
    {
        if (opt == '?' || opt == ':')
        {
            alod_show_help(options);
            exit(1);
        }
        else if (opt == 'h')
        {
            alod_show_help(options);
            exit(1);
        }
        else if (opt == 'V')
        {
            cout << PACKAGE_STRING << "\n";
            exit(0);
        }
        else
        {
            int i;
            if (opt >= 256) i = opt - 256;
            else
            {
                for (i = 0; i < opt_count; i++)
                    if (options[i].short_opt == opt) break;
                if (i == opt_count) continue;
            }
            options[i].data->used = true;
            if (options[i].has_arg && optarg)
            {
                options[i].data->string_value = optarg;
                options[i].data->arg = true;
                char *endptr;
                switch (options[i].arg_type)
                {
                case PROGRAM_ARGUMENT_DOUBLE:
                    options[i].data->double_value = strtod(optarg, &endptr);
                    if (endptr == optarg || *endptr)
                    {
                        fflush(stdout);
                        fprintf(stderr, "Error: expected a real number for -%c\n",
                                options[i].short_opt);
                        alod_show_help(options);
                        exit(1);
                    }
                    break;
                case PROGRAM_ARGUMENT_INT:
                    options[i].data->int_value = strtol(optarg, &endptr, 0);
                    if (endptr == optarg || *endptr)
                    {
                        fflush(stdout);
                        fprintf(stderr, "Error: expected an integer for -%c\n",
                                options[i].short_opt);
                        alod_show_help(options);
                        exit(1);
                    }
                    break;
                case PROGRAM_ARGUMENT_STRING:
                    break;
                }
            }
        }
#if GNU_OPTIONS
        opt = getopt_long(argc, argv, optstring.c_str(), long_opts, NULL);
#else
        opt = getopt(argc, argv, optstring.c_str());
#endif
    }
    return optind;
}

string alod_generation_string(int argc, const char * const *argv)
{
    ostringstream s;
    string ans;
    time_t t;

    t = time(NULL);

    s << "Generated by '";
    for (int i = 0; i < argc - 1; i++)
        s << argv[i] << ' ';
    s << argv[argc - 1] << "' at " << ctime(&t);
    ans = s.str();

    // some bright spark decided that ctime adds a \n. Let's get rid of it...
    while (!ans.empty() && ans[ans.length() - 1] == '\n')
        ans.erase(ans.length() - 1);
    return ans;
}
