/*! \file
 * \brief Options processing support
 */

#ifndef ALOD_COMMON_OPTIONS_H
#define ALOD_COMMON_OPTIONS_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <string>

//! Predefined option types, for help and for auto-conversion
enum AlodProgramArgumentType
{
    PROGRAM_ARGUMENT_STRING,
    PROGRAM_ARGUMENT_DOUBLE,
    PROGRAM_ARGUMENT_INT
};

//! The output from a program option
struct AlodProgramOption
{
    bool used;                 //!< True if the option was specified
    bool arg;                  //!< True if the option had an argument
    std::string string_value;  //!< The value of the argument, if any
    //! The value converted to double, if the type was \c PROGRAM_ARGUMENT_DOUBLE
    double double_value;
    //! The value converted to int, if the type was \c PROGRAM_ARGUMENT_INT
    int int_value;
};

//! The input from the program to the option processor
struct AlodProgramOptionInfo
{
    char short_opt;         //!< Short option letter, or \c \\0 for none
    std::string long_opt;   //!< The GNU-style long option, or \c "" for none
    std::string help;       //!< The help string to show with \c -h
    int has_arg;            //!< 0 for no arg, 1 for required, 2 for optional
    //! The type of argument (string is most generic)
    AlodProgramArgumentType arg_type;

    //! The area in which the processor will write its results
    AlodProgramOption *data;

    AlodProgramOptionInfo();
    AlodProgramOptionInfo(char short_opt, const std::string &long_opt,
                          const std::string &help, int has_arg,
                          AlodProgramArgumentType arg_type,
                          AlodProgramOption *data);
};

//! Process command line options
/*! \param argc the number of arguments (arg 0 is ignored)
 * \param argv a pointer to the arguments, which may be permuted
 * \param options the options that the program accepts (not \c -h or \c -V)
 * \returns the first element of \a argv that contains a non-option
 *
 * The last element of \a options must have NULL for \c data, indicating that
 * it is a terminator.
 *
 * \test Check that null short or long option works
 */
int alod_process_options(int argc, char * const argv[],
                         const AlodProgramOptionInfo *options);

//! Shows usage information on stderr
void alod_show_help(const AlodProgramOptionInfo *options);

//! Turns argv into a string for writing generation comments
std::string alod_generation_string(int argc, const char * const *argv);

#endif /* ALOD_COMMON_OPTIONS_H */
