/*! \addtogroup format
 * @{
 */

/*! \file
 * \brief Internal ALM file format
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <iostream>
#include <algorithm>
#include <vector>
#include <string>
#include <cstddef>
#include <glib.h>
#include <cstdio>
#include <sstream>
#include <iterator>
#include "formats.h"
#include "alm.h"
#include "common/maputils.h"
#include "mesh/mesh.h"
#include "geom/coords.h"
#include "geom/anim.h"
#include "geom/skeleton.h"
using namespace std;

// the minimum version that can be read
#define ALM_MIN_VERSION 2
// the maximum version that can be read, and the version that is written
#define ALM_VERSION 10

#if ALM_DEBUG
# define format_read_binary(in, big_endian, fmt) \
    ({ double ans = format_read_binary(in, big_endian, fmt); cerr << "Read " << ans << "\n"; ans; })
#endif

static inline gint32 alm_read_int(istream &in) throw (bad_format)
{
    return (gint32) format_read_binary(in, true, FORMAT_TYPE_INT32);
}

static inline guint32 alm_read_uint(istream &in) throw (bad_format)
{
    return (guint32) format_read_binary(in, true, FORMAT_TYPE_UINT32);
}

static inline double alm_read_float(istream &in, guint32 version) throw (bad_format)
{
    if (version <= 2)
        return format_read_binary(in, true, FORMAT_TYPE_FLOAT64);
    else
        return format_read_binary(in, true, FORMAT_TYPE_FLOAT32);
}

static Matrix3 alm_read_matrix3(istream &in, guint32 version) throw (bad_format)
{
    Matrix3 ans;

    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
            ans[i][j] = alm_read_float(in, version);
    return ans;
}

static Matrix4 alm_read_matrix4(istream &in, guint32 version) throw (bad_format)
{
    Matrix4 ans;

    for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++)
            ans[i][j] = alm_read_float(in, version);
    return ans;
}

static Vector3 alm_read_vector3(istream &in, guint32 version) throw (bad_format)
{
    Vector3 ans;

    for (int i = 0; i < 3; i++) ans[i] = alm_read_float(in, version);
    return ans;
}

static VectorPosition alm_read_vector(istream &in,
                                      const Skeleton &skeleton,
                                      guint32 version) throw (bad_format)
{
    VectorPosition ans(skeleton);

    guint32 count = alm_read_uint(in);

    for (guint32 i = 0; i < count; i++)
    {
        guint32 index = alm_read_uint(in);
        if (index >= skeleton.size()) throw bad_format("Bone index is out of range");
        index = skeleton.get_new_index(index);
        ans[4 * index + 0] = alm_read_float(in, version);
        ans[4 * index + 1] = alm_read_float(in, version);
        ans[4 * index + 2] = alm_read_float(in, version);
        ans[4 * index + 3] = alm_read_float(in, version);
    }
    return ans;
}

static MeshTangentMapEntry alm_read_tangent_map_entry(istream &in,
                                                      const Skeleton &skeleton,
                                                      guint32 version) throw (bad_format)
{
    MeshTangentMapEntry ans;

    ans.patch = alm_read_uint(in);
    ans.tangent_plane.tangents[0] = alm_read_vector(in, skeleton, version);
    ans.tangent_plane.tangents[1] = alm_read_vector(in, skeleton, version);
    return ans;
}

void alm_load_mesh(istream &in, Mesh &m, MeshLevel level,
                   vector<string> &comments,
                   const string &path,
                   const vector<string> &args) throw(bad_alm)
{
    char magic[4];
    guint32 version;
    guint32 num_comments, num_bones, num_positions, num_vertices,
        num_textures, num_materials, num_triangles, num_collapses;
    guint32 num_frames = 0;   // version 4
    guint32 num_tangent_maps = 0; // version 8
    double frame_rate = 30.0; // version 6
    guint32 override_P = 0;   // version 6
    m.clear();

    try
    {
        in.read(magic, 4);
        if (in.gcount() < 4) throw bad_format("Signature not found");
        if (magic[0] != 'A' || magic[1] != 'L' || magic[2] != 'M'
            || magic[3] != '\0')
            throw bad_format("Signature not found");
        version = alm_read_uint(in);
        if (version < ALM_MIN_VERSION || version > ALM_VERSION)
        {
            ostringstream o;
            o << "version mismatch: got " << version << ", expected " << ALM_MIN_VERSION << " to " << ALM_VERSION;
            throw bad_format(o.str());
        }

        num_comments = alm_read_uint(in);
        num_bones = alm_read_uint(in);
        num_positions = alm_read_uint(in);
        num_vertices = alm_read_uint(in);
        num_textures = alm_read_uint(in);
        if (version >= 8) num_tangent_maps = alm_read_uint(in);
        num_materials = alm_read_uint(in);
        num_triangles = alm_read_uint(in);
        num_collapses = alm_read_uint(in);
        if (version >= 4) num_frames = alm_read_uint(in);
        if (version >= 6) frame_rate = alm_read_float(in, version);
        if (version >= 6) override_P = alm_read_uint(in);

        // comments
        for (guint32 i = 0; i < num_comments; i++)
        {
            guint32 l = alm_read_uint(in);
            char buffer[l];
            in.read(buffer, l);
            comments.push_back(string(buffer, l));
        }

        // bones
        vector<SkeletonBoneInfo> bones(num_bones);
        for (guint32 i = 0; i < num_bones; i++)
        {
            bones[i].parent = alm_read_int(in);
            if (version < 5)
                bones[i].rest_transform = matrix3_translate3_to_matrix4(alm_read_matrix3(in, version),
                                                                        alm_read_vector3(in, version));
            else
                bones[i].rest_transform = alm_read_matrix4(in, version);

            if (version < 6)
            {
                bones[i].flags = BONE_FLAG_UNIFORM_SCALE | BONE_FLAG_FIXED_SCALE;
                bones[i].tail.zero();
                bones[i].max_singular = 1.0;
                bones[i].translate_bounds.centre = matrix4_to_translate3(bones[i].rest_transform);
                bones[i].translate_bounds.radius = 0.0;
                bones[i].expect_L.identity();
                bones[i].expect_GTG.identity();
                bones[i].expect_LTGTGL = bones[i].translate_bounds.centre * bones[i].translate_bounds.centre;
            }
            else
            {
                bones[i].flags = alm_read_uint(in);
                if (version >= 10)
                    bones[i].tail = alm_read_vector3(in, version);
                else
                    bones[i].tail.zero();
                bones[i].max_singular = alm_read_float(in, version);
                bones[i].translate_bounds.centre = alm_read_vector3(in, version);
                bones[i].translate_bounds.radius = alm_read_float(in, version);
                bones[i].expect_L = alm_read_matrix4(in, version);
                bones[i].expect_GTG = alm_read_matrix4(in, version);
                bones[i].expect_LTGTGL = alm_read_float(in, version);
            }
        }
        Skeleton &skeleton = m.skeleton.get_write();
        skeleton.set_bones(bones);
        if (version < 10)
            skeleton.fix_tails();
        if (override_P)
        {
            MAT *P = m_get(4 * num_bones, 4 * num_bones);
            for (guint32 i = 0; i < num_bones; i++)
                for (guint32 j = 0; j < num_bones; j++)
                {
                    int i2 = skeleton.get_new_index(i);
                    int j2 = skeleton.get_new_index(j);
                    for (int k = 0; k < 4; k++)
                        for (int l = 0; l < 4; l++)
                            m_set_val(P, 4 * i2 + k, 4 * j2 + l,
                                      alm_read_float(in, version));
                }
            skeleton.set_P(P);
        }

        // positions
        MeshPositionIterator positions[num_positions];
        for (guint32 i = 0; i < num_positions; i++)
            positions[i] = m.positions.insert_lod(MeshPosition(alm_read_vector(in, skeleton, version)));

        // vertices
        MeshVertexIterator vertices[num_vertices];
        for (guint32 i = 0; i < num_vertices; i++)
        {
            guint32 p = alm_read_uint(in);
            if (p >= num_positions) throw bad_format("out of range position index");
            vertices[i] = m.vertices.insert_lod(MeshVertex());
            vertices[i]->position() = positions[p];
            if (version >= 7)
            {
                vertices[i]->data.tangent_plane.tangents[0] = alm_read_vector(in, skeleton, version);
                vertices[i]->data.tangent_plane.tangents[1] = alm_read_vector(in, skeleton, version);
            }
            else
            {
                vertices[i]->data.tangent_plane.tangents[0] = VectorPosition(skeleton);
                vertices[i]->data.tangent_plane.tangents[1] = VectorPosition(skeleton);
            }
            vertices[i]->data.texcoords[0] = alm_read_float(in, version);
            vertices[i]->data.texcoords[1] = alm_read_float(in, version);
        }

        // textures
        MeshTextureIterator textures[num_textures];
        for (guint32 i = 0; i < num_textures; i++)
        {
            textures[i] = m.textures.insert(m.textures.end(), MeshTexture());
            guint32 w = alm_read_uint(in);
            guint32 h = alm_read_uint(in);
            guint32 c = alm_read_uint(in);
            textures[i]->set_size(w, h, c);
            textures[i]->sign = alm_read_uint(in);
            in.read((char *) textures[i]->data(), w * h * c);
        }

        // tangent maps
        MeshTangentMapIterator tangent_maps[num_tangent_maps];
        for (guint32 i = 0; i < num_tangent_maps; i++)
        {
            tangent_maps[i] = m.tangent_maps.insert(m.tangent_maps.end(), MeshTangentMap());
            guint32 w = alm_read_uint(in);
            guint32 h = alm_read_uint(in);
            tangent_maps[i]->set_size(w, h);
            for (guint32 j = 0; j < w * h; j++)
                tangent_maps[i]->data()[j] = alm_read_tangent_map_entry(in, skeleton, version);
        }

        // materials
        MeshMaterialIterator materials[num_materials];
        for (guint32 i = 0; i < num_materials; i++)
        {
            materials[i] = m.materials.insert(m.materials.end(), MeshMaterial());
            gint32 t = alm_read_int(in);
            if (t == -1)
                materials[i]->texture() = mesh_null_texture;
            else if (t < -1 || t >= (gint32) num_textures)
                throw bad_format("out of range texture index");
            else
                materials[i]->texture() = textures[t];
            if (version < 8)
            {
                /* Suck out the data and ignore it */
                guint32 w = alm_read_uint(in);
                for (guint32 j = 0; j < w; j++)
                    alm_read_uint(in);
                for (guint32 j = 0; j < w; j++)
                    alm_read_uint(in);
                materials[i]->tangent_map() = mesh_null_tangent_map;
            }
            else
            {
                t = alm_read_int(in);
                if (t == -1)
                    materials[i]->tangent_map() = mesh_null_tangent_map;
                else if (t < -1 || t >= (gint32) num_tangent_maps)
                    throw bad_format("out of range tangent map index");
                else
                    materials[i]->tangent_map() = tangent_maps[t];
            }
        }

        // triangles
        for (guint32 i = 0; i < num_triangles; i++)
        {
            MeshFaceIterator f;
            MeshHalfEdgeIterator e[3];

            f = m.faces.insert_real(MeshFace());
            for (int j = 0; j < 3; j++)
            {
                guint32 id = alm_read_uint(in);
                if (id >= num_vertices) throw bad_format("out of range vertex index");
                e[j] = m.edges.insert_real(MeshHalfEdge());
                e[j]->face() = f;
                e[j]->vertex() = vertices[id];
                m.vertices.move_to_real(e[j]->vertex());
                m.positions.move_to_real(e[j]->vertex()->position());
            }
            for (int j = 0; j < 3; j++)
            {
                e[j]->next() = e[(j + 1) % 3];
                e[j]->prev() = e[(j + 2) % 3];
            }
            f->edges() = e[0];

            guint32 t = alm_read_uint(in);
            if (t >= num_materials) throw bad_format("out of range material index");
            f->material() = materials[t];
            f->patch = -1;
        }

        // collapses
        for (guint32 i = 0; i < num_collapses; i++)
        {
            guint32 p1, p2, vl, vr;
            MeshCollapse c;

            p1 = alm_read_uint(in);
            if (version <= 8 || p1 != G_MAXUINT32)
            {
                p2 = alm_read_uint(in);
                vl = alm_read_uint(in);
                vr = alm_read_uint(in);

                c.is_edge_collapse = true;
                MeshEdgeCollapse &ec = c.edge_collapse;
                if (p1 >= num_positions || p2 >= num_positions)
                    throw bad_format("out of range position index");
                if (vl >= num_vertices || vr >= num_vertices)
                    throw bad_format("out of range vertex index");

                ec.p1() = positions[p1];
                ec.p2() = positions[p2];
                ec.vl() = vertices[vl];
                ec.vr() = vertices[vr];
            }
            else
            {
                c.is_edge_collapse = false;
                MeshInfluenceCollapse &ic = c.influence_collapse;
                guint32 num = alm_read_uint(in);
                for (guint32 j = 0; j < num; j++)
                {
                    vl = alm_read_uint(in);
                    vr = alm_read_uint(in);
                    if (vl >= num_vertices || vr >= num_vertices)
                        throw bad_format("out of range vertex index");
                    ic.push_back(MeshInfluenceCollapseElement());
                    ic.back().before() = vertices[vl];
                    ic.back().after() = vertices[vr];
                    if (ic.back().before()->position() != ic.front().before()->position())
                        throw bad_format("before-vertices in influence collapse do not share position");
                    if (ic.back().after()->position() != ic.front().after()->position())
                        throw bad_format("after-vertices in influence collapse do not share position");
                }
            }
            c.error = alm_read_float(in, version);
            m.append_collapse(c);
        }

        // frames
        for (guint32 i = 0; i < num_frames; i++)
        {
            vector<Matrix4> transforms;
            transforms.resize(num_bones);
            for (guint32 b = 0; b < num_bones; b++)
            {
                if (version < 5)
                {
                    Matrix4 m = bones[b].rest_transform;
                    for (int j = 0; j < 3; j++)
                        for (int k = 0; k < 3; k++)
                            m[j][k] = alm_read_float(in, version);
                    transforms[skeleton.get_new_index(b)] = m;
                }
                else
                    transforms[skeleton.get_new_index(b)] = alm_read_matrix4(in, version);
            }
            skeleton.add_frame(transforms);
        }
        skeleton.set_frame_rate(frame_rate);

        mesh_complete(m, level);
    }
    catch (bad_format &e)
    {
        throw bad_alm(e.what());
    }
    catch (mesh_nonmanifold &e)
    {
        throw bad_alm(string("Non-manifold mesh: ") + e.what());
    }
}

static inline void alm_write_int(ostream &out, gint32 x) throw (bad_format)
{
    format_write_binary(out, true, FORMAT_TYPE_INT32, x);
}

static inline void alm_write_uint(ostream &out, guint32 x) throw (bad_format)
{
    format_write_binary(out, true, FORMAT_TYPE_UINT32, x);
}

static inline void alm_write_float(ostream &out, double x) throw (bad_format)
{
    format_write_binary(out, true, FORMAT_TYPE_FLOAT32, x);
}

static void alm_write_matrix4(ostream &out, const Matrix4 &m) throw (bad_format)
{
    for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++)
            alm_write_float(out, m[i][j]);
}

static void alm_write_vector3(ostream &out, const Vector3 &v) throw (bad_format)
{
    for (int i = 0; i < 3; i++)
        alm_write_float(out, v[i]);
}

static void alm_write_vector(ostream &out, const VectorPosition &v) throw (bad_format)
{
    VectorPosition v2 = v;
    v2.compress();

    alm_write_uint(out, distance(v.begin(), v.end())); // count

    for (VectorPosition::const_iterator i = v.begin(); i != v.end(); i++)
    {
        alm_write_uint(out, i->index / 4); // bone id
        for (int j = 0; j < 4; j++)
            alm_write_float(out, i->values[j]);
    }
}

void alm_save_mesh(ostream &out, const Mesh &m,
                   const vector<string> &comments,
                   const string &path) throw (bad_alm)
{
    try
    {
        const Skeleton &skeleton = m.skeleton.get_read();
        out.write("ALM", 4);
        alm_write_uint(out, ALM_VERSION);
        alm_write_uint(out, comments.size());
        alm_write_uint(out, skeleton.size());           // bones
        alm_write_uint(out, m.positions.all_size());
        alm_write_uint(out, m.vertices.all_size());
        alm_write_uint(out, m.textures.size());
        alm_write_uint(out, m.tangent_maps.size());
        alm_write_uint(out, m.materials.size());
        alm_write_uint(out, m.faces.all_size());
        alm_write_uint(out, mesh_collapses(m));
        alm_write_uint(out, skeleton.num_frames());
        alm_write_float(out, skeleton.get_frame_rate());
        alm_write_uint(out, skeleton.has_custom_P() ? 1 : 0);

        // comments
        for (size_t i = 0; i < comments.size(); i++)
        {
            alm_write_uint(out, comments[i].length());
            out.write(comments[i].data(), comments[i].length());
        }

        // bones
        for (Skeleton::const_iterator i = skeleton.begin(); i != skeleton.end(); i++)
        {
            alm_write_int(out, i->parent);
            alm_write_matrix4(out, i->rest_transform);
            alm_write_uint(out, i->flags);
            alm_write_vector3(out, i->tail);
            alm_write_float(out, i->max_singular);
            alm_write_vector3(out, i->translate_bounds.centre);
            alm_write_float(out, i->translate_bounds.radius);
            alm_write_matrix4(out, i->expect_L);
            alm_write_matrix4(out, i->expect_GTG);
            alm_write_float(out, i->expect_LTGTGL);
        }
        if (skeleton.has_custom_P())
        {
            MAT *P = skeleton.get_P();
            for (size_t i = 0; i < 4 * skeleton.size(); i++)
                for (size_t j = 0; j < 4 * skeleton.size(); j++)
                    alm_write_float(out, m_entry(P, i, j));
        }

        // positions
        id_map<MeshPositionConstIterator, guint32>::type position_map;
        guint32 position_id = 0;
        for (MeshPositionConstIterator p = m.positions.all_begin(); p != m.positions.all_end(); p++)
        {
            alm_write_vector(out, p->position);
            position_map[p] = position_id++;
        }
        g_assert(position_map.size() == position_id);

        // vertices
        id_map<MeshVertexConstIterator, guint32>::type vertex_map;
        guint32 vertex_id = 0;
        for (MeshVertexConstIterator v = m.vertices.all_begin(); v != m.vertices.all_end(); v++)
        {
            alm_write_uint(out, position_map[v->position()]);
            alm_write_vector(out, v->data.tangent_plane.tangents[0]);
            alm_write_vector(out, v->data.tangent_plane.tangents[1]);
            alm_write_float(out, v->data.texcoords[0]);
            alm_write_float(out, v->data.texcoords[1]);
            vertex_map[v] = vertex_id++;
        }
        g_assert(vertex_map.size() == vertex_id);

        // textures
        id_map<MeshTextureConstIterator, guint32>::type texture_map;
        guint32 texture_id = 0;
        for (MeshTextureConstIterator t = m.textures.begin(); t != m.textures.end(); t++)
        {
            alm_write_uint(out, t->width);
            alm_write_uint(out, t->height);
            alm_write_uint(out, t->channels);
            alm_write_uint(out, t->sign ? 1 : 0);
            out.write((const char *) t->data(), t->width * t->height * t->channels);
            texture_map[t] = texture_id++;
        }

        // tangent maps
        id_map<MeshTangentMapConstIterator, guint32>::type tangent_map_map;
        guint32 tangent_map_id = 0;
        for (MeshTangentMapConstIterator tm = m.tangent_maps.begin(); tm != m.tangent_maps.end(); tm++)
        {
            alm_write_uint(out, tm->width);
            alm_write_uint(out, tm->height);
            for (size_t i = 0; i < tm->width * tm->height; i++)
            {
                alm_write_int(out, tm->data()[i].patch);
                alm_write_vector(out, tm->data()[i].tangent_plane.tangents[0]);
                alm_write_vector(out, tm->data()[i].tangent_plane.tangents[1]);
            }
            tangent_map_map[tm] = tangent_map_id++;
        }

        // materials
        id_map<MeshMaterialConstIterator, guint32>::type material_map;
        guint32 material_id = 0;
        for (MeshMaterialConstIterator s = m.materials.begin(); s != m.materials.end(); s++)
        {
            if (s->texture() == mesh_null_texture)
                alm_write_int(out, -1);
            else
                alm_write_int(out, texture_map[s->texture()]);
            if (s->tangent_map() == mesh_null_tangent_map)
                alm_write_int(out, -1);
            else
                alm_write_int(out, tangent_map_map[s->tangent_map()]);
            material_map[s] = material_id++;
        }

        // triangles
        for (MeshFaceConstIterator f = m.faces.all_begin(); f != m.faces.all_end(); f++)
        {
            MeshHalfEdgeConstIterator e = f->edges();
            for (int i = 0; i < 3; i++, e = e->next())
            {
                g_assert(vertex_map.count(e->vertex()));
                alm_write_uint(out, vertex_map[e->vertex()]);
            }
            alm_write_uint(out, material_map[f->material()]);
        }

        // collapses
        for (MeshCollapseConstIterator c = m.collapses_begin(); c != m.collapses_end(); c++)
        {
            if (c->is_edge_collapse)
            {
                const MeshEdgeCollapse &ec = c->edge_collapse;
                g_assert(position_map.count(ec.p1()));
                g_assert(position_map.count(ec.p2()));
                g_assert(vertex_map.count(ec.vl()));
                g_assert(vertex_map.count(ec.vr()));
                alm_write_uint(out, position_map[ec.p1()]);
                alm_write_uint(out, position_map[ec.p2()]);
                alm_write_uint(out, vertex_map[ec.vl()]);
                alm_write_uint(out, vertex_map[ec.vr()]);
            }
            else
            {
                const MeshInfluenceCollapse &ic = c->influence_collapse;
                alm_write_uint(out, G_MAXUINT32);
                alm_write_uint(out, ic.size());
                for (size_t i = 0; i < ic.size(); i++)
                {
                    g_assert(vertex_map.count(ic[i].before()));
                    g_assert(vertex_map.count(ic[i].after()));
                    alm_write_uint(out, vertex_map[ic[i].before()]);
                    alm_write_uint(out, vertex_map[ic[i].after()]);
                }
            }
            alm_write_float(out, c->error);
        }

        // frames
        for (size_t f = 0; f < skeleton.num_frames(); f++)
        {
            const vector<Matrix4> &frame = skeleton.get_frame(f);
            for (size_t b = 0; b < skeleton.size(); b++)
                alm_write_matrix4(out, frame[b]);
        }
    }
    catch (bad_format &e)
    {
        throw bad_alm(e.what());
    }
}

/*! @} */
