/*! \addtogroup format
 * @{
 */

/*! \file
 * \brief Generic mesh loading and saving support
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <cstddef>
#include <string>
#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>
#include <glib.h>
#include "mesh/mesh.h"
#include "common/paths.h"
#include "formats.h"
using namespace std;

static size_t format_sizes[] =
{
    1, 1,     /* int8, uint8 */
    2, 2,     /* int16, uint16 */
    4, 4,     /* int32, uint32 */
    4, 8      /* float32, float64 */
};

// Low level I/O code
double format_read_binary(std::istream &in, bool big_endian,
                          FormatNumberType fmt) throw(bad_format)
{
    union
    {
        char buffer[8];
        gint8 i8;
        guint8 u8;
        gint16 i16;
        guint16 u16;
        gint32 i32;
        guint32 u32;
        float f32;
        double f64;
    } data;
    bool rev;

    in.read(data.buffer, format_sizes[fmt]);
    if (!in)
    {
        if (in.eof()) throw bad_format("unexpected end of file");
        else throw bad_format("I/O error reading file");
    }
    switch (fmt)
    {
    case FORMAT_TYPE_INT8: return data.i8;
    case FORMAT_TYPE_UINT8: return data.u8;
    case FORMAT_TYPE_INT16:
        return big_endian ? GINT16_FROM_BE(data.i16) : GINT16_FROM_LE(data.i16);
    case FORMAT_TYPE_UINT16:
        return big_endian ? GUINT16_FROM_BE(data.u16) : GUINT16_FROM_LE(data.u16);
    case FORMAT_TYPE_INT32:
        return big_endian ? GINT32_FROM_BE(data.i32) : GINT32_FROM_LE(data.i32);
    case FORMAT_TYPE_UINT32:
        return big_endian ? GUINT32_FROM_BE(data.u32) : GUINT32_FROM_LE(data.u32);
    case FORMAT_TYPE_FLOAT32:
    case FORMAT_TYPE_FLOAT64:
        /* We have to hope that things are in IEEE format. Also
         * ignore natively PDP endian machines since I have no idea
         * what they would do to floating point (and probably wouldn't
         * use IEEE anyway).
         */
        switch (G_BYTE_ORDER)
        {
        case G_LITTLE_ENDIAN: rev = big_endian; break;
        case G_BIG_ENDIAN: rev = !big_endian; break;
        default: g_assert_not_reached();
        }
        if (rev) reverse(data.buffer, data.buffer + format_sizes[fmt]);
        if (fmt == FORMAT_TYPE_FLOAT32) return data.f32;
        else return data.f64;
    }
    g_return_val_if_reached(0);
}

void format_write_binary(std::ostream &out, bool big_endian,
                         FormatNumberType fmt, double value) throw(bad_format)
{
    union
    {
        char buffer[8];
        gint8 i8;
        guint8 u8;
        gint16 i16;
        guint16 u16;
        gint32 i32;
        guint32 u32;
        float f32;
        double f64;
    } data;
    bool rev;

    switch (fmt)
    {
    case FORMAT_TYPE_INT8: data.i8 = (gint8) value;
    case FORMAT_TYPE_UINT8: data.u8 = (guint8) value;
    case FORMAT_TYPE_INT16:
        data.i16 = big_endian ?
            GINT16_TO_BE((gint16) value) : GINT16_TO_LE((gint16) value);
        break;
    case FORMAT_TYPE_UINT16:
        data.u16 = big_endian ?
            GUINT16_TO_BE((guint16) value) : GUINT16_TO_LE((guint16) value);
        break;
    case FORMAT_TYPE_INT32:
        data.i32 = big_endian ?
            GINT32_TO_BE((gint32) value) : GINT32_TO_LE((gint32) value);
        break;
    case FORMAT_TYPE_UINT32:
        data.u32 = big_endian ?
            GUINT32_TO_BE((guint32) value) : GUINT32_TO_LE((guint32) value);
        break;
    case FORMAT_TYPE_FLOAT32:
    case FORMAT_TYPE_FLOAT64:
        if (fmt == FORMAT_TYPE_FLOAT32) data.f32 = (float) value;
        else data.f64 = value;
        switch (G_BYTE_ORDER)
        {
        case G_LITTLE_ENDIAN: rev = big_endian; break;
        case G_BIG_ENDIAN: rev = !big_endian; break;
        default: g_assert_not_reached();
        }
        if (rev) reverse(data.buffer, data.buffer + format_sizes[fmt]);
    }
    out.write(data.buffer, format_sizes[fmt]);
    if (!out) throw bad_format("I/O error writing file");
}

/********************************
 * High level stuff starts here *
 ********************************/

struct FormatFileFormat
{
    string name;
    string ext;
    string desc;
    void (*load_mesh)(istream &, Mesh &, MeshLevel, vector<string> &, const string &, const vector<string> &);
    void (*save_mesh)(ostream &, const Mesh &, const vector<string> &, const string &);
};

// Wrappers to save PLY in specific formats
static void ply_save_mesh_ascii(ostream &out, const Mesh &m,
                                const vector<string> &comments,
                                const string &path)
{
    ply_save_mesh(out, m, PLY_FORMAT_ASCII, comments, path);
}

static void ply_save_mesh_binary(ostream &out, const Mesh &m,
                                 const vector<string> &comments,
                                 const string &path)
{
    ply_save_mesh(out, m, PLY_FORMAT_LITTLE_ENDIAN, comments, path);
}

static FormatFileFormat format_file_formats[] =
{
    { "md3", "md3", "Quake3 MD3", md3_load_mesh, NULL },
    { "md5", "md5mesh", "Doom3 md5mesh", md5_load_mesh, md5_save_mesh },
    { "md5anim", "md5anim", "Doom3 md5anim", md5_load_anim, NULL },
    { "obj", "obj", "Wavefront OBJ", NULL, obj_save_mesh },
    { "plybin", "ply", "Stanford PLY (binary)", ply_load_mesh, ply_save_mesh_binary },
    { "plyascii", "ply", "Stanford PLY (ASCII)", ply_load_mesh, ply_save_mesh_ascii },
    { "alm", "alm", "internal ALM format", alm_load_mesh, alm_save_mesh },
    { "", "", "", NULL, NULL }
};

static FormatFileFormat *format_find_file_format(const string &fmt,
                                                 const string &fname) throw(bad_format)
{
    FormatFileFormat *format;
    if (fmt == "auto")
    {
        for (format = format_file_formats; format->name != ""; format++)
            if (fname.length() >= format->ext.length() + 1)
            {
                string suffix = fname.substr(fname.length() - (format->ext.length() + 1));
                if (suffix == "." + format->ext) break;
            }
    }
    else
    {
        for (format = format_file_formats; format->name != ""; format++)
            if (fmt == format->name) break;
    }
    if (format->name == "")
    {
        if (fmt == "auto")
            throw bad_format("unable to determine format from extension");
        else
            throw bad_format("unknown format " + fmt);
    }
    return format;
}

void format_load_mesh_comments(const string &fmt, istream &in,
                               Mesh &m, MeshLevel level,
                               vector<string> &comments) throw(bad_format)
{
    FormatFileFormat *format = format_find_file_format(fmt, "");
    if (!format->load_mesh)
        throw bad_format("no input support for " + format->desc + " format");
    mesh_clear(m);
    (*format->load_mesh)(in, m, level, comments, "", vector<string>());
}

void format_load_mesh_file_comments(const string &fmt, const string &fname,
                                    Mesh &m, MeshLevel level,
                                    vector<string> &comments) throw(bad_format)
{
    long fcount = -1, icount = -1;
    size_t p;
    string name = fname;
    double dummy = 0.0;
    vector<string> args;
    long frame = -1;

    if ((p = name.rfind('%')) != string::npos)
    {
        const char *c = name.c_str() + p + 1;
        char *e;
        frame = strtol(c, &e, 0);
        if (c != e && e[0] == '\0' && frame >= 0)
            name.erase(p, string::npos);
        else
            frame = -1;
    }

    if ((p = name.rfind('@')) != string::npos)
    {
        const char *c = name.c_str() + p + 1;
        char *e;
        fcount = strtol(c, &e, 0);
        if (c != e && e[0] == '\0' && fcount >= 0)
            name.erase(p, string::npos);
        else
            fcount = -1;
    }
    else if ((p = name.rfind('#')) != string::npos)
    {
        const char *c = name.c_str() + p + 1;
        char *e;
        icount = strtol(c, &e, 0);
        if (c != e && e[0] == '\0' && icount >= 0)
            name.erase(p, string::npos);
        else
            icount = -1;
    }

    while ((p = name.rfind(':')) != string::npos)
    {
        args.push_back(name.substr(p + 1));
        name.erase(p);
    }
    reverse(args.begin(), args.end());

    FormatFileFormat *format = format_find_file_format(fmt, name);
    if (!format->load_mesh)
        throw bad_format("no input support for " + format->desc + " format");
    ifstream in(name.c_str(), ios::binary);
    if (!in)
        throw bad_format("could not open " + name);

    if (fcount != -1 || icount != -1)
        level = MESH_LEVEL_PATCHES;
    (*format->load_mesh)(in, m, level, comments, name, args);

    if (fcount != -1) mesh_reduce(m, fcount, true, dummy);
    if (icount != -1) mesh_reduce(m, icount, false, dummy);
    if (frame != -1) m.skeleton.get_write().set_to_frame(frame);
}

void format_load_mesh(const string &fmt, istream &in,
                      Mesh &m, MeshLevel level) throw(bad_format)
{
    vector<string> dummy;
    format_load_mesh_comments(fmt, in, m, level, dummy);
}

void format_load_mesh_file(const string &fmt, const string &fname,
                           Mesh &m, MeshLevel level) throw(bad_format)
{
    vector<string> dummy;
    format_load_mesh_file_comments(fmt, fname, m, level, dummy);
}

void format_save_mesh(const string &fmt, ostream &out,
                      const Mesh &m,
                      const vector<string> &comments) throw(bad_format)
{
    FormatFileFormat *format = format_find_file_format(fmt, "");
    if (!format->save_mesh)
        throw bad_format("no output support for " + format->desc + " format");
    (*format->save_mesh)(out, m, comments, "");
}

void format_save_mesh_file(const string &fmt, const string &fname,
                           const Mesh &m,
                           const vector<string> &comments) throw(bad_format)
{
    FormatFileFormat *format = format_find_file_format(fmt, fname);
    if (!format->save_mesh)
        throw bad_format("no output support for " + format->desc + " format");
    ofstream out(fname.c_str(), ios::binary);
    if (!out)
        throw bad_format("could not open " + fname);
    (*format->save_mesh)(out, m, comments, fname);
}

/*! @} */
