/*! \addtogroup format
 * @{
 */

/*! \file
 * \brief Generic mesh loading and saving support
 */

#ifndef ALOD_FORMATS_FORMATS_H
#define ALOD_FORMATS_FORMATS_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdexcept>
#include <iostream>
#include <string>
#include <vector>
#include "mesh/mesh.h"

//! Specifies that something went wrong loading or saving a file
class bad_format : public std::runtime_error
{
public:
    bad_format(const std::string &msg) : std::runtime_error(msg) {}
};

// If this table is changed, remember to update the size table in
// formats.cpp
enum FormatNumberType
{
    FORMAT_TYPE_INT8 = 0,
    FORMAT_TYPE_UINT8 = 1,
    FORMAT_TYPE_INT16 = 2,
    FORMAT_TYPE_UINT16 = 3,
    FORMAT_TYPE_INT32 = 4,
    FORMAT_TYPE_UINT32 = 5,
    FORMAT_TYPE_FLOAT32 = 6,
    FORMAT_TYPE_FLOAT64 = 7
};

// These functions are intended for use by the specific format handlers

//! Reads raw binary data from file.
/*! \param in the input stream
 * \param big_endian true to read big-endian, false to read little-endian
 * \param fmt the type to load
 * \returns the value cast to \c double
 * \exception bad_format if an I/O error occurred
 */
double format_read_binary(std::istream &in, bool big_endian,
                          FormatNumberType fmt) throw(bad_format);
//! Writes raw binary data to file.
/*! \param out the output stream
 * \param big_endian true to write big-endian, false to write little-endian
 * \param fmt the type of the data
 * \param value the value to write, cast to \c double
 * \exception bad_format if an I/O error occurred
 */
void format_write_binary(std::ostream &out, bool big_endian,
                         FormatNumberType fmt, double value) throw(bad_format);

//! Loads a mesh from a stream
/*! \param fmt the file format to use
 * \param in the input stream
 * \param m the mesh to overwrite with the file data
 * \param level how much data to fill in on mesh (see \c Mesh)
 * \exception bad_format if there was an I/O error or badly formatted file
 * \exception bad_format if the mesh was completely broken (e.g. out of range indices)
 * \exception bad_format if the mesh was not compliant to level \a level
 */
void format_load_mesh(const std::string &fmt, std::istream &in,
                      Mesh &m, MeshLevel level) throw(bad_format);

//! Similar to \c format_load_mesh, but also appends any comments to \a comments
void format_load_mesh_comments(const std::string &fmt, std::istream &in,
                               Mesh &m, MeshLevel level,
                               std::vector<std::string> &comments) throw(bad_format);

//! Loads a mesh from a file
/*! Similar to \c format_load_mesh, except that it takes a filename rather
 * than an open stream. In addition, \a fmt may be \c "auto", in which case
 * the format is detected from the filename.
 */
void format_load_mesh_file(const std::string &fmt, const std::string &fname,
                           Mesh &m, MeshLevel level) throw(bad_format);

//! Similar to \c format_load_mesh_file, but also appends any comments to \a comments
void format_load_mesh_file_comments(const std::string &fmt,
                                    const std::string &fname,
                                    Mesh &m,
                                    MeshLevel level,
                                    std::vector<std::string> &comments) throw(bad_format);

//! Write a mesh to a stream
/*! \param fmt the file format to use
 * \param out the stream to which to write the mesh
 * \param m the mesh to write
 * \param comments a list of comments to write into the file
 * \exception bad_format if an I/O error occurred
 */
void format_save_mesh(const std::string &fmt, std::ostream &out,
                      const Mesh &m,
                      const std::vector<std::string> &comments) throw(bad_format);
//! Write a mesh to a file
/*! Similar to \c format_save_mesh, except that it takes a filename rather
 * than an open stream. In addition, \a fmt may be \c "auto", in which case
 * the format is detected from the filename.
 */
void format_save_mesh_file(const std::string &fmt, const std::string &fname,
                           const Mesh &m,
                           const std::vector<std::string> &comments) throw(bad_format);

// so that other files can just #include "formats/formats.h"

#include "formats/ply.h"
#include "formats/obj.h"
#include "formats/md3.h"
#include "formats/md5.h"
#include "formats/alm.h"

#endif /* ALOD_FORMATS_FORMATS_H */

/*! @} */
