/*! \addtogroup format
 * @{
 */

/*! \file
 * \brief Quake 3 MD3 model format support
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <iostream>
#include <algorithm>
#include <vector>
#include <cstddef>
#include <glib.h>
#include "md3.h"
#include "mesh/mesh.h"
using namespace std;

//! The magic number stored in the first four bytes (little endian)
#define MD3_MAGIC 0x33504449 /* "IDP3" */
//! The version number of the file format, common to all models
#define MD3_VERSION 15
//! The unit of fixed point arithmetic for vertex positions
#define MD3_XYZ_SCALE (1.0/64.0)
//! The size of a number of strings in the file format
#define MAX_QPATH 64

//! Data about a single frame of an animation sequence
struct Md3Frame
{
    // have to hope for IEEE floats
    float min_bounds[3];     //!< Bounding box corner
    float max_bounds[3];     //!< Bounding box corner
    float local_origin[3];
    float radius;            //!< Bounding sphere
    char name[16];
};

//! Tags are used to align multiple MD3 files for simple skeletal animation
struct Md3Tag
{
    char name[MAX_QPATH];
    float origin[3];
    float axis[3][3];
};

//! Names the file describing how to shade an \c Md3Surface
struct Md3Shader
{
    char name[MAX_QPATH];
    gint32 shader_index;
};

//! Specifies the indices of a triangle, in clockwise order
struct Md3Triangle
{
    gint32 indices[3];
};

//! Holds normalised texture coordinates for a vertex
/*! Texture coordinates are held separately from vertex data because
 * vertex data is different for each frame, while texture coordinates
 * are constant over an animation. However, there is a one-to-one
 * relation between vertices and texture coordinates.
 */
struct Md3TexCoord
{
    float st[2];
};

//! Holds position and normal data for a vertex
struct Md3Vertex
{
    gint16 x;         //!< X coordinate, divided by \c MD3_XYZ_SCALE
    gint16 y;         //!< Y coordinate, divided by \c MD3_XYZ_SCALE
    gint16 z;         //!< Z coordinate, divided by \c MD3_XYZ_SCALE
    //! Lat-long encoded normal
    /*! The first byte is the latitude, measured from the north pole
     * (0, 0, 1) and scaled by \f$\frac{255}{2\pi}\f$. It is in the range
     * 0 to 127, or 128 to indicate the south pole (0, 0, -1). The
     * second coordinate is signed longitude, again scaled by
     * \f$\frac{255}{2\pi}\f$.
     */
    gint8 normal[2];
};

//! Describes a surface in an MD3 file.
/*! Each surface has a common shading function. */
struct Md3Surface
{
    gint32 ident;              //!< Must equal \c MD3_MAGIC
    char name[MAX_QPATH];      //!< A symbolic name for the surface
    gint32 flags;              //!< Unknown
    gint32 num_frames;         //!< Animation frames; must match \c Md3Mesh::num_frames
    gint32 num_shaders;        //!< Number of shaders in this surface
    gint32 num_verts;          //!< Number of vertices in this surface
    gint32 num_triangles;      //!< Number of triangles in this surface
    gint32 ofs_triangles;      //!< Offset in file from start of this surface to \c num_triangles contiguous \c Md3Triangle blocks
    gint32 ofs_shaders;        //!< Offset in file from start of this surface to \c num_shaders contiguous \c Md3Shader blocks
    gint32 ofs_texcoords;      //!< Offset in file from start of this surface to \c num_verts contiguous \c Md3TexCoord blocks
    gint32 ofs_verts;          //!< Offset in file from start of this surface to \c num_frames times \c num_verts contiguous \c Md3Vertex blocks
    gint32 ofs_end;            //!< Size of this surface, including linked data?

    Md3Triangle *triangles;
    Md3Shader *shaders;
    Md3TexCoord *texcoords;
    Md3Vertex *vertices;

    Md3Surface();
    ~Md3Surface();
};

//! A Quake 3 MD3 model
struct Md3Mesh
{
    gint32 ident;              //!< Must equal MD3_MAGIC
    gint32 version;            //!< Must equal MD3_VERSION
    char name[MAX_QPATH];      //!< Symbolic name for the model
    gint32 flags;              //!< Unknown
    gint32 num_frames;         //!< Number of Md3Frame blocks
    gint32 num_tags;           //!< Number of Md3Tag blocks
    gint32 num_surfaces;       //!< Number of Md3Surface blocks
    gint32 num_skins;          //!< Apparently an unused holdover from MD2
    gint32 ofs_frames;         //!< Offset from start to mesh to \c num_frames contiguous Md3Frame blocks
    gint32 ofs_tags;           //!< Offset from start to mesh to \c num_tags contiguous Md3Tag blocks
    gint32 ofs_surfaces;       //!< Offset from start to mesh to \c num_surfaces contiguous Md3Surface blocks
    gint32 ofs_eof;            //!< Size of the mesh, including linked data?

    Md3Frame *frames;
    Md3Tag *tags;
    Md3Surface *surfaces;

    Md3Mesh();
    ~Md3Mesh();
};

//! Reads a signed 32-bit little endian
static inline gint32 md3_load_int32(std::istream &in) throw(bad_format)
{
    return (gint32) format_read_binary(in, false, FORMAT_TYPE_INT32);
}

//! Reads a signed 16-bit little endian
static inline gint16 md3_load_int16(std::istream &in) throw(bad_format)
{
    return (gint16) format_read_binary(in, false, FORMAT_TYPE_INT16);
}

//! Reads a signed 8-bit
static inline gint8 md3_load_int8(std::istream &in) throw(bad_format)
{
    return (gint8) format_read_binary(in, false, FORMAT_TYPE_INT8);
}

//! Reads a little endian IEEE 32-bit float
static float md3_load_float(std::istream &in) throw (bad_format)
{
    return (float) format_read_binary(in, false, FORMAT_TYPE_FLOAT32);
}

//! Reads a string of exactly \a count bytes
static void md3_load_string(std::istream &in, char *buffer, int count) throw (bad_md3)
{
    in.read(buffer, count);
    if (!in)
    {
        if (in.eof()) throw bad_md3("premature end of file");
        else throw bad_md3("I/O error reading file");
    }
}

Md3Surface::Md3Surface()
: triangles(NULL), shaders(NULL), texcoords(NULL), vertices(NULL) {}

Md3Surface::~Md3Surface()
{
    if (triangles) delete[] triangles;
    if (shaders) delete[] shaders;
    if (vertices) delete[] vertices;
    if (texcoords) delete[] texcoords;
}

Md3Mesh::Md3Mesh() : frames(NULL), tags(NULL), surfaces(NULL) {}

Md3Mesh::~Md3Mesh()
{
    if (tags) delete[] tags;
    if (frames) delete[] frames;
    if (surfaces) delete[] surfaces;
}

//! Loads an MD3 file from \a in, without interpreting it
static Md3Mesh *md3_read_mesh(std::istream &in) throw(bad_md3)
{
    Md3Mesh *md3 = new Md3Mesh;

    try
    {
        streampos md3_start = in.tellg();

        md3->ident = md3_load_int32(in);
        if (md3->ident != MD3_MAGIC)
            throw bad_md3("magic identifier missing");
        md3->version = md3_load_int32(in);
        if (md3->version != MD3_VERSION)
            throw bad_md3("version mismatch");
        md3_load_string(in, md3->name, sizeof(md3->name));
        md3->flags = md3_load_int32(in);
        md3->num_frames = md3_load_int32(in);
        md3->num_tags = md3_load_int32(in);
        md3->num_surfaces = md3_load_int32(in);
        md3->num_skins = md3_load_int32(in);
        md3->ofs_frames = md3_load_int32(in);
        md3->ofs_tags = md3_load_int32(in);
        md3->ofs_surfaces = md3_load_int32(in);
        md3->ofs_eof = md3_load_int32(in);
    
        md3->frames = new Md3Frame[md3->num_frames];
        md3->tags = new Md3Tag[md3->num_tags];
        md3->surfaces = new Md3Surface[md3->num_surfaces];

        in.seekg(md3_start + streampos(md3->ofs_frames), ios::beg);
        for (int i = 0; i < md3->num_frames; i++)
        {
            Md3Frame *frame = md3->frames + i;
            for (int j = 0; j < 3; j++)
                frame->min_bounds[j] = md3_load_float(in);
            for (int j = 0; j < 3; j++)
                frame->max_bounds[j] = md3_load_float(in);
            for (int j = 0; j < 3; j++)
                frame->local_origin[j] = md3_load_float(in);
            frame->radius = md3_load_float(in);
            md3_load_string(in, frame->name, sizeof(frame->name));
        }

        in.seekg(md3_start + streampos(md3->ofs_tags), ios::beg);
        for (int i = 0; i < md3->num_tags; i++)
        {
            Md3Tag *tag = md3->tags + i;
            md3_load_string(in, tag->name, sizeof(tag->name));
            for (int j = 0; j < 3; j++)
                tag->origin[j] = md3_load_float(in);
            for (int j = 0; j < 3; j++)
                for (int k = 0; k < 3; k++)
                    tag->axis[j][k] = md3_load_float(in);
        }

        in.seekg(md3_start + streampos(md3->ofs_surfaces), ios::beg);
        for (int i = 0; i < md3->num_surfaces; i++)
        {
            Md3Surface *surface = md3->surfaces + i;
            streampos surface_start = in.tellg();

            surface->ident = md3_load_int32(in);
            if (surface->ident != MD3_MAGIC)
                throw bad_md3("wrong magic number in surface");
            md3_load_string(in, surface->name, sizeof(surface->name));
            surface->flags = md3_load_int32(in);
            surface->num_frames = md3_load_int32(in);
            surface->num_shaders = md3_load_int32(in);
            surface->num_verts = md3_load_int32(in);
            surface->num_triangles = md3_load_int32(in);
            surface->ofs_triangles = md3_load_int32(in);
            surface->ofs_shaders = md3_load_int32(in);
            surface->ofs_texcoords = md3_load_int32(in);
            surface->ofs_verts = md3_load_int32(in);
            surface->ofs_end = md3_load_int32(in);

            surface->triangles = new Md3Triangle[surface->num_triangles];
            surface->shaders = new Md3Shader[surface->num_shaders];
            surface->vertices = new Md3Vertex[surface->num_verts];
            surface->texcoords = new Md3TexCoord[surface->num_verts];

            in.seekg(surface_start + streampos(surface->ofs_shaders));
            for (int j = 0; j < surface->num_shaders; j++)
            {
                Md3Shader *shader = surface->shaders + j;
                md3_load_string(in, shader->name, sizeof(shader->name));
                shader->shader_index = md3_load_int32(in);
            }

            in.seekg(surface_start + streampos(surface->ofs_triangles));
            for (int j = 0; j < surface->num_triangles; j++)
                for (int k = 0; k < 3; k++)
                    surface->triangles[j].indices[k] = md3_load_int32(in);

            in.seekg(surface_start + streampos(surface->ofs_texcoords));
            for (int j = 0; j < surface->num_verts; j++)
                for (int k = 0; k < 2; k++)
                    surface->texcoords[j].st[k] = md3_load_float(in);

            in.seekg(surface_start + streampos(surface->ofs_verts));
            for (int j = 0; j < surface->num_verts; j++)
            {
                Md3Vertex *vertex = surface->vertices + j;
                vertex->x = md3_load_int16(in);
                vertex->y = md3_load_int16(in);
                vertex->z = md3_load_int16(in);
                vertex->normal[0] = md3_load_int8(in);
                vertex->normal[1] = md3_load_int8(in);
            }

            in.seekg(surface_start + streampos(surface->ofs_end));
        }
    }
    catch (bad_md3)
    {
        delete md3;
        throw;
    }
    return md3;
}

/*! See \c Mesh for a description of mesh levels. */
void md3_load_mesh(istream &in, Mesh &m, MeshLevel level,
                   vector<string> &comments,
                   const string &path,
                   const vector<string> &args) throw(bad_md3)
{
    Md3Mesh *md3 = md3_read_mesh(in);

    vector<MeshRawVertex> vertices;
    vector<vector<int> > faces;
    vector<int> materials;

    const Skeleton &skeleton = m.skeleton.get_read();
    try
    {
        for (int i = 0; i < md3->num_surfaces; i++)
        {
            Md3Surface *surface = md3->surfaces + i;
            size_t base_vertex = vertices.size();
            size_t base_face = faces.size();

            vertices.resize(base_vertex + surface->num_verts);
            faces.resize(base_face + surface->num_triangles, vector<int>(3));
            materials.resize(base_face + surface->num_triangles, 0);
            for (int j = 0; j < surface->num_verts; j++)
            {
                size_t inx = base_vertex + j;
                Md3Vertex *v = surface->vertices + j;
                vertices[inx].position = skeleton.ssd_vector();
                vertices[inx].position[0] = v->x * MD3_XYZ_SCALE;
                vertices[inx].position[1] = v->y * MD3_XYZ_SCALE;
                vertices[inx].position[2] = v->z * MD3_XYZ_SCALE;
                vertices[inx].position[3] = 1.0;

                vertices[inx].texcoords[0] = surface->texcoords[j].st[0];
                vertices[inx].texcoords[1] = surface->texcoords[j].st[1];
            }
            for (int j = 0; j < surface->num_triangles; j++)
            {
                for (int k = 0; k < 3; k++)
                {
                    int inx = surface->triangles[j].indices[k];
                    if (inx < 0 || inx >= surface->num_verts)
                        throw bad_md3("vertex index is out of range");
                    /* It also seems as though MD3 uses backwards
                     * faces. At least this makes things look right
                     * in Blender...
                     */
                    faces[base_face + j][2 - k] = base_vertex + surface->triangles[j].indices[k];
                }
                materials[base_face + j] = i;
            }
        }
    }
    catch (bad_format &e)
    {
        delete md3;
        throw bad_md3(e.what());
    }

    delete md3;
    try
    {
        mesh_create(m, vertices, faces, materials, level);
    }
    catch (mesh_nonmanifold &e)
    {
        throw bad_md3(string("non-manifold mesh: ") + e.what());
    }
}

/*! @} */
