/*! \addtogroup format
 * @{
 */

/*! \file
 * \brief Doom3 MD5 model format support
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <iostream>
#include <sstream>
#include <algorithm>
#include <vector>
#include <string>
#include <cstddef>
#include <glib.h>
#include <cstdio>
#include "md5.h"
#include "md5lex.h"
#include "common/maputils.h"
#include "mesh/mesh.h"
#include "geom/skeleton.h"
using namespace std;

/* Globals for the lexer (since we cannot pass anything in) */
istream *md5_input;
vector<string> *md5_comments;
Md5Yylval md5_yylval;

typedef struct Md5Weight
{
    int bone;
    double weight;
    Vector3 position;
};

static int md5_read_int() throw (bad_md5)
{
    int token = yylex();

    if (token == MD5_TOKEN_INTEGER)
        return md5_yylval.integer;
    else
        throw bad_md5("Expected an integer");
}

static double md5_read_number() throw(bad_md5)
{
    int token = yylex();

    if (token == MD5_TOKEN_INTEGER)
        return md5_yylval.integer;
    else if (token == MD5_TOKEN_NUMBER)
        return md5_yylval.number;
    else throw bad_md5("Expected a number");
}

static string md5_read_string() throw (bad_md5)
{
    int token = yylex();

    if (token == MD5_TOKEN_STRING)
        return md5_yylval.str;
    else
        throw bad_md5("Expected a string");
}

static void md5_expect(int token) throw (bad_md5)
{
    if (yylex() != token) throw bad_md5("Syntax error");
}

/* Quaternion routines. These all have the last coordinate as the real
 * part, to conform to the md5 documentation convention (md5 itself omits
 * the real part, so has no convention).
 */
static Vector4 quaternion_inverse(const Vector4 &q)
{
    // note: q * q is the vector, not quaternion product
    return Vector4(-q[0], -q[1], -q[2], q[3]) / (q * q);
}

static Vector4 quaternion_product(const Vector4 &p, const Vector4 &q)
{
    return Vector4(p[0] * q[3] + p[3] * q[0] + p[1] * q[2] - p[2] * q[1],
                   p[1] * q[3] + p[3] * q[1] + p[2] * q[0] - p[0] * q[2],
                   p[2] * q[3] + p[3] * q[2] + p[0] * q[1] - p[1] * q[0],
                   p[3] * q[3] - p[0] * q[0] - p[1] * q[1] - p[2] * q[2]);
}

static Matrix3 quaternion_to_matrix(const Vector4 &q)
{
    double xx = q[0] * q[0];
    double yy = q[1] * q[1];
    double zz = q[2] * q[2];
    double ww = q[3] * q[3];

    double xy = q[0] * q[1];
    double xz = q[0] * q[2];
    double xw = q[0] * q[3];
    double yz = q[1] * q[2];
    double yw = q[1] * q[3];
    double zw = q[2] * q[3];

    return Matrix3(2.0 * (xx + ww) - 1.0, 2.0 * (xy + zw)      , 2.0 * (xz - yw),
                   2.0 * (xy - zw)      , 2.0 * (yy + ww) - 1.0, 2.0 * (yz + xw),
                   2.0 * (xz + yw)      , 2.0 * (yz - xw)      , 2.0 * (zz + ww) - 1.0);
}

static Vector4 matrix_to_quaternion(const Matrix3 &m)
{
    /* Algorithm taken from Shoemake: Animating Rotations with Quaternion Curves */
    double w2 = 0.25 * (1.0 + m[0][0] + m[1][1] + m[2][2]);
    if (w2 > 1e-8)
    {
        double w = sqrt(w2);
        return Vector4((m[1][2] - m[2][1]) / (4.0 * w),
                       (m[2][0] - m[0][2]) / (4.0 * w),
                       (m[0][1] - m[1][0]) / (4.0 * w),
                       w);
    }
    else
    {
        double x2 = -0.5 * (m[1][1] + m[2][2]);
        if (x2 > 1e-8)
        {
            double x = sqrt(x2);
            return Vector4(x, m[0][1] / (2.0 * x), m[0][2] / (2.0 * x), 0.0);
        }
        else
        {
            double y2 = 0.5 * (1.0 - m[2][2]);
            if (y2 > 1e-8)
            {
                double y = sqrt(y2);
                return Vector4(0.0, y, m[1][2] / (2 * y), 0.0);
            }
            else
                return Vector4(0.0, 0.0, 1.0, 0.0);
        }
    }
}

/*! See \c Mesh for a description of mesh levels.
 */
void md5_load_mesh(istream &in, Mesh &m, MeshLevel level,
                   vector<string> &comments,
                   const string &path,
                   const vector<string> &args) throw(bad_md5)
{
    int token;
    int version = -1;
    int num_meshes = 1, num_joints = -1;
    vector<MeshRawVertex> vertices;
    vector<vector<int> > faces;
    vector<int> materials;
    vector<string> textures;
    vector<SkeletonBoneInfo> bones;

    md5_input = &in;
    md5_comments = &comments;

    md5_expect(MD5_TOKEN_MD5VERSION);
    version = md5_read_int();
    if (version != 10) throw bad_md5("version mismatch (expected version 10)");
    mesh_clear(m);
    while ((token = yylex()) != 0)
    {
        switch (token)
        {
        case MD5_TOKEN_COMMANDLINE:
            comments.push_back(md5_read_string());
            break;
        case MD5_TOKEN_NUMMESHES:
            num_meshes = md5_read_int();
            break;
        case MD5_TOKEN_NUMJOINTS:
            num_joints = md5_read_int();
            break;
        case MD5_TOKEN_JOINTS:
            {
                if (num_joints < 0) throw bad_md5("numJoints must precede joints");
                md5_expect('{');
                bones.resize(num_joints);
                Vector4 quat[num_joints];
                Vector3 bind[num_joints];
                for (int i = 0; i < num_joints; i++)
                {
                    string name = md5_read_string();
                    bones[i].parent = md5_read_int();
                    md5_expect('(');
                    bind[i][0] = md5_read_number();
                    bind[i][1] = md5_read_number();
                    bind[i][2] = md5_read_number();
                    md5_expect(')');
                    md5_expect('(');
                    quat[i][0] = md5_read_number();
                    quat[i][1] = md5_read_number();
                    quat[i][2] = md5_read_number();
                    md5_expect(')');
                    quat[i][3] = sqrt(max(0.0, 1.0 - quat[i][0] * quat[i][0] - quat[i][1] * quat[i][1] - quat[i][2] * quat[i][2]));
                }

                /* The data in the file are absolute; convert to relative */
                for (int i = 0; i < num_joints; i++)
                {
                    Vector4 quat_rel, par_inv;
                    Vector3 bind_rel;
                    int p = bones[i].parent;
                    if (p == -1)
                    {
                        quat_rel = quat[i];
                        bind_rel = bind[i];
                    }
                    else
                    {
                        par_inv = quaternion_inverse(quat[p]);
                        quat_rel = quaternion_product(quat[i], par_inv);
                        bind_rel = quaternion_to_matrix(par_inv) * (bind[i] - bind[p]);
                    }
                    bones[i].rest_transform = matrix3_translate3_to_matrix4(quaternion_to_matrix(quat_rel), bind_rel);
                    bones[i].tail.zero();
                    bones[i].flags = BONE_FLAG_UNIFORM_SCALE | BONE_FLAG_FIXED_SCALE;
                    bones[i].max_singular = 1.0;
                    bones[i].translate_bounds.centre = bind_rel;
                    bones[i].translate_bounds.radius = 0.0; // can be fixed from animation
                }
                m.skeleton.get_write().set_bones(bones);
                m.skeleton.get_write().fix_tails();
                md5_expect('}');
            }
            break;
        case MD5_TOKEN_MESH:
            if (num_joints < 0) throw bad_md5("joints must precede meshes");
            {
                int num_verts = -1, num_tris = -1, num_weights = -1;
                int index;
                vector<int> weight_start;
                vector<int> weight_count;
                vector<Md5Weight> weights;
                size_t vertex_base = vertices.size();
                size_t face_base = faces.size();
                textures.push_back("");

                md5_expect('{');
                while ((token = yylex()) != '}')
                {
                    switch (token)
                    {
                    case MD5_TOKEN_SHADER:
                        textures.back() = md5_read_string();
                        break;
                    case MD5_TOKEN_NUMVERTS:
                        num_verts = md5_read_int();
                        weight_start.resize(num_verts);
                        weight_count.resize(num_verts);
                        vertices.resize(vertex_base + num_verts);
                        break;
                    case MD5_TOKEN_VERT:
                        index = md5_read_int();
                        md5_expect('(');

                        vertices[index + vertex_base].position = VectorPosition(m.skeleton.get_read());
                        vertices[index + vertex_base].tangents[0] = VectorPosition(m.skeleton.get_read());
                        vertices[index + vertex_base].tangents[1] = VectorPosition(m.skeleton.get_read());
                        vertices[index + vertex_base].texcoords[0] = md5_read_number();
                        vertices[index + vertex_base].texcoords[1] = md5_read_number();
                        md5_expect(')');
                        weight_start[index] = md5_read_int();
                        weight_count[index] = md5_read_int();
                        break;
                    case MD5_TOKEN_NUMTRIS:
                        num_tris = md5_read_int();
                        faces.resize(face_base + num_tris, vector<int>(3));
                        materials.resize(face_base + num_tris, textures.size() - 1);
                        break;
                    case MD5_TOKEN_TRI:
                        index = md5_read_int();
                        faces[index + face_base][2] = vertex_base + md5_read_int();
                        faces[index + face_base][1] = vertex_base + md5_read_int();
                        faces[index + face_base][0] = vertex_base + md5_read_int();
                        materials[index + face_base] = textures.size() - 1;
                        break;
                    case MD5_TOKEN_NUMWEIGHTS:
                        num_weights = md5_read_int();
                        weights.resize(num_weights);
                        break;
                    case MD5_TOKEN_WEIGHT:
                        index = md5_read_int();
                        weights[index].bone = md5_read_int();
                        weights[index].weight = md5_read_number();
                        md5_expect('(');
                        weights[index].position[0] = md5_read_number();
                        weights[index].position[1] = md5_read_number();
                        weights[index].position[2] = md5_read_number();
                        md5_expect(')');
                        break;
                    default:
                        throw bad_md5("Unexpected token");
                    }
                }
                for (int i = 0; i < num_verts; i++)
                {
                    size_t vindex = i + vertex_base;
                    for (int j = weight_start[i]; j < weight_start[i] + weight_count[i]; j++)
                    {
                        int b = m.skeleton.get_read().get_new_index(weights[j].bone);
                        double w = weights[j].weight;
                        vertices[vindex].position[4 * b + 0] = w * weights[j].position[0];
                        vertices[vindex].position[4 * b + 1] = w * weights[j].position[1];
                        vertices[vindex].position[4 * b + 2] = w * weights[j].position[2];
                        vertices[vindex].position[4 * b + 3] = w;
                    }
                }
            }
            break;
        default:
            throw bad_md5("Unexpected token");
        }
    }

    if (!args.empty())
    {
        Mesh anim;

        try
        {
            format_load_mesh_file("md5anim", args[0], anim, MESH_LEVEL_NONE);
        }
        catch (bad_format &e)
        {
            throw bad_md5(string("Could not load animation: ") + e.what());
        }
        m.skeleton.get_write().copy_frames(anim.skeleton.get_read());
    }

    try
    {
        mesh_create(m, vertices, faces, materials, level);
        MeshMaterialIterator mat = m.materials.begin();
        map<string, MeshTextureIterator> texture_cache;
        for (size_t i = 0; i < textures.size(); i++, mat++)
        {
            map<string, MeshTextureIterator>::iterator t = texture_cache.find(textures[i]);
            if (t == texture_cache.end())
            {
                mat->texture() = m.textures.insert(m.textures.end(), MeshTexture());
                try
                {
                    mesh_load_texture(textures[i] + ".tga",
                                      path,
                                      *mat->texture());
                }
                catch (ios::failure &e)
                {
                    g_warning("failed to load texture: %s", e.what());
                    m.textures.erase(mat->texture());
                    mat->texture() = mesh_null_texture;
                }
                catch (bad_format &e)
                {
                    g_warning("failed to load texture: %s", e.what());
                    m.textures.erase(mat->texture());
                    mat->texture() = mesh_null_texture;
                }
                texture_cache[textures[i]] = mat->texture();
            }
            else
                mat->texture() = t->second;
        }
    }
    catch (mesh_nonmanifold &e)
    {
        throw bad_md5(string("non-manifold mesh: ") + e.what());
    }
}

void md5_load_anim(istream &in, Mesh &m, MeshLevel level,
                   vector<string> &comments,
                   const string &path,
                   const vector<string> &anim) throw(bad_md5)
{
    int token;
    int version;
    int num_frames = -1, num_joints = -1, num_components = -1, index;
    double frame_rate = 0.0;
    vector<Vector4> quat_rel;
    vector<Vector3> bind_rel;
    vector<vector<double> > frames;
    vector<SkeletonBoneInfo> bones;
    vector<int> first;
    vector<int> bits;

    md5_input = &in;
    md5_comments = &comments;

    md5_expect(MD5_TOKEN_MD5VERSION);
    version = md5_read_int();
    if (version != 10) throw bad_md5("version mismatch (expected version 10)");
    while ((token = yylex()) != 0)
    {
        switch (token)
        {
        case MD5_TOKEN_COMMANDLINE:
            comments.push_back(md5_read_string());
            break;
        case MD5_TOKEN_FRAMERATE:
            frame_rate = md5_read_number();
            break;
        case MD5_TOKEN_NUMFRAMES:
            num_frames = md5_read_int();
            frames.resize(num_frames);
            break;
        case MD5_TOKEN_NUMJOINTS:
            num_joints = md5_read_int();
            bones.resize(num_joints);
            first.resize(num_joints);
            bits.resize(num_joints);
            break;
        case MD5_TOKEN_NUMANIMATEDCOMPONENTS:
            num_components = md5_read_int();
            break;
        case MD5_TOKEN_HIERARCHY:
            if (num_joints == -1) throw bad_md5("numJoints must precede hierarchy");
            md5_expect('{');
            for (int i = 0; i < num_joints; i++)
            {
                md5_read_string();
                bones[i].parent = md5_read_int();
                bits[i] = md5_read_int();
                first[i] = md5_read_int();
            }
            md5_expect('}');
            break;
        case MD5_TOKEN_BOUNDS:
            if (num_frames == -1) throw bad_md5("numFrames must precede bounds");
            md5_expect('{');
            for (int i = 0; i < num_frames; i++)
            {
                md5_expect('(');
                md5_read_number();
                md5_read_number();
                md5_read_number();
                md5_expect(')');
                md5_expect('(');
                md5_read_number();
                md5_read_number();
                md5_read_number();
                md5_expect(')');
            }
            md5_expect('}');
            break;
        case MD5_TOKEN_BASEFRAME:
            if (num_joints == -1) throw bad_md5("numJoints must precede baseframe");
            {
                md5_expect('{');
                quat_rel.resize(num_joints);
                bind_rel.resize(num_joints);
                for (int i = 0; i < num_joints; i++)
                {
                    md5_expect('(');
                    bind_rel[i][0] = md5_read_number();
                    bind_rel[i][1] = md5_read_number();
                    bind_rel[i][2] = md5_read_number();
                    md5_expect(')');
                    md5_expect('(');
                    quat_rel[i][0] = md5_read_number();
                    quat_rel[i][1] = md5_read_number();
                    quat_rel[i][2] = md5_read_number();
                    md5_expect(')');
                    quat_rel[i][3] = sqrt(max(0.0, 1.0 - quat_rel[i][0] * quat_rel[i][0] - quat_rel[i][1] * quat_rel[i][1] - quat_rel[i][2] * quat_rel[i][2]));
                }

                for (int i = 0; i < num_joints; i++)
                {
                    bones[i].rest_transform = matrix3_translate3_to_matrix4(quaternion_to_matrix(quat_rel[i]),
                                                                            bind_rel[i]);
                    bones[i].flags = BONE_FLAG_UNIFORM_SCALE | BONE_FLAG_FIXED_SCALE;
                    bones[i].max_singular = 1.0;
                    bones[i].translate_bounds.centre = bind_rel[i];
                    bones[i].translate_bounds.radius = 0.0;
                }
                m.skeleton.get_write().set_bones(bones);
                md5_expect('}');
            }
            break;
        case MD5_TOKEN_FRAME:
            if (num_components == -1) throw bad_md5("numAnimatedComponents must precede frames");
            index = md5_read_int();
            md5_expect('{');
            frames[index].resize(num_components);
            for (int i = 0; i < num_components; i++)
                frames[index][i] = md5_read_number();
            md5_expect('}');
            break;
        default:
            throw bad_md5("Unexpected token");
        }
    }

    for (int i = 0; i < num_frames; i++)
    {
        vector<Matrix4> frame(num_joints);

        for (int j = 0; j < num_joints; j++)
        {
            int p = first[j];
            if (bits[j] & 0x01) bind_rel[j][0] = frames[i][p++];
            if (bits[j] & 0x02) bind_rel[j][1] = frames[i][p++];
            if (bits[j] & 0x04) bind_rel[j][2] = frames[i][p++];
            if (bits[j] & 0x08) quat_rel[j][0] = frames[i][p++];
            if (bits[j] & 0x10) quat_rel[j][1] = frames[i][p++];
            if (bits[j] & 0x20) quat_rel[j][2] = frames[i][p++];
            quat_rel[j][3] = sqrt(max(0.0, 1.0 - quat_rel[j][0] * quat_rel[j][0] - quat_rel[j][1] * quat_rel[j][1] - quat_rel[j][2] * quat_rel[j][2]));
        }
        for (int j = 0; j < num_joints; j++)
        {
            frame[m.skeleton.get_read().get_new_index(j)]
                = matrix3_translate3_to_matrix4(quaternion_to_matrix(quat_rel[j]),
                                                bind_rel[j]);
        }
        m.skeleton.get_write().add_frame(frame);
    }
    m.skeleton.get_write().set_frame_rate(frame_rate);
}

static string md5_vector_string(const Vector3 &v)
{
    ostringstream o;

    o << '(' << v[0] << ' ' << v[1] << ' ' << v[2] << ')';
    return o.str();
}

void md5_save_mesh(ostream &out, const Mesh &m,
                   const vector<string> &comments,
                   const string &path)
{
    const Skeleton &skeleton = m.skeleton.get_read();

    out << "MD5Version 10\n";
    for (vector<string>::const_iterator i = comments.begin(); i != comments.end(); i++)
        out << "// " << *i << "\n";
    out << "numJoints " << skeleton.size() << "\n"
        << "numMeshes " << m.materials.size() << "\n"
        << "joints\n"
        << "{\n";

    int index = 0;
    for (Skeleton::const_iterator i = skeleton.begin(); i != skeleton.end(); i++)
    {
        Matrix4 G = skeleton.get_model_transform() * skeleton.get_absolute(index);
        Vector4 q = matrix_to_quaternion(matrix4_to_matrix3(G));
        /* The implementation of matrix_to_quaternion ensures w >= 0 */
        out << "    \"bone" << index << "\" " << i->parent << " "
            << md5_vector_string(matrix4_to_translate3(G)) << " "
            << md5_vector_string(vector4_to_vector3(q)) << "\n";
        index++;
    }
    out << "}\n\n";

    for (MeshMaterialConstIterator mat = m.materials.begin(); mat != m.materials.end(); mat++)
    {
        out << "mesh\n"
            << "{\n"
            << "    shader \"\"\n\n";

        /* Collect the relevant vertices and faces. Note that we cannot
         * use first_edge, as this may not be a complete mesh, or even that
         * it is manifold.
         */
        vector<MeshFaceConstIterator> faces;
        vector<MeshVertexConstIterator> vertices;
        id_map<MeshVertexConstIterator, size_t>::type indices;
        int weight_offset = 0;

        for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
            if (f->material() == mat)
            {
                faces.push_back(f);
                MeshHalfEdgeConstIterator e = f->edges();
                for (int i = 0; i < 3; i++, e = e->next())
                {
                    MeshVertexConstIterator v = e->vertex();
                    if (!indices.count(v))
                    {
                        indices[v] = vertices.size();
                        vertices.push_back(v);
                    }
                }
            }

        /* Write the vertices */
        out << "    numverts " << vertices.size() << "\n";
        for (size_t i = 0; i < vertices.size(); i++)
        {
            MeshVertexConstIterator v = vertices[i];
            int d = distance(v->position()->position.begin(), v->position()->position.end());
            out << "    vert " << i << " "
                << "( " << v->data.texcoords[0] << " "
                << v->data.texcoords[1] << " ) "
                << weight_offset << " " << d << "\n";
            weight_offset += d;
        }

        /* Write the triangles */
        out << "    numtris " << faces.size() << "\n";
        for (size_t i = 0; i < faces.size(); i++)
        {
            MeshFaceConstIterator f = faces[i];
            MeshHalfEdgeConstIterator e = f->edges();
            out << "    tri " << i;
            for (int j = 0; j < 3; j++, e = e->prev())
                out << " " << indices[e->vertex()];
            out << "\n";
        }

        /* Write the weights */
        out << "    numweights " << weight_offset << "\n";
        weight_offset = 0;
        int w = 0;
        for (size_t i = 0; i < vertices.size(); i++)
        {
            MeshVertexConstIterator v = vertices[i];
            const VectorPosition &p = v->position()->position;
            for (VectorPosition::const_iterator j = p.begin(); j != p.end(); j++)
            {
                out << "    weight " << w << " " << j->index / 4 << " "
                    << j->values[3] << " (";
                for (int k = 0; k < 3; k++)
                    out << " " << j->values[k] / j->values[3];
                out << " )\n";
                w++;
            }
        }

        /* Mesh footer */
        out << "}\n";
    }
}

/*! @} */
