/*! \addtogroup format
 * @{
 */

/*! \file
 * \brief Support for Stanford's PLY format
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "ply.h"
#include "common/maputils.h"
#include <string>
#include <cstring>
#include <iostream>
#include <sstream>
#include <stdexcept>
#include <cstddef>
#include <list>
#include <vector>
#include <algorithm>
#include <iterator>
#include <glib.h>

using namespace std;

enum PlyType
{
    PLY_TYPE_NUMERIC,
    PLY_TYPE_LIST,
    PLY_TYPE_STRING
};

struct PlyProperty
{
    string name;
    PlyType type;
    FormatNumberType list_type;
    FormatNumberType number_type;
};

struct PlyElement
{
    string name;
    size_t count;
    list<PlyProperty> properties;
};

static FormatNumberType ply_string_to_type(const string &t) throw(bad_ply)
{
    if (t == "int8" || t == "char") return FORMAT_TYPE_INT8;
    else if (t == "uint8" || t == "uchar") return FORMAT_TYPE_UINT8;
    else if (t == "int16") return FORMAT_TYPE_INT16;
    else if (t == "uint16") return FORMAT_TYPE_UINT16;
    else if (t == "int32" || t == "int") return FORMAT_TYPE_INT32;
    else if (t == "uint32" || t == "uint") return FORMAT_TYPE_UINT32;
    else if (t == "float32" || t == "float") return FORMAT_TYPE_FLOAT32;
    else if (t == "float64") return FORMAT_TYPE_FLOAT64;
    else throw bad_ply("Unknown type " + t);
}

static double ply_read_field(istream &in,
                             PlyFormat format,
                             FormatNumberType type)
{
    float answer;
    if (format == PLY_FORMAT_ASCII)
    {
        in >> answer;
        if (!in) throw bad_ply("error while reading file");
        return answer;
    }
    else
    {
        try
        {
            bool big_endian = format == PLY_FORMAT_BIG_ENDIAN;
            return format_read_binary(in, big_endian, type);
        }
        catch (bad_format &e)
        {
            throw bad_ply(e.what());
        }
    }
}

static string ply_read_string(istream &in,
                              PlyFormat format)
{
    string out = "";
    if (format == PLY_FORMAT_ASCII)
    {
        char cur;

        // search for opening "
        do
        {
            in >> cur;
        } while (in && cur != '"');
        if (!in) throw bad_ply("failed to read string");
        in >> cur;
        while (in && cur != '"')
        {
            out += cur;
            in >> cur;
        }
        if (!in) throw bad_ply("failed to read string");
        return out;
    }
    else
    {
        try
        {
            guint32 len;
            bool big_endian = format == PLY_FORMAT_BIG_ENDIAN;
            len = (guint32) format_read_binary(in, big_endian, FORMAT_TYPE_UINT32);
            char buffer[len];
            in.read(buffer, len);
            if ((guint32) in.gcount() != len)
                throw bad_ply("short read while loading string");
            if (!in)
                throw bad_ply("failed to read string");
            if (buffer[len - 1] != '\0')
                throw bad_ply("binary string not NULL-terminated");
            return buffer;
        }
        catch (bad_format &e)
        {
            throw bad_ply(e.what());
        }
    }
}

static void ply_write_field(ostream &out,
                            PlyFormat format,
                            FormatNumberType type,
                            double value)
{
    if (format == PLY_FORMAT_ASCII)
    {
        if (type == FORMAT_TYPE_FLOAT32 || type == FORMAT_TYPE_FLOAT64)
            out << value;
        else if (type == FORMAT_TYPE_UINT32) // might lose precision
            out << (unsigned long) value;
        else
            out << (long) value;
        out << " ";
    }
    else
    {
        try
        {
            bool big_endian = (format == PLY_FORMAT_BIG_ENDIAN);
            format_write_binary(out, big_endian, type, value);
        }
        catch (bad_format &e)
        {
            throw bad_ply(e.what());
        }
    }
}

static void ply_write_string(ostream &out,
                             PlyFormat format,
                             const string &value)
{
    if (format == PLY_FORMAT_ASCII)
        out << "\"" << value << "\" ";
    else
    {
        try
        {
            bool big_endian = (format == PLY_FORMAT_BIG_ENDIAN);
            /* Note: the binary form of strings is not specified, and
             * this was determined from the source of Georgia Tech's
             * PLY utils. We're guessing at 32 bits, since the people
             * in Georgia seem to be trying to make this a
             * platform-dependent format by using sizeof(int).
             */
            format_write_binary(out, big_endian, FORMAT_TYPE_UINT32,
                                value.length() + 1);
            out.write(value.c_str(), value.length() + 1);
        }
        catch (bad_format &e)
        {
            throw bad_ply(e.what());
        }
    }
    if (!out) throw bad_ply("I/O error writing file");
}

static inline void ply_write_newline(ostream &o, PlyFormat format)
{
    if (format == PLY_FORMAT_ASCII)
        o << "\n";
}

void ply_load_mesh(istream &in, Mesh &m, MeshLevel level,
                   vector<string> &comments,
                   const string &path,
                   const vector<string> &args) throw(bad_ply)
{
    PlyFormat format = PLY_FORMAT_ASCII;
    list<PlyElement> elements;
    vector<MeshRawVertex> vertices;
    vector<vector<int> > faces;
    vector<int> materials;
    vector<string> textures, normals;

    const Skeleton &skeleton = m.skeleton.get_read();

    try
    {
        try
        {
            string line;
            getline(in, line);
            if (line != "ply") throw bad_ply("PLY header missing");

            // read the header
            while (1)
            {
                getline(in, line);
                istringstream tokens(line);
                string token;
                tokens >> token;
                if (token == "end_header") break;
                else if (token == "format")
                {
                    string fmt;
                    string version;
                    tokens >> fmt >> version;
                    if (version != "1.0")
                        throw bad_ply("version mismatch (expected 1.0, got " + version + ")");
                    if (fmt == "ascii")
                        format = PLY_FORMAT_ASCII;
                    else if (fmt == "binary_big_endian")
                        format = PLY_FORMAT_BIG_ENDIAN;
                    else if (fmt == "binary_little_endian")
                        format = PLY_FORMAT_LITTLE_ENDIAN;
                    else
                        throw bad_ply("unknown format " + format);
                }
                else if (token == "element")
                {
                    PlyElement e;
                    tokens >> e.name >> e.count;
                    elements.push_back(e);
                }
                else if (token == "property")
                {
                    PlyProperty p;
                    string type;
                    tokens >> type;
                    if (type == "list")
                    {
                        p.type = PLY_TYPE_LIST;
                        tokens >> type;
                        p.list_type = ply_string_to_type(type);
                        tokens >> type;
                        p.number_type = ply_string_to_type(type);
                    }
                    else if (type == "string")
                        p.type = PLY_TYPE_STRING;
                    else
                    {
                        p.type = PLY_TYPE_NUMERIC;
                        p.number_type = ply_string_to_type(type);
                    }
                    tokens >> p.name;
                    if (elements.empty())
                        throw bad_ply("property " + p.name + " preceeds all elements");
                    elements.back().properties.push_back(p);
                }
                else if (token == "comment")
                {
                    string comment;
                    tokens >> ws;
                    getline(tokens, comment);
                    while (!comment.empty() && isspace(comment[comment.length() - 1]))
                        comment.erase(comment.length() - 1);
                    comments.push_back(comment);
                }
                else if (token == "obj_info")
                {
                    /* Do nothing */
                }
                else
                    throw bad_ply("unknown header line " + line);
            }

            // read the body, extracting vertices and faces
            for (list<PlyElement>::iterator e = elements.begin();
                 e != elements.end(); e++)
            {
                for (size_t i = 0; i < e->count; i++)
                {
                    MeshRawVertex vertex;
                    Vector3 normal(0.0, 0.0, 0.0);
                    vertex.position = skeleton.ssd_vector();
                    vertex.position[3] = 1.0;
                    vector<int> face;
                    int mat = 0;
                    for (list<PlyProperty>::iterator p = e->properties.begin();
                         p != e->properties.end(); p++)
                    {
                        if (p->type == PLY_TYPE_NUMERIC)
                        {
                            double val = ply_read_field(in, format, p->number_type);
                            if (e->name == "vertex")
                            {
                                if (p->name == "x")
                                    vertex.position[0] = val;
                                else if (p->name == "y")
                                    vertex.position[1] = val;
                                else if (p->name == "z")
                                    vertex.position[2] = val;
                                else if (p->name == "s")
                                    vertex.texcoords[0] = val;
                                else if (p->name == "t")
                                    vertex.texcoords[1] = val;
                                else if (p->name == "nx")
                                    normal[0] = val;
                                else if (p->name == "ny")
                                    normal[1] = val;
                                else if (p->name == "nz")
                                    normal[2] = val;
                            }
                            else if (e->name == "face" && p->name == "material_index")
                                mat = (int) val;
                        }
                        else if (p->type == PLY_TYPE_LIST)
                        {
                            int count = (int) ply_read_field(in, format, p->list_type);
                            for (int j = 0; j < count; j++)
                            {
                                double val = ply_read_field(in, format, p->number_type);
                                if (e->name == "face" && p->name == "vertex_indices")
                                    face.push_back((int) val);
                            }
                        }
                        else
                        {
                            string str = ply_read_string(in, format);
                            if (e->name == "material")
                            {
                                if (p->name == "texture_map")
                                    textures.push_back(str);
                                else if (p->name == "normal_map")
                                    normals.push_back(str);
                            }
                        }
                    }
                    if (e->name == "vertex")
                    {
                        vertex.tangents[0] = VectorPosition(skeleton);
                        vertex.tangents[1] = VectorPosition(skeleton);
                        if (length(normal) > 1e-9)
                        {
                            Vector3 t[2];
                            complete_basis(normal, t[0], t[1]);
                            for (int j = 0; j < 2; j++)
                                for (int k = 0; k < 3; k++)
                                    vertex.tangents[j][k] = t[j][k];
                        }
                        vertices.push_back(vertex);
                    }
                    else if (e->name == "face" && face.size() >= 3)
                    {
                        faces.push_back(face);
                        materials.push_back(mat);
                    }
                }
            }
            mesh_create(m, vertices, faces, materials, level);

            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO,
                  "loaded %d positions, %d vertices and %d faces",
                  (int) m.positions.real_size(),
                  (int) m.vertices.real_size(),
                  (int) m.faces.real_size());
        }
        catch (mesh_nonmanifold &e)
        {
            throw bad_ply(string("non-manifold mesh: ") + e.what());
        }
    }
    catch (bad_ply &e)
    {
        in.setstate(ios::failbit);
        throw;
    }
}

void ply_save_mesh(ostream &out, const Mesh &m, PlyFormat format,
                   const vector<string> &comments,
                   const string &path) throw(bad_format)
{
    size_t true_faces = m.faces.real_size();

    out << "ply\n";

    switch (format)
    {
    case PLY_FORMAT_ASCII:
        out << "format ascii 1.0\n";
        break;
    case PLY_FORMAT_LITTLE_ENDIAN:
        out << "format binary_little_endian 1.0\n";
        break;
    case PLY_FORMAT_BIG_ENDIAN:
        out << "format binary_big_endian 1.0\n";
        break;
    }
    for (vector<string>::const_iterator i = comments.begin(); i != comments.end(); i++)
        out << "comment " << *i << "\n";
    out << "element vertex " << m.vertices.real_size() << "\n"
        << "property float32 x\n"
        << "property float32 y\n"
        << "property float32 z\n"
        << "property float32 nx\n"
        << "property float32 ny\n"
        << "property float32 nz\n"
        << "property float32 s\n"
        << "property float32 t\n";
    out << "element face " << true_faces << "\n"
        << "property list uint8 int32 vertex_indices\n"
        << "property int32 material_index\n"
        << "end_header\n";

    id_map<MeshVertexConstIterator, size_t>::type indices;
    fast_map<int, Vector3>::type patch_colours;
    size_t index = 0;
    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
    {
        Vector3 pos = m.skeleton.get_read().ssd_to_3d(v->position()->position);
        Vector3 normal = m.skeleton.get_read().ssd_to_normal(v->data.tangent_plane);
        ply_write_field(out, format, FORMAT_TYPE_FLOAT32, pos[0]);
        ply_write_field(out, format, FORMAT_TYPE_FLOAT32, pos[1]);
        ply_write_field(out, format, FORMAT_TYPE_FLOAT32, pos[2]);
        ply_write_field(out, format, FORMAT_TYPE_FLOAT32, normal[0]);
        ply_write_field(out, format, FORMAT_TYPE_FLOAT32, normal[1]);
        ply_write_field(out, format, FORMAT_TYPE_FLOAT32, normal[2]);
        ply_write_field(out, format, FORMAT_TYPE_FLOAT32, v->data.texcoords[0]);
        ply_write_field(out, format, FORMAT_TYPE_FLOAT32, v->data.texcoords[1]);

        ply_write_newline(out, format);
        indices[v] = index++;
    }

    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
    {
        ply_write_field(out, format, FORMAT_TYPE_UINT8, 3);
        MeshHalfEdgeConstIterator e = f->edges();
        do
        {
            ply_write_field(out, format, FORMAT_TYPE_INT32, indices[e->vertex()]);
            e = e->next();
        } while (e != f->edges());
        int mat = distance(m.materials.begin(), f->material());
        ply_write_field(out, format, FORMAT_TYPE_INT32, mat);
        ply_write_newline(out, format);
    }

    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO,
          "wrote %d positions, %d vertices and %d faces",
          (int) m.positions.real_size(),
          (int) m.vertices.real_size(),
          (int) true_faces);
}

/*! @} */
