/*! \addtogroup format
 * @{
 */

/*! \file
 * \brief Support for Stanford's PLY format
 *
 * The PLY format is used by Stanford and Georgia Tech for large models.
 * It is a very generic format, and only naming conventions allow it to
 * work at all. It consists of an ASCII header describing the fields
 * available in the type, following by either ASCII or binary data. The
 * header describes <em>elements</em>, which are built up of <em>properties</em>.
 * Each property is either a simple scalar or a variable-length list of
 * scalars (it's not clear whether lists of lists are legal, but they are
 * not supported by this code). The following properties and elements (amongst
 * others) are standard.
 *
 * - \c vertex
 *   - \c x, \c y, \c z: position
 *   - \c nx, \c ny, \c nz: normal
 *   - \c s, \c t: texture coordinates
 *   - \c red, \c green, \c blue: colour
 * - \c face
 *   - \c vertex_indices: a list of indices into the vertex list
 *
 * \see http://www.cc.gatech.edu/projects/large_models/ply.html
 */

#ifndef ALOD_FORMATS_PLY_H
#define ALOD_FORMATS_PLY_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "formats/formats.h"
#include <iosfwd>
#include <vector>
#include <string>

enum PlyFormat
{
    PLY_FORMAT_ASCII,
    PLY_FORMAT_LITTLE_ENDIAN,
    PLY_FORMAT_BIG_ENDIAN
};

class bad_ply : public bad_format
{
public:
    bad_ply(const std::string &msg) : bad_format(msg) {}
};

//! Loads a mesh from a Stanford PLY file.
void ply_load_mesh(std::istream &in, Mesh &m, MeshLevel level,
                   std::vector<std::string> &comments,
                   const std::string &path,
                   const std::vector<std::string> &args) throw(bad_ply);

//! Writes a mesh back to Stanford PLY format
void ply_save_mesh(std::ostream &out, const Mesh &m, PlyFormat format,
                   const std::vector<std::string> &comments,
                   const std::string &path) throw(bad_format);

#endif /* ALOD_FORMATS_PLY_H */

/*! @} */
