/*! \addtogroup geom
 * @{
 */

/*! \file
 * \brief Bounding volumes.
 */

#ifndef GEOM_BOUNDING_H
#define GEOM_BOUNDING_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "coords.h"

struct BoundingSphere
{
    Vector3 centre;
    double radius;

    BoundingSphere() {}
    BoundingSphere(const Vector3 &c, double r) : centre(c), radius(r) {}
};

//! Sphere embedded in a hyper-plane of fixed w
struct BoundingSphereW
{
    Vector4 centre;
    double radius;

    BoundingSphereW() {}
    BoundingSphereW(const Vector4 &c, double r) : centre(c), radius(r) {}
};

/* BoundingSphere */

static inline BoundingSphere operator +(const BoundingSphere &a,
                                        const BoundingSphere &b)
{
    return BoundingSphere(a.centre + b.centre,
                          a.radius + b.radius);
}

static inline BoundingSphere operator +(const BoundingSphere &a,
                                        const Vector3 &b)
{
    return BoundingSphere(a.centre + b, a.radius);
}

static inline BoundingSphere operator +(const Vector3 &a,
                                        const BoundingSphere &b)
{
    return BoundingSphere(a + b.centre, b.radius);
}

static inline BoundingSphere &operator +=(BoundingSphere &a,
                                          const Vector3 &b)
{
    a.centre += b;
    return a;
}

static inline BoundingSphere &operator +=(BoundingSphere &a,
                                          const BoundingSphere &b)
{
    a.centre += b.centre;
    a.radius += b.radius;
    return a;
}

static inline BoundingSphere operator *(const BoundingSphere &a,
                                        double s)
{
    return BoundingSphere(Vector3(s * a.centre[0],
                                  s * a.centre[1],
                                  s * a.centre[2]),
                          fabs(s) * a.radius);
}

static inline BoundingSphere operator *(double s, const BoundingSphere &b)
{
    return b * s;
}

static inline BoundingSphere &operator *=(BoundingSphere &a, double s)
{
    a.centre *= s;
    a.radius *= fabs(s);
    return a;
}

static inline BoundingSphere operator *(const Matrix3 &m, const BoundingSphere &a)
{
    double max_singular = matrix_norm2(m);
    return BoundingSphere(m * a.centre, a.radius * max_singular);
}

// must be an affine transform
static inline BoundingSphere operator *(const Matrix4 &m, const BoundingSphere &a)
{
    double max_singular = matrix_norm2(matrix4_to_matrix3(m));
    return BoundingSphere(vector4_to_vector3(m * vector3_to_vector4(a.centre, 1.0)),
                          a.radius * max_singular);
}

static inline double inf_norm(const BoundingSphere &s)
{
    return length(s.centre) + s.radius;
}

/* BoundingSphereW */

static inline BoundingSphereW operator +(const BoundingSphereW &a,
                                         const BoundingSphereW &b)
{
    return BoundingSphereW(a.centre + b.centre,
                           a.radius + b.radius);
}

/* This is interpreted as if a translate matrix was applied to a,
 * whose translation vector lies inside b.
 */
static inline BoundingSphereW operator *(const BoundingSphereW &a,
                                         const BoundingSphere &b)
{
    return BoundingSphereW(a.centre
                           + fabs(a.centre[3]) * vector3_to_vector4(b.centre, 0.0),
                           a.radius + fabs(a.centre[3]) * b.radius);
}

static inline BoundingSphereW operator +(const BoundingSphereW &a,
                                         const Vector4 &b)
{
    return BoundingSphereW(a.centre + b, a.radius);
}

static inline BoundingSphereW operator +(const Vector4 &a,
                                         const BoundingSphereW &b)
{
    return BoundingSphereW(a + b.centre, b.radius);
}

static inline BoundingSphereW &operator +=(BoundingSphereW &a,
                                           const Vector4 &b)
{
    a.centre += b;
    return a;
}

static inline BoundingSphereW &operator +=(BoundingSphereW &a,
                                           const BoundingSphereW &b)
{
    a.centre += b.centre;
    a.radius += b.radius;
    return a;
}

static inline BoundingSphereW &operator *=(BoundingSphereW &a,
                                           const BoundingSphere &b)
{
    a.centre += a.centre[3] * vector3_to_vector4(b.centre, 0.0);
    a.radius += fabs(a.centre[3]) * b.radius;
    return a;
}

static inline BoundingSphereW operator *(const BoundingSphereW &a,
                                         double s)
{
    return BoundingSphereW(Vector4(s * a.centre[0],
                                   s * a.centre[1],
                                   s * a.centre[2],
                                   a.centre[3]),
                           fabs(s) * a.radius);
}

static inline BoundingSphereW operator *(double s, const BoundingSphereW &b)
{
    return b * s;
}

static inline BoundingSphereW &operator *=(BoundingSphereW &a, double s)
{
    a.centre[0] *= s;
    a.centre[1] *= s;
    a.centre[2] *= s;
    a.radius *= fabs(s);
    return a;
}

// must be an affine transform
static inline BoundingSphereW operator *(const Matrix4 &m, const BoundingSphereW &a)
{
    double max_singular = matrix_norm2(matrix4_to_matrix3(m));
    return BoundingSphereW(m * a.centre, a.radius * max_singular);
}

static inline double inf_norm(const BoundingSphereW &s)
{
    return length(vector4_to_vector3(s.centre)) + s.radius;
}

#endif /* GEOM_BOUNDING_H */

/*! @} */
