/*! \addtogroup geom
 * @{
 */

/*! \file
 * \brief Simple vector code
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "coords.h"
#include "common/mesch.h"
#include <iostream>
#include <cmath>
#include <algorithm>
#include <glib.h>

using namespace std;

double cross(const Vector2 &a, const Vector2 &b)
{
    return a[0] * b[1] - a[1] * b[0];
}

Vector3 cross(const Vector3 &a, const Vector3 &b)
{
    return Vector3(a[1] * b[2] - a[2] * b[1],
                   a[2] * b[0] - a[0] * b[2],
                   a[0] * b[1] - a[1] * b[0]);
}

static inline double inf_norm(const Vector2 &v)
{
    return max(fabs(v[0]), fabs(v[1]));
}

static inline double inf_norm(const Vector3 &v)
{
    return max(max(fabs(v[0]), fabs(v[1])), fabs(v[2]));
}

static inline double inf_norm(const Vector4 &v)
{
    return max(max(fabs(v[0]), fabs(v[1])), max(fabs(v[2]), fabs(v[3])));
}

/* These functions seem inefficient but are very stable. Dividing by
 * the infinity norm guarantees that there will be no overflow
 * (unless the length itself overflows, and even then the normal is
 * correct), that length is always >= every element, and that normals
 * are correct even for very short vectors.
 */

double length(const Vector2 &v)
{
    double s = inf_norm(v);
    if (s == 0.0) return 0.0;
    double si = 1.0 / s;
    const Vector2 w(v[0] * si, v[1] * si);
    return length_fast(w) * s;
}

double length(const Vector3 &v)
{
    double s = inf_norm(v);
    if (s == 0.0) return 0.0;
    double si = 1.0 / s;
    const Vector3 w(v[0] * si, v[1] * si, v[2] * si);
    return length_fast(w) * s;
}

double length(const Vector4 &v)
{
    double s = inf_norm(v);
    if (s == 0.0) return 0.0;
    double si = 1.0 / s;
    const Vector4 w(v[0] * si, v[1] * si, v[2] * si, v[3] * si);
    return length_fast(w) * s;
}

Vector2 normalise(const Vector2 &v)
{
    double s = inf_norm(v);
    g_return_val_if_fail(s > 0.0, Vector2(0.0, 0.0));
    double si = 1.0 / s;
    const Vector2 w(v[0] * si, v[1] * si);
    return w / length_fast(w);
}

Vector3 normalise(const Vector3 &v)
{
    double s = inf_norm(v);
    g_return_val_if_fail(s > 0.0, Vector3(0.0, 0.0, 0.0));
    double si = 1.0 / s;
    const Vector3 w(v[0] * si, v[1] * si, v[2] * si);
    return w / length_fast(w);
}

Vector4 normalise(const Vector4 &v)
{
    double s = inf_norm(v);
    g_return_val_if_fail(s > 0.0, Vector4(0.0, 0.0, 0.0, 0.0));
    double si = 1.0 / s;
    const Vector4 w(v[0] * si, v[1] * si, v[2] * si, v[3] * si);
    return w / length_fast(w);
}

/*! If \a open is true, the intersection occurs only if the
 * lines are not parallel and if their interiors intersect.
 * Otherwise the intersection occurs if any point (including
 * the endpoints) is common.
 */
bool intersects(const Vector2 &p1, const Vector2 &p2,
                const Vector2 &q1, const Vector2 &q2,
                bool open)
{
    if (min(p1[0], p2[0]) > max(q1[0], q2[0])) return false;
    if (min(q1[0], q2[0]) > max(p1[0], p2[0])) return false;
    if (min(p1[1], p2[1]) > max(q1[1], q2[1])) return false;
    if (min(q1[1], q2[1]) > max(p1[1], p2[1])) return false;

    double s1 = cross(q1 - p1, p2 - p1);
    double s2 = cross(q2 - p1, p2 - p1);
    double s3 = cross(p1 - q1, q2 - q1);
    double s4 = cross(p2 - q2, q2 - q1);
    if (!s1 || !s2 || !s3 || !s4) return !open;
    return (((s1 < 0) ^ (s2 < 0))
            && ((s3 < 0) ^ (s4 < 0)));
}

/*! If the lines are parallel, then both \a s and \a t will be
 * non-finite.
 */
Vector2 intersection(const Vector2 &p1, const Vector2 &p2,
                     const Vector2 &q1, const Vector2 &q2,
                     double &s, double &t)
{
    Vector2 np(p1[1] - p2[1], p2[0] - p1[0]);
    Vector2 nq(q1[1] - q2[1], q2[0] - q1[0]);
    double dp = p1 * np;
    double dq = q1 * nq;

    double det = np[0] * nq[1] - np[1] * nq[0];
    Vector2 ans((nq[1] * dp - np[1] * dq) / det,
                (np[0] * dq - nq[0] * dp) / det);
    s = ((ans - p1) * (p2 - p1)) / (np * np);
    t = ((ans - q1) * (q2 - q1)) / (nq * nq);
    return ans;
}

ostream &operator <<(ostream &o, const Vector2 &v)
{
    int old = o.precision(6);
    o << "(" << v[0] << ", " << v[1] << ")";
    o.precision(old);
    return o;
}

ostream &operator <<(ostream &o, const Vector3 &v)
{
    int old = o.precision(6);
    o << "(" << v[0] << ", " << v[1] << ", " << v[2] << ")";
    o.precision(old);
    return o;
}

ostream &operator <<(ostream &o, const Vector4 &v)
{
    int old = o.precision(6);
    o << "(" << v[0] << ", " << v[1] << ", " << v[2] << ", " << v[3] << ")";
    o.precision(old);
    return o;
}

void complete_basis(const Vector3 &x, Vector3 &y, Vector3 &z)
{
    if (fabs(x[0]) < fabs(x[1]) && fabs(x[0]) < fabs(x[2]))
        z.assign(1.0, 0.0, 0.0);
    else if (fabs(x[1]) < fabs(x[2]))
        z.assign(0.0, 1.0, 0.0);
    else z.assign(0.0, 0.0, 1.0);
    y = cross(z, x);
    z = cross(x, y);
    y = normalise(y);
    z = normalise(z);
}

double matrix_norm2(const Matrix2 &m)
{
    static MAT *m2 = MNULL;
    static VEC *sing = VNULL;

    m2 = m_resize(m2, 2, 2);
    for (int i = 0; i < 2; i++)
        for (int j = 0; j < 2; j++)
            m_set_val(m2, i, j, m[i][j]);
    sing = svd(m2, MNULL, MNULL, sing);
    return v_entry(sing, 0);
}

double matrix_norm2(const Matrix3 &m)
{
    /* First check for pure axial shear, the most common case being no shear */
    double maxl = max(max(m[0] * m[0], m[1] * m[1]), m[2] * m[2]);
    if (fabs(m[0] * m[1]) < 1e-6 * maxl
        && fabs(m[1] * m[2]) < 1e-6 * maxl
        && fabs(m[2] * m[0]) < 1e-6 * maxl)
        return sqrt(maxl);

    static MAT *m2 = MNULL;
    static VEC *sing = VNULL;

    m2 = m_resize(m2, 3, 3);
    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
            m_set_val(m2, i, j, m[i][j]);
    sing = svd(m2, MNULL, MNULL, sing);
    return v_entry(sing, 0);
}

double matrix_norm2(const Matrix4 &m)
{
    static MAT *m2 = MNULL;
    static VEC *sing = VNULL;

    m2 = m_resize(m2, 4, 4);
    for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++)
            m_set_val(m2, i, j, m[i][j]);
    sing = svd(m2, MNULL, MNULL, sing);
    return v_entry(sing, 0);
}

double determinant(const Matrix2 &m)
{
    return m[0][0] * m[1][1] - m[0][1] * m[1][0];
}

double determinant(const Matrix3 &m)
{
    return m[0][0] * (m[1][1] * m[2][2] - m[1][2] * m[2][1])
        +  m[0][1] * (m[1][2] * m[2][0] - m[1][0] * m[2][2])
        +  m[0][2] * (m[1][0] * m[2][1] - m[1][1] * m[2][0]);
}

double determinant(const Matrix4 &m)
{
    return m[0][0] * (  m[1][1] * (m[2][2] * m[3][3] - m[2][3] * m[3][2])
                      + m[1][2] * (m[2][3] * m[3][1] - m[2][1] * m[3][3])
                      + m[1][3] * (m[2][1] * m[3][2] - m[2][2] * m[3][1]))
        -  m[0][1] * (  m[1][2] * (m[2][3] * m[3][0] - m[2][0] * m[3][3])
                      + m[1][3] * (m[2][0] * m[3][2] - m[2][2] * m[3][0])
                      + m[1][0] * (m[2][2] * m[3][3] - m[2][3] * m[3][2]))
        +  m[0][2] * (  m[1][3] * (m[2][0] * m[3][1] - m[2][1] * m[3][0])
                      + m[1][0] * (m[2][1] * m[3][3] - m[2][3] * m[3][1])
                      + m[1][1] * (m[2][3] * m[3][0] - m[2][0] * m[3][3]))
        -  m[0][3] * (  m[1][0] * (m[2][1] * m[3][2] - m[2][2] * m[3][1])
                      + m[1][1] * (m[2][2] * m[3][0] - m[2][0] * m[3][2])
                      + m[1][2] * (m[2][0] * m[3][1] - m[2][1] * m[3][0]));
}

Matrix2 inverse(const Matrix2 &m)
{
    return Matrix2( m[1][1], -m[0][1],
                   -m[1][0],  m[0][0]) * (1.0 / determinant(m));
}

Matrix3 inverse(const Matrix3 &m)
{
    Matrix3 adj;

    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
        {
            int i1 = (i + 1) % 3;
            int i2 = (i + 2) % 3;
            int j1 = (j + 1) % 3;
            int j2 = (j + 2) % 3;
            adj[j][i] = m[i1][j1] * m[i2][j2] - m[i1][j2] * m[i2][j1];
        }
    return adj * (1.0 / determinant(m));
}

Matrix4 inverse(const Matrix4 &m)
{
    Matrix4 adj;
    Matrix3 sub;

    for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++)
        {
            for (int k = 0; k < 3; k++)
                for (int l = 0; l < 3; l++)
                    sub[k][l] = m[k + (k >= i)][l + (l >= j)];
            double s = determinant(sub);
            if ((i ^ j) & 1) adj[j][i] = -s; else adj[j][i] = s;
        }
    return adj * (1.0 / determinant(m));
}

/*! Returns R such that m = RS where R is a rotation and S is symmetric */
Matrix3 polar_rotation(const Matrix3 &m)
{
    static MAT *M = m_get(3, 3);
    static VEC *d = v_get(3);
    static MAT *U = m_get(3, 3), *V = m_get(3, 3);
    Matrix3 out;

    for (int j = 0; j < 3; j++)
        for (int k = 0; k < 3; k++)
            m_set_val(M, j, k, m[j][k]);
    d = svd(M, U, V, d);
    M = mtrm_mlt(U, V, M);
    for (int j = 0; j < 3; j++)
        for (int k = 0; k < 3; k++)
            out[j][k] = m_entry(M, j, k);
    if (determinant(out) < 0.0)
    {
        for (int i = 0; i < 3; i++)
            m_set_val(V, 2, i, -m_entry(V, 2, i));
        M = mtrm_mlt(U, V, M);
        for (int j = 0; j < 3; j++)
            for (int k = 0; k < 3; k++)
                out[j][k] = m_entry(M, j, k);
    }
    return out;
}

bool matrix_symmetric(const Matrix2 &m, double tol)
{
    return matrix_norm2(m - transpose(m)) < tol * max(matrix_norm2(m), 1.0);
}

bool matrix_symmetric(const Matrix3 &m, double tol)
{
    return matrix_norm2(m - transpose(m)) < tol * max(matrix_norm2(m), 1.0);
}

bool matrix_symmetric(const Matrix4 &m, double tol)
{
    return matrix_norm2(m - transpose(m)) < tol * max(matrix_norm2(m), 1.0);
}

double matrix_uniform_scale(const Matrix3 &m, double tol)
{
    double d = determinant(m);
    if (fabs(d) < tol) return 0.0;
    if (d <= 0.0) return -1.0;
    d = pow(d, 2.0 / 3.0);   /* Expected scale factor */
    Matrix3 m2 = m * transpose(m);
    if (matrix_norm2(m2 - Matrix3() * d) < 1e-6 * d)
        return d;
    else
        return -1.0;
}

Matrix4 matrix_scale(double s)
{
    return Matrix4(  s, 0.0, 0.0, 0.0,
                   0.0,   s, 0.0, 0.0,
                   0.0, 0.0,   s, 0.0,
                   0.0, 0.0, 0.0, 1.0);
}

Matrix4 matrix_translate(const Vector3 &t)
{
    return Matrix4(1.0, 0.0, 0.0, t[0],
                   0.0, 1.0, 0.0, t[1],
                   0.0, 0.0, 1.0, t[2],
                   0.0, 0.0, 0.0, 1.0);
}

Matrix4 matrix_rotate(double ang, const Vector3 &axis)
{
    double c = cos(ang);
    double s = sin(ang);
    Vector3 u = normalise(axis);
    Matrix3 S( 0,    -u[2],  u[1],
               u[2],  0,    -u[0],
              -u[1],  u[0],  0);

    Matrix4 R;

    for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++)
        {
            if (i == 3 || j == 3) R[i][j] = i == j;
            else
            {
                R[i][j] = S[i][j] * s + u[i] * u[j] * (1 - c);
                if (i == j) R[i][j] += c;
            }
        }
    return R;
}

Matrix4 matrix_look_at(const Vector3 &eye,
                       const Vector3 &centre,
                       const Vector3 &up)
{
    Vector3 f = normalise(centre - eye);
    Vector3 s, u;

    s = cross(f, normalise(up));
    u = cross(s, f);
    return Matrix4(s[0], s[1], s[2], 0.0,
                   u[0], u[1], u[2], 0.0,
                   -f[0], -f[1], -f[2], 0.0,
                   0.0, 0.0, 0.0, 1.0) * matrix_translate(-eye);
}


/**************************************************************
 * Test code
 **************************************************************/

#if HAVE_CPPUNIT
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>

class CoordTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(CoordTest);
    CPPUNIT_TEST(determinant);
    CPPUNIT_TEST(symmetric);
    CPPUNIT_TEST(inverse);
    CPPUNIT_TEST(polar_rotation);
    CPPUNIT_TEST(norm);
    CPPUNIT_TEST_SUITE_END();
private:
    Matrix4 A;
    Matrix3 A3;

    void assert_identity(const Matrix3 &m)
    {
        for (int i = 0; i < 3; i++)
            for (int j = 0; j < 3; j++)
                CPPUNIT_ASSERT_DOUBLES_EQUAL(m[i][j], (i == j) ? 1.0 : 0.0, 1e-7);
    }

    void assert_identity(const Matrix4 &m)
    {
        for (int i = 0; i < 4; i++)
            for (int j = 0; j < 4; j++)
                CPPUNIT_ASSERT_DOUBLES_EQUAL(m[i][j], (i == j) ? 1.0 : 0.0, 1e-7);
    }

    void assert_zero(const Matrix3 &m)
    {
        for (int i = 0; i < 3; i++)
            for (int j = 0; j < 3; j++)
                CPPUNIT_ASSERT_DOUBLES_EQUAL(m[i][j], 0.0, 1e-7);
    }

    void assert_zero(const Matrix4 &m)
    {
        for (int i = 0; i < 4; i++)
            for (int j = 0; j < 4; j++)
                CPPUNIT_ASSERT_DOUBLES_EQUAL(m[i][j], 0.0, 1e-7);
    }

public:
    void setUp()
    {
        /* Random matrix generated by Octave */
        A = Matrix4(Vector4(0.865846, 0.306571, 0.195325, 0.672628),
                    Vector4(0.063441, 0.547351, 0.882569, 0.167962),
                    Vector4(0.991781, 0.373255, 0.425959, 0.272165),
                    Vector4(0.0,      0.0,      0.0,      1.0));
        A3 = matrix4_to_matrix3(A);
    }

    void determinant()
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(::determinant(A), 0.0752956605754505, 1e-7);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(::determinant(A3), 0.0752956605754505, 1e-7);
    }

    void inverse()
    {
        assert_identity(::inverse(A) * A);
        assert_identity(::inverse(A3) * A3);
    }

    void symmetric()
    {
        CPPUNIT_ASSERT(!matrix_symmetric(A));
        CPPUNIT_ASSERT(!matrix_symmetric(A3));
        CPPUNIT_ASSERT(matrix_symmetric(A + transpose(A)));
        CPPUNIT_ASSERT(matrix_symmetric(A3 + transpose(A3)));
    }

    void polar_rotation()
    {
        Matrix3 R = ::polar_rotation(A3);
        assert_identity(R * transpose(R));
        CPPUNIT_ASSERT(::determinant(R) > 0.0);
        CPPUNIT_ASSERT(matrix_symmetric(::inverse(R) * A3));
    }

    void norm()
    {
        CPPUNIT_ASSERT_DOUBLES_EQUAL(matrix_norm2(A), 1.7859746663901506, 1e-7);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(matrix_norm2(A3), 1.5975222865434506, 1e-7);
    }
};

CPPUNIT_TEST_SUITE_REGISTRATION(CoordTest);

#endif /* HAVE_CPPUNIT */

/*! @} */
