/*! \addtogroup geom
 * @{
 */

/*! \file
 * \brief Simple vector code
 */

#ifndef ALOD_GEOM_COORDS_H
#define ALOD_GEOM_COORDS_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <cstddef>
#include <cmath>
#include <iostream>
#include <algorithm>

//! A 2D vector of \c double
class Vector2
{
private:
    double v[2];

public:
    Vector2() { v[0] = v[1] = 0.0; }
    Vector2(double x, double y) { v[0] = x; v[1] = y; }
    Vector2 &assign(double x, double y) { v[0] = x; v[1] = y; return *this; }
    Vector2 &zero() { v[0] = 0.0; v[1] = 0.0; return *this; }
    double &x() { return v[0]; }
    double &y() { return v[1]; }
    double &operator[](std::size_t index) { return v[index]; }
    double operator[](std::size_t index) const { return v[index]; }

    Vector2 operator +(const Vector2 &b) const
    {
        return Vector2(v[0] + b.v[0],
                       v[1] + b.v[1]);
    }

    Vector2 operator -(const Vector2 &b) const
    {
        return Vector2(v[0] - b.v[0],
                       v[1] - b.v[1]);
    }

    Vector2 operator -() const
    { return Vector2(-v[0], -v[1]); }

    Vector2 &operator +=(const Vector2 &b)
    { v[0] += b.v[0]; v[1] += b.v[1]; return *this; }

    Vector2 &operator -=(const Vector2 &b)
    { v[0] -= b.v[0]; v[1] -= b.v[1]; return *this; }

    Vector2 operator *(double f) const
    { return Vector2(f * v[0], f * v[1]); }

    Vector2 &operator *=(double f)
    { v[0] *= f; v[1] *= f; return *this; }

    Vector2 operator /(double f) const
    {
        double inv = 1.0 / f;
        return Vector2(inv * v[0], inv * v[1]);
    }

    Vector2 &operator /=(double f)
    {
        double inv = 1.0 / f;
        v[0] *= inv; v[1] *= inv; return *this;
    }

    //! zero test
    bool operator !() const { return !v[0] && !v[1]; }

    bool operator == (const Vector2 &b) const
    { return v[0] == b.v[0] && v[1] == b.v[1]; }

    //! Dot product
    double operator *(const Vector2 &b) const
    { return v[0] * b.v[0] + v[1] * b.v[1]; }
};

static inline Vector2 operator *(double f, const Vector2 &b)
{
    return Vector2(f * b[0], f * b[1]);
}

//! Computes length in an unstable manner
static inline double length_fast(const Vector2 &v)
{ return std::sqrt(v[0] * v[0] + v[1] * v[1]); }
//! Returns the lower corner of a bounding box containing two points
static inline Vector2 vector_min(const Vector2 &a, const Vector2 &b)
{ return Vector2(std::min(a[0], b[0]), std::min(a[1], b[1])); }
//! Returns the upper corner of a bounding box containing two points
static inline Vector2 vector_max(const Vector2 &a, const Vector2 &b)
{ return Vector2(std::max(a[0], b[0]), std::max(a[1], b[1])); }

//! Computes length in a slower, stable manner
double length(const Vector2 &v);
//! Returns \a v scaled to unit length
/*! \pre \a v is not zero.
 */
Vector2 normalise(const Vector2 &v);
//! Return the double-area of the enclosed triangle
double cross(const Vector2 &a, const Vector2 &b);
//! Determines whether \a p1 \a p2 intersects \a q1 \a q2
bool intersects(const Vector2 &p1, const Vector2 &p2,
                const Vector2 &q1, const Vector2 &q2,
                bool open);
//! Computes the intersection of \a p1 \a p2 with \a q1 \a q2
Vector2 intersection(const Vector2 &p1, const Vector2 &p2,
                     const Vector2 &q1, const Vector2 &q2,
                     double &s, double &t);

//! A 3D vector of \c double
class Vector3
{
private:
    double v[3];

public:
    Vector3() { v[0] = v[1] = v[2] = 0.0; }
    Vector3(double x, double y, double z)
    { v[0] = x; v[1] = y; v[2] = z; }
    Vector3 &assign(double x, double y, double z)
    { v[0] = x; v[1] = y; v[2] = z; return *this; }
    Vector3 &zero()
    { v[0] = 0.0; v[1] = 0.0; v[2] = 0.0; return *this; }
    double &x() { return v[0]; }
    double &y() { return v[1]; }
    double &z() { return v[2]; }
    double &operator[](std::size_t index) { return v[index]; }
    double operator[](std::size_t index) const { return v[index]; }

    Vector3 operator +(const Vector3 &b) const
    {
        return Vector3(v[0] + b.v[0],
                       v[1] + b.v[1],
                       v[2] + b.v[2]);
    }

    Vector3 operator -(const Vector3 &b) const
    {
        return Vector3(v[0] - b.v[0],
                       v[1] - b.v[1],
                       v[2] - b.v[2]);
    }

    Vector3 operator -() const
    { return Vector3(-v[0], -v[1], -v[2]); }

    Vector3 &operator +=(const Vector3 &b)
    { v[0] += b.v[0]; v[1] += b.v[1]; v[2] += b.v[2]; return *this; }

    Vector3 &operator -=(const Vector3 &b)
    { v[0] -= b.v[0]; v[1] -= b.v[1]; v[2] -= b.v[2]; return *this; }

    Vector3 operator *(double f) const
    { return Vector3(f * v[0], f * v[1], f * v[2]); }

    Vector3 &operator *=(double f)
    { v[0] *= f; v[1] *= f; v[2] *= f; return *this; }

    Vector3 operator /(double f) const
    {
        double inv = 1.0 / f;
        return Vector3(inv * v[0], inv * v[1], inv * v[2]);
    }

    Vector3 &operator /=(double f)
    {
        double inv = 1.0 / f;
        v[0] *= inv; v[1] *= inv; v[2] *= inv; return *this;
    }

    //! zero test
    bool operator !() const { return !v[0] && !v[1] && !v[2]; }

    bool operator == (const Vector3 &b) const
    { return v[0] == b.v[0] && v[1] == b.v[1] && v[2] == b.v[2]; }

    //! Dot product
    double operator *(const Vector3 &b) const
    { return v[0] * b.v[0] + v[1] * b.v[1] + v[2] * b.v[2]; }
};

static inline Vector3 operator *(double f, const Vector3 &b)
{
    return Vector3(f * b[0], f * b[1], f * b[2]);
}

//! Computes length in an unstable manner
static inline double length_fast(const Vector3 &v)
{ return std::sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]); }
//! Returns the lower corner of a bounding box containing two points
static inline Vector3 vector_min(const Vector3 &a, const Vector3 &b)
{ return Vector3(std::min(a[0], b[0]), std::min(a[1], b[1]), std::min(a[2], b[2])); }
//! Returns the upper corner of a bounding box containing two points
static inline Vector3 vector_max(const Vector3 &a, const Vector3 &b)
{ return Vector3(std::max(a[0], b[0]), std::max(a[1], b[1]), std::max(a[2], b[2])); }

//! Computes length in a slower, stable manner
double length(const Vector3 &v);
//! Returns \a v scaled to unit length
/*! \pre \a v is not zero.
 */
Vector3 normalise(const Vector3 &v);
//! Returns the cross product of \a a and \a b
Vector3 cross(const Vector3 &a, const Vector3 &b);
//! Returns twice the area of the triangle formed by vectors \a a, \a b
static inline double double_area(const Vector3 &a, const Vector3 &b)
{ return length(cross(a, b)); }

//! A 4D vector of \c double
class Vector4
{
private:
    double v[4];

public:
    Vector4() { v[0] = v[1] = v[2] = v[3] = 0.0; }
    Vector4(double x, double y, double z, double w)
    { v[0] = x; v[1] = y; v[2] = z; v[3] = w; }
    Vector4 &assign(double x, double y, double z, double w)
    { v[0] = x; v[1] = y; v[2] = z; v[3] = w; return *this; }
    Vector4 &zero()
    { v[0] = 0.0; v[1] = 0.0; v[2] = 0.0; v[3] = 0.0; return *this; }
    double &x() { return v[0]; }
    double &y() { return v[1]; }
    double &z() { return v[2]; }
    double &w() { return v[3]; }
    double &operator[](std::size_t index) { return v[index]; }
    double operator[](std::size_t index) const { return v[index]; }

    Vector4 operator +(const Vector4 &b) const
    {
        return Vector4(v[0] + b.v[0],
                       v[1] + b.v[1],
                       v[2] + b.v[2],
                       v[3] + b.v[3]);
    }

    Vector4 operator -(const Vector4 &b) const
    {
        return Vector4(v[0] - b.v[0],
                       v[1] - b.v[1],
                       v[2] - b.v[2],
                       v[3] - b.v[3]);
    }

    Vector4 operator -() const
    { return Vector4(-v[0], -v[1], -v[2], -v[3]); }

    Vector4 &operator +=(const Vector4 &b)
    { v[0] += b.v[0]; v[1] += b.v[1]; v[2] += b.v[2]; v[3] += b.v[3]; return *this; }

    Vector4 &operator -=(const Vector4 &b)
    { v[0] -= b.v[0]; v[1] -= b.v[1]; v[2] -= b.v[2]; v[3] -= b.v[3]; return *this; }

    Vector4 operator *(double f) const
    { return Vector4(f * v[0], f * v[1], f * v[2], f * v[3]); }

    Vector4 &operator *=(double f)
    { v[0] *= f; v[1] *= f; v[2] *= f; v[3] *= f; return *this; }

    Vector4 operator /(double f) const
    {
        double inv = 1.0 / f;
        return Vector4(inv * v[0], inv * v[1], inv * v[2], inv * v[3]);
    }

    Vector4 &operator /=(double f)
    {
        double inv = 1.0 / f;
        v[0] *= inv; v[1] *= inv; v[2] *= inv; v[3] *= inv; return *this;
    }

    //! zero test
    bool operator !() const { return !v[0] && !v[1] && !v[2] && !v[3]; }

    bool operator == (const Vector4 &b) const
    { return v[0] == b.v[0] && v[1] == b.v[1]
        && v[2] == b.v[2] && v[3] == b.v[3]; }

    //! Dot product
    double operator *(const Vector4 &b) const
    { return v[0] * b.v[0] + v[1] * b.v[1] + v[2] * b.v[2] + v[3] * b.v[3]; }
};

static inline Vector4 operator *(double f, const Vector4 &b)
{
    return Vector4(f * b[0], f * b[1], f * b[2], f * b[3]);
}

//! Computes length in an unstable manner
static inline double length_fast(const Vector4 &v)
{ return std::sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2] + v[3] * v[3]); }
//! Returns the lower corner of a bounding box containing two points
static inline Vector4 vector_min(const Vector4 &a, const Vector4 &b)
{ return Vector4(std::min(a[0], b[0]), std::min(a[1], b[1]),
                 std::min(a[2], b[2]), std::min(a[3], b[3])); }
//! Returns the upper corner of a bounding box containing two points
static inline Vector4 vector_max(const Vector4 &a, const Vector4 &b)
{ return Vector4(std::max(a[0], b[0]), std::max(a[1], b[1]),
                 std::max(a[2], b[2]), std::max(a[3], b[3])); }

//! Computes length in a slower, stable manner
double length(const Vector4 &v);
//! Returns \a v scaled to unit length
/*! \pre \a v is not zero.
 */
Vector4 normalise(const Vector4 &v);

std::ostream &operator <<(std::ostream &, const Vector2 &);
std::ostream &operator <<(std::ostream &, const Vector3 &);
std::ostream &operator <<(std::ostream &, const Vector4 &);

class Matrix2;
static inline Vector2 operator *(const Vector2 &a, const Matrix2 &b);
//! A 2 by 2 matrix of doubles
class Matrix2
{
private:
    Vector2 row[2];

public:
    Matrix2()
    {
        assign(1.0, 0.0,
               0.0, 1.0);
    }

    Matrix2(double aa, double ab,
            double ba, double bb)
    {
        row[0].assign(aa, ab);
        row[1].assign(ba, bb);
    }

    Matrix2(const Vector2 &r1, const Vector2 &r2)
    {
        row[0] = r1;
        row[1] = r2;
    }

    Matrix2 &assign(double aa, double ab,
                    double ba, double bb)
    {
        row[0].assign(aa, ab);
        row[1].assign(ba, bb);
        return *this;
    }

    Matrix2 &assign(const Vector2 &r1, const Vector2 &r2)
    {
        row[0] = r1;
        row[1] = r2;
        return *this;
    }

    Matrix2 &zero()
    {
        row[0].zero();
        row[1].zero();
        return *this;
    }

    Matrix2 &identity()
    {
        assign(1.0, 0.0,
               0.0, 1.0);
        return *this;
    }

    Vector2 &operator[](int index) { return row[index]; }
    const Vector2 &operator[](int index) const { return row[index]; }

    Matrix2 operator +(const Matrix2 &b) const
    { return Matrix2(row[0] + b[0], row[1] + b[1]); }
    Matrix2 &operator +=(const Matrix2 &b)
    { row[0] += b[0]; row[1] += b[1]; return *this; }
    Matrix2 operator -(const Matrix2 &b) const
    { return Matrix2(row[0] - b[0], row[1] - b[1]); }
    Matrix2 &operator -=(const Matrix2 &b)
    { row[0] -= b[0]; row[1] -= b[1]; return *this; }
    Matrix2 operator *(double b) const
    { return Matrix2(row[0] * b, row[1] * b); }
    Matrix2 operator *=(double b)
    { row[0] *= b; row[1] *= b; return *this; }
    Vector2 operator *(const Vector2 &b) const
    { return Vector2(row[0] * b, row[1] * b); }
    Matrix2 operator *(const Matrix2 &b) const
    { return Matrix2(row[0] * b, row[1] * b); }
    Matrix2 &operator *=(const Matrix2 &b)
    { return *this = *this * b; }
};

static inline Vector2 operator *(const Vector2 &a, const Matrix2 &b)
{
    return Vector2(a[0] * b[0][0] + a[1] * b[1][0],
                   a[0] * b[0][1] + a[1] * b[1][1]);
}

static inline Matrix2 operator *(double a, const Matrix2 &b)
{
    return b * a;
}

static inline Matrix2 transpose(const Matrix2 &m)
{
    return Matrix2(m[0][0], m[1][0],
                   m[0][1], m[1][1]);
}

class Matrix3;
static inline Vector3 operator *(const Vector3 &a, const Matrix3 &b);
//! A 3 by 3 matrix of doubles
class Matrix3
{
private:
    Vector3 row[3];

public:
    Matrix3()
    {
        assign(1.0, 0.0, 0.0,
               0.0, 1.0, 0.0,
               0.0, 0.0, 1.0);
    }

    Matrix3(double aa, double ab, double ac,
            double ba, double bb, double bc,
            double ca, double cb, double cc)
    {
        row[0].assign(aa, ab, ac);
        row[1].assign(ba, bb, bc);
        row[2].assign(ca, cb, cc);
    }

    Matrix3(const Vector3 &r1, const Vector3 &r2, const Vector3 &r3)
    {
        row[0] = r1;
        row[1] = r2;
        row[2] = r3;
    }

    Matrix3 &assign(double aa, double ab, double ac,
                    double ba, double bb, double bc,
                    double ca, double cb, double cc)
    {
        row[0].assign(aa, ab, ac);
        row[1].assign(ba, bb, bc);
        row[2].assign(ca, cb, cc);
        return *this;
    }

    Matrix3 &assign(const Vector3 &r1, const Vector3 &r2, const Vector3 &r3)
    {
        row[0] = r1;
        row[1] = r2;
        row[2] = r3;
        return *this;
    }

    Matrix3 &zero()
    {
        row[0].zero();
        row[1].zero();
        row[2].zero();
        return *this;
    }

    Matrix3 &identity()
    {
        assign(1.0, 0.0, 0.0,
               0.0, 1.0, 0.0,
               0.0, 0.0, 1.0);
        return *this;
    }

    Vector3 &operator[](int index) { return row[index]; }
    const Vector3 &operator[](int index) const { return row[index]; }

    Matrix3 operator +(const Matrix3 &b) const
    { return Matrix3(row[0] + b[0], row[1] + b[1], row[2] + b[2]); }
    Matrix3 &operator +=(const Matrix3 &b)
    { row[0] += b[0]; row[1] += b[1]; row[2] += b[2]; return *this; }
    Matrix3 operator -(const Matrix3 &b) const
    { return Matrix3(row[0] - b[0], row[1] - b[1], row[2] - b[2]); }
    Matrix3 &operator -=(const Matrix3 &b)
    { row[0] -= b[0]; row[1] -= b[1]; row[2] -= b[2]; return *this; }
    Matrix3 operator *(double b) const
    { return Matrix3(row[0] * b, row[1] * b, row[2] * b); }
    Matrix3 operator *=(double b)
    { row[0] *= b; row[1] *= b; row[2] *= b; return *this; }
    Vector3 operator *(const Vector3 &b) const
    { return Vector3(row[0] * b, row[1] * b, row[2] * b); }
    Matrix3 operator *(const Matrix3 &b) const
    { return Matrix3(row[0] * b, row[1] * b, row[2] * b); }
    Matrix3 &operator *=(const Matrix3 &b)
    { return *this = *this * b; }
};

static inline Vector3 operator *(const Vector3 &a, const Matrix3 &b)
{
    return Vector3(a[0] * b[0][0] + a[1] * b[1][0] + a[2] * b[2][0],
                   a[0] * b[0][1] + a[1] * b[1][1] + a[2] * b[2][1],
                   a[0] * b[0][2] + a[1] * b[1][2] + a[2] * b[2][2]);
}

static inline Matrix3 operator *(double a, const Matrix3 &b)
{
    return b * a;
}

static inline Matrix3 transpose(const Matrix3 &m)
{
    return Matrix3(m[0][0], m[1][0], m[2][0],
                   m[0][1], m[1][1], m[2][1],
                   m[0][2], m[1][2], m[2][2]);
}

class Matrix4;
static inline Vector4 operator *(const Vector4 &a, const Matrix4 &b);
//! A 4 by 4 matrix of doubles
class Matrix4
{
private:
    Vector4 row[4];

public:
    Matrix4()
    {
        assign(1.0, 0.0, 0.0, 0.0,
               0.0, 1.0, 0.0, 0.0,
               0.0, 0.0, 1.0, 0.0,
               0.0, 0.0, 0.0, 1.0);
    }

    Matrix4(double aa, double ab, double ac, double ad,
            double ba, double bb, double bc, double bd,
            double ca, double cb, double cc, double cd,
            double da, double db, double dc, double dd)
    {
        row[0].assign(aa, ab, ac, ad);
        row[1].assign(ba, bb, bc, bd);
        row[2].assign(ca, cb, cc, cd);
        row[3].assign(da, db, dc, dd);
    }

    Matrix4(const Vector4 &r1, const Vector4 &r2,
            const Vector4 &r3, const Vector4 &r4)
    {
        row[0] = r1;
        row[1] = r2;
        row[2] = r3;
        row[3] = r4;
    }

    Matrix4 &assign(double aa, double ab, double ac, double ad,
                    double ba, double bb, double bc, double bd,
                    double ca, double cb, double cc, double cd,
                    double da, double db, double dc, double dd)
    {
        row[0].assign(aa, ab, ac, ad);
        row[1].assign(ba, bb, bc, bd);
        row[2].assign(ca, cb, cc, cd);
        row[3].assign(da, db, dc, dd);
        return *this;
    }

    Matrix4 &assign(const Vector4 &r1, const Vector4 &r2,
                    const Vector4 &r3, const Vector4 &r4)
    {
        row[0] = r1;
        row[1] = r2;
        row[2] = r3;
        row[3] = r4;
        return *this;
    }

    Matrix4 &zero()
    {
        row[0].zero();
        row[1].zero();
        row[2].zero();
        row[3].zero();
        return *this;
    }

    Matrix4 &identity()
    {
        assign(1.0, 0.0, 0.0, 0.0,
               0.0, 1.0, 0.0, 0.0,
               0.0, 0.0, 1.0, 0.0,
               0.0, 0.0, 0.0, 1.0);
        return *this;
    }

    Vector4 &operator[](int index) { return row[index]; }
    const Vector4 &operator[](int index) const { return row[index]; }

    Matrix4 operator +(const Matrix4 &b) const
    { return Matrix4(row[0] + b[0], row[1] + b[1], row[2] + b[2], row[3] + b[3]); }
    Matrix4 &operator +=(const Matrix4 &b)
    { row[0] += b[0]; row[1] += b[1]; row[2] += b[2]; row[3] += b[3]; return *this; }
    Matrix4 operator -(const Matrix4 &b) const
    { return Matrix4(row[0] - b[0], row[1] - b[1], row[2] - b[2], row[3] - b[3]); }
    Matrix4 &operator -=(const Matrix4 &b)
    { row[0] -= b[0]; row[1] -= b[1]; row[2] -= b[2]; row[3] -= b[3]; return *this; }
    Matrix4 operator *(double b) const
    { return Matrix4(row[0] * b, row[1] * b, row[2] * b, row[3] * b); }
    Matrix4 operator *=(double b)
    { row[0] *= b; row[1] *= b; row[2] *= b; row[3] *= b; return *this; }
    Vector4 operator *(const Vector4 &b) const
    { return Vector4(row[0] * b, row[1] * b, row[2] * b, row[3] * b); }
    Matrix4 operator *(const Matrix4 &b) const
    { return Matrix4(row[0] * b, row[1] * b, row[2] * b, row[3] * b); }
    Matrix4 &operator *=(const Matrix4 &b)
    { return *this = *this * b; }
};

static inline Vector4 operator *(const Vector4 &a, const Matrix4 &b)
{
    return Vector4(a[0] * b[0][0] + a[1] * b[1][0] + a[2] * b[2][0] + a[3] * b[3][0],
                   a[0] * b[0][1] + a[1] * b[1][1] + a[2] * b[2][1] + a[3] * b[3][1],
                   a[0] * b[0][2] + a[1] * b[1][2] + a[2] * b[2][2] + a[3] * b[3][2],
                   a[0] * b[0][3] + a[1] * b[1][3] + a[2] * b[2][3] + a[3] * b[3][3]);
}

static inline Matrix4 operator *(double a, const Matrix4 &b)
{
    return b * a;
}

static inline Matrix4 transpose(const Matrix4 &m)
{
    return Matrix4(m[0][0], m[1][0], m[2][0], m[3][0],
                   m[0][1], m[1][1], m[2][1], m[3][1],
                   m[0][2], m[1][2], m[2][2], m[3][2],
                   m[0][3], m[1][3], m[2][3], m[3][3]);
}

/* Homogeneous coordinate routines */

static inline Matrix4 matrix3_to_matrix4(const Matrix3 &m)
{
    return Matrix4(m[0][0], m[0][1], m[0][2], 0.0,
                   m[1][0], m[1][1], m[1][2], 0.0,
                   m[2][0], m[2][1], m[2][2], 0.0,
                   0.0, 0.0, 0.0, 1.0);
}

static inline Matrix4 matrix3_translate3_to_matrix4(const Matrix3 &m,
                                                    const Vector3 &t)
{
    return Matrix4(m[0][0], m[0][1], m[0][2], t[0],
                   m[1][0], m[1][1], m[1][2], t[1],
                   m[2][0], m[2][1], m[2][2], t[2],
                   0.0, 0.0, 0.0, 1.0);
}

static inline Matrix3 matrix4_to_matrix3(const Matrix4 &m)
{
    return Matrix3(m[0][0], m[0][1], m[0][2],
                   m[1][0], m[1][1], m[1][2],
                   m[2][0], m[2][1], m[2][2]);
}

static inline Vector3 vector4_to_vector3(const Vector4 &v)
{
    return Vector3(v[0], v[1], v[2]);
}

static inline Vector4 vector3_to_vector4(const Vector3 &v, double w)
{
    return Vector4(v[0], v[1], v[2], w);
}

static inline Vector3 matrix4_to_translate3(const Matrix4 &m)
{
    return Vector3(m[0][3], m[1][3], m[2][3]);
}

static inline Vector4 matrix4_to_translate4(const Matrix4 &m, double w)
{
    return Vector4(m[0][3], m[1][3], m[2][3], w);
}

/* Misc */

double determinant(const Matrix2 &m);
double determinant(const Matrix3 &m);
double determinant(const Matrix4 &m);
Matrix2 inverse(const Matrix2 &m);
Matrix3 inverse(const Matrix3 &m);
Matrix4 inverse(const Matrix4 &m);

//! Extracts the rotation from m
Matrix3 polar_rotation(const Matrix3 &m);

//! Returns true if the matrix is symmetric
bool matrix_symmetric(const Matrix2 &m, double tol = 1e-6);
//! Returns true if the matrix is symmetric
bool matrix_symmetric(const Matrix3 &m, double tol = 1e-6);
//! Returns true if the matrix is symmetric
bool matrix_symmetric(const Matrix4 &m, double tol = 1e-6);

//! Returns common singular value, or -1.0 if non-uniform scaling
double matrix_uniform_scale(const Matrix3 &m, double tol = 1e-6);

//! Returns \a y and \a z normal to \a x and each other, in a right-handed system
void complete_basis(const Vector3 &x, Vector3 &y, Vector3 &z);

//! Largest singular value of \a m
double matrix_norm2(const Matrix2 &m);
//! Largest singular value of \a m
double matrix_norm2(const Matrix3 &m);
//! Largest singular value of \a m
double matrix_norm2(const Matrix4 &m);

//! An equivalent to \c glTranslated
Matrix4 matrix_translate(const Vector3 &t);

//! An equivalent to \c glRotated, but with an angle in radians
Matrix4 matrix_rotate(double ang, const Vector3 &axis);

//! An equivalent to \c glScaled, but with uniform scaling
Matrix4 matrix_scale(double s);

//! An equivalent to \c gluLookAt
Matrix4 matrix_look_at(const Vector3 &eye,
                       const Vector3 &centre,
                       const Vector3 &up);

#include "sparse.h"

#endif /* ALOD_GEOM_COORDS_H */

/*! @} */
