/*! \addtogroup geom
 * @{
 */

/*! \file
 * \brief Code to find the intersection of a number of half-planes
 */

#include "geom/coords.h"
#include "kernel.h"
#include <vector>
#include <cmath>
#include <deque>
#include <glib.h>

using namespace std;

//! Represents a half-plane with quantised normals
/*! We quantise the normals, so that 'parallel' is consistently defined,
 * and to reduce numeric instability.
 */
struct HalfPlane
{
    double d;
    double dn;
    int nx;
    int ny;

    HalfPlane() {}
    HalfPlane(const Vector2 &pnt, const Vector2 &n);
    HalfPlane operator -() const; //!< Complementary half-plane
};

HalfPlane::HalfPlane(const Vector2 &pnt, const Vector2 &n)
{
    static const int qbits = 14;
    static const int quant = (1 << qbits);
    Vector2 nn = normalise(n);

    dn = pnt * nn;
    nx = (int) rint(quant * nn[0]);
    ny = (int) rint(quant * nn[1]);
    d = nx * pnt[0] + ny * pnt[1];
}

HalfPlane HalfPlane::operator -() const
{
    HalfPlane out;

    out.d = -d;
    out.dn = -dn;
    out.nx = -nx;
    out.ny = -ny;
    return out;
}

/*! Orders half-planes (like a < operator) by normal. A straight-up
 * normal is ordered first, then it goes ccw, ending with a normal
 * that is straight-up but skewed slightly to the right. If two
 * half-planes are parallel, the smaller is considered 'first'.
 */
static bool halfplane_compare(const HalfPlane &a, const HalfPlane &b)
{
    g_assert(a.nx != 0 || a.ny != 0);
    g_assert(b.nx != 0 || b.ny != 0);
    if (a.nx <= 0 && b.nx > 0) return true;
    if (b.nx <= 0 && a.nx > 0) return false;

    if (a.nx == 0 && b.nx == 0)
    {
        if (a.ny > 0 && b.ny < 0) return true;
        if (a.ny < 0 && b.ny > 0) return false;
    }
    else
    {
        int area = a.nx * b.ny - b.nx * a.ny;
        if (area != 0) return area > 0;
    }
    return a.dn > b.dn;
}

static bool halfplane_parallel(const HalfPlane &a, const HalfPlane &b)
{
    return (a.nx * b.ny == a.ny * b.nx
            && a.nx * b.nx + a.ny * b.ny > 0);
}

static Vector2 cut(const HalfPlane &a, const HalfPlane &b)
{
    int det = a.nx * b.ny - b.nx * a.ny;
    g_assert(det != 0);
    double idet = 1.0 / det;

    return Vector2((b.ny * a.d - a.ny * b.d) * idet,
                   (a.nx * b.d - b.nx * a.d) * idet);
}

static inline bool contains(const HalfPlane &h, const Vector2 &p)
{
    return h.nx * p[0] + h.ny * p[1] >= h.d;
}

//! Returns 1 for parallel, -1 for antiparallel, 0 otherwise
static int parallel(const HalfPlane &a, const HalfPlane &b)
{
    if (a.nx * b.ny == a.ny * b.nx)
    {
        if (a.nx * b.nx + a.ny * b.ny > 0) return 1;
        else return -1;
    }
    else return 0;
}

/* Determines whether a, b have an empty intersection */
static bool empty(const HalfPlane &a, const HalfPlane &b)
{
    return parallel(a, b) == -1 && -b.dn <= a.dn;
}

/* Determines whether a, b and c have an empty intersection.
 */
static bool empty(const HalfPlane &a, const HalfPlane &b, const HalfPlane &c)
{
    HalfPlane h[3] = {a, b, c};

    // Check for parallel planes
    for (int i = 0; i < 3; i++)
    {
        int j = (i == 2) ? 0 : i + 1;
        if (parallel(h[i], h[j]) == 1)
        {
            int k = 3 - i - j;
            if (h[i].dn > h[j].dn) return empty(h[i], h[k]);
            else return empty(h[j], h[k]);
        }
    }

    // Check for anti-parallel planes, knowing there are none parallel
    for (int i = 0; i < 3; i++)
    {
        int j = (i == 2) ? 0 : i + 1;
        if (parallel(h[i], h[j]) == -1)
            return empty(h[i], h[j]);
    }

    // No parallel or anti-parallel; go intersection hunting
    for (int i = 0; i < 3; i++)
    {
        int j = (i == 2) ? 0 : i + 1;
        int k = 3 - i - j;
        Vector2 c = cut(h[i], h[j]);
        if (contains(h[k], c)) return false;
    }
    return true;
}

/*! \param pnt points on the half-planes
 * \param normal normals to the half-plane boundaries, directed into plane
 * \returns a vector of intersection points
 * \pre \a pnt and \a normal have the same size
 * \pre \a pnt and \a normal define a closed polygon or an empty space
 * \post the result is ordered anti-clockwise
 * \post the returned vector is either empty (for empty intersection)
 * or has size 3 or greater.
 */
vector<Vector2> kernel(const vector<Vector2> &pnt,
                       const vector<Vector2> &normal) throw(open_kernel)
{
    g_assert(pnt.size() == normal.size());

    int H = normal.size();
    HalfPlane h[H];
    deque<int> chosen;

    for (int i = 0; i < H; i++)
        h[i] = HalfPlane(pnt[i], normal[i]);

    sort(h, h + H, halfplane_compare);
    H = unique(h, h + H, halfplane_parallel) - h; // eliminate parallel half-planes
    if (H < 3) throw open_kernel();

    // check that the kernel will be closed
    for (int i = 0; i < H; i++)
    {
        int j = i ? i - 1 : H - 1;
        if (h[j].nx * h[i].ny <= h[j].ny * h[i].nx)
            throw open_kernel();
    }

    if (empty(h[0], h[1])) return vector<Vector2>();
    chosen.push_back(0);
    chosen.push_back(1);
    for (int i = 2; i < H; i++)
    {
        while (chosen.size() >= 2)
        {
            int a = chosen[chosen.size() - 2];
            int b = chosen[chosen.size() - 1];
            if (empty(h[a], h[b], h[i])) return vector<Vector2>();
            if (!empty(h[a], -h[b], h[i])) break;
            chosen.pop_back();
        }
        while (chosen.size() >= 2)
        {
            int a = chosen[1];
            int b = chosen[0];
            if (empty(h[a], h[b], h[i])) return vector<Vector2>();
            if (!empty(h[a], -h[b], h[i])) break;
            chosen.pop_front();
        }
        if (chosen.size() >= 2
            && empty(h[chosen.back()], h[chosen.front()], h[i]))
            return vector<Vector2>();
        g_assert(chosen.size() >= 2
                 || parallel(h[chosen[0]], h[i]) == 0);
        if (chosen.size() == 1
            || !empty(h[chosen.back()], h[chosen.front()], -h[i]))
            chosen.push_back(i);
    }
    g_assert(chosen.size() >= 3);

    vector<Vector2> answer;
    for (size_t i = 0; i < chosen.size(); i++)
    {
        int j = i + 1;
        if (j == (int) chosen.size()) j = 0;
        answer.push_back(cut(h[chosen[i]], h[chosen[j]]));
    }
    return answer;
}

//! Determine if a point lies in an intersection of half-planes
/*! The first two parameters are the same as for \c kernel, but this
 * routine is much faster, as well as more robust. The boundaries of the
 * kernel are defined to lie outside.
 */
bool in_kernel(const vector<Vector2> &pnt,
               const vector<Vector2> &normal,
               const Vector2 &p) throw(open_kernel)
{
    g_assert(pnt.size() == normal.size());
    for (size_t i = 0; i < pnt.size(); i++)
        if ((pnt[i] - p) * normal[i] >= normal[i] * normal[i] * -1e-4)
            return false;
    return true;
}

/**************************************************************
 * Test code
 **************************************************************/

#if HAVE_CPPUNIT
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>

class KernelTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(KernelTest);
    CPPUNIT_TEST(open);
    CPPUNIT_TEST(empty);
    CPPUNIT_TEST(closed);
    CPPUNIT_TEST_SUITE_END();

    int kernel_size(Vector2 *arr, int n) throw(open_kernel)
    {
        vector<Vector2> pnts, normals;
        vector<Vector2> ans;
        for (int i = 0; i < n; i += 2)
        {
            pnts.push_back(arr[i]);
            normals.push_back(arr[i + 1]);
        }
        ans = kernel(pnts, normals);
        return ans.size();
    }

#define KERNEL_SIZE(a) kernel_size((a), sizeof(a) / sizeof((a)[0]))

public:
    void open()
    {
        Vector2 open1[] =
        {
            Vector2(0.0, 0.0), Vector2(0.0, 1.0),
            Vector2(1.0, 0.0), Vector2(-1.0, 0.0),
            Vector2(1.0, 1.0), Vector2(-1.0, 1.0)
        };
        CPPUNIT_ASSERT_THROW(KERNEL_SIZE(open1), open_kernel);
    }

    void closed()
    {
        Vector2 tri1[] =
        {
            Vector2(0.0, 0.0), Vector2(0.0, 1.0),
            Vector2(1.0, 0.0), Vector2(-1.0, 0.0),
            Vector2(1.0, 1.0), Vector2(1.0, -1.0)
        };

        Vector2 tri2[] =
        {
            Vector2(0.0, 0.0), Vector2(0.0, 1.0),
            Vector2(1.0, 0.0), Vector2(-1.0, 0.0),
            Vector2(1.0, 2.0), Vector2(-1.0, -1.0),
            Vector2(1.0, 1.0), Vector2(1.0, -1.0)
        };

        Vector2 quad1[] =
        {
            Vector2(0.0, 0.0), Vector2(0.0, 1.0),
            Vector2(1.0, 0.0), Vector2(-1.0, 0.0),
            Vector2(1.0, 1.0), Vector2(0.0, -1.0),
            Vector2(0.0, 1.0), Vector2(1.0, 0.0)
        };

        vector<Vector2> pnts, normals, ans;

        CPPUNIT_ASSERT_EQUAL(KERNEL_SIZE(tri1), 3);
        CPPUNIT_ASSERT_EQUAL(KERNEL_SIZE(tri2), 3);
        CPPUNIT_ASSERT_EQUAL(KERNEL_SIZE(quad1), 4);
    }

    void empty()
    {
        Vector2 empty1[] =
        {
            Vector2(0.0, 0.0), Vector2(0.0, 1.0),
            Vector2(0.0, -1.0), Vector2(0.0, -1.0),
            Vector2(1.0, 0.0), Vector2(-1.0, 0.0),
            Vector2(0.0, 1.0), Vector2(1.0, 0.0)
        };

        Vector2 empty2[] =
        {
            Vector2(0.0, 0.0), Vector2(0.0, 1.0),
            Vector2(1.0, 0.0), Vector2(-1.0, 0.0),
            Vector2(1.0, 2.0), Vector2(1.0, 1.0),
            Vector2(1.0, 1.0), Vector2(1.0, -1.0)
        };

        CPPUNIT_ASSERT_EQUAL(KERNEL_SIZE(empty1), 0);
        CPPUNIT_ASSERT_EQUAL(KERNEL_SIZE(empty2), 0);
    }
};

CPPUNIT_TEST_SUITE_REGISTRATION(KernelTest);

#endif /* HAVE_CPPUNIT */
/*! @} */
