/*! \addtogroup geom
 * @{
 */

/*! \file
 * \brief Various linear algebra tools built in meschach
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <glib.h>
#include <cmath>
#include <algorithm>
#include "geom/linear.h"
#include "common/mesch.h"

using namespace std;

void show_matrix(const char *name, MAT *matrix)
{
    printf("%s = [", name);
    for (size_t i = 0; i < matrix->m; i++)
    {
        for (size_t j = 0; j < matrix->n; j++)
            printf("% 17.14f ", m_entry(matrix, i, j));
        if (i + 1 < matrix->m) printf("\n");
    }
    printf("];\n");
}

void show_vector(const char *name, VEC *vec)
{
    printf("%s = [", name);
    for (size_t i = 0; i < vec->dim; i++)
    {
        printf("% 17.14f", v_entry(vec, i));
        if (i + 1 < vec->dim) printf("\n");
    }
    printf("];\n");
}

/* Takes the Q part of a Meschach QR-factorisation and computes Q^TB
 * May be in-situ.
 */
static MAT *QTmlt(MAT *QR, VEC *diag, MAT *B, MAT *X)
{
    VEC *tmp, *b = VNULL;

    if (X == MNULL || X->m != B->m || X->n != B->n)
        X = m_resize(X, B->m, B->n);
    tmp = v_get(QR->m);
    for (size_t j = 0; j < B->n; j++)
    {
        b = get_col(B, j, b);
        _Qsolve(QR, diag, b, b, tmp);
        set_col(X, j, b);
    }
    V_FREE(tmp);
    V_FREE(b);
    return X;
}

/* Meschach doesn't handle v_free of a sub_vec properly, which is
 * why we need to manually set ->ve to NULL. It gets matrices
 * right, however.
 */
static void v_free_sub(VEC *v)
{
    if (v)
    {
        v->ve = NULL;
        V_FREE(v);
    }
}

static MAT *sub_mat_copy(MAT *A, int r1, int c1, int r2, int c2, MAT *out)
{
    out = m_resize(out, r2 - r1 + 1, c2 - c1 + 1);
    for (int i = 0; i <= r2 - r1; i++)
        for (int j = 0; j <= c2 - c1; j++)
            m_set_val(out, i, j, m_entry(A, i + r1, j + c1));
    return out;
}

/*! \pre \a C has full rank and has at least as many columns as rows
 * \pre \a L is m by m and Q is m by n, where C is m by n
 * \post \f$ QQ^T = I\f$
 * \post \f$ Q = LC\f$
 * \post L is lower triangular
 * \note It is legal for \a C and \a Q to be the same
 */
void gram_schmidt(MAT *C, MAT *L, MAT *Q)
{
    VEC *v = VNULL, *v2 = VNULL, *l = VNULL, *l2 = VNULL;
    double n;

    g_assert(L != MNULL);
    g_assert(Q != MNULL);
    g_assert(L->m == C->m && L->n == C->m);
    g_assert(Q->m == C->m && Q->n == C->n);
    g_assert(C->m <= C->n);

    m_ident(L);
    if (Q != C) Q = m_copy(C, Q);

    l = v_get(C->m);
    for (size_t i = 0; i < Q->m; i++)
    {
        v = get_row(Q, i, v);
        n = v_norm2(v);
        g_assert(n >= 1e-18);
        v = sv_mlt(1.0 / n, v, v);
        v_zero(l);
        v_set_val(l, i, 1.0 / n);

        for (size_t j = 0; j < i; j++)
        {
            v2 = get_row(Q, j, v2);
            l2 = get_row(L, j, l2);
            n = in_prod(v, v2);
            v = v_mltadd(v, v2, -n, v);
            l = v_mltadd(l, l2, -n, l);
        }

        n = v_norm2(v);
        g_assert(n >= 1e-18);
        v = sv_mlt(1.0 / n, v, v);
        l = sv_mlt(1.0 / n, l, l);
        set_row(Q, i, v);
        set_row(L, i, l);
    }

    V_FREE(v);
    V_FREE(v2);
    V_FREE(l);
    V_FREE(l2);
}

struct UnweightedCLS
{
    MAT *L;
    MAT *Q;
};

UnweightedCLS *alloc_unweighted_cls(MAT *C)
{
    UnweightedCLS *handle = new UnweightedCLS;
    size_t r = C->m;

    handle->L = m_get(r, r);
    handle->Q = m_copy(C, MNULL);
    gram_schmidt(handle->Q, handle->L, handle->Q);
    return handle;
}

/*! We must minimise \f$||x||\f$ subject to \f$L^{-1}Qx = b\f$.
 * This means that \f$x\f$ is in the span of
 * \f$Q^T\f$ i.e. \f$x = Q^Tz\f$. But then \f$Lb = QQ^Tz = z\f$, so
 * \f$x = Q^TLb\f$.
 * \see \c alloc_unweighted_cls
 */
VEC *solve_unweighted_cls(UnweightedCLS *handle, VEC *b, double *cost, VEC *x)
{
    VEC *t;

    t = mv_mlt(handle->L, b, VNULL);
    x = vm_mlt(handle->Q, t, x);
    if (cost) *cost = in_prod(x, x);
    V_FREE(t);
    return x;
}

void free_unweighted_cls(UnweightedCLS *handle)
{
    M_FREE(handle->Q);
    M_FREE(handle->L);
    delete handle;
}

struct WeightedCLSNS
{
    VEC *rsqrts; //!< Reciprocal square root of s
    MAT *L;      //!< Gram-Schmidt factorisation of rs-scaled C
    MAT *Q;
};


WeightedCLSNS *alloc_weighted_cls_ns(VEC *s, MAT *C)
{
    WeightedCLSNS *handle = new WeightedCLSNS;
    size_t r = C->m;
    size_t c = C->n;
    g_assert(s->dim == c);

    handle->rsqrts = v_get(s->dim);
    handle->L = m_get(r, r);
    handle->Q = m_get(r, c);
    for (size_t j = 0; j < c; j++)
        v_set_val(handle->rsqrts, j, 1.0 / sqrt(v_entry(s, j)));
    for (size_t i = 0; i < r; i++)
        for (size_t j = 0; j < c; j++)
            m_set_val(handle->Q, i, j, m_entry(C, i, j) * v_entry(handle->rsqrts, j));
    gram_schmidt(handle->Q, handle->L, handle->Q);

    return handle;
}

/*! Let \f$y = S^{1/2}x\f$. We must minimise \f$||y||\f$ subject to
 * \f$L^{-1}Qy = b\f$. This means that \f$ y\f$ is in the span of
 * \f$Q^T\f$ i.e. \f$y = Q^Tz\f$. But then \f$Lb = QQ^Tz = z\f$, so
 * \f$y = Q^TLb\f$.
 * \see \c alloc_weighted_cls_ns
 */
VEC *solve_weighted_cls_ns(WeightedCLSNS *handle, VEC *b, double *cost, VEC *x)
{
    VEC *t;

    t = mv_mlt(handle->L, b, VNULL);
    x = vm_mlt(handle->Q, t, x);
    x = v_star(x, handle->rsqrts, x);
    if (cost)
    {
        *cost = 0.0;
        for (size_t i = 0; i < x->dim; i++)
        {
            double w = v_entry(x, i) / v_entry(handle->rsqrts, i);
            *cost += w * w;
        }
    }
    V_FREE(t);
    return x;
}

void free_weighted_cls_ns(WeightedCLSNS *handle)
{
    V_FREE(handle->rsqrts);
    M_FREE(handle->Q);
    M_FREE(handle->L);
    delete handle;
}

/*** Weighted constrained least squares with potentially singular weights ***/

struct WeightedCLS
{
    WeightedCLSNS *fallback;   //!< Fallback solver for non-singular case

    MAT *QR;      //!< QR-factorisation of B2
    MAT *B1upper;
    VEC *dw;      //!< Diagonal for QR-factorisation in \c QR
    PERM *pw;     //!< Permutation that has been applied to null-space params
    PERM *pweight; //!< Permutation to recover original variable order

    WeightedCLSNS *solvev;   //!< Handle to sub-problem for the span of P
    UnweightedCLS *solvew;   //!< Handle to sub-problem for the null-space of P
};

WeightedCLS *alloc_weighted_cls(VEC *s, MAT *C, double epss, double epsc)
{
    WeightedCLS *handle = new WeightedCLS;
    PERM *pweight, *pw;
    VEC *mys, *dw;
    MAT *B1, *B2, *C1, *C2, *R;
    int nv = 0, nw;      // number of weighted and unweighted variables
    int cw = 0, cv = 0;  // constraints that are unweighted and weighted
    double base;

    g_assert(s->dim == C->n);
    g_assert(C->m <= C->n);

    /* Sort the weights in descending order. We first negate the weights
     * so that we can use v_sort, which sorts in ascending order. The
     * permutation returned by v_sort maps from the unsorted order to the
     * sorted order.
     */
    mys = v_get(s->dim);
    for (size_t i = 0; i < s->dim; i++) v_set_val(mys, i, -v_entry(s, i));
    pweight = px_get(s->dim);
    mys = v_sort(mys, pweight);
    for (size_t i = 0; i < s->dim; i++) v_set_val(mys, i, -v_entry(mys, i));

    /* Identify "zero" weights */
    base = v_entry(mys, 0);
    g_assert(base > 0.0);
    g_assert(v_entry(mys, mys->dim - 1) >= 0.0);
    while (nv < (int) mys->dim && v_entry(mys, nv) >= epss * base) nv++;
    nw = mys->dim - nv;
    mys = v_resize(mys, nv);

    /* Check for non-singular case */
    if (nw == 0)
    {
        memset(handle, 0, sizeof(*handle));
        handle->fallback = alloc_weighted_cls_ns(s, C);
        return handle;
    }

    /* Apply the permutation to C and split into blocks */
    B1 = m_get(C->m, nv);
    B2 = m_get(C->m, C->n - nv);
    for (size_t r = 0; r < C->m; r++)
    {
        for (size_t c = 0; c < (size_t) nv; c++)
            m_set_val(B1, r, c, m_entry(C, r, pweight->pe[c]));
        for (size_t c = 0; c < C->n - nv; c++)
            m_set_val(B2, r, c, m_entry(C, r, pweight->pe[c + nv]));
    }
    /* QR-factorise the portion of the constraints dealing with unweighted
     * variables. Again, the returned permutation converts from the original
     * column order to the sorted column order.
     */
    pw = px_get(B2->n);
    dw = v_get(min(B2->m, B2->n));
    QRCPfactor(B2, dw, pw);         // B_2 now holds QR-factorisation of B2
    QTmlt(B2, dw, B1, B1);          // B_1 is now Q^TB_1

    /* Separate constraints on v, w from those purely on v (v being the
     * weighted eigenvectors).
     */
    while ((size_t) cw < B2->m && (size_t) cw < B2->n
           && fabs(m_entry(B2, cw, cw)) >= epsc) cw++;
    cv = B2->m - cw;
    C1 = cv ? sub_mat(B1, cw, 0, cw + cv - 1, B1->n - 1, MNULL) : MNULL;
    R = makeR(B2, MNULL);
    C2 = cw ? sub_mat(R, 0, 0, cw - 1, R->n - 1, MNULL) : MNULL;

    handle->fallback = NULL;
    handle->QR = B2;
    handle->B1upper = cw ? sub_mat_copy(B1, 0, 0, cw - 1, B1->n - 1, MNULL) : MNULL;
    handle->dw = dw;
    handle->pw = pw;
    handle->pweight = pweight;
    handle->solvev = cv ? alloc_weighted_cls_ns(mys, C1) : NULL;
    handle->solvew = cw ? alloc_unweighted_cls(C2) : NULL;

    M_FREE(B1);
    M_FREE(C1);
    M_FREE(C2);
    M_FREE(R);
    V_FREE(mys);
    return handle;
}

VEC *solve_weighted_cls(WeightedCLS *handle, VEC *b, double *cost, VEC *x)
{
    if (handle->fallback)
        return solve_weighted_cls_ns(handle->fallback, b, cost, x);

    int cw, cv, nv, vars;
    VEC *qtb, *qtbv, *qtbw;
    VEC *v, *w, *B1v = VNULL;

    vars = handle->pweight->size;
    nv = vars - handle->pw->size;
    cw = handle->solvew ? handle->solvew->L->m : 0;
    cv = handle->solvev ? handle->solvev->L->m : 0;

    x = v_resize(x, vars);
    v = sub_vec(x, 0, nv - 1, VNULL);
    w = sub_vec(x, nv, vars - 1, VNULL);

    qtb = _Qsolve(handle->QR, handle->dw, b, VNULL, VNULL);
    qtbw = cw ? sub_vec(qtb, 0, cw - 1, VNULL) : VNULL;
    qtbv = cv ? sub_vec(qtb, cw, cw + cv - 1, VNULL) : VNULL;
    if (cv)
        solve_weighted_cls_ns(handle->solvev, qtbv, cost, v);
    else
    {
        v_zero(v);
        if (cost) *cost = 0.0;
    }
    if (cw)
    {
        B1v = mv_mlt(handle->B1upper, v, VNULL);
        v_sub(qtbw, B1v, qtbw);
        solve_unweighted_cls(handle->solvew, qtbw, NULL, w);
        pxinv_vec(handle->pw, w, w);
    }
    else
        v_zero(w);
    x = pxinv_vec(handle->pweight, x, x);

    v_free_sub(v);
    v_free_sub(w);
    v_free_sub(qtbv);
    v_free_sub(qtbw);
    V_FREE(qtb);
    V_FREE(B1v);
    return x;
}

void free_weighted_cls(WeightedCLS *handle)
{
    if (handle->fallback)
    {
        free_weighted_cls_ns(handle->fallback);
    }
    else
    {
        M_FREE(handle->QR);
        if (handle->B1upper) M_FREE(handle->B1upper);
        V_FREE(handle->dw);
        PX_FREE(handle->pw);
        PX_FREE(handle->pweight);
        if (handle->solvev) free_weighted_cls_ns(handle->solvev);
        if (handle->solvew) free_unweighted_cls(handle->solvew);
    }
    delete handle;
}

/*** Matrix-weighted constrained least squares with singular matrices ***/

struct MatrixCLS
{
    MAT *U;       //!< \f$P = USU^T\f$
    WeightedCLS *solver;
};

MatrixCLS *alloc_matrix_cls(MAT *P, MAT *C, double epss, double epsc)
{
    MatrixCLS *handle;
    MAT *U;
    MAT *CU;
    VEC *s;

    g_assert(P->m == P->n);
    g_assert(C->m <= C->n);

    /* Eigendecomposition of P */
    U = m_get(P->m, P->m);
    s = symmeig(P, U, VNULL);
    CU = m_mlt(C, U, MNULL);
    /* Need to make sure there are no marginally negative eigenvalues, to
     * keep alloc_weighted_cls happy.
     */
    g_assert(v_entry(s, 0) > 0.0);
    for (size_t i = 0; i < s->dim; i++)
    {
        g_assert(v_entry(s, i) > -1e-5 * v_entry(s, 0));
        if (v_entry(s, i) < 0.0) v_set_val(s, i, 0.0);
    }

    handle = new MatrixCLS;
    handle->U = U;
    handle->solver = alloc_weighted_cls(s, CU, epss, epsc);

    M_FREE(CU);
    V_FREE(s);
    return handle;
}

VEC *solve_matrix_cls(MatrixCLS *handle, VEC *b, double *cost, VEC *x)
{
    VEC *a;

    a = solve_weighted_cls(handle->solver, b, cost, VNULL);
    x = mv_mlt(handle->U, a, x);
    V_FREE(a);
    return x;
}

void free_matrix_cls(MatrixCLS *handle)
{
    M_FREE(handle->U);
    free_weighted_cls(handle->solver);
    delete handle;
}

/**************************************************************
 * Test code
 **************************************************************/

#if HAVE_CPPUNIT
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>
#include <sstream>
#include <vector>
#include <iterator>

class LinearTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(LinearTest);
    CPPUNIT_TEST(gram_schmidt);
    CPPUNIT_TEST(unweighted_cls);
    CPPUNIT_TEST(weighted_cls_ns);
    CPPUNIT_TEST(weighted_cls);
    CPPUNIT_TEST(matrix_cls);
    CPPUNIT_TEST_SUITE_END();

private:
    /* Randomly generated orthogonal matrix */
    static VEC *parse_vec(const char *str)
    {
        vector<double> t;
        istringstream in(str);
        copy(istream_iterator<double>(in), istream_iterator<double>(), back_inserter(t));
        VEC *out = v_get(t.size());
        copy(t.begin(), t.end(), out->ve);
        return out;
    }

    static MAT *parse_mat(const char *str)
    {
        vector<vector<double> > t;
        istringstream in(str);
        string line;
        size_t len = UINT_MAX;
        while (getline(in, line))
        {
            istringstream in2(line);
            t.push_back(vector<double>());
            copy(istream_iterator<double>(in2), istream_iterator<double>(), back_inserter(t.back()));
            g_assert(len == UINT_MAX || len == t.back().size());
            len = t.back().size();
        }
        MAT *out = m_get(t.size(), len);
        for (size_t i = 0; i < t.size(); i++)
            copy(t[i].begin(), t[i].end(), &out->me[i][0]);
        return out;
    }

    class DataSet
    {
    public:
        VEC *P;
        MAT *C;
        VEC *B;
        VEC *Y;

        DataSet(const char *p, const char *c, const char *b, const char *y) : P(parse_vec(p)), C(parse_mat(c)), B(parse_vec(b)), Y(parse_vec(y)) {}
        ~DataSet() { V_FREE(P); M_FREE(C); V_FREE(B); V_FREE(Y); }
    private:
        DataSet() {}
        DataSet(const DataSet &b) {}
        DataSet &operator =(const DataSet &b) { return *this; }
    };

    vector<DataSet *> data;
    MAT *U;

    void test_weighted_cls(VEC *s, MAT *C, VEC *b, VEC *yp)
    {
        WeightedCLS *handle;
        VEC *x;

        handle = alloc_weighted_cls(s, C);
        x = solve_weighted_cls(handle, b, NULL, VNULL);
        free_weighted_cls(handle);

        v_sub(x, yp, x);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(v_norm2(x), 0.0, 1e-6);

        V_FREE(x);
    }

    void test_matrix_cls(MAT *U, VEC *Pd, MAT *C, VEC *b, VEC *yp)
    {
        MAT *tmp = MNULL;
        VEC *x = VNULL, *y = VNULL;
        MatrixCLS *handle;

        MAT *P = m_get(U->m, U->m);
        for (size_t i = 0; i < P->m; i++)
            P->me[i][i] = v_entry(Pd, i);

        tmp = mmtr_mlt(P, U, tmp);
        P = m_mlt(U, tmp, P);
        tmp = mmtr_mlt(C, U, tmp);
        C = m_copy(tmp, C);
        M_FREE(tmp);

        handle = alloc_matrix_cls(P, C);
        M_FREE(P);
        x = solve_matrix_cls(handle, b, NULL, VNULL);
        free_matrix_cls(handle);

        y = vm_mlt(U, x, VNULL);
        v_sub(y, yp, y);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(v_norm2(y), 0.0, 1e-6);

        V_FREE(x);
        V_FREE(y);
    }

public:
    void setUp()
    {
        /* Random unitary matrix */
        U = parse_mat(
                      "  0.4860627737330025  -0.0515547265107234  -0.6541718167269790   0.1589184521690312  -0.5548776891583470\n"
                      "  0.2720554225715700   0.3827554844458788   0.0955002494212027   0.8157894684974557   0.3238075533787505\n"
                      "  0.5353387089327772  -0.5692096463704388  -0.1000771838640259  -0.1380193350049984   0.6002899846171442\n"
                      "  0.3389195613117070  -0.3454756364489906   0.7189644062230486   0.1533298024948445  -0.4747212548766653\n"
                      "  0.5369123063581714   0.6383473289675533   0.1897726967052870  -0.5164040118749159   0.0393837909022721\n"
                     );

        data.resize(6);
        /* C lies entirely within span of P */
        data[0] = new DataSet("3 2 1 1 0",

                              " 1  0  0  0  0\n"
                              " 0  1  1  1  0\n"
                              " 0  0  1 -1  0\n",

                              "2 1 1",

                              "2 0.2 0.9 -0.1 0");

        /* C lies entirely in null-space of P */
        data[1] = new DataSet("3 2 1 0 0",

                              "0 0 0 1 2\n",

                              "3",

                              "0 0 0 0.6 1.2");

        /* Similar to 0, but C has terms outside of the span that do not affect
         * the impact on the span.
         */
        data[2] = new DataSet("3 2 1 1 0",

                              " 1  0  0  0  0\n"
                              " 0  1  1  1  0\n"
                              " 0  0  1 -1  1\n"
                              " 1  0  0  0 -1\n",

                              "2 1 2 1",

                              "2 0.2 0.9 -0.1 1");

        /* Similar to 1, but with non-zeros in the span */
        data[3] = new DataSet("3 2 1 0 0",

                              "1 1 1 1 2\n",

                              "3",

                              "0 0 0 0.6 1.2");

        /* Like 0, but with the variables permuted */
        data[4] = new DataSet("1 3 0 2 1",

                              " 0  1  0  0  0\n"
                              " 1  0  0  1  1\n"
                              "-1  0  0  0  1\n",

                              "2 1 1",

                              "-0.1 2 0 0.2 0.9");

        /* Non-singular case */
        data[5] = new DataSet("1 1 1 1 1",

                              " 1  1  1  0  0\n"
                              " 0  0  0  1  1\n"
                              " 1  0  0  1  0\n",

                              "7 7 7",

                              "3 2 2 4 3");
    }

    void tearDown()
    {
        for (size_t i = 0; i < data.size(); i++)
            delete data[i];
        M_FREE(U);
    }

    void gram_schmidt()
    {
        unsigned int r = 5;
        unsigned int c = 10;
        MAT *C = MNULL, *Q = MNULL, *L = MNULL;
        MAT *tmp = MNULL;

        C = m_get(r, c);
        C = m_rand(C);
        Q = m_get(r, c);
        L = m_get(r, r);
        ::gram_schmidt(C, L, Q);
        CPPUNIT_ASSERT_EQUAL(L->m, r);
        CPPUNIT_ASSERT_EQUAL(L->n, r);
        CPPUNIT_ASSERT_EQUAL(Q->m, r);
        CPPUNIT_ASSERT_EQUAL(Q->n, c);
        for (unsigned int i = 0; i < r; i++)
            for (unsigned int j = i + 1; j < r; j++)
                CPPUNIT_ASSERT_EQUAL(m_entry(L, i, j), 0.0);
        tmp = m_mlt(L, C, tmp);
        tmp = m_sub(tmp, Q, tmp);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(m_norm_frob(tmp), 0.0, 1e-6);

        tmp = mmtr_mlt(Q, Q, tmp);
        for (unsigned int i = 0; i < tmp->m; i++)
            m_sub_val(tmp, i, i, 1.0);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(m_norm_frob(tmp), 0.0, 1e-6);

        M_FREE(C);
        M_FREE(Q);
        M_FREE(L);
        M_FREE(tmp);
    }

    void unweighted_cls()
    {
        UnweightedCLS *handle;

        VEC *x, *b;
        MAT *C;
        double cost;
        double out[6] = {0.0, 1.0, 8.0, 10.0, 10.0, -4.0};

        C = m_get(4, 6);
        m_zero(C);
        m_set_val(C, 0, 0, 1.0);
        m_set_val(C, 1, 1, 1.0);
        m_set_val(C, 2, 2, 1.0);
        m_set_val(C, 2, 3, 1.0);
        m_set_val(C, 2, 4, 1.0);
        m_set_val(C, 3, 2, 1.0);
        m_set_val(C, 3, 5, 2.0);

        b = parse_vec("0.0 1.0 28.0 0.0");

        handle = alloc_unweighted_cls(C);
        x = solve_unweighted_cls(handle, b, &cost, VNULL);
        free_unweighted_cls(handle);

        for (int i = 0; i < 6; i++)
            CPPUNIT_ASSERT_DOUBLES_EQUAL(v_entry(x, i), out[i], 1e-9);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(cost, 281.0, 1e-6);

        V_FREE(x);
        V_FREE(b);
        M_FREE(C);
    }

    void weighted_cls_ns()
    {
        WeightedCLSNS *handle;
        VEC *s, *x, *b;
        MAT *C;
        double cost;
        double out[5] = {0.0, 1.0, 3.0, 2.0, 1.0};

        s = parse_vec("1.0 1.0 2.0 3.0 6.0");

        C = m_get(3, 5);
        m_zero(C);
        m_set_val(C, 0, 0, 1.0);
        m_set_val(C, 1, 1, 1.0);
        m_set_val(C, 2, 2, 1.0);
        m_set_val(C, 2, 3, 1.0);
        m_set_val(C, 2, 4, 1.0);

        b = parse_vec("0.0 1.0 6.0");

        handle = alloc_weighted_cls_ns(s, C);
        x = solve_weighted_cls_ns(handle, b, &cost, VNULL);
        free_weighted_cls_ns(handle);
        for (int i = 0; i < 5; i++)
            CPPUNIT_ASSERT_DOUBLES_EQUAL(v_entry(x, i), out[i], 1e-9);
        CPPUNIT_ASSERT_DOUBLES_EQUAL(cost, 37.0, 1e-6);

        V_FREE(s);
        V_FREE(b);
        V_FREE(x);
        M_FREE(C);
    }

    void weighted_cls()
    {
        for (size_t i = 0; i < data.size(); i++)
            test_weighted_cls(data[i]->P, data[i]->C, data[i]->B, data[i]->Y);
    }

    void matrix_cls()
    {
        for (size_t i = 0; i < data.size(); i++)
            test_matrix_cls(U, data[i]->P, data[i]->C, data[i]->B, data[i]->Y);
    }
};

CPPUNIT_TEST_SUITE_REGISTRATION(LinearTest);

#endif /* HAVE_CPPUNIT */

/*! @} */
