/*! \addtogroup geom
 * @{
 */

/*! \file
 * \brief Various linear algebra tools built in meschach
 */

#ifndef ALOD_GEOM_LINEAR_H
#define ALOD_GEOM_LINEAR_H

#include "common/mesch.h"

//! Dumps a matrix to \c stdout in a format suitable for Octave
void show_matrix(const char *name, MAT *matrix);

//! Dumps a vector to \c stdout in a format suitable for Octave
void show_vector(const char *name, VEC *vec);

//! Factorises \a C as \f$ C = L^{-1}Q \f$.
void gram_schmidt(MAT *C, MAT *L, MAT *Q);

/*** Unweighted constrained least squares with non-singular matrices **/

struct UnweightedCLS;

//! Prepares a handle for an unweighted constrained least-squares problem.
/*! The problem has the form \f$ \sum x_i^2\f$ subject to the constraint
 * \f$Cx = b\f$.
 * \pre \a C has full rank and at least as many columns as rows
 */

UnweightedCLS *alloc_unweighted_cls(MAT *C);

//! Solves an unweighted constrained least-squares problem.
VEC *solve_unweighted_cls(UnweightedCLS *handle, VEC *b, double *cost, VEC *x);

//! Cleans up a handle returned by \c alloc_unweighted_cls_ns
void free_unweighted_cls(UnweightedCLS *handle);

/*** Weighted constrained least squares with non-singular weights ***/

struct WeightedCLSNS;

//! Prepares a handle for a weighted constrained least-squares problem.
/*! The problem has the form \f$ \sum s_i x_i^2\f$ subject to
 * the constraint \f$ Cx = b\f$.
 * \pre \a s has strictly positive elements.
 * \pre \a C has full rank and at least as many columns as rows
 */
WeightedCLSNS *alloc_weighted_cls_ns(VEC *s, MAT *C);

//! Solves a weighted constrained least-squares problem.
VEC *solve_weighted_cls_ns(WeightedCLSNS *handle, VEC *b, double *cost, VEC *x);

//! Cleans up a handle returned by \c alloc_weighted_cls_ns
void free_weighted_cls_ns(WeightedCLSNS *handle);

/*** Weighted constrained least squares with possibly singular weights ***/

struct WeightedCLS;

//! Prepares a handle for a weighted constrained least-squares problem.
/*! The problem has the same form as for \c alloc_weighted_cls_ns, except
 * that \a s may have zero elements.
 */
WeightedCLS *alloc_weighted_cls(VEC *s, MAT *C, double epss = 1e-6, double epsc = 1e-3);

//! Solves a weighted constrained least-squares problem
VEC *solve_weighted_cls(WeightedCLS *handle, VEC *b, double *cost, VEC *x);

//! Cleans up a handle returned by \c alloc_weighted_cls
void free_weighted_cls(WeightedCLS *handle);

/*** Matrix-weighted constrained least squares with singular matrices ***/

struct MatrixCLS;

//! Prepares a handle for a matrix-weighted constrained least squares problem.
/*! The problem minimises \f$x^TPx\f$ subject to the constraint \f$Cx = b\f$.
 * \pre P is non-negative definite and symmetric
 * \pre \a C has full rank and at least as many columns as rows
 */
MatrixCLS *alloc_matrix_cls(MAT *P, MAT *C, double epss = 1e-6, double epsc = 1e-3);

//! Solves a matrix-weighted constrained least squares problem.
VEC *solve_matrix_cls(MatrixCLS *handle, VEC *b, double *cost, VEC *x);

//! Cleans up a handle returned by \c alloc_matrix_cls
void free_matrix_cls(MatrixCLS *handle);

#endif /* !ALOD_GEOM_LINEAR_H */

/*! @} */
