#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "geom/plane.h"
#include "geom/skeleton.h"
#include "geom/sparse.h"
#include "geom/anim.h"
#include "common/mesch.h"
#include <vector>

using namespace std;

TangentPlane operator *(double s, const TangentPlane &b)
{
    return TangentPlane(s * b.tangents[0], s * b.tangents[1]);
}

TangentPlane fit_tangent_plane(const Skeleton &skeleton,
                               vector<TangentPlane> &planes)
{
    TangentPlane ans;

    size_t sz = skeleton.size();
    int compress[sz];   // maps bone indices to matrix indices
    int decompress[sz]; // maps matrix indices to bone indices
    int ncompress;

    MAT *A = MNULL, *B = MNULL, *C = MNULL, *U = MNULL, *V = MNULL;
    VEC *sing = VNULL, *t[2] = { VNULL, VNULL }, *tmp = VNULL;
    const VectorPosition zero(skeleton);

    /* Identify the relevant bones. We initially just flag them as
     * seen or unseen, and assign the matrix indices after sorting.
     */
    ncompress = 0;
    fill(compress, compress + sz, -1);
    for (size_t i = 0; i < planes.size(); i++)
        for (int j = 0; j < 2; j++)
        {
            const VectorPosition &t = planes[i].tangents[j];
            for (VectorPosition::const_iterator k = t.begin(); k != t.end(); k++)
                if (compress[k->index / 4] == -1)
                {
                    compress[k->index / 4] = 0;
                    decompress[ncompress++] = k->index / 4;
                }
        }
    if (!ncompress)  // No non-zero planes. Return a zero plane
    {
        ans.tangents[0] = VectorPosition(skeleton);
        ans.tangents[1] = VectorPosition(skeleton);
        return ans;
    }
    sort(decompress, decompress + ncompress);
    for (int i = 0; i < ncompress; i++)
        compress[decompress[i]] = i;

    /* Set up the matrix */
    A = m_resize(A, 4 * ncompress, 4 * ncompress);
    m_zero(A);
    for (size_t i = 0; i < planes.size(); i++)
    {
        const VectorPosition &t1 = planes[i].tangents[0];
        const VectorPosition &t2 = planes[i].tangents[1];

        VectorPosition::const_iterator t1j = t1.begin();
        VectorPosition::const_iterator t2j = t2.begin();
        for (int j = 0; j < ncompress; j++)
        {
            size_t js = decompress[j] * 4;
            while (t1j != t1.end() && t1j->index < js) t1j++;
            while (t2j != t2.end() && t2j->index < js) t2j++;
            bool t1j_valid = (t1j != t1.end() && t1j->index == js);
            bool t2j_valid = (t2j != t2.end() && t2j->index == js);
            VectorPosition::const_iterator t1k = t1.begin();
            VectorPosition::const_iterator t2k = t2.begin();
            for (int k = 0; k < ncompress; k++)
            {
                size_t ks = decompress[k] * 4;
                while (t1k != t1.end() && t1k->index < ks) t1k++;
                while (t2k != t2.end() && t2k->index < ks) t2k++;
                bool t1k_valid = (t1k != t1.end() && t1k->index == ks);
                bool t2k_valid = (t2k != t2.end() && t2k->index == ks);
                for (int l = 0; l < 4; l++)
                    for (int m = 0; m < 4; m++)
                    {
                        double t1l = t1j_valid ? t1j->values[l] : 0.0;
                        double t1m = t1k_valid ? t1k->values[m] : 0.0;
                        double t2l = t2j_valid ? t2j->values[l] : 0.0;
                        double t2m = t2k_valid ? t2k->values[m] : 0.0;
                        m_add_val(A, j * 4 + l, k * 4 + m,
                                  t1l * t2m - t2l * t1m);
                        // svd can go into an infinite loop if this does not hold
                        g_assert(isfinite(m_entry(A, j * 4 + l, k * 4 + m)));
                    }
            }
        }
    }

    /* Extract the relevant components of C */
    C = m_resize(C, 4 * ncompress, 4 * ncompress);
    for (int i = 0; i < ncompress; i++)
        for (int j = 0; j < ncompress; j++)
        {
            int i2 = decompress[i] * 4;
            int j2 = decompress[j] * 4;
            for (int k = 0; k < 4; k++)
                for (int l = 0; l < 4; l++)
                    m_set_val(C, i * 4 + k, j * 4 + l,
                              m_entry(skeleton.get_C(), i2 + k, j2 + l));
        }

    /* Compute B = CAC^T, using V as a temporary */
    V = mmtr_mlt(A, C, V);
    B = m_mlt(C, V, B);

    /* Solve the system */
    U = m_resize(U, B->n, B->n);
    V = m_resize(V, B->n, B->n);
    sing = svd(B, U, V, sing);

    /* Extract solutions. Unlike Octave, meschach seems to return
     * right singular vectors in the ROWS of V. We use sing as
     * temporary.
     */
    double s = v_entry(sing, 0);
    tmp = v_resize(tmp, V->n);
    tmp = get_row(V, 0, tmp);
    sing = vm_mlt(C, tmp, sing);
    t[0] = mv_mlt(A, sing, t[0]);
    t[0] = sv_mlt(1.0 / s, t[0], t[0]);

    tmp = get_row(U, 0, tmp);
    sing = vm_mlt(C, tmp, sing);
    t[1] = vm_mlt(A, sing, t[1]);
    t[1] = sv_mlt(1.0 / s, t[1], t[1]);

    /* Copy to the answer */
    for (int i = 0; i < 2; i++)
    {
        ans.tangents[i] = VectorPosition(skeleton);
        for (int j = 0; j < ncompress; j++)
        {
            int j2 = 4 * decompress[j];
            for (int k = 0; k < 4; k++)
            {
                double e = v_entry(t[i], 4 * j + k);
                if (fabs(e) > 1e-9)
                    ans.tangents[i][j2 + k] = e;
            }
        }
        ans.tangents[i] = normalise(ans.tangents[i]);
    }

    /* Canonicalise, based on the longest projection of any axis into the plane */
    double best_proj = -1.0;
    int best = 0;
    /* Making const copies prevents [] from introducing elements */
    const VectorPosition t0 = ans.tangents[0];
    const VectorPosition t1 = ans.tangents[1];
    for (int i = 0; i < ncompress; i++)
        for (int j = 0; j < 4; j++)
        {
            int a = 4 * decompress[i] + j;
            double proj = sqrt(t0[a] * t0[a] + t1[a] * t1[a]);
            if (proj > best_proj)
            {
                best_proj = proj;
                best = a;
            }
        }

    double cs = t0[best] / best_proj;
    double sn = t1[best] / best_proj;
    ans.tangents[0] = cs * t0 + sn * t1;
    ans.tangents[1] = -sn * t0 + cs * t1;

    V_FREE(sing);
    V_FREE(t[0]);
    V_FREE(t[1]);
    V_FREE(tmp);
    M_FREE(A);
    M_FREE(B);
    M_FREE(C);
    M_FREE(U);
    M_FREE(V);
    return ans;
}


/**************************************************************
 * Test code
 **************************************************************/

#if HAVE_CPPUNIT
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>

class PlaneTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(PlaneTest);
    CPPUNIT_TEST(test3d);
    CPPUNIT_TEST(test2b);
    CPPUNIT_TEST_SUITE_END();
public:
    void test3d()
    {
        int R = 50;
        Skeleton skeleton;

        for (int r = 0; r < R; r++)
        {
            int N = 6;
            Vector3 normals[N];
            Vector3 normals_fit(0.0, 0.0, 0.0);
            vector<TangentPlane> tp(N, TangentPlane(skeleton));
            TangentPlane tp_fit;
            Vector3 compare;

            for (int i = 0; i < N; i++)
            {
                for (int j = 0; j < 3; j++)
                    normals[i][j] = rand();
                normals[i] = normalise(normals[i]);
                normals_fit += normals[i];

                int best_axis = 0;
                for (int j = 1; j < 3; j++)
                    if (fabs(normals[i][j]) < fabs(normals[i][best_axis]))
                        best_axis = j;
                Vector3 b(0.0, 0.0, 0.0);
                b[best_axis] = 1.0;
                Vector3 p = normalise(cross(normals[i], b));
                Vector3 q = normalise(cross(normals[i], p));
                for (int j = 0; j < 3; j++)
                {
                    tp[i].tangents[0][j] = p[j];
                    tp[i].tangents[1][j] = q[j];
                }
            }
            normals_fit = normalise(normals_fit);
            tp_fit = fit_tangent_plane(skeleton, tp);
            compare = normalise(cross(skeleton.ssd_to_3d(tp_fit.tangents[0]),
                                      skeleton.ssd_to_3d(tp_fit.tangents[1])));
            CPPUNIT_ASSERT(length(compare - normals_fit) < 1e-8);
        }
    }

    void test2b()
    {
        Skeleton skeleton;
        vector<SkeletonBoneInfo> bones(2);
        double len = 10.0;

        bones[0].parent = -1;
        bones[0].flags = BONE_FLAGS_FIXED_UNIFORM;
        bones[0].max_singular = 1.0;
        bones[0].translate_bounds.centre.assign(0.0, 0.0, 0.0);
        bones[0].translate_bounds.radius = 0.0;

        bones[1].parent = 0;
        bones[1].flags = BONE_FLAG_UNIFORM_SCALE | BONE_FLAG_FIXED_SCALE;
        bones[1].max_singular = 1.0;
        bones[1].translate_bounds.centre.assign(0.5 * len, 0.0, 0.0);
        bones[1].translate_bounds.radius = 0.0;
        bones[1].rest_transform.identity();
        bones[1].rest_transform[0][3] = len * 0.5;
        skeleton.set_bones(bones);

        int P = 6;
        vector<TangentPlane> tp(P, TangentPlane(skeleton));
        vector<TangentPlane> fits(P, TangentPlane(skeleton));
        for (int i = 0; i < P; i++)
        {
            for (int j = 0; j < 2; j++)
            {
                for (int k = 0; k < 7; k++)
                    tp[i].tangents[j][k] = rand() / (double) RAND_MAX;
                tp[i].tangents[j][7] = -tp[i].tangents[j][3];
                tp[i].tangents[j] = normalise(tp[i].tangents[j]);
            }
        }

        for (int i = 0; i < P; i++)
        {
            fits[i] = fit_tangent_plane(skeleton, tp);
            random_shuffle(tp.begin(), tp.end());
        }
        for (int i = 0; i + 1 < P; i++)
            for (int j = i + 1; j < P; j++)
            {
                double corr = (fits[i].tangents[0] * fits[j].tangents[0]) * (fits[i].tangents[1] * fits[j].tangents[1])
                    - (fits[i].tangents[0] * fits[j].tangents[1]) * (fits[i].tangents[1] * fits[j].tangents[0]);
                CPPUNIT_ASSERT_DOUBLES_EQUAL(corr, 1.0, 1e-8);
            }
    }
};

CPPUNIT_TEST_SUITE_REGISTRATION(PlaneTest);

#endif /* HAVE_CPPUNIT */
