#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "skeleton.h"
#include "sparse.h"
#include "coords.h"
#include "bounding.h"
#include "common/mesch.h"
#include "mesh/mesh.h"
#include <cstddef>
#include <vector>
#include <stack>

using namespace std;

Skeleton::MetricMode Skeleton::metric_mode = Skeleton::MODE_STATISTICAL;

void Skeleton::set_metric_mode(Skeleton::MetricMode mode)
{
    metric_mode = mode;
}

SkeletonBoneInfo &SkeletonBoneInfo::operator =(const SkeletonBone &bone)
{
    parent = bone.parent;
    rest_transform = bone.rest_transform;
    flags = bone.flags;
    tail = bone.tail;
    max_singular = bone.max_singular;
    translate_bounds = bone.translate_bounds;
    expect_L = bone.expect_L;
    expect_GTG = bone.expect_GTG;
    expect_LTGTGL = bone.expect_LTGTGL;
    return *this;
}

Skeleton::Skeleton(const Skeleton &s)
{
    frame_rate = s.frame_rate;
    bones = s.bones;
    frames = s.frames;
    custom_P = s.custom_P;
    if (s.P) P = m_copy(s.P, MNULL); else P = MNULL;
    if (s.C) C = m_copy(s.C, MNULL); else C = MNULL;
    if (s.M) M = m_copy(s.M, MNULL); else M = MNULL;
    if (P) ssd_inner.set_matrix(P);
}

Skeleton &Skeleton::operator =(const Skeleton &s)
{
    bones = s.bones;
    frames = s.frames;
    frame_rate = s.frame_rate;
    custom_P = s.custom_P;
    if (s.P) P = m_copy(s.P, P); else if (P) M_FREE(P);
    if (s.C) C = m_copy(s.C, C); else if (C) M_FREE(C);
    if (s.M) M = m_copy(s.M, M); else if (M) M_FREE(M);
    ssd_inner.set_matrix(P);
    return *this;
}

Skeleton::Skeleton(const std::vector<SkeletonBoneInfo> &b)
: P(MNULL), C(MNULL), M(MNULL), frame_rate(30.0), custom_P(false)
{
    set_bones(b);
}

Skeleton::Skeleton() : P(MNULL), C(MNULL), M(MNULL), frame_rate(30.0), custom_P(false)
{
    vector<SkeletonBoneInfo> b;

    b.push_back(SkeletonBoneInfo());
    b[0].parent = -1;
    b[0].flags = BONE_FLAGS_FIXED_UNIFORM;
    b[0].tail = Vector3(0.0, 1.0, 0.0);
    b[0].max_singular = 1.0;
    b[0].translate_bounds.centre.assign(0.0, 0.0, 0.0);
    b[0].translate_bounds.radius = 0.0;
    b[0].rest_transform = Matrix4();

    set_bones(b);
}

void Skeleton::update_G()
{
    bones[0].G = Matrix4();
    for (size_t i = 1; i < size(); i++)
        bones[i].G = bones[bones[i].parent].G * bones[i].L;
}

void Skeleton::update_M(std::size_t bone)
{
    for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++)
            m_set_val(M, i, 4 * bone + j, bones[bone].G[i][j]);
}

void Skeleton::update_M()
{
    M = m_resize(M, 4, 4 * size());
    for (size_t i = 0; i < size(); i++)
        update_M(i);
}

void Skeleton::update_P()
{
    if (custom_P) return;

    size_t sz = size();
    Matrix4 expect1[sz];      // E[L_j]
    vector<vector<Matrix4> > expect2(sz, vector<Matrix4>(sz)); // E[G_i^TG_j]
    // Note: the above is a vector not an array to avoid stack overflows

    if (metric_mode == MODE_3D)
        P = mtrm_mlt(M, M, P);
    else
    {
        P = m_resize(P, 4 * sz, 4 * sz);
        for (size_t i = 0; i < sz; i++)
        {
            Matrix3 Ll; // expected linear part of L
            Vector3 Lt; // expected translation part of L
            guint32 flags = bones[i].flags;
            if (metric_mode == MODE_ANIMATION)
                flags &= BONE_FLAGS_FIXED_UNIFORM;

            if (flags & BONE_FLAG_EXPECT_L_LINEAR)
                Ll = matrix4_to_matrix3(bones[i].expect_L);
            else if (flags & BONE_FLAG_FIXED_ROTATE)
            {
                // Only valid if there is no variable scaling, but it's a
                // better guess than zero if not.
                Ll = matrix4_to_matrix3(bones[i].L);
            }
            else
                Ll.zero();

            if (flags & BONE_FLAG_EXPECT_L_TRANSLATE)
                Lt = matrix4_to_translate3(bones[i].expect_L);
            else
                Lt = matrix4_to_translate3(bones[i].rest_transform);

            expect1[i] = matrix3_translate3_to_matrix4(Ll, Lt);
            bones[i].expect_L = expect1[i];

            // put in identity so long, then fix what we can
            expect2[i][i].identity();
            if (i != 0)
            {
                int p = bones[i].parent;
                Matrix3 pGTGl = matrix4_to_matrix3(expect2[p][p]);
                Vector3 pGTGt = matrix4_to_translate3(expect2[p][p]);
                double pGTGw = expect2[p][p][3][3];
                Matrix3 GTGl;
                Vector3 GTGt;
                double GTGw;

                if (flags & BONE_FLAG_EXPECT_GTG_LINEAR)
                    GTGl = matrix4_to_matrix3(bones[i].expect_GTG);
                else if (flags & BONE_FLAG_FIXED_ROTATE)
                {
                    // Only valid if there is no variable scaling, but it's
                    // a better guess than the alternative if not.
                    GTGl = transpose(Ll) * pGTGl * Ll;
                }
                else
                {
                    GTGl.identity();
                    /* Ideally should solve for expected value of the linear
                     * part of L^TL, but we don't have the flags for that.
                     * This also fails if the rotations are non-uniform
                     * AND there is non-uniform scaling in an ancestor.
                     */
                    Matrix4 tmp = transpose(bones[i].rest_transform) * bones[i].rest_transform;
                    double s = (tmp[0][0] + tmp[1][1] + tmp[2][2]) / 3.0;
                    GTGl *= s * (pGTGl[0][0] + pGTGl[1][1] + pGTGl[2][2]) / 3.0;
                }
                if (flags & BONE_FLAG_EXPECT_GTG_TRANSLATE)
                    GTGt = matrix4_to_translate3(bones[i].expect_GTG);
                else
                    GTGt = transpose(Ll) * (pGTGt + pGTGl * Lt);

                if (flags & BONE_FLAG_EXPECT_GTG_WEIGHT)
                    GTGw = bones[i].expect_GTG[3][3];
                else
                {
                    /* The formula below assumes translate is independent
                     * of linear, or that scaling is always uniform.
                     * If neither is true then there may be insufficient
                     * information anyway.
                     */
                    double LTGTGL;
                    if (flags & BONE_FLAG_EXPECT_LTGTGL)
                        LTGTGL = bones[i].expect_LTGTGL;
                    else
                        bones[i].expect_LTGTGL = LTGTGL = Lt * (GTGl * Lt);
                    GTGw = pGTGw + 2.0 * (pGTGt * Lt) + LTGTGL;
                }

                for (int j = 0; j < 3; j++)
                    for (int k = 0; k < 3; k++)
                        expect2[i][i][j][k] = GTGl[j][k];
                for (int j = 0; j < 3; j++)
                {
                    expect2[i][i][j][3] = GTGt[j];
                    expect2[i][i][3][j] = GTGt[j];
                }
                expect2[i][i][3][3] = GTGw;

                bones[i].expect_GTG = expect2[i][i];
            }
        }
        for (size_t i = 0; i + 1 < sz; i++)
            for (size_t j = i + 1; j < sz; j++)
            {
                int p = bones[j].parent;
                expect2[i][j] = expect2[i][p] * expect1[j];
                expect2[j][i] = transpose(expect2[i][j]);
            }

        for (size_t i = 0; i < sz; i++)
            for (size_t j = 0; j < sz; j++)
                for (int k = 0; k < 4; k++)
                    for (int l = 0; l < 4; l++)
                        m_set_val(P, i * 4 + k, j * 4 + l,
                                  expect2[i][j][k][l]);
    }

    ssd_inner.set_matrix(P);
}

/* A Cholesky factorisation would probably be favourite, but there are
 * numerical issues when applying it to matrices that are only
 * semi-definite. Instead we compute the square root from the SVD.
 * We also recompute P by setting negative eigenvalues (which should
 * only arise from numerical instability) to zero.
 */
void Skeleton::update_C()
{
    VEC *eig = VNULL;
    MAT *ev = MNULL;

    g_assert(P);
    C = m_resize(C, P->n, P->n);
    ev = m_get(P->n, P->n);
    eig = symmeig(P, ev, eig);
    m_zero(C);
    m_zero(P);
    g_assert(v_entry(eig, 0) > 0.0);
    for (size_t i = 0; i < P->n; i++)
    {
        double e = v_entry(eig, i);
        double ep = e;
        g_assert(e > -1e-3 * v_entry(eig, 0));
        if (e < 0.0) e = ep = 0.0; else e = sqrt(e);
        for (size_t j = 0; j < P->n; j++)
            for (size_t k = 0; k < P->n; k++)
            {
                m_add_val(C, j, k, e * m_entry(ev, j, i) * m_entry(ev, k, i));
                m_add_val(P, j, k, ep * m_entry(ev, j, i) * m_entry(ev, k, i));
            }
    }

    M_FREE(ev);
    V_FREE(eig);
}

void Skeleton::set_P(MAT *new_P)
{
    g_assert(new_P);

    if (P) M_FREE(P);
    P = new_P;
    ssd_inner.set_matrix(P);
    update_C();
    custom_P = true;
}

void Skeleton::load_bones(const std::vector<SkeletonBoneInfo> &b)
{
    const size_t sz = b.size();
    size_t roots = 0;

    for (size_t i = 0; i < sz; i++)
        if (b[i].parent == -1) roots++;
    if (roots != 1) g_error("Must be exactly one root");

    /* If the bones are already in topological order, then we don't
     * mess with them. This ensures that the sorting is idempotent.
     */
    bool needs_sort = false;
    for (size_t i = 0; i < sz; i++)
        if (b[i].parent >= (int) i) needs_sort = true;

    bones.resize(sz);
    if (needs_sort)
    {
        int valence[sz];
        int new_index[sz];
        stack<int> st;
        int pos = sz;

        fill(valence, valence + sz, 0);
        for (size_t i = 0; i < sz; i++)
            if (b[i].parent != -1) valence[b[i].parent]++;
        for (size_t i = 0; i < sz; i++)
            if (valence[i] == 0) st.push(i);

        while (!st.empty())
        {
            int cur = st.top();
            st.pop();

            pos--;
            bones[pos].parent = -1; // to be fixed later
            bones[pos].original_index = cur;
            bones[pos].flags = b[cur].flags;
            bones[pos].tail = b[cur].tail;

            bones[pos].max_singular = b[cur].max_singular;
            bones[pos].translate_bounds = b[cur].translate_bounds;
            bones[pos].expect_L = b[cur].expect_L;
            bones[pos].expect_GTG = b[cur].expect_GTG;
            bones[pos].expect_LTGTGL = b[cur].expect_LTGTGL;

            bones[pos].rest_transform = b[cur].rest_transform;
            bones[pos].L = b[cur].rest_transform;

            new_index[cur] = pos;

            if (b[cur].parent != -1)
            {
                valence[b[cur].parent]--;
                if (valence[b[cur].parent] == 0)
                    st.push(b[cur].parent);
            }
        }

        if (pos != 0) g_error("Cycle in skeleton");

        // fix up the parent and new_index pointers
        for (size_t i = 0; i < sz; i++)
        {
            if (b[bones[i].original_index].parent != -1)
                bones[i].parent = new_index[b[bones[i].original_index].parent];
            bones[bones[i].original_index].new_index = i;
        }
    }
    else
    {
        for (size_t i = 0; i < sz; i++)
        {
            bones[i].parent = b[i].parent;
            bones[i].original_index = i;
            bones[i].new_index = i;
            bones[i].flags = b[i].flags;
            bones[i].tail = b[i].tail;

            bones[i].max_singular = b[i].max_singular;
            bones[i].translate_bounds = b[i].translate_bounds;
            bones[i].expect_L = b[i].expect_L;
            bones[i].expect_GTG = b[i].expect_GTG;
            bones[i].expect_LTGTGL = b[i].expect_LTGTGL;

            bones[i].rest_transform = b[i].rest_transform;
            bones[i].L = b[i].rest_transform;
        }
    }
}

void Skeleton::set_bones(const std::vector<SkeletonBoneInfo> &b)
{
    load_bones(b);
    update_G();
    update_M();
    update_P();
    update_C();
    frames.clear();
}

void Skeleton::set_transform(std::size_t bone, const Matrix4 &m)
{
    g_assert(bone < size());
    bones[bone].L = m;
    update_G();
    // Could update just the children, but this function isn't used at
    // the moment so I haven't bothered
    update_M(); 
}

Skeleton::~Skeleton()
{
    if (P) M_FREE(P);
    if (C) M_FREE(C);
    if (M) M_FREE(M);
}

void Skeleton::clear()
{
    if (P) M_FREE(P);
    if (C) M_FREE(C);
    if (M) M_FREE(M);
    ssd_inner.set_matrix(MNULL);
    bones.clear();
}

Matrix4 Skeleton::get_absolute(std::size_t b) const
{
    g_assert(b <= size());
    return bones[b].G;
}

Vector3 Skeleton::ssd_to_3d(const Skeleton::SsdVector &s) const
{
    SparseVector<double, 4, DotInner> out(4, DotInner<double, 4>());
    matrix_vector(M, s, out);
    return Vector3(out[0], out[1], out[2]);
}

Vector3 Skeleton::ssd_to_3d_expected(const Skeleton::SsdVector &s) const
{
    Skeleton::SsdVector t = s;
    Vector4 in, out;

    for (size_t i = size() - 1; i > 0; i--)
    {
        for (int j = 0; j < 4; j++)
            in[j] = t[4 * i + j];
        out = bones[i].expect_L * in;
        for (int j = 0; j < 4; j++)
            t[4 * bones[i].parent + j] += out[j];
    }
    return Vector3(t[0], t[1], t[2]);
}

Vector3 Skeleton::ssd_to_normal(const TangentPlane &t) const
{
    const Vector3 t0 = ssd_to_3d(t.tangents[0]);
    const Vector3 t1 = ssd_to_3d(t.tangents[1]);
    return normalise(cross(t0, t1));
}

double Skeleton::inf_metric(const Skeleton::SsdVector &s) const
{
    size_t sz = size();
    BoundingSphereW spheres[sz];

    for (size_t i = 0; i < sz; i++)
    {
        for (int j = 0; j < 4; j++)
            spheres[i].centre[j] = s[4 * i + j];
        spheres[i].radius = 0.0;
    }
    for (size_t i = sz - 1; i > 0; i--)
    {
        int p = bones[i].parent;
        Matrix4 m = bones[i].rest_transform;
        m[0][3] = m[1][3] = m[2][3] = 0.0;    /* Linear part in a 4x4 matrix */
        if ((bones[i].flags & BONE_FLAGS_FIXED_LINEAR) == BONE_FLAGS_FIXED_LINEAR)
        {
            spheres[p] += (m * spheres[i]) * bones[i].translate_bounds;
        }
        else if ((bones[i].flags & BONE_FLAG_FIXED_ROTATE)
                 && (bones[i].flags & BONE_FLAG_UNIFORM_SCALE))
        {
            double s = matrix_norm2(matrix4_to_matrix3(m));
            BoundingSphereW up = (m * spheres[i]) * (bones[i].max_singular / s);
            /* If the centre is not covered, then a shrinking transform
             * could extend the sphere into an ice-cream cone. We cover this
             * by moving the centre to the midpoint of the main axis.
             */
            Vector3 c = vector4_to_vector3(up.centre);
            if (length(c) > up.radius && !(bones[i].flags & BONE_FLAG_FIXED_SCALE))
            {
                up.radius = 0.5 * (length(c) + up.radius);
                up.centre = vector3_to_vector4(up.radius * normalise(c), up.centre[3]);
            }
            up *= bones[i].max_singular;
            up *= bones[i].translate_bounds;
            spheres[p] += up;
        }
        else
        {
            Vector4 c(0.0, 0.0, 0.0, spheres[i].centre[3]);
            double r = inf_norm(spheres[i]);
            BoundingSphereW up(c, r);
            up *= bones[i].max_singular;
            up *= bones[i].translate_bounds;
            spheres[p] += up;
        }
    }
    return inf_norm(spheres[0]);
}

void Skeleton::set_to_frame(size_t frame_number)
{
    g_assert(frame_number >= 0 && frame_number < num_frames());
    const vector<Matrix4> &data = frames[frame_number];

    for (size_t i = 0; i < size(); i++)
        bones[i].L = data[i];
    update_G();
    update_M();
}

void Skeleton::set_to_rest()
{
    for (size_t i = 0; i < size(); i++)
        bones[i].L = bones[i].rest_transform;
    update_G();
    update_M();
}

const vector<Matrix4> &Skeleton::get_frame(size_t frame_number) const
{
    g_assert(frame_number >= 0 && frame_number < num_frames());
    return frames[frame_number];
}

void Skeleton::add_frame(const std::vector<Matrix4> &frame)
{
    g_assert(frame.size() == size());
    frames.push_back(frame);
}

void Skeleton::copy_frames(const Skeleton &reference)
{
    g_assert(size() == reference.size());
    frames = reference.frames;
    frame_rate = reference.frame_rate;
}

/*** Higher-level modifications ***/

void Skeleton::fix_tails()
{
    for (size_t i = 0; i < size(); i++)
        if (!bones[i].tail)
        {
            int c = 0;
            for (size_t j = 0; j < size(); j++)
                if (bones[j].parent == (int) i)
                    if (!frames.empty())
                        for (size_t f = 0; f < frames.size(); f++)
                        {
                            c++;
                            for (int k = 0; k < 3; k++)
                                bones[i].tail[k] += frames[f][j][k][3];
                        }
                    else
                    {
                        c++;
                        for (int k = 0; k < 3; k++)
                            bones[i].tail[k] += bones[j].rest_transform[k][3];
                    }
            if (c)
                bones[i].tail /= c;
            else
                bones[i].tail.assign(0.0, 1.0, 0.0);
        }
}

void Skeleton::clear_expect()
{
    for (size_t b = 0; b < size(); b++)
    {
        bones[b].flags &= ~(BONE_FLAG_EXPECT_L_LINEAR
                            | BONE_FLAG_EXPECT_L_TRANSLATE
                            | BONE_FLAG_EXPECT_GTG_LINEAR
                            | BONE_FLAG_EXPECT_GTG_TRANSLATE
                            | BONE_FLAG_EXPECT_GTG_WEIGHT
                            | BONE_FLAG_EXPECT_LTGTGL);

        bones[b].expect_L.zero();
        bones[b].expect_GTG.zero();
        bones[b].expect_LTGTGL = 0.0;
    }
    update_P();
    update_C();
}

void Skeleton::update_expect()
{
    if (!num_frames()) return;

    for (size_t b = 0; b < size(); b++)
    {
        bones[b].flags |=
            BONE_FLAG_EXPECT_L_LINEAR
            | BONE_FLAG_EXPECT_L_TRANSLATE
            | BONE_FLAG_EXPECT_GTG_LINEAR
            | BONE_FLAG_EXPECT_LTGTGL;
        bones[b].flags &= ~(BONE_FLAG_EXPECT_GTG_TRANSLATE
                            | BONE_FLAG_EXPECT_GTG_WEIGHT);
        bones[b].expect_L.zero();
        bones[b].expect_GTG.zero();
        bones[b].expect_LTGTGL = 0.0;
        for (size_t f = 0; f < num_frames(); f++)
        {
            set_to_frame(f);  /* FIXME: write a get_absolute_frame to avoid this */
            const Matrix4 &L = get_frame(f)[b];
            const Matrix4 &G = get_absolute(b);

            bones[b].expect_L += L;
            bones[b].expect_GTG += transpose(G) * G;
            double l = length(matrix4_to_matrix3(G) * matrix4_to_translate3(L));
            bones[b].expect_LTGTGL += l * l;
        }
        bones[b].expect_L *= 1.0 / num_frames();
        bones[b].expect_GTG *= 1.0 / num_frames();
        bones[b].expect_LTGTGL *= 1.0 / num_frames();
    }
    update_P();
    update_C();
}

void Skeleton::clear_bounds()
{
    for (size_t b = 0; b < size(); b++)
    {
        bones[b].max_singular = matrix_norm2(matrix4_to_matrix3(bones[b].rest_transform));
        bones[b].translate_bounds.centre = matrix4_to_translate3(bones[b].rest_transform);
        bones[b].translate_bounds.radius = 0.0;
    }
    update_P();
    update_C();
}

void Skeleton::update_bounds()
{
    if (!num_frames()) return;

    for (size_t b = 0; b < size(); b++)
    {
        bones[b].max_singular = 0.0;
        bones[b].translate_bounds.centre.zero();
        bones[b].translate_bounds.radius = 0.0;
        for (size_t f = 0; f < num_frames(); f++)
        {
            const Matrix4 &m = get_frame(f)[b];
            bones[b].translate_bounds.centre += matrix4_to_translate3(m);
            bones[b].max_singular = max(bones[b].max_singular, matrix_norm2(matrix4_to_matrix3(m)));
        }
        bones[b].translate_bounds.centre /= num_frames();
        for (size_t f = 0; f < num_frames(); f++)
        {
            const Matrix4 &m = get_frame(f)[b];
            bones[b].translate_bounds.radius = max(bones[b].translate_bounds.radius,
                                                   length(matrix4_to_translate3(m) - bones[b].translate_bounds.centre));
        }
    }
    update_P();
    update_C();
}

void Skeleton::update_flags()
{
    if (!num_frames()) return;  // not enough information to update

    bones[0].flags |= BONE_FLAGS_FIXED_UNIFORM;
    for (size_t b = 1; b < size(); b++)
    {
        const Matrix4 &rest = bones[b].rest_transform;
        int flags = BONE_FLAGS_FIXED_UNIFORM;

        double uscale = matrix_uniform_scale(matrix4_to_matrix3(rest));
        if (uscale < 0.0)
            flags &= ~BONE_FLAG_UNIFORM_SCALE;
        for (size_t f = 0; f < num_frames(); f++)
        {
            Matrix4 T = get_frame(f)[b];
            Matrix3 linear = matrix4_to_matrix3(T) * inverse(matrix4_to_matrix3(rest));
            Vector3 trans = matrix4_to_translate3(T) - matrix4_to_translate3(rest);
            if (length(trans) >= 1e-6 * (length(matrix4_to_translate3(rest)) + 1.0))
                flags &= ~BONE_FLAG_FIXED_TRANSLATE;
            if (!matrix_symmetric(linear))
                flags &= ~BONE_FLAG_FIXED_ROTATE;
            double uscale2 = matrix_uniform_scale(linear);
            if (uscale2 < 0.0)
                flags &= ~BONE_FLAG_UNIFORM_SCALE;
            if ((flags & BONE_FLAG_UNIFORM_SCALE) && fabs(uscale2 - uscale) > 1e-6 * uscale)
                flags &= ~BONE_FLAG_FIXED_SCALE;
            if ((flags & BONE_FLAG_FIXED_ROTATE)
                && matrix_norm2(linear - Matrix3()) > 1e-6 * matrix_norm2(linear))
                flags &= ~BONE_FLAG_FIXED_SCALE;
        }
        if ((flags & BONE_FLAGS_FIXED_UNIFORM_LINEAR) == BONE_FLAG_FIXED_ROTATE)
            flags &= ~BONE_FLAG_FIXED_SCALE; // flag is only valid with one of the others
        bones[b].flags &= ~BONE_FLAGS_FIXED_UNIFORM;
        bones[b].flags |= flags;
    }

    update_P();
    update_C();
}

void Skeleton::set_rest_frame(int frame)
{
    g_assert(frame >= 0 && frame < (int) frames.size());
    for (size_t i = 0; i < size(); i++)
        bones[i].rest_transform = frames[frame][i];
}

/**************************************************************
 * Test code
 **************************************************************/

#if HAVE_CPPUNIT
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>
#include "formats/formats.h"
#include "mesh/mesh.h"
#include "common/paths.h"

class SkeletonTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(SkeletonTest);
    CPPUNIT_TEST(flags);
    CPPUNIT_TEST(clear_expect);
    CPPUNIT_TEST(update_expect);
    CPPUNIT_TEST(update_bounds);
    CPPUNIT_TEST_SUITE_END();

public:
    void flags()
    {
        const unsigned int target[6] =
        {
            BONE_FLAGS_FIXED_UNIFORM,
            BONE_FLAGS_FIXED_UNIFORM,
            BONE_FLAG_FIXED_TRANSLATE | BONE_FLAG_FIXED_ROTATE,
            BONE_FLAG_FIXED_TRANSLATE | BONE_FLAG_FIXED_SCALE | BONE_FLAG_UNIFORM_SCALE,
            BONE_FLAG_UNIFORM_SCALE,
            BONE_FLAG_FIXED_ROTATE | BONE_FLAG_FIXED_SCALE | BONE_FLAG_UNIFORM_SCALE
        };
        Mesh m;
        string filename = resource_path("srcdir", ".", "tests/meshes/cross.alm");
        format_load_mesh_file("auto", filename, m,
                              MESH_LEVEL_PATCHES);
        Skeleton &skeleton = m.skeleton.get_write();
        skeleton.update_flags();

        for (int i = 0; i < 6; i++)
            CPPUNIT_ASSERT_EQUAL(skeleton.get_bone(i).flags & BONE_FLAGS_FIXED_UNIFORM, target[i]);
    }

    void clear_expect()
    {
        Mesh m;
        string filename = resource_path("srcdir", ".", "tests/meshes/cross.alm");
        format_load_mesh_file("auto", filename, m,
                              MESH_LEVEL_PATCHES);
        Skeleton &skeleton = m.skeleton.get_write();
        skeleton.clear_expect();

        // FIXME: finish
    }

    void update_expect()
    {
        // FIXME
    }

    void update_bounds()
    {
        // FIXME
    }
};

CPPUNIT_TEST_SUITE_REGISTRATION(SkeletonTest);

#endif /* HAVE_CPPUNIT */

/*! @} */
