/*! \addtogroup geom
 * @{
 */

/*! \file
 * \brief A class holding skeletal information for character animation
 */

#ifndef GEOM_SKELETON_H
#define GEOM_SKELETON_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "geom/sparse.h"
#include "geom/coords.h"
#include "geom/bounding.h"
#include "common/mesch.h"
#include <cstddef>
#include <vector>
#include <glib.h>

struct TangentPlane;
class Mesh;

//!< See doc/alm.txt for descriptions of these flags
enum SkeletonBoneFlags
{
    BONE_FLAG_FIXED_TRANSLATE          = 0x00000001,
    BONE_FLAG_FIXED_ROTATE             = 0x00000002,
    BONE_FLAG_FIXED_SCALE              = 0x00000004,
    BONE_FLAG_UNIFORM_SCALE            = 0x00000008,

    BONE_FLAGS_FIXED_LINEAR            = 0x00000006,
    BONE_FLAGS_FIXED                   = 0x00000007,
    BONE_FLAGS_FIXED_UNIFORM_LINEAR    = 0x0000000e,
    BONE_FLAGS_FIXED_UNIFORM           = 0x0000000f,

    BONE_FLAG_EXPECT_L_LINEAR          = 0x00010000,
    BONE_FLAG_EXPECT_L_TRANSLATE       = 0x00020000,
    BONE_FLAG_EXPECT_GTG_LINEAR        = 0x00040000,
    BONE_FLAG_EXPECT_GTG_TRANSLATE     = 0x00080000,
    BONE_FLAG_EXPECT_GTG_WEIGHT        = 0x00100000,
    BONE_FLAG_EXPECT_LTGTGL            = 0x00200000
};

//! Stores internal information on one coordinate frame in a skeleton
struct SkeletonBone
{
    int parent;              //!< Parent bone index, or -1 for the root bone
    std::size_t original_index;
    //! An ugly hack. \c bones[old_index].new_index is the new index for \c old_index
    std::size_t new_index;
    guint32 flags;
    Vector3 tail;            //!< Tail position in local coordinates

    double max_singular;
    BoundingSphere translate_bounds;
    Matrix4 expect_L;
    Matrix4 expect_GTG;
    double expect_LTGTGL;

    Matrix4 L;               //!< Transforms local coordinates to parent (or world)
    Matrix4 G;               //!< Absolute transform to model space
    Matrix4 rest_transform;  //!< Rest pose relative transformation
};

//! Used to pass information to the skeleton (see \c SkeletonBone as well)
struct SkeletonBoneInfo
{
    int parent;
    Matrix4 rest_transform;
    guint32 flags;
    Vector3 tail;            //!< Tail position in local coordinates

    // bounding volume info
    double max_singular;
    BoundingSphere translate_bounds;

    // expectations
    Matrix4 expect_L;
    Matrix4 expect_GTG;
    double expect_LTGTGL;

    SkeletonBoneInfo &operator =(const SkeletonBone &bone);
};

template<class T, int B, template<class, int> class I> T inf_norm(const SparseVector<T, B, I> &a)
{
    typedef typename SparseVector<T, B, I>::const_iterator _const_iterator;
    T big = T();
    for (_const_iterator i = a.begin(); i != a.end(); i++)
        for (std::size_t j = 0; j < B; j++)
            big = std::max(std::abs(i->values[j]), big);
    return big;
}

template<class T, int B, template<class, int> class I> T length_euclidean(const SparseVector<T, B, I> &v)
{
    typedef typename SparseVector<T, B, I>::const_iterator _const_iterator;

    T inf = inf_norm(v);
    if (inf == T()) return inf;
    T inv = 1.0 / inf;

    T ans = T();
    _const_iterator e = v.end();
    for (_const_iterator i = v.begin(); i != e; i++)
        for (std::size_t j = 0; j < B; j++)
        {
            T tmp = i->values[j] * inv;
            ans += tmp * tmp;
        }
    return std::sqrt(ans) * inf;
}

template<class T, int B, template<class, int> class I> T dot_euclidean(const SparseVector<T, B, I> &a,
                                                                       const SparseVector<T, B, I> &b)
{
    typedef typename SparseVector<T, B, I>::const_iterator _const_iterator;

    T ans = T();
    _const_iterator i1 = a.begin(), i2 = b.begin();
    _const_iterator e1 = a.end(), e2 = b.end();

    while (i1 != e1 && i2 != e2)
    {
        if (i1->index == i2->index)
        {
            for (int j = 0; j < B; j++)
                ans += i1->values[j] * i2->values[j];
            i1++;
            i2++;
        }
        else if (i1->index < i2->index) i1++;
        else i2++;
    }
    return ans;
}

template<class T, int B> class DotInner
{
public:
    typedef SparseVector<T, B, ::DotInner> value_type;

    double length(const value_type &v) const
    { return length_euclidean(v); }

    double dot(const value_type &a,
               const value_type &b) const
    { return dot_euclidean(a, b); }

    bool operator ==(const DotInner<T, B> &b) const { return true; }
};

template<class T, int B> class MatrixInner
{
public:
    typedef SparseVector<T, B, ::MatrixInner> value_type;
private:
    MAT *M; // NB: MatrixInner does not own the matrix

public:
    MatrixInner() : M(MNULL) {}
    explicit MatrixInner(MAT *m) : M(m) { g_assert(!M || M->m == M->n); }

    void set_matrix(MAT *m) { M = m; g_assert(!M || M->m == M->n); }

    double length(const value_type &v) const
    {
        typedef typename value_type::const_iterator _const_iterator;
        if (!M) return length_euclidean(v);
        g_assert(v.size() == M->m);

        T inf = inf_norm(v);
        if (inf == T()) return inf;
        T inv = 1.0 / inf;

        T ans = T();
        _const_iterator e = v.end();
        for (_const_iterator i = v.begin(); i != e; i++)
        {
            double scale = 1.0;
            for (_const_iterator j = i; j != e; j++)
            {
                double **px = M->me + i->index;
                for (int x = 0; x < B; x++, px++)
                {
                    double *py = *px + j->index;
                    for (int y = 0; y < B; y++, py++)
                        ans += scale * (i->values[x] * inv) * (j->values[y] * inv) * *py;
                }
                scale = 2.0;
            }
        }
        return std::sqrt(ans) * inf;
    }

    double dot(const value_type &a, const value_type &b) const
    {
        typedef typename value_type::const_iterator _const_iterator;
        if (!M) return dot_euclidean(a, b);
        g_assert(a.size() == M->m);

        T ans = T();
        _const_iterator e1 = a.end();
        _const_iterator e2 = b.end();
        for (_const_iterator i = a.begin(); i != e1; i++)
            for (_const_iterator j = b.begin(); j != e2; j++)
                for (int x = 0; x < B; x++)
                    for (int y = 0; y < B; y++)
                        ans += i->values[x] * j->values[y] * M->me[i->index + x][j->index + y];
        return ans;
    }

    bool operator ==(const MatrixInner<T, B> &b) const
    { return M == b.M; }
};

class Skeleton
{
public:
    enum MetricMode
    {
        MODE_3D,
        MODE_ANIMATION,
        MODE_STATISTICAL
    };
private:
    MAT *P;                 //!< covariance matrix for norms in animation space
    MAT *C;                 //!< some matrix such that C^TC = P
    MAT *M;                 //!< converts SSD space to (homogeneous) 3D
    static MetricMode metric_mode;
    double frame_rate;
    bool custom_P;          //!< True if P has been replaced by set_P

    std::vector<SkeletonBone> bones;
    MatrixInner<double, 4> ssd_inner;

    //! Copies the data from \a b into \c bones, topologically sorting on the way.
    void load_bones(const std::vector<SkeletonBoneInfo> &b);
    void update_G();
    void update_M(std::size_t bone);
    void update_M();
    void update_P(); //! Update expectations when flags or expectations or modified
    void update_C(); //! Update square root of P when P changes

    std::vector<std::vector<Matrix4> > frames;

    friend void mesh_transform(Mesh &m, const Matrix4 &transform);

public:
    static void set_metric_mode(MetricMode mode);

    typedef SparseVector<double, 4, MatrixInner> SsdVector;
    typedef std::vector<SkeletonBone>::const_iterator const_iterator;

    Skeleton();
    Skeleton(const Skeleton &s);
    Skeleton(const std::vector<SkeletonBoneInfo> &b);
    ~Skeleton();
    Skeleton &operator =(const Skeleton &s);

    //! Specifies the bones in the skeleton
    /*! You do not need to fill in the \c SkeletonBone::transform
     * field. It will be copied from \c SkeletonBone::rest_transform.
     * There must be exactly one root bone.
     */
    void set_bones(const std::vector<SkeletonBoneInfo> &bone);
    void set_transform(std::size_t bone, const Matrix4 &m);
    std::size_t size() const { return bones.size(); }
    void clear();
    const_iterator begin() const { return bones.begin(); }
    const_iterator end() const { return bones.end(); }
    const SkeletonBone &get_bone(std::size_t index) const
    { g_assert(index < size()); return bones[index]; }
    std::size_t get_new_index(size_t bone) const
    { g_assert(bone < size()); return bones[bone].new_index; }
    std::size_t get_original_index(size_t bone) const
    { g_assert(bone < size()); return bones[bone].original_index; }

    SsdVector ssd_vector() const { return SsdVector(4 * size(), ssd_inner); }
    Vector3 ssd_to_3d(const SsdVector &s) const;
    Vector3 ssd_to_3d_expected(const SsdVector &s) const;
    Vector3 ssd_to_normal(const TangentPlane &t) const;
    Matrix4 get_absolute(std::size_t bone) const;
    Matrix4 get_model_transform() const { return bones[0].L; }
    MAT *get_M() const { return M; }
    MAT *get_P() const { return P; }
    MAT *get_C() const { return C; }
    //! Override the computed P.
    /*! Call this after loading the bones. The skeleton takes over ownership
     * of the memory, and deletes the old P.
     */
    void set_P(MAT *P);
    //! True if P has been replaced by set_P.
    bool has_custom_P() const { return custom_P; }

    const MatrixInner<double, 4> &get_ssd_inner() const { return ssd_inner; }
    double inf_metric(const SsdVector &v) const;

    void clear_frames() { frames.clear(); }
    size_t num_frames() const { return frames.size(); }
    const std::vector<Matrix4> &get_frame(size_t frame_number) const;
    void set_to_frame(size_t frame_number);
    void set_to_rest();
    void add_frame(const std::vector<Matrix4> &frame);
    void copy_frames(const Skeleton &reference);
    void set_frame_rate(double rate) { frame_rate = rate; }
    double get_frame_rate() const { return frame_rate; }

    /*** Methods to compute higher level stuff ***/

    //! Tries to estimate the tail for all bones with zero tail
    void fix_tails();

    //! Clear all the expectation flags and recompute from the bare minimum
    void clear_expect();
    //! Estimate the expectations from the animation
    void update_expect();
    //! Set bounds to cover just the rest pose
    void clear_bounds();
    //! Estimate the bounds from the animation
    void update_bounds();
    //! Estimate the fixed/uniform flags from the rest pose and animation
    void update_flags();
    //! Copies a particular frame over the rest position
    void set_rest_frame(int frame);
};

#endif /* !GEOM_SKELETON_H */

/*! @} */
