/*! \addtogroup geom
 * @{
 */

/*! \file
 * \brief A sparse vector class.
 */

#ifndef ALOD_GEOM_SPARSE_H
#define ALOD_GEOM_SPARSE_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "common/refcount.h"
#include "common/mesch.h"
#include <cstddef>
#include <vector>
#include <functional>
#include <numeric>
#include <cmath>
#include <iostream>
#include <algorithm>
#include <glib.h>

template<class T, int B, template<class, int> class I> class SparseVector;

template<class T, int B, template<class, int> class I>
static bool vector_compare(const SparseVector<T, B, I> &a,
                           const SparseVector<T, B, I> &b);

template<class T, int B, template<class, int> class I>
static bool vector_compare_float(const SparseVector<T, B, I> &a,
                                 const SparseVector<T, B, I> &b);

// For some reason std::min and std::max don't want to work as function
// objects. I have no idea why these work and std:: doesn't.
template<class T> inline T _vector_min(const T &a, const T &b)
{ if (a < b) return a; return b; }
template<class T> inline T _vector_max(const T &a, const T &b)
{ if (a < b) return b; return a; }

template<class T, int B> struct _sparse_element
{
    std::size_t index;
    T values[B];

    _sparse_element() {}
    explicit _sparse_element(std::size_t i) : index(i) {}
    _sparse_element(std::size_t i, const T &v) : index(i)
    { std::fill(values, values + B, v); }
};

template<class T, int B> struct _compare_element : std::binary_function<_sparse_element<T, B>, std::size_t, bool>
{
    bool operator()(const _sparse_element<T, B> &a, const std::size_t &b) const
    { return a.index < b; }
};

template<class T, int B, template<class, int> class InnerProduct> class SparseVector
{
public:
    typedef T value_type;
    typedef T *pointer_type;
    typedef T &reference;
    typedef const T &const_reference;

private:
    typedef SparseVector<T, B, InnerProduct> _self;
    typedef _sparse_element<T, B> _block;
    typedef std::vector<_block> _base;
    typedef Refcount<_base> _ref;
    typedef _compare_element<T, B> _compare;
    template<class T2, int B2, template<class, int> class IP2> friend class SparseVector;

public:
    typedef typename _base::const_iterator const_iterator;

private:
    typedef typename _base::iterator iterator;
    _ref v;
    std::size_t n;
    InnerProduct<T, B> inner;

    bool valid() const { return n != 0; }

    static std::size_t get_index(const_iterator i1, const_iterator i2,
                                 const_iterator e1, const_iterator e2)
    {
        if (i1 == e1) return i2->index;
        else if (i2 == e2) return i1->index;
        else return std::min(i1->index, i2->index);
    }

    static T get_value(const_iterator i1, const_iterator i2,
                       const_iterator e1, const_iterator e2,
                       int k)
    {
        if (i1 == e1) return T();
        else if (i2 == e2 || i1->index <= i2->index) return i1->values[k];
        else return T();
    }

    static void advance_pair(const_iterator &i1, const_iterator &i2,
                             const_iterator e1, const_iterator e2)
    {
        if (i1 == e1) i2++;
        else if (i2 == e2) i1++;
        else if (i1->index == i2->index) i1++, i2++;
        else if (i1->index < i2->index) i1++;
        else i2++;
    }

public:
    template<class Op> _self unary_op(Op op) const
    {
        g_assert(valid());

        _self ans(size(), inner);
        _base &out = ans.v.get_write();
        const _base &in = v.get_read();

        out.resize(in.size());
        iterator o = out.begin();
        for (const_iterator i = in.begin(); i != in.end(); i++, o++)
        {
            o->index = i->index;
            for (int j = 0; j < B; j++)
                o->values[j] = op(i->values[j]);
        }
        return ans;
    }

    template<class Op> _self binary_op(const _self &b, Op op) const
    {
        g_assert(valid() && b.valid());
        g_assert(size() == b.size());
        g_assert(inner == b.inner);

        _self ans(size(), inner);
        _base &out = ans.v.get_write();
        const _base &in1 = v.get_read();
        const _base &in2 = b.v.get_read();

        const_iterator i = in1.begin();
        const_iterator j = in2.begin();
        while (i != in1.end() || j != in2.end())
        {
            out.push_back(_block(get_index(i, j, in1.end(), in2.end())));
            for (int k = 0; k < B; k++)
                out.back().values[k] = op(get_value(i, j, in1.end(), in2.end(), k),
                                          get_value(j, i, in2.end(), in1.end(), k));
            advance_pair(i, j, in1.end(), in2.end());
        }
        return ans;
    }

    void compress(double tol = 1e-12)
    {
        _base &me = v.get_write();
        iterator out = me.begin();
        for (const_iterator in = me.begin(); in != me.end(); in++)
        {
            bool zeros = true;
            for (int j = 0; j < B; j++)
                if (fabs(in->values[j]) > tol) zeros = false;
            if (!zeros) *out++ = *in;
        }
        me.erase(out, me.end());
    }

    SparseVector() : n(0) {}
    SparseVector(std::size_t s, const InnerProduct<T, B> &i) : n(s), inner(i)
    { g_assert(n > 0); }
    ~SparseVector() {}

    const_iterator begin() const { return v.get_read().begin(); }
    const_iterator end() const { return v.get_read().end(); }

    std::size_t size() const { return n; }
    T operator [](std::size_t index) const
    {
        g_assert(valid());
        g_assert(index < size());
        std::size_t rem = index % B;
        std::size_t block = index - rem;
        const_iterator i = std::lower_bound(v.get_read().begin(),
                                            v.get_read().end(),
                                            block,
                                            _compare());
        if (i != v.get_read().end() && i->index == block)
            return i->values[rem];
        else
            return T();
    }

    T &operator [](std::size_t index)
    {
        g_assert(valid());
        g_assert(index < size());

        _base &base = v.get_write();
        std::size_t rem = index % B;
        std::size_t block = index - rem;
        iterator i = std::lower_bound(base.begin(),
                                      base.end(),
                                      block,
                                      _compare());
        if (i == base.end() || i->index != block)
            i = base.insert(i, _block(block, T()));
        return i->values[rem];
    }

    void zero()
    { g_assert(valid()); v.get_write().clear(); }

    const InnerProduct<T, B> &get_inner() const
    { return inner; }

    void set_inner(const InnerProduct<T, B> &in)
    { inner = in; }

    _self operator -() const
    { return unary_op(std::negate<T>()); }

    _self operator +() const
    { g_assert(valid()); return *this; }

    _self operator *(T scale) const
    { return unary_op(std::bind2nd(std::multiplies<T>(), scale)); }

    _self operator /(T scale) const
    { return unary_op(std::bind2nd(std::multiplies<T>(), 1.0 / scale)); }

    _self operator +(const _self &b) const
    { return binary_op(b, std::plus<T>()); }

    _self operator -(const _self &b) const
    { return binary_op(b, std::minus<T>()); }

    _self &operator +=(const _self &b)
    { return *this = *this + b; }

    _self &operator -=(const _self &b)
    { return *this = *this - b; }

    _self &operator *=(T scale)
    {
        g_assert(valid());
        _base &b = v.get_write();
        for (iterator i = b.begin(); i != b.end(); i++)
            for (int j = 0; j < B; j++)
                i->values[j] *= scale;
        return *this;
    }

    _self &operator /=(T scale)
    { return *this *= 1.0 / scale; }

    bool operator !() const
    {
        const _base &in = v.get_read();
        for (const_iterator i = in.begin(); i != in.end(); i++)
            for (int j = 0; j < B; j++)
                if (i->values[j] != T()) return false;
        return true;
    }

    bool operator ==(const _self &b)
    {
        g_assert(valid() && b.valid());
        g_assert(size() == b.size());
        g_assert(inner == b.inner);

        const _base &in1 = v.get_read();
        const _base &in2 = b.v.get_read();

        const_iterator i = in1.begin();
        const_iterator j = in2.begin();
        const_iterator e1 = in1.end();
        const_iterator e2 = in2.end();
        while (i != e1 || j != e2)
        {
            for (int k = 0; k < B; k++)
            {
                if (get_value(i, j, e1, e2, k) != get_value(j, i, e2, e1, k))
                    return false;
            }
            advance_pair(i, j, e1, e2);
        }
        return true;
    }

    bool operator !=(const _self &b)
    { return !(*this == b); }

    T length() const
    { g_assert(valid()); return inner.length(*this); }

    T dot(const _self &b) const
    {
        g_assert(valid() && b.valid());
        g_assert(size() == b.size());
        g_assert(inner == b.inner);

        return inner.dot(*this, b);
    }

    template<int B2, template<class, int> class InnerProduct2> void matrix_vector(MAT *m, SparseVector<T, B2, InnerProduct2> &out) const
    {
        typedef SparseVector<T, B2, InnerProduct2> Out;
        const _base &in = v.get_read();

        g_assert(valid());
        g_assert(out.size() == m->m && m->n == size());
        out.zero();

        for (std::size_t i = 0; i < out.size(); i += B2)
        {
            bool any = false;

            typename Out::_block b;
            b.index = i;
            for (int c = 0; c < B2; c++)
            {
                b.values[c] = T();
                for (const_iterator j = in.begin(); j != in.end(); j++)
                {
                    double *in2 = &m->me[i + c][j->index];
                    for (int k = 0; k < B; k++, in2++)
                        b.values[c] += *in2 * j->values[k];
                }
                if (b.values[c] != T()) any = true;
            }
            if (any) out.v.get_write().push_back(b);
        }
    }

    friend bool vector_compare<T, B, InnerProduct>(const SparseVector<T, B, InnerProduct> &a,
                                                   const SparseVector<T, B, InnerProduct> &b);
    friend bool vector_compare_float<T, B, InnerProduct>(const SparseVector<T, B, InnerProduct> &a,
                                                         const SparseVector<T, B, InnerProduct> &b);
};

template<class T, int B, template<class, int> class I>
inline SparseVector<T, B, I> operator *(T scale, const SparseVector<T, B, I> &b)
{ return b * scale; }

template<class T, int B, template<class, int> class I>
inline T length(const SparseVector<T, B, I> &v)
{ return v.length(); }

template<class T, int B, template<class, int> class I>
inline SparseVector<T, B, I> normalise(const SparseVector<T, B, I> &v)
{ return v / length(v); }

template<class T, int B, template<class, int> class I>
inline SparseVector<T, B, I> vector_min(const SparseVector<T, B, I> &a,
                                        const SparseVector<T, B, I> &b)
{ return a.binary_op(b, _vector_min<T>); }

template<class T, int B, template<class, int> class I>
inline SparseVector<T, B, I> vector_max(const SparseVector<T, B, I> &a,
                                        const SparseVector<T, B, I> &b)
{ return a.binary_op(b, _vector_max<T>); }

// lexicographic less
template<class T, int B, template<class, int> class I>
static bool vector_compare(const SparseVector<T, B, I> &a,
                           const SparseVector<T, B, I> &b)
{
    g_assert(a.size() == b.size());

    typename SparseVector<T, B, I>::const_iterator ai = a.begin();
    typename SparseVector<T, B, I>::const_iterator bi = b.begin();
    typename SparseVector<T, B, I>::const_iterator ae = a.end();
    typename SparseVector<T, B, I>::const_iterator be = b.end();

    while (ai != ae || bi != be)
    {
        for (int j = 0; j < B; j++)
        {
            T v1 = SparseVector<T, B, I>::get_value(ai, bi, ae, be, j);
            T v2 = SparseVector<T, B, I>::get_value(bi, ai, be, ae, j);
            if (v1 != v2) return v1 < v2;
        }
        SparseVector<T, B, I>::advance_pair(ai, bi, ae, be);
    }

    return false;
}

// Similar, but first quantises to float
template<class T, int B, template<class, int> class I>
static bool vector_compare_float(const SparseVector<T, B, I> &a,
                                 const SparseVector<T, B, I> &b)
{
    g_assert(a.size() == b.size());

    typename SparseVector<T, B, I>::const_iterator ai = a.begin();
    typename SparseVector<T, B, I>::const_iterator bi = b.begin();
    typename SparseVector<T, B, I>::const_iterator ae = a.end();
    typename SparseVector<T, B, I>::const_iterator be = b.end();

    while (ai != ae || bi != be)
    {
        for (int j = 0; j < B; j++)
        {
            T v1 = SparseVector<T, B, I>::get_value(ai, bi, ae, be, j);
            T v2 = SparseVector<T, B, I>::get_value(bi, ai, be, ae, j);
            if ((float) v1 != (float) v2) return (float) v1 < (float) v2;
        }
        SparseVector<T, B, I>::advance_pair(ai, bi, ae, be);
    }

    return false;
}

template<class T, int B, template<class, int> class I>
inline T dot(const SparseVector<T, B, I> &a,
             const SparseVector<T, B, I> &b)
{ return a.dot(b); }

template<class T, int B, template<class, int> class I>
inline T double_area(const SparseVector<T, B, I> &a,
                     const SparseVector<T, B, I> &b)
{
    T d = dot(a, b);
    T s = dot(a, a) * dot(b, b) - d * d;
    return std::sqrt(std::max(s, T()));
}

template<class T, int B, template<class, int> class I>
std::ostream &operator <<(std::ostream &o, const SparseVector<T, B, I> &v)
{
    int old = o.precision(3);
    o << "(";

    std::size_t cur = 0;
    typename SparseVector<T, B, I>::const_iterator i = v.begin();
    typename SparseVector<T, B, I>::const_iterator e = v.end();
    for (; i != e; i++)
    {
        while (cur < i->index)
        {
            if (cur++) o << ", ";
            o << "[0]";
        }
        for (int j = 0; j < B; j++)
        {
            if (cur++) o << ", ";
            o << i->values[j];
        }
    }
    while (cur < v.size())
    {
        if (cur++) o << ", ";
        o << "[0]";
    }

    o << ")";
    o.precision(old);
    return o;
}

template<class T, int B, template<class, int> class I>
T operator *(const SparseVector<T, B, I> &a,
             const SparseVector<T, B, I> &b)
{ return a.dot(b); }

template<class T, int B, template<class, int> class I, int B2, template<class, int> class I2>
void matrix_vector(MAT *m,
                   const SparseVector<T, B, I> &v,
                   SparseVector<T, B2, I2> &out)
{ v.matrix_vector(m, out); }

#endif /* !ALOD_GEOM_SPARSE_H */

/*! @} */
