/*! \addtogroup mesh
 * @{
 */

// #define DUMP_APS

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <string>
#include <vector>
#include <utility>
#include "aps.h"
#include "geom/coords.h"
#include "geom/skeleton.h"
#include "geom/anim.h"
#include "common/maputils.h"
#include <glib.h>
#ifdef DUMP_APS
# include <fstream>
#endif
using namespace std;

struct CellCorner
{
    Vector2 uv;
    VectorPosition delta;

    vector<MeshFaceConstIterator> old_faces;
    vector<MeshFaceConstIterator> new_faces;

    CellCorner() {}
    //! Intersection constructor
    CellCorner(const Vector2 &uv, const VectorPosition &delta,
               MeshHalfEdgeConstIterator old_edge,
               MeshHalfEdgeConstIterator new_edge);

    //! Existing vertex constructor
    CellCorner(const Vector2 &uv, MeshFaceConstIterator old_face,
               MeshFaceConstIterator new_face,
               const Skeleton &skeleton);
};

CellCorner::CellCorner(const Vector2 &uv, const VectorPosition &delta,
                       MeshHalfEdgeConstIterator old_edge,
                       MeshHalfEdgeConstIterator new_edge)
: uv(uv), delta(delta)
{
    old_faces.push_back(old_edge->face());
    old_faces.push_back(old_edge->dual()->face());
    new_faces.push_back(new_edge->face());
    new_faces.push_back(new_edge->dual()->face());
}

CellCorner::CellCorner(const Vector2 &uv, MeshFaceConstIterator old_face,
                       MeshFaceConstIterator new_face,
                       const Skeleton &skeleton)
: uv(uv), delta(skeleton)
{
    delta.zero();
    old_faces.push_back(old_face);
    new_faces.push_back(new_face);
}

#ifdef DUMP_APS
static void dump_edge(ostream &out, const Vector2 &a,
                      const Vector2 &b, int lt, int lw)
{
    out << a[0] << "+" << b[0] - a[0] << "*t,"
        << a[1] << "+" << b[1] - a[1] << "*t lt " << lt << " lw " << lw << " ";
}

template<class InputIterator>
static void dump_neighbourhood(MeshHalfEdgeConstIterator edge,
                               InputIterator cell_first,
                               InputIterator cell_last)
{
    Vector2 uv_new[2] =
    {
        edge->prev()->vertex()->data.texcoords,
        edge->dual()->next()->vertex()->data.texcoords
    };
    bool first = true;

    ofstream f("aps.gplot");
    f << "set parametric\n";
    f << "set nokey\n";
    f << "set nolabel\n";

    f << "plot [t=0:1]";

    MeshHalfEdgeConstIterator e = edge;
    do
    {
        Vector2 a = e->vertex()->data.texcoords;
        Vector2 b = e->next()->vertex()->data.texcoords;
        Vector2 c = e->prev()->vertex()->data.texcoords;
        f << (first ? ' ' : ','); first = false;
        dump_edge(f, a, b, 1, 3);
        f << ',';
        dump_edge(f, a, c, 1, 3);
        f << ',';
        dump_edge(f, b, c, 1, 3);
        e = e->next()->dual();
    } while (e != edge);

    e = edge->prev()->dual()->prev();
    while (e != edge->dual())
    {
        Vector2 a = e->vertex()->data.texcoords;
        Vector2 b = e->next()->vertex()->data.texcoords;
        Vector2 c = e->prev()->vertex()->data.texcoords;
        Vector2 n = (e->face()->patch == edge->face()->patch) ? uv_new[0] : uv_new[1];
        f << ',';
        dump_edge(f, a, b, 1, 3);
        f << ',';
        dump_edge(f, b, c, 1, 3);
        f << ',';
        dump_edge(f, a, c, 1, 3);
        f << ',';
        dump_edge(f, b, n, 2, 1);
        e = e->dual()->prev();
    }

    for (InputIterator i = cell_first; i != cell_last; i++)
    {
        const Vector2 &uv = i->uv;
        f << ',';
        f << uv[0] << "," << uv[1] << " with points ps 4";
    }

    f << "\n";
}
#endif /* DUMP_APS */

/*! Holds a bounding box of deviations. The deviations are transformed
 * by the C matrix of the skeleton so that a Euclidean metric is in use
 * (which bounds the over-estimation to sqrt(d) in d dims), and also aligns
 * the box with the principle components.
 */
class BoundingBox
{
private:
    const Skeleton &skeleton;
    VEC *low, *high;
    static VEC *tmp;            //!< temporary storage during calculations
    bool init;

public:
    explicit BoundingBox(const Skeleton &skeleton);
    BoundingBox(const BoundingBox &bb);
    BoundingBox &operator =(const BoundingBox &bb);
    BoundingBox operator +(const VectorPosition &v) const;
    BoundingBox operator -(const VectorPosition &v) const;
    BoundingBox &operator |=(const BoundingBox &bb);  //!< Union
    BoundingBox &operator &=(const BoundingBox &bb);  //!< Intersection
    ~BoundingBox();

    double length() const;
};

VEC *BoundingBox::tmp = NULL;

BoundingBox::BoundingBox(const Skeleton &skeleton)
: skeleton(skeleton),
low(v_get(4 * skeleton.size())),
high(v_get(4 * skeleton.size())),
init(false)
{
    v_zero(low);
    v_zero(high);
}

BoundingBox::~BoundingBox()
{
    V_FREE(low);
    V_FREE(high);
}

BoundingBox::BoundingBox(const BoundingBox &bb)
: skeleton(bb.skeleton),
low(v_copy(bb.low, NULL)),
high(v_copy(bb.high, NULL)),
init(bb.init)
{
}

BoundingBox &BoundingBox::operator =(const BoundingBox &bb)
{
    low = v_copy(bb.low, low);
    high = v_copy(bb.high, high);
    init = bb.init;
    return *this;
}

BoundingBox BoundingBox::operator +(const VectorPosition &v) const
{
    BoundingBox ans = *this;

    MAT *C = skeleton.get_C();
    tmp = v_resize(tmp, low->dim);
    v_zero(tmp);
    for (VectorPosition::const_iterator i = v.begin(); i != v.end(); i++)
        for (int j = 0; j < 4; j++)
            for (size_t k = 0; k < tmp->dim; k++)
                v_add_val(tmp, k, m_entry(C, k, i->index + j) * i->values[j]);

    if (ans.init)
    {
        ans.low = v_add(low, tmp, ans.low);
        ans.high = v_add(high, tmp, ans.high);
    }
    else
    {
        ans.init = true;
        ans.low = v_copy(tmp, ans.low);
        ans.high = v_copy(tmp, ans.high);
    }
    return ans;
}

BoundingBox BoundingBox::operator -(const VectorPosition &v) const
{
    return (*this) + (-v);
}

BoundingBox &BoundingBox::operator |=(const BoundingBox &bb)
{
    if (!bb.init) return *this;
    if (!init) return *this = bb;

    for (size_t i = 0; i < low->dim; i++)
    {
        v_set_val(low, i, min(v_entry(low, i), v_entry(bb.low, i)));
        v_set_val(high, i, max(v_entry(high, i), v_entry(bb.high, i)));
    }
    return *this;
}

BoundingBox &BoundingBox::operator &=(const BoundingBox &bb)
{
    if (!bb.init) return *this;
    if (!init) return *this = bb;

    for (size_t i = 0; i < low->dim; i++)
    {
        v_set_val(low, i, max(v_entry(low, i), v_entry(bb.low, i)));
        v_set_val(high, i, min(v_entry(high, i), v_entry(bb.high, i)));
    }
    return *this;
}

double BoundingBox::length() const
{
    if (!init)
        return 0.0;
    tmp = v_resize(tmp, low->dim);
    for (size_t i = 0; i < low->dim; i++)
        v_set_val(tmp, i, max(-v_entry(low, i), v_entry(high, i)));
    return v_norm2(tmp);
}

class MeshSimplifierAps : public MeshSimplifier
{
private:
    //! True if this is a memory-full APS.
    /*! Memory-full APS maintains, for each face, a bounding box on the
     * deviations for UV coordinates currently covered by that face.
     */
    bool memory;
    vector<BoundingBox *> deviations;
    typedef id_map<MeshFaceConstIterator, BoundingBox>::type BoundingBoxMap;

    mutable InfluenceSimplificationCache influence_cache;

    bool foldover(MeshHalfEdgeConstIterator edge) const;
    template<class OutputIterator> void cell_corners(MeshHalfEdgeConstIterator edge,
                                                     OutputIterator out) const;

    //! Updates memory-full bounding box information
    /*! For each corner in the range, the bounding box of the old faces is
     * used to enlarge the bounding box of the new faces.
     */
    template<class InputIterator> void propagate(InputIterator first_corner,
                                                 InputIterator last_corner,
                                                 BoundingBoxMap &out) const;
    MeshEdgeCollapseData *evaluate_edge_collapse(MeshHalfEdgeConstIterator edge) const;
    MeshInfluenceCollapseData *evaluate_influence_collapse(MeshPositionConstIterator pos) const;
    BoundingBox &get_deviations(MeshFaceConstIterator f) const
    {
        g_assert(memory);
        g_assert(deviations[GPOINTER_TO_SIZE(f->internal_data)]);
        return *deviations[GPOINTER_TO_SIZE(f->internal_data)];
    }
    void remove_deviations(MeshFaceConstIterator f)
    {
        g_assert(memory);
        g_assert(deviations[GPOINTER_TO_SIZE(f->internal_data)]);
        delete deviations[GPOINTER_TO_SIZE(f->internal_data)];
        deviations[GPOINTER_TO_SIZE(f->internal_data)] = NULL;
    }

    void simplify_tangent_plane(const VectorPosition &pos, MeshVertexData &data) const;

protected:
    MeshEdgeCollapseData *evaluate_collapse_half(MeshHalfEdgeConstIterator edge) const;
    MeshPositionIterator collapse_edge(MeshHalfEdgeIterator edge, const MeshEdgeCollapseData &col);
    MeshPositionIterator collapse_influence(MeshPositionIterator old_pos, const MeshInfluenceCollapseData &col);

public:
    MeshSimplifierAps(Mesh &m, bool memory, bool do_influence, bool do_edge);
    ~MeshSimplifierAps();
};

MeshSimplifierAps::MeshSimplifierAps(Mesh &m, bool memory, bool do_influence, bool do_edge)
: MeshSimplifier(m, do_influence, do_edge), memory(memory), influence_cache(m.skeleton.get_read(), 1024)
{
    if (memory)
    {
        deviations.resize(m.faces.real_size());
        int idx = 0;
        for (MeshFaceIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++, idx++)
        {
            deviations[idx] = new BoundingBox(m.skeleton.get_read());
            f->internal_data = GSIZE_TO_POINTER(idx);
        }
    }
}

MeshSimplifierAps::~MeshSimplifierAps()
{
    for (size_t i = 0; i < deviations.size(); i++)
        delete deviations[i];
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "ISC miss rate: %.6g", 1.0 - influence_cache.hit_rate());
}

//! Checks whether moving the clockwise vertex of \a edge to the counterclockwise one will cause a foldover.
bool MeshSimplifierAps::foldover(MeshHalfEdgeConstIterator edge) const
{
    Vector2 uv_new[2] =
    {
        edge->prev()->vertex()->data.texcoords,
        edge->dual()->next()->vertex()->data.texcoords
    };
    MeshHalfEdgeConstIterator e = edge->prev()->dual()->prev();

    while (e != edge->dual())
    {
        Vector2 uv0 = e->vertex()->data.texcoords;
        Vector2 uv1 = e->next()->vertex()->data.texcoords;
        Vector2 uv2 = (e->face()->patch == edge->face()->patch) ? uv_new[0] : uv_new[1];
        if (cross(uv1 - uv0, uv2 - uv0) <= 1e-5 * length(uv1 - uv0) * length(uv2 - uv0))
            return true;
        e = e->dual()->prev();
    }
    return false;
}

/* There seems to be some weird problem, possibly with gcc 3.4, that
 * makes (((*x).*forw)()) not work. However, I haven't been able to
 * reproduce it with a microprogram.
 */
#define APS_FORW(x) (((&*x)->*forw)())
#define APS_BACK(x) (((&*x)->*back)())

//! Determines cell corners generated by moving the clockwise to the ccw vertex
template<class OutputIterator>
void MeshSimplifierAps::cell_corners(MeshHalfEdgeConstIterator edge,
                                     OutputIterator out) const
{
    const Skeleton &skeleton = get_mesh().skeleton.get_read();
    VectorPosition old_pos = edge->next()->vertex()->position()->position;
    VectorPosition new_pos = edge->prev()->vertex()->position()->position;
    typedef MeshHalfEdgeConstIterator (MeshHalfEdge::*edge_ptr)() const;

#ifdef DUMP_APS
    vector<CellCorner> dummy;
    dump_neighbourhood(edge, dummy.begin(), dummy.end());
#endif
    bool cross_patch = edge->face()->patch != edge->dual()->face()->patch;

    /* Generate data on each side of the split */
    MeshHalfEdgeConstIterator splits[2];
    CellCorner old_corners[2];
    bool split_old_corners = false;
    for (int side = 0; side < 2; side++)
    {
        MeshHalfEdgeConstIterator base = side ? edge->dual() : edge;
        edge_ptr forw = side ? (edge_ptr) &MeshHalfEdge::prev : (edge_ptr) &MeshHalfEdge::next;
        edge_ptr back = side ? (edge_ptr) &MeshHalfEdge::next : (edge_ptr) &MeshHalfEdge::prev;
        Vector2 old_uv = APS_FORW(base)->vertex()->data.texcoords;
        Vector2 new_uv = APS_BACK(base)->vertex()->data.texcoords;

        vector<MeshHalfEdgeConstIterator> seen;
        MeshHalfEdgeConstIterator e = APS_BACK(base);
        if (e->dual()->face()->patch == e->face()->patch)
        {
            /* The test above can fail. This means that we have generated
             * a close to degenerate triangle, two of whose edges belong to
             * the same patch boundary. Since the middle vertex was not
             * labelled as a patch corner, it *should* be safe to eliminate
             * it.
             */
            if (memory)
                *out++ = CellCorner(new_uv, e->face(), e->dual()->face(),
                                    skeleton);
            seen.push_back(e);
            e = APS_BACK(e->dual());
            while (e != base->dual())
            {
                Vector2 uv1 = APS_FORW(e)->vertex()->data.texcoords;
                VectorPosition pos1 = APS_FORW(e)->vertex()->position()->position;
                if (e->dual()->face()->patch != e->face()->patch) break;
                double dot = (uv1 - old_uv) * (new_uv - old_uv);
                double area = cross(uv1 - old_uv, new_uv - old_uv);
                if (side == 0 && area > 1e-6 * dot) break;
                if (side == 1 && area < -1e-6 * dot) break;

                for (size_t i = 0; i < seen.size(); i++)
                {
                    MeshHalfEdgeConstIterator e2 = seen[i];
                    Vector2 uv2 = APS_FORW(e2)->vertex()->data.texcoords;
                    VectorPosition pos2 = APS_FORW(e2)->vertex()->position()->position;
                    double s, t;

                    Vector2 uv = intersection(uv1, new_uv, uv2, old_uv, s, t);
                    VectorPosition delta = (1.0 - s) * pos1 + s * new_pos
                        - ((1.0 - t) * pos2 + t * old_pos);
                    g_assert(s >= -1e-5 && s <= 1.0 + 1e-5);
                    g_assert(t >= -1e-5 && t <= 1.0 + 1e-5);
                    *out++ = CellCorner(uv, delta, e2, e);
                }
                if (memory)
                {
                    *out++ = CellCorner(uv1, e->face(), e->dual()->face(),
                                        skeleton);
                    *out++ = CellCorner(new_uv, base->face(), e->dual()->face(),
                                        skeleton);
                }

                seen.push_back(e);
                e = APS_BACK(e->dual());
            }
        }
        g_assert(e != base->dual());
        splits[side] = e;

        /* Generate corner data for the old point */
        int target = side;

        if (side == 1 && splits[0] != splits[1]->dual() && !cross_patch) target = 0;
        old_corners[target].uv = old_uv;
        for (size_t i = 0; i < seen.size(); i++)
            old_corners[target].old_faces.push_back(seen[i]->face());

        if (side == target)
        {
            old_corners[target].old_faces.push_back(e->face());
            old_corners[target].new_faces.push_back(e->face());
        }

        if (side == 0)
        {
            VectorPosition p(skeleton);
            MeshVertexConstIterator v[3];
            v[0] = e->vertex();
            v[1] = APS_FORW(e)->vertex();
            v[2] = APS_BACK(base)->vertex();
            p.zero();
            for (int i = 0; i < 3; i++)
            {
                int i2 = (i == 2) ? 0 : i + 1;
                int i3 = 3 - i - i2;
                p += cross(v[i2]->data.texcoords - v[i]->data.texcoords,
                           old_uv - v[i]->data.texcoords) * v[i3]->position()->position;
            }
            p /= cross(v[1]->data.texcoords - v[0]->data.texcoords,
                       v[2]->data.texcoords - v[0]->data.texcoords);
            old_corners[target].delta = p - old_pos;
        }
        else if (target == 1)
        {
            old_corners[target].delta = old_corners[0].delta;
            split_old_corners = true;
        }
    }
    *out++ = old_corners[0];
    if (split_old_corners) *out++ = old_corners[1];
}

template<class InputIterator>
void MeshSimplifierAps::propagate(InputIterator first_corner,
                                  InputIterator last_corner,
                                  MeshSimplifierAps::BoundingBoxMap &out) const
{
    g_assert(memory);
    const Skeleton &skeleton = get_mesh().skeleton.get_read();

    BoundingBox bbox(skeleton);
    for (InputIterator c = first_corner; c != last_corner; c++)
    {
        for (size_t i = 0; i < c->old_faces.size(); i++)
        {
            MeshFaceConstIterator f = c->old_faces[i];
            const BoundingBox &old_dev = get_deviations(f);
            bbox |= old_dev - c->delta;
        }
        for (size_t i = 0; i < c->new_faces.size(); i++)
        {
            MeshFaceConstIterator f = c->new_faces[i];
            BoundingBoxMap::iterator outbb = out.find(f);
            if (outbb == out.end())
                outbb = out.insert(make_pair(f, get_deviations(f))).first;
            outbb->second |= bbox;
        }
    }
}

//! Computes the cost of collapsing the clockwise vertex of \a edge to the counterclockwise one
/*! \pre \c foldover(edge) must return false
 */
MeshEdgeCollapseData *MeshSimplifierAps::evaluate_collapse_half(MeshHalfEdgeConstIterator edge) const
{
    MeshEdgeCollapseData *ans = new MeshEdgeCollapseData(0.0);
    ans->position = edge->prev()->vertex()->position()->position;
    ans->vertex_data[0] = edge->prev()->vertex()->data;
    ans->vertex_data[1] = edge->dual()->next()->vertex()->data;
    ans->t = 1;

    vector<CellCorner> corners;
    cell_corners(edge, back_inserter(corners));
#ifdef DUMP_APS
    dump_neighbourhood(edge, corners.begin(), corners.end());
#endif

    if (memory)
    {
        BoundingBoxMap bbox;

        propagate(corners.begin(), corners.end(), bbox);
        for (BoundingBoxMap::const_iterator b = bbox.begin(); b != bbox.end(); b++)
            ans->cost = max(ans->cost, b->second.length());
    }
    else
    {
        for (vector<CellCorner>::iterator c = corners.begin(); c != corners.end(); c++)
            ans->cost = max(ans->cost, length(c->delta));
    }

    return ans;
}

MeshEdgeCollapseData *MeshSimplifierAps::evaluate_edge_collapse(MeshHalfEdgeConstIterator edge) const
{
    if (!legal_collapse(edge)) return new MeshEdgeCollapseData(HUGE_VAL);
    int forced_t = edge_forced_t(edge);
    g_assert(forced_t != -2);

    MeshEdgeCollapseData *ans = NULL;
    for (int i = 0; i < 2; i++)
    {
        if (forced_t != i)
        {
            MeshHalfEdgeConstIterator e = (i == 1) ? edge->dual() : edge;
            if (!foldover(e))
            {
                MeshEdgeCollapseData *cur = evaluate_collapse_half(e);
                if (cur && (!ans || cur->cost < ans->cost))
                {
                    if (ans) delete ans;
                    ans = cur;
                    if (i == 1)
                    {
                        swap(ans->vertex_data[0],
                             ans->vertex_data[1]);
                        if (ans->t >= 0) ans->t = !ans->t;
                    }
                }
                else if (cur)
                    delete cur;
            }
        }
    }
    if (!ans) ans = new MeshEdgeCollapseData(HUGE_VAL);
    return ans;
}

MeshPositionIterator MeshSimplifierAps::collapse_edge(MeshHalfEdgeIterator edge, const MeshEdgeCollapseData &col)
{
    if (memory)
    {
        vector<CellCorner> corners;
        BoundingBoxMap bboxs;
        MeshHalfEdgeIterator e;

        if (col.t == 1)
            e = edge;
        else
            e = edge->dual();
        cell_corners(e, back_inserter(corners));
        propagate(corners.begin(), corners.end(), bboxs);
        for (BoundingBoxMap::const_iterator b = bboxs.begin(); b != bboxs.end(); b++)
            get_deviations(b->first) = b->second;

        remove_deviations(edge->face());
        remove_deviations(edge->dual()->face());
    }

    return MeshSimplifier::collapse_edge(edge, col);
}

void MeshSimplifierAps::simplify_tangent_plane(const VectorPosition &pos, MeshVertexData &data) const
{
    TangentPlane &tp = data.tangent_plane;
    vector<int> pinf;
    VectorPosition::const_iterator j;
    for (j = pos.begin(); j != pos.end(); j++)
        pinf.push_back(j->index);

    for (int i = 0; i < 2; i++)
    {
        vector<int> tinf, dinf;
        for (j = tp.tangents[i].begin(); j != tp.tangents[i].end(); j++)
            tinf.push_back(j->index);
        set_difference(tinf.begin(), tinf.end(), pinf.begin(), pinf.end(),
                       back_inserter(dinf));
        tp.tangents[i] = simplify_position_cache_remove(influence_cache, tp.tangents[i], dinf);
        tp.tangents[i] = normalise(tp.tangents[i]);
    }
    tp.tangents[1] -= (tp.tangents[0] * tp.tangents[1]) * tp.tangents[1];
    tp.tangents[1] = normalise(tp.tangents[1]);
}

MeshInfluenceCollapseData *MeshSimplifierAps::evaluate_influence_collapse(MeshPositionConstIterator pos) const
{
    MeshHalfEdgeConstIterator e;
    VectorPosition old_pos = pos->position;

    // If only one influence, cannot be simplified
    if (distance(old_pos.begin(), old_pos.end()) <= 1)
        return new MeshInfluenceCollapseData(HUGE_VAL);

    MeshInfluenceCollapseData *ans = new MeshInfluenceCollapseData();
    ans->position = simplify_position_cache(influence_cache, old_pos);
    VectorPosition delta = ans->position - old_pos;

    e = pos->first_vertex()->first_edge();
    /* Walk clockwise around the vertices */
    do
    {
        MeshVertexConstIterator v = e->next()->vertex();
        if (mesh_true_vertex(v))
        {
            MeshVertexData data = v->data;
            simplify_tangent_plane(ans->position, data);
            ans->vertex_data.push_back(data);
        }

        e = e->dual()->prev()->vertex()->first_edge();
    } while (e != pos->first_vertex()->first_edge());

    if (memory)
    {
        /* Bounding box for the deviation at the vertex */
        BoundingBox bbox(get_mesh().skeleton.get_read());
        /* Walk anticlockwise around the faces */
        do
        {
            if (e->face()->patch != 0)
                bbox &= get_deviations(e->face()) - delta;
            e = e->prev()->dual();
        } while (e != pos->first_vertex()->first_edge());

        ans->cost = 0.0;
        /* Consider max deviation over each face in turn */
        do
        {
            if (e->face()->patch != 0)
            {
                BoundingBox tmp = bbox;
                tmp |= get_deviations(e->face());
                ans->cost = max(ans->cost, tmp.length());
            }
            e = e->prev()->dual();
        } while (e != pos->first_vertex()->first_edge());
    }
    else
        ans->cost = length(delta);
    return ans;
}

MeshPositionIterator MeshSimplifierAps::collapse_influence(MeshPositionIterator pos, const MeshInfluenceCollapseData &col)
{
    if (memory)
    {
        VectorPosition delta = col.position - pos->position;
        MeshHalfEdgeIterator e = pos->first_vertex()->first_edge();

        BoundingBox bbox(get_mesh().skeleton.get_read());
        /* Obtain a tight bounding box for the vertex itself, as for
         * evaluate_influence_collapse.
         */
        do
        {
            if (e->face()->patch != 0)
                bbox &= get_deviations(e->face()) - delta;
            e = e->prev()->dual();
        } while (e != pos->first_vertex()->first_edge());
        /* Propagate this to all the faces */
        do
        {
            if (e->face()->patch != 0)
            {
                BoundingBox &dev = get_deviations(e->face());
                dev |= bbox;
            }
            e = e->prev()->dual();
        } while (e != pos->first_vertex()->first_edge());
    }
    return MeshSimplifier::collapse_influence(pos, col);
}

MeshSimplifier *mesh_simplifier_new_aps(Mesh &m, const string &method, bool do_influence, bool do_edge)
{
    if (method == "aps") return new MeshSimplifierAps(m, true, do_influence, do_edge);
    else return new MeshSimplifierAps(m, false, do_influence, do_edge);
}

/*! @} */
