/*! \addtogroup mesh
 * @{
 */

/*! \file iheap.h
 * \brief A priority queue where the priority may be changed
 */

#ifndef ALOD_MESH_IHEAP_H
#define ALOD_MESH_IHEAP_H

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <utility>
#include <vector>
#include <functional>
#include <algorithm>
#include <iterator>
#include <map>

//! \internal A reference wrapper for \c indexed_priority_queue
template<typename _ipq>
class _ipq_reference
{
private:
    typedef typename _ipq::mapped_type _mapped_type;
    typedef typename _ipq::iterator _iterator;
    _ipq &_parent;
    _iterator _it;

public:
    _ipq_reference(_ipq &__parent, _iterator __it) : _parent(__parent), _it(__it) {}

    operator _mapped_type() const { return _it->second; }

    _ipq_reference &operator =(const _mapped_type &__data)
    {
        _it->second = __data;
        _parent._bubble_up_down(_it);
        return *this;
    }

    _ipq_reference &operator =(const _ipq_reference<_ipq> &__x)
    {
        return *this = _mapped_type(__x);
    }

    bool operator ==(const _ipq_reference<_ipq> &__x) const
    {
        return _mapped_type(*this) == _mapped_type(__x);
    }
};

//! A priority queue where the priority may be changed
/*! This is similar to a \c map (in terms of <tt>operator []</tt>), but
 * one can extract the element with the largest value.
 * \arg \c _K the key type
 * \arg \c _V the value type
 * \arg \c _Sequence the underlying Random Access Container
 * \arg \c _Compare the Strict Weak Ordering using to compare values
 * \arg \c _Map the Unique Associative Container used to index the keys
 *
 * Undocumented member functions match the functions in either
 * \c std::map or \c std::priority_queue.
 */
template<
    typename _K,
    typename _V,
    typename _Sequence = std::vector<std::pair<_K, _V> >,
    typename _Compare = std::less<_V>,
    typename _Map = std::map<_K, typename std::iterator_traits<typename _Sequence::iterator>::difference_type> >
class indexed_priority_queue
{
public:
    typedef _K key_type;
    typedef _V mapped_type;
    typedef typename _Sequence::value_type value_type;
    typedef typename _Sequence::iterator iterator;
    typedef typename _Sequence::const_iterator const_iterator;
    typedef typename _Sequence::reference reference;
    typedef typename _Sequence::const_reference const_reference;
    typedef typename _Sequence::size_type size_type;
    typedef _Sequence container_type;

private:
    typedef _ipq_reference<indexed_priority_queue<_K, _V, _Sequence, _Compare, _Map> > _proxy_reference;
    friend class _ipq_reference<indexed_priority_queue<_K, _V, _Sequence, _Compare, _Map> >;

    typedef typename _Map::mapped_type _index_type;
    _Sequence _data;
    _Compare _comp;
    _Map _map;

    void _bubble_up(iterator __i)
    {
        value_type __value = *__i;
        iterator __first = _data.begin();
        _index_type __index = __i - __first;
        _index_type __parent = (__index - 1) / 2;
        while (__index > 0 && _comp((__first + __parent)->second, __value.second))
        {
            *(__first + __index) = *(__first + __parent);
            _map[(__first + __index)->first] = __index;
            __index = __parent;
            __parent = (__index - 1) / 2;
        }
        *(__first + __index) = __value;
        _map[__value.first] = __index;
    }

    void _bubble_down(iterator __i)
    {
        value_type __value = *__i;
        iterator __first = _data.begin();
        _index_type __len = _data.size();
        _index_type __index = __i - __first;
        _index_type __child = 2 * __index + 1;
        while (__child < __len)
        {
            if (__child + 1 < __len
                && _comp((__first + __child)->second, (__first + __child + 1)->second))
                __child++;
            if (_comp(__value.second, (__first + __child)->second))
            {
                *(__first + __index) = *(__first + __child);
                _map[(__first + __index)->first] = __index;
                __index = __child;
                __child = __index * 2 + 1;
            }
            else
                break;
        }
        *(__first + __index) = __value;
        _map[__value.first] = __index;
    }

    void _bubble_up_down(iterator __i)
    {
        _index_type __parent = (__i - _data.begin() - 1) / 2;
        if (__i != _data.begin()
            && _comp((_data.begin() + __parent)->second, __i->second))
            _bubble_up(__i);
        else
            _bubble_down(__i);
    }

    void _erase(_index_type __index)
    {
        iterator __first = _data.begin();
        iterator __last = _data.end() - 1;
        iterator __cur = __first + __index;
        _map.erase(__cur->first);
        if (__cur != __last)
        {
            *__cur = *__last;
            _map[__cur->first] = __index;
            _data.pop_back();
            _bubble_up_down(__cur);
        }
        else
            _data.pop_back();
    }

public:
    //! Constructs the data from \a __data, eliminating duplicates
    explicit indexed_priority_queue(const _Sequence &__data = _Sequence()) : _data(__data)
    {
        for (const_iterator __i = __data.begin(); __i != __data.end(); __i++)
        {
            typename _Map::iterator __pos = _map.find(__i->first);
            if (__pos != _map.end())
            {
                _map[__i->first] = _data.size();
                _data.push_back(*__i);
            }
        }
        for (iterator __i = _data.end(); __i != _data.begin(); __i--)
            _bubble_down(__i - 1);
    }

    bool empty() const { return _data.empty(); }
    size_type size() const { return _data.size(); }

    //! The highest priority pair in the priority queue
    const_reference top() const { return _data.front(); }

    //! Add an element to the priority queue
    void push(const value_type &__x)
    {
        _map[__x.first] = _data.size();
        _data.push_back(__x);
        _bubble_up(_data.end() - 1);
    }

    //! Remove the highest priority element from the queue
    void pop()
    {
        iterator __front = _data.begin();
        iterator __back = _data.end() - 1;
        _map.erase(__front->first);
        if (__front != __back)
        {
            *__front = *__back;
            _map[__front->first] = (_index_type) 0;
            _data.pop_back();
            _bubble_down(__front);
        }
        else
            _data.pop_back();
    }

    const mapped_type &operator[](const key_type &__key) const
    {
        typename _Map::iterator __pos = _map.find(__key);
        if (__pos == _map.end()) return top()->second; // have to return something...
        else return (_data.begin() + __pos->second)->second;
    }

    void erase(const key_type &__key)
    {
        typename _Map::iterator __loc = _map.find(__key);
        if (__loc != _map.end())
            _erase(__loc->second);
    }

    size_type count(const key_type &__key)
    {
        typename _Map::iterator __pos = _map.find(__key);
        return __pos != _map.end();
    }

    _proxy_reference operator[](const key_type &__key)
    {
        typename _Map::iterator __pos = _map.find(__key);
        if (__pos == _map.end())
        {
            push(value_type(__key, mapped_type()));
            __pos = _map.find(__key);
        }
        return _proxy_reference(*this, _data.begin() + __pos->second);
    }
};

#endif /* ALOD_MESH_IHEAP_H */

/*! @} */
