/*! \addtogroup mesh
 * @{
 */

/*! \file
 * \brief Functions to validate a mesh
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "manifold.h"
#include <utility>
#include <stdexcept>
#include <sstream>
#include <glib.h>

using namespace std;

//! Determines whether \a p is valid
/*! \a p must be a an element of \a s. It should first be compared to
 * \a end (a non-dereferencable iterator) to prevent an illegal dereference.
 */
template<typename Iterator, typename Iterator2>
static bool mesh_valid_pointer(Iterator p, Iterator2 end,
                               const typename id_set<Iterator>::type &s)
{
    if (p == end) return false;
    typename id_set<Iterator>::type::const_iterator i = s.find(p);
    if (i == s.end()) return false;
    // not trivial, since an id_set maps on id's, not actual values
    return (*i == p);
}

//! Throws a \c mesh_nonmanifold exception with message \a msg if \a expr is false
static void mesh_manifold_assert(bool expr, const string &msg) throw(mesh_nonmanifold)
{
    if (!expr)
        throw mesh_nonmanifold(msg);
}

//! Returns number of sharing vertices, capped at 3
static int mesh_position_vertex_count(MeshPositionConstIterator pos)
{
    int count = 0;
    MeshVertexConstIterator v = pos->first_vertex();
    do
    {
        if (++count >= 3) return count;
        v = v->first_edge()->dual()->prev()->vertex();
    } while (v != pos->first_vertex());
    return count;
}

/*! See doc/mesh_invariant.txt for details of the invariants that
 * are checked.
 *
 * We try to check the conditions in a way than will guarantee
 * correct termination even in the event of a broken structure.
 * \pre \a level >= \c MESH_LEVEL_MANIFOLD
 */
void mesh_check_manifold(const Mesh &m, MeshLevel level) throw(mesh_nonmanifold)
{
    /* We build sets of valid pointers. Since id_set tests for
     * equality by ID, rather than by iterator, we must still explicitly
     * check the values. We throw in a check ID uniqueness as well.
     */
    id_set<MeshPositionConstIterator>::type position_it;
    id_set<MeshVertexConstIterator>::type vertex_it;
    id_set<MeshHalfEdgeConstIterator>::type edge_it;
    id_set<MeshFaceConstIterator>::type face_it;
    bool unique;

    g_assert(level >= MESH_LEVEL_MANIFOLD);
    for (MeshPositionConstIterator p = m.positions.total_begin(); p != m.positions.total_end(); p++)
    {
        unique = position_it.insert(p).second;
        mesh_manifold_assert(unique, "non-unique position ID");
    }
    for (MeshVertexConstIterator v = m.vertices.total_begin(); v != m.vertices.total_end(); v++)
    {
        unique = vertex_it.insert(v).second;
        mesh_manifold_assert(unique, "non-unique vertex ID");
    }
    for (MeshHalfEdgeConstIterator e = m.edges.total_begin(); e != m.edges.total_end(); e++)
    {
        unique = edge_it.insert(e).second;
        mesh_manifold_assert(unique, "non-unique edge ID");
    }
    for (MeshFaceConstIterator f = m.faces.total_begin(); f != m.faces.total_end(); f++)
    {
        unique = face_it.insert(f).second;
        mesh_manifold_assert(unique, "non-unique face ID");
    }

    /* There must be a mesh_manifold_assert statement below corresponding
     * to each invariant listed in doc/mesh_invariant.txt.
     */

    // First check that all pointers point to legal elements
    for (MeshPositionConstIterator p = m.positions.whole_begin(); p != m.positions.whole_end(); p++)
        mesh_manifold_assert(mesh_valid_pointer(p->first_vertex(), mesh_null_vertex,
                                                vertex_it),
                             "p->first_vertex is invalid");
    for (MeshVertexConstIterator v = m.vertices.whole_begin(); v != m.vertices.whole_end(); v++)
    {
        mesh_manifold_assert(mesh_valid_pointer(v->position(), mesh_null_position,
                                                position_it),
                             "v->position is invalid");
        mesh_manifold_assert(mesh_valid_pointer(v->first_edge(), mesh_null_half_edge,
                                                edge_it),
                             "v->first_edge is invalid");
    }
    for (MeshHalfEdgeConstIterator e = m.edges.whole_begin(); e != m.edges.whole_end(); e++)
    {
        mesh_manifold_assert(mesh_valid_pointer(e->prev(), mesh_null_half_edge, edge_it),
                             "e->prev is invalid");
        mesh_manifold_assert(mesh_valid_pointer(e->next(), mesh_null_half_edge, edge_it),
                             "e->next is invalid");
        mesh_manifold_assert(mesh_valid_pointer(e->dual(), mesh_null_half_edge, edge_it),
                             "e->dual is invalid");
        mesh_manifold_assert(mesh_valid_pointer(e->vertex(), mesh_null_vertex, vertex_it),
                             "e->vertex is invalid");
        mesh_manifold_assert(mesh_valid_pointer(e->face(), mesh_null_face, face_it),
                             "e->face is invalid");
    }
    for (MeshFaceConstIterator f = m.faces.whole_begin(); f != m.faces.whole_end(); f++)
    {
        mesh_manifold_assert(mesh_valid_pointer(f->edges(), mesh_null_half_edge, edge_it),
                             "f->edges is invalid");
    }

    // check that we have the right number of faces for the edges
    mesh_manifold_assert(m.edges.whole_size()
                         == 3 * m.faces.whole_size(),
                         "not 3 edges per face");

    // Recover the memory
    position_it.clear();
    vertex_it.clear();
    edge_it.clear();
    face_it.clear();

    // Next, do the "easy" tests (no loops, no risk of invalid pointers)
    for (MeshPositionConstIterator p = m.positions.whole_begin(); p != m.positions.whole_end(); p++)
    {
        mesh_manifold_assert(p->first_vertex()->position() == p,
                             "p->first_vertex->position != p");
        if (level >= MESH_LEVEL_PATCHES && p->patch_corner)
            mesh_manifold_assert(mesh_position_vertex_count(p) >= 2,
                                 "patch_corner set for interior position");
        if (mesh_position_vertex_count(p) >= 3 && level >= MESH_LEVEL_PATCHES)
            mesh_manifold_assert(p->patch_corner,
                                 "patch_corner not set for boundary meeting point");
    }
    for (MeshVertexConstIterator v = m.vertices.whole_begin(); v != m.vertices.whole_end(); v++)
    {
        mesh_manifold_assert(v->first_edge()->next()->vertex() == v,
                             "v->first_edge->next->vertex != v");
        mesh_manifold_assert(v->first_edge()->face()->patch != 0
                             || v->position()->first_vertex() == v,
                             "v->position->first_vertex != v on hole");
    }
    for (MeshHalfEdgeConstIterator e = m.edges.whole_begin(); e != m.edges.whole_end(); e++)
    {
        mesh_manifold_assert(e->prev()->next() == e, "e->prev->next != e");
        mesh_manifold_assert(e->next()->prev() == e, "e->next->prev != e");
        mesh_manifold_assert(e->next()->face() == e->face(), "e->next->face != e->face");
        mesh_manifold_assert(e->next() != e, "e->next == e");
        mesh_manifold_assert(e->next()->next()->next() == e, "non-triangle");
        mesh_manifold_assert(e->next()->vertex() != e->vertex(), "degenerate triangle");
        mesh_manifold_assert(e->face()->patch != 0
                             || e->prev()->dual()->prev()->vertex() == e->vertex()
                             || e->vertex()->first_edge() == e->prev(),
                             "incorrect value for first_edge on hole");
        if (level >= MESH_LEVEL_PATCHES)
        {
            mesh_manifold_assert(e->prev()->dual()->prev()->vertex() == e->vertex()
                                 || e->vertex()->first_edge() == e->prev(),
                                 "incorrect value for first_edge");
            mesh_manifold_assert((e->next()->vertex() == e->dual()->prev()->vertex())
                                 ^ (e->face()->patch != e->dual()->face()->patch),
                                 "inconsistent vertices in a patch");
            mesh_manifold_assert(e->face()->patch == e->vertex()->first_edge()->face()->patch,
                                 "vertex shared across patches");
            mesh_manifold_assert(e->face()->patch != e->dual()->face()->patch
                                 || e->face()->material() == e->dual()->face()->material(),
                                 "material changes within patch");
        }
        mesh_manifold_assert(e->next()->vertex()->position() == e->dual()->prev()->vertex()->position(),
                             "illegal dual pointer");
        mesh_manifold_assert(e->prev()->vertex()->position() == e->dual()->next()->vertex()->position(),
                             "illegal dual pointer");
    }
    for (MeshFaceConstIterator f = m.faces.whole_begin(); f != m.faces.whole_end(); f++)
        mesh_manifold_assert(f->edges()->face() == f, "f->edges->face != f");

    // now we do the trickier tests, starting with edge uniqueness
    typedef pair<MeshPositionConstIterator, MeshPositionConstIterator> Edge;
    id_set<Edge>::type edge_set;
    for (MeshHalfEdgeConstIterator e = m.edges.whole_begin(); e != m.edges.whole_end(); e++)
    {
        Edge me(e->next()->vertex()->position(), e->prev()->vertex()->position());
        ostringstream label1, label2;
        label1 << e->next()->vertex()->position()->position;
        label2 << e->prev()->vertex()->position()->position;
        mesh_manifold_assert(!edge_set.count(me),
                             "duplicate edge from " + label1.str() + " to " + label2.str());
        edge_set.insert(me);
    }
    edge_set.clear(); // recover memory

    /* (patch, position) must uniquely define vertex
     * We don't have a map that takes an int and MeshElement, so
     * we explicit map on the position ID.
     */
    if (level >= MESH_LEVEL_PATCHES)
    {
        typedef fast_map<pair<int, size_t>, MeshVertexConstIterator>::type VertexMap;
        VertexMap vertex_map;
        for (MeshHalfEdgeConstIterator e = m.edges.whole_begin(); e != m.edges.whole_end(); e++)
        {
            pair<int, size_t> me(e->vertex()->position()->id, e->face()->patch);

            VertexMap::iterator i = vertex_map.find(me);
            if (i == vertex_map.end())
                i = vertex_map.insert(make_pair(me, e->vertex())).first;
            mesh_manifold_assert(i->second == e->vertex(),
                                 "patch border duplicates vertex");
        }
        vertex_map.clear(); // recover memory
    }

    /* Position must have connected neighbourhood. We check this by a counting
     * argument. For efficiency, we walk the edges rather than counting up
     * to m.edges.size
     */
    MeshHalfEdgeConstIterator counter = m.edges.whole_begin();
    for (MeshPositionConstIterator p = m.positions.whole_begin(); p != m.positions.whole_end(); p++)
    {
        MeshHalfEdgeConstIterator e = p->first_vertex()->first_edge();
        do
        {
            mesh_manifold_assert(counter != m.edges.whole_end(),
                                 "double-counted edge (probably a bug)");
            counter++;
            e = e->dual()->next();
        } while (e != p->first_vertex()->first_edge());
    }
    mesh_manifold_assert(counter == m.edges.whole_end(),
                         "disconnected vertex neighbourhood (touching points?)");

    /* Extra checks: make sure that the hole vs real distinction is correct */
    for (MeshPositionConstIterator p = m.positions.hole_begin(); p != m.positions.hole_end(); p++)
        g_assert(!mesh_true_position(p));
    for (MeshPositionConstIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
        g_assert(mesh_true_position(p));
    for (MeshVertexConstIterator v = m.vertices.hole_begin(); v != m.vertices.hole_end(); v++)
        g_assert(!mesh_true_vertex(v));
    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
        g_assert(mesh_true_vertex(v));
}

bool mesh_manifold(const Mesh &m, MeshLevel level)
{
    try
    {
        mesh_check_manifold(m, level);
    }
    catch (mesh_nonmanifold &e)
    {
        g_warning("non-manifold mesh: %s", e.what());
        return false;
    }

    return true;
}

/*! @} */
