/*! \addtogroup mesh
 * @{
 */

/*! \file
 * \brief %Mesh data structures
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "common/maputils.h"
#include <cstddef>
#include <vector>
#include <utility>
#include <map>
#include <stack>
#include <sstream>
#include <glib.h>

using namespace std;

size_t MeshElement::id_pool = 0;

// a convenient container to define the null iterator types
static Mesh null_mesh;
const MeshPositionIterator mesh_null_position = null_mesh.positions.total_end();
const MeshVertexIterator mesh_null_vertex = null_mesh.vertices.total_end();
const MeshHalfEdgeIterator mesh_null_half_edge = null_mesh.edges.total_end();
const MeshFaceIterator mesh_null_face = null_mesh.faces.total_end();
const MeshTextureIterator mesh_null_texture = null_mesh.textures.end();
const MeshTangentMapIterator mesh_null_tangent_map = null_mesh.tangent_maps.end();
const MeshMaterialIterator mesh_null_material = null_mesh.materials.end();

template<typename T> MeshCollection<T>::MeshCollection()
{
    real_start = Base::end();
    lod_start = Base::end();
}

template<typename T> MeshCollection<T>::MeshCollection(const MeshCollection<T> &c) : Base()
{
    *this = c;
}

template<typename T> MeshCollection<T> &MeshCollection<T>::operator =(const MeshCollection<T> &c)
{
    clear();
    Base::insert(Base::begin(), c.lod_begin(), c.lod_end());
    lod_start = Base::begin();
    Base::insert(Base::begin(), c.real_begin(), c.real_end());
    real_start = Base::begin();
    Base::insert(Base::begin(), c.hole_begin(), c.hole_end());
    return *this;
}

template<typename T> typename MeshCollection<T>::iterator MeshCollection<T>::insert_hole(const MeshCollection<T>::value_type &v)
{
    return insert(Base::begin(), v);
}

template<typename T> typename MeshCollection<T>::iterator MeshCollection<T>::insert_real(const MeshCollection<T>::value_type &v)
{
    return real_start = insert(real_start, v);
}

template<typename T> typename MeshCollection<T>::iterator MeshCollection<T>::insert_lod(const MeshCollection<T>::value_type &v)
{
    iterator i = insert(lod_start, v);
    if (real_start == lod_start) real_start = i;
    return lod_start = i;
}

template<typename T> void MeshCollection<T>::erase(MeshCollection<T>::iterator pos)
{
    if (pos == real_start) real_start++;
    if (pos == lod_start) lod_start++;
    Base::erase(pos);
}

template<typename T> void MeshCollection<T>::move_to_hole(MeshCollection<T>::iterator pos)
{
    if (pos == real_start) real_start++;
    if (pos == lod_start) lod_start++;
    Base::splice(Base::begin(), *this, pos);
}

template<typename T> void MeshCollection<T>::move_to_real(MeshCollection<T>::iterator pos)
{
    if (pos == real_start) real_start++;
    if (pos == lod_start) lod_start++;
    Base::splice(real_start, *this, pos);
    real_start = pos;
}

template<typename T> void MeshCollection<T>::move_to_lod(MeshCollection<T>::iterator pos)
{
    if (pos == real_start) real_start++;
    if (pos == lod_start) lod_start++;
    Base::splice(lod_start, *this, pos);
    if (real_start == lod_start) real_start = pos;
    lod_start = pos;
}

template<typename T> void MeshCollection<T>::clear()
{
    Base::clear();
    real_start = Base::end();
    lod_start = Base::end();
}

template<class T> void MeshCollection<T>::clear_lod()
{
    Base::erase(lod_start, Base::end());
    lod_start = Base::end();
}

MeshTexture::MeshTexture() : MeshElement(), _data(NULL), width(0), height(0), channels(0), sign(false)
{
}

MeshTexture::~MeshTexture()
{
    if (_data) delete[] _data;
}

MeshTexture::MeshTexture(const MeshTexture &t) : MeshElement()
{
    width = t.width;
    height = t.height;
    channels = t.channels;
    sign = t.sign;
    id = t.id;
    if (t._data)
    {
        _data = new guint8[width * height * channels];
        copy(t._data, t._data + width * height * channels, _data);
    }
    else
        _data = NULL;
}

MeshTexture &MeshTexture::operator =(const MeshTexture &t)
{
    if (_data)
    {
        delete[] _data;
        _data = NULL;
    }
    width = t.width;
    height = t.height;
    channels = t.channels;
    sign = t.sign;
    id = t.id;
    if (t._data)
    {
        _data = new guint8[width * height * channels];
        copy(t._data, t._data + width * height * channels, _data);
    }
    return *this;
}

void MeshTexture::set_size(std::size_t w, std::size_t h, std::size_t c)
{
    if (_data) delete[] _data;
    width = w;
    height = h;
    channels = c;
    _data = new guint8[w * h * c];
}

MeshTangentMap::MeshTangentMap() : MeshElement(), _data(NULL), width(0), height(0)
{
}

MeshTangentMap::~MeshTangentMap()
{
    if (_data) delete[] _data;
}

MeshTangentMap::MeshTangentMap(const MeshTangentMap &tm) : MeshElement()
{
    width = tm.width;
    height = tm.height;
    id = tm.id;
    if (tm._data)
    {
        _data = new MeshTangentMapEntry[width * height];
        copy(tm._data, tm._data + width * height, _data);
    }
    else
        _data = NULL;
}

MeshTangentMap &MeshTangentMap::operator =(const MeshTangentMap &tm)
{
    if (_data)
    {
        delete[] _data;
        _data = NULL;
    }
    width = tm.width;
    height = tm.height;
    id = tm.id;
    if (tm._data)
    {
        _data = new MeshTangentMapEntry[width * height];
        copy(tm._data, tm._data + width * height, _data);
    }
    return *this;
}

void MeshTangentMap::set_size(std::size_t w, std::size_t h)
{
    if (_data) delete[] _data;
    width = w;
    height = h;
    _data = new MeshTangentMapEntry[w * h];
}

Mesh::Mesh() : level(MESH_LEVEL_NONE)
{
}

Mesh::Mesh(const Mesh &m)
{
    *this = m;
}

/* This is a particularly complicated function because the mesh structure
 * uses iterators internally. The iterators need to be mapped from the
 * old structures to the new.
 */
Mesh &Mesh::operator =(const Mesh &m)
{
    fast_map<size_t, MeshPositionIterator>::type position_map;
    fast_map<size_t, MeshVertexIterator>::type vertex_map;
    fast_map<size_t, MeshHalfEdgeIterator>::type edge_map;
    fast_map<size_t, MeshFaceIterator>::type face_map;
    fast_map<size_t, MeshTextureIterator>::type texture_map;
    fast_map<size_t, MeshTangentMapIterator>::type tangent_map_map;
    fast_map<size_t, MeshMaterialIterator>::type material_map;

    skeleton = m.skeleton;
    positions = m.positions;
    vertices = m.vertices;
    edges = m.edges;
    faces = m.faces;
    textures = m.textures;
    tangent_maps = m.tangent_maps;
    materials = m.materials;
    collapses = m.collapses;
    level = m.level;
    // fix up the internal pointers
    for (MeshPositionIterator p = positions.total_begin(); p != positions.total_end(); p++)
        position_map[p->id] = p;
    for (MeshVertexIterator v = vertices.total_begin(); v != vertices.total_end(); v++)
        vertex_map[v->id] = v;
    for (MeshHalfEdgeIterator e = edges.total_begin(); e != edges.total_end(); e++)
        edge_map[e->id] = e;
    for (MeshFaceIterator f = faces.total_begin(); f != faces.total_end(); f++)
        face_map[f->id] = f;
    for (MeshTextureIterator t = textures.begin(); t != textures.end(); t++)
        texture_map[t->id] = t;
    for (MeshTangentMapIterator tm = tangent_maps.begin(); tm != tangent_maps.end(); tm++)
        tangent_map_map[tm->id] = tm;
    for (MeshMaterialIterator s = materials.begin(); s != materials.end(); s++)
        material_map[s->id] = s;

    for (MeshPositionIterator p = positions.total_begin(); p != positions.total_end(); p++)
    {
        if (p->first_vertex() != mesh_null_vertex)
            p->first_vertex() = vertex_map[p->first_vertex()->id];
        p->position.set_inner(skeleton.get_read().get_ssd_inner());
    }
    for (MeshVertexIterator v = vertices.total_begin(); v != vertices.total_end(); v++)
    {
        if (v->position() != mesh_null_position)
            v->position() = position_map[v->position()->id];
        if (v->first_edge() != mesh_null_half_edge)
            v->first_edge() = edge_map[v->first_edge()->id];
    }
    for (MeshHalfEdgeIterator e = edges.total_begin(); e != edges.total_end(); e++)
    {
        if (e->dual() != mesh_null_half_edge)
            e->dual() = edge_map[e->dual()->id];
        if (e->prev() != mesh_null_half_edge)
            e->prev() = edge_map[e->prev()->id];
        if (e->next() != mesh_null_half_edge)
            e->next() = edge_map[e->next()->id];
        if (e->vertex() != mesh_null_vertex)
            e->vertex() = vertex_map[e->vertex()->id];
        if (e->face() != mesh_null_face)
            e->face() = face_map[e->face()->id];
    }
    for (MeshFaceIterator f = faces.total_begin(); f != faces.total_end(); f++)
    {
        if (f->edges() != mesh_null_half_edge)
            f->edges() = edge_map[f->edges()->id];
        if (f->material() != mesh_null_material)
            f->material() = material_map[f->material()->id];
    }
    for (MeshMaterialIterator s = materials.begin(); s != materials.end(); s++)
    {
        if (s->texture() != mesh_null_texture)
            s->texture() = texture_map[s->texture()->id];
        if (s->tangent_map() != mesh_null_tangent_map)
            s->tangent_map() = tangent_map_map[s->tangent_map()->id];
    }
    for (MeshCollapseIterator c = collapses.begin(); c != collapses.end(); c++)
    {
        switch (c->is_edge_collapse)
        {
        case true:
            if (c->edge_collapse.p1() != mesh_null_position)
                c->edge_collapse.p1() = position_map[c->edge_collapse.p1()->id];
            if (c->edge_collapse.p2() != mesh_null_position)
                c->edge_collapse.p2() = position_map[c->edge_collapse.p2()->id];
            if (c->edge_collapse.vl() != mesh_null_vertex)
                c->edge_collapse.vl() = vertex_map[c->edge_collapse.vl()->id];
            if (c->edge_collapse.vr() != mesh_null_vertex)
                c->edge_collapse.vr() = vertex_map[c->edge_collapse.vr()->id];
        case false:
            for (vector<MeshInfluenceCollapseElement>::iterator i = c->influence_collapse.begin();
                 i != c->influence_collapse.end(); i++)
            {
                if (i->before() != mesh_null_vertex)
                    i->before() = vertex_map[i->before()->id];
                if (i->after() != mesh_null_vertex)
                    i->after() = vertex_map[i->after()->id];
            }
        }
    }

    return *this;
}

void Mesh::clear()
{
    skeleton.get_write().clear();
    positions.clear();
    vertices.clear();
    edges.clear();
    faces.clear();
    textures.clear();
    tangent_maps.clear();
    materials.clear();
    collapses.clear();

    level = MESH_LEVEL_NONE;
}

void Mesh::clear_lod()
{
    positions.clear_lod();
    vertices.clear_lod();
    edges.clear_lod();
    faces.clear_lod();
    collapses.clear();
}

void mesh_clear(Mesh &m)
{
    m.clear();
}

bool mesh_true_face(MeshFaceConstIterator f)
{
    return f->patch != 0;
}

bool mesh_true_vertex(MeshVertexConstIterator v)
{
    return (v->first_edge() == mesh_null_half_edge || v->first_edge()->face()->patch);
}

bool mesh_true_position(MeshPositionConstIterator p)
{
    return (p->first_vertex() == mesh_null_vertex
            || p->first_vertex()->first_edge() == mesh_null_half_edge
            || p->first_vertex()->first_edge()->face()->patch
            || p->first_vertex()->first_edge()->dual()->face()->patch);
}

size_t mesh_collapses(const Mesh &m)
{
    return distance(m.collapses_begin(), m.collapses_end());
}

double mesh_face_double_area(MeshFaceConstIterator f)
{
    const VectorPosition &pos0 = f->edges()->vertex()->position()->position;
    const VectorPosition &pos1 = f->edges()->next()->vertex()->position()->position;
    const VectorPosition &pos2 = f->edges()->next()->next()->vertex()->position()->position;
    return double_area(pos1 - pos0, pos2 - pos0);
}

double mesh_texture_double_area(MeshFaceConstIterator f)
{
    const Vector2 &uv0 = f->edges()->vertex()->data.texcoords;
    const Vector2 &uv1 = f->edges()->next()->vertex()->data.texcoords;
    const Vector2 &uv2 = f->edges()->next()->next()->vertex()->data.texcoords;
    return (uv1[0] - uv0[0]) * (uv2[1] - uv0[1])
        - (uv2[0] - uv0[0]) * (uv1[1] - uv0[1]);
}

pair<Vector3, Vector3> mesh_bounding_box_3d(const Mesh &m)
{
    if (m.positions.real_begin() == m.positions.real_end())
        return make_pair(Vector3(), Vector3());

    pair<Vector3, Vector3> bounds(
                                  Vector3(HUGE_VAL, HUGE_VAL, HUGE_VAL),
                                  Vector3(-HUGE_VAL, -HUGE_VAL, -HUGE_VAL)
                                 );

    const Skeleton &skeleton = m.skeleton.get_read();
    for (MeshPositionConstIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
    {
        Vector3 pos = skeleton.ssd_to_3d(p->position);
        bounds.first = vector_min(bounds.first, pos);
        bounds.second = vector_max(bounds.second, pos);
    }
    return bounds;
}

/*! The centre is taken as the area-weighted centroid of the base level of the mesh */
pair<VectorPosition, double> mesh_bounding_sphere_as(const Mesh &m)
{
    const Skeleton &skeleton = m.skeleton.get_read();
    VectorPosition centroid = VectorPosition(skeleton);
    double weight = 0.0;
    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
    {
        VectorPosition c(skeleton);
        c.zero();
        MeshHalfEdgeConstIterator e = f->edges();
        for (int i = 0; i < 3; i++, e = e->next())
            c += e->vertex()->position()->position;
        c /= 3.0;
        double area2 = mesh_face_double_area(f);
        centroid += c * area2;
        weight += area2;
    }
    centroid /= weight;

    double radius = 0.0;
    for (MeshPositionConstIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
        radius = max(radius, skeleton.inf_metric(p->position - centroid));

    return make_pair(centroid, radius);
}

BoundingSphere mesh_bounding_sphere_3d(const Mesh &m)
{
    const Skeleton &skeleton = m.skeleton.get_read();
    /* Get the centroid as a starting point */
    VectorPosition centroid = mesh_bounding_sphere_as(m).first;

    /* Push everything towards the root. This is similar to
     * Skeleton::ssd_to_3d_expected, but ignores any estimated expectations
     * since average-case behaviour is irrelevant.
     */
    for (size_t i = skeleton.size() - 1; i > 0; i--)
    {
        const SkeletonBone &bone = skeleton.get_bone(i);
        Matrix3 l = matrix4_to_matrix3(bone.rest_transform);
        l = bone.max_singular * polar_rotation(l);
        Vector3 t = bone.translate_bounds.centre;
        if (!(skeleton.get_bone(i).flags & BONE_FLAG_FIXED_ROTATE))
            l.zero();

        Vector4 up;
        for (int j = 0; j < 4; j++)
            up[j] = centroid[4 * i + j];
        up = matrix3_translate3_to_matrix4(l, t) * up;
        for (int j = 0; j < 4; j++)
            centroid[4 * bone.parent + j] += up[j];
    }

    Vector3 centre(centroid[0], centroid[1], centroid[2]);
    VectorPosition centre_as = VectorPosition(skeleton);
    for (int j = 0; j < 3; j++)
        centre_as[j] = centre[j];
    centre_as[3] = 1.0;

    double radius = 0.0;
    for (MeshPositionConstIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
        radius = max(radius, skeleton.inf_metric(p->position - centre_as));
    return BoundingSphere(centre, radius);
}

/* Scales the translation part of m by s */
static void scale_translation(Matrix4 &m, double s)
{
    m[0][3] *= s;
    m[1][3] *= s;
    m[2][3] *= s;
}

/*! Handles uniform scaling, rotations and translations. Rather than
 * introducing scaling into the bones, we scale the translation parts
 * of the local matrices and handle the scaling in the coordinates.
 * The translation and rotation are applied to the bones immediately
 * under the root, so that the transformation happens in model space
 * rather than world space.
 * \pre transform must have uniform scaling
 */
void mesh_transform(Mesh &m, const Matrix4 &transform)
{
    Skeleton &skeleton = m.skeleton.get_write();
    size_t sz = skeleton.size();
    size_t nf = skeleton.num_frames();
    vector<SkeletonBone> &bones = skeleton.bones;
    vector<vector<Matrix4> > &frames = skeleton.frames;
    vector<Matrix4> expect_G(sz);  // Expected value after munging scaling but before translation-and-rotate
    MAT *P = skeleton.get_P();
    Matrix3 rot;

    // Extract scale factor
    double s = cbrt(determinant(transform));
    g_assert(s >= -1e-6);
    if (fabs(s) > 1e-6)
        rot = matrix4_to_matrix3(transform) * (1.0 / s);
    Vector3 offset = matrix4_to_translate3(transform);
    Matrix4 rigid = matrix3_translate3_to_matrix4(rot, offset);

    bones[0].tail = vector4_to_vector3(transform * vector3_to_vector4(bones[0].tail, 1.0));
    expect_G[0].identity();
    for (size_t i = 1; i < sz; i++)
    {
        int p = bones[i].parent;
        /* Pre-scaling E[G] */
        for (int r = 0; r < 4; r++)
            for (int c = 0; c < 4; c++)
                expect_G[i][r][c] = m_entry(P, r, 4 * i + c);
        /* Push scaling into translation parts */
        scale_translation(bones[i].rest_transform, s);
        scale_translation(bones[i].L, s);
        scale_translation(bones[i].expect_L, s);
        scale_translation(bones[i].expect_GTG, s);
        for (int j = 0; j < 3; j++)
            bones[i].expect_GTG[3][j] *= s;
        bones[i].expect_LTGTGL *= s * s;
        bones[i].expect_GTG[3][3] = 1.0 + (bones[i].expect_GTG[3][3] - 1.0) * s * s;
        bones[i].translate_bounds.centre *= s;
        bones[i].translate_bounds.radius *= s;
        bones[i].tail *= s;
        for (size_t k = 0; k < nf; k++)
            scale_translation(frames[k][i], s);
        for (int j = 0; j < 3; j++)
            expect_G[i][j][3] *= s;

        /* Apply rotate and translate to sub-roots. Update only pure-L
         * expectations and bounds. */
        if (p == 0)
        {
            bones[i].rest_transform = rigid * bones[i].rest_transform;
            bones[i].L = rigid * bones[i].L;
            bones[i].expect_L = rigid * bones[i].expect_L;
            bones[i].translate_bounds.centre = vector4_to_vector3(rigid * vector3_to_vector4(bones[i].translate_bounds.centre, 1.0));
            bones[i].tail = rot * bones[i].tail;
            for (size_t k = 0; k < nf; k++)
                frames[k][i] = rigid * frames[k][i];
        }

        /* Apply rotation and translation to global expectations */
        Vector3 GTGt_adj = transpose(matrix4_to_matrix3(expect_G[i]))
            * transpose(rot) * offset;
        double GTGw_adj = 2.0 * offset * rot * matrix4_to_translate3(expect_G[i])
            + offset * offset;
        for (int j = 0; j < 3; j++)
        {
            bones[i].expect_GTG[j][3] += GTGt_adj[j];
            bones[i].expect_GTG[3][j] += GTGt_adj[j];
        }
        bones[i].expect_GTG[3][3] += GTGw_adj;
        bones[i].expect_LTGTGL = bones[i].expect_GTG[3][3] - bones[p].expect_GTG[3][3]
            - 2.0 * (matrix4_to_translate3(bones[p].expect_GTG) * matrix4_to_translate3(bones[i].expect_L));
    }

    for (MeshPositionIterator p = m.positions.all_begin();
         p != m.positions.all_end(); p++)
    {
        VectorPosition &pos = p->position;
        for (VectorPosition::const_iterator i = pos.begin(); i != pos.end(); i++)
        {
            if (i->index != 0)
                for (int j = 0; j < 3; j++)
                    pos[i->index + j] = i->values[j] * s;
            else
            {
                Vector4 pos4(i->values[0], i->values[1], i->values[2], i->values[3]);
                pos4 = transform * pos4;
                for (int j = 0; j < 3; j++)
                    pos[i->index + j] = pos4[j];
            }
        }
    }

    /* Apply scale to model offsets */
    for (size_t i = 0; i < 3; i++)
    {
        bones[0].rest_transform[i][3] *= s;
        bones[0].L[i][3] *= s;
        for (size_t k = 0; k < nf; k++)
            frames[k][0][i][3] *= s;
    }

    skeleton.update_G();
    skeleton.update_M();
    skeleton.update_P();
    skeleton.update_C();
}

int mesh_position_patch_valence(MeshPositionConstIterator p)
{
    int answer = 0;
    MeshVertexConstIterator start = p->first_vertex();
    MeshVertexConstIterator v = start;
    do
    {
        answer++;
        v = v->first_edge()->dual()->prev()->vertex();
    } while (v != start);
    return answer;
}


//! Fix up the dual structures.
/*! This sets the dual structures in a mesh. It is not yet guaranteed
 * that holes have been filled. Edges with no dual set the dual to
 * mesh_null_half_edge.
 */
static void mesh_set_neighbours(Mesh &m) throw(mesh_nonmanifold)
{
    typedef pair<MeshPositionIterator, MeshPositionIterator> Edge;
    id_map<Edge, MeshHalfEdgeIterator>::type edges;
    id_map<Edge, MeshHalfEdgeIterator>::type::iterator hedge;

    /* The coding here isn't pretty, but
     * a) it is efficient
     * b) it avoids the errors generated by using -D_GLIBCXX_DEBUG when
     * using operator [] - note that it gets upset even if it doesn't have
     * to insert an element, due to the *internal* implementation.
     */
    for (MeshHalfEdgeIterator e = m.edges.total_begin(); e != m.edges.total_end(); e++)
    {
        MeshHalfEdgeIterator cx = e->next();
        MeshHalfEdgeIterator cy = e->prev();
        MeshPositionIterator x = cx->vertex()->position();
        MeshPositionIterator y = cy->vertex()->position();

        if (edges.count(Edge(x, y)))
            throw mesh_nonmanifold("duplicated edge");
        hedge = edges.find(Edge(y, x));
        if (hedge != edges.end())
        {
            e->dual() = hedge->second;
            e->dual()->dual() = e;
        }
        else
            e->dual() = mesh_null_half_edge;
        edges.insert(make_pair(Edge(x, y), e));
    }
}

//! Tile the holes in the mesh with patch-0 triangles
/*! It is assumed that the dual fields have been set (by
 * \c mesh_set_neighbours), but \c mesh_set_edges
 * has not yet been run.
 *
 * To be safe, a new position is added. We would prefer to
 * add a triangle strip, but some of the edges may already
 * be used. If the high valence of the new position is a
 * problem, it can be avoided by instead constructing a
 * triangle strip then splitting problematic edges.
 */
static void mesh_tile_holes(Mesh &m) throw(mesh_nonmanifold)
{
    /* In a badly formed (non-manifold) mesh, the algorithm could go
     * into an infinite loop. However, on a well-formed mesh it cannot
     * visit any half-edge more than once. Rather than a counter, we
     * use an iterator so that we don't have to invoke list::size.
     */
    MeshHalfEdgeIterator loop_detector = m.edges.total_begin();

    // We find the first edge of each hole, then fill in from there.
    for (MeshHalfEdgeIterator e = m.edges.real_begin(); e != m.edges.real_end(); e++)
        if (e->dual() == mesh_null_half_edge)
        {
            MeshPositionIterator np = m.positions.insert_hole(MeshPosition(m));
            MeshVertexIterator nv = m.vertices.insert_hole(MeshVertex(m));
            nv->position() = np;
            // We use nv->first_edge to tie up the ends once we've gone
            // the full circle.
            nv->first_edge() = mesh_null_half_edge;

            MeshHalfEdgeIterator e0 = mesh_null_half_edge; // previous e1
            MeshHalfEdgeIterator e1 = e; // current boundary edge
            int count = 0;
            while (true)
            {
                MeshFaceIterator nf;
                MeshHalfEdgeIterator ne[3];
                nf = m.faces.insert_hole(MeshFace());
                nf->patch = 0;
                nf->material() = mesh_null_material;
                count++;
                np->position += e1->next()->vertex()->position()->position;
                for (int i = 0; i < 3; i++)
                {
                    ne[i] = m.edges.insert_hole(MeshHalfEdge());
                    ne[i]->face() = nf;
                    ne[i]->dual() = mesh_null_half_edge;
                }
                nf->edges() = ne[0];
                ne[0]->next() = ne[1]; ne[1]->next() = ne[2]; ne[2]->next() = ne[0];
                ne[0]->prev() = ne[2]; ne[1]->prev() = ne[0]; ne[2]->prev() = ne[1];
                ne[0]->dual() = e1;
                e1->dual() = ne[0];
                ne[0]->vertex() = nv;
                if (e0 != mesh_null_half_edge)
                {
                    ne[2]->dual() = e0->dual()->next();
                    e0->dual()->next()->dual() = ne[2];
                    ne[1]->vertex() = e0->dual()->prev()->vertex();
                }
                else
                {
                    nv->first_edge() = ne[2];
                    ne[1]->vertex() = m.vertices.insert_hole(MeshVertex(m));
                    ne[1]->vertex()->position() = e1->prev()->vertex()->position();
                }
                if (e0 != mesh_null_half_edge
                    && nv->first_edge()->prev()->vertex()->position() == e1->next()->vertex()->position())
                {
                    ne[1]->dual() = nv->first_edge();
                    nv->first_edge()->dual() = ne[1];
                    ne[2]->vertex() = nv->first_edge()->prev()->vertex();
                    break;
                }
                else
                {
                    ne[2]->vertex() = m.vertices.insert_hole(MeshVertex(m));
                    ne[2]->vertex()->position() = e1->next()->vertex()->position();
                }

                e0 = e1;
                e1 = e1->prev();
                while (e1->dual() != mesh_null_half_edge)
                {
                    e1 = e1->dual()->prev();
                    if (++loop_detector == m.edges.total_end())
                        throw mesh_nonmanifold("poorly formed neighbourhood");
                }
            }
            np->position /= count;
        }
}

//! Sets up the \c first_vertex and \c first_edge structures in a mesh
/*! It is assumed that \c mesh_set_neighbours and \c mesh_tile_holes
 * have already been run.
 */
static void mesh_set_edges(Mesh &m) throw(mesh_nonmanifold)
{
    // Set them all to something illegal, so that we can tell what is not set
    for (MeshPositionIterator p = m.positions.total_begin(); p != m.positions.total_end(); p++)
        p->first_vertex() = mesh_null_vertex;
    for (MeshVertexIterator v = m.vertices.total_begin(); v != m.vertices.total_end(); v++)
        v->first_edge() = mesh_null_half_edge;

    // Go looking for patch border edges
    for (MeshHalfEdgeIterator e = m.edges.whole_begin(); e != m.edges.whole_end(); e++)
    {
        MeshHalfEdgeIterator test = e->prev();
        if (test->dual() == mesh_null_half_edge)
            throw mesh_nonmanifold("hole remains after hole tiling");
        if (test->dual()->prev()->vertex() != e->vertex())
            e->vertex()->first_edge() = test;
    }

    // Look for hole borders
    for (MeshVertexIterator v = m.vertices.whole_begin(); v != m.vertices.whole_end(); v++)
        if (v->first_edge() != mesh_null_half_edge
            && v->first_edge()->face()->patch == 0)
            v->position()->first_vertex() = v;

    // Now set everything else to something arbitrary but legal
    // Also check for anything isolated
    for (MeshHalfEdgeIterator e = m.edges.whole_begin(); e != m.edges.whole_end(); e++)
        if (e->vertex()->first_edge() == mesh_null_half_edge)
            e->vertex()->first_edge() = e->prev();
    for (MeshVertexIterator v = m.vertices.whole_begin(); v != m.vertices.whole_end(); v++)
    {
        if (v->position()->first_vertex() == mesh_null_vertex)
            v->position()->first_vertex() = v;
        if (v->first_edge() == mesh_null_half_edge)
            throw mesh_nonmanifold("isolated vertex");
    }
    for (MeshPositionIterator p = m.positions.whole_begin(); p != m.positions.whole_end(); p++)
        if (p->first_vertex() == mesh_null_vertex)
            throw mesh_nonmanifold("isolated position");
}

//! Fills in the patch fields by dfs
/* It also attempts to handle the case where boundary curves do not
 * cut the mesh neatly into patches, especially when a patch would be
 * its own neighbour across a sharp edge. It does this by
 * - refusing to make a patch its own neighbour (introducing an
 * artificial patch if necessary).
 * - refusing to add a face to a patch if it will make it revisit
 * a single position without the intervening faces.
 * - splitting vertices to create these extra boundaries
 */
static void mesh_set_patches(Mesh &m) throw(mesh_nonmanifold)
{
    stack<MeshFaceIterator> faces;
    int pool = 0;

    // propagate patches
    for (MeshFaceIterator start = m.faces.real_begin(); start != m.faces.real_end(); start++)
        if (start->patch == -1)
        {
            id_set<MeshPositionIterator>::type used_positions;
            faces.push(start);
            start->patch = ++pool;
            MeshHalfEdgeIterator etmp = start->edges();
            for (int i = 0; i < 3; i++, etmp = etmp->next())
                used_positions.insert(etmp->vertex()->position());
            while (!faces.empty())
            {
                MeshFaceIterator f = faces.top();
                faces.pop();

                MeshHalfEdgeIterator e = f->edges();
                for (int i = 0; i < 3; i++, e = e->next())
                {
                    MeshHalfEdgeIterator dual = e->dual();
                    if (dual->face()->patch != -1) continue;
                    if (dual->face()->material() != start->material()) continue;
                    if (e->next()->vertex() != dual->prev()->vertex()) continue;
                    if (e->prev()->vertex() != dual->next()->vertex()) continue;

                    MeshHalfEdgeIterator e2 = dual;
                    bool neighbour[3];
                    for (int j = 0; j < 3; j++, e2 = e2->next())
                    {
                        bool sharp;
                        MeshHalfEdgeIterator dual2 = e2->dual();
                        sharp = e2->next()->vertex() != dual2->prev()->vertex();
                        sharp |= e2->prev()->vertex() != dual2->next()->vertex();
                        sharp |= e2->face()->material() != dual2->face()->material();
                        neighbour[j] = dual2->face()->patch == pool;
                        if (sharp && neighbour[j]) goto unusable;
                    }
                    if (used_positions.count(dual->vertex()->position())
                        && !neighbour[1] && !neighbour[2])
                        goto unusable;

                    dual->face()->patch = pool;
                    faces.push(dual->face());
                    used_positions.insert(dual->vertex()->position());
                unusable: ;
                }
            }
        }

    // split vertices where necessary
    for (MeshPositionIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
    {
        MeshVertexIterator v = p->first_vertex(), vnew;
        MeshHalfEdgeIterator first_edge = v->first_edge();
        MeshHalfEdgeIterator e = first_edge;
        int patch = e->face()->patch;

        /* If we have only a single vertex but multiple patches, we
         * will want to start at a patch boundary
         */
        if (first_edge->dual()->prev()->vertex() == v)
        {
            bool found = false;
            do
            {
                if (e->face()->patch != e->dual()->face()->patch)
                {
                    found = true;
                    break;
                }
                e = e->dual()->next();
            } while (e != first_edge);
            if (!found) continue; // normal vertex, don't bother
            v->first_edge() = first_edge = e;
            patch = e->dual()->face()->patch; // see below
        }

        /* We walk around the position, and every time there is a patch
         * change we check that there is also a vertex change.
         * In the code below, v always refers to the _original_ value
         * of the vertex field, while v1 is its updated value.
         *
         * To complicate matters, when splitting a solitary vertex we
         * only introduce _one fewer_ vertices than patch boundaries.
         * We handle this by using the dual patch in the above code, so
         * that we don't make a split right at the start
         */
        vnew = e->dual()->prev()->vertex();
        do
        {
            MeshVertexIterator v1 = e->next()->vertex(); // new value
            MeshVertexIterator &v2 = e->dual()->prev()->vertex();
            int patch2 = e->dual()->face()->patch;
            if (v != v2)
                vnew = v2;
            else if (patch != patch2)
            {
                // split
                vnew = m.vertices.insert_real(MeshVertex());
                // cannot do *vnew = *v, since that overwrites ID
                vnew->position() = v1->position();
                vnew->data = v1->data;
                vnew->first_edge() = v1->first_edge();
                v1->first_edge() = e;
            }
            v = v2;
            v2 = vnew;
            e = e->dual()->next();
            patch = patch2;
        } while (e != first_edge);
    }
}

//! Returns the cosine of the angle between two edges in texture space
static double mesh_edge_angle_texture(MeshHalfEdgeConstIterator e1,
                                      MeshHalfEdgeConstIterator e2)
{
    Vector2 d1, d2;
    double len2;

    d1 = e1->prev()->vertex()->data.texcoords - e1->next()->vertex()->data.texcoords;
    d2 = e2->prev()->vertex()->data.texcoords - e2->next()->vertex()->data.texcoords;
    len2 = (d1 * d1) * (d2 * d2);
    if (len2 < 1e-18) return 1.0;
    else return d1 * d2 / sqrt(len2);
}

//! True if \a pos lies on any texture space patch corner
static bool mesh_position_boundary_corner(MeshPositionConstIterator pos)
{
    int val = mesh_position_patch_valence(pos);
    if (val != 2) return val > 2;

    MeshHalfEdgeConstIterator edges[4];
    edges[0] = pos->first_vertex()->first_edge();
    edges[2] = edges[0]->dual();
    edges[3] = edges[2]->prev()->vertex()->first_edge();
    edges[1] = edges[3]->dual();
    if (edges[0]->face()->patch && mesh_edge_angle_texture(edges[0], edges[1]) < 0.9999)
        return true;
    if (edges[2]->face()->patch && mesh_edge_angle_texture(edges[2], edges[3]) < 0.9999)
        return true;
    return false;
}

void mesh_set_patch_corners(Mesh &m)
{
    for (MeshPositionIterator p = m.positions.whole_begin(); p != m.positions.whole_end(); p++)
        p->patch_corner = mesh_position_boundary_corner(p);
}

void mesh_complete(Mesh &m, MeshLevel level) throw(mesh_nonmanifold)
{
    if (level >= MESH_LEVEL_MANIFOLD && m.level < MESH_LEVEL_MANIFOLD)
    {
        mesh_set_neighbours(m);
        mesh_tile_holes(m);
        mesh_set_edges(m);
        mesh_compute_normals(m);
        mesh_check_manifold(m, MESH_LEVEL_MANIFOLD);
    }
    if (level >= MESH_LEVEL_PATCHES && m.level < MESH_LEVEL_PATCHES)
    {
        mesh_set_patches(m);
        mesh_set_patch_corners(m);
        mesh_check_manifold(m, level);
    }
    m.level = level;
}

class mesh_compare_raw_vertex
{
private:
    mesh_compare_vector_position comp_pos;
public:
    bool operator()(const MeshRawVertex * const &a,
                    const MeshRawVertex * const &b) const
    {
        if (comp_pos(a->position, b->position)) return true;
        if (comp_pos(b->position, a->position)) return false;
        if (a->texcoords[0] != b->texcoords[0])
            return a->texcoords[0] < b->texcoords[0];
        return a->texcoords[1] < b->texcoords[1];
    }
};

/*! The skeleton for the mesh should already exist, and the raw positions
 * must belong to this skeleton.
 */
void mesh_create(Mesh &m, const std::vector<MeshRawVertex> &vertices,
                 const std::vector<std::vector<int> > &faces,
                 const std::vector<int> &materials,
                 MeshLevel level) throw(mesh_nonmanifold)
{
    // So that we can do O(1) lookups of the vertex indices
    vector<MeshVertexIterator> vertex_iterators;
    // So that we can merge positions and vertices
    map<VectorPosition, MeshPositionIterator, mesh_compare_vector_position> position_map;
    map<const MeshRawVertex *, MeshVertexIterator, mesh_compare_raw_vertex> vertex_map;

    // Create vertices and positions
    for (size_t i = 0; i < vertices.size(); i++)
    {
        MeshPositionIterator p;
        MeshVertexIterator v;
        map<VectorPosition, MeshPositionIterator, mesh_compare_vector_position>::iterator ppos;
        map<const MeshRawVertex *, MeshVertexIterator, mesh_compare_raw_vertex>::iterator vpos;

        vpos = vertex_map.find(&vertices[i]);
        if (vpos == vertex_map.end())
        {
            v = m.vertices.insert_real(MeshVertex());
            v->data.texcoords = vertices[i].texcoords;
            v->data.tangent_plane.tangents[0] = vertices[i].tangents[0];
            v->data.tangent_plane.tangents[1] = vertices[i].tangents[1];
            vertex_map[&vertices[i]] = v;
        }
        else
            v = vpos->second;
        vertex_iterators.push_back(v);
        ppos = position_map.find(vertices[i].position);
        if (ppos == position_map.end())
        {
            p = m.positions.insert_real(MeshPosition(vertices[i].position));
            p->position.compress();
            position_map[vertices[i].position] = p;
        }
        else
            p = ppos->second;
        v->position() = p;
    }

    fast_map<int, MeshMaterialIterator>::type material_map;
    for (size_t i = 0; i < faces.size(); i++)
    {
        for (size_t j = 1; j + 1 < faces[i].size(); j++)
        {
            MeshFace f;
            MeshFaceIterator f_it;
            MeshHalfEdgeIterator edges[3];
            VectorPosition pos[3];

            for (int k = 0; k < 3; k++)
                edges[k] = m.edges.insert_real(MeshHalfEdge());
            for (int k = 0; k < 3; k++)
            {
                edges[k]->next() = edges[(k + 1) % 3];
                edges[k]->prev() = edges[(k + 2) % 3];
            }
            if (faces[i][0] < 0 || (size_t) faces[i][j] >= vertex_iterators.size())
                throw mesh_nonmanifold("out of range vertex index");
            if (faces[i][j] < 0 || (size_t) faces[i][j] >= vertex_iterators.size())
                throw mesh_nonmanifold("out of range vertex index");
            if (faces[i][j + 1] < 0 || (size_t) faces[i][j + 1] >= vertex_iterators.size())
                throw mesh_nonmanifold("out of range vertex index");
            edges[0]->vertex() = vertex_iterators[faces[i][0]];
            edges[1]->vertex() = vertex_iterators[faces[i][j]];
            edges[2]->vertex() = vertex_iterators[faces[i][j + 1]];
            for (int k = 0; k < 3; k++)
                pos[k] = edges[k]->vertex()->position()->position;

            f.edges() = edges[0];
            f.patch = -1; // indicates unassigned patch
            if (material_map.find(materials[i]) == material_map.end())
            {
                material_map[materials[i]] = m.materials.insert(m.materials.end(), MeshMaterial());
                material_map[materials[i]]->texture() = mesh_null_texture;
                material_map[materials[i]]->tangent_map() = mesh_null_tangent_map;
            }
            f.material() = material_map[materials[i]];

            f_it = m.faces.insert_real(f);
            for (int k = 0; k < 3; k++)
                edges[k]->face() = f_it;
        }
    }

    if (level >= MESH_LEVEL_MANIFOLD)
        mesh_complete(m, level);
    else
    {
        // put in semi-reasonable values for debugging
        for (MeshPositionIterator p = m.positions.total_begin(); p != m.positions.total_end(); p++)
            p->first_vertex() = mesh_null_vertex;
        for (MeshVertexIterator v = m.vertices.total_begin(); v != m.vertices.total_end(); v++)
            v->first_edge() = mesh_null_half_edge;
        for (MeshHalfEdgeIterator e = m.edges.total_begin(); e != m.edges.total_end(); e++)
            e->dual() = mesh_null_half_edge;
    }
}

static void mesh_compute_normal(const Skeleton &skeleton, MeshVertexIterator v)
{
    if (!!v->data.tangent_plane.tangents[0] && !!v->data.tangent_plane.tangents[1])
        return;
    if (v->first_edge()->face()->patch == 0) // hole
        return;

    vector<TangentPlane> planes;
    MeshHalfEdgeIterator start = v->position()->first_vertex()->first_edge();
    MeshHalfEdgeIterator e = start;
    const VectorPosition &p = v->position()->position;

    do
    {
        if (e->face()->patch != 0) // ignore holes
        {
            TangentPlane cur;
            const VectorPosition &p2 = e->prev()->vertex()->position()->position;
            const VectorPosition &p3 = e->vertex()->position()->position;
            cur.tangents[0] = p2 - p;
            cur.tangents[1] = p3 - p;
            double d = double_area(cur.tangents[0], cur.tangents[1]);
            if (fabs(d) > 1e-12)  // Ignore degenerate faces
            {
                cur.tangents[1] /= d;
                planes.push_back(cur);
            }
        }
        e = e->dual()->next();
    } while (e != start);

    if (!planes.empty())
        v->data.tangent_plane = fit_tangent_plane(skeleton, planes);
    else
    {
        v->data.tangent_plane.tangents[0] = VectorPosition(skeleton);
        v->data.tangent_plane.tangents[1] = VectorPosition(skeleton);
    }
}

void mesh_compute_normals(Mesh &m)
{
    const Skeleton &skeleton = m.skeleton.get_read();

    for (MeshVertexIterator v = m.vertices.whole_begin(); v != m.vertices.whole_end(); v++)
    {
        mesh_compute_normal(skeleton, v);
    }

    /* Internal_data gets copied. This links copies back to the originals */
    for (MeshVertexIterator v = m.vertices.total_begin(); v != m.vertices.total_end(); v++)
        v->internal_data = &*v;

    Mesh m2 = m;   // destructively simplified to extract LOD connectivity
    while (m2.collapses_begin() != m2.collapses_end())
    {
        MeshCollapseIterator c = m2.collapses_begin();
        mesh_apply_collapse(m2, *c, mesh_null_half_edge, false);
        if (c->is_edge_collapse)
        {
            MeshVertexIterator vl = c->edge_collapse.vl();
            MeshVertexIterator vr = c->edge_collapse.vr();
            mesh_compute_normal(skeleton, vl);
            mesh_compute_normal(skeleton, vr);
            ((MeshVertex *) vl->internal_data)->data.tangent_plane = vl->data.tangent_plane;
            ((MeshVertex *) vr->internal_data)->data.tangent_plane = vr->data.tangent_plane;
        }
        else
        {
            /* Don't re-solve when doing influence simplifications */
            for (size_t i = 0; i < c->influence_collapse.size(); i++)
            {
                MeshVertexIterator before = c->influence_collapse[i].before();
                MeshVertexIterator after = c->influence_collapse[i].after();
                after->data.tangent_plane = before->data.tangent_plane;
                ((MeshVertex *) after->internal_data)->data.tangent_plane = after->data.tangent_plane;
            }
        }

        m2.pop_collapse();
    }
}

template<class T> static bool null_internal_data(const T &x)
{
    return !x.internal_data;
}

template<class SourceIterator, class T>
static void mesh_garbage_collect1(SourceIterator sfirst, SourceIterator slast,
                                  list<T> &target, typename list<T>::iterator &(iterator_traits<SourceIterator>::value_type::*accessor)())
{
    for (typename list<T>::iterator i = target.begin(); i != target.end(); i++)
        i->internal_data = NULL;
    for (SourceIterator i = sfirst; i != slast; i++)
        ((*i).*accessor)()->internal_data = &accessor; // any non-NULL
    target.remove_if(null_internal_data<T>);
}

static void mesh_garbage_collect_bones_update_vector(Skeleton &new_skeleton, const vector<int> &fmap,
                                                     VectorPosition &vect)
{
    VectorPosition old = vect;
    vect = VectorPosition(new_skeleton);
    for (VectorPosition::const_iterator i = old.begin(); i != old.end(); i++)
    {
        int b = 4 * fmap[i->index / 4];
        if (b < 0) continue; // can happen for hole positions that depend on a deleted bone
        for (int j = 0; j < 4; j++)
            vect[b + j] = i->values[j];
    }
}

static void mesh_garbage_collect_bones(Mesh &m)
{
    /* Hold a ref to the original skeleton, so that the original positions
     * are not invalidated.
     */
    Refcount<Skeleton> old_ref = m.skeleton;
    const Skeleton &old_skeleton = old_ref.get_read();
    size_t sz = old_skeleton.size();

    vector<bool> used(sz, false);

    // Must keep the root
    used[0] = true;
    // Identify bones that are directly used
    for (MeshPositionConstIterator p = m.positions.all_begin(); p != m.positions.all_end(); p++)
    {
        for (VectorPosition::const_iterator i = p->position.begin(); i != p->position.end(); i++)
            used[i->index / 4] = true;
    }
    for (MeshVertexConstIterator v = m.vertices.all_begin(); v != m.vertices.all_end(); v++)
        for (int t = 0; t < 2; t++)
        {
            const VectorPosition &tan = v->data.tangent_plane.tangents[t];
            for (VectorPosition::const_iterator i = tan.begin(); i != tan.end(); i++)
                used[i->index / 4] = true;
        }
    for (MeshTangentMapConstIterator tm = m.tangent_maps.begin(); tm != m.tangent_maps.end(); tm++)
    {
        size_t size = tm->width * tm->height;
        for (size_t i = 0; i < size; i++)
            for (int t = 0; t < 2; t++)
            {
                const VectorPosition &tan = tm->data()[i].tangent_plane.tangents[t];
                for (VectorPosition::const_iterator i = tan.begin(); i != tan.end(); i++)
                    used[i->index / 4] = true;
            }
    }
    // Identify bones that need to be present for their children
    for (size_t i = sz - 1; i > 0; i--)
        if (used[i] && old_skeleton.get_bone(i).parent >= 0)
            used[old_skeleton.get_bone(i).parent] = true;

    vector<int> rmap;
    vector<int> fmap(sz, -1);
    for (size_t i = 0; i < sz; i++)
        if (used[i])
        {
            fmap[i] = rmap.size();
            rmap.push_back(i);
        }
    if (rmap.size() == sz) return;  // nothing to erase

    vector<SkeletonBoneInfo> bones(rmap.size());
    for (size_t i = 0; i < rmap.size(); i++)
    {
        bones[i] = old_skeleton.get_bone(rmap[i]);
        if (bones[i].parent >= 0)
            bones[i].parent = fmap[bones[i].parent];
    }

    vector<vector<Matrix4> > frames(old_skeleton.num_frames());
    for (size_t f = 0; f < frames.size(); f++)
    {
        frames[f].resize(rmap.size());
        for (size_t i = 0; i < rmap.size(); i++)
            frames[f][i] = old_skeleton.get_frame(f)[rmap[i]];
    }

    Skeleton &new_skeleton = m.skeleton.get_write(); // should make a new copy, since we hold a second ref
    new_skeleton.clear_frames();
    new_skeleton.set_bones(bones);
    for (size_t f = 0; f < frames.size(); f++)
        new_skeleton.add_frame(frames[f]);

    // Update the positions and tangent planes to the new skeleton
    for (MeshPositionIterator p = m.positions.total_begin(); p != m.positions.total_end(); p++)
        mesh_garbage_collect_bones_update_vector(new_skeleton, fmap, p->position);
    for (MeshVertexIterator v = m.vertices.total_begin(); v != m.vertices.total_end(); v++)
        for (int t = 0; t < 2; t++)
            mesh_garbage_collect_bones_update_vector(new_skeleton, fmap,
                                                     v->data.tangent_plane.tangents[t]);
    for (MeshTangentMapIterator tm = m.tangent_maps.begin(); tm != m.tangent_maps.end(); tm++)
    {
        size_t size = tm->width * tm->height;
        for (size_t i = 0; i < size; i++)
            for (int t = 0; t < 2; t++)
                mesh_garbage_collect_bones_update_vector(new_skeleton, fmap,
                                                         tm->data()[i].tangent_plane.tangents[t]);
    }
}

void mesh_garbage_collect(Mesh &m)
{
    mesh_garbage_collect1(m.faces.real_begin(), m.faces.real_end(),
                          m.materials, &MeshFace::material);
    mesh_garbage_collect1(m.materials.begin(), m.materials.end(),
                          m.textures, &MeshMaterial::texture);
    mesh_garbage_collect1(m.materials.begin(), m.materials.end(),
                          m.tangent_maps, &MeshMaterial::tangent_map);

    mesh_garbage_collect_bones(m);
}

MeshVertexData mesh_vertex_data_interpolate(const Skeleton &skeleton,
                                            const MeshVertexData &a,
                                            const MeshVertexData &b,
                                            double t)
{
    if (t == 0.0) return a;
    if (t == 1.0) return b;

    double t2 = 1.0 - t;
    vector<TangentPlane> tps(2);
    tps[0] = t2 * a.tangent_plane;
    tps[1] = t * b.tangent_plane;
    return MeshVertexData(t2 * a.texcoords + t * b.texcoords,
                          fit_tangent_plane(skeleton, tps));
}

template class MeshCollection<MeshPosition>;
template class MeshCollection<MeshVertex>;
template class MeshCollection<MeshHalfEdge>;
template class MeshCollection<MeshFace>;


/**************************************************************
 * Test code
 **************************************************************/

#if HAVE_CPPUNIT
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>
#include "geom/skeleton.h"
#include "formats/formats.h"
#include "common/paths.h"

class MeshTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(MeshTest);
    CPPUNIT_TEST(garbage_collect);
    CPPUNIT_TEST(mesh_transform);
    CPPUNIT_TEST(bounding_sphere_as);
    CPPUNIT_TEST(bounding_sphere_3d);
    CPPUNIT_TEST_SUITE_END();
public:
    void garbage_collect()
    {
        Mesh m;
        m.textures.push_back(MeshTexture());
        m.textures.push_back(MeshTexture());
        m.tangent_maps.push_back(MeshTangentMap());
        m.tangent_maps.push_back(MeshTangentMap());
        m.materials.push_back(MeshMaterial());
        m.materials.push_back(MeshMaterial());
        m.materials.front().texture() = m.textures.begin();
        m.materials.front().tangent_map() = --m.tangent_maps.end();
        mesh_garbage_collect(m);
        CPPUNIT_ASSERT(m.materials.empty());
        CPPUNIT_ASSERT(m.textures.empty());
        CPPUNIT_ASSERT(m.tangent_maps.empty());
    }

    void mesh_transform()
    {
        Mesh m;
        vector<double> l2, linf;
        vector<Vector4> pos;
        vector<Vector4> epos;
        double s = 2.0;
        Matrix4 T = matrix_scale(s);
        T *= matrix_rotate(1.3, Vector3(1.0, 0.5, 0.2));
        T *= matrix_translate(Vector3(-1.0, 2.0, -3.0));

        string filename = resource_path("srcdir", ".", "tests/meshes/cross.alm");
        format_load_mesh_file("auto", filename, m,
                              MESH_LEVEL_PATCHES);
        Skeleton &skeleton = m.skeleton.get_write();
        skeleton.update_bounds();
        skeleton.update_expect();
        skeleton.set_to_rest();

        for (MeshPositionConstIterator p1 = m.positions.real_begin(); p1 != m.positions.real_end(); p1++)
        {
            for (MeshPositionConstIterator p2 = p1; p2 != m.positions.real_end(); p2++)
            {
                VectorPosition delta = p1->position - p2->position;
                l2.push_back(length(delta));
                linf.push_back(skeleton.inf_metric(delta));
            }
            pos.push_back(vector3_to_vector4(skeleton.ssd_to_3d(p1->position), 1.0));
            epos.push_back(vector3_to_vector4(skeleton.ssd_to_3d_expected(p1->position), 1.0));
        }

        ::mesh_transform(m, T);

        int index = 0, index2 = 0;
        for (MeshPositionConstIterator p1 = m.positions.real_begin(); p1 != m.positions.real_end(); p1++)
        {
            for (MeshPositionConstIterator p2 = p1; p2 != m.positions.real_end(); p2++)
            {
                VectorPosition delta = p1->position - p2->position;
                CPPUNIT_ASSERT_DOUBLES_EQUAL(l2[index2] * s, length(delta), 1e-6);
                CPPUNIT_ASSERT_DOUBLES_EQUAL(linf[index2] * s, skeleton.inf_metric(delta), 1e-6);
                index2++;
            }
            Vector4 p = vector3_to_vector4(skeleton.ssd_to_3d(p1->position), 1.0);
            Vector4 ep = vector3_to_vector4(skeleton.ssd_to_3d_expected(p1->position), 1.0);
            CPPUNIT_ASSERT_DOUBLES_EQUAL(length(p - T * pos[index]), 0.0, 1e-6);
            CPPUNIT_ASSERT_DOUBLES_EQUAL(length(ep - T * epos[index]), 0.0, 1e-6);
            index++;
        }
    }

    void bounding_sphere_as()
    {
        Mesh m;
        string filename = resource_path("srcdir", ".", "tests/meshes/tbounds.alm");
        format_load_mesh_file("auto", filename, m, MESH_LEVEL_PATCHES);

        pair<VectorPosition, double> bs = mesh_bounding_sphere_as(m);
        Skeleton &skeleton = m.skeleton.get_write();
        for (size_t i = 0; i < skeleton.num_frames(); i++)
        {
            skeleton.set_to_frame(i);
            Vector3 centroid = skeleton.ssd_to_3d(bs.first);
            for (MeshPositionConstIterator p = m.positions.all_begin(); p != m.positions.all_end(); p++)
            {
                Vector3 pos = skeleton.ssd_to_3d(p->position);
                double dist = length(pos - centroid);
                CPPUNIT_ASSERT(dist < bs.second + 1e-6);
            }
        }
    }

    void bounding_sphere_3d()
    {
        Mesh m;
        string filename = resource_path("srcdir", ".", "tests/meshes/tbounds.alm");
        format_load_mesh_file("auto", filename, m, MESH_LEVEL_PATCHES);

        BoundingSphere bs = mesh_bounding_sphere_3d(m);
        Skeleton &skeleton = m.skeleton.get_write();
        for (size_t i = 0; i < skeleton.num_frames(); i++)
        {
            skeleton.set_to_frame(i);
            for (MeshPositionConstIterator p = m.positions.all_begin(); p != m.positions.all_end(); p++)
            {
                Vector3 pos = skeleton.ssd_to_3d(p->position);
                double dist = length(pos - bs.centre);
                CPPUNIT_ASSERT(dist < bs.radius + 1e-6);
            }
        }
        CPPUNIT_ASSERT(bs.radius < 0.8); // conversative bound, but useful sanity check
    }
};

CPPUNIT_TEST_SUITE_REGISTRATION(MeshTest);

#endif /* HAVE_CPPUNIT */
/*! @} */
