/*! \addtogroup mesh
 * @{
 */

/*! \file
 * \brief %Mesh data structures
 *
 * See <tt>doc/mesh_invariant.txt</tt> for details of the mesh data
 * structures.
 */

#ifndef ALOD_MESH_MESH_H
#define ALOD_MESH_MESH_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <list>
#include <deque>
#include <vector>
#include <cstddef>
#include <utility>
#include <stdexcept>
#include <algorithm>
#include <iterator>
#include <glib.h>
#include "geom/coords.h"
#include "geom/skeleton.h"
#include "geom/anim.h"
#include "geom/plane.h"
#include "geom/bounding.h"
#include "common/refcount.h"
#include "common/maputils.h"

template<typename T>
class MeshCollection : private std::list<T>
{
public:
    typedef std::list<T> Base;
    typedef typename Base::iterator iterator;
    typedef typename Base::const_iterator const_iterator;
    typedef typename iterator::value_type value_type;

private:
    iterator real_start;
    iterator lod_start;

    template<class BinaryPredicate> void sort(iterator &first, iterator last, BinaryPredicate p)
    {
        if (first == last) return;

        std::list<T> tmp;
        tmp.splice(tmp.begin(), *this, first, last);
        tmp.sort(p);
        first = tmp.begin();
        splice(last, tmp);
    }

public:
    iterator hole_begin() { return Base::begin(); }
    iterator hole_end() { return real_start; }
    iterator real_begin() { return real_start; }
    iterator real_end() { return lod_start; }
    iterator lod_begin() { return lod_start; }
    iterator lod_end() { return Base::end(); }
    iterator whole_begin() { return Base::begin(); }
    iterator whole_end() { return lod_start; }
    iterator all_begin() { return real_start; }
    iterator all_end() { return Base::end(); }
    iterator total_begin() { return Base::begin(); }
    iterator total_end() { return Base::end(); }

    const_iterator hole_begin() const { return Base::begin(); }
    const_iterator hole_end() const { return real_start; }
    const_iterator real_begin() const { return real_start; }
    const_iterator real_end() const { return lod_start; }
    const_iterator lod_begin() const { return lod_start; }
    const_iterator lod_end() const { return Base::end(); }
    const_iterator whole_begin() const { return Base::begin(); }
    const_iterator whole_end() const { return lod_start; }
    const_iterator all_begin() const { return real_start; }
    const_iterator all_end() const { return Base::end(); }
    const_iterator total_begin() const { return Base::begin(); }
    const_iterator total_end() const { return Base::end(); }

    size_t hole_size() const { return std::distance(hole_begin(), hole_end()); }
    size_t real_size() const { return std::distance(real_begin(), real_end()); }
    size_t lod_size() const { return std::distance(lod_begin(), lod_end()); }
    size_t whole_size() const { return std::distance(whole_begin(), whole_end()); }
    size_t all_size() const { return std::distance(all_begin(), all_end()); }
    size_t total_size() const { return std::distance(total_begin(), total_end()); }

    iterator insert_hole(const value_type &v);
    iterator insert_real(const value_type &v);
    iterator insert_lod(const value_type &v);

    void erase(iterator pos);
    void move_to_hole(iterator pos);
    void move_to_real(iterator pos);
    void move_to_lod(iterator pos);

    void clear();
    void clear_lod();

    MeshCollection();
    MeshCollection(const MeshCollection<T> &c);
    MeshCollection &operator =(const MeshCollection<T> &c);

    template<class BinaryPredicate> void sort(BinaryPredicate p)
    {
        iterator first = this->begin();
        iterator old_real_start = real_start;
        iterator old_lod_start = lod_start;

        sort(lod_start, this->end(), p);

        if (real_start != old_lod_start) sort(real_start, lod_start, p);
        else real_start = lod_start;

        if (first != old_real_start) sort(first, real_start, p);
    }
};

class Mesh;
class MeshPosition;
class MeshVertex;
class MeshFace;
class MeshHalfEdge;
class MeshTexture;
class MeshTangentMap;
class MeshMaterial;
class MeshCollapse;

typedef MeshCollection<MeshPosition>::iterator MeshPositionIterator;
typedef MeshCollection<MeshPosition>::const_iterator MeshPositionConstIterator;
typedef MeshCollection<MeshVertex>::iterator MeshVertexIterator;
typedef MeshCollection<MeshVertex>::const_iterator MeshVertexConstIterator;
typedef MeshCollection<MeshFace>::iterator MeshFaceIterator;
typedef MeshCollection<MeshFace>::const_iterator MeshFaceConstIterator;
typedef MeshCollection<MeshHalfEdge>::iterator MeshHalfEdgeIterator;
typedef MeshCollection<MeshHalfEdge>::const_iterator MeshHalfEdgeConstIterator;
typedef std::list<MeshTexture>::iterator MeshTextureIterator;
typedef std::list<MeshTexture>::const_iterator MeshTextureConstIterator;
typedef std::list<MeshTangentMap>::iterator MeshTangentMapIterator;
typedef std::list<MeshTangentMap>::const_iterator MeshTangentMapConstIterator;
typedef std::list<MeshMaterial>::iterator MeshMaterialIterator;
typedef std::list<MeshMaterial>::const_iterator MeshMaterialConstIterator;

typedef std::deque<MeshCollapse> MeshCollapseGroup;
typedef MeshCollapseGroup::iterator MeshCollapseIterator;
typedef MeshCollapseGroup::const_iterator MeshCollapseConstIterator;

#define CREATE_POINTER(type, name) \
    private: \
    type ## Iterator _ ## name; \
    public: \
    type ## Iterator &name() { return _ ## name; } \
    type ## ConstIterator name() const { return _ ## name; }

// There is no such thing as a NULL iterator, so we provide magic values
extern const MeshPositionIterator mesh_null_position;
extern const MeshVertexIterator mesh_null_vertex;
extern const MeshHalfEdgeIterator mesh_null_half_edge;
extern const MeshFaceIterator mesh_null_face;
extern const MeshTextureIterator mesh_null_texture;
extern const MeshTangentMapIterator mesh_null_tangent_map;
extern const MeshMaterialIterator mesh_null_material;

//! Class used only for input of meshes from files
struct MeshRawVertex
{
    VectorPosition position;
    VectorPosition tangents[2];
    Vector2 texcoords;
};

//! A base class that has an ID, useful for \c id_map and \c id_set
class MeshElement
{
private:
    static std::size_t id_pool;

public:
    std::size_t id;
    //! Used internally by algorithms to attach extra data
    void *internal_data;
    MeshElement() : id(id_pool++) {}

    bool operator ==(const MeshElement &b) const { return id == b.id; }
    bool operator <(const MeshElement &b) const { return id < b.id; }
};

//! Stored XYZ coordinates for a vertex
/*! In the invariants below, \c POSITION stands for an iterator to this
 * position. C is the number of vertices that share this position.
 * \invariant
 * - \c first_vertex points to an element of Mesh::vertices.
 * - \c first_vertex->position == \c POSITION.
 * - The following loop will terminate and will iterate over
 * all edges for which <tt>EDGE->next->vertex->position == POSITION</tt>:
 * \code
 *     e = POSITION->first_vertex->first_edge;
 *     do
 *     {
 *       e = e->dual->next;
 *     } while (e != POSITION->first_vertex->first_edge)
 * \endcode
 * - If \c patch_corner then C > 1.
 * - If C > 2 then \c patch_corner is true.
 */
class MeshPosition : public MeshElement
{
public:
    CREATE_POINTER(MeshVertex, first_vertex);

    VectorPosition position;
    //! A vertex corresponding to this position
    /*! If this is a boundary position, then first_vertex->first_edge
     * is the boundary edge most clockwise of the mesh (counter-clockwise
     * of the hole). Otherwise it is an arbitrary vertex.
     */

    //! True if this mesh lies on a patch boundary which is not straight in texture space
    bool patch_corner;

    MeshPosition() : patch_corner(false) {}
    explicit MeshPosition(const Mesh &m);
    explicit MeshPosition(VectorPosition pos)
        : _first_vertex(mesh_null_vertex), position(pos), patch_corner(false) {}
};

//! Holds the coordinate data for a vertex.
/*! This is a separate structure because the simplifier needs to
 * keep track of where it intends to place new vertices, without
 * needing the pointer and id data maintained by the vertices.
 */
struct MeshVertexData
{
    Vector2 texcoords;
    TangentPlane tangent_plane;

    MeshVertexData(const Vector2 &t, const Skeleton &skeleton)
        : texcoords(t), tangent_plane(skeleton) {}
    explicit MeshVertexData(const Vector2 &t = Vector2(), const TangentPlane &tp = TangentPlane())
        : texcoords(t), tangent_plane(tp) {}
};

//! Stores auxiliary data about a position, from a subset of incident faces.
/*! In the invariants below, \c VERTEX stands for an iterator to this
 * vertex.
 * \invariant
 * - \c position points to an element of \c Mesh::positions.
 * - \c first_edge points to an element of \c Mesh::edges.
 * - \c first_edge->next->vertex == \c VERTEX.
 * - If \c first_edge->face->patch == 0.
 *   then \c position->first_vertex == \c VERTEX.
 */
class MeshVertex : public MeshElement
{
public:
    CREATE_POINTER(MeshHalfEdge, first_edge);
    CREATE_POINTER(MeshPosition, position);

    MeshVertexData data;

    MeshVertex() : _first_edge(mesh_null_half_edge), _position(mesh_null_position) {}
    explicit MeshVertex(const Mesh &m);
};

//! Describes a half-edge.
/*! A half-edge is an edge of a face. Every edge of the mesh is represented
 * by two half-edges (or some other number, for non-manifold meshes).
 * Half-edges are considered to be directed counter-clockwise around the
 * face. In the invariants below, EDGE represents an iterator to this
 * half-edge.
 * \invariant
 * - \c prev points to an element of Mesh::edges
 * - \c next points to an element of Mesh::edges
 * - \c dual points to an element of Mesh::edges
 * - \c vertex points to an element of Mesh::vertices
 * - \c face points to an element of Mesh::faces
 * - \c prev->next == \c EDGE
 * - \c next->prev == \c EDGE
 * - \c face == \c next->face
 * - \c next != \c EDGE
 * - \c next->next->next == \c EDGE
 * - \c next->vertex != \c vertex
 * - If \c prev->dual->prev->vertex != \c vertex then
 *   \c vertex->first_edge == \c prev
 * - \c next->vertex == \c dual->prev->vertex if and only if
 *   \c face->patch == \c dual->face->patch
 * - \c face->patch == \c vertex_first_edge->face->patch
 * - If
 *   - \c EDGE1->vertex->position == \c EDGE2->vertex->position and
 *   - \c EDGE1->face->patch == \c EDGE2->face->patch
 *   .
 *   then \c EDGE1->vertex == \c EDGE2->vertex
 * - next->vertex->position == dual->prev->vertex->position
 * - prev->vertex->position == dual->next->vertex->position
 * - the pair (\c next->vertex, \c prev->vertex) is unique to each edge
 */
class MeshHalfEdge : public MeshElement
{
public:
    CREATE_POINTER(MeshVertex, vertex);
    CREATE_POINTER(MeshHalfEdge, dual);
    CREATE_POINTER(MeshFace, face);
    CREATE_POINTER(MeshHalfEdge, prev);
    CREATE_POINTER(MeshHalfEdge, next);

    MeshHalfEdge()
        : _vertex(mesh_null_vertex),
        _dual(mesh_null_half_edge),
        _face(mesh_null_face),
        _prev(mesh_null_half_edge),
        _next(mesh_null_half_edge)
    {}
};

//! Describes a triangular face of a mesh
/*! In the invariants below, FACE refers to an iterator to this face.
 * \invariant
 * - \c edge points to an element of \c Mesh::edges
 * - \c edge->face == FACE
 * - The length of \c Mesh::edges is triple that of \c Mesh::faces
 */
class MeshFace : public MeshElement
{
public:
    CREATE_POINTER(MeshHalfEdge, edges);
    CREATE_POINTER(MeshMaterial, material);

    int patch;                  //!< ID number of patch (0 for hole)

    MeshFace() : _edges(mesh_null_half_edge), _material(mesh_null_material), patch(-1) {}
};

class MeshTexture : public MeshElement
{
private:
    guint8 *_data;
public:
    std::size_t width;
    std::size_t height;
    int channels;
    bool sign;    // true for signed texture, false for unsigned

    guint8 *data() { return _data; }
    const guint8 *data() const { return _data; }

    MeshTexture();
    ~MeshTexture();
    MeshTexture(const MeshTexture &t);
    MeshTexture &operator =(const MeshTexture &t);
    void set_size(std::size_t w, std::size_t h, std::size_t c);
};

struct MeshTangentMapEntry
{
    int patch;
    TangentPlane tangent_plane;
};

class MeshTangentMap : public MeshElement
{
private:
    MeshTangentMapEntry *_data;
public:
    std::size_t width;
    std::size_t height;

    MeshTangentMapEntry *data() { return _data; }
    const MeshTangentMapEntry *data() const { return _data; }

    MeshTangentMap();
    ~MeshTangentMap();
    MeshTangentMap(const MeshTangentMap &tm);
    MeshTangentMap &operator =(const MeshTangentMap &tm);
    void set_size(std::size_t w, std::size_t h);
};

class MeshMaterial : public MeshElement
{
public:
    CREATE_POINTER(MeshTexture, texture);
    CREATE_POINTER(MeshTangentMap, tangent_map);

    MeshMaterial() : _texture(mesh_null_texture), _tangent_map(mesh_null_tangent_map) {}
};

//! Describes an edge collapse in a progressive mesh
/*! The edge to collapse is the one joining \a _p1 and \a _p2. Visualise the
 * edge with \a _p1 below \a _p2. Then the new position has vertex \a _vl
 * on the left and \a _vr on the right (where left and right refer to the
 * patches owning the collapsing triangles). If there are any incident
 * patches other than these, then this must be a half-edge collapse with
 * one of \a _p1 and \a _p2 being the position of \a _vl and \a _vr.
 */
class MeshEdgeCollapse
{
public:
    CREATE_POINTER(MeshPosition, p1);
    CREATE_POINTER(MeshPosition, p2);
    CREATE_POINTER(MeshVertex, vl);
    CREATE_POINTER(MeshVertex, vr);

    MeshEdgeCollapse()
        : _p1(mesh_null_position), _p2(mesh_null_position),
        _vl(mesh_null_vertex), _vr(mesh_null_vertex)
    {}
};

class MeshInfluenceCollapseElement
{
public:
    CREATE_POINTER(MeshVertex, before);
    CREATE_POINTER(MeshVertex, after);

    MeshInfluenceCollapseElement() : _before(mesh_null_vertex), _after(mesh_null_vertex) {}
};

typedef std::vector<MeshInfluenceCollapseElement> MeshInfluenceCollapse;

class MeshCollapse
{
public:
    bool is_edge_collapse;
    MeshEdgeCollapse edge_collapse;
    MeshInfluenceCollapse influence_collapse;
    double error;

    MeshCollapse() : is_edge_collapse(true), edge_collapse(), influence_collapse()
    { error = -1.0; /* error is a macro, so cannot be put in initialiser list */ }
};

//! Describes how much state is available in the mesh
enum MeshLevel
{
    MESH_LEVEL_NONE,
    MESH_LEVEL_MANIFOLD,
    MESH_LEVEL_PATCHES
};

//! Describes a 3D triangle mesh
/*! Meshes can legally exist at one of three levels: none, manifold or
 * patches. The first has essentially no structure. The fields below
 * may be empty or undefined, and the normal invariants do not necessary
 * hold.
 * - \c MeshPosition::first_vertex (\c mesh_null_vertex)
 * - \c MeshPosition::patch_corner (undefined)
 * - \c MeshVertex::first_edge (\c mesh_null_half_edge)
 * - \c MeshHalfEdge::dual (\c mesh_null_half_edge)
 * - \c MeshFace::patch (\c 1)
 * .
 * Manifold and patch-level meshes differ in that the latter has
 * well defined patches. Non-patch meshes should consider the patch
 * field as purely a boolean value used to distinguish holes.
 *
 * All the lists are broken into three sections (although in some
 * cases a section is identically empty).
 * 1. Fake stuff introduced to close holes.
 * 2. The real stuff in the full-res mesh.
 * 3. New vertices and positions used by the collapses.
 * But these are hidden in the private section. The following conventions
 * are used in naming the accessors:
 * - real: group 2
 * - whole: group 1 and 2
 * - all: group 2 and 3
 * - total: all three
 */
class Mesh
{
private:
    MeshCollapseGroup collapses;

public:
    Refcount<Skeleton> skeleton;
    MeshCollection<MeshPosition> positions;
    MeshCollection<MeshVertex> vertices;
    MeshCollection<MeshHalfEdge> edges;
    MeshCollection<MeshFace> faces;
    std::list<MeshTexture> textures;
    std::list<MeshTangentMap> tangent_maps;
    std::list<MeshMaterial> materials;
    MeshLevel level;

    MeshCollapseIterator collapses_begin() { return collapses.begin(); }
    MeshCollapseConstIterator collapses_begin() const { return collapses.begin(); }
    MeshCollapseIterator collapses_end() { return collapses.end(); }
    MeshCollapseConstIterator collapses_end() const { return collapses.end(); }
    void prepend_collapse(const MeshCollapse &c) { collapses.push_front(c); }
    void append_collapse(const MeshCollapse &c) { collapses.push_back(c); }
    void pop_collapse() { g_assert(!collapses.empty()); collapses.pop_front(); }

    Mesh();
    Mesh(const Mesh &m);
    Mesh &operator =(const Mesh &m);
    void clear();
    void clear_lod();
};

class mesh_nonmanifold : public std::runtime_error
{
public:
    mesh_nonmanifold(const std::string &s) : std::runtime_error(s) {}
};

inline MeshPosition::MeshPosition(const Mesh &m)
: _first_vertex(mesh_null_vertex), position(m.skeleton.get_read()) {}

inline MeshVertex::MeshVertex(const Mesh &m)
: _first_edge(mesh_null_half_edge), _position(mesh_null_position),
    data(Vector2(0.0, 0.0), TangentPlane(m.skeleton.get_read())) {}

/* Functions in mesh.cpp */

//! Remove all positions, vertices and faces from a mesh
void mesh_clear(Mesh &m);

//! Returns true if the face is not inside a hole
bool mesh_true_face(MeshFaceConstIterator p);

//! Returns true if the vertex is not inside a hole
bool mesh_true_vertex(MeshVertexConstIterator p);

//! Returns true if the position is not inside a hole
bool mesh_true_position(MeshPositionConstIterator p);

//! Returns the number of collapses in the mesh (linear time)
std::size_t mesh_collapses(const Mesh &m);

//! Returns twice the area of the given face (undirected)
double mesh_face_double_area(MeshFaceConstIterator f);

//! Returns twice the texture space area of the given face (directed)
double mesh_texture_double_area(MeshFaceConstIterator f);

//! Returns the corners of a 3D bounding box around the current pose and level
std::pair<Vector3, Vector3> mesh_bounding_box_3d(const Mesh &m);

//! Returns an animation-space bounding sphere around the mesh
std::pair<VectorPosition, double> mesh_bounding_sphere_as(const Mesh &m);

//! Returns a 3D bounding sphere around the mesh, valid for all poses
BoundingSphere mesh_bounding_sphere_3d(const Mesh &m);

//! Applies a transformation to a mesh
void mesh_transform(Mesh &m, const Matrix4 &transform);

//! Returns the number of patches incident on \a p, including holes
int mesh_position_patch_valence(MeshPositionConstIterator p);

//! Takes an incomplete mesh and puts in the extra structures
void mesh_complete(Mesh &m, MeshLevel level) throw(mesh_nonmanifold);

//! Create a mesh from a list of positions and of faces
void mesh_create(Mesh &m, const std::vector<MeshRawVertex> &vertices,
                 const std::vector<std::vector<int> > &faces,
                 const std::vector<int> &materials,
                 MeshLevel level = MESH_LEVEL_PATCHES) throw(mesh_nonmanifold);

//! Fills in any uncomputed tangent planes
void mesh_compute_normals(Mesh &m);

//! Deletes any unreferenced textures, materials or tangent maps
void mesh_garbage_collect(Mesh &m);

//! Fills in the \c MeshPosition::patch_corner fields.
void mesh_set_patch_corners(Mesh &m);

//! Interpolate between two vertex data entries
MeshVertexData mesh_vertex_data_interpolate(const Skeleton &skeleton,
                                            const MeshVertexData &a,
                                            const MeshVertexData &b,
                                            double t);

//! A structure for mapping on (exact) \c VectorPosition keys
struct mesh_compare_vector_position
{
    bool operator ()(const VectorPosition &a,
                     const VectorPosition &b) const
    {
        return vector_compare(a, b);
    }
};

//! Similar to mesh_compare_vector_position, but quantises to float
struct mesh_compare_vector_position_float
{
    bool operator ()(const VectorPosition &a,
                     const VectorPosition &b) const
    {
        return vector_compare_float(a, b);
    }
};

#include "mesh/manifold.h"
#include "mesh/param.h"
#include "mesh/simplify.h"
#include "mesh/aps.h"
#include "mesh/images.h"

#endif /* ALOD_INCLUDE_MESH_H */

/*! @} */
