/*! \addtogroup mesh
 * @{
 */

/*! \file
 * \brief Parametrise a Mesh into a 2D domain.
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <vector>
#include <stack>
#include <list>
#include <utility>
#include <iterator>
#include <queue>
#include <cmath>
#include <glib.h>
#include "param.h"
#include "mesh/iheap.h"
#include "common/maputils.h"
#include "common/mesch.h"
#include "common/timing.h"

// #define SANDER_PLANAR    // Use the least-squares plane-fit merge metric
// #define SANDER_ORIENT    // Bias the metric with Garland's orientation bias
#define AS_PATCH_METRIC     // compute the flatness metric in animation space

using namespace std;

struct MeshMergePatch;
struct MeshMergePatchEdge;
typedef list<MeshMergePatch> MeshMergePatchGroup;
typedef MeshMergePatchGroup::iterator MeshMergePatchIterator;
typedef MeshMergePatchGroup::const_iterator MeshMergePatchConstIterator;
typedef list<MeshMergePatchEdge> MeshMergePatchEdgeGroup;
typedef MeshMergePatchEdgeGroup::iterator MeshMergePatchEdgeIterator;
typedef MeshMergePatchEdgeGroup::const_iterator MeshMergePatchEdgeConstIterator;

struct MeshMergePatchEdge : public MeshElement
{
    //! The perimeter of the section of the patch represented by this edge
    double length;
    //! The position on the ccw end of this section
    MeshPositionIterator position;
    //! The ccw edge
    MeshMergePatchEdgeIterator next;
    //! The clockwise edge
    MeshMergePatchEdgeIterator prev;
    //! The corresponding edge on the neighbour patch
    MeshMergePatchEdgeIterator dual;
    //! The patch containing this edge
    MeshMergePatchIterator patch;
    //! True if the patches on either side may be merged
    bool mergeable;
};

/* Wrapper around MAT * that handles copying and a few operators. This
 * isn't a general class because in most cases using a C++ wrapper is too
 * inefficient.
 */
struct MergeMatrix
{
    MAT *M;

    MergeMatrix() : M(MNULL) {}
    MergeMatrix(const MergeMatrix &m) : M(MNULL)
    { if (m.M) M = m_copy(m.M, MNULL); }
    ~MergeMatrix() { if (M) M_FREE(M); }

    MergeMatrix &operator =(const MergeMatrix &m)
    {
        if (m.M) M = m_copy(m.M, m_resize(M, m.M->m, m.M->n));
        else if (M) M_FREE(M);
        return *this;
    }
    MergeMatrix &operator +=(const MergeMatrix &m)
    { g_assert(M && m.M); M = m_add(M, m.M, M); return *this; }
    MergeMatrix &operator -=(const MergeMatrix &m)
    { g_assert(M && m.M); M = m_sub(M, m.M, M); return *this; }
    MergeMatrix operator +(const MergeMatrix &m) const
    { MergeMatrix ans; g_assert(M && m.M); ans.M = m_add(M, m.M, MNULL); return ans; }
    MergeMatrix operator -(const MergeMatrix &m) const
    { MergeMatrix ans; g_assert(M && m.M); ans.M = m_sub(M, m.M, MNULL); return ans; }
    MergeMatrix operator -() const
    { MergeMatrix ans; g_assert(M); ans.M = sm_mlt(-1.0, M, MNULL); return ans; }

    void clear() { if (M) M_FREE(M); }
};

#ifdef SANDER_PLANAR

class MeshMergePatchGeometry
{
public:
    MergeMatrix q;     //!< Integral of x over surface
    MergeMatrix Q;     //!< Integral of xx^T over surface
    MergeMatrix D;     //!< Quadratic term of orientation quadric
    MergeMatrix E;     //!< Linear term of orientation quadric
    double F;          //!< Constant term of orientation quadric

public:
    void clear()
    {
        q.clear();
        Q.clear();
        D.clear();
        E.clear();
        F = 0.0;
    }

    MeshMergePatchGeometry() : F(0.0) {}

    MeshMergePatchGeometry operator +(const MeshMergePatchGeometry &b) const
    {
        MeshMergePatchGeometry ans;
        ans.q.M = m_add(q.M, b.q.M, MNULL);
        ans.Q.M = m_add(Q.M, b.Q.M, MNULL);
        ans.D.M = m_add(D.M, b.D.M, MNULL);
        ans.E.M = m_add(E.M, b.E.M, MNULL);
        ans.F = F + b.F;
        return ans;
    }

    MeshMergePatchGeometry operator -() const
    {
        MeshMergePatchGeometry ans;
        ans.q.M = sm_mlt(-1.0, q.M, MNULL);
        ans.Q.M = sm_mlt(-1.0, Q.M, MNULL);
        ans.D.M = sm_mlt(-1.0, D.M, MNULL);
        ans.E.M = sm_mlt(-1.0, E.M, MNULL);
        ans.F = -F;
        return ans;
    }

    MeshMergePatchGeometry &operator +=(const MeshMergePatchGeometry &b)
    { q += b.q; Q += b.Q; D += b.D; E += b.E; F += b.F; return *this; }
};

#else /* !SANDER_PLANAR */
#ifdef AS_PATCH_METRIC

//! A sparse matrix structure for merging in animation space
class MeshMergePatchGeometry
{
public:
    MAT *P;                 //!< Skeleton's P
    MergeMatrix A;          //!< Sum of (pq^T - qp^T)/2 over triangles
    vector<int> occupied;   //!< The rows+columns that are present in A

private:
    //! A += that assumes that the target is a superset of the argument */
    MeshMergePatchGeometry &add(const MeshMergePatchGeometry &b);
    //! Expands the occupation to include all of occ
    MeshMergePatchGeometry &expand(const vector<int> &occ);

public:
    MeshMergePatchGeometry() : P(MNULL) {}
    void clear() { P = MNULL; A.clear(); occupied.clear(); }

    MeshMergePatchGeometry operator +(const MeshMergePatchGeometry &b) const;
    MeshMergePatchGeometry &operator +=(const MeshMergePatchGeometry &b);
    MeshMergePatchGeometry operator -() const;
};

MeshMergePatchGeometry &MeshMergePatchGeometry::add(const MeshMergePatchGeometry &b)
{
    int isize = b.occupied.size();
    int ocmap[isize];
    int o = 0;

    for (int i = 0; i < isize; i++)
    {
        while (occupied[o] != b.occupied[i]) o++;
        ocmap[i] = o;
    }

    for (int i = 0; i < isize; i++)
        for (int j = 0; j < isize; j++)
            m_add_val(A.M, ocmap[i], ocmap[j], m_entry(b.A.M, i, j));

    g_assert(A.M->m == occupied.size());
    return *this;
}

MeshMergePatchGeometry MeshMergePatchGeometry::operator +(const MeshMergePatchGeometry &b) const
{
    MeshMergePatchGeometry ans;

    ans.P = P;
    set_union(occupied.begin(), occupied.end(), b.occupied.begin(), b.occupied.end(),
              back_inserter(ans.occupied));
    ans.A.M = m_resize(ans.A.M, ans.occupied.size(), ans.occupied.size());
    m_zero(ans.A.M);
    g_assert(ans.A.M->m == ans.occupied.size());
    ans.add(*this);
    g_assert(ans.A.M->m == ans.occupied.size());
    ans.add(b);
    g_assert(ans.A.M->m == ans.occupied.size());
    return ans;
}

MeshMergePatchGeometry MeshMergePatchGeometry::operator -() const
{
    MeshMergePatchGeometry ans;

    ans.P = P;
    ans.occupied = occupied;
    ans.A.M = sm_mlt(-1.0, A.M, MNULL);
    return ans;
}

MeshMergePatchGeometry &MeshMergePatchGeometry::expand(const vector<int> &occ)
{
    if (!includes(occupied.begin(), occupied.end(), occ.begin(), occ.end()))
    {
        vector<int> n;
        set_union(occupied.begin(), occupied.end(), occ.begin(), occ.end(),
                  back_inserter(n));
        int isize = occupied.size();
        int nsize = n.size();
        int ocmap[isize];
        int o = 0;
        MAT *nA = m_get(nsize, nsize);

        for (int i = 0; i < isize; i++)
        {
            while (n[o] != occupied[i]) o++;
            ocmap[i] = o;
        }

        m_zero(nA);
        for (int i = 0; i < isize; i++)
            for (int j = 0; j < isize; j++)
                m_set_val(nA, ocmap[i], ocmap[j], m_entry(A.M, i, j));
        M_FREE(A.M);
        A.M = nA;
        occupied = n;
    }
    return *this;
}

MeshMergePatchGeometry &MeshMergePatchGeometry::operator +=(const MeshMergePatchGeometry &b)
{
    expand(b.occupied);
    return add(b);
}

#else /* !AS_PATCH_METRIC */

class MeshMergePatchGeometry
{
public:
    Vector3 normal_sum;

private:
    explicit MeshMergePatchGeometry(const Vector3 &n) : normal_sum(n) {}
public:
    MeshMergePatchGeometry() { normal_sum.zero(); }
    void clear() { normal_sum.zero(); }

    MeshMergePatchGeometry operator +(const MeshMergePatchGeometry &b) const
    { return MeshMergePatchGeometry(normal_sum + b.normal_sum); }
    MeshMergePatchGeometry &operator +=(const MeshMergePatchGeometry &b)
    { normal_sum += b.normal_sum; return *this; }
    MeshMergePatchGeometry operator -() const
    { return MeshMergePatchGeometry(-normal_sum); }
};
#endif /* !AS_PATCH_METRIC */
#endif /* !SANDER_PLANAR */

//! Structure to abstract information related to planarity metric
/*! \sa the report for the equations used */
struct MeshMergePatchMetrics
{
    MeshMergePatchGeometry geometry;
    double area;
    double perimeter;

    void clear() { geometry.clear(); area = 0.0; perimeter = 0.0; }
    MeshMergePatchMetrics() : geometry(), area(0.0), perimeter(0.0) {}
    MeshMergePatchMetrics(const MeshMergePatchGeometry &g, double a, double p)
    : geometry(g), area(a), perimeter(p) {}

    MeshMergePatchMetrics operator +(const MeshMergePatchMetrics &b) const
    { return MeshMergePatchMetrics(geometry + b.geometry, area + b.area, perimeter + b.perimeter); }
    MeshMergePatchMetrics &operator +=(const MeshMergePatchMetrics &b)
    {
        geometry += b.geometry;
        area += b.area;
        perimeter += b.perimeter;
        return *this;
    }
    MeshMergePatchMetrics operator -() const
    { return MeshMergePatchMetrics(-geometry, -area, -perimeter); }
};

//! A structure used internally by the chart-merging algorithm
struct MeshMergePatch : public MeshElement
{
    MeshMergePatchMetrics metrics;
    int patch;         //!< Patch id during write-back

    //! List of all edges, in no particular order
    MeshMergePatchEdgeGroup edges;
    //! List of all corners, in increasing order by id
    list<MeshPositionIterator> corners;

    MeshMergePatch *parent; //!< hierarchy structure

    MeshMergePatch() : parent(NULL) {}
};

static inline void get_bounding_box(const Vector2 *coords,
                                    int count,
                                    Vector2 &bmin, Vector2 &bmax)
{
    bmin.assign(HUGE_VAL, HUGE_VAL);
    bmax.assign(-HUGE_VAL, -HUGE_VAL);
    for (int i = 0; i < count; i++)
    {
        bmin = vector_min(bmin, coords[i]);
        bmax = vector_max(bmax, coords[i]);
    }
}

#ifdef SANDER_PLANAR
//! Computes the cost factors, but does not set edges
static void triangle_metrics(const Skeleton &skeleton,
                             MeshFaceConstIterator f,
                             MeshMergePatchMetrics &out)
{
    int dims = 3;
    static MAT *qMat = MNULL;
    static MAT *QMat = MNULL;
    static MAT *V = MNULL;
    static MAT *tmp = MNULL;
    MeshHalfEdgeConstIterator e = f->edges();

    // initialise constant matrices
    if (qMat == MNULL)
    {
        qMat = m_get(3, 1);
        QMat = m_get(3, 3);
        for (int i = 0; i < 3; i++)
        {
            m_set_val(qMat, i, 0, 1.0 / 3.0);
            for (int j = 0; j < 3; j++)
                m_set_val(QMat, i, j, (i == j) ? 1.0 / 6.0 : 1.0 / 12.0);
        }
    }

    // initialise the matrix of positions
    V = m_resize(V, dims, 3);
    Vector3 pos[3];
    for (int i = 0; i < 3; i++, e = e->next())
    {
        pos[i] = skeleton.ssd_to_3d(e->vertex()->position()->position);
        for (int j = 0; j < dims; j++)
            m_set_val(V, j, i, pos[i][j]);
    }

    // compute q and Q
    out.geometry.q.M = m_mlt(V, qMat, MNULL);
    out.geometry.Q.M = m_mlt(V, tmp = mmtr_mlt(QMat, V, tmp), MNULL);
    out.area = double_area(pos[1] - pos[0], pos[2] - pos[0]) * 0.5;
    Vector3 normal = normalise(cross(pos[1] - pos[0], pos[2] - pos[0]));
    out.geometry.q.M = sm_mlt(out.area, out.geometry.q.M, out.geometry.q.M);
    out.geometry.Q.M = sm_mlt(out.area, out.geometry.Q.M, out.geometry.Q.M);
    out.geometry.D.M = m_get(3, 3);
    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 3; j++)
            m_set_val(out.geometry.D.M, i, j, out.area * normal[i] * normal[j]);
    out.geometry.E.M = m_get(1, 3);
    for (int i = 0; i < 3; i++)
        m_set_val(out.geometry.E.M, 0, i, -2.0 * out.area * normal[i]);
    out.geometry.F = out.area;

    // compute perimeter
    Vector3 d[3] = {pos[0] - pos[1], pos[1] - pos[2], pos[2] - pos[0]};
    out.perimeter = length(d[0]) + length(d[1]) + length(d[2]);
}

static double evaluate_metric(const MeshMergePatchMetrics &m,
                              double compact_weight)
{
    static MAT *R = MNULL;
    static MAT *qqT = MNULL;
    static VEC *eig = VNULL;
    static MAT *eigv = MNULL;
    static VEC *tmp = VNULL;
    static VEC *e = VNULL;
    double small;
    size_t smalli = 0;

    qqT = mmtr_mlt(m.geometry.q.M, m.geometry.q.M, qqT);
    R = ms_mltadd(m.geometry.Q.M, qqT, -1.0 / m.area, R);
    eigv = m_resize(eigv, 3, 3);
    eig = symmeig(R, eigv, eig);

    small = HUGE_VAL;
    for (size_t i = 0; i < 3; i++)
    {
        if (v_entry(eig, i) < small)
        {
            small = v_entry(eig, i);
            smalli = i;
        }
    }

    /* The m.area term makes no difference to the simplification process
     * (since it cancels out the areas from the two sub-charts that will
     * be merged), but it makes the base metric equal to surface area when
     * all the charts are planar. This is used together with the limit factor
     * to determine when to stop.
     */
    double main_term = m.area + small;
    double compact_term = compact_weight / (4.0 * G_PI) * m.perimeter * m.perimeter;
#ifndef SANDER_ORIENT
    return main_term + compact_term;
#else
    double orient_term;
    eig = get_col(eigv, smalli, eig); // eig is now the normal
    eig = sv_mlt(1.0 / v_norm2(eig), eig, eig);
    tmp = mv_mlt(m.geometry.D.M, eig, tmp);
    e = get_row(m.geometry.E.M, 0, e);
    orient_term = in_prod(eig, tmp) - fabs(in_prod(e, eig)) + m.geometry.F;
    return main_term + compact_term + orient_term;
#endif
}

#else   /* !SANDER_METRIC */

//! Computes the cost factors, but does not set edges
static void triangle_metrics(const Skeleton &skeleton,
                             MeshFaceConstIterator f,
                             MeshMergePatchMetrics &out)
{
    MeshHalfEdgeConstIterator e = f->edges();
#ifdef AS_PATCH_METRIC
    VectorPosition pos[3];
    for (int i = 0; i < 3; i++, e = e->next())
        pos[i] = e->vertex()->position()->position;
    VectorPosition d1 = pos[1] - pos[0];
    VectorPosition d2 = pos[2] - pos[0];

    VectorPosition test = d1 + d2;
    out.geometry.occupied.clear();
    for (VectorPosition::const_iterator i = test.begin(); i != test.end(); i++)
        for (int j = 0; j < 4; j++)
            out.geometry.occupied.push_back(i->index + j);
    int dims = out.geometry.occupied.size();
    out.geometry.A.M = m_get(dims, dims);

    for (int i = 0; i < dims; i++)
        for (int j = 0; j < dims; j++)
        {
            int oi = out.geometry.occupied[i];
            int oj = out.geometry.occupied[j];
            m_set_val(out.geometry.A.M, i, j, (d1[oi] * d2[oj] - d2[oi] * d1[oj]) * 0.5);
        }
    out.geometry.P = skeleton.get_P();
#else
    Vector3 pos[3];
    for (int i = 0; i < 3; i++, e = e->next())
        pos[i] = skeleton.ssd_to_3d(e->vertex()->position()->position);
    Vector3 d1 = pos[1] - pos[0];
    Vector3 d2 = pos[2] - pos[0];
    out.geometry.normal_sum = cross(d1, d2) * 0.5;
#endif

    out.area = double_area(d1, d2) * 0.5;
    out.perimeter = length(d1) + length(d2) + length(d1 - d2);
}

#ifdef AS_PATCH_METRIC

//! Computes maximum projected area of surface
static double projected_area(const MeshMergePatchGeometry &g)
{
    static MAT *P1 = MNULL;  /* Submatrix of P */
    static MAT *C1 = MNULL;  /* Square root of P1 */
    static VEC *eig = VNULL;
    static MAT *ev = MNULL;

    int d = g.occupied.size();

    P1 = m_resize(P1, d, d);
    for (int i = 0; i < d; i++)
        for (int j = 0; j < d; j++)
        {
            int oi = g.occupied[i];
            int oj = g.occupied[j];
            m_set_val(P1, i, j, m_entry(g.P, oi, oj));
        }

    /* Compute the square root of P1, using the eigenvalue decomposition */
    C1 = m_resize(C1, d, d);
    ev = m_resize(ev, d, d);
    eig = symmeig(P1, ev, eig);
    m_zero(C1);
    for (int i = 0; i < d; i++)
    {
        double e = v_entry(eig, i);
        g_assert(e > -1e-3);
        if (e < 0.0) e = 0.0; else e = sqrt(e);
        for (int j = 0; j < d; j++)
            for (int k = 0; k < d; k++)
                m_add_val(C1, j, k, e * m_entry(ev, j, i) * m_entry(ev, k, i));
    }

    /* Compute C1.A.C1^T, using P1 as a temporary and overwriting ev */
    P1 = m_mlt(C1, g.A.M, P1);
    ev = mmtr_mlt(P1, C1, ev);

    /* Extract the largest singular value of C1.A.C1^T */
    eig = svd(ev, NULL, NULL, eig);
    return v_entry(eig, 0);
}

#else

//! Computes maximum projected area of surface
static double projected_area(const MeshMergePatchGeometry &g)
{
    return length(g.normal_sum);
}

#endif /* !AS_PATCH_METRIC */

static double evaluate_metric(const MeshMergePatchMetrics &m,
                              double compact_weight)
{
    double parea = projected_area(m.geometry);
    if (parea <= 0.0) return HUGE_VAL;
    double base_compact = compact_weight * (m.perimeter * m.perimeter / (4.0 * G_PI));
    return (base_compact + m.area) * (m.area / parea);
}

#endif /* !SANDER_PLANAR */

static void merge_metrics(MeshMergePatchEdgeIterator edge,
                          MeshMergePatchMetrics &out)
{
    MeshMergePatchIterator a = edge->patch;
    MeshMergePatchIterator b = edge->dual->patch;
    g_assert(a != b);
    g_assert(edge->mergeable);

    out = a->metrics + b->metrics;
    out.perimeter = a->metrics.perimeter + b->metrics.perimeter - 2 * edge->length;
}

//! Used by the straightener to add and subtract faces
/*! The \a InputIterator must point to \c MeshFaceIterator */
template<class InputIterator>
static void adjust_metrics(const Skeleton &skeleton,
                           MeshMergePatchMetrics &m,
                           InputIterator add_first, InputIterator add_last,
                           InputIterator remove_first, InputIterator remove_last,
                           double delta_perimeter)
{
    MeshMergePatchMetrics tmp;
    double old_perimeter = m.perimeter;
    for (InputIterator f_it = add_first; f_it != add_last; f_it++)
    {
        MeshFaceConstIterator f = *f_it;
        tmp.clear();
        triangle_metrics(skeleton, f, tmp);
        m += tmp;
    }
    for (InputIterator f_it = remove_first; f_it != remove_last; f_it++)
    {
        MeshFaceConstIterator f = *f_it;
        tmp.clear();
        triangle_metrics(skeleton, f, tmp);
        m += -tmp;
    }
    m.perimeter = old_perimeter + delta_perimeter;
}

//! Determines the cost of merging two patches
/*! \param edge the patch edge over which to merge
 * \param compact_weight weight to give to term that favours circular patches
 * \returns the cost if the merge is legal
 * \returns \c HUGE_VAL if the merge is illegal
 * \note Can in theory contribute cubic time to the simplification
 */
static double merge_cost(MeshMergePatchEdgeIterator edge, double compact_weight)
{
    static MeshMergePatchMetrics m;
    MeshMergePatchIterator a, b;

    a = edge->patch;
    b = edge->dual->patch;
    g_assert(a != b);
    if (!edge->mergeable) return HUGE_VAL;

    // check that merged patch will be a disc.
    int shared = 0;
    list<MeshPositionIterator>::const_iterator ia, ib;
    ia = a->corners.begin();
    ib = b->corners.begin();
    less_id<MeshPositionIterator> compare;
    while (ia != a->corners.end() && ib != b->corners.end())
    {
        if (*ia == *ib) shared++;
        compare(*ia, *ib) ? ia++ : ib++;
    }
    if (shared != 2) return HUGE_VAL;

    merge_metrics(edge, m);
    return evaluate_metric(m, compact_weight)
        - (evaluate_metric(a->metrics, compact_weight)
           + evaluate_metric(b->metrics, compact_weight));
}

//! Merges \a e1 with the edge clockwise of it
static void patch_coalesce_edges(MeshMergePatchEdgeIterator e1)
{
    MeshMergePatchEdgeIterator e2 = e1->prev;
    MeshMergePatchEdgeIterator d1 = e1->dual;
    MeshMergePatchEdgeIterator d2 = e2->dual;
    g_assert(d1->next == d2);

    e1->length += e2->length;
    e1->mergeable &= e2->mergeable;
    d1->length = e1->length;
    d1->mergeable = e1->mergeable;
    d1->position = d2->position;

    e1->prev = e2->prev;
    e1->prev->next = e1;
    d1->next = d2->next;
    d1->next->prev = d1;

    e1->patch->corners.remove(e2->position);
    d1->patch->corners.remove(e2->position);
    e1->patch->edges.erase(e2);
    d1->patch->edges.erase(d2);
}

//! Merges two patches permanently in the mesh
/*! \param edge the patch edge over which to merge
 * \param out the merged patch
 * \pre \c mesh_merge_cost does not return \c HUGE_VAL for this
 * collapse.
 */
static void mesh_merge_patches(MeshMergePatchEdgeIterator edge,
                               MeshMergePatchIterator out)
{
    MeshMergePatchIterator a, b;
    a = edge->patch;
    b = edge->dual->patch;
    g_assert(a != b);
    if (a->corners.size() > a->edges.size())
        g_error("%ld > %ld", (long) a->corners.size(), (long) a->edges.size());
    g_assert(a->edges.size() >= 2);
    g_assert(b->edges.size() >= 2);

    // merge cost data
    merge_metrics(edge, out->metrics);

    // merge corner lists
    set_union(a->corners.begin(), a->corners.end(),
              b->corners.begin(), b->corners.end(),
              back_inserter(out->corners),
              less_id<MeshPositionIterator>());

    // merge up edge lists
    out->edges.splice(out->edges.end(), a->edges);
    out->edges.splice(out->edges.end(), b->edges);
    // cut out joined edges
    edge->next->prev = edge->dual->prev;
    edge->prev->next = edge->dual->next;
    edge->dual->next->prev = edge->prev;
    edge->dual->prev->next = edge->next;
    // update self references
    for (MeshMergePatchEdgeIterator e = out->edges.begin(); e != out->edges.end(); e++)
        e->patch = out;

    /* Identify edges near join */
    MeshMergePatchEdgeIterator e[4];
    e[0] = edge->next;
    e[1] = e[0]->prev;
    e[2] = edge->dual->next;
    e[3] = e[2]->prev;

    /* Coalesce edges that are adjacent and between the same neighbours */
    if (e[0]->dual->patch == e[1]->dual->patch)
    {
        if (e[1] == e[2]) e[2] = e[0];
        g_assert(e[0]->prev == e[1]);
        patch_coalesce_edges(e[0]);
    }
    if (e[2]->dual->patch == e[3]->dual->patch)
    {
        g_assert(e[2] != e[3]); // otherwise only one neighbour
        g_assert(e[2]->prev == e[3]);
        patch_coalesce_edges(e[2]);
    }

    // erase join
    out->edges.erase(edge->dual);
    out->edges.erase(edge);

    a->parent = b->parent = &*out;

    // check some sanity conditions
    g_assert(out->corners.size() == out->edges.size());
    g_assert(out->corners.size() >= 2);
}

//! Determines whether it is safe to collapse edge
/*! It particularly checks for conditions that would lead to
 * fewer than 3 corners on a patch. These conditions cannot always
 * be checked from \c merge_cost, since they depend on factors
 * outside the two patches to be merged.
 */
static bool mesh_can_merge_patches(MeshMergePatchEdgeIterator edge)
{
    MeshMergePatchIterator a = edge->patch;
    MeshMergePatchIterator b = edge->dual->patch;
    g_assert(a != b);

    if (!edge->mergeable) return false;
    g_assert(edge->dual->mergeable);
    int our_corners = a->edges.size() + b->edges.size() - 2;
    g_assert(our_corners >= 2);

    /* The code below is closely modelled on the code in
     * mesh_merge_patches, to simulate what would happen to the surrounding
     * patches.
     */
    MeshMergePatchEdgeIterator e[4];
    e[0] = edge->next;
    e[1] = edge->dual->prev;
    e[2] = edge->dual->next;
    e[3] = edge->prev;

    /* Coalesce edges that are adjacent and between the same neighbours */
    if (e[0]->dual->patch == e[1]->dual->patch)
    {
        if (e[1] == e[2]) e[2] = e[0];
        if (e[0]->dual->patch->edges.size() <= 2) return false;
        our_corners--;
    }
    if (e[2]->dual->patch == e[3]->dual->patch)
    {
        if (e[2] != e[3])
        {
            size_t min = 3;
            if (e[0]->dual->patch == e[1]->dual->patch
                && e[2]->dual->patch == e[0]->dual->patch) min = 5;
            if (e[2]->dual->patch->edges.size() < min) return false;
            our_corners--;
        }
        else return false;
        // The alternative above means that we would have only one neighbour
    }
    return our_corners >= 2;
}

static void mesh_split_vertex(Mesh &m, MeshPositionIterator p)
{
    MeshVertexIterator v = p->first_vertex(), vnew;
    MeshHalfEdgeIterator first_edge = v->first_edge();
    MeshHalfEdgeIterator e = first_edge;
    int patch = e->face()->patch;
    int valence = 0;

    /* If we have only a single vertex but multiple patches, we
     * will want to start at a patch boundary
     */
    if (first_edge->dual()->prev()->vertex() == v)
    {
        bool found = false;
        do
        {
            if (e->face()->patch != e->dual()->face()->patch)
            {
                found = true;
                break;
            }
            e = e->dual()->next();
        } while (e != first_edge);
        if (!found) return; // normal vertex, don't bother
        v->first_edge() = first_edge = e;
        patch = e->dual()->face()->patch; // see below
        valence++;
    }

    /* We walk around the position, and every time there is a patch
     * change we check that there is also a vertex change.
     * In the code below, v always refers to the _original_ value
     * of the vertex field, while v1 is its updated value.
     *
     * To complicate matters, when splitting a solitary vertex we
     * only introduce _one fewer_ vertices than patch boundaries.
     * We handle this by using the dual patch in the above code, so
     * that we don't make a split right at the start
     */
    vnew = e->dual()->prev()->vertex();
    do
    {
        MeshVertexIterator v1 = e->next()->vertex(); // new value
        MeshVertexIterator &v2 = e->dual()->prev()->vertex();
        int patch2 = e->dual()->face()->patch;
        if (v != v2)
        {
            vnew = v2;
            valence++;
        }
        else if (patch != patch2)
        {
            // split
            vnew = m.vertices.insert_real(MeshVertex());
            // cannot do *vnew = *v, since that overwrites ID
            vnew->position() = v1->position();
            vnew->data = v1->data;
            vnew->first_edge() = v1->first_edge();
            v1->first_edge() = e;
            valence++;
        }
        v = v2;
        v2 = vnew;
        e = e->dual()->next();
        patch = patch2;
    } while (e != first_edge);

    if (valence >= 3) p->patch_corner = true;
}

static void mesh_split_vertices(Mesh &m)
{
    // split vertices where necessary
    for (MeshPositionIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
        mesh_split_vertex(m, p);
}

template<class ForwardIterator>
static void mesh_split_vertices(Mesh &m, ForwardIterator first, ForwardIterator last)
{
    id_set<MeshPositionIterator>::type seen;

    for (ForwardIterator fi = first; fi != last; fi++)
    {
        MeshHalfEdgeIterator e = (*fi)->edges();
        for (int i = 0; i < 3; i++, e = e->next())
        {
            MeshPositionIterator p = e->vertex()->position();
            if (seen.insert(p).second)
                mesh_split_vertex(m, p);
        }
    }
}

//! Merge vertices with the same position, ignoring texcoords
/*! Any materials and simplifications are also wiped out.
 */
static void mesh_remove_patches(Mesh &m)
{
    vector<MeshRawVertex> vertices;
    vector<vector<int> > faces;
    vector<int> materials;
    id_map<MeshVertexIterator, MeshVertexIterator>::type vertex_remap;

    /* Rip out LOD and materials */
    m.clear_lod();
    m.materials.clear();
    m.textures.clear();
    m.tangent_maps.clear();
    m.materials.push_back(MeshMaterial());
    m.materials.front().texture() = mesh_null_texture;
    m.materials.front().tangent_map() = mesh_null_tangent_map;
    for (MeshFaceIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
    {
        f->material() = f->patch ? m.materials.begin() : mesh_null_material;
        f->patch = f->patch ? 1 : 0;
    }

    /* Don't rely too heavily on first_vertex and first_edge (just assume they
     * are some random vertex and edge), since they could be borked by a crummy
     * existing parametrisation.
     */
    for (MeshPositionIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
    {
        MeshHalfEdgeIterator start = p->first_vertex()->first_edge();
        g_assert(start->next()->vertex()->position() == p);
        if (start->face()->patch == 0)
        {
            MeshHalfEdgeIterator e = start;
            do
            {
                e = e->dual()->next();
            } while (e->face()->patch == 0 && e != start);
            start = e;
            g_assert(start->face()->patch != 0);
        }
        MeshVertexIterator key = start->next()->vertex();
        key->first_edge() = start;
        MeshHalfEdgeIterator e = start;
        do
        {
            if (e->face()->patch != 0 && e->next()->vertex() != key)
                vertex_remap[e->next()->vertex()] = key;
            if (e->face()->patch != 0 && e->dual()->face()->patch == 0)
                key->first_edge() = e;
            e = e->dual()->next();
        } while (e != start);
        if (p->first_vertex()->first_edge()->face()->patch != 0)
            p->first_vertex() = key;
    }

    // Apply remapping
    for (MeshHalfEdgeIterator e = m.edges.real_begin(); e != m.edges.real_end(); e++)
        if (vertex_remap.count(e->vertex()))
            e->vertex() = vertex_remap[e->vertex()];

    // Eliminate vertices that were mapped away
    for (id_map<MeshVertexIterator, MeshVertexIterator>::type::iterator i = vertex_remap.begin(); i != vertex_remap.end(); i++)
        m.vertices.erase(i->first);

    mesh_set_patch_corners(m);
    g_assert(mesh_manifold(m, MESH_LEVEL_PATCHES));
}

//! Compare a pair on the first element
template<typename T>
class greater_1st : binary_function<T, T, bool>
{
public:
    bool operator() (const T &a, const T &b)
    { return a.first > b.first; }
};

struct StraightenItem
{
    MeshPositionIterator cur;
    MeshPositionIterator prev;
    MeshHalfEdgeIterator prev_edge;
    double prio;
    bool done;
};

//! Straightens the edge between \a start and \a end
/*! The edge between \a start and \a end between \a patch1 and \a patch2 is
 * replaced by the shortest possible one, as long as the edge does not
 * meet any other boundaries.
 *
 * An A* algorithm is used. The whole process is complicated by the
 * possible existence of multiple edges between the two patches, only
 * one of which is desired.
 *
 * \param start, end The endpoints of the current path. \a start->\a end must
 * be clockwise on \a p1.
 * \param skeleton The skeleton used to weight the edges
 * \param p1, p2 The two patches bordering the edge
 * \param constrained Set to true to limit the vertices to those on the
 * current path
 * \param update Set to true to make permanent changes, false to only
 * compute the metrics.
 * \param compact_weight The weight to use to favour circular patches
 * \returns The modified metrics of \a p1 and \a p2
 *
 * \pre \c MeshPosition::internal_data must be \c NULL everywhere on the
 * two patches.
 * \pre \c MeshHalfEdge::internal_data must be \c NULL everywhere on the two
 * patches.
 * \pre \c p1->patch!=p2->patch
 * \pre \c start!=end
 * \post \c MeshPosition::internal_data is again \c NULL everywhere
 * \post \c MeshHalfEdge::internal_data is again \c NULL everywhere
 * \note The function does not depend on \c MeshVertex::first_edge other
 * than to start walking the neighbourhood of a vertex, because it does
 * not yet have the correct value.
 */
static pair<double, double> mesh_patch_straighten_edge(const Skeleton &skeleton,
                                                       MeshPositionIterator start,
                                                       MeshPositionIterator end,
                                                       MeshMergePatchIterator p1,
                                                       MeshMergePatchIterator p2,
                                                       bool constrained,
                                                       bool update,
                                                       double compact_weight)
{
    typedef pair<double, MeshPositionIterator> pq_item;
    priority_queue<pq_item, vector<pq_item>, greater_1st<pq_item> > pq;
    // Must never be a vector, as iterators would become invalid
    list<StraightenItem> infos;
    StraightenItem item;
    MeshPositionIterator cur;
    MeshHalfEdgeIterator e;
    double old_perimeter = 0.0, new_perimeter = 0.0;
    int patch1 = p1->patch;
    int patch2 = p2->patch;

    g_assert(start != end);
    g_assert(patch1 != patch2);

    /* We must first identify the original path, so that it is not
     * confused with other boundaries between the same pair of patches.
     * We identify them by giving them info records with infinite
     * priority. Illegal points are similarly identified, but with
     * done == true.
     */
    cur = start;
    item.prio = HUGE_VAL;
    item.prev = mesh_null_position;
    item.prev_edge = mesh_null_half_edge;
    item.done = false;
    while (cur != end)
    {
        item.cur = cur;
        infos.push_back(item);
        cur->internal_data = &infos.back();
        e = cur->first_vertex()->first_edge();
        g_assert(e->next()->vertex()->position() == cur);
        while (e->face()->patch != patch2 || e->dual()->face()->patch != patch1)
            e = e->dual()->next();
        old_perimeter += length(skeleton.ssd_to_3d(e->prev()->vertex()->position()->position)
                                - skeleton.ssd_to_3d(cur->position));
        cur = e->prev()->vertex()->position();
    }
    item.cur = end;
    infos.push_back(item);
    end->internal_data = &infos.back();

    infos.front().prio = length(skeleton.ssd_to_3d(end->position)
                                - skeleton.ssd_to_3d(start->position));
    pq.push(make_pair(infos.front().prio, start));
    while (!pq.empty())
    {
        cur = pq.top().second;
        if (cur == end) break;
        pq.pop();
        StraightenItem *info = (StraightenItem *) cur->internal_data;
        g_assert(info);
        if (info->done) continue;
        info->done = true;

        e = cur->first_vertex()->first_edge();
        g_assert(e->next()->vertex()->position() == cur);
        do
        {
            if ((e->face()->patch == patch1 || e->face()->patch == patch2)
                && (e->dual()->face()->patch == patch1 || e->dual()->face()->patch == patch2))
            {
                MeshPositionIterator nxt = e->prev()->vertex()->position();
                StraightenItem *nxt_info = (StraightenItem *) nxt->internal_data;
                if (!nxt_info)
                {
                    /* First check that nxt isn't an unwanted boundary vertex.
                     * We have pre-identified the old path (including end),
                     * so any other boundary vertex is a no-go.
                     */
                    item.cur = nxt;
                    item.done = false;

                    if (!constrained)
                    {
                        MeshHalfEdgeIterator e2 = e->dual();
                        do
                        {
                            if (e2->face()->patch != e2->dual()->face()->patch)
                            {
                                item.done = true; // tag as illegal
                                break;
                            }
                            e2 = e2->dual()->next();
                        } while (e2 != e->dual());
                    }
                    else
                        item.done = true;
                    infos.push_back(item);
                    nxt->internal_data = nxt_info = &infos.back();
                }

                double nxt_prio = info->prio
                    + length(skeleton.ssd_to_3d(cur->position) - skeleton.ssd_to_3d(nxt->position))
                    + length(skeleton.ssd_to_3d(nxt->position) - skeleton.ssd_to_3d(end->position))
                    - length(skeleton.ssd_to_3d(cur->position) - skeleton.ssd_to_3d(end->position));
                if (!nxt_info->done && nxt_prio < nxt_info->prio)
                {
                    nxt_info->prev = cur;
                    nxt_info->prev_edge = e->dual();
                    nxt_info->prio = nxt_prio;
                    pq.push(make_pair(nxt_prio, nxt));
                }
            }
            e = e->dual()->next();
        } while (e != cur->first_vertex()->first_edge());
    }
    g_assert(!pq.empty());

    /* We now floodfill the areas between the original and modified paths.
     * The floodfill is started from the modified path. The edges on the
     * new path their internal_data set to some arbitrary non-NULL value,
     * to act as sentinels.
     */
    queue<MeshFaceIterator> q;
    vector<MeshFaceIterator> move_p1_p2, move_p2_p1;
    while (cur != start)
    {
        StraightenItem *info = (StraightenItem *) cur->internal_data;
        g_assert(info);
        if (info->prev_edge->face()->patch == patch2)
        {
            info->prev_edge->face()->patch = patch1;
            move_p2_p1.push_back(info->prev_edge->face());
            q.push(info->prev_edge->face());
        }
        if (info->prev_edge->dual()->face()->patch == patch1)
        {
            info->prev_edge->dual()->face()->patch = patch2;
            move_p1_p2.push_back(info->prev_edge->dual()->face());
            q.push(info->prev_edge->dual()->face());
        }
        info->prev_edge->internal_data = &q;
        info->prev_edge->dual()->internal_data = &q;
        cur = info->prev;
    }
    while (!q.empty())
    {
        MeshFaceIterator cur = q.front();
        q.pop();
        MeshHalfEdgeIterator e = cur->edges();
        for (int i = 0; i < 3; i++, e = e->next())
            if (e->internal_data == NULL)
            {
                MeshFaceIterator nxt = e->dual()->face();
                if (nxt->patch != cur->patch)
                {
                    nxt->patch = cur->patch;
                    if (nxt->patch == patch1)
                        move_p2_p1.push_back(nxt);
                    else
                        move_p1_p2.push_back(nxt);
                    q.push(nxt);
                }
            }
    }

    pair<MeshMergePatchMetrics, MeshMergePatchMetrics> new_metrics;
    new_metrics.first = p1->metrics;
    adjust_metrics(skeleton,
                   new_metrics.first,
                   move_p2_p1.begin(), move_p2_p1.end(),
                   move_p1_p2.begin(), move_p1_p2.end(),
                   new_perimeter - old_perimeter);
    new_metrics.second = p2->metrics;
    adjust_metrics(skeleton,
                   new_metrics.second,
                   move_p1_p2.begin(), move_p1_p2.end(),
                   move_p2_p1.begin(), move_p2_p1.end(),
                   new_perimeter - old_perimeter);

    // Either do or don't update (currently we're half-way)
    if (!update)
    {
        for (size_t i = 0; i < move_p1_p2.size(); i++)
            move_p1_p2[i]->patch = patch1;
        for (size_t i = 0; i < move_p2_p1.size(); i++)
            move_p2_p1[i]->patch = patch2;
    }
    else
    {
        p1->metrics = new_metrics.first;
        p2->metrics = new_metrics.second;
    }

    // Clean up sentinels
    cur = end;
    while (cur != start)
    {
        StraightenItem *info = (StraightenItem *) cur->internal_data;
        info->prev_edge->internal_data = NULL;
        info->prev_edge->dual()->internal_data = NULL;
        cur = info->prev;
    }

    // Clean up internal_data
    for (list<StraightenItem>::iterator i = infos.begin(); i != infos.end(); i++)
        i->cur->internal_data = NULL;

    return make_pair(evaluate_metric(new_metrics.first, compact_weight),
                     evaluate_metric(new_metrics.second, compact_weight));
}

/*! \returns the number of patches, excluding holes
 * \post the boundary vertices have not been split
 *
 * \c MeshFace::internal_data is used to map faces to patches.
 * \c MeshPosition::internal_data is used for edge straightening;
 */
static int mesh_create_patches(Mesh &m, int num_patches, double limit,
                               double compact_weight,
                               double constrain_tolerance)
{
    // the list of top-level patches
    MeshMergePatchGroup patches;
    double total_metric = 0.0;
    double limit_metric = HUGE_VAL;
    double total_area = 0.0;
    const Skeleton &skeleton = m.skeleton.get_read();

    for (MeshFaceIterator f = m.faces.whole_begin(); f != m.faces.whole_end(); f++)
    {
        MeshMergePatchEdgeIterator b[3];
        // We go backwards since we push patch edges in from the front
        MeshHalfEdgeIterator e = f->edges()->prev();
        MeshMergePatchIterator p;

        patches.push_front(MeshMergePatch());
        p = patches.begin();
        f->internal_data = &patches.front();
        triangle_metrics(skeleton, f, p->metrics);
        if (mesh_true_face(f))
        {
            total_metric += evaluate_metric(p->metrics, compact_weight);
            total_area += p->metrics.area;
        }
        for (int i = 0; i < 3; i++, e = e->prev())
        {
            MeshFaceIterator f2 = e->dual()->face();

            p->corners.push_back(e->vertex()->position());
            p->edges.push_front(MeshMergePatchEdge());
            b[i] = p->edges.begin();
            b[i]->position = e->prev()->vertex()->position();
            b[i]->mergeable = (f->patch
                               && f2->patch
                               && e->next()->vertex() == e->dual()->prev()->vertex()
                               && e->prev()->vertex() == e->dual()->next()->vertex()
                               && f->material() == f2->material());
            b[i]->patch = p;
        }
        for (int i = 0; i < 3; i++)
        {
            b[i]->prev = b[(i + 1) % 3];
            b[i]->next = b[(i + 2) % 3];
        }
        p->parent = NULL;
        g_assert(p->corners.size() <= p->edges.size());
    }

    /* Fill in the dual edges. Note that the loop structure must not be
     * messed with.
     * Also sort the corner lists and figure out lengths.
     */
    for (MeshFaceIterator f = m.faces.whole_begin(); f != m.faces.whole_end(); f++)
    {
        MeshMergePatch *p = (MeshMergePatch *) f->internal_data;
        MeshHalfEdgeIterator e = f->edges();
        MeshMergePatchEdgeIterator b = p->edges.begin();
        for (int i = 0; i < 3; i++, e = e->next(), b++)
        {
            MeshHalfEdgeIterator e2 = e->dual();
            MeshFaceIterator f2 = e2->face();
            MeshMergePatch *p2 = (MeshMergePatch *) f2->internal_data;
            MeshMergePatchEdgeIterator b2 = p2->edges.begin();
            while (e2 != f2->edges())
            {
                b2++;
                e2 = e2->prev();
            }
            b->dual = b2;
            b->length = length(skeleton.ssd_to_3d(b->position->position)
                               - skeleton.ssd_to_3d(b->prev->position->position));
        }
        p->corners.sort(less_id<MeshPositionIterator>());
    }

    // Set up the priority queue
    indexed_priority_queue<
        MeshMergePatchEdgeIterator,
        double,
        vector<pair<MeshMergePatchEdgeIterator, double> >,
        greater<double>,
        id_map<MeshMergePatchEdgeIterator, vector<pair<MeshMergePatchEdgeIterator, double> >::difference_type>::type> pq;

    for (MeshMergePatchIterator p = patches.begin(); p != patches.end(); p++)
    {
        for (MeshMergePatchEdgeIterator e = p->edges.begin(); e != p->edges.end(); e++)
            if (e->mergeable && e->patch->id < e->dual->patch->id)
            {
                double cost = merge_cost(e, compact_weight);
                if (finite(cost))
                    pq[e] = cost;
            }
    }

    // Go for it
    int left = m.faces.real_size();
    if (finite(limit))
        limit_metric = total_area * limit;
    while (!pq.empty() && left > num_patches && total_metric + pq.top().second <= limit_metric)
    {
        double delta = pq.top().second;
        MeshMergePatchEdgeIterator e = pq.top().first;
        MeshMergePatchIterator a = e->patch;
        MeshMergePatchIterator b = e->dual->patch;
        MeshMergePatchIterator c;
        pq.pop();
        // check if it is still valid (produces enough corners everywhere)
        if (!mesh_can_merge_patches(e)) continue;
        total_metric += delta;

        // nuke all other edges to/from a and b
        for (MeshMergePatchEdgeIterator e2 = a->edges.begin(); e2 != a->edges.end(); e2++)
        {
            pq.erase(e2);
            pq.erase(e2->dual);
        }
        for (MeshMergePatchEdgeIterator e2 = b->edges.begin(); e2 != b->edges.end(); e2++)
        {
            pq.erase(e2);
            pq.erase(e2->dual);
        }

        // do the merge
        patches.push_front(MeshMergePatch());
        c = patches.begin();
        mesh_merge_patches(e, c);
        c->parent = NULL; // top-level
        left--;

        // free some memory from the old patches (still needed for hierarchy)
        a->metrics.clear(); a->corners.clear(); a->edges.clear();
        b->metrics.clear(); b->corners.clear(); b->edges.clear();

        // update the priority queue
        for (MeshMergePatchEdgeIterator e2 = c->edges.begin(); e2 != c->edges.end(); e2++)
            if (e2->mergeable)
            {
                MeshMergePatchEdgeIterator e3;
                e3 = (c->id < e2->dual->patch->id) ? e2 : e2->dual;
                double cost = merge_cost(e3, compact_weight);
                if (finite(cost))
                    pq[e3] = cost;
            }

        if (left % 100 == 0)
            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO,
                  "%d patches remaining", left);
    }

    /* Write back patch ids, using path compression to ensure linear time
     * performance.
     */
    int pool_id = 1;
    for (MeshMergePatchIterator p = patches.begin(); p != patches.end(); p++)
        p->patch = 0;
    for (MeshFaceIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
    {
        MeshMergePatch *cur, *root, *nxt;

        cur = (MeshMergePatch *) f->internal_data;
        root = cur;
        while (root->parent) root = root->parent;
        if (root->patch == 0) root->patch = pool_id++;
        f->patch = root->patch;
        // path compression
        while (cur != root)
        {
            nxt = cur->parent;
            cur->parent = root;
            cur = nxt;
        }
    }

    /* Straighten out the edges.
     * mesh_patch_straighten_edge wants NULL internal_data.
     */
    for (MeshPositionIterator p = m.positions.whole_begin(); p != m.positions.whole_end(); p++)
        p->internal_data = NULL;
    for (MeshHalfEdgeIterator e = m.edges.whole_begin(); e != m.edges.whole_end(); e++)
        e->internal_data = NULL;

    for (MeshMergePatchIterator p = patches.begin(); p != patches.end(); p++)
    {
        if (p->patch == 0) continue;
        for (MeshMergePatchEdgeIterator e = p->edges.begin(); e != p->edges.end(); e++)
        {
            MeshMergePatchIterator p2 = e->dual->patch;
            if (p2->patch <= p->patch) continue;
            if (e->position == e->dual->position) continue; /* 2-patch sphere */

            pair<double, double> old_metrics, new_metrics;
            old_metrics = make_pair(evaluate_metric(p->metrics, compact_weight),
                                    evaluate_metric(p->metrics, compact_weight));
            new_metrics = mesh_patch_straighten_edge(skeleton,
                                                     e->position, e->dual->position,
                                                     p, p2,
                                                     false, false,
                                                     compact_weight);
            double new_sum = new_metrics.first + new_metrics.second;
            double old_sum = old_metrics.first + old_metrics.second;
            if (new_sum < constrain_tolerance * old_sum)
                new_metrics = mesh_patch_straighten_edge(skeleton,
                                                         e->position, e->dual->position,
                                                         p, p2,
                                                         false, true,
                                                         compact_weight);
            else
                new_metrics = mesh_patch_straighten_edge(skeleton,
                                                         e->position, e->dual->position,
                                                         p, p2,
                                                         true, true,
                                                         compact_weight);
        }
    }

    g_assert(pool_id == left + 1);

    /* Split boundary vertices */
    mesh_split_vertices(m);

    return left;
}

double mesh_segmentation_metric(const Mesh &m, double compact_weight)
{
    typedef fast_map<int, vector<MeshFaceConstIterator> >::type patch_map_t;
    patch_map_t patches;
    const Skeleton &skeleton = m.skeleton.get_read();

    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
        patches[f->patch].push_back(f);

    double total = 0.0;
    double area = 0.0;
    for (patch_map_t::const_iterator p = patches.begin(); p != patches.end(); p++)
    {
        MeshMergePatchMetrics metrics;
        double perimeter = 0.0;

        for (vector<MeshFaceConstIterator>::const_iterator f = p->second.begin(); f != p->second.end(); f++)
        {
            MeshHalfEdgeConstIterator e = (*f)->edges();
            for (int i = 0; i < 3; i++, e = e->next())
                if (e->dual()->face()->patch != p->first)
                {
                    VectorPosition pos1 = e->next()->vertex()->position()->position;
                    VectorPosition pos2 = e->prev()->vertex()->position()->position;
#if defined(SANDER) || defined(AS_METRIC)
                    perimeter += length(pos1 - pos2);
#else
                    perimeter += length(skeleton.ssd_to_3d(pos1 - pos2));
#endif
                }
        }

        triangle_metrics(skeleton, p->second.front(), metrics);
        metrics.perimeter = 0.0;
        adjust_metrics(skeleton, metrics, p->second.begin() + 1, p->second.end(),
                       p->second.begin(), p->second.begin(), perimeter);
        total += evaluate_metric(metrics, compact_weight);
        area += metrics.area;
    }
    return total / area;
}

//! Determines whether the boundary of the specified patch is free of self-intersections
/*! \note \c ForwardIterator is an iterator pointing to \c MeshFaceIterator
 */
template<class ForwardIterator>
static bool mesh_patch_legal_boundary(ForwardIterator first,
                                      ForwardIterator last,
                                      MeshHalfEdgeIterator &edge1,
                                      MeshHalfEdgeIterator &edge2)
{
    vector<MeshHalfEdgeIterator> boundary;

    // Find the boundary edges
    for (ForwardIterator fi = first; fi != last; fi++)
    {
        MeshHalfEdgeIterator e = (*fi)->edges();
        int i;
        for (i = 0; i < 3; i++, e = e->next())
        {
            if (e->dual()->face()->patch != (*fi)->patch)
                boundary.push_back(e);
        }
    }

    // Check them for intersections
    for (vector<MeshHalfEdgeIterator>::const_iterator ei1 = boundary.begin(); ei1 != boundary.end(); ei1++)
        for (vector<MeshHalfEdgeIterator>::const_iterator ei2 = ei1 + 1; ei2 != boundary.end(); ei2++)
        {
            MeshHalfEdgeIterator e1 = *ei1;
            MeshHalfEdgeIterator e2 = *ei2;
            // ignore adjacent edges
            if (e1->next()->vertex() == e2->prev()->vertex()
                || e1->prev()->vertex() == e2->next()->vertex()) continue;

            // grab coordinates
            Vector2 uv1[2], uv2[2];
            uv1[0] = e1->next()->vertex()->data.texcoords;
            uv1[1] = e1->prev()->vertex()->data.texcoords;
            uv2[0] = e2->next()->vertex()->data.texcoords;
            uv2[1] = e2->prev()->vertex()->data.texcoords;

            if (intersects(uv1[0], uv1[1], uv2[0], uv2[1], false))
            {
                edge1 = e1;
                edge2 = e2;
                return false;
            }
        }
    return true;
}

//! Determines whether there are any obstructions to straight parametric boundaries
/*! \note \c ForwardIterator is an iterator pointing to \c MeshFaceIterator
 * \post \c MeshVertex::internal_data is left meaningless
 */
template<class ForwardIterator>
static bool mesh_patch_straight_boundary(ForwardIterator first,
                                         ForwardIterator last)
{
    /* We first find a corner vertex, then walk the boundary.
     * We use internal_data to tag the vertices along each patch edge.
     * Finally, internal edges are checked for whether they connect
     * two boundary vertices from the same patch edge.
     */
    MeshVertexIterator start = mesh_null_vertex, corner;
    char *id = NULL;

    for (ForwardIterator i = first; i != last; i++)
    {
        MeshHalfEdgeIterator e = (*i)->edges();
        for (int j = 0; j < 3; j++, e = e->next())
            if (e->vertex()->position()->patch_corner)
            {
                start = e->vertex();
                break;
            }
    }

    if (start == mesh_null_vertex)
    {
        g_warning("No corners on patch");
        return false;
    }

    corner = start;
    do
    {
        id++;
        MeshVertexIterator v = corner;
        do
        {
            v->internal_data = id;
            v = v->first_edge()->prev()->vertex();
        } while (!v->position()->patch_corner);
        corner = v;
    } while (corner != start);
    if (id < (char *) NULL + 3)
    {
        g_warning("Fewer than 3 corners on patch");
        return false;
    }

    for (ForwardIterator i = first; i != last; i++)
    {
        MeshHalfEdgeConstIterator e = (*i)->edges();
        for (int j = 0; j < 3; j++, e = e->next())
        {
            if (e->dual()->face()->patch == e->face()->patch)
            {
                MeshVertexConstIterator v1 = e->next()->vertex();
                MeshVertexConstIterator v2 = e->prev()->vertex();
                if (mesh_position_patch_valence(v1->position()) >= 2
                    && mesh_position_patch_valence(v2->position()) >= 2)
                {
                    if (v1->internal_data == v2->internal_data)
                        return false;
                    if (v2->position()->patch_corner
                        && ((char *) v1->internal_data + 1 == (char *) v2->internal_data
                            || (v1->internal_data == id
                                && v2->internal_data == (char *) NULL + 1)))
                        return false;
                }
            }
        }
    }
    return true;
}


//! Determines which vertices to pin down for \c mesh_parametrise_flatten
/*! The result depends on the value of \a pass.
 * - Pass 0 pins two widely separated boundary vertices
 * - Pass 1 pins all boundary vertices, with the corners left as is and
 *   the edges straightened
 * - Pass 2 is similar to pass 1, but places the corners around a circle.
 */
template<class ForwardIterator, class OutputIterator>
static void pin_vertices(ForwardIterator first, ForwardIterator last,
                         OutputIterator out, int pass)
{
    VectorPosition pin0pos;
    MeshVertexIterator pin1;
    double pin_dist = -1.0;

    if (pass == 0)
    {
        /* Go and find two boundary vertices. We want a widely separated pair,
         * so we choose an arbitrary one then the most distant one from it.
         */
        for (ForwardIterator f = first; f != last; f++)
        {
            MeshHalfEdgeIterator e = (*f)->edges();
            for (int i = 0; i < 3; i++, e = e->next())
            {
                if (e->dual()->face()->patch != (*f)->patch)
                {
                    pin0pos = e->next()->vertex()->position()->position;
                    e->next()->vertex()->data.texcoords.assign(0.0, 0.0);
                    *out++ = e->next()->vertex();
                    goto found1;
                }
            }
        }
        g_assert_not_reached();
    found1:
        for (ForwardIterator f = first; f != last; f++)
        {
            MeshHalfEdgeIterator e = (*f)->edges();
            for (int i = 0; i < 3; i++, e = e->next())
            {
                if (e->dual()->face()->patch != (*f)->patch)
                {
                    VectorPosition pos = e->next()->vertex()->position()->position;
                    if (length(pos - pin0pos) > pin_dist)
                    {
                        pin1 = e->next()->vertex();
                        pin_dist = length(pos - pin0pos);
                    }
                }
            }
        }
        g_assert(pin_dist > 0.0);
        pin1->data.texcoords.assign(0.0, 1.0);
        *out++ = pin1;
    }
    else
    {
        /* Find patch corner */
        for (ForwardIterator f = first; f != last; f++)
        {
            MeshHalfEdgeIterator e = (*f)->edges();
            for (int i = 0; i < 3; i++, e = e->next())
            {
                pin1 = e->prev()->vertex();
                if (pin1->position()->patch_corner)
                    goto found2;
            }
        }
        return;

    found2:
        double total_arc = 0.0, cur_arc = 0.0;

        if (pass == 2)
        {
            MeshVertexIterator v = pin1, tmp;
            MeshHalfEdgeIterator e;
            do
            {
                e = v->first_edge();
                tmp = e->prev()->vertex();
                total_arc += length(v->position()->position - tmp->position()->position);
                v = tmp;
            } while (v != pin1);
            pin1->data.texcoords.assign(1.0, 0.0);
        }

        MeshVertexIterator v0 = pin1;
        do
        {
            MeshHalfEdgeIterator e;
            MeshVertexIterator v = v0, v1;
            double arc = 0.0, arc2 = 0.0;
            Vector2 delta;
            do
            {
                e = v->first_edge();
                MeshVertexIterator tmp = e->prev()->vertex();
                arc += length(v->position()->position - tmp->position()->position);
                v = tmp;
            } while (!v->position()->patch_corner);
            v1 = v;

            if (pass == 2)
            {
                cur_arc += arc;
                double ang = G_PI * 2.0 * cur_arc / total_arc;
                v1->data.texcoords.assign(cos(ang), sin(ang));
            }

            delta = (v1->data.texcoords - v0->data.texcoords) / arc;
            v = v0;
            do
            {
                e = v->first_edge();
                MeshVertexIterator tmp = e->prev()->vertex();
                if (arc > 0.0)
                    v->data.texcoords = v0->data.texcoords + delta * arc2;
                *out++ = v;
                arc2 += length(v->position()->position - tmp->position()->position);
                v = tmp;
            } while (v != v1);
            v0 = v1;
        } while (v0 != pin1);
    }
}

//! A drop-in wrapper around \c iter_splsqr.
/*! The original LSQR implementation is buggy. The bug seems to
 * appear mainly with small matrices, and leads to a failure to
 * converge. Thus if convergence fails, we switch to QR.
 * In this case \a steps is set to -1.
 */
static VEC *solve_least_squares(SPMAT *A, VEC *b, double tol,
                                VEC *x, int limit, int *steps)
{
    x = v_resize(x, A->n);
    x = iter_splsqr(A, b, tol, x, limit, steps);
    if (*steps > limit)
    {
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "LSQR failed, falling back to QR");

        MAT *Ad = sp_m2dense(A, NULL);
        VEC *diag = v_get(A->m);
        Ad = QRfactor(Ad, diag);
        x = QRsolve(Ad, diag, b, x);
        *steps = -1;
        M_FREE(Ad);
        V_FREE(diag);
    }
    return x;
}

//! Implement Least Squares Conformal Mapping
/*! \c MeshVertex::internal_data is used to encode vertex ids. The
 * number is added to ((char *) NULL) + 2
 */
template<class ForwardIterator>
static void mesh_parametrise_flatten(ForwardIterator first,
                                     ForwardIterator last,
                                     int pass)
{
    int N = 0;
    int F = 0;
    char * const base = ((char *) NULL) + 2;
    vector<MeshVertexIterator> inverse_id;
    vector<MeshVertexIterator> pinned;

    pin_vertices(first, last, back_inserter(pinned), pass);
    int P = pinned.size();
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "P = %d pass = %d", P, pass);
    if (P < 2) return;

    // tag all vertices as unseen, and count faces
    for (ForwardIterator f = first; f != last; f++, F++)
    {
        MeshHalfEdgeIterator e = (*f)->edges();
        for (int i = 0; i < 3; i++, e = e->next())
            e->vertex()->internal_data = NULL;
    }

    // reserve space in the table, although we don't know the count yet
    for (int i = 0; i < P; i++)
        pinned[i]->internal_data = base - 1;
    // compact vertex ids for addressing the sparse matrix
    for (ForwardIterator f = first; f != last; f++)
    {
        MeshHalfEdgeIterator e = (*f)->edges();
        for (int i = 0; i < 3; i++, e = e->next())
            if (!e->vertex()->internal_data)
            {
                inverse_id.push_back(e->vertex());
                e->vertex()->internal_data = base + N++;
            }
    }
    // now we can assign id's to the pinned vertices
    for (int i = 0; i < P; i++)
    {
        inverse_id.push_back(pinned[i]);
        pinned[i]->internal_data = base + N++;
    }

    int NP = N - P;    // number of free vertices
    if (NP == 0) return; // nothing to solve for

    // obtain memory
    SPMAT *A = sp_get(2 * F, 2 * NP, 6);
    VEC *b = v_get(2 * F);
    // these are used to compute b
    MAT *fixed = m_get(2 * F, 2 * P);
    VEC *ufixed = v_get(2 * P);
    // fill in the sparse matrix
    int fi = 0;
    for (ForwardIterator f = first; f != last; f++, fi++)
    {
        MeshHalfEdgeIterator e = (*f)->edges();
        VectorPosition pos[3];       // positions of vertices
        Vector2 local[3];     // coordinates of vertices in a plane
        VectorPosition xaxis, yaxis; // basis for the triangle
        int index[3];         // compacted vertex indices
        for (int i = 0; i < 3; i++, e = e->next())
        {
            pos[i] = e->vertex()->position()->position;
            index[i] = ((char *) e->vertex()->internal_data) - base;
        }
        xaxis = pos[1] - pos[0]; g_assert(length(xaxis) > 0);
        yaxis = pos[2] - pos[0]; g_assert(length(yaxis) > 0);
        xaxis = normalise(xaxis);
        yaxis = normalise(yaxis - (xaxis * yaxis) * xaxis);
        g_assert(fabs(xaxis * yaxis) < 1e-8);
        for (int i = 0; i < 3; i++)
        {
            local[i][0] = pos[i] * xaxis;
            local[i][1] = pos[i] * yaxis;
        }

        double d = 0.0;
        for (int i = 0; i < 3; i++)
        {
            int j = (i + 1) % 3;
            d += local[i][0] * local[j][1] - local[i][1] * local[j][0];
        }
        g_assert(d > 0.0);
        double isqrtd = sqrt(1.0 / d);

        for (int i = 0; i < 3; i++)
        {
            int j = (i + 1) % 3;
            int k = 3 - i - j;
            Vector2 m = (local[k] - local[j]) * isqrtd;

            if (index[i] < NP) // free vertices
            {
                int r = fi;
                int c = index[i];
                sp_set_val(A, r, c, m[0]);
                sp_set_val(A, r, c + NP, -m[1]);
                sp_set_val(A, r + F, c, m[1]);
                sp_set_val(A, r + F, c + NP, m[0]);
            }
            else
            {
                int r = fi;
                int c = index[i] - NP;
                m_set_val(fixed, r, c, -m[0]);
                m_set_val(fixed, r, c + P, m[1]);
                m_set_val(fixed, r + F, c, -m[1]);
                m_set_val(fixed, r + F, c + P, -m[0]);
            }
        }
    }

    for (int i = 0; i < P; i++)
    {
        v_set_val(ufixed, i, pinned[i]->data.texcoords[0]);
        v_set_val(ufixed, i + P, pinned[i]->data.texcoords[1]);
    }
    mv_mlt(fixed, ufixed, b);

    // Solve Ax = b
    int steps;
    int limit = 3 * (NP + F) + 10;
    VEC *x = solve_least_squares(A, b, 1e-12, NULL, limit, &steps);
    if (steps < 0)
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG,
              "solved with QR");
    else
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG,
              "%d step(s) of LSQR (limit was %d)",
              steps, limit);

    // Write back results
    for (int i = 0; i < NP; i++)
    {
        MeshVertexIterator v = inverse_id[i];
        v->data.texcoords[0] = v_entry(x, i);
        v->data.texcoords[1] = v_entry(x, i + NP);
    }

    V_FREE(x);
    V_FREE(b);
    V_FREE(ufixed);
    M_FREE(fixed);
    SP_FREE(A);
}

//! Splits a patch edge of [\a first, \a last)
/*! If \a force3 is true, then this only occurs for patches with fewer
 * than three existing corners. Returns true if a corner was added.
 * The split is chosen to maximise the object space perimeter of the
 * polygon with corners at the UV corners.
 */
template<class ForwardIterator>
bool add_extra_corner(ForwardIterator first,
                      ForwardIterator last,
                      bool force3 = false)
{
    MeshVertexIterator start, v, corner, corner2;
    int segment = 0;
    MeshVertexIterator best;
    double best_score = -HUGE_VAL;

    // Find a first corner
    v = mesh_null_vertex;
    for (ForwardIterator i = first; i != last; i++)
    {
        MeshHalfEdgeIterator e = (*i)->edges();
        for (int j = 0; j < 3; j++, e = e->next())
            if (e->vertex()->position()->patch_corner)
            {
                v = e->vertex();
                break;
            }
    }
    g_assert(v != mesh_null_vertex);
    start = corner = v;

    do
    {
        if (force3 && segment == 2) return false; // 3 corners, nothing to do
        // Find the next corner
        do
        {
            v = v->first_edge()->prev()->vertex();
        } while (!v->position()->patch_corner);

        // Search for optimal split point
        corner2 = v;
        v = corner;
        const VectorPosition &c1 = corner->position()->position;
        const VectorPosition &c2 = corner2->position()->position;
        double base = length(c2 - c1);
        do
        {
            v = v->first_edge()->prev()->vertex();
            if (v != corner2)
            {
                double score =
                    length(c1 - v->position()->position)
                    + length(c2 - v->position()->position)
                    - base;
                if (score > best_score)
                {
                    best_score = score;
                    best = v;
                }
            }
        } while (v != corner2);
        segment++;
        corner = corner2;
    } while (corner != start);
    if (best_score == -HUGE_VAL) return false;
    else
    {
        g_assert(!best->position()->patch_corner);
        best->position()->patch_corner = true;
        return true;
    }
}

template<class ForwardIterator>
void force_three_corners(ForwardIterator first,
                         ForwardIterator last)
{
    int segment = 0;
    MeshVertexIterator corners[2]; // first two found
    double segment_len[2] = {0.0, 0.0};
    MeshVertexIterator v, fallback = mesh_null_vertex;

    v = mesh_null_vertex;
    for (ForwardIterator i = first; i != last; i++)
    {
        MeshHalfEdgeIterator e = (*i)->edges();
        for (int j = 0; j < 3; j++, e = e->next())
            if (e->vertex()->position()->patch_corner)
            {
                v = e->vertex();
                break;
            }
            else if (mesh_position_patch_valence(e->vertex()->position()) >= 2)
                fallback = e->vertex();
    }
    /* Things go somewhat wrong when we have holes in the mesh. The holes
     * are triangulated and generally register as lots of mini-charts, but
     * once we have finished segmentation it is treated as a single patch.
     * The end result is that we can have a region with 0 corners.
     */
    if (v == mesh_null_vertex)
    {
        g_assert(fallback != mesh_null_vertex);
        v = fallback;
        v->position()->patch_corner = true;
    }
    g_assert(v != mesh_null_vertex);

    do
    {
        if (segment == 2) return; // 3 corners, nothing to do
        corners[segment] = v;
        do
        {
            MeshVertexIterator nxt = v->first_edge()->prev()->vertex();
            segment_len[segment] += length(v->position()->position - nxt->position()->position);
            v = nxt;
        } while (!v->position()->patch_corner);
        segment++;
    } while (v != corners[0]);

    // find the half-way vertex
    int base = (segment_len[0] > segment_len[1]) ? 0 : 1;
    double target = segment_len[base] / 2;
    v = corners[base];
    while (v != corners[1 - base])
    {
        MeshVertexIterator nxt = v->first_edge()->prev()->vertex();
        double len = length(v->position()->position - nxt->position()->position);
        if (len > target)
        {
            MeshVertexIterator n;
            if (nxt->position()->patch_corner) n = v;
            else if (v->position()->patch_corner) n = nxt;
            else n = (len > 2 * target) ? v : nxt;
            g_assert(!n->position()->patch_corner);
            n->position()->patch_corner = true;
            return;
        }
        target -= len;
        v = nxt;
    }
    g_assert_not_reached();
}


/* debug code */
#if 0
static void dump_faces(const vector<MeshFaceIterator> &faces)
{
    bool first = true;

    ofstream f("param.gplot");
    f << "set parametric\n";
    f << "set nokey\n";
    f << "set nolabel\n";

    for (size_t i = 0; i < faces.size(); i++)
    {
        Vector2 centroid(0.0, 0.0);
        MeshHalfEdgeIterator e = faces[i]->edges;
        f << "# triangle ";
        for (int j = 0; j < 3; j++, e = e->next)
        {
            f << e->vertex->data.texcoords << " ";
            centroid += e->vertex->data.texcoords;
        }
        centroid /= 3.0;
        f << "\nset label \"face" << i;
        if (mesh_texture_double_area(faces[i]) > 0.0)
            f << "(ccw)\" at ";
        else
            f << "(cw)\" at ";
        f << centroid[0] << "," << centroid[1] << "\n";
    }

    f << "plot [t=0:1]";

    for (size_t i = 0; i < faces.size(); i++)
    {
        MeshHalfEdgeIterator e = faces[i]->edges;
        for (int j = 0; j < 3; j++, e = e->next)
        {
            const Vector2 &a = e->vertex->data.texcoords;
            const Vector2 &b = e->next->vertex->data.texcoords;
            f << (first ? ' ' : ',');
            first = false;
            f   << a[0] << "+" << b[0] - a[0] << "*t,"
                << a[1] << "+" << b[1] - a[1] << "*t lt 1";
        }
    }
    f << "\n";
}
#endif

//! A structure that computes local stretch values with precomputed data
/*! To understand how this class works, refer to the internals of
 * mesh_texture_triangle_stretch. Instead of using the \c J vectors in
 * animation space, we consider them on the basis pos[j].
 */
class MeshLocalStretch
{
private:
    struct Triangle
    {
        double dot_pos[3][3];
        Vector2 uv[3];
        double spatial_area2;   //! Double-area in animation space, not parametric space

        explicit Triangle(MeshHalfEdgeConstIterator e);
    };

    vector<Triangle> triangles;
public:
    //! Precomputes data for the neighbourhood of v
    explicit MeshLocalStretch(MeshVertexConstIterator v);

    //! Returns a measure of stretch when placing the vertex at uv
    /*! \note the actual value is meaningless except for ordering */
    double operator()(const Vector2 &uv, bool infinity);
};

MeshLocalStretch::Triangle::Triangle(MeshHalfEdgeConstIterator e)
{
    MeshVertexConstIterator v[3];

    v[0] = e->next()->vertex();
    v[1] = e->prev()->vertex();
    v[2] = e->vertex();
    for (int i = 0; i < 3; i++)
    {
        for (int j = i; j < 3; j++)
        {
            dot_pos[i][j] = v[i]->position()->position * v[j]->position()->position;
            if (i != j) dot_pos[j][i] = dot_pos[i][j];
        }
        uv[i] = v[i]->data.texcoords;
    }
    spatial_area2 = mesh_face_double_area(e->face());
}

MeshLocalStretch::MeshLocalStretch(MeshVertexConstIterator v)
{
    MeshHalfEdgeConstIterator e = v->first_edge();
    do
    {
        triangles.push_back(Triangle(e));
        e = e->dual()->next();
    } while (e != v->first_edge());
}

double MeshLocalStretch::operator()(const Vector2 &uv, bool infinity)
{
    double ans = 0.0;
    for (vector<Triangle>::iterator t = triangles.begin(); t != triangles.end(); t++)
    {
        Vector2 j[3];  // coefficients of J on the basis pos[0], pos[1], pos[2]

        t->uv[0] = uv;
        j[0] = t->uv[1] - t->uv[2];
        j[1] = t->uv[2] - t->uv[0];
        j[2] = t->uv[0] - t->uv[1];
        /* mesh_texture_double_area */
        double A2 = (t->uv[1][0] - t->uv[0][0]) * (t->uv[2][1] - t->uv[0][1])
            - (t->uv[2][0] - t->uv[0][0]) * (t->uv[1][1] - t->uv[0][1]);
        if (A2 <= 0.0) return HUGE_VAL;
        double iA22 = 1.0 / (A2 * A2); // inverse square of double-area

        double a = 0.0, b = 0.0, c = 0.0;
        for (int p = 0; p < 3; p++)
            for (int q = 0; q < 3; q++)
            {
                a += t->dot_pos[p][q] * j[p][0] * j[q][0];
                b += t->dot_pos[p][q] * j[p][0] * j[q][1];
                c += t->dot_pos[p][q] * j[p][1] * j[q][1];
            }
        a *= iA22;
        b *= -iA22;
        c *= iA22;
        if (infinity)
        {
            double d = sqrt((a - c) * (a - c) + 4.0 * b * b);
            ans = max(ans, a + c + d);
        }
        else
            ans += t->spatial_area2 * (a + c);
    }
    return ans;
}

template<class ForwardIterator>
static void mesh_parametrise_optimise(ForwardIterator first,
                                      ForwardIterator last)
{
    vector<MeshVertexIterator> interior;
    id_set<MeshVertexIterator>::type seen;
    const bool infinity = false; // whether to use Linf or L2

    // first identify interior and boundary vertices
    for (ForwardIterator fi = first; fi != last; fi++)
    {
        MeshHalfEdgeIterator e = (*fi)->edges();
        for (int i = 0; i < 3; i++, e = e->next())
        {
            MeshVertexIterator v = e->vertex();
            if (seen.insert(v).second) // true if new element
            {
                if (v->first_edge()->dual()->face()->patch == (*fi)->patch)
                    interior.push_back(v);
            }
        }
    }

    for (int iter = 0; iter < 20; iter++)
    {
        /* Optimise interior vertices */
        for (size_t i = 0; i < interior.size(); i++)
        {
            MeshVertexIterator v = interior[i];
            Vector2 uv = v->data.texcoords;
            Vector2 dir;
            double kernel[2];  // min and max shift to avoid foldover
            double ang;
            MeshLocalStretch stretch(v);

            ang = rand() / (RAND_MAX + 1.0) * G_PI * 2.0;
            dir[0] = cos(ang);
            dir[1] = sin(ang);
            kernel[0] = -HUGE_VAL;
            kernel[1] = HUGE_VAL;
            MeshHalfEdgeIterator e = v->first_edge();
            do
            {
                Vector2 edge[2];
                Vector2 normal;
                edge[0] = e->prev()->vertex()->data.texcoords;
                edge[1] = e->vertex()->data.texcoords;
                normal = edge[1] - edge[0];
                normal = Vector2(-normal[1], normal[0]);

                /* We want to solve (uv + s * dir).normal >= edge[0].normal), which
                 * is the same as s(dir.normal) >= (edge[0] - uv).normal
                 * We don't actually normalise the normal, since we're only interested
                 * in direction.
                 */
                double num = (edge[0] - uv) * normal;
                double den = dir * normal;
                if (den > 0.0)
                    kernel[0] = max(kernel[0], num / den);
                else if (den < 0.0)
                    kernel[1] = min(kernel[1], num / den);

                e = e->dual()->next();
            } while (e != v->first_edge());

            double prev = stretch(uv, infinity);
            if (finite(kernel[0]) && finite(kernel[1]) && kernel[0] < kernel[1])
            {
                double l = kernel[0];
                double r = kernel[1];
                /* Ternary search optimisation */
                for (int k = 0; k < 30; k++)
                {
                    double p = (2 * l + r) / 3.0;
                    double q = (l + 2 * r) / 3.0;
                    double sp = stretch(uv + p * dir, infinity);
                    double sq = stretch(uv + q * dir, infinity);
                    if (!finite(sp)) l = p;
                    else if (!finite(sq)) r = q;
                    else if (sq < sp) l = p;
                    else r = q;
                }
                Vector2 target = uv + (l + r) * 0.5 * dir;
                double cur = stretch(target, infinity);
                if (cur < prev)
                    v->data.texcoords = target;
            }
        }
    }
}

template<class ForwardIterator>
static double mesh_parametrise_l2stretch(ForwardIterator first,
                                         ForwardIterator last)
{
    double texture_area = 0.0;
    double surface_area = 0.0;
    double A = 0.0;
    double l2sum = 0.0;
    for (ForwardIterator fi = first; fi != last; fi++)
    {
        A = 0.5 * mesh_face_double_area(*fi);
        surface_area += A;
        texture_area += 0.5 * mesh_texture_double_area(*fi);
        double l = mesh_texture_triangle_stretch(*fi, false);
        l2sum += A * l * l;
    }
    if (!finite(l2sum)) return l2sum;
    else return sqrt(l2sum * texture_area) / surface_area;
}

template<class ForwardIterator>
static double mesh_parametrise_l2scale(ForwardIterator first,
                                       ForwardIterator last)
{
    double texture_area = 0.0;
    double A = 0.0;
    double l2sum = 0.0;
    for (ForwardIterator fi = first; fi != last; fi++)
    {
        A = 0.5 * mesh_face_double_area(*fi);
        texture_area += 0.5 * mesh_texture_double_area(*fi);
        double l = mesh_texture_triangle_stretch(*fi, false);
        l2sum += A * l * l;
    }
    if (!finite(l2sum)) return l2sum;
    else return sqrt(sqrt(l2sum / texture_area));
}

template<class ForwardIterator>
static void mesh_parametrise_rescale(ForwardIterator first,
                                     ForwardIterator last)
{
    // Rescale the coordinates
    double uv_ratio = mesh_parametrise_l2scale(first, last);
    g_assert(finite(uv_ratio));
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "scaling by %f", uv_ratio);
    for (ForwardIterator fi = first; fi != last; fi++)
    {
        MeshHalfEdgeIterator e = (*fi)->edges();
        for (int i = 0; i < 3; i++, e = e->next())
        {
            MeshVertexIterator v = e->next()->vertex();
            if (v->first_edge() == e)
                v->data.texcoords *= uv_ratio;
        }
    }
}

static inline bool is_patch(MeshFaceConstIterator a, int p)
{
    return a->patch == p;
}

//! Flattens patches, and splits those with self-intersecting boundaries
/*! The splitting process works by doing a 2-source BFS from the two
 * intersection edges.
 * \param m The mesh on which to operate
 * \param tolerance The ratio by which straightening may increase stretch
 * \param first, last Iterators to MeshFaceIterators bounding the range
 * \param patch_pool The next available patch id (updated)
 * \param no_split If true, no splitting is done
 * \param split_points An iterator to the ForwardIterator type. It is
 * filled in order with the splits at which to split [first, last) to
 * obtain the new face ranges (the iterators are permuted).
 * \returns the updated value of \a split_points.
 */
template<class ForwardIterator, class OutputIterator>
static OutputIterator mesh_parametrise_split(Mesh &m,
                                             double tolerance,
                                             ForwardIterator first,
                                             ForwardIterator last,
                                             int &patch_pool,
                                             bool no_split,
                                             OutputIterator split_points)
{
    MeshHalfEdgeIterator cuts[2];
    int ids[2];
    double baseline;
    bool repeat;

    g_assert(first != last);
    mesh_parametrise_flatten(first, last, 0);
    mesh_parametrise_optimise(first, last);
    baseline = mesh_parametrise_l2stretch(first, last);

    if (!finite(baseline))
    {
        g_warning("persistent foldover, trying convex baseline");
        mesh_parametrise_flatten(first, last, 2);
        mesh_parametrise_optimise(first, last);
        baseline = mesh_parametrise_l2stretch(first, last);
        if (!finite(baseline))
            g_error("foldover could not be avoided");
    }

    if (tolerance > 0.0)
        do
        {
            repeat = false;
            mesh_parametrise_flatten(first, last, 1);
            mesh_parametrise_optimise(first, last);
            if (mesh_parametrise_l2stretch(first, last) > tolerance * baseline)
            {
                g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "straightening failed, falling back to convex (L2 = %g base = %g)",
                      mesh_parametrise_l2stretch(first, last), baseline);
                mesh_parametrise_flatten(first, last, 2);
                mesh_parametrise_optimise(first, last);
                if (mesh_parametrise_l2stretch(first, last) > tolerance * baseline)
                {
                    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "convex failed, adding a corner (L2 = %g base = %g)",
                          mesh_parametrise_l2stretch(first, last), baseline);
                    if (!add_extra_corner(first, last))
                    {
                        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "no more corners to add; using baseline");
                        g_assert(mesh_patch_straight_boundary(first, last));
                        mesh_parametrise_flatten(first, last, 0);
                        mesh_parametrise_optimise(first, last);
                        if (!finite(mesh_parametrise_l2stretch(first, last)))
                        {
                            mesh_parametrise_flatten(first, last, 2);
                            mesh_parametrise_optimise(first, last);
                        }
                    }
                    else
                        mesh_parametrise_flatten(first, last, 0); // re-establish coordinates
                    repeat = true;
                }
            }
        } while(repeat);

    if (no_split || mesh_patch_legal_boundary(first, last, cuts[0], cuts[1]))
        mesh_parametrise_rescale(first, last);
    else
    {
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "Splitting a patch");

        queue<MeshFaceIterator> faceq;
        ForwardIterator mid;

        faceq.push(cuts[0]->face());
        faceq.push(cuts[1]->face());
        ids[0] = (*first)->patch;
        ids[1] = patch_pool++;
        for (ForwardIterator fi = first; fi != last; fi++)
            (*fi)->patch = -1; /* indicates unset */
        cuts[0]->face()->patch = ids[0];
        cuts[1]->face()->patch = ids[1];
        while (!faceq.empty())
        {
            MeshFaceIterator cur = faceq.front();
            MeshHalfEdgeIterator e = cur->edges();
            faceq.pop();
            for (int i = 0; i < 3; i++, e = e->next())
            {
                MeshFaceIterator nxt = e->dual()->face();
                if (nxt->patch == -1)
                {
                    nxt->patch = cur->patch;
                    faceq.push(nxt);
                }
            }
        }
        mid = partition(first, last, bind2nd(ptr_fun(is_patch), ids[0]));
        g_assert(mid != first && mid != last);
        // split boundary vertices (only needed from one side)
        mesh_split_vertices(m, first, mid);

        force_three_corners(first, mid);
        force_three_corners(mid, last);
        split_points = mesh_parametrise_split(m, tolerance, first, mid,
                                              patch_pool, no_split, split_points);
        *split_points++ = mid;
        split_points = mesh_parametrise_split(m, tolerance, mid, last,
                                              patch_pool, no_split, split_points);
    }
    return split_points;
}

struct PackInfo
{
    Vector2 bsize;         //!< bounding rect size; based at (0, 0)
    Vector2 shift;         //!< shift to pack in, before scaling
    vector<MeshVertexIterator> vertices;
};

static inline bool compare_pack_info(const PackInfo &a, const PackInfo &b)
{
    return a.bsize[1] > b.bsize[1];
}

//! Packs charts into a texture atlas
static ParametrisationStatistics mesh_pack_patches(const list<vector<MeshFaceIterator> > &faces)
{
    size_t size = faces.size();
    PackInfo patches[size];
    double widthl = 0.0, widthr = 0.0, width;
    double rect_area = 0.0;
    double gutter_area = 0.0;

    /* Estimate the final size, so that we can add an appropriate border. */
    double tex_area = 0.0;
    double border_size;
    for (list<vector<MeshFaceIterator> >::const_iterator fl = faces.begin();
         fl != faces.end(); fl++)
        for (vector<MeshFaceIterator>::const_iterator fi = fl->begin();
             fi != fl->end(); fi++)
            tex_area += mesh_texture_double_area(*fi);
    tex_area *= 0.5;
    border_size = sqrt(tex_area) / 128.0;

    /* Identify vertices */
    list<vector<MeshFaceIterator> >::const_iterator fl = faces.begin();
    for (size_t i = 0; i < size; i++, fl++)
    {
        for (size_t j = 0; j < fl->size(); j++)
        {
            MeshHalfEdgeIterator e = (*fl)[j]->edges();
            for (int k = 0; k < 3; k++, e = e->next())
                if (e->vertex()->first_edge()->next() == e)
                    patches[i].vertices.push_back(e->vertex());
        }
    }

    /* Rotate charts into minimum area rectangles, and find bounding boxes */
    for (size_t i = 0; i < size; i++)
    {
        double min_area = HUGE_VAL;
        double besta = 0.0;
        Vector2 bmin;
        for (int r = 0; r < 16; r++)
        {
            Vector2 cmin(HUGE_VAL, HUGE_VAL);
            Vector2 cmax(-HUGE_VAL, -HUGE_VAL);
            double a = G_PI * r / 32.0;
            double c = cos(a);
            double s = sin(a);
            for (size_t j = 0; j < patches[i].vertices.size(); j++)
            {
                Vector2 uv0 = patches[i].vertices[j]->data.texcoords;
                Vector2 uv(c * uv0[0] - s * uv0[1],
                           s * uv0[0] + c * uv0[1]);
                cmin[0] = min(cmin[0], uv[0]);
                cmin[1] = min(cmin[1], uv[1]);
                cmax[0] = max(cmax[0], uv[0]);
                cmax[1] = max(cmax[1], uv[1]);
            }
            double area = (cmax[0] - cmin[0]) * (cmax[1] - cmin[1]);
            if (area < min_area)
            {
                min_area = area;
                if (cmax[0] - cmin[0] > cmax[1] - cmin[1])
                {
                    besta = a + 0.5 * G_PI;
                    bmin[0] = -cmax[1];
                    bmin[1] = cmin[0];
                    patches[i].bsize[0] = cmax[1] - cmin[1];
                    patches[i].bsize[1] = cmax[0] - cmin[0];
                }
                else
                {
                    besta = a;
                    bmin = cmin;
                    patches[i].bsize = cmax - cmin;
                }
            }
        }

        /* Add some extra space for sampling artefacts. */
        bmin -= Vector2(border_size, border_size);
        rect_area += patches[i].bsize[0] * patches[i].bsize[1];
        patches[i].bsize += Vector2(2 * border_size, 2 * border_size);
        gutter_area += patches[i].bsize[0] * patches[i].bsize[1];
        /* Apply rotation and translation */
        double c = cos(besta);
        double s = sin(besta);
        for (size_t j = 0; j < patches[i].vertices.size(); j++)
        {
            Vector2 &uv = patches[i].vertices[j]->data.texcoords;
            Vector2 old = uv;
            uv.assign(c * old[0] - s * old[1] - bmin[0],
                      s * old[0] + c * old[1] - bmin[1]);
        }

        /* Determine useful bounds for the binary search */
        widthr += patches[i].bsize[0];
        widthl = max(widthl, patches[i].bsize[0]);
    }

    /* Sort charts by size */
    sort(patches, patches + size, compare_pack_info);

    double x, maxh;
    for (int c = 0; c < 10; c++)
    {
        width = (widthl + widthr) * 0.5;
        x = maxh = 0.0;
        bool l2r = true; // direction
        // place patches
        for (size_t i = 0; i < size; i++)
        {
            if (l2r)
            {
                if (x + patches[i].bsize[0] > width)
                {
                    x = width;
                    l2r = false;
                    i--; // cancel i++
                    continue;
                }
                patches[i].shift = Vector2(x, 0);
                double right = x + patches[i].bsize[0];
                for (size_t j = 0; j < i; j++)
                {
                    Vector2 jmin = patches[j].shift;
                    Vector2 jmax = jmin + patches[j].bsize;
                    if (jmin[0] < right - 1e-9 && jmax[0] > x + 1e-9)
                        patches[i].shift[1] = max(patches[i].shift[1], jmax[1]);
                }
                x = right;
                maxh = max(maxh, patches[i].shift[1] + patches[i].bsize[1]);
            }
            else
            {
                if (x - patches[i].bsize[0] < 0.0)
                {
                    x = 0.0;
                    l2r = true;
                    i--;
                    continue;
                }
                double left = x - patches[i].bsize[0];
                patches[i].shift = Vector2(left, 0);
                for (size_t j = 0; j < i; j++)
                {
                    Vector2 jmin = patches[j].shift;
                    Vector2 jmax = jmin + patches[j].bsize;
                    if (jmin[0] < x - 1e-9 && jmax[0] > left + 1e-9)
                        patches[i].shift[1] = max(patches[i].shift[1], jmax[1]);
                }
                x = left;
                maxh = max(maxh, patches[i].shift[1] + patches[i].bsize[1]);
            }
        }
        if (maxh > width) widthl = width;
        else widthr = width;
    }

    double scale = 1.0 / max(width, maxh);
    rect_area *= scale * scale;
    tex_area *= scale * scale;
    gutter_area *= scale * scale;

    for (size_t i = 0; i < size; i++)
    {
        for (size_t j = 0; j < patches[i].vertices.size(); j++)
        {
            Vector2 &uv = patches[i].vertices[j]->data.texcoords;
            uv = (uv + patches[i].shift) * scale;
        }
    }

    ParametrisationStatistics stats;
    stats.texture_area = tex_area;
    stats.rectangle_area = rect_area;
    stats.gutter_area = gutter_area;
    return stats;
}

/*! The parametrisation proceeds in three steps
 * - creating patches that are topological discs
 * - flatten patches into charts
 * - pack charts into an atlas
 *
 * \param m The mesh to parametrise. All existing texture information is
 * removed, and replaced with the new information in place.
 * \param patches The minimum number of patches to create (actual number
 * may be slightly higher). Set to 0 to use the original patches.
 * \param limit An alternative stopping mechanism for segmentation.
 * This limits the segmentation metric (the lower bound for the metric
 * is 1.0 + compact_weight).
 * \param compact_weight The weight to give the compacting term. Higher
 * values favour more circular patches. A value of 1.0 gives equal weight
 * to compactness and planarity.
 * \param straighten_tolerance Specifies a ratio by which the segmentation
 * metric may be increased by straightening chart edges in the mesh.
 * \param constrain_tolerance Specifies a ratio by which the L2 metric may
 * be increased by straightening chart edges in parameter space.
 * \param no_split If true, patches with self-intersecting boundaries will
 * not be split.
 *
 * Internally, patches are represented as vectors of \c MeshFaceIterators.
 */
ParametrisationStatistics mesh_parametrise(Mesh &m, int patches,
                                           double limit, double compact_weight,
                                           double straighten_tolerance,
                                           double constrain_tolerance,
                                           bool no_split)
{
    Timer parametrise_timer;
    /* We don't use a vector on the outside, as patch splitting will cause
     * the whole thing to be copied. That messes with the iterators.
     */
    list<vector<MeshFaceIterator> > faces;

    // remove LOD information, since it will be invalidated
    m.clear_lod();

    // create patches
    if (patches != 0)
    {
        mesh_remove_patches(m);
        patches = mesh_create_patches(m, patches, limit, compact_weight,
                                      constrain_tolerance);
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Created %d patches", patches);
    }
    else
    {
        mesh_complete(m, MESH_LEVEL_PATCHES);
        for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
            if (f->patch > patches) patches = f->patch;
        mesh_check_manifold(m, MESH_LEVEL_PATCHES);
    }

    faces.resize(patches);
    list<vector<MeshFaceIterator> >::iterator face_ptr[patches]; // used only for extraction
    face_ptr[0] = faces.begin();
    for (int i = 1; i < patches; i++)
    {
        face_ptr[i] = face_ptr[i - 1];
        face_ptr[i]++;
    }
    for (MeshFaceIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
    {
        g_assert(f->patch > 0);
        face_ptr[f->patch - 1]->push_back(f);
    }

    /* Determine the corners in texture space. Patch boundaries get to be
     * corners for free, but if we have any patches with only two neighbours
     * then we have to force in an extra corner.
     */
    list<vector<MeshFaceIterator> >::iterator fl;
    for (MeshPositionIterator p = m.positions.whole_begin(); p != m.positions.whole_end(); p++)
        p->patch_corner = (mesh_position_patch_valence(p) >= 3);
    for (fl = faces.begin(); fl != faces.end(); fl++)
        force_three_corners(fl->begin(), fl->end());

    // flatten patches
    Timer flatten_timer;
    int patch_pool = patches + 1;
    int logcount = 0;
    for (fl = faces.begin(); fl != faces.end(); fl++)
    {
        vector<vector<MeshFaceIterator>::iterator> splits;
        vector<vector<MeshFaceIterator>::iterator>::iterator j, k;
        vector<MeshFaceIterator>::iterator l;
        mesh_parametrise_split(m, straighten_tolerance,
                               fl->begin(), fl->end(), patch_pool, no_split,
                               back_inserter(splits));
        for (j = splits.begin(); j != splits.end(); j++)
        {
            k = j;
            k++;
            if (k == splits.end()) l = fl->end(); else l = *k;
            /* push in front, so that we don't revisit it in the loop */
            faces.push_front(vector<MeshFaceIterator>(*j, l));
        }
        if (!splits.empty())
            fl->erase(splits.front(), fl->end());
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Flattened patch %d (L2 = %g)",
              ++logcount, mesh_parametrise_l2stretch(fl->begin(), fl->end()));

    }
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Flattened patches in %.3fs", flatten_timer.elapsed());

    // pack patches
    ParametrisationStatistics stats = mesh_pack_patches(faces);
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Packed %ld patches", (long) faces.size());

    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Total parametrisation time: %.3fs", parametrise_timer.elapsed());
    return stats;
}

double mesh_texture_triangle_stretch(MeshFaceConstIterator f, bool infinity)
{
    VectorPosition pos[3];
    Vector2 uv[3];
    VectorPosition J[2]; // columns of Jacobian, from Sander et al
    double A2;

    MeshHalfEdgeConstIterator e = f->edges();
    for (int i = 0; i < 3; i++, e = e->next())
    {
        pos[i] = e->vertex()->position()->position;
        uv[i] = e->vertex()->data.texcoords;
    }

    // Same as mesh_texture_double_area, but duplicated to avoid a second
    // walk around the face
    A2 = (uv[1][0] - uv[0][0]) * (uv[2][1] - uv[0][1])
        - (uv[2][0] - uv[0][0]) * (uv[1][1] - uv[0][1]);
    if (A2 <= 0.0) return HUGE_VAL; // flipped/degenerate triangle
    for (int i = 0; i < 2; i++)
    {
        J[i] = pos[0] * (uv[1][i] - uv[2][i])
            +  pos[1] * (uv[2][i] - uv[0][i])
            +  pos[2] * (uv[0][i] - uv[1][i]);
        J[i] /= A2;
    }

    if (!infinity)
        return sqrt((J[0] * J[0] + J[1] * J[1]) * 0.5);

    double a = J[0] * J[0];
    double b = J[0] * J[1];
    double c = J[1] * J[1];
    double d = sqrt((a - c) * (a - c) + 4.0 * b * b);
    return sqrt(0.5 * (a + c + d));
}

double mesh_texture_stretch(const Mesh &m, bool infinity)
{
    double surface_area = 0.0;
    double texture_area = 0.0;
    double stretch = 0.0;

    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
    {
        double A2s = mesh_face_double_area(f);
        double A2t = mesh_texture_double_area(f);
        surface_area += A2s;
        texture_area += A2t;
        double l = mesh_texture_triangle_stretch(f, infinity);
        if (!infinity)
            stretch += l * l * A2s;
        else
            stretch = max(stretch, l);
    }
    g_assert(surface_area > 0.0);
    g_assert(texture_area > 0.0);
    if (!infinity)
        stretch = sqrt(stretch / surface_area);
    return stretch * sqrt(texture_area / surface_area);
}

/*! Quantise texture coordinates to a more or less 2^23 x 2^23 grid
 * so that we can perform exact computations with doubles.
 */
static Vector2 quantise(const Vector2 &v)
{
    Vector2 ans;
    static const double scale = 1 << 23;

    ans[0] = ((int) (v[0] * scale)) / scale;
    ans[1] = ((int) (v[1] * scale)) / scale;
    return ans;
}

static inline void get_texcoords(MeshFaceConstIterator f, Vector2 *coords)
{
    MeshHalfEdgeConstIterator e = f->edges();
    for (int i = 0; i < 3; i++, e = e->next())
        coords[i] = e->vertex()->data.texcoords;
}

/*! To be legal, a parametrisation must
 * - have no flipped or degenerate triangles
 * - have no texture coordinates outside of the unit square
 * - have no intersecting triangles
 */
bool mesh_valid_parametrisation(const Mesh &mesh)
{
    Mesh &m = const_cast<Mesh &>(mesh);

    bool answer = true;
    static const int GRIDSIZE = 32;
    vector<MeshFaceIterator> faces[GRIDSIZE][GRIDSIZE];

    Vector2 old[m.vertices.real_size()];
    // save original values, then quantise
    size_t index = 0;
    for (MeshVertexIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
    {
        old[index++] = v->data.texcoords;
        v->data.texcoords = quantise(v->data.texcoords);
    }

    // check for flipped and degenerate triangles
    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
    {
        if (mesh_texture_double_area(f) <= 0.0)
        {
            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "illegal parametrisation: flipped triangle");
            answer = false;
            goto cleanup;
        }
    }

    // check for out-of-bounds texcoords
    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
        if (!finite(v->data.texcoords[0])
            || !finite(v->data.texcoords[1])
            || v->data.texcoords[0] < 0.0
            || v->data.texcoords[1] < 0.0
            || v->data.texcoords[0] > 1.0
            || v->data.texcoords[1] > 1.0)
        {
            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "illegal parametrisation: outside of unit square");
            answer = false;
            goto cleanup;
        }

    /* Rather than going to the effort of implementing a k-d tree or using
     * an O(n.log n) all intersections algorithm, we use a grid approach.
     */

    // bucketing
    for (MeshFaceIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
    {
        Vector2 coords[3];
        Vector2 tmin, tmax;
        int minx, maxx, miny, maxy;

        get_texcoords(f, coords);
        get_bounding_box(coords, 3, tmin, tmax);

        minx = (int) floor(tmin[0] * GRIDSIZE);
        maxx = (int) ceil(tmax[0] * GRIDSIZE);
        miny = (int) floor(tmin[1] * GRIDSIZE);
        maxy = (int) ceil(tmax[1] * GRIDSIZE);
        for (int x = minx; x < maxx; x++)
            for (int y = miny; y < maxy; y++)
                faces[x][y].push_back(f);
    }

    // intersection testing
    for (int x = 0; x < GRIDSIZE; x++)
        for (int y = 0; y < GRIDSIZE; y++)
        {
            vector<MeshFaceIterator>::const_iterator fa, fb;
            for (fa = faces[x][y].begin(); fa != faces[x][y].end(); fa++)
                for (fb = fa + 1; fb != faces[x][y].end(); fb++)
                {
                    Vector2 ca[3], cb[3];
                    get_texcoords(*fa, ca);
                    get_texcoords(*fb, cb);
                    // point inside tests
                    for (int i = 0; i < 3; i++)
                    {
                        int j;
                        for (j = 0; j < 3; j++)
                            if (cross(ca[(j + 1) % 3] - ca[j], cb[i] - ca[j]) <= 0.0) break;
                        if (j == 3)
                        {
                            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "illegal parametrisation: overlapping triangles");
                            answer = false;
                            goto dumptris;
                        }
                        for (j = 0; j < 3; j++)
                            if (cross(cb[(j + 1) % 3] - cb[j], ca[i] - cb[j]) <= 0.0) break;
                        if (j == 3)
                        {
                            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "illegal parametrisation: overlapping triangles");
                            answer = false;
                            goto dumptris;
                        }
                    }
                    // edge intersection tests
                    for (int i = 0; i < 3; i++)
                        for (int j = 0; j < 3; j++)
                        {
                            Vector2 p1 = ca[i];
                            Vector2 p2 = ca[(i + 1) % 3];
                            Vector2 q1 = cb[j];
                            Vector2 q2 = cb[(j + 1) % 3];
                            if (intersects(p1, p2, q1, q2, true))
                            {
                                g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "illegal parametrisation: overlapping triangles");
                                answer = false;
                                goto dumptris;
                            }
                        }
                    // identical triangles
                    for (int i = 0; i < 3; i++)
                    {
                        int j;
                        for (j = 0; j < 3; j++)
                            if (ca[j][0] != cb[(i + j) % 3][0]
                                || ca[j][1] != cb[(i + j) % 3][1]) break;
                        if (j == 3)
                        {
                            g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "illegal parametrisation: identical triangles");
                            answer = false;
                            goto dumptris;
                        }
                    }
                    continue;
                dumptris:
                    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG,
                          "(%f %f) (%f %f) (%f %f)",
                          ca[0][0], ca[0][1], ca[1][0], ca[1][1], ca[2][0], ca[2][1]);
                    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG,
                          "(%f %f) (%f %f) (%f %f)",
                          cb[0][0], cb[0][1], cb[1][0], cb[1][1], cb[2][0], cb[2][1]);
                    goto cleanup;
                }
        }

cleanup:
    // reassign old coordinates
    index = 0;
    for (MeshVertexIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
        v->data.texcoords = old[index++];

    return answer;
}

/*!
 @}
 */
