/*! \addtogroup mesh
 * @{
 */

/*! \file
 * \brief %Mesh simplification using progressive meshes
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <vector>
#include <map>
#include <set>
#include <algorithm>
#include <utility>
#include <cstddef>
#include <glib.h>
#include "mesh/mesh.h"
#include "mesh/iheap.h"
#include "mesh/aps.h"
#include "common/mesch.h"
#include "common/timing.h"
#include "geom/skeleton.h"

using namespace std;

//! Resets the \c first_vertex and \c first_edge fields around \a p
/*! Since the fields are currently assumed to be invalid, an edge is
 * required to start the walk. The edge \a edge must have \a p on its
 * clockwise end.
 */
static void repair_first(MeshPositionIterator p, MeshHalfEdgeIterator edge)
{
    MeshHalfEdgeIterator e = edge;
    bool found = false;

    p->first_vertex() = mesh_null_vertex;
    do
    {
        g_assert(e->next()->vertex()->position() == p);
        if (e->dual()->prev()->vertex() != e->next()->vertex())
        {
            found = true;
            e->next()->vertex()->first_edge() = e;
            if (e->face()->patch == 0)
                p->first_vertex() = e->next()->vertex();
        }
        e = e->dual()->next();
    } while (e != edge);
    // the case of only one patch
    if (!found) e->next()->vertex()->first_edge() = e;
    if (p->first_vertex() == mesh_null_vertex)
        p->first_vertex() = e->next()->vertex();
}

static MeshHalfEdgeIterator find_edge(MeshPositionIterator p1, MeshPositionIterator p2,
                                      MeshHalfEdgeIterator hint) throw (bad_collapse)
{
    if (hint != mesh_null_half_edge
        && hint->next()->vertex()->position() == p1
        && hint->prev()->vertex()->position() == p2)
        return hint;

    MeshHalfEdgeIterator e = p1->first_vertex()->first_edge();
    do
    {
        if (e->prev()->vertex()->position() == p2) return e;
        e = e->dual()->next();
    } while (e != p1->first_vertex()->first_edge());
    throw bad_collapse("edge not found in find_edge");
}

//! Applies an influence collapse
/*! Set \a erase to \c true to erase all the old data; otherwise it is
 * moved to the lod section of the mesh.
 */
static void mesh_apply_collapse_influence(Mesh &m, MeshInfluenceCollapse &c,
                                          bool erase) throw (bad_collapse)
{
    id_map<MeshVertexIterator, MeshVertexIterator>::type replacements;

    g_assert(!c.empty());
    MeshPositionIterator old_pos = c.front().before()->position();
    MeshPositionIterator new_pos = c.front().after()->position();
    g_assert(old_pos != new_pos);

    m.positions.move_to_real(new_pos);
    // We need a default, because if this is on a boundary then there will
    // be no replacement for first_vertex
    new_pos->first_vertex() = old_pos->first_vertex();
    for (MeshInfluenceCollapse::iterator i = c.begin(); i != c.end(); i++)
    {
        replacements[i->before()] = i->after();
        i->after()->first_edge() = i->before()->first_edge();
        if (old_pos->first_vertex() == i->before())
            new_pos->first_vertex() = i->after();
    }

    MeshHalfEdgeIterator e0 = old_pos->first_vertex()->first_edge()->next();  // edge opposite the vertex
    MeshHalfEdgeIterator e = e0;
    do
    {
        if (mesh_true_vertex(e->vertex()))
        {
            g_assert(replacements.count(e->vertex()));
            e->vertex() = replacements[e->vertex()];
        }
        e->vertex()->position() = new_pos;
        e = e->next()->dual()->next();
    } while (e != e0);

    for (MeshInfluenceCollapse::iterator i = c.begin(); i != c.end(); i++)
    {
        g_assert(mesh_true_vertex(i->after()));
        m.vertices.move_to_real(i->after());

        if (erase)
            m.vertices.erase(i->before());
        else
            m.vertices.move_to_lod(i->before());
    }
    if (erase)
        m.positions.erase(old_pos);
    else
        m.positions.move_to_lod(old_pos);
}

//! Applies an edge collapse
/*! ehint is only a hint; if it is wrong then we search the neighbourhood
 * of p1 for p2. It should be the case that ehint->next->vertex->position = p1,
 * ehint->prev->vertex->position = p2.
 * Set \a erase to \c true to erase all the old data; otherwise it is
 * moved to the lod section of the mesh.
 */
static void mesh_apply_collapse_edge(Mesh &m, MeshEdgeCollapse &c,
                                     MeshHalfEdgeIterator ehint,
                                     bool erase) throw(bad_collapse)
{
    g_assert(c.vl()->position() == c.vr()->position());

    MeshHalfEdgeIterator edge, dual;
    edge = find_edge(c.p1(), c.p2(), ehint);
    dual = edge->dual();

    m.positions.move_to_real(c.vl()->position());
    m.vertices.move_to_real(c.vl());
    m.vertices.move_to_real(c.vr());

    /* The collapse is done in four stages:
     * 1. Cut the collapsed edges and face out of the mesh and delete
     * them. The topology is now correct if one never uses the vertices
     * and positions.
     * 2. Walk the new neighbourhood, and set the edge->vertex fields
     * to point to the new vertices where appropriate.
     * 3. Recompute the first_vertex and first_edge fields for the
     * new position and the positions opposite the old edge.
     * 4. Delete all the old stuff.
     */

    // 1
    edge->prev()->dual()->dual() = edge->next()->dual();
    edge->next()->dual()->dual() = edge->prev()->dual();
    dual->prev()->dual()->dual() = dual->next()->dual();
    dual->next()->dual()->dual() = dual->prev()->dual();

    // 2
    MeshHalfEdgeIterator e = edge->next()->dual()->prev();
    do
    {
        if (e->vertex() == edge->next()->vertex()
            || e->vertex() == edge->prev()->vertex())
            e->vertex() = c.vl();
        else if (e->vertex() == dual->next()->vertex()
                 || e->vertex() == dual->prev()->vertex())
            e->vertex() = c.vr();
        e = e->prev()->dual()->prev();
    } while (e != edge->next()->dual()->prev());

    // 3
    repair_first(c.vl()->position(), edge->prev()->dual());
    repair_first(edge->vertex()->position(), edge->next()->dual());
    repair_first(dual->vertex()->position(), dual->next()->dual());

    // 4
    m.faces.erase(edge->face());
    m.faces.erase(dual->face());

    void (MeshCollection<MeshVertex>::*erase_vertex)(MeshVertexIterator);
    void (MeshCollection<MeshPosition>::*erase_position)(MeshPositionIterator);
    if (erase)
    {
        erase_vertex = &MeshCollection<MeshVertex>::erase;
        erase_position = &MeshCollection<MeshPosition>::erase;
    }
    else
    {
        erase_vertex = &MeshCollection<MeshVertex>::move_to_lod;
        erase_position = &MeshCollection<MeshPosition>::move_to_lod;
    }

    if (c.p1() != c.vl()->position()) (m.positions.*erase_position)(c.p1());
    if (c.p2() != c.vl()->position()) (m.positions.*erase_position)(c.p2());

    if (edge->next()->vertex() != c.vl()
        && edge->next()->vertex() != dual->prev()->vertex())
        (m.vertices.*erase_vertex)(edge->next()->vertex());
    if (dual->prev()->vertex() != c.vr())
        (m.vertices.*erase_vertex)(dual->prev()->vertex());
    if (edge->prev()->vertex() != c.vl()
        && edge->prev()->vertex() != dual->next()->vertex())
        (m.vertices.*erase_vertex)(edge->prev()->vertex());
    if (dual->next()->vertex() != c.vr())
        (m.vertices.*erase_vertex)(dual->next()->vertex());

    m.edges.erase(edge->next());
    m.edges.erase(edge->prev());
    m.edges.erase(dual->next());
    m.edges.erase(dual->prev());
    m.edges.erase(edge);
    m.edges.erase(dual);
}

void mesh_apply_collapse(Mesh &m, MeshCollapse &c,
                         MeshHalfEdgeIterator ehint,
                         bool erase) throw(bad_collapse)
{
    if (c.is_edge_collapse)
        mesh_apply_collapse_edge(m, c.edge_collapse, ehint, erase);
    else
        mesh_apply_collapse_influence(m, c.influence_collapse, erase);
}

static void mesh_apply_split_vertex(Mesh &m, const MeshVertexSplit &s) throw(bad_collapse)
{
    MeshVertexIterator v[2] = {s.vl, s.vr};
    MeshHalfEdgeIterator e[2];

    for (int i = 0; i < 2; i++)
        e[i] = find_edge(s.p, v[i]->position(), mesh_null_half_edge);

    m.vertices.move_to_real(s.vl1);
    m.vertices.move_to_real(s.vr1);
    m.vertices.move_to_real(s.vl2);
    m.vertices.move_to_real(s.vr2);
    m.positions.move_to_real(s.vl1->position());
    m.positions.move_to_real(s.vl2->position());
    g_assert(s.vl1->position() == s.vr1->position());
    g_assert(s.vl2->position() == s.vr2->position());

    /* The steps here are as follows:
     * 1. Move the old stuff out to the LOD section
     * 2. Create new faces with edges, and insert them into the mesh
     * 3. Walk the two neighbourhoods to update the vertex() fields
     * 4. Repair first_vertex and first_edge.
     */

    // 1
    if (s.vl1->position() != s.p
        && s.vl2->position() != s.p)
    {
        m.positions.move_to_lod(s.p);
        MeshHalfEdgeIterator edge = s.p->first_vertex()->first_edge();
        do
        {
            m.vertices.move_to_lod(edge->next()->vertex());
            edge = edge->dual()->prev()->vertex()->first_edge();
        } while (edge != s.p->first_vertex()->first_edge());
    }

    // 2
    MeshFaceIterator new_faces[2];
    MeshHalfEdgeIterator new_edges[2][3];

    for (int i = 0; i < 2; i++)
    {
        for (int j = 0; j < 3; j++)
            new_edges[i][j] = m.edges.insert_real(MeshHalfEdge());
        for (int j = 0; j < 3; j++)
        {
            new_edges[i][j]->next() = new_edges[i][(j + 1) % 3];
            new_edges[i][j]->prev() = new_edges[i][(j + 2) % 3];
        }
        new_edges[i][1]->dual() = e[i]->dual();
        e[i]->dual()->dual() = new_edges[i][1];
        new_edges[i][2]->dual() = e[i];
        e[i]->dual() = new_edges[i][2];
        new_edges[i][0]->vertex() = v[i];

        MeshFaceIterator reff = v[i]->first_edge()->face();
        if (reff->patch)
            new_faces[i] = m.faces.insert_real(MeshFace());
        else
            new_faces[i] = m.faces.insert_hole(MeshFace());
        new_faces[i]->patch = reff->patch;
        new_faces[i]->material() = reff->material();
        new_faces[i]->edges() = new_edges[i][0];

        for (int j = 0; j < 3; j++)
            new_edges[i][j]->face() = new_faces[i];
    }
    new_edges[0][0]->dual() = new_edges[1][0];
    new_edges[1][0]->dual() = new_edges[0][0];

    // 3
    for (int i = 0; i < 2; i++)
    {
        MeshHalfEdgeIterator edge = e[i];
        do
        {
            if (edge->face()->patch == new_faces[0]->patch)
                edge->next()->vertex() = i ? s.vl2 : s.vl1;
            else if (edge->face()->patch == new_faces[1]->patch)
                edge->next()->vertex() = i ? s.vr2 : s.vr1;
            edge = edge->dual()->next();
        } while (edge != e[i]);
    }

    // 4
    repair_first(s.vl1->position(), e[0]);
    repair_first(s.vl2->position(), e[1]);
    repair_first(s.vl->position(), e[0]->dual());
    repair_first(s.vr->position(), e[1]->dual());
}

static void mesh_apply_split(Mesh &m, MeshSplit &s) throw (bad_collapse)
{
    if (s.is_vertex_split)
        mesh_apply_split_vertex(m, s.vertex_split);
    else
        mesh_apply_collapse_influence(m, s.influence_split, false);
}

//! True if \a edge is part of a patch boundary
static inline bool mesh_edge_sharp(MeshHalfEdgeConstIterator edge)
{
    return edge->face()->patch != edge->dual()->face()->patch;
}

//! True if \a pos lies on a patch boundary
static inline bool mesh_position_boundary(MeshPositionConstIterator pos)
{
    return pos->first_vertex() != pos->first_vertex()->first_edge()->dual()->prev()->vertex();
}

//! Determines whether one end of an edge must be kept fixed
/*! \returns -1 if there are no constraints
 * \returns -2 if both ends must be kept fixed (no collapse)
 * \returns 0 if the first (clockwise) end must be kept fixed
 * \returns 1 if the second (ccw) end must be kept fixed
 */

int MeshSimplifier::edge_forced_t(MeshHalfEdgeConstIterator edge)
{
    MeshPositionConstIterator p1 = edge->next()->vertex()->position();
    MeshPositionConstIterator p2 = edge->prev()->vertex()->position();
    bool fix1, fix2;

    if (mesh_edge_sharp(edge))
    {
        fix1 = p1->patch_corner;
        fix2 = p2->patch_corner;
    }
    else
    {
        fix1 = mesh_position_boundary(p1);
        fix2 = mesh_position_boundary(p2);
    }
    if (fix1 && fix2) return -2;
    else if (!fix1 && !fix2) return -1;
    else return fix1 ? 0 : 1;
}

bool MeshSimplifier::legal_collapse(MeshHalfEdgeConstIterator edge)
{
    MeshHalfEdgeConstIterator dual = edge->dual();

    // Don't bother collapsing edges entirely in the hole
    if (edge->face()->patch == 0 && dual->face()->patch == 0) return false;

    // Don't collapse a component consisting of two back-to-back triangles
    if (edge->next()->dual()->face() == dual->face()) return false;
    g_assert(edge->vertex()->position() != dual->vertex()->position());

    /* Meshes are smegging evil. If there is any position (other than
     * the obvious two) that are connected to both ends of the collapsing
     * edge, then the collapse would pinch out a (non-manifold) fin.
     */
    id_set<MeshPositionConstIterator>::type seen;
    MeshHalfEdgeConstIterator e = edge;
    do
    {
        seen.insert(e->vertex()->position());
        e = e->dual()->prev();
    } while (e != edge);
    e = edge;
    do
    {
        if (e->vertex()->position() != edge->vertex()->position()
            && e->vertex()->position() != dual->vertex()->position()
            && seen.count(e->vertex()->position())) return false;
        e = e->dual()->next();
    } while (e != edge);

    // Disallow patch topology changes (see Hoppe: Progressive Meshes)
    if (mesh_edge_sharp(edge->next()) && mesh_edge_sharp(edge->prev())) return false;
    if (mesh_edge_sharp(dual->next()) && mesh_edge_sharp(dual->prev())) return false;
    if (edge_forced_t(edge) == -2) return false;
    // we don't need Hoppe's other conditions since boundary curves don't end

    return true;
}

//! Returns true for exactly one of a pair of dual halfedges
static inline bool mesh_is_canonical_edge(MeshHalfEdgeConstIterator edge)
{
    return edge->id < edge->dual()->id;
}

//! Returns \c edge or \c edge->dual, whichever is canonical
static inline MeshHalfEdgeConstIterator mesh_get_canonical_edge(MeshHalfEdgeConstIterator edge)
{
    return (edge->id < edge->dual()->id) ? edge : edge->dual();
}

//! Returns \c edge or \c edge->dual, whichever is canonical
static inline MeshHalfEdgeIterator mesh_get_canonical_edge(MeshHalfEdgeIterator edge)
{
    return (edge->id < edge->dual()->id) ? edge : edge->dual();
}

//! Collapse an edge
/*! \param edge the edge to collapse
 * \param col the data about how to collapse the edge
 * \returns an iterator to the new position
 * \pre \c legal_collapse( \a edge ) is true
 * \pre \c evaluate_collapse( \a edge ) == \a col
 */
MeshPositionIterator MeshSimplifier::collapse_edge(MeshHalfEdgeIterator edge,
                                                   const MeshEdgeCollapseData &col)
{
    g_assert(legal_collapse(edge));
    g_assert(edge_forced_t(edge) == -1
             || edge_forced_t(edge) == col.t);

    MeshPositionIterator new_pos;
    MeshCollapse c;
    MeshEdgeCollapse &ec = c.edge_collapse;
    MeshSplit split;
    MeshVertexSplit &vs = split.vertex_split;

    c.is_edge_collapse = true;
    ec.p1() = edge->next()->vertex()->position();
    ec.p2() = edge->prev()->vertex()->position();
    c.error = col.cost;
    if (col.t == 0)
    {
        ec.vl() = edge->next()->vertex();
        ec.vr() = edge->dual()->prev()->vertex();
        new_pos = ec.vl()->position();
    }
    else if (col.t == 1)
    {
        ec.vl() = edge->prev()->vertex();
        ec.vr() = edge->dual()->next()->vertex();
        new_pos = ec.vl()->position();
    }
    else
    {
        new_pos = m.positions.insert_real(MeshPosition());
        new_pos->position = col.position;
        new_pos->patch_corner =
            edge->next()->vertex()->position()->patch_corner
            || edge->prev()->vertex()->position()->patch_corner;

        ec.vl() = m.vertices.insert_real(MeshVertex());
        ec.vl()->position() = new_pos;
        ec.vl()->data = col.vertex_data[0];
        ec.vl()->first_edge() = mesh_null_half_edge;
        if (edge->face()->patch == edge->dual()->face()->patch)
            ec.vr() = ec.vl();
        else
        {
            ec.vr() = m.vertices.insert_real(MeshVertex());
            ec.vr()->position() = new_pos;
            ec.vr()->data = col.vertex_data[1];
            ec.vr()->first_edge() = mesh_null_half_edge;
        }
    }

    split.is_vertex_split = true;
    vs.p = new_pos;
    vs.vl = edge->vertex();
    vs.vr = edge->dual()->vertex();
    vs.vl1 = edge->next()->vertex();
    vs.vl2 = edge->prev()->vertex();
    vs.vr1 = edge->dual()->prev()->vertex();
    vs.vr2 = edge->dual()->next()->vertex();
    split.error = c.error;
    if (splits.empty()) split.max_error = c.error;
    else split.max_error = max(splits.top().max_error, c.error);
    splits.push(split);
    m.append_collapse(c);
    mesh_apply_collapse_edge(m, ec, edge, false);

    return new_pos;
}

MeshPositionIterator MeshSimplifier::collapse_influence(MeshPositionIterator old_pos,
                                                        const MeshInfluenceCollapseData &col)
{
    MeshCollapse c;
    MeshInfluenceCollapse &ic = c.influence_collapse;
    MeshSplit split;
    MeshInfluenceCollapse &is = split.influence_split;
    MeshPositionIterator new_pos;

    // g_assert(mesh_manifold(m, MESH_LEVEL_PATCHES));
    g_assert(distance(col.position.begin(), col.position.end())
             < distance(old_pos->position.begin(), old_pos->position.end()));

    c.is_edge_collapse = false;
    c.error = col.cost;
    new_pos = m.positions.insert_real(MeshPosition());
    new_pos->position = col.position;
    new_pos->patch_corner = old_pos->patch_corner;
    new_pos->first_vertex() = mesh_null_vertex;

    /* Create new vertices */
    MeshHalfEdgeIterator e0 = old_pos->first_vertex()->first_edge();
    MeshHalfEdgeIterator e = e0;
    std::vector<MeshVertexData>::const_iterator cur = col.vertex_data.begin();
    do
    {
        MeshVertexIterator old_vertex = e->next()->vertex();
        if (mesh_true_vertex(old_vertex))
        {
            MeshVertexIterator new_vertex = m.vertices.insert_real(MeshVertex());
            new_vertex->data = *cur++;
            new_vertex->position() = new_pos;
            new_vertex->first_edge() = old_vertex->first_edge();
            if (old_pos->first_vertex() == old_vertex)
                new_pos->first_vertex() = new_vertex;

            ic.push_back(MeshInfluenceCollapseElement());
            ic.back().before() = old_vertex;
            ic.back().after() = new_vertex;
            is.push_back(MeshInfluenceCollapseElement());
            is.back().before() = new_vertex;
            is.back().after() = old_vertex;
        }
        // Work clockwise by patch
        e = e->dual()->prev()->vertex()->first_edge();
    } while (e != e0);

    split.is_vertex_split = false;
    split.error = c.error;
    if (splits.empty()) split.max_error = c.error;
    else split.max_error = max(splits.top().error, c.error);
    splits.push(split);
    m.append_collapse(c);
    mesh_apply_collapse_influence(m, ic, false);

    return new_pos;
}

class MeshSimplifierSimple : public MeshSimplifier
{
public:
    MeshSimplifierSimple(Mesh &m, bool do_influence, bool do_edge) : MeshSimplifier(m, do_influence, do_edge) {}
    virtual ~MeshSimplifierSimple() {}
    virtual MeshEdgeCollapseData *evaluate_edge_collapse(MeshHalfEdgeConstIterator edge) const;
    virtual MeshInfluenceCollapseData *evaluate_influence_collapse(MeshPositionConstIterator pos) const;
};

//! Determine how to perform a collapse
/*! The return may be \c HUGE_VAL to indicate an illegal collapse
 * \post if \c legal_collapse (\a edge) then the result is \c HUGE_VAL
 */
MeshEdgeCollapseData *MeshSimplifierSimple::evaluate_edge_collapse(MeshHalfEdgeConstIterator edge) const
{
    MeshEdgeCollapseData *answer = new MeshEdgeCollapseData(HUGE_VAL);

    if (!legal_collapse(edge)) return answer;

    MeshVertexConstIterator v1 = edge->next()->vertex();
    MeshVertexConstIterator v2 = edge->prev()->vertex();
    MeshVertexConstIterator d1 = edge->dual()->prev()->vertex();
    MeshVertexConstIterator d2 = edge->dual()->next()->vertex();
    MeshPositionConstIterator p1 = v1->position();
    MeshPositionConstIterator p2 = v2->position();

    double t = 0.5;
    int forced_t = edge_forced_t(edge);
    g_assert(forced_t != -2); // would indicate illegal collapse
    if (forced_t >= 0) t = forced_t;

    const Skeleton &skeleton = get_mesh().skeleton.get_read();
    answer->cost = length(p1->position - p2->position);
    answer->position = (1.0 - t) * p1->position + t * p2->position;
    answer->vertex_data[0] = mesh_vertex_data_interpolate(skeleton, v1->data, v2->data, t);
    if (d1 != v1)
        answer->vertex_data[1] = mesh_vertex_data_interpolate(skeleton, d1->data, d2->data, t);
    answer->t = forced_t;
    return answer;
}

MeshInfluenceCollapseData *MeshSimplifierSimple::evaluate_influence_collapse(MeshPositionConstIterator pos) const
{
    MeshInfluenceCollapseData *answer = new MeshInfluenceCollapseData(HUGE_VAL);
    return answer;
}

MeshSimplifier *mesh_simplifier_new_simple(Mesh &m, const string &method,
                                           bool do_influence, bool do_edge)
{
    return new MeshSimplifierSimple(m, do_influence, do_edge);
}

struct MeshCollapseMethod
{
    const char *name;
    MeshSimplifier *(*simplifier_new)(Mesh &m, const string &method,
                                      bool do_influence, bool do_edge);
};

/* The callbacks must allocate and return an object of a subclass of
 * MeshSimplifier, but not call initialise() on it.
 */
static const MeshCollapseMethod collapsers[] =
{
    { "simple", mesh_simplifier_new_simple },
    { "aps-nomem", mesh_simplifier_new_aps },
    { "aps", mesh_simplifier_new_aps },
    { NULL, NULL }
};

MeshSimplifier::MeshSimplifier(Mesh &m, bool do_influence, bool do_edge)
: m(m), do_influence(do_influence), do_edge(do_edge)
{
    /* The subclass constructor must call initialise(), since
     * at this point the VMT is incomplete and hence we cannot call
     * evaluate_collapse to build the priority queue.
     */
    max_metric = -HUGE_VAL;
    faces_left = m.faces.real_size();
    influences_left = 0;
    for (MeshPositionConstIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
        influences_left += distance(p->position.begin(), p->position.end());
    influences_report = influences_left / 1000 * 1000;
}

void MeshSimplifier::initialise()
{
    m.clear_lod();
    if (do_edge)
        for (MeshHalfEdgeIterator e = m.edges.whole_begin(); e != m.edges.whole_end(); e++)
            if (mesh_is_canonical_edge(e))
            {
                MeshEdgeCollapseData *data = evaluate_edge_collapse(e);
                if (finite(data->cost))
                    edge_collapses.push(make_pair(e, data));
                else
                    delete data;
            }
    if (do_influence)
        for (MeshPositionIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
        {
            MeshInfluenceCollapseData *data = evaluate_influence_collapse(p);
            if (finite(data->cost))
                influence_collapses.push(make_pair(p, data));
            else
                delete data;
        }
}

MeshSimplifier::~MeshSimplifier()
{
    while (!edge_collapses.empty())
    {
        delete edge_collapses.top().second;
        edge_collapses.pop();
    }
    while (!influence_collapses.empty())
    {
        delete influence_collapses.top().second;
        influence_collapses.pop();
    }
    while (!splits.empty())
    {
        mesh_apply_split(m, splits.top());
        splits.pop();
    }
}

void MeshSimplifier::simplify()
{
    while (simplify_once());
}

static inline void edge_collapse_erase(MeshEdgeCollapseQueue &q,
                                       MeshHalfEdgeIterator e)
{
    if (q.count(e))
    {
        delete q[e];
        q.erase(e);
    }
}

static inline void influence_collapse_erase(MeshInfluenceCollapseQueue &q,
                                            MeshPositionIterator p)
{
    if (q.count(p))
    {
        delete q[p];
        q.erase(p);
    }
}

void MeshSimplifier::edge_collapse_update(MeshHalfEdgeIterator e)
{
    if (!do_edge) return;

    g_assert(mesh_is_canonical_edge(e));
    if (edge_collapses.count(e))
        delete edge_collapses[e];

    MeshEdgeCollapseData *data = evaluate_edge_collapse(e);
    if (finite(data->cost))
        edge_collapses[e] = data;
    else
    {
        delete data;
        edge_collapses.erase(e);
    }
}

void MeshSimplifier::influence_collapse_update(MeshPositionIterator p)
{
    if (!do_influence) return;

    if (influence_collapses.count(p))
        delete influence_collapses[p];
    if (!mesh_true_position(p)) return;

    MeshInfluenceCollapseData *data = evaluate_influence_collapse(p);
    if (finite(data->cost))
        influence_collapses[p] = data;
    else
    {
        delete data;
        influence_collapses.erase(p);
    }
}

void MeshSimplifier::neighbourhood_update(MeshPositionIterator new_pos)
{
    MeshHalfEdgeIterator e1 = new_pos->first_vertex()->first_edge();
    do  // go around the new position
    {
        MeshHalfEdgeIterator e2 = e1;
        do  // go around one of the neighbour positions
        {
            MeshHalfEdgeIterator cedge = mesh_get_canonical_edge(e2);
            // In some extremely small cases this might lead to duplication,
            // but not enough to worry about
            edge_collapse_update(cedge);
            influence_collapse_update(e2->prev()->vertex()->position());
            e2 = e2->dual()->prev();
        } while (e2 != e1);
        e1 = e1->dual()->next();
    } while (e1 != new_pos->first_vertex()->first_edge());
    influence_collapse_update(new_pos);
}

bool MeshSimplifier::simplify_once()
{
    MeshPositionIterator new_pos;

    if (edge_collapses.empty()
        && influence_collapses.empty()) return false;

    // find a collapse
    GreaterCollapsePointer<MeshEdgeCollapseData, MeshInfluenceCollapseData> comparator;
    pair<MeshHalfEdgeIterator, MeshEdgeCollapseData *> ecoll(mesh_null_half_edge, NULL);
    pair<MeshPositionIterator, MeshInfluenceCollapseData *> icoll(mesh_null_position, NULL);

    if (!edge_collapses.empty()) ecoll = edge_collapses.top();
    if (!influence_collapses.empty()) icoll = influence_collapses.top();

    if (!comparator(ecoll.second, icoll.second))
    {
        g_assert(ecoll.second != NULL);
        g_assert(finite(ecoll.second->cost));
        g_assert(mesh_is_canonical_edge(ecoll.first));
        g_assert(legal_collapse(ecoll.first));
        edge_collapses.pop();
        max_metric = max(max_metric, ecoll.second->cost);

        // remove all old collapses from the neighbourhood
        MeshHalfEdgeIterator edge = ecoll.first;
        MeshHalfEdgeIterator dual = edge->dual();
        MeshHalfEdgeIterator nedge1 = edge->prev()->dual();
        MeshHalfEdgeIterator nedge2 = dual->prev()->dual();
        g_assert(edge_collapses.count(dual) == 0);
        edge_collapse_erase(edge_collapses, edge->prev());
        edge_collapse_erase(edge_collapses, edge->next());
        edge_collapse_erase(edge_collapses, dual->prev());
        edge_collapse_erase(edge_collapses, dual->next());
        influence_collapse_erase(influence_collapses, edge->prev()->vertex()->position());
        influence_collapse_erase(influence_collapses, edge->next()->vertex()->position());

        // perform the collapse and update counters
        MeshPositionConstIterator p1 = edge->prev()->vertex()->position();
        MeshPositionConstIterator p2 = edge->next()->vertex()->position();
        influences_left -= distance(p1->position.begin(), p1->position.end());
        influences_left -= distance(p2->position.begin(), p2->position.end());
        if (edge->face()->patch) faces_left--;
        if (edge->dual()->face()->patch) faces_left--;
        new_pos = collapse_edge(edge, *ecoll.second);
        influences_left += distance(new_pos->position.begin(), new_pos->position.end());
        // g_assert(mesh_manifold(m, MESH_LEVEL_PATCHES));
        if (influences_left <= influences_report)
        {
            g_message("%ld influences remain", (long) influences_report);
            influences_report -= 1000;
        }

        // remove collapses that are no longer canonical
        edge_collapse_erase(edge_collapses, mesh_get_canonical_edge(nedge1)->dual());
        edge_collapse_erase(edge_collapses, mesh_get_canonical_edge(nedge2)->dual());

        neighbourhood_update(new_pos);
        delete ecoll.second;
    }
    else
    {
        g_assert(icoll.second != NULL);
        g_assert(finite(icoll.second->cost));
        influence_collapses.pop();
        max_metric = max(max_metric, icoll.second->cost);

        MeshPositionIterator old_pos = icoll.first;
        influences_left -= distance(old_pos->position.begin(), old_pos->position.end());
        new_pos = collapse_influence(old_pos, *icoll.second);
        influences_left += distance(new_pos->position.begin(), new_pos->position.end());
        if (influences_left <= influences_report)
        {
            g_message("%ld influences remain", (long) influences_report);
            influences_report -= 1000;
        }

        /* Some of the edges attached to new_pos might have changed their
         * canonical orientation. Delete any that are no longer canonical
         */
        MeshHalfEdgeIterator e = new_pos->first_vertex()->first_edge();
        do
        {
            edge_collapse_erase(edge_collapses, mesh_get_canonical_edge(e)->dual());
            e = e->dual()->next(); // work clockwise by edge
        } while (e != new_pos->first_vertex()->first_edge());
        neighbourhood_update(new_pos);
        delete icoll.second;
    }

    return true;
}

MeshSimplifier *mesh_simplifier_new(Mesh &m, const string &method,
                                    bool do_influence, bool do_edge)
{
    const MeshCollapseMethod *C = collapsers;
    while (C->name && C->name != method) C++;
    if (!C->name)
        throw invalid_argument("illegal collapse method " + method);
    else
    {
        MeshSimplifier *s = C->simplifier_new(m, method, do_influence, do_edge);
        s->initialise();
        return s;
    }
}

/*! \param m the mesh to simplify
 * \param method the simplification method to use
 * \param do_influence generate influence simplifications
 * \param do_edge generate edge collapses
 * \pre m must be a complete mesh.
 */
double mesh_simplify(Mesh &m, const string &method,
                     bool do_influence, bool do_edge)
{
    Timer simplify_timer;

    MeshSimplifier *s = mesh_simplifier_new(m, method, do_influence, do_edge);
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Simpification setup time: %.3fs", simplify_timer.elapsed());
    s->simplify();
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Total simplification time: %.3fs", simplify_timer.elapsed());
    double ans = s->get_metric();
    delete s;
    return ans;
}

size_t mesh_reduce(Mesh &m, size_t items, bool do_faces, double &error, bool erase)
{
    size_t left = 0;

    if (do_faces) left = m.faces.real_size();
    else
    {
        for (MeshPositionConstIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
            left += distance(p->position.begin(), p->position.end());
    }

    while (left > items && m.collapses_begin() != m.collapses_end())
    {
        MeshCollapseIterator c = m.collapses_begin();
        MeshHalfEdgeIterator e = mesh_null_half_edge;
        if (c->is_edge_collapse)
        {
            MeshEdgeCollapse &ec = c->edge_collapse;
            e = find_edge(ec.p1(), ec.p2(), mesh_null_half_edge);
            if (do_faces)
            {
                if (e->face()->patch) left--;
                if (e->dual()->face()->patch) left--;
            }
            else
            {
                left += distance(ec.vl()->position()->position.begin(), ec.vl()->position()->position.end());
                left -= distance(ec.p1()->position.begin(), ec.p1()->position.end());
                left -= distance(ec.p2()->position.begin(), ec.p2()->position.end());
            }
            // Update the internal_data fields
            if (!erase)
            {
                ec.p1()->internal_data = ec.p2()->internal_data = &*ec.vl()->position();
                e->next()->vertex()->internal_data = e->prev()->vertex()->internal_data = &*ec.vl();
                e->dual()->next()->vertex()->internal_data = e->dual()->prev()->vertex()->internal_data = &*ec.vr();
            }
        }
        else
        {
            MeshInfluenceCollapse &ic = c->influence_collapse;
            MeshPositionIterator old_pos = ic.front().before()->position();
            MeshPositionIterator new_pos = ic.front().after()->position();
            if (!do_faces)
            {
                left += distance(new_pos->position.begin(), new_pos->position.end());
                left -= distance(old_pos->position.begin(), old_pos->position.end());
            }
            if (!erase)
            {
                for (MeshInfluenceCollapse::iterator i = ic.begin(); i != ic.end(); i++)
                    i->before()->internal_data = &*i->after();
                old_pos->internal_data = &*new_pos;
            }
        }
        // Do the collapse
        error = max(error, c->error);
        mesh_apply_collapse(m, *c, e, erase);
        m.pop_collapse();
    }
    return left;
}

/*** Influence simplifications ***/

InfluenceSimplificationCacheEntry::InfluenceSimplificationCacheEntry(const InfluenceSimplificationCacheEntry &e)
{
    g_assert(e.handles.empty());
}

InfluenceSimplificationCacheEntry &InfluenceSimplificationCacheEntry::operator =(InfluenceSimplificationCacheEntry &e)
{
    g_assert(e.handles.empty());
    g_assert(handles.empty());
    return *this;
}

InfluenceSimplificationCacheEntry::InfluenceSimplificationCacheEntry()
{
    /* No initialisation required */
}

InfluenceSimplificationCacheEntry::~InfluenceSimplificationCacheEntry()
{
    for (size_t i = 0; i < handles.size(); i++)
        free_matrix_cls(handles[i]);
}

InfluenceSimplificationCache::InfluenceSimplificationCache(const Skeleton &s, std::size_t c)
: skeleton(s), cache(), size(0), capacity(c), hits(0), misses(0)
{
}

const Skeleton &InfluenceSimplificationCache::get_skeleton() const
{
    return skeleton;
}

InfluenceSimplificationCacheEntry &InfluenceSimplificationCache::get_entry(const std::vector<int> &inf)
{
    list<pair<vector<int>, InfluenceSimplificationCacheEntry> >::iterator it;

    for (it = cache.begin(); it != cache.end(); it++)
    {
        if (it->first == inf)
        {
            list<pair<vector<int>, InfluenceSimplificationCacheEntry> > tmp;
            tmp.splice(tmp.begin(), cache, it);
            cache.splice(cache.begin(), tmp, it);
            hits++;
            return it->second;
        }
    }

    if (size == capacity)
        cache.pop_back();
    else
        size++;

    misses++;
    cache.push_front(make_pair(inf, InfluenceSimplificationCacheEntry()));
    return cache.begin()->second;
}

double InfluenceSimplificationCache::hit_rate() const
{
    return hits / (double) (hits + misses);
}

static void prepare_influence_simplification_cache_entry(const Skeleton &skeleton,
                                                         InfluenceSimplificationCacheEntry &e,
                                                         const vector<int> &inf)
{
    int xb = inf.size();
    MAT *P = skeleton.get_P();
    MAT *Px = MNULL;
    MAT *D = MNULL;

    g_assert(xb > 1);
    Px = m_get(4 * xb, 4 * xb);
    D = m_get(5, 4 * xb);

    int row = 0;
    for (vector<int>::const_iterator r = inf.begin(); r != inf.end(); r++)
        for (int rk = 0; rk < 4; rk++, row++)
        {
            int col = 0;
            for (vector<int>::const_iterator c = inf.begin(); c != inf.end(); c++)
                for (int ck = 0; ck < 4; ck++, col++)
                    m_set_val(Px, row, col, m_entry(P, *r + rk, *c + ck));
        }

    m_zero(D);
    for (int j = 3; j < 4 * xb; j += 4)
        m_set_val(D, 4, j, 1.0);
    int col = 0;
    for (vector<int>::const_iterator i = inf.begin(); i != inf.end(); i++, col += 4)
    {
        for (int ik = 0; ik < 4; ik++)
            m_set_val(D, ik, col + ik, 1.0);
        e.handles.push_back(alloc_matrix_cls(Px, D, 1e-8, 1e-2));
        for (int ik = 0; ik < 4; ik++)
            m_set_val(D, ik, col + ik, 0.0);
    }
    M_FREE(Px);
    M_FREE(D);
}

VectorPosition simplify_position_cache(InfluenceSimplificationCache &cache, const VectorPosition &x)
{
    const Skeleton &skeleton = cache.get_skeleton();
    vector<int> inf;
    int xb = distance(x.begin(), x.end());
    VEC *b = VNULL, *y = VNULL, *besty = VNULL;
    VectorPosition::const_iterator besti = x.end();
    double best_cost = HUGE_VAL;
    double cost;

    if (xb == 1) return x;

    for (VectorPosition::const_iterator i = x.begin(); i != x.end(); i++)
        inf.push_back(i->index);
    InfluenceSimplificationCacheEntry &e = cache.get_entry(inf);
    if (e.handles.empty())
        prepare_influence_simplification_cache_entry(skeleton, e, inf);

    b = v_get(5);
    y = v_get(4 * xb);
    besty = v_get(4 * xb);
    v_set_val(b, 4, 0.0);
    int index = 0;
    for (VectorPosition::const_iterator i = x.begin(); i != x.end(); i++, index++)
    {
        for (size_t ik = 0; ik < 4; ik++)
            v_set_val(b, ik, -i->values[ik]);
        y = solve_matrix_cls(e.handles[index], b, &cost, y);
        g_assert((int) y->dim == 4 * xb);
        if (cost < best_cost)
        {
            best_cost = cost;
            v_copy(y, besty);
            besti = i;
        }
    }

    VectorPosition q(skeleton), d(skeleton);
    int col = 0;
    for (VectorPosition::const_iterator j = x.begin(); j != x.end(); j++, col += 4)
        if (j != besti)
            for (int jk = 0; jk < 4; jk++)
                q[j->index + jk] = j->values[jk] + v_entry(besty, col + jk);

    V_FREE(b);
    V_FREE(y);
    V_FREE(besty);
    return q;
}

VectorPosition simplify_position_cache_remove(InfluenceSimplificationCache &cache,
                                              const VectorPosition &x,
                                              const std::vector<int> &remove)
{
    const Skeleton &skeleton = cache.get_skeleton();
    VEC *b = VNULL, *y = VNULL;
    VectorPosition ans = x;

    b = v_get(5);
    v_set_val(b, 4, 0.0);

    for (size_t i = 0; i < remove.size(); i++)
    {
        vector<int> inf;
        int rinf = remove[i];
        int rindex = -1;
        int xb = 0;
        for (VectorPosition::const_iterator i = ans.begin(); i != ans.end(); i++, xb++)
        {
            inf.push_back(i->index);
            if ((int) i->index == rinf) rindex = xb;
        }
        if (rindex == -1) continue;

        InfluenceSimplificationCacheEntry &e = cache.get_entry(inf);
        if (e.handles.empty())
            prepare_influence_simplification_cache_entry(skeleton, e, inf);

        for (size_t ik = 0; ik < 4; ik++)
            v_set_val(b, ik, -ans[rinf + ik]);
        y = solve_matrix_cls(e.handles[rindex], b, NULL, y);

        VectorPosition next(skeleton);
        int col = 0;
        for (VectorPosition::const_iterator j = ans.begin(); j != ans.end(); j++, col += 4)
            if ((int) j->index != rinf)
                for (int jk = 0; jk < 4; jk++)
                    next[j->index + jk] = j->values[jk] + v_entry(y, col + jk);

        ans = next;
    }

    V_FREE(y);
    V_FREE(b);
    return ans;
}

VectorPosition simplify_position_cache_keep(InfluenceSimplificationCache &cache,
                                            const VectorPosition &x,
                                            const vector<int> &keep)
{
    vector<int> remove;
    set<int> keep_set(keep.begin(), keep.end());
    for (VectorPosition::const_iterator i = x.begin(); i != x.end(); i++)
        if (!keep_set.count(i->index))
            remove.push_back(i->index);
    return simplify_position_cache_remove(cache, x, remove);
}

VectorPosition simplify_position(const Skeleton &skeleton, const VectorPosition &x)
{
    InfluenceSimplificationCache cache(skeleton, 1);
    return simplify_position_cache(cache, x);
}

/*! @} */
