/*! \addtogroup mesh
 * @{
 */

/*! \file
 * \brief Utility code for appearance preserving simplification
 */

#ifndef ALOD_MESH_SIMPLIFY_H
#define ALOD_MESH_SIMPLIFY_H

#include "mesh/mesh.h"
#include "mesh/iheap.h"
#include "geom/coords.h"
#include "geom/linear.h"
#include "common/maputils.h"
#include <vector>
#include <utility>
#include <stack>
#include <algorithm>
#include <iterator>
#include <string>
#include <functional>
#include <list>
#include <utility>
#include <stdexcept>
#include <cstddef>

class bad_collapse : public std::runtime_error
{
public:
    bad_collapse(const std::string &s) : std::runtime_error(s) {}
};

//! Data associated with an edge collapse
/*! Only data on legal collapses is stored. The actual edge to collapse
 * is implicit from the data structure containing this structure.
 */
struct MeshEdgeCollapseData
{
    double cost;                //!< Collapse metric (less is better)
    VectorPosition position;    //!< Location for new position
    int t;                      //!< 0 or 1 for half-collapse, -1 otherwise
    /*!< if forced_t is 0 or 1 then it must match \a t. */

    //! New texture coordinates for the new vertices
    /*! Only used when collapsing an edge of the patch boundary or
     * an internal edge of a patch.
     * \c new_vertices[0] is the data for the patch that
     * this edge belongs to, while \c new_vertices[1] is the
     * coordinates for the opposite patch if this is a patch boundary.
     * If one end of the edge is on a patch boundary but the other is not,
     * then the new vertex has the same data as the boundary
     * position.
     */
    MeshVertexData vertex_data[2];

    explicit MeshEdgeCollapseData(double c = -1.0) : cost(c) {}
};

struct MeshInfluenceCollapseData
{
    double cost;                //!< Collapse metric (less is better)
    VectorPosition position;    //!< New coordinates for the position

    //! New tangents for the new vertices
    /*! The elements are assigned starting from
     * \c MeshPosition::first_vertex on the original position, and moving
     * clockwise around the position. Data is excluded for holes.
     */
    std::vector<MeshVertexData> vertex_data;

    explicit MeshInfluenceCollapseData(double c = -1.0) : cost(c) {}
};

//! Acts as the comparator for the collapse queues
/*! The implementation is complicated by the fact that some items can be
 * inserted before being assigned priorities. This is handled by giving
 * NULL pointers the lowest priority.
 */
template<typename S, typename T>
struct GreaterCollapsePointer : std::binary_function<S *, T *, bool>
{
    bool operator()(S * const &a,
                    T * const &b)
    {
        if (!a && !b) return false;
        else if (!a) return true;
        else if (!b) return false;
        else return a->cost > b->cost;
    }
};

//! Describes a vertex split, the inverse of an edge collapse
/*! The vertex to split is \a p, and the edges that expand to faces are
 * (\a p, \a vl) and (\a p, \a vr). Visualise \a vl and \a vr as being
 * on the left and right of \a p respectively. Then \a p splits to the
 * positions \a vl1 and \a vr1 below and \a vl2 and \a vr2 above. To
 * clarify, the new triangles are (\a vl, \a vl1, \a vl2) and
 * (\a vr, \a vr2, \a vr1). In many cases, \a vl1 = \a vr1 and
 * \a vl2 = \a vr2, but not necessarily when dealing with patch
 * boundaries.
 */
struct MeshVertexSplit
{
public:
    MeshPositionIterator p;
    MeshVertexIterator vl, vr, vl1, vr1, vl2, vr2;
};

//! The inverse of a \c MeshCollapse.
/*! If \c is_vertex_split is true, then this is a vertex split and
 * \c vertex_split contains the data. Otherwise, it is an influence split,
 * and influence_split contains the data in a form suitable for passing to
 * mesh_apply_collapse_influence i.e., the before and after fields have been
 * reversed.
 */
struct MeshSplit
{
    bool is_vertex_split;
    MeshVertexSplit vertex_split;
    MeshInfluenceCollapse influence_split;
    double error;      //!< Copy of the error from the corresponding collapse
    double max_error;  //!< Cumulative error up to this point
};

//! \internal A class used internally by the simplifier. Not be used externally.
typedef indexed_priority_queue<
MeshHalfEdgeIterator,
MeshEdgeCollapseData *,
std::vector<std::pair<MeshHalfEdgeIterator, MeshEdgeCollapseData *> >,
GreaterCollapsePointer<MeshEdgeCollapseData, MeshEdgeCollapseData>,
id_map<MeshHalfEdgeIterator, std::iterator_traits<std::vector<std::pair<MeshHalfEdgeIterator, MeshEdgeCollapseData *> >::iterator>::difference_type>::type
> MeshEdgeCollapseQueue;

typedef indexed_priority_queue<
MeshPositionIterator,
MeshInfluenceCollapseData *,
std::vector<std::pair<MeshPositionIterator, MeshInfluenceCollapseData *> >,
GreaterCollapsePointer<MeshInfluenceCollapseData, MeshInfluenceCollapseData>,
id_map<MeshPositionIterator, std::iterator_traits<std::vector<std::pair<MeshPositionIterator, MeshInfluenceCollapseData *> >::iterator>::difference_type>::type
> MeshInfluenceCollapseQueue;

//! Driver class for simplification
/*! The driver uses the \c internal_data fields of the mesh. The mesh
 * should never be modified in any way during the lifetime of the
 * \c MeshSimplifier object. If you want to keep the original mesh,
 * you should first make a copy.
 */
class MeshSimplifier
{
private:
    Mesh &m;
    MeshEdgeCollapseQueue edge_collapses;
    MeshInfluenceCollapseQueue influence_collapses;
    std::stack<MeshSplit> splits;    //!< Those collapses that have happened
    std::size_t faces_left, influences_left;
    std::size_t influences_report;   //!< The next threshhold to report influences left
    bool do_influence, do_edge;
    double max_metric;

    bool simplify_once();

    //! Re-evaluates and updates the queue for the given edge
    /*! \pre \a e must be canonical
     */
    void edge_collapse_update(MeshHalfEdgeIterator e);

    //! Re-evaluates and updates the queue for the given position
    void influence_collapse_update(MeshPositionIterator p);

    //! Updates edge and influence collapses in the neighbourhood of \a new_pos
    void neighbourhood_update(MeshPositionIterator new_pos);

protected:
    //! Provided by children
    virtual MeshEdgeCollapseData *evaluate_edge_collapse(MeshHalfEdgeConstIterator edge) const = 0;
    virtual MeshInfluenceCollapseData *evaluate_influence_collapse(MeshPositionConstIterator pos) const = 0;

    //! May be extended by children, to track metrics
    virtual MeshPositionIterator collapse_edge(MeshHalfEdgeIterator edge, const MeshEdgeCollapseData &col);

    //! May be extended by children, to track metrics
    virtual MeshPositionIterator collapse_influence(MeshPositionIterator pos, const MeshInfluenceCollapseData &col);

    const Mesh &get_mesh() const { return m; }
    Mesh &get_mesh() { return m; }

    void reset_normals();
    //! Determines whether one end of an edge must be kept fixed
    static int edge_forced_t(MeshHalfEdgeConstIterator edge);
    //! Determines whether a collapse would violate some topology constraint
    static bool legal_collapse(MeshHalfEdgeConstIterator edge);

public:
    MeshSimplifier(Mesh &m, bool do_influence, bool do_edge);
    virtual ~MeshSimplifier();

    //! Must be called before any simplification is done (called by mesh_simplifier_new)
    virtual void initialise();
    //! Creates as many simplification records as possible
    void simplify();

    std::size_t get_faces() const { return faces_left; }
    std::size_t get_influences() const { return influences_left; }
    double get_metric() const { return max_metric; }
};

//! Create a simplifier by name
MeshSimplifier *mesh_simplifier_new(Mesh &m, const std::string &method,
                                    bool do_influence, bool do_edge);

//! Apply either an edge collapse or an influence collapse
void mesh_apply_collapse(Mesh &m, MeshCollapse &c,
                         MeshHalfEdgeIterator ehint,
                         bool erase) throw(bad_collapse);

//! Create a LOD sequence from \a m using \a method
double mesh_simplify(Mesh &m, const std::string &method,
                     bool do_influence, bool do_edge);

//! Use an existing LOD sequence to reduce a certain number of items
/*! Each \c MeshVertex and \c MeshPosition that is collapsed has its
 * \c internal_data field set to point at the new vertex/position. Note
 * that you may have to follow a chain to reach the final target. Vertices
 * and positions that remain have undefined \c internal_data.
 *
 * Set \a erase to \c false to keep the old vertices and positions. There is
 * no easy way to identify these kept vertices, so it is best used if the
 * entire mesh will be thrown away.
 *
 * If \a faces is true, \a items is a target face count, otherwise it is a
 * target influence count. The return value is the actual number left, which
 * may be slightly less than the target, or may be more if there was a lack
 * of collapses.
 *
 * \warning You must initialise \a error. On return, it will be set to
 * the maximum of the given value and any collapse error.
 */
std::size_t mesh_reduce(Mesh &m, std::size_t items, bool do_faces,
                        double &error, bool erase = true);

/*** Standalone influence simplification code ***/

class InfluenceSimplificationCache;

/* The keys into the cache are an ordered list of influences. The indices
 * are as stored in the index fields of the sparse vertices i.e. are
 * pre-multiplied by 4.
 */

class InfluenceSimplificationCacheEntry
{
private:
    friend class InfluenceSimplificationCache;
public:
    std::vector<MatrixCLS *> handles;

    InfluenceSimplificationCacheEntry();
    ~InfluenceSimplificationCacheEntry();

    /* Prevent copies except of empty cache entries */
    InfluenceSimplificationCacheEntry(const InfluenceSimplificationCacheEntry &e);
    InfluenceSimplificationCacheEntry &operator =(InfluenceSimplificationCacheEntry &e);
};

class InfluenceSimplificationCache
{
private:
    const Skeleton &skeleton;
    std::list<std::pair<std::vector<int>, InfluenceSimplificationCacheEntry> > cache;
    std::size_t size;
    std::size_t capacity;
    std::size_t hits, misses;
public:
    explicit InfluenceSimplificationCache(const Skeleton &s, std::size_t capacity);

    const Skeleton &get_skeleton() const;
    InfluenceSimplificationCacheEntry &get_entry(const std::vector<int> &inf);

    double hit_rate() const;
};

//! Uses the supplied cache to remove one influence from \a x
VectorPosition simplify_position_cache(InfluenceSimplificationCache &cache, const VectorPosition &x);

//! Removes the influences in \a remove and returns the optimised result
/*! \note The values in \a remove are pre-multiplied by 4 */
VectorPosition simplify_position_cache_remove(InfluenceSimplificationCache &cache,
                                              const VectorPosition &x,
                                              const std::vector<int> &remove);

//! Wrapper around \c simplify_position_cache_remove that says what to keep
/*! \note The values in \a keep are pre-multiplied by 4 */
VectorPosition simplify_position_cache_keep(InfluenceSimplificationCache &cache,
                                            const VectorPosition &x,
                                            const std::vector<int> &keep);

//! Wrapper around \c simplify_position_cache that uses a once-off cache (inefficient)
VectorPosition simplify_position(const Skeleton &skeleton, const VectorPosition &x);

#endif /* !ALOD_MESH_SIMPLIFY_H */

/*!
 * @}
 */
