#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <glib.h>
#include <string>
#include <vector>
#include "render/meshrender.h"
#include "render/interfaces.h"
#include "render/geometry.h"

using namespace std;

void RenderHelperClassGeometry::begin_state(CGprogram vp, CGprogram fp, int flags)
{
    const char *names[3] = {"i_position", "i_tangents[0]", "i_tangents[1]"};

    RenderHelperClass::begin_state(vp, fp, flags);
    buffer_count = 0;
    int reps = (flags & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    glGenBuffers(reps, buffers);
    for (int r = 0; r < reps; r++)
    {
        parameters[r] = cgGetNamedParameter(vp, names[r]);
        glBindBuffer(GL_ARRAY_BUFFER, buffers[r]);
        cgGLSetParameterPointer(parameters[r], 4, GL_FLOAT, 0, NULL);
        cgGLEnableClientState(parameters[r]);
    }
    fill(parameters + reps, parameters + 3, (CGparameter) NULL);

    if (flags & RENDER_FLAG_TEXCOORDS)
    {
        texcoord_parameter = cgGetNamedParameter(vp, "i_texcoords");
        cgGLEnableClientState(texcoord_parameter);
    }
    flod_parameter = cgGetNamedParameter(vp, "u_flod");

    stage_buffer = NULL;
}

void RenderHelperClassGeometry::end_state()
{
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    glDeleteBuffers(reps, buffers);
    for (int r = 0; r < reps; r++)
        cgGLDisableClientState(parameters[r]);
    if (global_flags() & RENDER_FLAG_TEXCOORDS)
        cgGLDisableClientState(texcoord_parameter);
    free(stage_buffer);
    RenderHelperClass::end_state();
}

VectorPosition RenderHelperGeometryBase::influences(const MeshVertex *base,
                                                    const MeshVertex *target)
{
    VectorPosition test = target->position()->position - base->position()->position;
    if (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES)
    {
        test += target->data.tangent_plane.tangents[0] - base->data.tangent_plane.tangents[0];
        test += target->data.tangent_plane.tangents[1] - base->data.tangent_plane.tangents[1];
    }
    return test;
}

VectorPosition RenderHelperGeometryBase::influences(const pair<MeshVertex *, MeshVertex *> &vpair)
{
    return influences(vpair.first, vpair.second);
}

const VectorPosition &RenderHelperGeometryBase::get_channel(const MeshVertex *v, int c)
{
    switch (c)
    {
    case 0:
        return v->position()->position;
    default:
        return v->data.tangent_plane.tangents[c - 1];
    }
}


RenderHelperGeometry::RenderHelperGeometry(RenderHelperClassGeometry &parent, const RenderMesh &m)
: RenderHelperGeometryBase(parent, m)
{
    if (parent.global_flags() & RENDER_FLAG_TEXCOORDS)
    {
        texcoord_buffers.resize(m.levels.size());
        int i;
        vector<RenderMeshLevel>::const_iterator l;
        for (i = 0, l = m.levels.begin(); l != m.levels.end(); i++, l++)
        {
            glGenBuffers(1, &texcoord_buffers[i]);
            glBindBuffer(GL_ARRAY_BUFFER, texcoord_buffers[i]);
            glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(GLfloat) * l->vertices.size(), NULL, parent.buffer_usage());
            GLfloat *data = (GLfloat *) glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
            for (size_t j = 0; j < l->vertices.size(); j++)
            {
                data[0] = l->vertices[j].first->data.texcoords[0];
                data[1] = l->vertices[j].first->data.texcoords[1];
                data[2] = l->vertices[j].second->data.texcoords[0] - data[0];
                data[3] = l->vertices[j].second->data.texcoords[1] - data[1];
                data += 4;
            }
            glUnmapBuffer(GL_ARRAY_BUFFER);
        }
    }
}

RenderHelperGeometry::~RenderHelperGeometry()
{
    for (size_t i = 0; i < texcoord_buffers.size(); i++)
        glDeleteBuffers(1, &texcoord_buffers[i]);
}

void RenderHelperGeometry::begin_object(const Skeleton &skeleton, double lod, double z,
                                        size_t buffer_min)
{
    RenderHelperClassGeometry &klass = get_class_geometry();
    size_t &buffer_count = klass.buffer_count;
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    if (buffer_min > buffer_count)
    {
        buffer_count = MAX(buffer_count * 2, buffer_min);
        for (int r = 0; r < reps; r++)
        {
            glBindBuffer(GL_ARRAY_BUFFER, klass.get_buffer(r));
            glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(GLfloat) * buffer_count, NULL, klass.buffer_usage());
        }
        if (klass.staged)
        {
            free(klass.stage_buffer);
            posix_memalign((void **) &klass.stage_buffer, 16, 4 * sizeof(GLfloat) * buffer_count);
        }
    }

    int ilod = rmesh.int_lod(lod);
    cgGLSetParameter1f(klass.flod_parameter, lod - ilod);

    if (global_flags() & RENDER_FLAG_TEXCOORDS)
    {
        glBindBuffer(GL_ARRAY_BUFFER, texcoord_buffers[ilod]);
        cgGLSetParameterPointer(get_class_geometry().texcoord_parameter,
                                4, GL_FLOAT, 0, NULL);
    }
}

void RenderHelperGeometry::begin_object(const Skeleton &skeleton, double lod, double z)
{
    int ilod = rmesh.int_lod(lod);

    const RenderMeshLevel &level = rmesh.levels[ilod];
    begin_object(skeleton, lod, z, level.vertices.size());
}

float *RenderHelperGeometry::begin_buffer()
{
    RenderHelperClassGeometry &klass = get_class_geometry();
    if (klass.staged)
        return klass.stage_buffer;
    else
    {
        // Hint that the old data may be discarded
        glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(GLfloat) * klass.buffer_count, NULL, klass.buffer_usage());
        float *ptr = (float *) glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
        if (!ptr)
            g_error("glMapBuffer failed - possibly out of memory");
        if (GPOINTER_TO_SIZE(ptr) & 0xf)
            g_error("Rendering method unavailable (glMapBuffer returned unaligned pointer)");
        return ptr;
    }
}

void RenderHelperGeometry::end_buffer(size_t num_vertices)
{
    RenderHelperClassGeometry &klass = get_class_geometry();
    if (klass.staged)
        glBufferSubData(GL_ARRAY_BUFFER, 0, num_vertices * 4 * sizeof(GLfloat), klass.stage_buffer);
    else
    {
        GLboolean success = glUnmapBuffer(GL_ARRAY_BUFFER);
        if (!success)
            g_warning("glUnmapBuffer failed - corruption possible");
    }
}
