#ifndef ALOD_RENDER_GEOMETRY_H
#define ALOD_RENDER_GEOMETRY_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "render/meshrender.h"
#include "render/interfaces.h"
#include <glib.h>
#include <vector>
#include <string>
#include <cstddef>
#include <utility>
#include <GL/gl.h>

/* In general, the 3 channels are:
 * 0: position
 * 1: tangent A
 * 2: tangent B
 */

class RenderHelperClassGeometry : public RenderHelperClass
{
private:
    //! prevent default construction
    RenderHelperClassGeometry() {} 
protected:
    bool staged;
    GLuint buffers[3];
    float *stage_buffer;          //!< system memory temporary buffer
    std::size_t buffer_count;

    CGparameter parameters[3];
    CGparameter texcoord_parameter;
    CGparameter flod_parameter;

    //! Overridden by child classes to indicate how the dynamic buffers should be allocated
    virtual GLenum buffer_usage() = 0;

    virtual std::string vertex_program_class() const { return "GeometryVP"; }

    GLuint get_buffer(int c)
    { g_assert(0 <= c && c < 3); return buffers[c]; }

    friend class RenderHelperGeometry;
public:
    explicit RenderHelperClassGeometry(bool staged) : RenderHelperClass(), staged(staged) {}

    virtual RenderHelperType type() const { return RENDER_HELPER_GEOMETRY; }
    virtual void begin_state(CGprogram vp, CGprogram fp, int flags);
    virtual void end_state();
};

//! Utility class that provides some helper methods to geometry classes
class RenderHelperGeometryBase : public RenderHelper
{
public:
    //! Returns a position whose influences are the unions of those of interest
    VectorPosition influences(const MeshVertex *base, const MeshVertex *target);

    //! Returns a position whose influences are the unions of those of interest
    VectorPosition influences(const std::pair<MeshVertex *, MeshVertex *> &vpair);

    //! Returns either the position or one of the tangents
    static const VectorPosition &get_channel(const MeshVertex *v, int c);

    RenderHelperGeometryBase(RenderHelperClass &parent, const RenderMesh &m)
        : RenderHelper(parent, m) {}
};

class RenderHelperGeometry : public RenderHelperGeometryBase
{
private:
    std::vector<GLuint> texcoord_buffers;
public:
    RenderHelperGeometry(RenderHelperClassGeometry &parent, const RenderMesh &m);
    virtual ~RenderHelperGeometry();

    RenderHelperClassGeometry &get_class_geometry()
    { return static_cast<RenderHelperClassGeometry &>(get_class()); }

    //! Internal version that takes buffer size lower bound as parameter
    void begin_object(const Skeleton &skeleton, double lod, double z, std::size_t buffer_min);
    virtual void begin_object(const Skeleton &skeleton, double lod, double z);
protected:
    GLuint get_buffer(int c) { return get_class_geometry().get_buffer(c); }

    //! Returns an output buffer pointer
    /*! If \c staged is true, it is the staging buffer, otherwise the
     * return from glMapBuffer. It will be 16-byte aligned.
     * \pre The appropriate VBO must be bound to \c GL_ARRAY_BUFFER
     */
    float *begin_buffer();
    //! Release the buffer allocated by \c begin_buffer e.g. call \c glUnmapBuffer
    void end_buffer(std::size_t num_vertices);
};

#endif /* !ALOD_RENDER_GEOMETRY_H */
