#if HAVE_CONFIG_H
# include <config.h>
#endif
#ifndef _XOPEN_SOURCE
# define _XOPEN_SOURCE 600
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <cstddef>
#include <glib.h>
#include <string>
#include <cstdlib>
#include <xmmintrin.h>
#include "render/interfaces.h"
#include "render/geometry.h"
#include "render/geometry_col.h"

using namespace std;

class RenderHelperGeometryCol : public RenderHelperGeometry
{
    struct Stripe               //!< Data about a single bone
    {
        size_t size;            //!< number of influences from this stripe
        __m128 *vertex_data[2]; //!< base, offset
        int *indices;           //!< indices into output buffer
    };
    struct Channel
    {
        std::vector<Stripe> stripes;
    };
    struct Level
    {
        std::vector<Channel> channels;  // v, t0, t1
    };

    std::vector<Level> levels;
public:
    RenderHelperGeometryCol(RenderHelperClassGeometryCol &parent, const RenderMesh &m);
    virtual ~RenderHelperGeometryCol();

    virtual void begin_object(const Skeleton &skeleton, double lod, double z);
};


RenderHelper *RenderHelperClassGeometryCol::new_instance(const RenderMesh &m)
{
    return new RenderHelperGeometryCol(*this, m);
}

RenderHelperGeometryCol::RenderHelperGeometryCol(RenderHelperClassGeometryCol &parent, const RenderMesh &m)
: RenderHelperGeometry(parent, m)
{
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;
    int bones = m.mesh.skeleton.get_read().size();

    levels.resize(m.levels.size());
    for (size_t i = 0; i < m.levels.size(); i++)
    {
        const RenderMeshLevel &rl = m.levels[i];
        Level &l = levels[i];
        int num_vertices = rl.vertices.size();

        l.channels.resize(reps);
        for (int r = 0; r < reps; r++)
        {
            size_t counts[bones];
            Channel &ch = l.channels[r];

            ch.stripes.resize(bones);
            fill(counts, counts + bones, 0U);
            for (int i = 0; i < num_vertices; i++)
            {
                VectorPosition test = get_channel(rl.vertices[i].second, r) - get_channel(rl.vertices[i].first, r);
                for (VectorPosition::const_iterator j = test.begin(); j != test.end(); j++)
                    counts[j->index / 4]++;
            }

            for (int i = 0; i < bones; i++)
            {
                posix_memalign((void **) &ch.stripes[i].vertex_data[0], 16, counts[i] * sizeof(__m128));
                posix_memalign((void **) &ch.stripes[i].vertex_data[1], 16, counts[i] * sizeof(__m128));
                ch.stripes[i].indices = new int[counts[i]];
                ch.stripes[i].size = 0;
            }

            for (int i = 0; i < num_vertices; i++)
            {
                const VectorPosition &vbase = get_channel(rl.vertices[i].first, r);
                const VectorPosition &vtarget = get_channel(rl.vertices[i].second, r);
                VectorPosition offset = vtarget - vbase;
                for (VectorPosition::const_iterator j = offset.begin(); j != offset.end(); j++)
                {
                    Stripe &st = ch.stripes[j->index / 4];
                    g_assert(st.size < counts[j->index / 4]);
                    st.vertex_data[0][st.size] = _mm_setr_ps(vbase[j->index + 0], vbase[j->index + 1], vbase[j->index + 2], vbase[j->index + 3]);
                    st.vertex_data[1][st.size] = _mm_setr_ps(offset[j->index + 0], offset[j->index + 1], offset[j->index + 2], offset[j->index + 3]);
                    st.indices[st.size] = i;
                    st.size++;
                }
            }
        }
    }
}

/* Takes a 4x4 matrix in column-major form (one column per __m128) and
 * transforms v.
 */
static inline __m128 matrix_transform_sse(const __m128 m[],
                                          const __m128 &v)
{
    __m128 out, vx;
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(0, 0, 0, 0));
    out = _mm_mul_ps(m[0], vx);
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(1, 1, 1, 1));
    out = _mm_add_ps(out, _mm_mul_ps(m[1], vx));
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(2, 2, 2, 2));
    out = _mm_add_ps(out, _mm_mul_ps(m[2], vx));
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(3, 3, 3, 3));
    out = _mm_add_ps(out, _mm_mul_ps(m[3], vx));
    return out;
}

/* Transforms a pair, applying a suitable flod blend */
static inline __m128 transform_pair_sse(const __m128 &a,
                                        const __m128 &b, const __m128 m[],
                                        const __m128 &flod)
{
    __m128 v = _mm_add_ps(a, _mm_mul_ps(b, flod));
    return matrix_transform_sse(m, v);
}

void RenderHelperGeometryCol::begin_object(const Skeleton &skeleton, double lod, double z)
{
    int ilod = rmesh.int_lod(lod);
    size_t num_vertices = rmesh.levels[ilod].vertices.size();
    int bones = skeleton.size();
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;
    __m128 flod = _mm_set1_ps(CLAMP(lod - ilod, 0.0, 1.0));

    RenderHelperGeometry::begin_object(skeleton, lod, z);

    for (int r = 0; r < reps; r++)
    {
        glBindBuffer(GL_ARRAY_BUFFER, get_buffer(r));
        __m128 *ptr = (__m128 *) begin_buffer();
        memset(ptr, 0, num_vertices * sizeof(__m128));
        for (int b = 0; b < bones; b++)
        {
            const Stripe &st = levels[ilod].channels[r].stripes[b];
            const Matrix4 &cm = skeleton.get_absolute(b);
            __m128 transform[4];
            for (int j = 0; j < 4; j++)
                transform[j] = _mm_setr_ps(cm[0][j], cm[1][j], cm[2][j], cm[3][j]);

            for (size_t i = 0; i < st.size; i++)
            {
                __m128 &out = ptr[st.indices[i]];
                out = _mm_add_ps(out, transform_pair_sse(st.vertex_data[0][i],
                                                         st.vertex_data[1][i],
                                                         transform,
                                                         flod));
            }
        }
        end_buffer(num_vertices);
    }
}

RenderHelperGeometryCol::~RenderHelperGeometryCol()
{
    for (size_t i = 0; i < levels.size(); i++)
        for (size_t j = 0; j < levels[i].channels.size(); j++)
            for (size_t k = 0; k < levels[i].channels[j].stripes.size(); k++)
            {
                free(levels[i].channels[j].stripes[k].vertex_data[0]);
                free(levels[i].channels[j].stripes[k].vertex_data[1]);
                delete[] levels[i].channels[j].stripes[k].indices;
            }
}
