#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include "render/interfaces.h"
#include "render/meshrender.h"
#include "render/geometry.h"
#include "render/geometry_rtvb.h"
#include "render/geometry_frag.h"
#include "geom/skeleton.h"
#include <glib.h>
#include <GL/gl.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <cstddef>
#include <string>

using namespace std;

class RenderHelperGeometryFrag : public RenderHelperGeometryRTVB
{
protected:
    struct Level
    {
        GLint tex_width, tex_height;
        GLuint inf_tex[3][2];   //!< indexed by channel and [base, offset]
        GLuint bone_tex;
        GLuint draw_list;       //!< Display list that issues the draw commands
    };

    std::vector<Level> levels;
public:
    RenderHelperGeometryFrag(RenderHelperClassGeometryFrag &parent, const RenderMesh &m);
    virtual ~RenderHelperGeometryFrag();

    RenderHelperClassGeometryFrag &get_class_geometry_frag()
    { return static_cast<RenderHelperClassGeometryFrag &>(get_class()); }

    virtual void begin_object(const Skeleton &skeleton, double lod, double z);
};


RenderHelper *RenderHelperClassGeometryFrag::new_instance(const RenderMesh &m)
{
    return new RenderHelperGeometryFrag(*this, m);
}

void RenderHelperClassGeometryFrag::begin_state(CGprogram vp, CGprogram fp, int flags)
{
    int reps = (flags & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    RenderHelperClassGeometryRTVB::begin_state(vp, fp, flags);

    vp = rtvb_vertex_program();
    fp = rtvb_fragment_program();
    CGparameter inf = cgGetNamedParameter(fp, "inf_tex");
    g_assert(inf);
    for (int r = 0; r < reps; r++)
        for (int j = 0; j < 2; j++)
            inf_parameter[r][j] = cgGetArrayParameter(cgGetArrayParameter(inf, r), j);
    bone_parameter = cgGetNamedParameter(fp, "bone_tex");
    transform_parameter = cgGetNamedParameter(fp, "transform_tex");
    flod_parameter = cgGetNamedParameter(fp, "flod");
}

void RenderHelperClassGeometryFrag::end_state()
{
    RenderHelperClassGeometryRTVB::end_state();
}


static void do_block(int rbo_row, int tex_row, int inf,
                     int width, int rbo_height, bool start)
{
    GLfloat y = ((GLfloat) rbo_row / rbo_height) * 2.0f - 1.0f;
    GLfloat xl = -1.0f;
    GLfloat xh = 1.0f;

    GLfloat tl = (GLfloat) tex_row - (inf - 1) * 0.5f;
    GLfloat th = tl + inf - 0.5f;
    GLfloat sl = 0.0f;
    GLfloat sh = width;

    if (!start)
    {
        swap(xl, xh);
        swap(sl, sh);
    }
    glMultiTexCoord3f(GL_TEXTURE0, sl, tl, th);
    glVertex2f(xl, y);
    glMultiTexCoord3f(GL_TEXTURE0, sh, tl, th);
    glVertex2f(xh, y);
}

static void compute_sizes(const vector<int> &partial,
                          GLint &width, GLint &tex_height, GLint &rbo_height)
{
    GLint max_tex_height, max_rbo_height, max_width;

    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &max_tex_height);
    glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE_EXT, &max_rbo_height);
    max_width = min(max_tex_height, max_rbo_height);
    // Can only reliably address 512 with half precision
    max_tex_height = min(512, max_tex_height);
    for (width = 4; width <= max_width; width *= 2)
    {
        tex_height = 0;
        for (size_t i = 1; i < partial.size(); i++)
            tex_height += (partial[i] + width - 1) / width;
        rbo_height = (partial[1] + width - 1) / width;
        if (rbo_height <= max_rbo_height
            && tex_height <= max_tex_height)
            return;
    }
    g_error("Too many vertices for GeometryFrag processor");
}

RenderHelperGeometryFrag::RenderHelperGeometryFrag(RenderHelperClassGeometryFrag &parent, const RenderMesh &m)
: RenderHelperGeometryRTVB(parent, m)
{
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;
    levels.resize(m.levels.size());
    for (size_t i = 0; i < m.levels.size(); i++)
    {
        const RenderMeshLevel &rl = m.levels[i];
        Level &l = levels[i];
        GLint &width = l.tex_width, &tex_height = l.tex_height;
        GLint rbo_height;

        vector<int> partial;   // number of vertices with i or more influences
        for (size_t v = 0; v < rl.vertices.size(); v++)
        {
            const VectorPosition inf = influences(rl.vertices[v]);
            size_t d = distance(inf.begin(), inf.end());
            if (d >= partial.size()) partial.resize(d + 1, 0);
            partial[d] = v + 1;
        }
        /* If there is a count that does not appear at all, it will be missed above */
        for (size_t j = (int) partial.size() - 1; j > 0; j--)
            partial[j - 1] = max(partial[j - 1], partial[j]);

        g_assert(partial.size() <= 256);

        compute_sizes(partial, width, tex_height, rbo_height);
        set_rbo_size(i, width, rbo_height);

        GLfloat *inf_data[3][2];
        for (int r = 0; r < reps; r++)
            for (int j = 0; j < 2; j++)
            {
                inf_data[r][j] = new GLfloat[width * tex_height * 4];
                fill(inf_data[r][j], inf_data[r][j] + width * tex_height * 4, 0.0f);
            }
        GLubyte *bone_data = new GLubyte[width * tex_height];
        fill(bone_data, bone_data + width * tex_height, 0);

        l.draw_list = glGenLists(1);
        glNewList(l.draw_list, GL_COMPILE);
        glBegin(GL_QUADS);

        int rbo_row = 0, tex_row = 0, col = 0;
        int last_inf = -1;    // influence count of the current row
        for (size_t v = 0; v < rl.vertices.size(); v++)
        {
            const VectorPosition inf = influences(rl.vertices[v]);
            int d = distance(inf.begin(), inf.end());

            if (last_inf != d && col == 0)
            {
                /* Start of a new block */
                if (last_inf != -1)
                    do_block(rbo_row, tex_row, last_inf, width, rbo_height, false);
                last_inf = d;
                do_block(rbo_row, tex_row, last_inf, width, rbo_height, true);
            }

            int ofs = 0;
            for (VectorPosition::const_iterator p = inf.begin(); p != inf.end(); p++, ofs++)
            {
                for (int r = 0; r < reps; r++)
                {
                    const VectorPosition base = get_channel(rl.vertices[v].first, r);
                    const VectorPosition target = get_channel(rl.vertices[v].second, r);
                    const VectorPosition delta = target - base;
                    for (int j = 0; j < 2; j++)
                        for (int k = 0; k < 4; k++)
                            inf_data[r][j][(((tex_row + ofs) * width) + col) * 4 + k] =
                                (j ? delta : base)[p->index + k];
                }
                bone_data[((tex_row + ofs) * width) + col] = p->index / 4;
            }
            col++;
            if (col == width)
            {
                tex_row += last_inf;
                rbo_row++;
                col = 0;
            }
        }
        if (col)
        {
            tex_row += last_inf;
            rbo_row++;
        }
        g_assert(rbo_row <= rbo_height);
        g_assert(tex_row <= tex_height);
        if (last_inf != -1)
            do_block(rbo_row, tex_row, last_inf, width, rbo_height, false);
        glEnd();
        glEndList();

        for (int r = 0; r < reps; r++)
            for (int j = 0; j < 2; j++)
            {
                glGenTextures(1, &l.inf_tex[r][j]);
                glBindTexture(GL_TEXTURE_RECTANGLE_ARB, l.inf_tex[r][j]);
                glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
                glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
                glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
                glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
                glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA16F_ARB,
                             width, tex_height, 0, GL_RGBA, GL_FLOAT,
                             inf_data[r][j]);
                delete[] inf_data[r][j];
            }
        glGenTextures(1, &l.bone_tex);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, l.bone_tex);
        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_LUMINANCE8,
                     width, tex_height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE,
                     bone_data);
        delete[] bone_data;
    }
}

RenderHelperGeometryFrag::~RenderHelperGeometryFrag()
{
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;
    for (size_t i = 0; i < levels.size(); i++)
    {
        Level &l = levels[i];
        for (int r = 0; r < reps; r++)
            glDeleteTextures(2, &l.inf_tex[r][0]);
        glDeleteTextures(1, &l.bone_tex);
    }
}

void RenderHelperGeometryFrag::begin_object(const Skeleton &skeleton, double lod, double z)
{
    TextureManager manager;

    RenderHelperGeometryRTVB::begin_object(skeleton, lod, z);

    RenderHelperClassGeometryFrag &klass = get_class_geometry_frag();
    CGprogram vp = rtvb_vertex_program();
    CGprogram fp = rtvb_fragment_program();
    int ilod = rmesh.int_lod(lod);
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;
    const Level &l = levels[ilod];

    GLuint transform_tex = update_transform_texture(skeleton);

    cgGLBindProgram(vp);
    cgGLBindProgram(fp);
    cgGLSetParameter1f(klass.flod_parameter, lod - ilod);
    for (int r = 0; r < reps; r++)
        for (int j = 0; j < 2; j++)
            manager.bind_texture(klass.inf_parameter[r][j], l.inf_tex[r][j]);
    manager.bind_texture(klass.bone_parameter, l.bone_tex);
    manager.bind_texture(klass.transform_parameter, transform_tex);

    glCallList(l.draw_list);

    manager.reset();
    RenderHelperGeometryRTVB::restore_object(skeleton, lod, z);
}
