#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include "render/interfaces.h"
#include "render/meshrender.h"
#include "render/geometry.h"
#include "render/geometry_rtvb.h"
#include <glib.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <string>

using namespace std;

void RenderHelperClassGeometryRTVB::begin_state(CGprogram vp, CGprogram fp, int flags)
{
    static const GLenum buffers[3] =
    {
        GL_COLOR_ATTACHMENT0_EXT,
        GL_COLOR_ATTACHMENT1_EXT,
        GL_COLOR_ATTACHMENT2_EXT
    };
    int reps = (flags & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    RenderHelperClassGeometry::begin_state(vp, fp, flags);
    glGenFramebuffersEXT(1, &fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    glGenRenderbuffersEXT(reps, rbo);
    rbo_width = 16;
    rbo_height = 16;
    for (int i = 0; i < reps; i++)
    {
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, rbo[i]);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_RGBA32F_ARB,
                                 rbo_width, rbo_height);
        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, buffers[i],
                                     GL_RENDERBUFFER_EXT, rbo[i]);
    }
    glDrawBuffers(reps, buffers);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

    string arg_string = string("-DREPS=") + (reps == 3 ? "3" : "1");
    const char *args[2] = { arg_string.c_str(), NULL };

    rtvb_vp = load_cg_program(rtvb_vertex_program_file(),
                              global_cg_vertex_profile(),
                              args);
    check_cg_error();
    cgCompileProgram(rtvb_vp);
    cgGLLoadProgram(rtvb_vp);
    check_cg_error();

    rtvb_fp = load_cg_program(rtvb_fragment_program_file(),
                              global_cg_fragment_profile(),
                              args);
    check_cg_error();
    cgCompileProgram(rtvb_fp);
    cgGLLoadProgram(rtvb_fp);
    check_cg_error();
}

void RenderHelperClassGeometryRTVB::end_state()
{
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    glDeleteRenderbuffersEXT(reps, rbo);
    glDeleteFramebuffersEXT(1, &fbo);
    cgDestroyProgram(rtvb_vp);
    cgDestroyProgram(rtvb_fp);
    transform_texture.end();
    RenderHelperClassGeometry::end_state();
}


RenderHelperGeometryRTVB::RenderHelperGeometryRTVB(RenderHelperClassGeometryRTVB &parent, const RenderMesh &m)
: RenderHelperGeometry(parent, m)
{
    rtvb_levels.resize(m.levels.size());
    for (size_t i = 0; i < m.levels.size(); i++)
    {
        GLint v = m.levels[i].vertices.size();
        GLint w = 4;
        while (w * w < v)
            w *= 2;
        rtvb_levels[i].rbo_width = w;
        rtvb_levels[i].rbo_height = (v + w - 1) / w;
    }
}

RenderHelperGeometryRTVB::~RenderHelperGeometryRTVB()
{
}

void RenderHelperGeometryRTVB::begin_object(const Skeleton &skeleton, double lod, double z)
{
    /* Note: we chain to the internal version of
     * RenderHelperGeometry::begin_object rather than the public version,
     * because the latter does not allocate the padded VBO necessary to do
     * a single ReadPixels without overrunning it.
     *
     * Also, the caller is responsible for switching vertex and fragment
     * programs.
     */
    RenderHelperClassGeometryRTVB &klass = get_class_geometry_rtvb();
    GLint &rbo_width = klass.rbo_width;
    GLint &rbo_height = klass.rbo_height;
    int ilod = rmesh.int_lod(lod);
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    const RTVBLevel &level = rtvb_levels[ilod];
    if (level.rbo_width > rbo_width || level.rbo_height > rbo_height)
    {
        rbo_width = max(rbo_width, level.rbo_width);
        rbo_height = max(rbo_height, level.rbo_height);
        for (int r = 0; r < reps; r++)
        {
            glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, get_rbo(r));
            glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_RGBA32F_ARB,
                                     rbo_width, rbo_height);
        }
    }
    RenderHelperGeometry::begin_object(skeleton, lod, z, rbo_width * rbo_height);

    /* Prepare for render-to-vertex-buffer */
    glPushAttrib(GL_VIEWPORT_BIT);
    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, klass.fbo);
    glDisable(GL_DEPTH_TEST);
    glViewport(0, 0, level.rbo_width, level.rbo_height);
    glClear(GL_COLOR_BUFFER_BIT);
}

void RenderHelperGeometryRTVB::restore_object(const Skeleton &skeleton, double lod, double z)
{
    int ilod = rmesh.int_lod(lod);
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    for (int r = 0; r < reps; r++)
    {
        glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, get_buffer(r));
        glReadBuffer(GL_COLOR_ATTACHMENT0_EXT + r);
        glReadPixels(0, 0, rtvb_levels[ilod].rbo_width, rtvb_levels[ilod].rbo_height, GL_RGBA, GL_FLOAT, NULL);
    }
    glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, 0);

    glEnable(GL_DEPTH_TEST);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glPopAttrib();
    glPopClientAttrib();
    cgGLBindProgram(get_class().get_vertex_program());
    cgGLBindProgram(get_class().get_fragment_program());
}

void RenderHelperGeometryRTVB::get_rbo_size(int level, GLint &rbo_width, GLint &rbo_height) const
{
    g_assert(level >= 0 && level < (int) rtvb_levels.size());
    rbo_width = rtvb_levels[level].rbo_width;
    rbo_height = rtvb_levels[level].rbo_height;
}

void RenderHelperGeometryRTVB::set_rbo_size(int level, GLint rbo_width, GLint rbo_height)
{
    g_assert(level >= 0 && level < (int) rtvb_levels.size());
    rtvb_levels[level].rbo_width = rbo_width;
    rtvb_levels[level].rbo_height = rbo_height;
}

GLuint RenderHelperGeometryRTVB::update_transform_texture(const Skeleton &skeleton)
{
    RenderHelperClassGeometryRTVB &klass = get_class_geometry_rtvb();
    return klass.transform_texture.update(skeleton);
}
