#if HAVE_CONFIG_H
# include <config.h>
#endif
#ifndef _XOPEN_SOURCE
# define _XOPEN_SOURCE 600
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <cstddef>
#include <glib.h>
#include <string>
#include <cstdlib>
#include <xmmintrin.h>
#include "render/interfaces.h"
#include "render/geometry.h"
#include "render/geometry_sse.h"

using namespace std;

class RenderHelperGeometrySSE : public RenderHelperGeometry
{
private:
    struct Channel
    {
        __m128 *vertex_data[2]; // base, offset
        GLubyte *bone_data;
        GLubyte *counts;
    };
    struct Level
    {
        std::vector<Channel> channels;  // v, t0, t1
    };

    std::vector<Level> levels;
public:
    RenderHelperGeometrySSE(RenderHelperClassGeometrySSE &parent, const RenderMesh &m);
    virtual ~RenderHelperGeometrySSE();

    virtual void begin_object(const Skeleton &skeleton, double lod, double z);
};


RenderHelper *RenderHelperClassGeometrySSE::new_instance(const RenderMesh &m)
{
    return new RenderHelperGeometrySSE(*this, m);
}


RenderHelperGeometrySSE::RenderHelperGeometrySSE(RenderHelperClassGeometrySSE &parent, const RenderMesh &m)
: RenderHelperGeometry(parent, m)
{
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    levels.resize(m.levels.size());
    for (size_t i = 0; i < m.levels.size(); i++)
    {
        const RenderMeshLevel &rl = m.levels[i];
        Level &l = levels[i];

        l.channels.resize(reps);
        for (int r = 0; r < reps; r++)
        {
            Channel &ch = l.channels[r];

            size_t total_influences = 0;
            size_t num_vertices = rl.vertices.size();
            for (size_t i = 0; i < num_vertices; i++)
            {
                VectorPosition test = get_channel(rl.vertices[i].second, r) - get_channel(rl.vertices[i].first, r);
                total_influences += distance(test.begin(), test.end());
            }

            ch.counts = new GLubyte[num_vertices];
            ch.bone_data = new GLubyte[total_influences];
            posix_memalign((void **) &ch.vertex_data[0], 16, total_influences * sizeof(__m128));
            posix_memalign((void **) &ch.vertex_data[1], 16, total_influences * sizeof(__m128));

            size_t pos = 0;
            for (size_t v = 0; v < num_vertices; v++)
            {
                const VectorPosition &vbase = get_channel(rl.vertices[v].first, r);
                const VectorPosition &vtarget = get_channel(rl.vertices[v].second, r);
                VectorPosition offset = vtarget - vbase;

                ch.counts[v] = 0;
                for (VectorPosition::const_iterator i = offset.begin(); i != offset.end(); i++)
                {
                    union
                    {
                        __m128 sse[2];
                        float scalar[2][4];
                    } cur;

                    for (int j = 0; j < 4; j++)
                    {
                        cur.scalar[0][j] = vbase[i->index + j];
                        cur.scalar[1][j] = offset[i->index + j];
                    }
                    ch.vertex_data[0][pos] = cur.sse[0];
                    ch.vertex_data[1][pos] = cur.sse[1];
                    ch.bone_data[pos] = i->index / 4;
                    ch.counts[v]++;
                    pos++;
                }
            }
        }
    }
}

/* Takes a 4x4 matrix in column-major form (one column per __m128) and
 * transforms v.
 */
static inline __m128 matrix_transform_sse(const __m128 m[],
                                          const __m128 &v)
{
    __m128 out, vx;
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(0, 0, 0, 0));
    out = _mm_mul_ps(m[0], vx);
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(1, 1, 1, 1));
    out = _mm_add_ps(out, _mm_mul_ps(m[1], vx));
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(2, 2, 2, 2));
    out = _mm_add_ps(out, _mm_mul_ps(m[2], vx));
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(3, 3, 3, 3));
    out = _mm_add_ps(out, _mm_mul_ps(m[3], vx));
    return out;
}

/* Transforms a pair, applying a suitable flod blend */
static inline __m128 transform_pair_sse(const __m128 &a,
                                        const __m128 &b, const __m128 m[],
                                        const __m128 &flod)
{
    __m128 v = _mm_add_ps(a, _mm_mul_ps(b, flod));
    return matrix_transform_sse(m, v);
}

RenderHelperGeometrySSE::~RenderHelperGeometrySSE()
{
    for (size_t i = 0; i < levels.size(); i++)
        for (size_t j = 0; j < levels[i].channels.size(); j++)
        {
            free(levels[i].channels[j].vertex_data[0]);
            free(levels[i].channels[j].vertex_data[1]);
            delete [] levels[i].channels[j].bone_data;
            delete [] levels[i].channels[j].counts;
        }
}

void RenderHelperGeometrySSE::begin_object(const Skeleton &skeleton, double lod, double z)
{
    RenderHelperGeometry::begin_object(skeleton, lod, z);

    int ilod = rmesh.int_lod(lod);
    size_t num_vertices = rmesh.levels[ilod].vertices.size();
    int bones = skeleton.size();
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    __m128 transforms[bones][4];
    __m128 flod = _mm_set1_ps(CLAMP(lod - ilod, 0.0, 1.0));
    for (int i = 0; i < bones; i++)
    {
        const Matrix4 &cm = skeleton.get_absolute(i);
        for (int j = 0; j < 4; j++)
            transforms[i][j] = _mm_setr_ps(cm[0][j], cm[1][j], cm[2][j], cm[3][j]);
    }

    for (int r = 0; r < reps; r++)
    {
        const Channel &ch = levels[ilod].channels[r];
        const __m128 *base = ch.vertex_data[0];
        const __m128 *offset = ch.vertex_data[1];
        float *ptr;

        glBindBuffer(GL_ARRAY_BUFFER, get_buffer(r));
        ptr = begin_buffer();
        int pos = 0;
        for (size_t i = 0; i < num_vertices; i++)
        {
            __m128 cur = _mm_setzero_ps();
            for (int j = ch.counts[i]; j != 0; j--, pos++)
                cur = _mm_add_ps(cur, transform_pair_sse(base[pos],
                                                         offset[pos],
                                                         transforms[ch.bone_data[pos]],
                                                         flod));
            _mm_stream_ps(ptr, cur);
            ptr += 4;
        }

        end_buffer(num_vertices);
    }
}
