#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <cstddef>
#include <glib.h>
#include <string>
#include <cstdlib>
#include <vector>
#include <sstream>
#include "render/interfaces.h"
#include "render/geometry.h"
#include "render/geometry_texture.h"

using namespace std;

/* The three channels are kept in corresponding textures with shared bone
 * data. This is in contrast to the SSE and RTVB shaders, because in this
 * case everything happens in one pass so we can save on computations by
 * sharing bone data.
 */

class RenderHelperGeometryTexture : public RenderHelperGeometryBase
{
private:
    struct Level
    {
        GLuint influence_texture[3];  // pos, tan0, tan1
        GLuint bone_texture;
        GLuint offset_buffer;         // x,y = bone+base; x,z = offset, w = final x
        GLuint texcoord_buffer;
        GLint width;                  // width of the texture maps
    };
    vector<Level> levels;

public:
    RenderHelperGeometryTexture(RenderHelperClassGeometryTexture &parent, const RenderMesh &m);
    virtual ~RenderHelperGeometryTexture();

    virtual void begin_object(const Skeleton &skeleton, double lod, double z);
};


CGparameter RenderHelperClassGeometryTexture::begin_vertex_program(CGprogram vp)
{
    CGparameter cg_class = RenderHelperClass::begin_vertex_program(vp);

    CGparameter ipa = cgGetNamedParameter(vp, "influences");
    int reps = cgGetArraySize(ipa, 0);
    for (int c = 0; c < reps; c++)
        influence_parameter[c] = cgGetArrayParameter(ipa, c);
    bone_parameter = cgGetNamedParameter(vp, "bones");
    offset_parameter = cgGetNamedParameter(vp, "i_offsets");

    cgSetParameterSemantic(bone_parameter, "TEXUNIT8");
    for (int c = 0; c < reps; c++)
    {
        ostringstream sname;
        sname << "TEXUNIT" << 9 + c;
        cgSetParameterSemantic(influence_parameter[c], sname.str().c_str());
    }

    return cg_class;
}

void RenderHelperClassGeometryTexture::begin_state(CGprogram vp, CGprogram fp, int flags)
{
    RenderHelperClass::begin_state(vp, fp, flags);

    cgGLEnableClientState(offset_parameter);
    if (flags & RENDER_FLAG_TEXCOORDS)
    {
        texcoord_parameter = cgGetNamedParameter(vp, "i_texcoords");
        cgGLEnableClientState(texcoord_parameter);
    }

    transform_parameter = cgGetNamedParameter(vp, "u_transforms");
    flod_parameter = cgGetNamedParameter(vp, "u_flod");
    xstep_parameter = cgGetNamedParameter(vp, "u_xstep");
}

void RenderHelperClassGeometryTexture::end_state()
{
    cgGLDisableClientState(offset_parameter);
    if (global_flags() & RENDER_FLAG_TEXCOORDS)
        cgGLDisableClientState(texcoord_parameter);
}

RenderHelper *RenderHelperClassGeometryTexture::new_instance(const RenderMesh &m)
{
    return new RenderHelperGeometryTexture(*this, m);
}

RenderHelperGeometryTexture::RenderHelperGeometryTexture(RenderHelperClassGeometryTexture &parent, const RenderMesh &m)
: RenderHelperGeometryBase(parent, m)
{
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;
    GLint old;
    glGetIntegerv(GL_TEXTURE_BINDING_2D, &old);

    levels.resize(m.levels.size());
    for (size_t i = 0; i < m.levels.size(); i++)
    {
        const RenderMeshLevel &rl = m.levels[i];
        Level &l = levels[i];
        size_t num_vertices = rl.vertices.size();

        GLint width = 0, x = 0, y = 0;
        glGetIntegerv(GL_MAX_TEXTURE_SIZE, &width);
        for (size_t j = 0; j < num_vertices; j++)
        {
            VectorPosition test = influences(rl.vertices[j]);
            int inf = distance(test.begin(), test.end());
            g_assert(inf <= width);
            x += inf;
            if (x > width)
            {
                x = inf;
                y++;
            }
        }
        GLint height = y + 1;
        if (height == 1) width = x;

        GLfloat *bone_data = new GLfloat[width * height];
        memset(bone_data, 0, width * height * sizeof(GLfloat));
        GLfloat (*offset_data)[4] = new GLfloat[num_vertices][4];
        GLfloat (*texcoord_data)[4] = NULL;
        GLfloat (*influence_data[3])[4];
        for (int c = 0; c < reps; c++)
        {
            influence_data[c] = new GLfloat[width * height * 2][4];
            memset(influence_data[c], 0, width * height * 8 * sizeof(GLfloat));
        }
        if (parent.global_flags() & RENDER_FLAG_TEXCOORDS)
            texcoord_data = new GLfloat[num_vertices][4];

        y = 0;
        x = 0;
        for (size_t j = 0; j < num_vertices; j++)
        {
            VectorPosition test = influences(rl.vertices[j]);
            int inf = distance(test.begin(), test.end());
            if (x + inf > width)
            {
                y++;
                x = 0;
            }
            int addr1 = y * width + x;
            int addr2 = y * 2 * width + x;
            for (VectorPosition::const_iterator k = test.begin(); k != test.end(); k++)
            {
                bone_data[addr1] = k->index / 4;
                for (int c = 0; c < reps; c++)
                    for (int l = 0; l < 4; l++)
                    {
                        influence_data[c][addr2][l] = get_channel(rl.vertices[j].first, c)[k->index + l];
                        influence_data[c][addr2 + width][l] = get_channel(rl.vertices[j].second, c)[k->index + l] - influence_data[c][addr2][l];
                    }
                addr1++;
                addr2++;
            }
            offset_data[j][0] = (x + 0.5) / (GLfloat) width;
            offset_data[j][1] = (y + 0.25) / (GLfloat) height;
            offset_data[j][2] = (y + 0.75) / (GLfloat) height;
            offset_data[j][3] = (x + inf) / (GLfloat) width;

            if (parent.global_flags() & RENDER_FLAG_TEXCOORDS)
            {
                texcoord_data[j][0] = rl.vertices[j].first->data.texcoords[0];
                texcoord_data[j][1] = rl.vertices[j].first->data.texcoords[1];
                texcoord_data[j][2] = rl.vertices[j].second->data.texcoords[0] - texcoord_data[j][0];
                texcoord_data[j][3] = rl.vertices[j].second->data.texcoords[1] - texcoord_data[j][1];
            }

            x += inf;
        }

        glGenTextures(reps, &l.influence_texture[0]);
        for (int c = 0; c < reps; c++)
        {
            glBindTexture(GL_TEXTURE_2D, l.influence_texture[c]);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, 2 * height, 0,
                         GL_RGBA, GL_FLOAT, influence_data[c]);
            delete[] influence_data[c];
        }

        glGenTextures(reps, &l.bone_texture);
        glBindTexture(GL_TEXTURE_2D, l.bone_texture);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE32F_ARB, width, height, 0,
                     GL_LUMINANCE, GL_FLOAT, bone_data);
        delete[] bone_data;

        glGenBuffers(1, &l.offset_buffer);
        glBindBuffer(GL_ARRAY_BUFFER, l.offset_buffer);
        glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(GLfloat) * num_vertices, offset_data, GL_STATIC_DRAW);
        delete[] offset_data;

        if (parent.global_flags() & RENDER_FLAG_TEXCOORDS)
        {
            glGenBuffers(1, &l.texcoord_buffer);
            glBindBuffer(GL_ARRAY_BUFFER, l.texcoord_buffer);
            glBufferData(GL_ARRAY_BUFFER, 4 * sizeof(GLfloat) * num_vertices, texcoord_data, GL_STATIC_DRAW);
            delete[] texcoord_data;
        }

        l.width = width;
    }

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindTexture(GL_TEXTURE_2D, old);
}

void RenderHelperGeometryTexture::begin_object(const Skeleton &skeleton, double lod, double z)
{
    RenderHelperClassGeometryTexture &klass = static_cast<RenderHelperClassGeometryTexture &>(get_class());

    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;
    int ilod = rmesh.int_lod(lod);
    Level l = levels[ilod];
    cgGLSetParameter1f(klass.flod_parameter, lod - ilod);
    cgGLSetParameter1f(klass.xstep_parameter, 1.0 / l.width);

    for (int c = 0; c < reps; c++)
    {
        cgGLSetTextureParameter(klass.influence_parameter[c], l.influence_texture[c]);
        cgGLEnableTextureParameter(klass.influence_parameter[c]);
    }
    cgGLSetTextureParameter(klass.bone_parameter, l.bone_texture);
    cgGLEnableTextureParameter(klass.bone_parameter);
    glBindBuffer(GL_ARRAY_BUFFER, l.offset_buffer);
    cgGLSetParameterPointer(klass.offset_parameter, 4, GL_FLOAT, 0, NULL);
    if (global_flags() & RENDER_FLAG_TEXCOORDS)
    {
        glBindBuffer(GL_ARRAY_BUFFER, l.texcoord_buffer);
        cgGLSetParameterPointer(klass.texcoord_parameter, 4, GL_FLOAT, 0, NULL);
    }

    g_assert(skeleton.size() <= 128);
    double mat_data[128][3][4];
    for (size_t i = 0; i < skeleton.size(); i++)
    {
        const Matrix4 cur = skeleton.get_absolute(i);
        for (int j = 0; j < 3; j++)
            for (int k = 0; k < 4; k++)
                mat_data[i][j][k] = cur[j][k];
    }
    cgGLSetMatrixParameterArraydr(klass.transform_parameter, 0, skeleton.size(), &mat_data[0][0][0]);
}

RenderHelperGeometryTexture::~RenderHelperGeometryTexture()
{
    int reps = (global_flags() & RENDER_FLAGS_VERTEX_DERIVATIVES) ? 3 : 1;

    for (size_t i = 0; i < levels.size(); i++)
    {
        Level &l = levels[i];
        glDeleteTextures(reps, &l.influence_texture[0]);
        glDeleteTextures(1, &l.bone_texture);
        glDeleteBuffers(1, &l.offset_buffer);
        if (global_flags() & RENDER_FLAG_TEXCOORDS)
            glDeleteBuffers(1, &l.texcoord_buffer);
    }
}
