#ifndef ALOD_RENDER_INTERFACES_H
#define ALOD_RENDER_INTERFACES_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "render/meshrender.h"
#include "render/rendermesh.h"
#include <Cg/cg.h>
#include <GL/gl.h>
#include <string>

//! Enum for the types of render helpers.
/*! In general, there will be
 * - one geometry helper
 * - one normal helper
 * - one lighting helper
 * - an arbitrary number of shader helpers.
 */
enum RenderHelperType
{
    RENDER_HELPER_GEOMETRY,     // produces vertex data
    RENDER_HELPER_NORMAL,       // produces a per-fragment normal
    RENDER_HELPER_LIGHTING,     // converts normal into light level
    RENDER_HELPER_SHADER,       // produces colours for modulation
    RENDER_HELPER_OTHER         // non-shading visualisations
};

class RenderHelper;
class RenderMesh;

//! Returns an instantiation of \a cg_class from \a program, or NULL
CGparameter create_cg_parameter(CGprogram program, const std::string &cg_class);

//! A generic render helper, not tied to a specific mesh.
/*! In general there will be one instance of each subclass, although
 * for a parametric subclass there may be multiple instances. It is
 * responsible for holding the GPU programs, as well as other shared data
 * such as a lighting cube-map.
 */
class RenderHelperClass
{
protected:
    CGprogram vertex_program, fragment_program;
    int _global_flags;

    //! Returns the class name used for \c begin_vertex_program
    virtual std::string vertex_program_class() const { return std::string(); }
    //! Returns the class name used for \c begin_fragment_program
    virtual std::string fragment_program_class() const { return std::string(); }
public:
    //! Returns the type of this class
    virtual RenderHelperType type() const = 0;

    //! Returns the values expected from the geometry converter
    virtual int flags() const { return 0; }
    //! Returns the flags unioned across all helpers
    int global_flags() const { return _global_flags; }

    //! Instantiate a new instance of the class
    virtual RenderHelper *new_instance(const RenderMesh &m) = 0;

    //! Tie in vertex program
    /*! This is called before begin_state, since begin_state is only called
     * after the main programs are compiled. It returns either a parameter
     * which is an implementation of the appropriate interface, or NULL if
     * is does not wish to connect anything.
     */
    virtual CGparameter begin_vertex_program(CGprogram vertex_program)
    {
        return create_cg_parameter(vertex_program, vertex_program_class());
    }

    //! Tie in fragment program
    /*! This is called before begin_state, since begin_state is only called
     * after the main programs are compiled. It returns either a parameter
     * which is an implementation of the appropriate interface, or NULL if
     * is does not wish to connect anything.
     */
    virtual CGparameter begin_fragment_program(CGprogram fragment_program)
    {
        return create_cg_parameter(fragment_program, fragment_program_class());
    }

    //! Set up state that will exist for the lifetime of this state configuration
    /* All vertex attribute state should be done with generic attributes
     * via Cg. The geometry helper should set the POSITION semantic for the
     * position, but otherwise Cg should be allowed to do allocations.
     *
     * \param vp the global vertex program
     * \param fp the global fragment program
     * \param flags the union of all the flags
     */
    virtual void begin_state(CGprogram vp, CGprogram fp, int flags)
    { vertex_program = vp; fragment_program = fp; _global_flags = flags; }

    //! Clean up after begin_state
    virtual void end_state() {}

    //! Called after begin_state to pass handles to the connected parameter instances
    virtual void set_parameters(CGparameter vparameter, CGparameter fparameter) {}

    //! Called after begin_state and set_parameters to control the light parameters
    virtual void set_light(float diffuse, float ambient, const Vector3 &direction) {}

    CGprogram get_vertex_program() { return vertex_program; }
    CGprogram get_fragment_program() { return fragment_program; }

    virtual ~RenderHelperClass() {}
};

//! Base class for rendering plugins
class RenderHelper
{
private:
    RenderHelperClass &klass;

protected:
    const RenderMesh &rmesh;

public:
    RenderHelper(RenderHelperClass &parent, const RenderMesh &m) : klass(parent), rmesh(m) {}

    RenderHelperClass &get_class() { return klass; }
    int global_flags() { return klass.global_flags(); }

    //! Do per-instance setup (called once per frame)
    /* \a flags is set to the union of the flags of all the helpers. It is
     * intended to allow geometry helpers to know whether tangents and/or
     * normals need to be produced.
     */
    virtual void begin_object(const Skeleton &skeleton, double lod, double z) {}
    //! Reverse effects of \c beginObject
    virtual void end_object() {}

    virtual ~RenderHelper() {}
};

#endif /* !ALOD_RENDER_INTERFACES_H */
