#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include "render/interfaces.h"
#include "render/meshrender.h"
#include "render/lighting_cube.h"
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <algorithm>

using namespace std;

RenderHelper *RenderHelperClassLightingCube::new_instance(const RenderMesh &m)
{
    return new RenderHelper(*this, m);
}

void RenderHelperClassLightingCube::set_parameters(CGparameter vparameter, CGparameter fparameter)
{
    glGenTextures(1, &lighting_texture);
    glBindTexture(GL_TEXTURE_CUBE_MAP, lighting_texture);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_TRUE);

    /* Allocate storage; the actual data is provided by set_light */
    for (int i = 0; i < 6; i++)
        glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_LUMINANCE8, size, size, 0,
                     GL_LUMINANCE, GL_FLOAT, NULL);

    CGparameter tex_parameter = cgGetNamedStructParameter(fparameter, "lighting_texture");
    check_cg_error();
    g_assert(tex_parameter);
    cgGLSetTextureParameter(tex_parameter, lighting_texture);
    cgGLEnableTextureParameter(tex_parameter);
    check_cg_error();
}

void RenderHelperClassLightingCube::set_light(float diffuse, float ambient, const Vector3 &direction)
{
    /* The values for each face are xmap, ymap, other. xmap and ymap
     * are the axes corresponding to X and Y in the texture image
     * (1=X, 2=Y, 3=Z, negated for inverses). other is the value of
     * the third coordinate.
     */
    static const int faces[6][3] =
    {
        { -3, -2,  1 },
        {  3, -2, -1 },
        {  1,  3,  1 },
        {  1, -3, -1 },
        {  1, -2,  1 },
        { -1, -2, -1 }
    };
    GLint old_tex;

    glGetIntegerv(GL_TEXTURE_BINDING_CUBE_MAP, &old_tex);
    glBindTexture(GL_TEXTURE_CUBE_MAP, lighting_texture);
    double scale = 2.0 / size;
    double bias = -1.0 + 1.0 / size;
    GLfloat data[size][size];

    for (int i = 0; i < 6; i++)
    {
        for (int j = 0; j < size; j++)
            for (int k = 0; k < size; k++)
            {
                Vector3 view;
                int ax1 = abs(faces[i][0]) - 1;
                int ax2 = abs(faces[i][1]) - 1;
                int ax3 = 3 - ax1 - ax2;
                view[ax1] = j * scale + bias;
                if (faces[i][0] < 0) view[ax1] = -view[ax1];
                view[ax2] = k * scale + bias;
                if (faces[i][1] < 0) view[ax2] = -view[ax2];
                view[ax3] = faces[i][2];
                view = normalise(view);
                data[k][j] = max(0.0, view * direction) * diffuse
                    + max(0.0, view[2]) * diffuse + ambient;
            }
        glTexSubImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0,
                        0, 0, size, size, GL_LUMINANCE, GL_FLOAT, data);
    }
    glBindTexture(GL_TEXTURE_CUBE_MAP, old_tex);
}

void RenderHelperClassLightingCube::end_state()
{
    glDeleteTextures(1, &lighting_texture);
}
