#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include "render/interfaces.h"
#include "render/factory.h"
#include "mesh/mesh.h"
#include "geom/coords.h"
#include "geom/skeleton.h"
#include "common/paths.h"
#include "common/options.h"
#include "common/logging.h"
#include "common/timing.h"
#include "alodgl/glmisc.h"
#include "formats/formats.h"
#include <glib.h>
#include <gtk/gtk.h>
#include <gtk/gtkgl.h>
#include <gdk/gdkglglext.h>
#include <gdk/gdkkeysyms.h>
#ifdef G_OS_UNIX
# include <gdk/gdkx.h>
#endif
#include <vector>
#include <string>
#include <utility>
#include <sstream>
#include <libxml/parser.h>
#include <libxml/xmlmemory.h>

#define MAX_TOLERANCE 50

using namespace std;

static CGcontext cg_context;
static GdkCursor *blank_cursor;
static AlodProgramOption in_format, verbose, debug;
static AlodProgramOption camerapath, draw_width, draw_height;
static AlodProgramOption initial_lod, initial_shaders;
static AlodProgramOption show_influences, show_lod, white_bg;
static Timer render_timer;

/*** Utility functions ***/

CGcontext global_cg_context() { return cg_context; }
CGprofile global_cg_vertex_profile() { return CG_PROFILE_VP40; }
CGprofile global_cg_fragment_profile() { return CG_PROFILE_FP40; }

void check_cg_error()
{
    CGerror err = cgGetError();
    if (err != CG_NO_ERROR)
    {
        const char *details = cgGetLastListing(global_cg_context());
        if (details && details[0])
            g_error("CG error: %s\nDetails:\n%s", cgGetErrorString(err), details);
        else
            g_error("CG error: %s", cgGetErrorString(err));
    }
}

CGprogram load_cg_program(const string &source_file, CGprofile profile,
                          const char *args[])
{
    CGprogram program;

    string file = resource_path("ALOD_SHADERS", PKGDATADIR, "meshrender/" + source_file);
    program = cgCreateProgramFromFile(global_cg_context(), CG_SOURCE, file.c_str(),
                                      profile, NULL, args);
    check_cg_error();
    g_assert(program != NULL);
    return program;
}

void TextureManager::bind_texture(CGparameter param, GLuint texture)
{
    static CGtype type2D = cgGetType("sampler2D");
    static CGtype typeRECT = cgGetType("samplerRECT");
    CGtype type;
    GLint unit;
    save old;
    GLenum binding;

    glGetIntegerv(GL_ACTIVE_TEXTURE, &unit);
    old.unit = cgGLGetTextureEnum(param);
    type = cgGetParameterType(param);
    if (old.unit != (GLenum) unit)
        glActiveTexture(old.unit);
    if (type == type2D)
    {
        old.target = GL_TEXTURE_2D;
        binding = GL_TEXTURE_BINDING_2D;
    }
    else if (type == typeRECT)
    {
        old.target = GL_TEXTURE_RECTANGLE_ARB;
        binding = GL_TEXTURE_BINDING_RECTANGLE_ARB;
    }
    else
        g_assert_not_reached();
    glGetIntegerv(binding, &old.texture);
    if (old.unit != (GLenum) unit)
        glActiveTexture(unit);
    saved.push_back(old);

    cgGLSetTextureParameter(param, texture);
    cgGLEnableTextureParameter(param);
}

void TextureManager::reset()
{
    GLint unit;
    glGetIntegerv(GL_ACTIVE_TEXTURE, &unit);

    for (vector<save>::const_reverse_iterator i = saved.rbegin(); i != saved.rend(); i++)
    {
        glActiveTexture(i->unit);
        glBindTexture(i->target, i->texture);
    }
    glActiveTexture(unit);
    saved.clear();
}

void load_simple_cg_programs(CGprogram &vp, CGprogram &fp)
{
    static CGprogram _vp = NULL;
    static CGprogram _fp = NULL;

    if (!_vp)
    {
        _vp = load_cg_program("simple_vp.cg", global_cg_vertex_profile(), NULL);
        _fp = load_cg_program("simple_fp.cg", global_cg_fragment_profile(), NULL);
        cgCompileProgram(_vp);
        cgGLLoadProgram(_vp);
        check_cg_error();
        cgCompileProgram(_fp);
        cgGLLoadProgram(_fp);
        check_cg_error();
    }
    vp = _vp;
    fp = _fp;
}

GLuint sphere_list()
{
    static GLuint list = 0;
    int sub = 50;

    if (!list)
    {
        list = glGenLists(1);
        glNewList(list, GL_COMPILE);
        glBegin(GL_QUADS);
        for (int i = 0; i < sub; i++)
            for (int j = 0; j < sub; j++)
            {
                double x[2], y[2], z[2], s[2];
                int i2 = (i + 1) % sub;

                s[0] = sin(j * G_PI / sub);
                s[1] = sin((j + 1) * G_PI / sub);

                x[0] = cos(i * G_PI * 2.0 / sub);
                x[1] = cos(i2 * G_PI * 2.0 / sub);
                y[0] = sin(i * G_PI * 2.0 / sub);
                y[1] = sin(i2 * G_PI * 2.0 / sub);
                z[0] = cos(j * G_PI / sub);
                z[1] = cos((j + 1) * G_PI / sub);

                Vector3 v[4] =
                {
                    Vector3(x[0] * s[0], y[0] * s[0], z[0]),
                    Vector3(x[1] * s[0], y[1] * s[0], z[0]),
                    Vector3(x[1] * s[1], y[1] * s[1], z[1]),
                    Vector3(x[0] * s[1], y[0] * s[1], z[1])
                };
                for (int k = 0; k < 4; k++)
                    glVertex3f(v[k][0], v[k][1], v[k][2]);
            }
        glEnd();
        glEndList();
    }
    return list;
}

static void check_extension(GLboolean present, const char *name)
{
    if (!present)
        g_error("meshrender requires %s", name);
}

static gboolean redraw(GtkWidget *widget)
{
    if (widget->window)
    {
        gdk_window_invalidate_rect(widget->window,
                                   &widget->allocation,
                                   FALSE);
    }
    return TRUE;
}

static void create_blank_cursor()
{
    static const gchar bits[] = {0};
    static const gchar mask_bits[] = {0};
    GdkColor dummy = {0, 0, 0, 0};
    GdkPixmap *source, *mask;

    source = gdk_bitmap_create_from_data(NULL, bits, 1, 1);
    mask = gdk_bitmap_create_from_data(NULL, mask_bits, 1, 1);
    blank_cursor = gdk_cursor_new_from_pixmap(source, mask, &dummy, &dummy, 0, 0);
    gdk_pixmap_unref(source);
    gdk_pixmap_unref(mask);
}

/*** Cg helper handling ***/

CGparameter create_cg_parameter(CGprogram program, const string &cg_class)
{
    CGtype type;
    CGparameter parameter, child;

    if (cg_class.empty()) return NULL;

    type = cgGetNamedUserType(program, cg_class.c_str());
    if (type == CG_UNKNOWN_TYPE)
        g_error("CG type %s is not defined", cg_class.c_str());

    parameter = cgCreateParameter(cgGetProgramContext(program), type);
    check_cg_error();
    g_assert(parameter);
    /* Set all fields to literal, except for uniforms which cannot be */
    for (child = cgGetFirstStructParameter(parameter); child; child = cgGetNextParameter(child))
    {
        switch (cgGetParameterClass(child))
        {
        case CG_PARAMETERCLASS_SCALAR:
        case CG_PARAMETERCLASS_VECTOR:
        case CG_PARAMETERCLASS_MATRIX:
            cgSetParameterVariability(child, CG_LITERAL);
            check_cg_error();
        default:;
        }
    }
    return parameter;
}

/*** GTK handling */

#define VIEWER_WRAP(function) \
    static void viewer_ ## function(GtkWidget *widget, gpointer user_data) \
    { static_cast<RenderViewer *>(user_data)->function(widget); }
#define VIEWER_WRAP_EVENT(function, eventtype) \
    static gboolean viewer_ ## function(GtkWidget *widget, eventtype *event, gpointer user_data) \
    { return static_cast<RenderViewer *>(user_data)->function(widget, event); }

VIEWER_WRAP(realize)
VIEWER_WRAP_EVENT(configure_event, GdkEventConfigure)
VIEWER_WRAP_EVENT(expose_event, GdkEventExpose)
VIEWER_WRAP_EVENT(button_event, GdkEventButton)
VIEWER_WRAP_EVENT(motion_notify_event, GdkEventMotion)
VIEWER_WRAP_EVENT(key_event, GdkEventKey)
static void viewer_notebook_switch_page(GtkWidget *notebook, GtkWidget *notebook_page, gint page, gpointer user_data)
{
    static_cast<RenderViewer *>(user_data)->notebook_switch_page(notebook, notebook_page, page);
}
static void viewer_toggle_helper(GtkToggleAction *action, gpointer user_data)
{
    static_cast<RenderViewer *>(user_data)->toggle_helper(action);
}

void RenderViewer::update_camera(double elapsed)
{
    Matrix4 &view = camera.view_matrix;
    if (camerapath_frame == -1)  // free camera
    {
        double dy = 0.0;
        double dx = 0.0;

        if (keys_down.count(GDK_Up)) dy += 2.0 * elapsed;
        if (keys_down.count(GDK_Down)) dy -= 2.0 * elapsed;
        if (keys_down.count(GDK_Left)) dx += 2.0 * elapsed;
        if (keys_down.count(GDK_Right)) dx -= 2.0 * elapsed;

        view = matrix_rotate(camera.elevation - G_PI * 0.5, Vector3(1.0, 0.0, 0.0));
        view *= matrix_rotate(camera.azimuth, Vector3(0.0, 0.0, 1.0));

        camera.position -= dx * Vector3(view[0][0], view[0][1], view[0][2]);
        camera.position -= dy * Vector3(view[2][0], view[2][1], view[2][2]);

        view *= matrix_translate(-camera.position);
    }
    else
    {
        if (camerapath_frame == 5)
            render_timer.reset();
        view = camerapath_frames[camerapath_frame++];
    }

    // Update frustum
    Matrix4 vp = camera.projection_matrix * view;

    for (int i = 0; i < 3; i++)
    {
        camera.frustum[i * 2 + 0].normal = vp[3] - vp[i];
        camera.frustum[i * 2 + 1].normal = vp[3] + vp[i];
    }
    for (int i = 0; i < 6; i++)
        camera.frustum[i].normal /= length(vector4_to_vector3(camera.frustum[i].normal));
}

void RenderViewer::update_animation(double elapsed)
{
    if (camerapath_frame != -1)
    {
        elapsed = 0.0;
        advance_frames = 1;
    }
    for (list<RenderScene>::iterator scene = scenes.begin();
         scene != scenes.end(); scene++)
        for (list<RenderInstance>::iterator i = scene->instances.begin();
             i != scene->instances.end(); i++)
            i->update_animation(elapsed, advance_frames);
    advance_frames = 0;
}

void RenderViewer::reset_animation()
{
    for (list<RenderScene>::iterator scene = scenes.begin();
         scene != scenes.end(); scene++)
        for (list<RenderInstance>::iterator i = scene->instances.begin();
             i != scene->instances.end(); i++)
            i->reset_animation();
    advance_frames = 0;
}

void RenderViewer::begin_capture(guint32 time)
{
    if (capturing) return;
    capturing = true;
    capture_first = true;

    GdkGrabStatus status;
    status = gdk_pointer_grab(draw->window,
                              FALSE,
                              (GdkEventMask) (GDK_POINTER_MOTION_MASK
                                              | GDK_BUTTON1_MOTION_MASK
                                              | GDK_BUTTON_PRESS_MASK
                                              | GDK_BUTTON_RELEASE_MASK),
                              draw->window,
                              blank_cursor,
                              time);
    if (status != GDK_GRAB_SUCCESS)
        g_warning("grab failed; status was %d", (int) status);
}

void RenderViewer::end_capture(guint32 time)
{
    if (!capturing) return;
    capturing = false;

    gdk_pointer_ungrab(time);
}

void RenderViewer::realize(GtkWidget *widget)
{
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext)) return;

    cgGLSetOptimalOptions(global_cg_vertex_profile());
    cgGLSetOptimalOptions(global_cg_fragment_profile());
    rc.begin_state();

    check_extension(GLEE_VERSION_1_5, "OpenGL 1.5");
    check_extension(GLEE_ARB_vertex_program, "GL_ARB_vertex_program");
    check_extension(GLEE_ARB_fragment_program, "GL_ARB_fragment_program");
    check_extension(GLEE_NV_texture_expand_normal, "GL_NV_texture_expand_normal");

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    if (white_bg.used)
        glClearColor(1.0, 1.0, 1.0, 0.0);

    gdk_gl_drawable_gl_end(gldrawable);
}

gboolean RenderViewer::configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext)) return FALSE;

    glViewport(0, 0, widget->allocation.width, widget->allocation.height);
    GLdouble w = widget->allocation.width;
    GLdouble h = widget->allocation.height;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0, w / h, 0.01, 100.0);
    camera.projection_matrix = alodgl_get_matrix(GL_PROJECTION_MATRIX);
    glMatrixMode(GL_MODELVIEW);

    double theta = w / h * 60.0 * 0.5 * G_PI / 180.0;
    camera.lod_tolerance_base = 2.0 * sin(theta) / w;

    gdk_gl_drawable_gl_end(gldrawable);
    return TRUE;
}

gboolean RenderViewer::expose_event(GtkWidget *widget, GdkEventExpose *event)
{
    size_t influences = 0;
    RenderScene &scene = *current_scene;
    GdkGLContext *glcontext = gtk_widget_get_gl_context(widget);
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(widget);
    if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext)) return FALSE;

    GTimeVal now;
    double elapsed;
    g_get_current_time(&now);
    elapsed = (now.tv_sec - last_update.tv_sec)
        + 1e-6 * (now.tv_usec - last_update.tv_usec);
    update_camera(elapsed);
    update_animation(elapsed);
    last_update = now;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    for (list<RenderInstance>::iterator i = scene.instances.begin();
         i != scene.instances.end(); i++)
    {
        if (i->get_object().helpers.empty())
            rc.create_instances(i->get_object());
        influences += i->influences();
        i->render();
    }
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "%lu influences", (unsigned long) influences);

    gdk_gl_drawable_swap_buffers(gldrawable);
    gdk_gl_drawable_gl_end(gldrawable);
    if (camerapath_frame == (int) camerapath_frames.size())
        gtk_main_quit();
    return TRUE;
}

gboolean RenderViewer::motion_notify_event(GtkWidget *widget, GdkEventMotion *event)
{
    if (capturing)
    {
        gint centrex = widget->allocation.width / 2;
        gint centrey = widget->allocation.height / 2;
        if (event->x != centrex || event->y != centrey)
        {
            if (capture_first) capture_first = false;
            else
            {
                Vector2 delta(event->x - centrex, event->y - centrey);
                camera.azimuth += delta[0] * 0.01;
                camera.elevation += delta[1] * 0.01;
                camera.elevation = CLAMP(camera.elevation,
                                         -G_PI * 0.5, G_PI * 0.5);
            }

#ifdef G_OS_UNIX
            XWarpPointer(GDK_WINDOW_XDISPLAY(widget->window),
                         None, GDK_WINDOW_XID(widget->window),
                         0, 0, 0, 0,
                         centrex, centrey);
#endif
#ifdef G_OS_WIN32
            gint originx, originy;
            gdk_window_get_origin(widget->window, &originx, &originy);
            SetCursorPos(centrex + originx, centrey + originy);
#endif
        }
        return TRUE;
    }
    return FALSE;
}

gboolean RenderViewer::button_event(GtkWidget *widget, GdkEventButton *event)
{
    if (event->button == 2 && event->type == GDK_BUTTON_PRESS)
    {
        fullscreen = !fullscreen;
        if (fullscreen)
        {
            gtk_widget_hide(menu);
            gtk_widget_hide(notebook);
            gtk_widget_hide(lod_slider);
            gtk_window_fullscreen(GTK_WINDOW(window));
            gtk_window_set_decorated(GTK_WINDOW(window), false);
            begin_capture(event->time);
        }
        else
        {
            end_capture(event->time);
            gtk_window_unfullscreen(GTK_WINDOW(window));
            gtk_window_set_decorated(GTK_WINDOW(window), true);
            gtk_widget_show(menu);
            gtk_widget_show(notebook);
            gtk_widget_show(lod_slider);
        }
        return TRUE;
    }
    else if (event->button == 1)
    {
        if (event->type == GDK_BUTTON_PRESS)
        {
            gtk_widget_grab_focus(draw);
            begin_capture(event->time);
        }
        else
            end_capture(event->time);
        return TRUE;
    }
    return FALSE;
}

gboolean RenderViewer::key_event(GtkWidget *widget, GdkEventKey *event)
{
    switch (event->keyval)
    {
    case GDK_Up:
    case GDK_Down:
    case GDK_Left:
    case GDK_Right:
        if (event->type == GDK_KEY_PRESS)
            keys_down.insert(event->keyval);
        else
            keys_down.erase(event->keyval);
        return TRUE;
    case GDK_KP_Add:
    case GDK_KP_Subtract:
        if (event->type == GDK_KEY_PRESS)
            advance_frames += (event->keyval == GDK_KP_Add ? 1 : -1);
        return TRUE;
    case GDK_F9:
        if (fullscreen && event->type == GDK_KEY_PRESS)
        {
            if (GTK_WIDGET_VISIBLE(menu))
                gtk_widget_hide(menu);
            else
                gtk_widget_show(menu);
            return TRUE;
        }
        return FALSE;
    default:
        return FALSE;
    }
}

void RenderViewer::notebook_switch_page(GtkWidget *notebook,
                                        GtkWidget *notebook_page,
                                        gint page)
{
    current_scene = scenes.begin();
    advance(current_scene, page);
    redraw(draw);
}

void RenderViewer::toggle_option(GtkToggleAction *action, RenderOption option)
{
    bool active = gtk_toggle_action_get_active(action);
    render_options[option] = active;
    redraw(draw);
}

void RenderViewer::rebuild_helpers()
{
    rc.end_state();
    rc = RenderClasses();  // clear out the old

    // invalidate the current instantiations
    for (list<RenderScene>::iterator i = scenes.begin(); i != scenes.end(); i++)
        for (list<RenderObject>::iterator j = i->objects.begin(); j != i->objects.end(); j++)
        {
            for (vector<RenderHelper *>::iterator k = j->helpers.begin(); k != j->helpers.end(); k++)
                delete *k;
            j->helpers.clear();
        }
    // create the new ones
    for (size_t i = 0; i < helper_actions.size(); i++)
    {
        if (gtk_toggle_action_get_active(helper_actions[i]))
            rc.add_helper(factory[i].instantiate());
    }

    GdkGLContext *glcontext = gtk_widget_get_gl_context(draw);
    GdkGLDrawable *gldrawable = gtk_widget_get_gl_drawable(draw);
    if (!gdk_gl_drawable_gl_begin(gldrawable, glcontext))
        g_assert_not_reached();
    rc.begin_state();
}

void RenderViewer::toggle_helper(GtkToggleAction *action)
{
    if (GTK_IS_RADIO_ACTION(action)
        && !gtk_toggle_action_get_active(action))
    {
        /* When changing radio actions, there is a deactivation and an
         * activation event. We only process the activation to avoid the
         * double invalidation.
         */
        return;
    }

    rebuild_helpers();
}

//! Fills in r.bounds and normalise_transform
static void normalise_mesh(RenderObject &r, const Mesh &m)
{
    BoundingSphere bs = mesh_bounding_sphere_3d(m);

    r.bounds.centre = vector3_to_vector4(bs.centre, 1.0);
    r.bounds.radius = bs.radius;
    r.normalise_transform = matrix_scale(1.0 / bs.radius) * matrix_translate(-bs.centre);
}

double RenderViewer::lod_tolerance()
{
    return gtk_range_get_value(GTK_RANGE(lod_slider));
}

/*** File loading code ***/

static double get_float(const xmlChar *value) throw(bad_format)
{
    istringstream in((const char *) value);
    double v;

    in >> v >> ws;
    if (!in || !in.eof() || !isfinite(v))
        throw bad_format(string("expected a floating point value, received \"") + (const char *) value + "\"");
    return v;
}

static int get_int(const xmlChar *value) throw(bad_format)
{
    istringstream in((const char *) value);
    int v;

    in >> v >> ws;
    if (!in || !in.eof())
        throw bad_format(string("expected an integer value, received \"") + (const char *) value + "\"");
    return v;
}

static Vector3 get_triple(const xmlChar *value) throw(bad_format)
{
    istringstream in((const char *) value);
    Vector3 v;

    in >> v[0] >> v[1] >> v[2] >> ws;
    if (!in || !in.eof() || !isfinite(v[0]) || !isfinite(v[1]) || !isfinite(v[2]))
        throw bad_format(string("expected a triple of floating point values, received \"") + (const char *) value + "\"");
    return v;
}

#define TO_XML(x) ((const xmlChar *)(x))

ModelTransform RenderScene::parse_transform(xmlNodePtr cur) throw(bad_format)
{
    xmlChar *prop = NULL;
    ModelTransform ans;

    if (cur->type != XML_ELEMENT_NODE)
        return ModelTransform();
    if (!xmlStrcmp(cur->name, TO_XML("scale")))
    {
        prop = xmlGetProp(cur, TO_XML("value"));
        ans.spatial = matrix_scale(get_float(prop));
    }
    else if (!xmlStrcmp(cur->name, TO_XML("rotate")))
    {
        prop = xmlGetProp(cur, TO_XML("angle"));
        double angle = get_float(prop);
        xmlFree(prop);

        prop = xmlGetProp(cur, TO_XML("axis"));
        ans.spatial = matrix_rotate(angle * M_PI / 180.0, get_triple(prop));
    }
    else if (!xmlStrcmp(cur->name, TO_XML("translate")))
    {
        prop = xmlGetProp(cur, TO_XML("offset"));
        ans.spatial = matrix_translate(get_triple(prop));
    }
    else if (!xmlStrcmp(cur->name, TO_XML("time")))
    {
        prop = xmlGetProp(cur, TO_XML("value"));
        ans.temporal = get_float(prop);
    }
    else
        g_assert_not_reached();

    if (prop) xmlFree(prop);
    return ans;
}

void RenderScene::parse_model(xmlNodePtr cur) throw(bad_format)
{
    ModelTransform transform;
    xmlNodePtr child;
    xmlChar *normalise, *name, *filename;

    child = cur->children;
    while (child)
    {
        transform *= parse_transform(child);
        child = child->next;
    }

    normalise = xmlGetProp(cur, TO_XML("normalise"));
    name = xmlGetProp(cur, TO_XML("name"));
    filename = xmlGetProp(cur, TO_XML("filename"));

    string real_filename = relative_path(this->filename, (const char *) filename);
    try
    {
        Mesh mesh;
        format_load_mesh_file(in_format.string_value, real_filename,
                              mesh, MESH_LEVEL_PATCHES);
        objects.push_back(RenderObject());
        objects.back().load_mesh(mesh);
        objects.back().name = (const char *) name;
        objects.back().mesh_filename = real_filename;
        objects.back().default_transform = transform;
        normalise_mesh(objects.back(), mesh);
        if (!normalise || xmlStrcmp(normalise, TO_XML("no")) != 0)
            objects.back().default_transform.spatial *= objects.back().normalise_transform;
    }
    catch (...)
    {
        xmlFree(normalise);
        xmlFree(name);
        xmlFree(filename);
        throw;
    }
}

void RenderScene::parse_instance(xmlNodePtr cur, const ModelTransform &tree_transform) throw(bad_format)
{
    xmlChar *model_name;
    list<RenderObject>::iterator model;

    model_name = xmlGetProp(cur, TO_XML("model"));
    for (model = objects.begin(); model != objects.end(); model++)
        if (model->name == (const char *) model_name) break;
    if (model == objects.end())
    {
        string error = string("No model named \"") + (const char *) model_name + "\"";
        xmlFree(model_name);
        throw bad_format(error);
    }
    xmlFree(model_name);

    ModelTransform transform = tree_transform;
    xmlNodePtr child = cur->children;
    while (child)
    {
        transform *= parse_transform(child);
        child = child->next;
    }
    transform *= model->default_transform;
    instances.push_back(RenderInstance(model, transform, viewer));
}

/* Handles transformations, instances and higher level nodes. The return is
 * the identity for non-transformations.
 */
ModelTransform RenderScene::parse_tree(xmlNodePtr cur, const ModelTransform &tree_transform) throw(bad_format)
{
    if (!xmlStrcmp(cur->name, TO_XML("instance")))
    {
        parse_instance(cur, tree_transform);
        return ModelTransform();
    }
    else if (!xmlStrcmp(cur->name, TO_XML("repeat")))
    {
        ModelTransform transform = tree_transform;
        xmlChar *prop;
        int c;

        prop = xmlGetProp(cur, TO_XML("count"));
        c = get_int(prop);
        xmlFree(prop);
        for (int i = 0; i < c; i++)
        {
            xmlNodePtr child = cur->children;
            while (child)
            {
                transform *= parse_tree(child, transform);
                child = child->next;
            }
        }
        return ModelTransform();
    }
    else
    {
        /* Assume that everything else is a transformation */
        return parse_transform(cur);
    }
}

RenderScene::RenderScene(RenderViewer *viewer, const string &filename)
: viewer(viewer), filename(filename)
{
}

void RenderScene::load(xmlNodePtr root)
{
    xmlNodePtr child = root->children;

    while (child)
    {
        if (!xmlStrcmp(child->name, TO_XML("model")))
            parse_model(child);
        else
            parse_tree(child, ModelTransform());
        child = child->next;
    }
}

void RenderViewer::add_scene(const char *filename, xmlNodePtr root)
{
    GtkWidget *vbox, *label;
    /* We don't actually put the drawing area inside the notebook,
     * because we use a single drawing area for all the meshes.
     */
    vbox = gtk_vbox_new(FALSE, 0);
    gchar *basename = g_path_get_basename(filename);
    label = gtk_label_new(basename);
    g_free(basename);
    gtk_widget_show(vbox);
    gtk_widget_show(label);
    gtk_notebook_append_page(GTK_NOTEBOOK(notebook),
                             vbox,
                             label);

    bool first_scene = scenes.empty();
    scenes.push_back(RenderScene(this, filename));
    scenes.back().load(root);
    if (first_scene) current_scene = scenes.begin();
}

static xmlParserInputPtr external_entity_loader(const char *url, const char *id,
                                                xmlParserCtxtPtr ctxt)
{
    xmlParserInputPtr ret;

    string filename = resource_path("ALOD_SHADERS", PKGDATADIR, url);
    ret = xmlNoNetExternalEntityLoader(filename.c_str(), id, ctxt);
    if (ret) return ret;

    /* Oh well... */
    if (url)
        g_warning("failed to load external entity \"%s\"\n", url);
    else if (id)
        g_warning("failed to load external entity \"%s\"\n", id);
    return NULL;
}

void RenderViewer::load_scene(const string &scene_filename)
{
    xmlDocPtr doc;
    xmlDtdPtr dtd;
    xmlValidCtxtPtr valid;
    const gchar *name = scene_filename.c_str();

    xmlSetExternalEntityLoader(external_entity_loader);
    doc = xmlParseFile(name);
    if (!doc)
        g_error("Failed to load scene \"%s\"", name);
    dtd = xmlParseDTD(NULL, TO_XML("scene.dtd"));
    if (!dtd)
        g_error("Failed to load the scene DTD");
    valid = xmlNewValidCtxt();
    if (!xmlValidateDtd(valid, doc, dtd))
        g_error("Scene \"%s\" does not conform to the DTD", name);
    xmlFreeValidCtxt(valid);
    xmlFreeDtd(dtd);

    add_scene(name, xmlDocGetRootElement(doc));

    xmlFreeDoc(doc);
}

void RenderViewer::load_camerapath(const string &camerapath_filename)
{
    xmlDocPtr doc;
    xmlDtdPtr dtd;
    xmlValidCtxtPtr valid;
    const gchar *name = camerapath_filename.c_str();

    xmlSetExternalEntityLoader(external_entity_loader);
    doc = xmlParseFile(name);
    if (!doc)
        g_error("Failed to load camera path \"%s\"", name);
    dtd = xmlParseDTD(NULL, TO_XML("camerapath.dtd"));
    if (!dtd)
        g_error("Failed to load the camera path DTD");
    valid = xmlNewValidCtxt();
    if (!xmlValidateDtd(valid, doc, dtd))
        g_error("Camera path \"%s\" does not conform to the DTD", name);
    xmlFreeValidCtxt(valid);
    xmlFreeDtd(dtd);

    xmlNodePtr root = xmlDocGetRootElement(doc);
    xmlNodePtr child = root->children;
    while (child)
    {
        if (!xmlStrcmp(child->name, TO_XML("cameraframe")))
        {
            xmlNodePtr cur = child->children;
            Matrix4 transform;
            while (cur)
            {
                xmlChar *prop;
                if (!xmlStrcmp(cur->name, TO_XML("rotate")))
                {
                    prop = xmlGetProp(cur, TO_XML("angle"));
                    double angle = get_float(prop);
                    xmlFree(prop);

                    prop = xmlGetProp(cur, TO_XML("axis"));
                    transform *= matrix_rotate(angle * M_PI / 180.0, get_triple(prop));
                    xmlFree(prop);
                }
                else if (!xmlStrcmp(cur->name, TO_XML("translate")))
                {
                    prop = xmlGetProp(cur, TO_XML("offset"));
                    transform *= matrix_translate(get_triple(prop));
                    xmlFree(prop);
                }
                cur = cur->next;
            }
            transform = inverse(transform);
            camerapath_frames.push_back(transform);
        }
        child = child->next;
    }
    camerapath_frame = 0;

    xmlFreeDoc(doc);
}

void RenderViewer::load_model(const string &model_filename)
{
    xmlDocPtr doc;
    xmlNodePtr model, instance;
    const char *name = model_filename.c_str();
    const xmlChar *xmlname = TO_XML(name);

    doc = xmlNewDoc(TO_XML("1.0"));
    doc->children = xmlNewDocNode(doc, NULL, TO_XML("scene"), NULL);
    model = xmlNewChild(doc->children, NULL, TO_XML("model"), NULL);
    xmlSetProp(model, TO_XML("name"), xmlname);
    xmlSetProp(model, TO_XML("filename"), xmlname);
    instance = xmlNewChild(doc->children, NULL, TO_XML("instance"), NULL);
    xmlSetProp(instance, TO_XML("model"), xmlname);

    add_scene(name, xmlDocGetRootElement(doc));

    xmlFreeDoc(doc);
}

GtkWidget *RenderViewer::create_draw()
{
    GtkWidget *draw;

    draw = gtk_drawing_area_new();
    gtk_widget_set_size_request(draw, draw_width.int_value, draw_height.int_value);
    GdkGLConfig *glconfig = gdk_gl_config_new_by_mode((GdkGLConfigMode) (GDK_GL_MODE_RGB
                                                                         | GDK_GL_MODE_DEPTH
                                                                         | GDK_GL_MODE_DOUBLE));
    if (glconfig == NULL)
        g_error("failed to initialise OpenGL config");
    gtk_widget_set_gl_capability(draw, glconfig, NULL, TRUE, GDK_GL_RGBA_TYPE);
    g_signal_connect_after(G_OBJECT(draw), "realize",
                           G_CALLBACK(viewer_realize), this);
    g_signal_connect(G_OBJECT(draw), "configure_event",
                     G_CALLBACK(viewer_configure_event), this);
    g_signal_connect(G_OBJECT(draw), "expose_event",
                     G_CALLBACK(viewer_expose_event), this);

    /* Disable user interface if there is a fixed camerapath */
    if (camerapath_frame == -1)
    {
        gtk_widget_add_events(draw,
                              GDK_BUTTON_PRESS_MASK
                              | GDK_BUTTON_RELEASE_MASK
                              | GDK_KEY_PRESS_MASK
                              | GDK_KEY_RELEASE_MASK);
        g_signal_connect(G_OBJECT(draw), "button_press_event",
                         G_CALLBACK(viewer_button_event), this);
        g_signal_connect(G_OBJECT(draw), "button_release_event",
                         G_CALLBACK(viewer_button_event), this);
        g_signal_connect(G_OBJECT(draw), "motion_notify_event",
                         G_CALLBACK(viewer_motion_notify_event), this);
        g_signal_connect(G_OBJECT(draw), "key_press_event",
                         G_CALLBACK(viewer_key_event), this);
        g_signal_connect(G_OBJECT(draw), "key_release_event",
                         G_CALLBACK(viewer_key_event), this);
        GTK_WIDGET_SET_FLAGS(draw, GTK_CAN_FOCUS);
    }

    return draw;
}

static const gchar *ui_desc =
"<ui>"
"  <menubar name='MenuBar'>"
"    <menu action='FileMenu'>"
"      <menuitem action='Quit' />"
"    </menu>"
"    <menu action='ViewMenu'>"
"      <menuitem action='Animate' />"
"      <separator />"
"      <menuitem action='DirectLOD' />"
"      <menuitem action='NoFrustumCull' />"
"      <menuitem action='NoLOD' />"
"    </menu>"
"    <menu action='ShaderMenu'>"
"      <placeholder name='Geometry' />"
"      <separator />"
"      <placeholder name='Normal' />"
"      <separator />"
"      <placeholder name='Lighting' />"
"      <separator />"
"      <placeholder name='Shaders' />"
"      <separator />"
"      <placeholder name='Other' />"
"    </menu>"
"  </menubar>"
"</ui>";

struct ViewOptionData
{
    RenderViewer *viewer;
    RenderOption option;
};

static void view_option_data_destroy(gpointer data, GClosure *closure)
{
    delete (ViewOptionData *) data;
}

static void viewer_toggle_option(GtkToggleAction *action, gpointer user_data)
{
    const ViewOptionData *data = (const ViewOptionData *) user_data;
    data->viewer->toggle_option(action, data->option);
}

GtkUIManager *RenderViewer::create_menu()
{
    GtkUIManager *ui;
    GtkActionGroup *menu_actions;
    GError *error = NULL;
    set<string> enabled;

    const string defaults[] = {"FP with loop", "Static pixel", "Dot product", "Texture"};
    const struct { RenderOption option; const char *action, *label; } view_options[] =
    {
        { RENDER_ANIMATE,         "Animate",       "Animate"            },
        { RENDER_DIRECT_LOD,      "DirectLOD",     "Directly set LOD"   },
        { RENDER_NO_FRUSTUM_CULL, "NoFrustumCull", "No frustum culling" },
        { RENDER_NO_LOD,          "NoLOD",         "No LOD"             }
    };

    if (initial_shaders.used)
    {
        string::size_type p = 0, q;
        for (;;)
        {
            q = initial_shaders.string_value.find(',', p);
            string s = initial_shaders.string_value.substr(p, q - p);
            // Check that it is valid
            bool valid = false;
            for (size_t i = 0; factory[i].name; i++)
                if (factory[i].name == s)
                {
                    valid = true;
                    break;
                }
            for (size_t i = 0; i < G_N_ELEMENTS(view_options); i++)
                if (view_options[i].action == s)
                {
                    valid = true;
                    break;
                }
            if (!valid)
                g_error("Unknown shader \"%s\".", s.c_str());
            enabled.insert(s);
            if (q == string::npos) break;
            p = q + 1;
        }
    }
    else
    {
        for (size_t i = 0; i < G_N_ELEMENTS(defaults); i++)
            enabled.insert(defaults[i]);
    }

    GtkActionEntry misc_menu_actions[] =
    {
        { "FileMenu",   NULL,           "_File",    NULL, NULL, NULL },
        { "ViewMenu"  , NULL,           "_View",    NULL, NULL, NULL },
        { "ShaderMenu", NULL,           "_Shaders", NULL, NULL, NULL },
        { "Quit",       GTK_STOCK_QUIT, "_Quit",    NULL, NULL, G_CALLBACK(gtk_main_quit) }
    };

    menu_actions = gtk_action_group_new("MenuActions");
    gtk_action_group_add_actions(menu_actions, misc_menu_actions, G_N_ELEMENTS(misc_menu_actions), NULL);
    for (size_t i = 0; i < G_N_ELEMENTS(view_options); i++)
    {
        GtkToggleAction *action;
        ViewOptionData *data;

        data = new ViewOptionData;
        data->viewer = this;
        data->option = view_options[i].option;

        action = gtk_toggle_action_new(view_options[i].action, view_options[i].label, NULL, NULL);
        g_signal_connect_data(G_OBJECT(action), "toggled",
                              G_CALLBACK(viewer_toggle_option), data,
                              view_option_data_destroy, (GConnectFlags) 0);
        gtk_action_group_add_action_with_accel(menu_actions, GTK_ACTION(action), NULL);
        if (enabled.count(view_options[i].action))
            gtk_action_activate(GTK_ACTION(action));
    }
    GSList *groups[4] = {NULL, NULL, NULL, NULL};
    for (size_t i = 0; factory[i].name; i++)
    {
        GtkToggleAction *action;
        GtkRadioAction *radio_action;

        switch (factory[i].type)
        {
        case RENDER_HELPER_SHADER:
        case RENDER_HELPER_OTHER:
            action = gtk_toggle_action_new(factory[i].name,
                                           factory[i].name,
                                           NULL, NULL);
            break;
        default:
            radio_action = gtk_radio_action_new(factory[i].name,
                                                factory[i].name,
                                                NULL, NULL, 0);
            gtk_radio_action_set_group(radio_action, groups[(int) factory[i].type]);
            groups[(int) factory[i].type] = gtk_radio_action_get_group(radio_action);
            action = GTK_TOGGLE_ACTION(radio_action);
            break;
        }
        if (enabled.count(factory[i].name))
        {
            gtk_action_activate(GTK_ACTION(action));
            rc.add_helper(factory[i].instantiate());
        }
        g_signal_connect(G_OBJECT(action), "toggled",
                         G_CALLBACK(viewer_toggle_helper), this);
        helper_actions.push_back(action);
        gtk_action_group_add_action_with_accel(menu_actions, GTK_ACTION(action), factory[i].accel);
    }

    ui = gtk_ui_manager_new();
    gtk_ui_manager_set_add_tearoffs(ui, TRUE);
    gtk_ui_manager_insert_action_group(ui, menu_actions, 0);
    if (!gtk_ui_manager_add_ui_from_string(ui, ui_desc, strlen(ui_desc), &error))
    {
        g_message("Failed to create UI: %s", error->message);
        g_error_free(error);
    }

    guint merge_id = gtk_ui_manager_new_merge_id(ui);
    for (size_t i = 0; factory[i].name; i++)
    {
        const gchar *path = "";
        switch (factory[i].type)
        {
        case RENDER_HELPER_GEOMETRY: path = "/ui/MenuBar/ShaderMenu/Geometry"; break;
        case RENDER_HELPER_NORMAL:   path = "/ui/MenuBar/ShaderMenu/Normal";   break;
        case RENDER_HELPER_LIGHTING: path = "/ui/MenuBar/ShaderMenu/Lighting"; break;
        case RENDER_HELPER_SHADER:   path = "/ui/MenuBar/ShaderMenu/Shaders";  break;
        case RENDER_HELPER_OTHER:    path = "/ui/MenuBar/ShaderMenu/Other";    break;
        }

        gtk_ui_manager_add_ui(ui, merge_id, path,
                              factory[i].name, factory[i].name,
                              GTK_UI_MANAGER_MENUITEM, FALSE);
    }

    gtk_ui_manager_ensure_update(ui);
    return ui;
}


RenderViewer::RenderViewer() :
fullscreen(false),
capturing(false),
capture_first(false),
advance_frames(0.0),
camerapath_frames(),
camerapath_frame(-1)
{
    GtkUIManager *ui;

    fill(render_options, render_options + RENDER_OPTION_COUNT, false);
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_container_set_reallocate_redraws(GTK_CONTAINER(window), TRUE);
    g_signal_connect(G_OBJECT(window), "destroy",
                     G_CALLBACK(gtk_main_quit), NULL);

    draw = create_draw();
    ui = create_menu();
    menu = gtk_ui_manager_get_widget(ui, "/ui/MenuBar");
    gtk_widget_hide(menu); // prevent spurious appearance with camerapath

    notebook = gtk_notebook_new();
    g_signal_connect(G_OBJECT(notebook), "switch_page",
                     G_CALLBACK(viewer_notebook_switch_page), this);

    lod_slider = gtk_hscale_new_with_range(0.0, MAX_TOLERANCE, 0.01);
    gtk_range_set_value(GTK_RANGE(lod_slider), initial_lod.double_value);

    GtkWidget *vbox = gtk_vbox_new(FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), menu, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), notebook, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), lod_slider, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), draw, TRUE, TRUE, 0);
    gtk_container_add(GTK_CONTAINER(window), vbox);
    gtk_widget_grab_focus(draw);
    gtk_window_add_accel_group(GTK_WINDOW(window),
                               gtk_ui_manager_get_accel_group(ui));
    gtk_widget_show(vbox);

    // necessary to do initialisation before calls to load_mesh
    gtk_widget_realize(draw);

    g_idle_add_full(GDK_PRIORITY_REDRAW, (GSourceFunc) redraw, draw, NULL);
    g_get_current_time(&last_update);

    current_scene = scenes.end();
}

/*** Main ***/

static int process_args(int argc, char **argv)
{
    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo('c', "camera", "recorded camera path file", 1, PROGRAM_ARGUMENT_STRING, &camerapath),
        AlodProgramOptionInfo('W', "width", "drawable width", 1, PROGRAM_ARGUMENT_INT, &draw_width),
        AlodProgramOptionInfo('H', "height", "drawable height", 1, PROGRAM_ARGUMENT_INT, &draw_height),
        AlodProgramOptionInfo('l', "initial-lod", "initial setting of pixel slider", 1, PROGRAM_ARGUMENT_DOUBLE, &initial_lod),
        AlodProgramOptionInfo('S', "initial-shaders", "comma-separated list of shaders", 1, PROGRAM_ARGUMENT_STRING, &initial_shaders),
        AlodProgramOptionInfo('\0',"show-influences", "show influence counts for various LODs", 2, PROGRAM_ARGUMENT_DOUBLE, &show_influences),
        AlodProgramOptionInfo('\0',"white-bg", "use a white background", 0, PROGRAM_ARGUMENT_STRING, &white_bg),
        AlodProgramOptionInfo()
    };

    int args = alod_process_options(argc, argv, opts);

    if (!in_format.used) in_format.string_value = "auto";
    if (!draw_width.used) draw_width.int_value = 720;
    if (!draw_height.used) draw_height.int_value = 576;
    if (!initial_lod.used) initial_lod.double_value = 1.0;
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);

    if (argc - args < 1)
    {
        cerr << "Error: must specify input file\n";
        alod_show_help(opts);
        exit(1);
    }
    return args;
}

void RenderViewer::show_all()
{
    if (camerapath_frame == -1)
        gtk_widget_show_all(window);
    else
    {
        gtk_window_fullscreen(GTK_WINDOW(window));
        gtk_window_set_decorated(GTK_WINDOW(window), false);
        gtk_widget_show(draw);
        gtk_widget_show(window);
    }
}

double RenderViewer::vertices_at_lod(double lod)
{
    RenderScene &scene = *current_scene;
    double total = 0.0;
    gtk_range_set_value(GTK_RANGE(lod_slider), lod);
    camerapath_frame = 0;
    reset_animation();
    for (size_t f = 0; f < camerapath_frames.size(); f++)
    {
        update_camera(0.0);
        update_animation(0.0);

        for (list<RenderInstance>::iterator i = scene.instances.begin();
             i != scene.instances.end(); i++)
            total += i->vertices();
    }
    return total / camerapath_frames.size();
}

double RenderViewer::influences_at_lod(double lod)
{
    RenderScene &scene = *current_scene;
    double total = 0.0;
    gtk_range_set_value(GTK_RANGE(lod_slider), lod);
    camerapath_frame = 0;
    reset_animation();
    for (size_t f = 0; f < camerapath_frames.size(); f++)
    {
        update_camera(0.0);
        update_animation(0.0);

        for (list<RenderInstance>::iterator i = scene.instances.begin();
             i != scene.instances.end(); i++)
            total += i->influences();
    }
    return total / camerapath_frames.size();
}

void RenderViewer::show_influences()
{
    if (::show_influences.arg && ::show_influences.double_value < 0.0)
    {
        double inf = -::show_influences.double_value;
        double l = 0.0;
        double r = MAX_TOLERANCE;
        for (int i = 0; i < 20; i++)
        {
            double m = (l + r) * 0.5;
            double mi = influences_at_lod(m);
            if (mi < inf)
                r = m;
            else
                l = m;
        }
        double li = influences_at_lod(l);
        double ri = influences_at_lod(r);
        if (fabs(ri - inf)
            < fabs(li - inf))
            li = ri, l = r;
        printf("%.3f\n", l);
    }
    else if (::show_influences.arg)
        printf("%.3f\n", influences_at_lod(::show_influences.double_value));
    else
    {
        RenderScene &scene = *current_scene;
        for (int l = 0; l <= MAX_TOLERANCE; l++)
            printf("%d %lu %.3f %.3f\n", 
                   l, (unsigned long) scene.instances.size(), vertices_at_lod(l), influences_at_lod(l));
    }
}

size_t RenderViewer::camerapath_num_frames()
{
    return camerapath_frames.size();
}

int main(int argc, char **argv)
{
    gtk_init(&argc, &argv);
    gtk_gl_init(&argc, &argv);
    cg_context = cgCreateContext();
    cgSetAutoCompile(cg_context, CG_COMPILE_MANUAL);
    cgSetErrorCallback(check_cg_error);

    int args = process_args(argc, argv);

    RenderViewer viewer;
    for (int i = args; i < argc; i++)
    {
        string file = argv[i];
        if (file.size() >= 4 && file.substr(file.size() - 4) == ".xml")
            viewer.load_scene(file);
        else
            viewer.load_model(file);
    }
    if (camerapath.used)
    {
        viewer.load_camerapath(camerapath.string_value);
        if (show_influences.used)
        {
            viewer.show_all();
            viewer.show_influences();
            return 0;
        }
    }

    create_blank_cursor();
    viewer.show_all();

    gtk_main();
    if (camerapath.used)
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Time per frame: %.7fs", render_timer.elapsed() / viewer.camerapath_num_frames());
    return 0;
}
