#ifndef ALOD_RENDER_MESHRENDER_H
#define ALOD_RENDER_MESHRENDER_H

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <glib.h>
#include <vector>
#include <list>
#include <set>
#include <string>
#include <Cg/cg.h>
#include <gtk/gtk.h>
#include "mesh/mesh.h"
#include "geom/skeleton.h"
#include "geom/coords.h"
#include "geom/bounding.h"
#include "render/rendermesh.h"
#include "formats/formats.h"
#include <GL/gl.h>
#include <libxml/tree.h>

#define RENDER_FLAG_VERTEX_TANGENTS     0x0001   //!< Need tangents from geometry converter
#define RENDER_FLAG_VERTEX_NORMALS      0x0002   //!< Need normals from geometry converter
#define RENDER_FLAG_TEXCOORDS           0x0004   //!< Need the parametrisation
#define RENDER_FLAG_SORT_INFLUENCES     0x0008   //!< Sort vertices by decreasing number of influences

#define RENDER_FLAGS_VERTEX_DERIVATIVES (RENDER_FLAG_VERTEX_TANGENTS | RENDER_FLAG_VERTEX_NORMALS)

enum RenderOption
{
    RENDER_ANIMATE,

    RENDER_DIRECT_LOD,
    RENDER_NO_FRUSTUM_CULL,
    RENDER_NO_LOD,

    RENDER_NULL
};
#define RENDER_OPTION_COUNT ((size_t) RENDER_NULL)

//! Specifies a plane in space, used for frustum culling
struct ClipPlane
{
    Vector4 normal; //!< First three coordinates are normalised
};

//! Information used for frame-to-frame updates
/*! This structure contains the information associated with the viewpoint.
 * The intention is that freezing this state will lock the camera without
 * having other adverse effects.
 */
struct Camera
{
    double azimuth;
    double elevation;
    Vector3 position;

    Matrix4 view_matrix;
    Matrix4 projection_matrix;
    ClipPlane frustum[6];

    //! Function of viewport size
    /*! This is the worse-case ratio of model-space to pixels when at a
     * depth of 1.
     */
    double lod_tolerance_base;

    Camera() :
        azimuth(G_PI * 1.25),
        elevation(G_PI * 0.25),
        position(Vector3(0.5, 0.0, 2.0)),
        lod_tolerance_base(0.0) {}
};

class RenderViewer;
class RenderScene
{
private:
    ModelTransform parse_transform(xmlNodePtr cur) throw(bad_format);
    void parse_model(xmlNodePtr cur) throw(bad_format);
    void parse_instance(xmlNodePtr cur, const ModelTransform &tree_transform) throw(bad_format);
    ModelTransform parse_tree(xmlNodePtr cur, const ModelTransform &tree_transform) throw(bad_format);
    RenderViewer *viewer;
public:
    std::string filename;
    std::list<RenderObject> objects;
    std::list<RenderInstance> instances;

    RenderScene(RenderViewer *viewer, const std::string &filename);
    void load(xmlNodePtr root);
};

class RenderViewer
{
private:
    /* User interface */
    GtkWidget *window, *draw, *lod_slider;
    GtkWidget *menu, *notebook;
    std::vector<GtkToggleAction *> helper_actions; //!< one action per factory item
    std::set<guint> keys_down;
    bool fullscreen : 1;
    bool capturing : 1;
    bool capture_first : 1;

    /* Raw data */
    std::list<RenderScene> scenes;
    std::list<RenderScene>::iterator current_scene;

    /* Rendering */
    RenderClasses rc;

    /* Timing */
    GTimeVal last_update;
    double advance_frames;

    /* Fixed camera stuff */
    std::vector<Matrix4> camerapath_frames;
    int camerapath_frame;

    GtkWidget *create_draw();
    GtkUIManager *create_menu();
    void update_camera(double elapsed);
    void update_animation(double elapsed);
    void reset_animation();

    void rebuild_helpers();
    void begin_capture(guint32 time);
    void end_capture(guint32 time);
    void add_scene(const char *filename, xmlNodePtr root);

    friend void scene_start_element(GMarkupParseContext *context,
                                    const gchar *element_name,
                                    const gchar **attribute_names,
                                    const gchar **attribute_values,
                                    gpointer user_data,
                                    GError **error);
public:
    /* User interface */
    bool render_options[RENDER_OPTION_COUNT];
    Camera camera;

    /* These are public only to facilitate the wrappers */
    void realize(GtkWidget *widget);
    void notebook_switch_page(GtkWidget *notebook,
                              GtkWidget *notebook_page,
                              gint page);
    void toggle_option(GtkToggleAction *action, RenderOption option);
    void toggle_helper(GtkToggleAction *action);
    gboolean configure_event(GtkWidget *widget, GdkEventConfigure *event);
    gboolean expose_event(GtkWidget *widget, GdkEventExpose *event);
    gboolean button_event(GtkWidget *widget, GdkEventButton *event);
    gboolean motion_notify_event(GtkWidget *widget, GdkEventMotion *event);
    gboolean key_event(GtkWidget *widget, GdkEventKey *event);

    double lod_tolerance();

    void load_scene(const std::string &scene_filename);
    void load_model(const std::string &model_filename);
    void load_camerapath(const std::string &camerapath_filename);

    void show_all();                      //!< wrapper around gtk_widget_show_all
    double influences_at_lod(double lod); //!< helper for \c show_influences
    double vertices_at_lod(double lod);   //!< helper for \c show_influences
    void show_influences();               //!< handles the \c --show-influences command-line option
    std::size_t camerapath_num_frames();  //!< number of frames in the camera path, or 0 if there isn't one

    RenderViewer();
};

/* Binds and enables the given parameter, saving the previous value. The
 * old values are restored on destruction.
 */
class TextureManager
{
private:
    struct save
    {
        GLenum unit;
        GLenum target;
        GLint texture;
    };

    std::vector<save> saved;
public:
    void bind_texture(CGparameter parameter, GLuint texture);
    void reset();
    TextureManager() {}
    ~TextureManager() { reset(); }
};

//! Checks for a CG error and calls \c g_error if there is one
void check_cg_error();

//! Returns the global CG context
CGcontext global_cg_context();

//! Returns the preferred vertex profile
CGprofile global_cg_vertex_profile();

//! Returns the preferred fragment profile
CGprofile global_cg_fragment_profile();

//! Loads and returns the specified program
CGprogram load_cg_program(const std::string &source_file, CGprofile profile,
                          const char *args[]);

//! Creates a simple shader pair that just transforms and passes through colour
/*! The returned shaders are statically allocated and should NOT be destroyed.
 * The returned programs are already compiled and loaded.
 */
void load_simple_cg_programs(CGprogram &vp, CGprogram &fp);

//! Returns a statically allocates display list that draws a unit sphere
GLuint sphere_list();

#include "render/interfaces.h"

#endif /* ALOD_RENDER_MESHRENDER_H */
