#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include "render/interfaces.h"
#include "render/meshrender.h"
#include "render/normal_dynamic.h"
#include "mesh/mesh.h"
#include "alodgl/glmisc.h"
#include <string>
#include <vector>
#include <algorithm>
#include <utility>
#include <queue>
#include <cstddef>
#include <glib.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>

using namespace std;

class RenderHelperNormalDynamic : public RenderHelper
{
private:
    struct Tile
    {
        int count;                 //!< Influence count for the tile
        int inu, inv;              //!< Texel coordinates in input map
        int outu, outv;            //!< Texel coordinates in output map

        bool operator <(const Tile &b) const { return count > b.count; }
    };

    struct Level
    {
        int tile_size;
        int out_width, out_height; //!< Size of generated normal map
        int in_width, in_height;   //!< Size of data textures
        GLuint bone_texture;
        GLuint tangent_textures[2];
        GLuint draw_list;
        std::vector<Tile> tiles;
    };

    std::vector<Level> levels;
    double texture_factor;   //!< converts model space distance to rough texture space distance
public:
    RenderHelperNormalDynamic(RenderHelperClassNormalDynamic &parent, const RenderMesh &m);
    virtual ~RenderHelperNormalDynamic();

    virtual void begin_object(const Skeleton &skeleton, double lod, double z);
};


RenderHelperClassNormalDynamic::RenderHelperClassNormalDynamic(int tile_size)
    : tile_size(tile_size), normal_textures()
{
    // empty
}

void RenderHelperClassNormalDynamic::set_parameters(CGparameter vparameter, CGparameter fparameter)
{
    normal_parameter = cgGetNamedStructParameter(fparameter, "normal_texture");
    check_cg_error();
    g_assert(normal_parameter);

    // Create the generation programs
    generate_vp = load_cg_program("dnm_vp.cg", global_cg_vertex_profile(), NULL);
    generate_fp = load_cg_program("dnm_fp.cg", global_cg_fragment_profile(), NULL);
    cgCompileProgram(generate_vp); check_cg_error();
    cgGLLoadProgram(generate_vp);  check_cg_error();
    cgCompileProgram(generate_fp); check_cg_error();
    cgGLLoadProgram(generate_fp);  check_cg_error();
    bone_parameter = cgGetNamedParameter(generate_fp, "bone_tex"); check_cg_error();
    tangent_parameters[0] = cgGetNamedParameter(generate_fp, "tangent_tex[0]"); check_cg_error();
    tangent_parameters[1] = cgGetNamedParameter(generate_fp, "tangent_tex[1]"); check_cg_error();
    transform_parameter = cgGetNamedParameter(generate_fp, "transform_tex"); check_cg_error();
}

void RenderHelperClassNormalDynamic::end_state()
{
    check_cg_error();
    for (size_t i = 0; i < normal_textures.size(); i++)
    {
        glDeleteTextures(1, &normal_textures[i].texture_id);
        glDeleteFramebuffersEXT(1, &normal_textures[i].fbo_id);
    }
    cgDestroyProgram(generate_vp);
    cgDestroyProgram(generate_fp);
    transform_texture.end();

    RenderHelperClass::end_state();
    check_cg_error();
}

void RenderHelperClassNormalDynamic::NormalTexture::initialise(int size)
{
    GLint old;
    glGetIntegerv(GL_TEXTURE_BINDING_2D, &old);

    glGenTextures(1, &texture_id);
    glBindTexture(GL_TEXTURE_2D, texture_id);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, size, size, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glBindTexture(GL_TEXTURE_2D, old);

    glGenFramebuffersEXT(1, &fbo_id);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo_id);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
                              GL_TEXTURE_2D, texture_id, 0);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
}

GLuint RenderHelperClassNormalDynamic::bind_fbo(int log_size)
{
    if (log_size >= (int) normal_textures.size())
        normal_textures.resize(log_size + 1);
    if (normal_textures[log_size].texture_id == 0)
        normal_textures[log_size].initialise(1 << log_size);

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, normal_textures[log_size].fbo_id);
    alodgl_check_framebuffer_status();
    return normal_textures[log_size].texture_id;
}

RenderHelper *RenderHelperClassNormalDynamic::new_instance(const RenderMesh &m)
{
    return new RenderHelperNormalDynamic(*this, m);
}

/**********************************************************************/

// Makes a half-size tangent map with a simple box filter */
static MeshTangentMap reduce_tangent_map(const Skeleton &skeleton,
                                         const MeshTangentMap &tm)
{
    int w = tm.width;
    int h = tm.height;
    int w2 = w / 2;
    int h2 = h / 2;
    MeshTangentMap out;

    out.set_size(w2, h2);
    for (int y = 0; y < h2; y++)
        for (int x = 0; x < w2; x++)
        {
            out.data()[y * w2 + x].patch = tm.data()[2 * y * w + 2 * x].patch;
            TangentPlane &sum = out.data()[y * w2 + x].tangent_plane;
            sum = tm.data()[2 * y * w + 2 * x].tangent_plane;
            for (int i = 1; i < 4; i++)
            {
                const TangentPlane &cur = tm.data()[(2 * y + (i >> 1)) * w + (2 * x + (i & 1))].tangent_plane;
                sum.tangents[0] += cur.tangents[0];
                sum.tangents[1] += cur.tangents[1];
            }
            sum.tangents[0] = normalise(sum.tangents[0]);
            sum.tangents[1] -= (sum.tangents[0] * sum.tangents[1]) * sum.tangents[0];
            sum.tangents[1] = normalise(sum.tangents[1]);
        }
    return out;
}

static void tile_vertex(int inu, int inv, int outu, int outv, int width, int height,
                        int count)
{
    glMultiTexCoord3f(GL_TEXTURE0, inu - 0.5f * (count - 1), inv, inu + 0.5f * count);
    glVertex2f(2.0f * outu / width - 1.0f, 2.0f * outv / height - 1.0f);
}

RenderHelperNormalDynamic::RenderHelperNormalDynamic(RenderHelperClassNormalDynamic &parent, const RenderMesh &m)
: RenderHelper(parent, m)
{
    GLint old;
    glGetIntegerv(GL_TEXTURE_BINDING_RECTANGLE_ARB, &old);

    MeshTangentMap tm;

    g_assert(m.mesh.tangent_maps.size() <= 1);
    if (m.mesh.tangent_maps.empty())
    {
        tm.set_size(1, 1);
        tm.data()[0].patch = 1;  /* Not used */
        /* Pick arbitrary non-zero tangent plane */
        tm.data()[0].tangent_plane.tangents[0] = VectorPosition(m.mesh.skeleton.get_read());
        tm.data()[0].tangent_plane.tangents[0][0] = 1.0;
        tm.data()[0].tangent_plane.tangents[1] = VectorPosition(m.mesh.skeleton.get_read());
        tm.data()[0].tangent_plane.tangents[1][1] = 1.0;
    }
    else
        tm = m.mesh.tangent_maps.front();
    int tile_size = parent.tile_size;

    if (tm.width != tm.height) g_error("Tangent map is not square");
    if (tm.width & (tm.width - 1)) g_error("Tangent map size is not a power of 2");

    for (int level = 0; ; level++)
    {
        levels.push_back(Level());
        Level &l = levels.back();
        tile_size = min(tile_size, (int) min(tm.width, tm.height));
        g_assert(!(tm.width % tile_size));
        g_assert(!(tm.height % tile_size));
        int xtiles = tm.width / tile_size;
        int ytiles = tm.height / tile_size;

        // Initial allocation of tiles, including influence counting
        l.tile_size = tile_size;
        l.tiles.resize(xtiles * ytiles);
        int i = 0;
        int count_sum = 0;
        for (int y = 0; y < ytiles; y++)
            for (int x = 0; x < xtiles; x++, i++)
            {
                int p0 = x * tile_size;
                int q0 = y * tile_size;
                l.tiles[i].outu = p0;
                l.tiles[i].outv = q0;
                l.tiles[i].count = 0;
                for (int q = q0; q < q0 + tile_size; q++)
                    for (int p = p0; p < p0 + tile_size; p++)
                    {
                        int addr = q * tm.width + p;
                        const TangentPlane &tp = tm.data()[addr].tangent_plane;
                        VectorPosition sum = tp.tangents[0] + tp.tangents[1];
                        l.tiles[i].count = max(l.tiles[i].count, (int) distance(sum.begin(), sum.end()));
                    }
                count_sum += l.tiles[i].count;
            }

        // Tetris-pack the tiles
        int oytiles = (int) sqrt(count_sum);
        int oxtiles = 0;
        priority_queue<pair<int, int>, vector<pair<int, int> >, greater<pair<int, int> > > q;   // items are used, row
        for (int i = 0; i < oytiles; i++)
            q.push(make_pair(0, i));
        sort(l.tiles.begin(), l.tiles.end());
        for (int i = 0; i < xtiles * ytiles; i++)
        {
            pair<int, int> place = q.top();
            q.pop();
            l.tiles[i].inu = tile_size * place.first;
            l.tiles[i].inv = tile_size * place.second;
            q.push(make_pair(place.first + l.tiles[i].count, place.second));
            oxtiles = max(oxtiles, place.first + l.tiles[i].count);
        }

        // fill in textures
        GLubyte *bones = new GLubyte[oxtiles * oytiles * tile_size * tile_size];
        GLfloat (*tangents[2])[4];
        tangents[0] = new GLfloat[oxtiles * oytiles * tile_size * tile_size][4];
        tangents[1] = new GLfloat[oxtiles * oytiles * tile_size * tile_size][4];
        fill(bones, bones + oxtiles * oytiles * tile_size * tile_size, 0);
        for (int i = 0; i < xtiles * ytiles; i++)
        {
            for (int q = 0; q < tile_size; q++)
                for (int p = 0; p < tile_size; p++)
                {
                    int addr = (q + l.tiles[i].outv) * tm.width + (p + l.tiles[i].outu);
                    const TangentPlane &tp = tm.data()[addr].tangent_plane;
                    const VectorPosition sum = tp.tangents[0] + tp.tangents[1];
                    VectorPosition::const_iterator s = sum.begin();
                    for (int c = 0; c < l.tiles[i].count; c++)
                    {
                        int addr2 = (q + l.tiles[i].inv) * (oxtiles * tile_size)
                            + (p * l.tiles[i].count + c + l.tiles[i].inu);
                        if (s != sum.end())
                        {
                            bones[addr2] = s->index / 4;
                            for (int j = 0; j < 2; j++)
                                for (int k = 0; k < 4; k++)
                                    tangents[j][addr2][k] = tp.tangents[j][s->index + k];
                            s++;
                        }
                        else
                        {
                            bones[addr2] = 0;
                            for (int j = 0; j < 2; j++)
                                for (int k = 0; k < 4; k++)
                                    tangents[j][addr2][k] = 0.0f;
                        }
                    }
                }
        }

        // allocate textures
        glGenTextures(1, &l.bone_texture);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, l.bone_texture);
        glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_LUMINANCE8,
                     oxtiles * tile_size, oytiles * tile_size, 0,
                     GL_LUMINANCE, GL_UNSIGNED_BYTE, bones);

        glGenTextures(2, l.tangent_textures);
        for (int i = 0; i < 2; i++)
        {
            glBindTexture(GL_TEXTURE_RECTANGLE_ARB, l.tangent_textures[i]);
            glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA16F_ARB,
                         oxtiles * tile_size, oytiles * tile_size, 0,
                         GL_RGBA, GL_FLOAT, tangents[i]);
        }
        delete[] bones;
        delete[] tangents[0];
        delete[] tangents[1];

        // Build the draw list
        l.draw_list = glGenLists(1);
        glNewList(l.draw_list, GL_COMPILE);
        glBegin(GL_QUADS);
        for (int i = 0; i < xtiles * ytiles; i++)
        {
            tile_vertex(l.tiles[i].inu, l.tiles[i].inv,
                        l.tiles[i].outu, l.tiles[i].outv,
                        tm.width, tm.height, l.tiles[i].count);
            tile_vertex(l.tiles[i].inu + l.tiles[i].count * tile_size, l.tiles[i].inv,
                        l.tiles[i].outu + tile_size, l.tiles[i].outv,
                        tm.width, tm.height, l.tiles[i].count);
            tile_vertex(l.tiles[i].inu + l.tiles[i].count * tile_size, l.tiles[i].inv + tile_size,
                        l.tiles[i].outu + tile_size, l.tiles[i].outv + tile_size,
                        tm.width, tm.height, l.tiles[i].count);
            tile_vertex(l.tiles[i].inu, l.tiles[i].inv + tile_size,
                        l.tiles[i].outu, l.tiles[i].outv + tile_size,
                        tm.width, tm.height, l.tiles[i].count);
        }
        glEnd();
        glEndList();
        l.in_width = oxtiles * tile_size;
        l.in_height = oytiles * tile_size;
        l.out_width = xtiles * tile_size;
        l.out_height = ytiles * tile_size;

        // Prepare for next iteration
        if (tm.width == 1 || tm.height == 1)
            break;
        tm = reduce_tangent_map(m.mesh.skeleton.get_read(), tm);
    }

    /* Compute the texture factor
     *
     * This is a bit of a thumb-suck, since it just has to be in the
     * right area, especially since both animation and camera orientation
     * make it impossible to be exact. We take each edge and measure the
     * texture rate across it, and take the smallest of these, favouring
     * fast and blurry over slow and aliased, and not trying to handle
     * super-stretched areas.
     *
     * The original mesh structure is destroyed by the decimation, so we
     * use the RenderMesh structure.
     */
    texture_factor = HUGE_VAL;
    for (vector<RenderMeshFace>::const_iterator f = m.levels[0].faces.begin(); f != m.levels[0].faces.end(); f++)
        for (int i = 0; i < 3; i++)
        {
            MeshVertexConstIterator a = f->v[i];
            MeshVertexConstIterator b = f->v[i ? i - 1 : 2];
            double spatial = length(a->position()->position - b->position()->position);
            double parametric = length(a->data.texcoords - b->data.texcoords);
            texture_factor = min(texture_factor, parametric / spatial);
        }

    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
}

RenderHelperNormalDynamic::~RenderHelperNormalDynamic()
{
    for (size_t i = 0; i < levels.size(); i++)
    {
        Level &l = levels[i];
        glDeleteTextures(1, &l.bone_texture);
        glDeleteTextures(2, l.tangent_textures);
        glDeleteLists(l.draw_list, 1);
    }
}

void RenderHelperNormalDynamic::begin_object(const Skeleton &skeleton, double lod, double z)
{
    RenderHelperClassNormalDynamic &klass = static_cast<RenderHelperClassNormalDynamic &>(get_class());
    z *= texture_factor;
    int lvl = 0;
    while (lvl < (int) levels.size() && z * levels[lvl].out_width >= 0.5)
        lvl++;

    const Level &l = levels[lvl];
    int log_size = 0;
    int tmp = l.out_width;
    while (tmp > 1)
    {
        log_size++;
        tmp >>= 1;
    }

    GLint old_viewport[4];
    GLuint normal_texture;
    TextureManager manager;

    check_cg_error();
    manager.bind_texture(klass.transform_parameter, klass.transform_texture.update(skeleton));
    manager.bind_texture(klass.bone_parameter, l.bone_texture);
    manager.bind_texture(klass.tangent_parameters[0], l.tangent_textures[0]);
    manager.bind_texture(klass.tangent_parameters[1], l.tangent_textures[1]);
    cgGLBindProgram(klass.generate_vp);
    cgGLBindProgram(klass.generate_fp);
    glDisable(GL_DEPTH_TEST);
    glGetIntegerv(GL_VIEWPORT, old_viewport);
    normal_texture = klass.bind_fbo(log_size);
    glViewport(0, 0, l.out_width, l.out_height);

    glClear(GL_COLOR_BUFFER_BIT);
    glCallList(l.draw_list);

    manager.reset();
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glEnable(GL_DEPTH_TEST);
    cgGLBindProgram(klass.vertex_program);
    cgGLBindProgram(klass.fragment_program);
    cgGLSetTextureParameter(klass.normal_parameter, normal_texture);
    cgGLEnableTextureParameter(klass.normal_parameter);
    glViewport(old_viewport[0], old_viewport[1], old_viewport[2], old_viewport[3]);
    check_cg_error();
}
