#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include "render/interfaces.h"
#include "render/meshrender.h"
#include "render/rendermesh.h"
#include "render/patches.h"
#include <map>
#include <vector>
#include <string>
#include <glib.h>
#include <time.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>

using namespace std;

class RenderHelperPatches : public RenderHelper
{
private:
    std::vector<GLuint> patch_buffers;
    struct Patch
    {
        VectorPosition seed;   //!< Sum of base vertices, used as a random seed
        Vector3 color;
        Patch() {}
        explicit Patch(const VectorPosition &v) : seed(v), color() {}
    };
public:
    RenderHelperPatches(RenderHelperClassPatches &parent, const RenderMesh &m);
    ~RenderHelperPatches();

    virtual void begin_object(const Skeleton &skeleton, double lod, double z);
};


RenderHelperClassPatches::RenderHelperClassPatches(bool random_colours)
: random_colours(random_colours)
{
}

RenderHelper *RenderHelperClassPatches::new_instance(const RenderMesh &m)
{
    return new RenderHelperPatches(*this, m);
}

void RenderHelperClassPatches::begin_state(CGprogram vp, CGprogram fp, int flags)
{
    patch_parameter = cgGetNamedParameter(vp, "patch_color");
    g_assert(patch_parameter);
    cgGLEnableClientState(patch_parameter);
    /* Cg has the unfortunate habit of setting normalisation to false.
     * We work around it by manually extracting the index.
     */
    if (cgGetParameterBaseResource(patch_parameter) == CG_ATTR0)
        patch_index = cgGetParameterResourceIndex(patch_parameter); // generic
    else
        patch_index = -1; // fixed-pipe
}

void RenderHelperClassPatches::end_state()
{
    cgGLDisableClientState(patch_parameter);
}


RenderHelperPatches::RenderHelperPatches(RenderHelperClassPatches &parent, const RenderMesh &m)
: RenderHelper(parent, m)
{
    srand(time(NULL));

    id_map<const MeshVertex *, int>::type patches; // maps vertices to their patches
    map<int, Patch> patch_map;                     // per-patch data

    /* Compute vertex-patch relationships. This is non-trivial because
     * vertices that are not on the last LOD level do not have edges
     */
    const RenderMeshLevel &ll = m.levels.back();
    for (size_t v = 0; v < ll.vertices.size(); v++)
        patches[ll.vertices[v].second] = ll.vertices[v].second->first_edge()->face()->patch;
    for (int i = (int) m.levels.size() - 1; i >= 0; i--)
    {
        const RenderMeshLevel &l = m.levels[i];
        for (size_t v = 0; v < l.vertices.size(); v++)
            patches[l.vertices[v].first] = patches[l.vertices[v].second];
    }

    for (size_t i = 0; i < m.levels[0].vertices.size(); i++)
    {
        const MeshVertex *v = m.levels[0].vertices[i].first;
        int p = patches[v];
        if (!patch_map.count(p))
            patch_map.insert(make_pair(p, Patch(v->position()->position)));
        else
            patch_map[p].seed += v->position()->position;
    }
    for (map<int, Patch>::iterator i = patch_map.begin(); i != patch_map.end(); i++)
    {
        int seed = 0;
        for (VectorPosition::const_iterator j = i->second.seed.begin(); j != i->second.seed.end(); j++)
            for (int k = 0; k < 4; k++) seed += (int) (4096.0 * j->values[k]);
        if (!parent.random_colours)
            srand(seed);
        for (int j = 0; j < 3; j++)
            i->second.color[j] = rand() / (RAND_MAX * 2.0) + 0.5;
    }

    patch_buffers.resize(m.levels.size(), 0);
    for (size_t i = 0; i < m.levels.size(); i++)
    {
        const RenderMeshLevel &rl = m.levels[i];
        glGenBuffers(1, &patch_buffers[i]);
        glBindBuffer(GL_ARRAY_BUFFER, patch_buffers[i]);
        glBufferData(GL_ARRAY_BUFFER, 3 * sizeof(GLubyte) * rl.vertices.size(), NULL, GL_STATIC_DRAW);
        GLubyte *ptr = (GLubyte *) glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
        g_assert(ptr);
        for (size_t v = 0; v < rl.vertices.size(); v++)
        {
            int p = patches[rl.vertices[v].first];
            g_assert(patch_map.count(p));
            for (int j = 0; j < 3; j++)
                *ptr++ = (GLubyte) (patch_map[p].color[j] * 256.0);
        }
        GLboolean success = glUnmapBuffer(GL_ARRAY_BUFFER);
        if (!success) g_error("glUnmapBuffer failed");
    }
}

RenderHelperPatches::~RenderHelperPatches()
{
    for (size_t i = 0; i < patch_buffers.size(); i++)
        glDeleteBuffers(1, &patch_buffers[i]);
}

void RenderHelperPatches::begin_object(const Skeleton &skeleton, double lod, double z)
{
    RenderHelperClassPatches &klass = static_cast<RenderHelperClassPatches &>(get_class());
    int ilod = rmesh.int_lod(lod);

    RenderHelper::begin_object(skeleton, lod, z);
    glBindBuffer(GL_ARRAY_BUFFER, patch_buffers[ilod]);
    /* See comments in begin_state */
    if (klass.patch_index >= 0)
        glVertexAttribPointer(klass.patch_index, 3, GL_UNSIGNED_BYTE, GL_TRUE, 0, NULL);
    else
        cgGLSetParameterPointer(klass.patch_parameter, 3, GL_UNSIGNED_BYTE, 0, NULL);
}
