#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include "render/interfaces.h"
#include "render/meshrender.h"
#include "render/rendermesh.h"
#include "common/paths.h"
#include "alodgl/glmisc.h"
#include "geom/coords.h"
#include "geom/bounding.h"
#include <GL/gl.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <glib.h>
#include <cmath>
#include <stack>

using namespace std;

void RenderMesh::load_mesh(const Mesh &m)
{
    double error = 0.0;
    levels.clear();
    mesh = m;

    for (int lod = 0; ; lod++)
    {
        size_t vnum = mesh.vertices.real_size();
        size_t fnum = mesh.faces.real_size();

        levels.push_back(RenderMeshLevel());
        RenderMeshLevel &l = levels.back();
        l.faces.reserve(3 * fnum);
        l.vertices.reserve(vnum);
        l.error = error;

        /* Build a list of vertices in a cache-friendly order by doing a
         * depth-first walk over the mesh. Triangles' internal_data is either
         * 'unseen' (0xffffffff) or NULL (for seen).
         */
        vector<MeshVertexIterator> prev_vertices;
        prev_vertices.reserve(vnum);
        stack<MeshHalfEdgeIterator> todo; // entry edge into the new triangle
        void *unseen = GSIZE_TO_POINTER(0xffffffffU);
        for (MeshVertexIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++)
            prev_vertices.push_back(v);
        for (MeshFaceIterator f = mesh.faces.real_begin(); f != mesh.faces.real_end(); f++)
            f->internal_data = unseen;
        // Mark hole faces as visited so that they are skipped
        for (MeshFaceIterator f = mesh.faces.hole_begin(); f != mesh.faces.hole_end(); f++)
            f->internal_data = NULL;

        // Do the DFS
        for (MeshFaceIterator f = mesh.faces.real_begin(); f != mesh.faces.real_end(); f++)
            if (f->internal_data == unseen)
            {
                f->internal_data = NULL;
                todo.push(f->edges());
                while (!todo.empty())
                {
                    MeshHalfEdgeIterator curedge, e, nextedge;
                    MeshFaceIterator next;

                    curedge = todo.top();
                    todo.pop();

                    e = curedge;
                    l.faces.push_back(RenderMeshFace());
                    for (int i = 0; i < 3; i++, e = e->next())
                    {
                        // Add this triangle to the index buffer
                        l.faces.back().v[i] = e->vertex();

                        // Add neighbouring triangles
                        nextedge = e->dual();
                        next = nextedge->face();
                        if (next->internal_data == unseen)
                        {
                            next->internal_data = NULL;
                            todo.push(nextedge);
                        }
                    }
                }
            }

        /* Count influences */
        size_t inum = 0;
        for (MeshPositionConstIterator p = mesh.positions.real_begin(); p != mesh.positions.real_end(); p++)
            inum += distance(p->position.begin(), p->position.end());
        /* Reduce size */
        for (MeshVertexIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++)
            v->internal_data = unseen; // helps with debugging
        size_t inum2 = mesh_reduce(mesh, inum / 2, false, error, false);
        /* Clear internal_data, since it is undefined for remaining vertices */
        for (MeshVertexIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++)
            v->internal_data = NULL;

        /* This is not the same as inum, because at patch boundaries we
         * have to count the influences multiple times.
         */
        l.influences = 0;
        /* Fill in vertices and deltas */
        for (size_t i = 0; i < vnum; i++)
        {
            /* Find descendant in new mesh */
            MeshVertexIterator v = prev_vertices[i];
            MeshVertex *v2 = &*v;
            while (v2->internal_data)
                v2 = (MeshVertex *) v2->internal_data;

            l.vertices.push_back(make_pair(&*v, v2));
            l.influences += distance(v->position()->position.begin(),
                                     v->position()->position.end());
        }

        /* Mark index buffer as uninitialised */
        l.index_buffer = 0;

        if (inum == inum2) break; // no further reduction possible
    }
}

class compare_influences
{
private:
    int flags;

    VectorPosition influences(const MeshVertex *base,
                              const MeshVertex *target) const
    {
        VectorPosition test = target->position()->position - base->position()->position;
        if (flags & RENDER_FLAGS_VERTEX_DERIVATIVES)
        {
            test += target->data.tangent_plane.tangents[0] - base->data.tangent_plane.tangents[0];
            test += target->data.tangent_plane.tangents[1] - base->data.tangent_plane.tangents[1];
        }
        return test;
    }
public:
    explicit compare_influences(int flags) : flags(flags) {}

    bool operator()(const pair<MeshVertex *, MeshVertex *> &a,
                    const pair<MeshVertex *, MeshVertex *> &b) const
    {
        const VectorPosition ia = influences(a.first, a.second);
        const VectorPosition ib = influences(b.first, b.second);
        return distance(ia.begin(), ia.end()) > distance(ib.begin(), ib.end());
    }
};

class compare_id
{
public:
    bool operator()(const pair<MeshVertex *, MeshVertex *> &a,
                    const pair<MeshVertex *, MeshVertex *> &b) const
    {
        gsize ida = GPOINTER_TO_SIZE(a.first->internal_data);
        gsize idb = GPOINTER_TO_SIZE(b.first->internal_data);
        return ida < idb;
    }
};

void RenderMesh::create_indices(int flags)
{
    for (size_t lod = 0; lod < levels.size(); lod++)
    {
        RenderMeshLevel &l = levels[lod];

        if (flags & RENDER_FLAG_SORT_INFLUENCES)
        {
            sort(l.vertices.begin(), l.vertices.end(), compare_influences(flags));
            for (size_t v = 0; v < l.vertices.size(); v++)
                l.vertices[v].first->internal_data = GSIZE_TO_POINTER(v);
        }
        else
        {
            gsize pool = 0;
            void *unseen = GSIZE_TO_POINTER(0xffffffffU);
            for (size_t v = 0; v < l.vertices.size(); v++)
                l.vertices[v].first->internal_data = unseen;
            for (size_t i = 0; i < l.faces.size(); i++)
                for (int j = 0; j < 3; j++)
                    if (l.faces[i].v[j]->internal_data == unseen)
                    {
                        l.faces[i].v[j]->internal_data = GSIZE_TO_POINTER(pool);
                        pool++;
                    }
            sort(l.vertices.begin(), l.vertices.end(), compare_id());
        }

        if (!l.index_buffer)
            glGenBuffers(1, &l.index_buffer);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, l.index_buffer);
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, 3 * l.faces.size() * sizeof(GLuint),
                     NULL, GL_STATIC_DRAW);
        GLuint *indices = (GLuint *) glMapBuffer(GL_ELEMENT_ARRAY_BUFFER, GL_WRITE_ONLY);
        for (size_t i = 0; i < l.faces.size(); i++)
            for (int j = 0; j < 3; j++)
                *indices++ = GPOINTER_TO_SIZE(l.faces[i].v[j]->internal_data);
        bool success = glUnmapBuffer(GL_ELEMENT_ARRAY_BUFFER);
        g_assert(success);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    }
}


RenderClasses::RenderClasses()
: flags(0), helpers(), vertex_program(NULL), fragment_program(NULL) {}

RenderClasses::~RenderClasses()
{
    helpers.clear();
    if (vertex_program) cgDestroyProgram(vertex_program);
    if (fragment_program) cgDestroyProgram(fragment_program);
}

void RenderClasses::add_helper(RenderHelperClass *h)
{
    helpers.push_back(h);
    flags |= h->flags();
}

void RenderClasses::create_instances(RenderObject &r)
{
    for (size_t i = 0; i < r.helpers.size(); i++)
        delete r.helpers[i];
    r.helpers.clear();
    r.create_indices(flags);
    for (size_t i = 0; i < helpers.size(); i++)
        r.helpers.push_back(helpers[i]->new_instance(r));
}

void RenderClasses::begin_state()
{
    const char *args[4] = {NULL, NULL, NULL, NULL};
    const char **a = args;
    static const float diffuse = 0.3;
    static const float ambient = 0.4;
    static const Vector3 direction = normalise(Vector3(1.0, 1.0, 1.0));

    if (flags & RENDER_FLAG_VERTEX_TANGENTS)
        *a++ = "-DVERTEX_TANGENTS";
    if (flags & RENDER_FLAG_VERTEX_NORMALS)
        *a++ = "-DVERTEX_NORMALS";
    if (flags & RENDER_FLAG_TEXCOORDS)
        *a++ = "-DTEXCOORDS";

    vertex_program = load_cg_program("vertex.cg", global_cg_vertex_profile(), args);
    fragment_program = load_cg_program("fragment.cg", global_cg_fragment_profile(), args);

    CGparameter geometry = NULL, normal = NULL, lighting = NULL;
    vector<CGparameter> vshaders, fshaders;
    vector<CGparameter> vshaders_all, fshaders_all; // matches helper indexing
    // maps the shared parameters to the shader-specific instances
    map<CGparameter, CGparameter> parameter_map;
    for (size_t i = 0; i < helpers.size(); i++)
    {
        CGparameter vp = helpers[i]->begin_vertex_program(vertex_program);
        CGparameter fp = helpers[i]->begin_fragment_program(fragment_program);
        vshaders_all.push_back(vp);
        fshaders_all.push_back(fp);
        switch (helpers[i]->type())
        {
        case RENDER_HELPER_GEOMETRY:
            g_assert(!geometry);
            g_assert(vp && !fp);
            geometry = vp;
            break;
        case RENDER_HELPER_NORMAL:
            g_assert(!normal);
            g_assert(!vp && fp);
            normal = fp;
            break;
        case RENDER_HELPER_LIGHTING:
            g_assert(!lighting);
            g_assert(!vp && fp);
            lighting = fp;
            break;
        case RENDER_HELPER_SHADER:
            if (vp) vshaders.push_back(vp);
            if (fp) fshaders.push_back(fp);
            break;
        case RENDER_HELPER_OTHER:
            g_assert(!vp && !fp);
            break;
        }
    }
    g_assert(geometry && normal && lighting);

    CGparameter vsp, fsp;
    vsp = cgGetNamedParameter(vertex_program, "shaders");
    fsp = cgGetNamedParameter(fragment_program, "shaders");
    g_assert(vsp);
    g_assert(fsp);
    cgSetArraySize(vsp, vshaders.size());
    cgSetArraySize(fsp, fshaders.size());

    parameter_map[geometry] = cgGetNamedParameter(vertex_program, "geometry");
    parameter_map[normal] = cgGetNamedParameter(fragment_program, "normal");
    parameter_map[lighting] = cgGetNamedParameter(fragment_program, "lighting");
    for (size_t i = 0; i < vshaders.size(); i++)
        parameter_map[vshaders[i]] = cgGetArrayParameter(vsp, i);
    for (size_t i = 0; i < fshaders.size(); i++)
        parameter_map[fshaders[i]] = cgGetArrayParameter(fsp, i);
    for (map<CGparameter, CGparameter>::const_iterator i = parameter_map.begin(); i != parameter_map.end(); i++)
    {
        cgConnectParameter(i->first, i->second);
        check_cg_error();
    }
    parameter_map[NULL] = NULL;

    cgCompileProgram(vertex_program);   check_cg_error();
    cgGLLoadProgram(vertex_program);    check_cg_error();
    cgCompileProgram(fragment_program); check_cg_error();
    cgGLLoadProgram(fragment_program);  check_cg_error();

    for (size_t i = 0; i < helpers.size(); i++)
    {
        helpers[i]->begin_state(vertex_program, fragment_program, flags);
        helpers[i]->set_parameters(parameter_map[vshaders_all[i]], parameter_map[fshaders_all[i]]);
        helpers[i]->set_light(diffuse, ambient, direction);
    }

    cgGLEnableProfile(global_cg_vertex_profile());
    cgGLEnableProfile(global_cg_fragment_profile());
    cgGLBindProgram(vertex_program);
    cgGLBindProgram(fragment_program);
#if 0
    cout << "Vertex program:\n"
        << cgGetProgramString(vertex_program, CG_COMPILED_PROGRAM) << "\n";
    cout << "Fragment program:\n"
        << cgGetProgramString(fragment_program, CG_COMPILED_PROGRAM) << "\n";
#endif
}

void RenderClasses::end_state()
{
    for (size_t i = 0; i < helpers.size(); i++)
        helpers[i]->end_state();
    cgGLDisableProfile(global_cg_vertex_profile());
    cgGLDisableProfile(global_cg_fragment_profile());

    /* Nuke the created parameters */
    CGparameter vsp = cgGetNamedParameter(vertex_program, "shaders");
    CGparameter fsp = cgGetNamedParameter(fragment_program, "shaders");
    int vs = cgGetArraySize(vsp, 0);
    int fs = cgGetArraySize(fsp, 0);
    for (int i = 0; i < vs; i++)
    {
        CGparameter dst = cgGetArrayParameter(vsp, i);
        CGparameter src = cgGetConnectedParameter(dst);
        cgDisconnectParameter(dst);
        cgDestroyParameter(src);
    }
    for (int i = 0; i < fs; i++)
    {
        CGparameter dst = cgGetArrayParameter(fsp, i);
        CGparameter src = cgGetConnectedParameter(dst);
        cgDisconnectParameter(dst);
        cgDestroyParameter(src);
    }
    cgDestroyProgram(vertex_program);
    cgDestroyProgram(fragment_program);
}

void RenderObject::render(const Skeleton &skeleton, double lod, double z) const
{
    for (size_t i = 0; i < helpers.size(); i++)
        helpers[i]->begin_object(skeleton, lod, z);

    int ilod = CLAMP((int) floor(lod), 0, (int) levels.size() - 1);
    const RenderMeshLevel &level = levels[ilod];
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, level.index_buffer);
    glDrawRangeElements(GL_TRIANGLES, 0, level.vertices.size() - 1,
                        3 * level.faces.size(), GL_UNSIGNED_INT, NULL);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

    for (int i = (int) helpers.size() - 1; i >= 0; i--)
        helpers[i]->end_object();
}

size_t RenderObject::influences(double lod) const
{
    int ilod = CLAMP((int) floor(lod), 0, (int) levels.size() - 1);
    const RenderMeshLevel &level = levels[ilod];
    return level.influences;
}

size_t RenderObject::vertices(double lod) const
{
    int ilod = CLAMP((int) floor(lod), 0, (int) levels.size() - 1);
    const RenderMeshLevel &level = levels[ilod];
    return level.vertices.size();
}


void RenderInstance::update_animation(double elapsed, double advance_frames)
{
    if (!skeleton.num_frames()) return;

    frame += advance_frames;
    if (viewer->render_options[RENDER_ANIMATE])
        frame += elapsed * skeleton.get_frame_rate();
    frame = fmod(frame, skeleton.num_frames());

    int int_frame = (int) floor(frame);
    while (int_frame < 0) int_frame += skeleton.num_frames();
    while (int_frame >= (int) skeleton.num_frames()) int_frame -= skeleton.num_frames();
    skeleton.set_to_frame(int_frame);
    model_matrix = transform * skeleton.get_model_transform();
}

void RenderInstance::reset_animation()
{
    frame = 0.0;
    skeleton.set_to_frame(0);
    model_matrix = transform * skeleton.get_model_transform();
}

double RenderInstance::compute_lod() const
{
    Matrix4 mv_matrix = viewer->camera.view_matrix * model_matrix;
    // converts from world scale to model scale
    double scale = 1.0 / matrix_norm2(matrix4_to_matrix3(mv_matrix));

    BoundingSphereW view_sphere = mv_matrix * object->bounds;
    double mindepth = -view_sphere.centre[2] - view_sphere.radius;
    double tol = viewer->camera.lod_tolerance_base * viewer->lod_tolerance();
    tol = max(1e-10, mindepth * tol * scale); // world space

    int lod = 0;
    double clod = 0.0;
    int nlevels = object->levels.size();
    while (lod + 1 < nlevels && object->levels[lod + 1].error < tol) lod++;
    if (lod + 1 == nlevels) clod = lod;
    else clod = lod + (tol - object->levels[lod].error) / (object->levels[lod + 1].error - object->levels[lod].error);

    if (viewer->render_options[RENDER_NO_LOD])
        clod = 0.0;
    else if (viewer->render_options[RENDER_DIRECT_LOD])
    {
        clod = viewer->lod_tolerance();
        if (clod >= nlevels) clod = nlevels - 1;
    }

    return clod;
}

double RenderInstance::compute_dist() const
{
    Matrix4 mv_matrix = viewer->camera.view_matrix * model_matrix;
    // converts from world scale to model scale
    double scale = 1.0 / matrix_norm2(matrix4_to_matrix3(mv_matrix));

    BoundingSphereW view_sphere = mv_matrix * object->bounds;
    double mindepth = -view_sphere.centre[2] - view_sphere.radius;
    return max(0.0, mindepth * viewer->camera.lod_tolerance_base * scale);
}

bool RenderInstance::frustum_cull() const
{
    /* Frustum culling */
    BoundingSphereW world_sphere = model_matrix * object->bounds;
    if (!viewer->render_options[RENDER_NO_FRUSTUM_CULL])
    {
        for (int i = 0; i < 6; i++)
            if (viewer->camera.frustum[i].normal * world_sphere.centre < -world_sphere.radius)
                return true;
    }
    return false;
}

void RenderInstance::render() const
{
    if (frustum_cull()) return;
    double lod = compute_lod();
    double z = compute_dist();

    Matrix4 mv_matrix = viewer->camera.view_matrix * model_matrix;
    alodgl_load_matrix(mv_matrix);

    object->render(get_skeleton_read(), lod, z);
}

size_t RenderInstance::influences() const
{
    if (frustum_cull()) return 0;
    double lod = compute_lod();
    return object->influences(lod);
}

size_t RenderInstance::vertices() const
{
    if (frustum_cull()) return 0;
    double lod = compute_lod();
    return object->vertices(lod);
}

RenderInstance::RenderInstance(list<RenderObject>::iterator object,
                               const ModelTransform &transform,
                               RenderViewer *viewer)
: object(object), transform(transform.spatial), skeleton(object->mesh.skeleton.get_read()), viewer(viewer)
{
    frame = transform.temporal * skeleton.get_frame_rate();

    if (skeleton.num_frames() != 0)
    {
        int int_frame = ((int) floor(frame)) % (int) skeleton.num_frames();
        if (int_frame < 0) int_frame += skeleton.num_frames();
        skeleton.set_to_frame(int_frame);
    }
    model_matrix = transform.spatial * this->skeleton.get_model_transform();
}
