#ifndef ALOD_RENDER_RENDERMESH_H
#define ALOD_RENDER_RENDERMESH_H

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "geom/bounding.h"
#include "geom/coords.h"
#include "geom/skeleton.h"
#include "render/meshrender.h"
#include "render/interfaces.h"
#include <vector>
#include <list>
#include <string>
#include <utility>
#include <Cg/cg.h>
#include <GL/gl.h>

class RenderHelper;
class RenderHelperClass;
class RenderObject;
class RenderViewer;

struct RenderMeshFace
{
    MeshVertexIterator v[3];
};

struct RenderMeshLevel
{
    GLuint index_buffer;
    //! A list of base, target pairs for the vertices at this level
    std::vector<std::pair<MeshVertex *, MeshVertex *> > vertices;
    std::vector<RenderMeshFace> faces;
    size_t influences;
    double error;
};

class RenderMesh
{
public:
    //! A copy of the underlying mesh, maximally decimated
    Mesh mesh;
    std::vector<RenderMeshLevel> levels;
    BoundingSphereW bounds;

    int int_lod(double lod) const
    { return CLAMP((int) floor(lod), 0, (int) levels.size() - 1); }

    void load_mesh(const Mesh &m);
    void create_indices(int flags);
    RenderMesh() {}
    explicit RenderMesh(const Mesh &m) { load_mesh(m); }
};

class RenderClasses
{
private:
    int flags;
    std::vector<RenderHelperClass *> helpers;
    CGprogram vertex_program, fragment_program;

public:
    RenderClasses();
    ~RenderClasses();

    void add_helper(RenderHelperClass *h);
    void create_instances(RenderObject &r);
    void begin_state();
    void end_state();
};

//! Structure representing a spatial transform plus a time offset
struct ModelTransform
{
    Matrix4 spatial;
    double temporal;

    ModelTransform() : spatial(), temporal(0.0) {}
    ModelTransform(const Matrix4 &s, double t) : spatial(s), temporal(t) {}
    ModelTransform operator *(const ModelTransform &b) const
    { return ModelTransform(spatial * b.spatial, temporal + b.temporal); }
    ModelTransform &operator *=(const ModelTransform &b)
    { spatial *= b.spatial; temporal += b.temporal; return *this; }
};

struct RenderObject : public RenderMesh
{
    std::vector<RenderHelper *> helpers;
    std::string mesh_filename;
    std::string name;

    Matrix4 normalise_transform;
    ModelTransform default_transform;

    void render(const Skeleton &skeleton, double lod, double z) const;
    size_t influences(double lod) const; //!< Number of influences at this LOD
    size_t vertices(double lod) const;   //!< Number of vertices at this LOD
};

class RenderInstance
{
private:
    std::list<RenderObject>::iterator object;
    Matrix4 transform;
    Matrix4 model_matrix;
    Skeleton skeleton;
    double frame;
    RenderViewer *viewer;

    bool frustum_cull() const;

    //! Returns a real-valued LOD level based on the current camera
    double compute_lod() const;
    //! Computes a distance metric based on the current camera
    /*! The value can be multiplied by a pixel count to obtain a
     * lower bound on the equivalent distance in model space.
     */
    double compute_dist() const;
public:
    RenderObject &get_object() { return *object; }
    const RenderObject &get_object() const { return *object; }
    const Skeleton &get_skeleton_read() const { return skeleton; }
    Skeleton &get_skeleton_write() { return skeleton; }
    void update_animation(double elapsed, double advance_frames);
    void reset_animation();

    void render() const;
    size_t influences() const;  //!< Number of influences that would be rendered
    size_t vertices() const;    //!< Number of vertices that would be rendered

    RenderInstance() {}
    RenderInstance(std::list<RenderObject>::iterator object,
                   const ModelTransform &transform,
                   RenderViewer *viewer);
};

#endif /* ALOD_RENDER_RENDERMESH_H */
