#!BPY

""" Registration info for Blender menus:
Name: 'Alod model (.alm)...'
Blender: 241
Group: 'Export'
Tooltip: 'Export to Alod model (*.alm)'
"""

__author__ = "Bruce Merry"
__version__ = "ALM version 10"
__bpydoc__ = """\
This script exports to the model format used by the authors system for
skeletally-animated level of detail.

Usage:

Select a mesh to be exported and run this script from "File->Export" menu.
If an armature is used, it should not be forced to rest position.
"""

import Blender, struct, as_utils
from Blender import *
from Blender.Mathutils import *
from as_utils import *

def alm_write_int(file, number):
        file.write(struct.pack('>i', number))

def alm_write_uint(file, number):
        file.write(struct.pack('>I', number))

def alm_write_float(file, number):
        file.write(struct.pack('>f', number))

def alm_write_vec3(file, vec):
        for i in range(3):
                alm_write_float(file, vec[i])

def alm_write_mat4(file, mat):
        for i in range(4):
                for j in range(4):
                        alm_write_float(file, mat[j][i])

def fc_key(mesh, face, corner):
        if mesh.faceUV:
                return (mesh.faces[face].verts[corner].index, tuple(mesh.faces[face].uv[corner]))
        else:
                return mesh.faces[face].verts[corner].index
        return mesh.faces[face].verts[corner].index

def add_position(coords, index, vec):
        if not coords.has_key(index):
                coords[index] = Vector([0.0, 0.0, 0.0, 0.0])
        coords[index] = coords[index] + vec

def fs_callback(filename):
        as_version = 1

	selected = Object.GetSelected()
        if not selected:
                Draw.PupMenu("Error|No object selected")
                return
        mesh_obj = selected[0]
        if mesh_obj.getType() != 'Mesh':
                Draw.PupMenu("Error|Selected object is not a mesh")
                return
        if as_utils.get_version(mesh_obj) != as_version:
                Draw.PupMenu("Error|Selected object is not an animation space model")
                return None
        
        mesh = mesh_obj.getData(False, True)  # Get a Mesh object
        obj_mat = mesh_obj.getMatrix()
        scene = Scene.GetCurrent()
        context = scene.getRenderingContext()

        arm = None
        has_arm = 0
        rest_bones = {}
        pose_bones = {}
        if mesh_obj.getParent() is not None:
                if mesh_obj.getParent().getType() == 'Armature':
                        has_arm = 1
                        arm_obj = mesh_obj.getParent()
                        arm = arm_obj.getData()
                        arm2mesh = arm_obj.getMatrix() * mesh_obj.getInverseMatrix()
                        rest_bones = arm.bones
                        pose_bones = arm_obj.getPose().bones
        influences = {}
        for b in rest_bones.keys():
                influences[b] = get_influence_coords(mesh, b)
        influences['Root'] = get_influence_coords(mesh, 'Root')

        num_tris = 0
        vertices = []     # List of tuples (face, corner)
        vertex_dict = {}  # Maps (vertex, uv) tuples to vertices
        # Construct vertices
        # At the same time, count triangles
        for f, face in enumerate(mesh.faces):
                num_tris += len(face.verts) - 2
                for i in range(len(face.verts)):
                        key = fc_key(mesh, f, i)
                        if not vertex_dict.has_key(key):
                                vertex_dict[key] = len(vertices)
                                vertices.append((f, i))

        if filename.find('.alm', -4) <= 0: filename += '.alm'
	file = open(filename, "wb")
        file.write(struct.pack('>4s', 'ALM'))
        alm_write_uint(file, 10)  # version; remember to update __version__

        alm_write_uint(file, 1)                  # comments
        if has_arm:
                alm_write_uint(file, len(rest_bones.keys()) + 1)  # bones
        else:
                alm_write_uint(file, 1)
        alm_write_uint(file, len(mesh.verts))    # positions
        alm_write_uint(file, len(vertices))      # vertices
        alm_write_uint(file, 0)                  # textures
        alm_write_uint(file, 0)                  # tangents
        alm_write_uint(file, 1)                  # materials
        alm_write_uint(file, num_tris)           # triangles
        alm_write_uint(file, 0)                  # collapses
        if has_arm:
                alm_write_uint(file, context.endFrame() - context.startFrame() + 1) # frames
        else:
                alm_write_uint(file, 0)
        alm_write_float(file, context.framesPerSec()) # frame rate
        alm_write_uint(file, 0)                       # override_P

        # comments
        comment = 'Created by alm_export.py'
        alm_write_uint(file, len(comment))
        file.write(struct.pack('>' + `len(comment)` + 's', comment))

        # bones
        # Regardless of any armature, we have a super-root bone
        # that applies object transforms.
        alm_write_int(file, -1)
        alm_write_mat4(file, obj_mat)
        alm_write_uint(file, 0)
        alm_write_vec3(file, Vector(0.0, 1.0, 0.0))
        alm_write_float(file, 1.0)
        alm_write_float(file, 0.0)
        alm_write_float(file, 0.0)
        alm_write_float(file, 0.0)
        alm_write_float(file, 0.0)
        alm_write_mat4(file, obj_mat)
        alm_write_mat4(file, obj_mat)
        alm_write_float(file, 0.0)
        if has_arm:
                bone_dict = {}
                inv_dict = {}
                order = rest_bones.items()
                for i, (name, bone) in enumerate(order):
                        bone_dict[name] = i + 1
                for name, bone in order:
                        mat = bone.matrix['ARMATURESPACE']
                        if bone.hasParent():
                                mat = mat * inverse(bone.parent.matrix['ARMATURESPACE'])
                        else:
                                mat = mat * arm2mesh
                        inv_dict[bone.name] = inverse(bone.matrix['ARMATURESPACE'] * arm2mesh)

                        if bone.hasParent():
                                alm_write_int(file, bone_dict[bone.parent.name])
                        else:
                                alm_write_int(file, 0)

                        alm_write_mat4(file, mat)
                        alm_write_uint(file, 0)          # flags
                        alm_write_vec3(file, extend(bone.tail['ARMATURESPACE'], 1.0) * inverse(bone.matrix['ARMATURESPACE']))
                        alm_write_float(file, 1.0)       # max_singular; may not be true
                        alm_write_float(file, mat[3][0])
                        alm_write_float(file, mat[3][1]) # translation bounding sphere
                        alm_write_float(file, mat[3][2]) # may not be true
                        alm_write_float(file, 0.0)

                        # Expectations; we don't have the flags for them so they don't matter
                        alm_write_mat4(file, mat)
                        alm_write_mat4(file, mat)
                        alm_write_float(file, 0.0)

        # positions
        for v in mesh.verts:
                coords = {}
                if influences['Root'].has_key(v.index):
                        add_position(coords, 0, influences['Root'][v.index])
                for bone in rest_bones.keys():
                        if influences[bone].has_key(v.index):
                                add_position(coords, bone_dict[bone], influences[bone][v.index])
                alm_write_uint(file, len(coords))
                for key, pos in coords.iteritems():
                        alm_write_uint(file, key)
                        alm_write_float(file, pos[0])
                        alm_write_float(file, pos[1])
                        alm_write_float(file, pos[2])
                        alm_write_float(file, pos[3])

        # vertices
        for vert in vertices:
                f = vert[0]
                i = vert[1]
                alm_write_uint(file, mesh.faces[f].verts[i].index)  # position index
                alm_write_uint(file, 0)   # Tangent vectors
                alm_write_uint(file, 0)
                if mesh.faceUV:
                        alm_write_float(file, mesh.faces[f].uv[i][0])         # s
                        alm_write_float(file, 1.0 - mesh.faces[f].uv[i][1])   # t
                else:
                        alm_write_float(file, 0.0)
                        alm_write_float(file, 0.0)

        # textures
        # tangents

        # materials
        alm_write_int(file, -1)
        alm_write_int(file, -1)

        # triangles
        for f, face in enumerate(mesh.faces):
                for i in range(len(face.verts) - 2):
                        for j in (0, i + 1, i + 2):
                                key = fc_key(mesh, f, j)
                                alm_write_uint(file, vertex_dict[key])   # vertex indices
                        alm_write_uint(file, 0)          # material

        # collapses

        # frames
        if has_arm:
                for f in range(context.startFrame(), context.endFrame() + 1):
                        arm_obj.evaluatePose(f)
                        obj_mat = mesh_obj.getMatrix()
                        arm2mesh = arm_obj.getMatrix() * mesh_obj.getInverseMatrix()
                        pose_bones = arm_obj.getPose().bones
                        alm_write_mat4(file, obj_mat) # super-root
                        for name, rest_bone in order:
                                bone = pose_bones[name]
                                mat = bone.poseMatrix
                                if rest_bone.hasParent():
                                        mat = mat * inverse(pose_bones[rest_bone.parent.name].poseMatrix)
                                else:
                                        mat = mat * arm2mesh
                                alm_write_mat4(file, mat)

        file.close()

def createALMPath():
        filename = Blender.Get('filename')
        if filename.find('.') != -1:
                return filename.split('.')[0] + ".alm"
        return filename

Window.FileSelector(fs_callback, "Export ALM", createALMPath());
