#!BPY

""" Registration info for Blender menus:
Name: 'Alod model (ALM)'
Blender: 241
Group: 'Import'
Submenu: 'Without sphere...' mesh
Submenu: 'With bounding sphere...' mesh_sphere
Tooltip: 'Import an animation level-of-detail (ALM) file.'
"""

__author__ = "Bruce Merry"
__version__ = "2007/05/24"
__bpydoc__ = """\
This script imports from the Animated LOD Mesh (ALM) format.
"""

# Copyright (C) 2005-2007: Bruce Merry, bmerry@cs.uct.ac.za
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.


# Portions of this code are taken from mod_meshtools.py in Blender
# 2.32.

import Blender
import meshtools, as_utils
import re, struct, StringIO, sys, string, os.path, math
from Blender import *
from Blender.Mathutils import *
from as_utils import *

MODE = 'mesh'

class AlmBone:
        def __init__(self, index):
                self.name = bone_name(index)
                self.index = index
                self.rest_transform = None
                self.abs_rest_transform = None
                self.parent = None
                self.posed_transforms = []
                self.abs_posed_transforms = []
                self.tail = Vector(0.0, 1.0, 0.0)
                self.reorient_transform = None

        def make_absolute(self):
                if self.abs_rest_transform is None:
                        if self.parent is not None:
                                self.parent.make_absolute()
                                self.abs_rest_transform = self.rest_transform * self.parent.abs_rest_transform
                                for i, t in enumerate(self.posed_transforms):
                                        self.abs_posed_transforms[i] = t * self.parent.abs_posed_transforms[i]
                        else:
                                self.abs_rest_transform = Matrix(self.rest_transform)
                                self.abs_posed_transforms = [Matrix(t) for t in self.posed_transforms]

def alm_read_int(file):
        return struct.unpack('>i', file.read(4))[0]

def alm_read_uint(file):
        return struct.unpack('>I', file.read(4))[0]

def alm_read_float(file):
        return struct.unpack('>f', file.read(4))[0]

def alm_read_vec3(file):
        x = alm_read_float(file)
        y = alm_read_float(file)
        z = alm_read_float(file)
        return Vector(x, y, z)

def alm_read_mat4(file):
        lists = [[], [], [], []]
        for i in range(4):
                for j in range(4):
                        lists[j].append(alm_read_float(file))
        return Matrix(lists[0], lists[1], lists[2], lists[3])

def alm_read_vector_element(file):
        e = AsVectorElement()
        e.bone = alm_read_uint(file)
        e.values = Vector([alm_read_float(file),
                           alm_read_float(file),
                           alm_read_float(file),
                           alm_read_float(file)])
        return e

def alm_read_vector(file):
        vec = AsVector()
        count = alm_read_uint(file)
        for j in range(count):
                vec.elements.append(alm_read_vector_element(file))
        return vec

def filesel_callback(filename):
        try:
                Window.DrawProgressBar(0.0, "Reading file")
                file = open(filename, "rb")
                signature = file.read(4)
                if (signature != "ALM\0"):
                        print "Signature line was invalid"
                        return
                version = alm_read_uint(file)
                if (version < 6 or version > 10):
                        print "Version is ", version
                        print "I only support versions 6 through 10"
                        return
                num_comments = alm_read_uint(file)
                num_bones = alm_read_uint(file)
                num_positions = alm_read_uint(file)
                num_vertices = alm_read_uint(file)
                num_textures = alm_read_uint(file)
                if (version >= 8): num_tangents = alm_read_uint(file)
                num_materials = alm_read_uint(file)
                num_triangles = alm_read_uint(file)
                num_collapses = alm_read_uint(file)
                num_frames = alm_read_uint(file)
                frame_rate = alm_read_float(file)
                override_P = alm_read_uint(file)

                comments = []
                for i in range(num_comments):
                        length = alm_read_uint(file)
                        comments.append(file.read(length))

                bones = []
                parents = []
                for i in range(num_bones):
                        bones.append(AlmBone(i))
                        parents.append(alm_read_int(file))
                        bones[i].rest_transform = alm_read_mat4(file)
                        alm_read_uint(file)  # flags
                        if version >= 10:
                                bones[i].tail = alm_read_vec3(file)
                        alm_read_float(file) # max singular
                        for j in range(4):   # translation sphere
                                alm_read_float(file)
                        for j in range(33):  # expectations
                                alm_read_float(file)
                for i in range(num_bones):
                        if parents[i] >= 0:
                                bones[i].parent = bones[parents[i]]

                if override_P:
                        for j in range(16 * num_bones):
                                alm_read_float(file)

                positions = []
                for i in range(num_positions):
                        positions.append(alm_read_vector(file))

                uvs = []
                vertices = []
                for i in range(num_vertices):
                        vertices.append(alm_read_uint(file)) # Position index
                        if version >= 7:
                                alm_read_vector(file) # Tangents
                                alm_read_vector(file)
                        uv = Vector(alm_read_float(file), 1.0 - alm_read_float(file))
                        uvs.append(uv)

                for i in range(num_textures):
                        width = alm_read_uint(file)
                        height = alm_read_uint(file)
                        channels = alm_read_uint(file)
                        alm_read_uint(file)                  # signed
                        file.read(width * height * channels) # data

                if (version >= 8):
                        for i in range(num_tangents):
                                width = alm_read_uint(file)
                                height = alm_read_uint(file)
                                for j in range(width * height):
                                        alm_read_int(file)           # patch
                                        alm_read_vector(file)        # tangents
                                        alm_read_vector(file)

                for i in range(num_materials):
                        if version < 8:
                                alm_read_int(file)           # texture
                                num_weights = alm_read_uint(file)
                                for i in range(num_weights): # bones
                                        alm_read_uint(file)
                                for i in range(num_weights): # maps
                                        alm_read_uint(file)
                        else:
                                alm_read_int(file)           # texture
                                alm_read_int(file)           # tangent map

                triangles = []
                for i in range(num_triangles):
                        indices = (alm_read_uint(file),
                                   alm_read_uint(file),
                                   alm_read_uint(file));
                        alm_read_uint(file)                  # material
                        triangles.append(indices)

                for i in range(num_collapses):
                        p1 = alm_read_uint(file)             # p1
                        if p1 == 0xffffffff:
                                count = alm_read_uint(file)
                                for j in range(count):
                                        alm_read_uint(file)  # old vertex
                                        alm_read_uint(file)  # new vertex
                        else:
                                alm_read_uint(file)          # p2
                                alm_read_uint(file)          # vl
                                alm_read_uint(file)          # vr
                        alm_read_float(file)                 # error

                for i in range(num_frames):
                        for j in range(num_bones):
                                bones[j].posed_transforms.append(alm_read_mat4(file))
                                bones[j].abs_posed_transforms.append(None)
        except IOError, (errno, strerror):
                file.close()
                Window.DrawProgressBar(1.0, '')
                return
        file.close()

        ### Start creating Blender data

        model_name = Blender.sys.splitext(Blender.sys.basename(filename))[0]
        mesh = Mesh.New(model_name)
        mesh_obj = Object.New('Mesh', model_name)
        mesh_obj.link(mesh)
        Scene.GetCurrent().link(mesh_obj)
        arm_obj = Object.New('Armature', 'Armature.' + model_name)
        arm = Armature.Armature('Armature.' + model_name)
        arm_obj.link(arm)
        Scene.GetCurrent().link(arm_obj)

        Window.DrawProgressBar(0.0, 'Processing data')

        # Ensure that at least one frame exists
        if num_frames == 0:
                for b in bones:
                        b.posed_transforms = [Matrix(b.rest_transform)]
                        b.abs_posed_transforms = None
                num_frames = 1

        arm_obj.makeParent([mesh_obj])

        # Eliminate root bone by transferring it to object keys
        scene = Blender.Scene.GetCurrent()
        context = scene.getRenderingContext()
        for i in range(num_frames):
                context.currentFrame(i + 1)
                arm_obj.setMatrix(bones[0].posed_transforms[i])
                arm_obj.insertIpoKey(Object.LOCROTSIZE)
                bones[0].posed_transforms[i].identity()
        bones[0].rest_transform.identity()

        # Convert relative to absolute transforms
        for b in bones:
                b.make_absolute()

        metaball_dynamic = None
        if string.find(MODE, 'sphere') != -1:
                Window.DrawProgressBar(0.0, 'Creating bounding spheres')
                for c in comments:
                        match = re.match(r"%BoundingSphere: \(([e\d.-]+), ([e\d.-]+), ([e\d.-]+)\), ([e\d.-]+)", c)
                        if match is not None:
                                metaball_obj = Object.New('Mball', 'AsSphere.' + model_name)
                                metaball = Metaball.New('AsSphere.' + model_name)
                                # Experimentally, setting the radius to twice the actual radius
                                # and the stiffness to sqrt(2) gives a metaball of the desired
                                # radius
                                metaball.addMetaelem([0,                     # Sphere
                                                      0.0,
                                                      0.0,                   # Centre
                                                      0.0,
                                                      2.0 * float(match.group(4)), # Radius
                                                      1,                     # 'Lay'???
                                                      math.sqrt(2.0),        # Stiffness
                                                      1.0, 1.0, 1.0])        # 'Exp'???
                                metaball_obj.link(metaball)
                                Scene.GetCurrent().link(metaball_obj)
                                mesh_obj.makeParent([metaball_obj], 1)
                                metaball_obj.setLocation(float(match.group(1)),
                                                         float(match.group(2)),
                                                         float(match.group(3)))

                        match = re.match(r"%BoundingSphereN: \(([\[\]e\d., -]+)\), ([e\d.-]+)", c)
                        if match is not None:
                                coords = string.split(match.group(1), ", ")
                                pos = AsVector()
                                for i in range(len(coords)):
                                        if coords[i][0] != '[':
                                                bone = i / 4
                                                dim = i % 4
                                                if not pos.elements or pos.elements[-1].bone != bone:
                                                        pos.elements.append(AsVectorElement())
                                                        pos.elements[-1].bone = bone
                                                pos.elements[-1].values[dim] = float(coords[i])
                                rest_pos = project(pos, [b.abs_rest_transform for b in bones])
                                positions.append(pos)

                                metaball_obj = Object.New('Mball', 'AsSphereN.' + model_name)
                                metaball = Metaball.New('AsSphereN.' + model_name)
                                metaball.addMetaelem([0,                     # Sphere
                                                      0.0,
                                                      0.0,                   # Centre
                                                      0.0,
                                                      2.0 * float(match.group(2)), # Radius
                                                      1,                     # 'Lay'???
                                                      math.sqrt(2.0),        # Stiffness
                                                      1.0, 1.0, 1.0])        # 'Exp'???
                                metaball_obj.link(metaball)
                                metaball_dynamic = metaball_obj
                                Scene.GetCurrent().link(metaball_obj)
                                mesh_obj.makeParent([metaball_obj], 1)
                                metaball_obj.setLocation(rest_pos[0], rest_pos[1], rest_pos[2])

        Window.DrawProgressBar(0.0, 'Creating skeleton')
        arm.makeEditable()
        for i in range(1, len(bones)):
                name = bone_name(i)
                arm.bones[name] = bone = Armature.Editbone()
                bone.head = bones[i].abs_rest_transform.translationPart()
                tail = contract(extend(bones[i].tail, 1.0) * bones[i].abs_rest_transform)
                if (tail - bone.head).length < 1e-6:
                        tail = tail + Vector(0.0, 0.0, 1.0)
                bone.tail = tail
                bone.roll = 0
        for i in range(1, len(bones)):
                name = bone_name(i)
                bone = bones[i]
                if bone.parent and bone.parent.index > 0:
                        arm.bones[name].parent = arm.bones[bone_name(bone.parent.index)]
                        if version < 10 and (arm.bones[name].parent.head - arm.bones[name].head).length > 1e-3:
                                arm.bones[name].parent.tail = arm.bones[name].head
        # In Blender, the matrix of a bone is closely tied to the head-tail
        # direction. Instead of enforcing our matrix on the bone, we see
        # what the existing matrix is and make appropriate adjustments to
        # the per-vertex data
        arm.update()
        for i in range(1, len(bones)):
                bone = arm.bones[bone_name(i)]
                bones[i].reorient_transform = bones[i].abs_rest_transform * inverse(bone.matrix['ARMATURESPACE'])

        Window.DrawProgressBar(0.0, 'Creating vertices')
        for pos in positions:
                pos = project(pos, [b.abs_rest_transform for b in bones])
                mesh.verts.extend(contract(pos))

        Window.DrawProgressBar(0.0, 'Creating triangles')
        # We create the triangles in one go then update them, since
        # the backend has quadratic-time performance otherwise
        triangle_list = []
        for t in triangles:
                triangle_list.append([mesh.verts[vertices[x]] for x in t])
        mesh.faces.extend(triangle_list)
        mesh.faceUV = 1
        for t, f in zip(triangles, mesh.faces):
                f.uv = [uvs[x] for x in t]
                f.mode = 0
                f.smooth = 1

        Window.DrawProgressBar(0.0, 'Creating vertex groups')
        for b in bones:
                name = b.name
                add_influence_groups(mesh, name)
        for p, pos in enumerate(positions):
                for e in pos.elements:
                        name = bones[e.bone].name
                        vec = e.values * bones[e.bone].reorient_transform
                        set_influence_coords(mesh, name, {p: vec})

        Window.DrawProgressBar(0.0, 'Setting pose keys')
        action = Armature.NLA.NewAction('Action.' + model_name)
        action.setActive(arm_obj)

        pose = arm_obj.getPose()
        for f in range(num_frames):
                for i in range(1, len(bones)):
                        rest_bone = arm.bones[bones[i].name]
                        pose_bone = pose.bones[bones[i].name]
                        mat = inverse(bones[i].reorient_transform) * bones[i].posed_transforms[f] * inverse(bones[i].rest_transform) * bones[i].reorient_transform
                        if mat.determinant() < 0:
                                pose_bone.size = Vector(1, -1, 1)
                                mat[1][0] = -mat[1][0]
                                mat[1][1] = -mat[1][1]
                                mat[1][2] = -mat[1][2]
                                print "Inversion!"
                        pose_bone.loc = mat.translationPart()
                        pose_bone.quat = mat.rotationPart().toQuat()
                        pose_bone.insertKey(arm_obj, f + 1, [Object.Pose.LOC, Object.Pose.ROT, Object.Pose.SIZE])

        as_utils.set_version(mesh_obj, 1)

        Window.DrawProgressBar(0.0, 'Setting frame count')
        context.startFrame(1)
        context.endFrame(int(num_frames))

        Window.DrawProgressBar(0.0, 'Setting initial positions')
        # Load scriptlink
        try:
                text = Text.Get('as_animate.py')
        except:
                scriptdir = Blender.Get('scriptsdir')
                text = Text.Load(os.path.join(scriptdir, 'as_animate.py'))
                text.setName('as_animate.py')
        mesh_obj.addScriptLink(text.getName(), 'FrameChanged')
        if metaball_dynamic is not None:
                metaball_dynamic.addScriptLink(text.getName(), 'FrameChanged')
        Blender.Set('curframe', 1)

        message = "Successfully imported " + Blender.sys.basename(filename)
        meshtools.print_boxed(message)
        Blender.Redraw()
        Window.DrawProgressBar(1.0, '')

def createALMPath():
        filename = Blender.Get('filename')
        if filename.find('.') != -1:
                return filename.split('.')[0] + ".alm"
        return filename

MODE = __script__['arg']
Window.FileSelector(filesel_callback, "Import ALM", createALMPath())
