#!BPY

""" Registration info for Blender menus:
Name: 'Animation space tools'
Blender: 241
Group: 'Animation'
Tooltip: 'Tools for manipulating an animation space model'
"""

__author__ = "Bruce Merry"
__version__ = "2007/01/22"
__bpydoc__ = """\
Tools for manipulating an animation space model
"""

import Blender, as_utils, os, sys
from Blender import *
from Blender.Mathutils import *
from as_utils import *

as_version = 1

def get_selected(as_required):
        selected = Object.GetSelected()
        if not selected:
                Draw.PupMenu("Error|No object selected")
                return None
        mesh_obj = selected[0]
        if as_required and as_utils.get_version(mesh_obj) != as_version:
                Draw.PupMenu("Error|Selected object is not an animation space model")
                return None
        if mesh_obj.getType() != 'Mesh':
                Draw.PupMenu("Error|Selected object is not a mesh")
                return None
        arm_obj = mesh_obj.getParent()
        if arm_obj is None:
                Draw.PupMenu("Error|Selected object has no parent")
                return None
        if arm_obj.getType() != 'Armature':
                Draw.PupMenu("Error|Selected object has no armature")
                return None
        return mesh_obj

def do_adjust():
        mesh_obj = get_selected(1)
        if mesh_obj is None: return
        arm_obj = mesh_obj.getParent()

        in_editmode = Window.EditMode()
        if in_editmode: Window.EditMode(0)
        mesh = mesh_obj.getData(False, True)
        arm = arm_obj.getData()

        coords = {}
        new_coords = {}
        for bone in arm.bones.values():
                coords[bone.name] = get_influence_coords(mesh, bone.name)
                new_coords[bone.name] = {}
        pose = arm_obj.getPose()
        for v in mesh.verts:
                i = v.index
                pos = AsVector()
                ref = Vector(0.0, 0.0, 0.0, 0.0)
                for bone in arm.bones.values():
                        if coords[bone.name].has_key(i):
                                ref += coords[bone.name][i] * pose.bones[bone.name].poseMatrix
                delta = extend(v.co, 1.0) - ref
                if delta.length > 1e-8:
                        for bone in arm.bones.values():
                                if coords[bone.name].has_key(i):
                                        adj = coords[bone.name][i] + coords[bone.name][i][3] * delta * inverse(pose.bones[bone.name].poseMatrix)
                                        new_coords[bone.name][i] = adj
        for bone in arm.bones.values():
                set_influence_coords(mesh, bone.name, new_coords[bone.name])

        if in_editmode: Window.EditMode(1)

def do_convert():
        mesh_obj = get_selected(0)
        if mesh_obj is None: return
        if as_utils.get_version(mesh_obj) != 0:
                Draw.PupMenu("Error|Selected object is already converted")
                return
        arm_obj = mesh_obj.getParent()

        in_editmode = Window.EditMode()
        if in_editmode: Window.EditMode(0)
        mesh = mesh_obj.getData(False, True)
        arm = arm_obj.getData()
        arm2obj = arm_obj.getMatrix() * mesh_obj.getInverseMatrix()

        groups = mesh.getVertGroupNames()
        total_weights = [0.0 for v in mesh.verts]
        for bone in arm.bones.values():
                if bone.name in groups:
                        group = mesh.getVertsFromGroup(bone.name, 1)
                        for v, w in group:
                                total_weights[v] += w

        for bone in arm.bones.values():
                add_influence_groups(mesh, bone.name)
                if bone.name in groups:
                        group = mesh.getVertsFromGroup(bone.name, 1)
                        weights = {}
                        tw = 0.0
                        for v, w in group: weights[v] = w
                        for v in mesh.verts:
                                if weights.has_key(v.index) and weights[v.index] > 1e-8:
                                        w = weights[v.index] / total_weights[v.index]
                                        el = w * extend(v.co, 1.0) * inverse(bone.matrix['ARMATURESPACE'] * arm2obj)
                                        set_influence_coords(mesh, bone.name, {v.index: el})
        add_influence_groups(mesh, 'Root')
        for v in mesh.verts:
                if total_weights[v.index] == 0.0:
                        set_influence_coords(mesh, 'Root', {v.index: extend(v.co, 1.0)})

        # Cancel the deformation, while keeping parenting
        # The computations above also account for the armature-to-object
        # offset, so we have to clear those from here.
        for modifier in mesh_obj.modifiers:
                if modifier.type == Modifier.Type.ARMATURE:
                        mesh_obj.modifiers.remove(modifier)
        mesh_obj.clrParent()
        mesh_obj.setLocation(0.0, 0.0, 0.0)
        mesh_obj.setSize(1.0, 1.0, 1.0)
        try:
                mesh_obj.setEuler(0.0, 0.0, 0.0)
        except RuntimeError:
                mesh_obj.setEuler(Euler(0.0, 0.0, 0.0))
        arm_obj.makeParent([mesh_obj], 1)

        try:
                text = Text.Get('as_animate.py')
        except:
                scriptdir = Blender.Get('scriptsdir')
                text = Text.Load(os.path.join(scriptdir, 'as_animate.py'))
                text.setName('as_animate.py')
        mesh_obj.addScriptLink(text.getName(), 'FrameChanged')
        Blender.Set('curframe', Blender.Get('curframe'))  # Force script link to activate

        as_utils.set_version(mesh_obj, as_version)

        if in_editmode: Window.EditMode(1)

def do_unconvert():
        mesh_obj = get_selected(1)
        if mesh_obj is None: return

def do_settails():
        mesh_obj = get_selected(1)
        if mesh_obj is None: return
        arm_obj = mesh_obj.getParent()

        in_editmode = Window.EditMode()
        if in_editmode: Window.EditMode(0)
        mesh = mesh_obj.getData(False, True)
        arm = arm_obj.getData()

        old = {}
        for bone in arm.bones.values():
                old[bone.name] = bone.matrix['ARMATURESPACE']
        arm.makeEditable()
        for bone in arm.bones.values():
                if bone.hasParent():
                        bone.parent.tail = bone.head
        arm.update()
        for bone in arm.bones.values():
                coords = get_influence_coords(mesh, bone.name)
                adjust = old[bone.name] * inverse(bone.matrix['ARMATURESPACE'])
                print adjust
                print
                for v in coords.keys():
                        coords[v] = coords[v] * adjust
                set_influence_coords(mesh, bone.name, coords)

        if in_editmode: Window.EditMode(1)

def do_reduce():
        mesh_obj = get_selected(1)
        if mesh_obj is None: return
        arm_obj = mesh_obj.getParent()

        in_editmode = Window.EditMode()
        if in_editmode: Window.EditMode(0)
        mesh = mesh_obj.getData(False, True)
        arm = arm_obj.getData()

        for bone in arm.bones.values():
                coords = get_influence_coords(mesh, bone.name)
                set_influence_coords(mesh, bone.name, coords)
        coords = get_influence_coords(mesh, 'Root')
        set_influence_coords(mesh, 'Root', coords)

handlers = [("Convert", "Convert a Blender-animated mesh to an animation-space mesh", do_convert, None),
            ("Reduce", "Cancel the positive and negative channels", do_reduce, None),
            ("Adjust", "Apply editing to the weight channels", do_adjust, None),
            ("Set tails", "Modify bones to make endpoints meet", do_settails, None),
            ("Exit", "Exit the script", Draw.Exit, None)
           ]
buttons = [None for x in handlers]

def draw_handler():
        global buttons
        global handlers

        BGL.glClearColor(0, 0.5, 0.5, 1)
        BGL.glClear(BGL.GL_COLOR_BUFFER_BIT)
        BGL.glColor3f(1, 1, 1)
        x = 20
        y = len(handlers) * 20 - 18
        for i, handler in enumerate(handlers):
                buttons[i] = Draw.PushButton(handler[0], i, x, y, 80, 20, handler[1])
                y -= 20

def event_handler(evt, val):
        pass

def button_handler(evt):
        handlers[evt][2]()

Draw.Register(draw_handler, event_handler, button_handler)
