#!BPY

""" Registration info for Blender menus:
Name: 'EPS...'
Blender: 236
Group: 'Export'
Tooltip: 'Export to EPS (*.eps)'
"""

import Blender
import Blender.Mathutils

scale = 800

# Matrix utilities
def inverse(mat):
        mat2 = Blender.Mathutils.CopyMat(mat)
        mat2.invert()
        return mat2

def extend(vec):
        return Blender.Mathutils.Vector([vec[0], vec[1], vec[2], 1.0])

# Applies camera transform and perspective projection
def ctransform(vertex, transform):
        v = Blender.Mathutils.VecMultMat(extend(vertex), transform)
        return (scale * v[0] / -v[2], scale * v[1] / -v[2])

# Returns the head and tail of a bone, in the model space of the
# armature. Blender stores them in the parent's space, for the rest pose.
# We convert the tail back to our own space, then apply the dynamic
# transformation to model space (don't believe the name getRestMatrix).
def bone_ends_blender(bone):
        unmap = inverse(bone.getRestMatrix('bonespace'))
        head = Blender.Mathutils.Vector([0, 0, 0, 1])
        tail = extend(Blender.Mathutils.VecMultMat(extend(bone.getTail()) - extend(bone.getHead()), unmap))
        A = bone.getRestMatrix('worldspace')
        return (Blender.Mathutils.VecMultMat(head, A), Blender.Mathutils.VecMultMat(tail, A))

def bone_transform_alm(obj, boneid, frame):
        t = [[0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
        for i in range(4):
                for j in range(4):
                        tname = 'almtrans_' + str(frame) + '_' + str(boneid) + '_' + str(i) + '_' + str(j)
                        t[i][j] = float(obj.getProperty(tname).getData())
        return Blender.Mathutils.Matrix(t[0], t[1], t[2], t[3])

def bone_head_alm(obj, boneid, frame):
        return Blender.Mathutils.Vector(list(bone_transform_alm(obj, boneid, frame)[3]))

def bone_ends_alm(obj, boneid, frame):
        parent = int(obj.getProperty('almparent_' + str(boneid)).getData())
        if parent == -1: parent = boneid
        head = bone_head_alm(obj, parent, frame)
        tail = bone_head_alm(obj, boneid, frame)
        return (head, tail)

def bone_ends(obj, bone, frame):
        try:
                obj.getProperty('almbones')
                return bone_ends_alm(obj, int(bone.getName()[4:]), frame)
        except:
                # print 'No ALM skeleton found'
                return bone_ends_blender(bone)

def file_callback(filename):
        if filename.find('.eps', -4) < 0: filename += '.eps'
        file = open(filename, "w")

        objects = Blender.Object.GetSelected()
        obj = objects[0]
        mesh = Blender.NMesh.GetRawFromObject(obj.name)
        camera = Blender.Scene.GetCurrent().getCurrentCamera()
        bones = []
        if obj.getParent() is not None:
                if obj.getParent().getType() == 'Armature':
                        arm = obj.getParent().data
                        A = obj.getParent().getMatrix() * camera.getInverseMatrix()
                        bones = [b for b in arm.getBones() if b.getBoneclass() != 'UNSKINNABLE']

        x1 = 100000
        y1 = 100000
        x2 = -100000
        y2 = -100000
        T = obj.getMatrix() * camera.getInverseMatrix()

        for v in mesh.verts:
                xy = ctransform(v.co, T)
                x1 = min(x1, xy[0])
                y1 = min(y1, xy[1])
                x2 = max(x2, xy[0])
                y2 = max(y2, xy[1])
        for b in bones:
                for v in bone_ends(obj, b, 0):
                        xy = ctransform(v, A)
                        x1 = min(x1, xy[0])
                        y1 = min(y1, xy[1])
                        x2 = max(x2, xy[0])
                        y2 = max(y2, xy[1])

        print >> file, '%!PS-Adobe-2.0 EPSF-1.2'
        print >> file, '%%%%BoundingBox: %.3f %.3f %.3f %.3f' % (x1 - 10, y1 - 10, x2 + 10, y2 + 10)
        print >> file, '%%EndComments'
        print >> file, 'gsave'
        print >> file, '1 setlinejoin 1 setlinecap'
        for f in mesh.faces:
                xy = []
                for i in range(len(f.v)):
                        xy.append(ctransform(f.v[i].co, T))
                area = 0.0
                for i in range(len(xy)):
                        area += xy[i - 1][0] * xy[i][1] - xy[i][0] * xy[i - 1][1]
                if area > 0.0:
                        print >> file, 'newpath'
                        for i in range(len(xy)):
                                print >> file, '%.6f %.6f' % xy[i],
                                if i > 0: print >> file, 'lineto'
                                else: print >> file, 'moveto'
                        print >> file, 'closepath'
                        print >> file, 'stroke'
        print >> file, 'gsave'
        print >> file, '3 setlinewidth 1.0 0.0 0.0 setrgbcolor'
        for b in bones:
                print >> file, 'newpath'
                ends = bone_ends(obj, b, 0)
                for i in range(len(ends)):
                        v = ends[i]
                        xy = ctransform(v, A)
                        print >> file, '%.6f %.6f' % xy,
                        if i > 0: print >> file, 'lineto'
                        else: print >> file, 'moveto'
                print >> file, 'stroke'
        print >> file, '8 setlinewidth'
        for b in bones:
                print >> file, 'newpath'
                ends = bone_ends(obj, b, 0)
                v = ends[1]
                xy = ctransform(v, A)
                print >> file, '%.6f %.6f' % xy, 'moveto'
                print >> file, '%.6f %.6f' % xy, 'lineto'
                print >> file, 'stroke'
        print >> file, 'grestore'
        print >> file, 'grestore'
	file.close()

def createEPSPath():
        filename = Blender.Get('filename')
        if filename.find('.') != -1:
                return filename.split('.')[0] + ".eps"
        return filename

Blender.Window.FileSelector(file_callback, "EPS Export", createEPSPath())
