import Blender

def add_influences(group):
        out = {}
        for inf in group:
                for key in inf.iterkeys():
                        if not out.has_key(key): out[key] = 0.0
                        out[key] += inf[key]
        return out

def scale_influences(inf, scale):
        out = {}
        for key in inf.iterkeys():
                out[key] = inf[key] * scale
        return out

def blend_influences(group, weights):
        tmpgroup = []
        weight = 0.0
        for w in weights: weight += w
        for (inf, w) in zip(group, weights):
                tmpgroup.append(scale_influences(inf, w / weight))
        return add_influences(tmpgroup)

def edge_pair(v1, v2):
        if v1.index < v2.index: return (v1, v2)
        else: return (v2, v1)

def edge_key(v1, v2):
        if v1.index < v2.index: return (v1.index, v2.index)
        else: return (v2.index, v1.index)

class SubVertex:
        """A vertex in a subdivision surface, recording just the
        contributions from original vertices. Each contribution is
        an {index: influence} map entry, and the influences must sum
        to 1."""
        def __init__(self):
                self.influences = {}
                self.index = -1
                self.edges = []
                self.faces = []
                self.newVertex = None

        def getInfluences(self): return self.influences

        def getNew(self):
                if self.newVertex is None:
                        n = len(self.edges)
                        inf = []
                        weights = []
                        for e in self.edges:
                                inf.append(e.getOtherVertexInfluences(self))
                                weights.append(1.0 / (n * n))
                        for f in self.faces:
                                inf.append(f.getNew())
                                weights.append(1.0 / (n * n))
                        inf.append(self.getInfluences())
                        weights.append(1.0 * (n - 2) / n)
                        return blend_influences(inf, weights)
                else:
                        return self.newVertex.influences

class SubEdge:
        def __init__(self, v0, v1):
                self.verts = [v0, v1]
                self.faces = []
                self.newVertex = None

        def getNew(self):
                if self.newVertex is None:
                        n = len(self.verts) + len(self.faces)
                        inf = []
                        weights = []
                        for v in self.verts:
                                inf.append(v.getInfluences())
                                weights.append(1.0 / n)
                        for f in self.faces:
                                inf.append(f.getNew())
                                weights.append(1.0 / n)
                        return blend_influences(inf, weights)
                else:
                        return self.newVertex.influences

        def getOtherVertexInfluences(self, v):
                if v is self.verts[0]: return self.verts[1].getInfluences()
                else: return self.verts[0].getInfluences()

class SubFace:
        def __init__(self, v):
                self.verts = v
                self.newVertex = None
        def getNew(self):
                if self.newVertex is None:
                        return blend_influences(map(SubVertex.getInfluences, self.verts), [1.0] * len(self.verts))
                else:
                        return self.newVertex.influences

class SubMesh:
        def __init__(self, mesh):
                self.mesh = mesh
                self.verts = []
                self.edges = {}    # Indexed by v0-v1 tuples (v0 < v1)
                self.faces = []
                self.faceUV = mesh.faceUV

        def addVertex(self, influences):
                vert = SubVertex()
                vert.influences = influences
                vert.index = len(self.verts)
                self.verts.append(vert)
                return vert

        def addFace(self, v):
                face = SubFace(v)
                for i in range(len(v)):
                        pair = edge_pair(v[i - 1], v[i])
                        key = (pair[0].index, pair[1].index)
                        if not self.edges.has_key(key):
                                self.edges[key] = SubEdge(pair[0], pair[1])
                                pair[0].edges.append(self.edges[key])
                                pair[1].edges.append(self.edges[key])
                        self.edges[key].faces.append(face)

                        v[i].faces.append(face)
                self.faces.append(face)
                return face

def subdivide(mesh, levels):
        smesh = SubMesh(mesh)
        for i in range(len(mesh.verts)):
                smesh.addVertex({i: 1.0})
        for f in mesh.faces:
                verts = []
                for v in f.verts:
                        verts.append(smesh.verts[v.index])
                nf = smesh.addFace(verts)
                if mesh.faceUV: nf.uv = f.uv

        for level in range(levels):
                smesh2 = SubMesh(mesh)
                for f in smesh.faces:
                        f.newVertex = smesh2.addVertex(f.getNew())
                for e in smesh.edges.itervalues():
                        e.newVertex = smesh2.addVertex(e.getNew())
                for v in smesh.verts:
                        v.newVertex = smesh2.addVertex(v.getNew())
                for f in smesh.faces:
                        if mesh.faceUV:
                                uvc = Blender.Mathutils.Vector([0.0, 0.0])
                                for uv in f.uv:
                                        uvc += Blender.Mathutils.Vector(list(uv))
                                uvc = uvc * (1.0 / len(f.uv))
                        for i in range(len(f.verts)):
                                v0 = f.verts[i - 2]
                                v1 = f.verts[i - 1]
                                v2 = f.verts[i]
                                e0 = smesh.edges[edge_key(v0, v1)]
                                e1 = smesh.edges[edge_key(v1, v2)]
                                nf = smesh2.addFace([e0.newVertex, v1.newVertex, e1.newVertex, f.newVertex])
                                if mesh.faceUV:
                                        uv0 = Blender.Mathutils.Vector(list(f.uv[i - 2]))
                                        uv1 = Blender.Mathutils.Vector(list(f.uv[i - 1]))
                                        uv2 = Blender.Mathutils.Vector(list(f.uv[i]))
                                        nf.uv = [tuple(0.5 * (uv0 + uv1)),
                                                 tuple(uv1),
                                                 tuple(0.5 * (uv1 + uv2)),
                                                 tuple(uvc)]
                smesh = smesh2

        return smesh
