/* Performs the core algorithms on a number of non-articulated models */

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <cppunit/ui/text/TestRunner.h>
#include <cppunit/extensions/HelperMacros.h>
#include <string>
#include "mesh/mesh.h"
#include "formats/formats.h"
#include "common/paths.h"

using namespace std;

static const char *legal_files[] =
{
    "box.ply",
    "cube.ply",
    "cube2.ply",
    "twocubes.ply",
    "halfsharp.ply",
    "openbox.ply",
    "overlap.ply",
    "sphere.ply",
    "tetra2.ply",
    "tetrahedron.ply",
    "torus.ply",
    "face2.ply",
    "triangle.ply",
    "cross.alm",
    "simpletube-open.alm"
};

static const char *illegal_files[] =
{
    "fin.ply",
    "isolated.ply",
    "touch1.ply",
    "touch2.ply",
    "touch3.ply"
};

class BasicTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(BasicTest);
    CPPUNIT_TEST(manifold);
    CPPUNIT_TEST(non_manifold);
    CPPUNIT_TEST(simplify);
    CPPUNIT_TEST(parametrise);
    CPPUNIT_TEST(stretch_reduce);
    CPPUNIT_TEST_SUITE_END();
private:
    size_t num_legal_files;
    size_t num_illegal_files;

    static string full_pathname(const char *fname)
    {
        return resource_path("srcdir", ".", string("tests/meshes/") + fname);
    }

public:
    void setUp()
    {
        num_legal_files = sizeof(legal_files) / sizeof(legal_files[0]);
        num_illegal_files = sizeof(illegal_files) / sizeof(illegal_files[0]);
    }

    void manifold()
    {
        for (size_t i = 0; i < num_legal_files; i++)
        {
            Mesh m;
            CPPUNIT_ASSERT_NO_THROW(format_load_mesh_file("auto", full_pathname(legal_files[i]), m,
                                                          MESH_LEVEL_PATCHES));
            cerr << '.';
        }
    }

    void non_manifold()
    {
        for (size_t i = 0; i < num_illegal_files;i++)
        {
            Mesh m;
            CPPUNIT_ASSERT_THROW(format_load_mesh_file("auto", full_pathname(illegal_files[i]), m,
                                                       MESH_LEVEL_MANIFOLD), bad_format);
            cerr << '.';
        }
    }

    void simplify()
    {
        for (size_t i = 0; i < num_legal_files; i++)
        {
            Mesh m;
            format_load_mesh_file("auto", full_pathname(legal_files[i]), m,
                                  MESH_LEVEL_PATCHES);
            mesh_simplify(m, "simple", true, true);
            mesh_simplify(m, "aps", true, true);
            mesh_simplify(m, "aps-nomem", true, true);
            cerr << '.';
        }
    }

    void parametrise()
    {
        for (size_t i = 0; i < num_legal_files; i++)
        {
            Mesh m;
            format_load_mesh_file("auto", full_pathname(legal_files[i]), m,
                                  MESH_LEVEL_MANIFOLD);
            mesh_parametrise(m, 1, HUGE_VAL, 0.2, 4.0, 1.2);
            cerr << '.';
        }
    }

    void stretch_reduce()
    {
        Mesh m;
        format_load_mesh_file("auto", full_pathname("hemisphere.ply"), m,
                              MESH_LEVEL_PATCHES);
        mesh_parametrise(m, 0, HUGE_VAL, 0.0, 0.0, 0.0);
        CPPUNIT_ASSERT(mesh_texture_stretch(m, false) < 1.05);
    }
};

CPPUNIT_TEST_SUITE_REGISTRATION(BasicTest);
