#if HAVE_CONFIG_H
# include <config.h>
#endif

// #define DUMPIT
// #define NULL_UPDATE
// #define NULL_RENDER
// #define TIMING

#include "glee/GLee.h"
#include "mesh/mesh.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include "common/paths.h"
#include "alodgl/glmisc.h"
#include "alodgl/shaders.h"
#include "mesh/iheap.h"
#include <GL/glut.h>
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <glib.h>
#include <string>
#include <utility>
#include <vector>
#include <queue>
#include <sstream>
#include <iterator>
#include <functional>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#ifdef __SSE__
# include <xmmintrin.h>
#endif

#ifndef GL_TIME_ELAPSED_EXT
# define GL_TIME_ELAPSED_EXT 0x88BF
#endif

using namespace std;

struct vector4
{
    GLfloat v[4];
};

class VertexProcessor
{
protected:
    size_t num_vertices;
    size_t max_inf;
    vector<int> freq;     /* number of vertices with exactly i influences */
    vector<int> partial;  /* number of vertices with at least i influences */
    /* Note: vertices are sorted by descending number of influences. Thus,
     * those with exactly i influences are in the interval
     * [ partial[i+1], partial[i] ),
     * while those with at least i influences are in the interval
     * [ 0, partial[i] )
     */

    void initialise_stats(Mesh &m);
    VertexProcessor() : num_vertices(0), max_inf(0), freq(), partial() {}
public:
    virtual void initialise(Mesh &m) = 0;
    virtual void update_vertex_buffer(const Skeleton &skeleton) = 0;
    virtual ~VertexProcessor();
};

/* Generic code for render-to-vertex-buffer approaches. It takes care of the
 * VBO setup, partitioning it into separate regions for the three channels.
 * It also does the allocation of the renderbuffers and handles the VBO copy.
 */
class RTVBVertexProcessor : public VertexProcessor
{
protected:
    GLint buffer_width, buffer_height;
    GLuint fbo, rbo[3], vbo;
    CGprogram vp, fp;

    /* Determines the size for input textures, assuming that influences are
     * stuffed in vertically. The width is suitable for passing to
     * initialise_gl, and will be the buffer_width, but the height will NOT
     * be the same as buffer_height (height depends on the number of vertices,
     * buffer_height on the number of influences).
     */
    void get_size(GLenum format, GLint &width, GLint &height);
    void initialise_gl(Mesh &m, int width, GLenum usage,
                       const string &vshader, const string &fshader);
    void do_readback();
public:
    ~RTVBVertexProcessor();
};

/* This is a utility processor for other processors that naturally spit
 * out 16-bit results (due to hardware limitations). It provides a
 * scratch framebuffer for this storage, plus code to convert it to 32-bit.
 */
class RTVBTwoPassVertexProcessor : public RTVBVertexProcessor
{
protected:
    GLuint middle_fbo, middle_tex[3];
    CGprogram middle_vp, middle_fp;

    void initialise_two_pass();
    void do_second_pass();
public:
    ~RTVBTwoPassVertexProcessor();
};

/* This processor does conversions in the fragment pipe using blending rather
 * than a loop to do the accumulation. The input textures are laid out in
 * vertical chunks. Chunk i corresponds to influence i of all vertices with
 * at least that many influences. MRTs are used to handle all 3 channels
 * at once. Transformations are loaded into a texture, since CG does not
 * appear to support relative addressing with the FP40 profile.
 */
class RTVBFragVertexProcessorBlend : public RTVBTwoPassVertexProcessor
{
private:
    struct Vertex
    {
        GLfloat v[2];
        GLfloat tex[2];
    };

    GLuint vertex_tex[3], bone_tex, transform_tex;
    vector<Vertex> chunks;
    CGparameter flod_parameter;
public:
    void initialise(Mesh &m);
    void update_vertex_buffer(const Skeleton &skeleton);
    ~RTVBFragVertexProcessorBlend();
};

/* This is a variation in the above, but using an explicit dynamic loop
 * rather than blending. This allows it to output directly to 32-bit and
 * still have a bandwidth advantage, at the expense of dynamic branching.
 *
 * Computes to the previous processor, the chunks are interleaved rather
 * than stacked. In other words, for a vertex with N influences, the
 * influence data are vertically adjacent.
 */
class RTVBFragVertexProcessorLoop : public RTVBVertexProcessor
{
    struct Vertex
    {
        GLfloat v[2];
        GLfloat tex[3];  /* Third coordinate holds loop count */
    };

    GLuint vertex_tex[3], bone_tex, transform_tex;
    vector<Vertex> chunks;
    CGparameter flod_parameter;
public:
    void initialise(Mesh &m);
    void update_vertex_buffer(const Skeleton &skeleton);
    ~RTVBFragVertexProcessorLoop();
};

class RTVBSpriteVertexProcessor : public RTVBTwoPassVertexProcessor
{
private:
    struct SpriteVertex
    {
        GLfloat pos_bone[3];  // x, y, bone
        GLfloat data[3][2][4];
    };

    GLuint vertex_vbo;
    size_t total_influences;

    CGparameter flod_parameter;
    CGparameter bones_parameter;

    void update_uniforms(const Skeleton &skeleton);
public:
    void initialise(Mesh &m);
    void update_vertex_buffer(const Skeleton &skeleton);
    ~RTVBSpriteVertexProcessor();
};

/* Basic CPU implementation. Unlike the RTVB processors, the VBO is
 * interleaved. The calculations are done one bone at a time. For each
 * bone, we loop over all vertices affected by that bone and add on the
 * contribution. The values are accumulated into buffer before being
 * passed to OpenGL.
 */
class CVertexProcessor : public VertexProcessor
{
protected:
    struct Vertex
    {
        GLfloat v[4];
        GLfloat t0[4];
        GLfloat t1[4];
    };
    struct VertexPair
    {
        Vertex a;
        Vertex b;
        int index;
    };

    GLuint vbo;
    vector<vector<VertexPair> > vertex_data;
    Vertex *buffer;
public:
    void initialise(Mesh &m);
    void update_vertex_buffer(const Skeleton &skeleton);
    ~CVertexProcessor();
};

#ifdef __SSE__

/* An SSE version of the previous method. To preserve alignment, the
 * index field of VertexPair is split out into a separate array.
 */
class SSEVertexProcessor : public VertexProcessor
{
protected:
    struct SSEVertex
    {
        __m128 v, t0, t1;
    };
    struct SSEVertexPair
    {
        SSEVertex a;
        SSEVertex b;
    };

    GLuint vbo;
    vector<SSEVertexPair *> vertex_data;
    vector<GLuint *> index_data;
    vector<size_t> counts;
    SSEVertex *buffer;

public:
    void initialise(Mesh &m);
    void update_vertex_buffer(const Skeleton &skeleton);
    ~SSEVertexProcessor();
};

/* A more straightforward SSE processor, that operates per-vertex rather
 * than per-bone. The advantage is that the results can be streamed
 * directly into the vertex array, at the expense of less predictable
 * branches. The influence data is all stored, one after the next, in
 * vertex_data. The bone indices are stored in parallel in bone_data, and
 * the influence counts in counts.
 */
class SSERowVertexProcessor : public VertexProcessor
{
protected:
    struct SSEVertex
    {
        __m128 v, t0, t1;
    };
    struct SSEVertexPair
    {
        SSEVertex a;
        SSEVertex b;
    };

    GLuint vbo;
    SSEVertexPair *vertex_data;
    GLubyte *bone_data;
    GLubyte *counts;
public:
    void initialise(Mesh &m);
    void update_vertex_buffer(const Skeleton &skeleton);
    ~SSERowVertexProcessor();
};
#endif

static AlodProgramOption in_format;
static AlodProgramOption verbose, debug;

static Mesh mesh;
static GLuint num_vertices, num_faces;
static int frame = 0;

static GLuint mesh_ibo;
static VertexProcessor *processor = NULL;
static CGcontext cg_ctx;
static CGprogram render_vp, render_fp;
static CGprofile vprofile = CG_PROFILE_ARBVP1;
static CGprofile fprofile = CG_PROFILE_FP40;

static int process_args(int argc, char **argv)
{
    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo()
    };

    int args = alod_process_options(argc, argv, opts);
    if (!in_format.used) in_format.string_value = "auto";
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);

    if (argc - args != 1)
    {
        cerr << "Error: must specify input file\n";
        alod_show_help(opts);
        exit(1);
    }
    return args;
}

static void check_cg_error()
{
    CGerror err = cgGetError();
    if (err != CG_NO_ERROR)
        g_error("CG error: %s", cgGetErrorString(err));
}

static void switch_processor(VertexProcessor *vp)
{
    if (processor) delete processor;
    processor = vp;
    processor->initialise(mesh);
}

static void key(unsigned char key, int x, int y)
{
    switch (key)
    {
    case '\027': exit(0);
    case 'C':
    case 'c': switch_processor(new CVertexProcessor); break;
#ifdef __SSE__
    case 'S': switch_processor(new SSERowVertexProcessor); break;
    case 's': switch_processor(new SSEVertexProcessor); break;
#endif
    case 'F': switch_processor(new RTVBFragVertexProcessorLoop); break;
    case 'f': switch_processor(new RTVBFragVertexProcessorBlend); break;
    case 'v':
    case 'V': switch_processor(new RTVBSpriteVertexProcessor); break;
    default: ;
    }
}

#ifdef DUMPIT
static void dumpit()
{
    FILE *dump = fopen("ashwtest.obj", "w");
    Vertex *vertices = new vector4[num_vertices];
    GLuint *indices = new GLuint[3 * num_faces];

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, rtvb_fbo);
    glReadPixels(0, 0, buffer_width, buffer_height, GL_RGBA, GL_FLOAT, vertices);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glGetBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, 3 * num_faces * sizeof(GLuint), indices);
    for (size_t i = 0; i < num_vertices; i++)
        fprintf(dump, "v %.9f %.9f %.9f\n",
                vertices[i].v[0], vertices[i].v[1], vertices[i].v[2]);
    for (size_t i = 0; i < 3 * num_faces; i += 3)
        fprintf(dump, "f %d %d %d\n",
                indices[i + 0] + 1, indices[i + 1] + 1, indices[i + 2] + 1);
    fclose(dump);
    delete[] indices;
    delete[] vertices;
    exit(0);
}
#endif

static queue<pair<GLuint, string> > timing_queue;
static vector<GLuint> timing_available;
static pair<GLuint, string> timing_active;

#ifndef TIMING
# define begin_timing(x) ((void) 0)
# define end_timing() ((void) 0)
#else
static void begin_timing(const char *desc)
{
    GLuint current;
    GLuint elapsed;
    GLint ready;

    /* For some reason, immediately reusing the last query causes a
     * GL_INVALID_OPERATION error on glEndQuery. I suspect this is a driver
     * bug. To work around it, we select the next query target before polling
     * for results.
     */
    if (timing_available.empty())
    {
        timing_available.resize(16);
        glGenQueries(16, &timing_available[0]);
    }
    timing_active.first = timing_available.back();
    timing_available.pop_back();
    while (!timing_queue.empty())
    {
        current = timing_queue.front().first;
        glGetQueryObjectiv(current, GL_QUERY_RESULT_AVAILABLE, &ready);
        if (!ready) break;
        glGetQueryObjectuiv(current, GL_QUERY_RESULT, &elapsed);
        printf("%s: %.6fms\n", timing_queue.front().second.c_str(), elapsed * 1e-6);
        timing_available.push_back(current);
        timing_queue.pop();
    }
    glBeginQuery(GL_TIME_ELAPSED_EXT, timing_active.first);
    timing_active.second = desc;
}

static void end_timing()
{
    glEndQuery(GL_TIME_ELAPSED_EXT);
    timing_queue.push(timing_active);
}
#endif /* TIMING */

static void display()
{
    Skeleton &skeleton = mesh.skeleton.get_write();
    skeleton.set_to_frame(frame);
    if (++frame == (int) skeleton.num_frames()) frame = 0;

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
#ifdef NULL_UPDATE
    static bool first = true;
    if (first) { processor->update_vertex_buffer(skeleton); first = false; }
#endif
    for (int i = -2; i <= 2; i++)
        for (int j = -2; j <= 2; j++)
        {
#ifndef NULL_UPDATE
            processor->update_vertex_buffer(skeleton);
#endif
#ifdef DUMPIT
            dumpit();
#endif
#ifndef NULL_RENDER
            cgGLBindProgram(render_vp);
            cgGLBindProgram(render_fp);
            glLoadIdentity();
            gluLookAt(10.0, 10.0, 10.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
            glTranslatef(5.0 * i, 5.0 * j, 0.0);

            begin_timing("draw");
            glDrawRangeElements(GL_TRIANGLES, 0, num_vertices - 1, num_faces * 3,
                                GL_UNSIGNED_INT, NULL);
            end_timing();
#endif // !NULL_RENDER
        }

    glutSwapBuffers();
}

static void reshape(int w, int h)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(90.0, (double) w / h, 0.01, 300.0);
    glMatrixMode(GL_MODELVIEW);
    glViewport(0, 0, w, h);
}

static void idle()
{
    glutPostRedisplay();
}

struct CompareVertices : public binary_function<MeshVertex, MeshVertex, bool>
{
    bool operator()(const MeshVertex &a, const MeshVertex &b) const
    {
        return GPOINTER_TO_SIZE(a.internal_data) > GPOINTER_TO_SIZE(b.internal_data);
    }
};

static void set_influences(Mesh &mesh)
{
    for (MeshVertexIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++)
    {
        /* Find influences affecting all three */
        VectorPosition test =
            v->position()->position
            + v->data.tangent_plane.tangents[0]
            + v->data.tangent_plane.tangents[1];

        v->internal_data = GSIZE_TO_POINTER(distance(test.begin(), test.end()));
    }
}

static void reorder_mesh(Mesh &mesh)
{
    set_influences(mesh);
    mesh.vertices.sort(CompareVertices());
}

static void initialise_gl_faces(Mesh &m)
{
    GLuint *faces = new GLuint[3 * num_faces];
    int seq = 0;
    for (MeshVertexIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++, seq++)
        v->internal_data = GSIZE_TO_POINTER(seq);
    seq = 0;
    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++, seq++)
    {
        MeshHalfEdgeConstIterator e = f->edges();
        for (int j = 0; j < 3; j++, e = e->next())
            faces[seq * 3 + j] = GPOINTER_TO_SIZE(e->vertex()->internal_data);
    }

    glGenBuffers(1, &mesh_ibo);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER_ARB, mesh_ibo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER_ARB, 3 * num_faces * sizeof(GLuint),
                 faces, GL_STATIC_DRAW);
    // Note: leaves buffer enabled for later glDrawRangeElements
    delete[] faces;
}

static void initialise_cg()
{
    string source;
    cg_ctx = cgCreateContext();

    source = alodgl_get_shader_file("ashwtest/render-vp.cg");
    render_vp = cgCreateProgram(cg_ctx, CG_SOURCE, source.c_str(),
                                vprofile, "main", NULL);
    check_cg_error();
    cgCompileProgram(render_vp);
    check_cg_error();
    cgGLLoadProgram(render_vp);

    source = alodgl_get_shader_file("ashwtest/render-fp.cg");
    render_fp = cgCreateProgram(cg_ctx, CG_SOURCE, source.c_str(),
                                fprofile, "main", NULL);
    check_cg_error();
    cgCompileProgram(render_fp);
    check_cg_error();
    cgGLLoadProgram(render_fp);

    cgGLEnableProfile(vprofile);
    cgGLEnableProfile(fprofile);
}

#define CHECK_GL_EXTENSION(e) \
    G_STMT_START{ \
    if (!GLEE_ ## e) \
    { \
        cerr << "GL_" #e " not supported\n"; \
        exit(1); \
    } \
    }G_STMT_END

static void initialise_gl(Mesh &m)
{
    GLeeInit();
    if (!GLEE_VERSION_2_0)
    {
        cerr << "OpenGL 2.0 not supported\n";
        exit(1);
    }

    CHECK_GL_EXTENSION(ARB_texture_rectangle);
    CHECK_GL_EXTENSION(ARB_vertex_program);
    CHECK_GL_EXTENSION(ARB_fragment_program);
    CHECK_GL_EXTENSION(ARB_pixel_buffer_object);
    CHECK_GL_EXTENSION(EXT_framebuffer_object);
    CHECK_GL_EXTENSION(ARB_color_buffer_float);
    CHECK_GL_EXTENSION(ARB_texture_float);

    num_vertices = m.vertices.real_size();
    num_faces = m.faces.real_size();

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);

    set_influences(m);
    reorder_mesh(m);
    initialise_gl_faces(m);
    initialise_cg();
    processor = new CVertexProcessor;
    // processor = new RTVBFragVertexProcessorBlend;
    processor->initialise(m);
}

static void update_transform_texture(GLuint texture, const Skeleton &skeleton)
{
    int bones = skeleton.size();

    GLfloat tex_data[12 * bones];
    int index = 0;
    for (int i = 0; i < bones; i++)
    {
        Matrix4 m = skeleton.get_absolute(i);
        for (int j = 0; j < 3; j++)
            for (int k = 0; k < 4; k++)
                tex_data[index++] = m[j][k];
    }
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, texture);
    glTexSubImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, 0, 0, 3, bones,
                    GL_RGBA, GL_FLOAT, tex_data);
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);

    int args = process_args(argc, argv);
    try
    {
        format_load_mesh_file(in_format.string_value, argv[args], mesh, MESH_LEVEL_MANIFOLD);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to load mesh: " << e.what() << endl;
        exit(1);
    }

    glutInitWindowSize(800, 600);
    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA);
    glutCreateWindow("ASHWTest");
    initialise_gl(mesh);
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);
    glutKeyboardFunc(key);
    glutMainLoop();

    return 0;
}

/*** Drivers appear below here ***/

void VertexProcessor::initialise_stats(Mesh &m)
{
    set_influences(m);
    max_inf = GPOINTER_TO_SIZE(m.vertices.real_begin()->internal_data);
    freq.clear();
    freq.resize(max_inf + 1, 0);
    partial.resize(max_inf + 1);
    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
        freq[GPOINTER_TO_SIZE(v->internal_data)]++;
    partial_sum(freq.rbegin(), freq.rend(), partial.rbegin());
    num_vertices = partial[0];
}

VertexProcessor::~VertexProcessor()
{
    /* To get predictable memory allocation, ensure that the old memory is
     * no longer being used so that new memory can be optimally allocated. */
    glFinish();
}


void RTVBVertexProcessor::initialise_gl(Mesh &m, int width, GLenum usage,
                                        const string &vshader, const string &fshader)
{
    static const GLenum buffers[3] =
    {
        GL_COLOR_ATTACHMENT0_EXT,
        GL_COLOR_ATTACHMENT1_EXT,
        GL_COLOR_ATTACHMENT2_EXT
    };

    g_assert(num_vertices > 0); // must have already called init_stats
    int height = (num_vertices + width - 1) / width;
    int buffer_size = width * height * sizeof(vector4);
    buffer_width = width;
    buffer_height = height;

    glGenBuffers(1, &vbo);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferData(GL_ARRAY_BUFFER, 3 * buffer_size, NULL, usage);
    glVertexPointer(3, GL_FLOAT, sizeof(vector4), GSIZE_TO_POINTER(0));
    glEnableClientState(GL_VERTEX_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glTexCoordPointer(3, GL_FLOAT, sizeof(vector4), GSIZE_TO_POINTER(buffer_size));
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE1);
    glTexCoordPointer(3, GL_FLOAT, sizeof(vector4), GSIZE_TO_POINTER(2 * buffer_size));
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    glGenFramebuffersEXT(1, &fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    glGenRenderbuffersEXT(3, rbo);
    for (int i = 0; i < 3; i++)
    {
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, rbo[i]);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_RGBA32F_ARB,
                                 width, height);

        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT + i,
                                     GL_RENDERBUFFER_EXT, rbo[i]);
    }
    glDrawBuffers(3, buffers);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

    ostringstream arg_stream;
    string arg;
    const char *args[2] = {NULL, NULL};
    arg_stream << "-DTOTAL_BONES=" << m.skeleton.get_read().size();
    arg = arg_stream.str();
    args[0] = arg.c_str();

    string source = alodgl_get_shader_file(vshader);
    vp = cgCreateProgram(cg_ctx, CG_SOURCE, source.c_str(),
                         vprofile, "main", args);
    check_cg_error();
    cgCompileProgram(vp);
    check_cg_error();
    cgGLLoadProgram(vp);
    source = alodgl_get_shader_file(fshader);
    fp = cgCreateProgram(cg_ctx, CG_SOURCE, source.c_str(),
                         fprofile, "main", args);
    check_cg_error();
    cgCompileProgram(fp);
    check_cg_error();
    cgGLLoadProgram(fp);
}

void RTVBVertexProcessor::get_size(GLenum format, GLint &width, GLint &height)
{
    GLint max_renderbuffer_size;
    GLint test;

    width = 4;
    glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE_EXT, &max_renderbuffer_size);
    do
    {
        width *= 2;
        height = 0;
        for (size_t i = 1; i <= max_inf; i++)
            height += (partial[i] + width - 1) / width;

        glTexImage2D(GL_PROXY_TEXTURE_RECTANGLE_ARB, 0, format,
                     width, height, 0, GL_RGBA, GL_FLOAT, NULL);
        glGetTexLevelParameteriv(GL_PROXY_TEXTURE_RECTANGLE_ARB, 0, GL_TEXTURE_WIDTH, &test);
    } while ((test <= 0 || height > max_renderbuffer_size) && width <= height);
    g_assert(test > 0 && height <= max_renderbuffer_size);
}

void RTVBVertexProcessor::do_readback()
{
    GLsizei buffer_size;

    buffer_size = buffer_width * buffer_height * sizeof(vector4);
    glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, vbo);
    // Tell the driver that we no longer care about the old values
    glBufferData(GL_PIXEL_PACK_BUFFER_ARB, 3 * buffer_size,
                 NULL, GL_STREAM_COPY);
    // Get pixels from each plane
    glReadPixels(0, 0, buffer_width, buffer_height,
                 GL_RGBA, GL_FLOAT, GSIZE_TO_POINTER(0));
    glReadBuffer(GL_COLOR_ATTACHMENT1_EXT);
    glReadPixels(0, 0, buffer_width, buffer_height,
                 GL_RGBA, GL_FLOAT, GSIZE_TO_POINTER(buffer_size));
    glReadBuffer(GL_COLOR_ATTACHMENT2_EXT);
    glReadPixels(0, 0, buffer_width, buffer_height,
                 GL_RGBA, GL_FLOAT, GSIZE_TO_POINTER(2 * buffer_size));
    glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
    glBindBuffer(GL_PIXEL_PACK_BUFFER_ARB, 0);
}

RTVBVertexProcessor::~RTVBVertexProcessor()
{
    glDisableClientState(GL_VERTEX_ARRAY);
    glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    glDeleteFramebuffersEXT(1, &fbo);
    glDeleteRenderbuffersEXT(3, rbo);
    glDeleteBuffers(1, &vbo);
    cgDestroyProgram(vp);
    cgDestroyProgram(fp);
}


/* Allocates a rectangle texture and sets NEAREST filtering; returns with
 * the texture still bound
 */
static GLuint get_texture_rectangle()
{
    GLuint ans;

    glGenTextures(1, &ans);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, ans);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_RECTANGLE_ARB, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    return ans;
}

void RTVBTwoPassVertexProcessor::initialise_two_pass()
{
    static const GLenum draw_buffers[3] =
    {
        GL_COLOR_ATTACHMENT0_EXT,
        GL_COLOR_ATTACHMENT1_EXT,
        GL_COLOR_ATTACHMENT2_EXT
    };

    for (int i = 0; i < 3; i++)
    {
        middle_tex[i] = get_texture_rectangle();
        glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA16F_ARB,
                     buffer_width, buffer_height,
                     0, GL_RGBA, GL_FLOAT, NULL);
    }
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);

    glGenFramebuffersEXT(1, &middle_fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, middle_fbo);
    glDrawBuffersARB(3, draw_buffers);
    for (int i = 0; i < 3; i++)
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT + i,
                                  GL_TEXTURE_RECTANGLE_ARB, middle_tex[i], 0);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

    string source = alodgl_get_shader_file("ashwtest/rtvb2pass-vp.cg");
    middle_vp = cgCreateProgram(cg_ctx, CG_SOURCE, source.c_str(), vprofile,
                                "main", NULL);
    check_cg_error();
    cgCompileProgram(middle_vp);
    check_cg_error();
    cgGLLoadProgram(middle_vp);
    check_cg_error();

    source = alodgl_get_shader_file("ashwtest/rtvb2pass-fp.cg");
    middle_fp = cgCreateProgram(cg_ctx, CG_SOURCE, source.c_str(), fprofile,
                                "main", NULL);
    check_cg_error();
    cgCompileProgram(middle_fp);
    check_cg_error();
    cgGLLoadProgram(middle_fp);
    check_cg_error();
}

void RTVBTwoPassVertexProcessor::do_second_pass()
{
    for (int i = 2; i >= 0; i--)
    {
        glActiveTexture(GL_TEXTURE0 + i);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, middle_tex[i]);
    }
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    glClear(GL_COLOR_BUFFER_BIT);
    cgGLBindProgram(middle_vp);
    cgGLBindProgram(middle_fp);
    begin_timing("generate2");
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 0.0f);
    glVertex2f(-1.0f, -1.0f);
    glTexCoord2f(buffer_width, 0.0f);
    glVertex2f(1.0f, -1.0f);
    glTexCoord2f(buffer_width, buffer_height);
    glVertex2f(1.0f, 1.0f);
    glTexCoord2f(0.0f, buffer_height);
    glVertex2f(-1.0f, 1.0f);
    glEnd();
    end_timing();
}

RTVBTwoPassVertexProcessor::~RTVBTwoPassVertexProcessor()
{
    glDeleteFramebuffersEXT(1, &middle_fbo);
    glDeleteTextures(3, middle_tex);
    cgDestroyProgram(middle_vp);
    cgDestroyProgram(middle_fp);
}

/********************************************************************/

void RTVBFragVertexProcessorBlend::initialise(Mesh &m)
{
    initialise_stats(m);
    int row_offset[max_inf + 1];   // first row of data for ith influence
    GLint width, height;

    get_size(GL_RGBA32F_ARB, width, height);
    initialise_gl(m, width, GL_STREAM_COPY,
                  "ashwtest/rtvbfrag-vp.cg", "ashwtest/rtvbfragblend-fp.cg");
    initialise_two_pass();

    /* Create the chunk list and row offset table */
    int curh, prevh = 0;
    chunks.clear();
    for (size_t i = 1; i <= max_inf; i++)
    {
        Vertex c[4];

        row_offset[i - 1] = prevh;
        int cur = (partial[i] + width - 1) / width;
        curh = prevh + cur;

        GLfloat endy = 2.0f * cur / buffer_height - 1.0f;

        c[0].tex[0] = 0.0f;         c[0].tex[1] = prevh; c[0].v[0] = -1.0f; c[0].v[1] = -1.0f;
        c[1].tex[0] = buffer_width; c[1].tex[1] = prevh; c[1].v[0] =  1.0f; c[1].v[1] = -1.0f;
        c[2].tex[0] = buffer_width; c[2].tex[1] = curh;  c[2].v[0] =  1.0f; c[2].v[1] = endy;
        c[3].tex[0] = 0.0f;         c[3].tex[1] = curh;  c[3].v[0] = -1.0f; c[3].v[1] = endy;
        for (int i = 0; i < 4; i++)
            chunks.push_back(c[i]);

        prevh = curh;
    }

    /* Create and populate the data arrays */
    GLfloat *vertex_data[3];
    GLubyte *bone_data;
    for (int i = 0; i < 3; i++)
    {
        vertex_data[i] = new GLfloat[width * height * 4];
        memset(vertex_data[i], 0, width * height * 4 * sizeof(GLfloat));
    }
    bone_data = new GLubyte[width * height];
    memset(bone_data, 0, width * height * sizeof(GLubyte));

    int r = 0, c = 0;
    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
    {
        const VectorPosition *fields[3] =
        {
            &v->position()->position,
            &v->data.tangent_plane.tangents[0],
            &v->data.tangent_plane.tangents[1]
        };
        VectorPosition test = *fields[0] + *fields[1] + *fields[2];
        int g = 0;
        for (VectorPosition::const_iterator i = test.begin(); i != test.end(); i++, g++)
        {
            bone_data[(r + row_offset[g]) * width + c] = i->index / 4;
            for (int j = 0; j < 3; j++)
                for (int k = 0; k < 4; k++)
                    vertex_data[j][((r + row_offset[g]) * width + c) * 4 + k]
                        = (*fields[j])[i->index + k];
        }
        if (++c == width)
        {
            r++;
            c = 0;
        }
    }

    /* Allocate and fill the textures */
    for (int i = 0; i < 3; i++)
    {
        vertex_tex[i] = get_texture_rectangle();
        glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA32F_ARB,
                     width, height, 0, GL_RGBA, GL_FLOAT, vertex_data[i]);
        delete[] vertex_data[i];
    }
    bone_tex = get_texture_rectangle();
    glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_LUMINANCE8,
                 width, height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, bone_data);
    delete[] bone_data;
    transform_tex = get_texture_rectangle();
    glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA16F_ARB,
                 3, m.skeleton.get_read().size(), 0, GL_RGBA, GL_FLOAT, NULL);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);

    /* Find the uniforms */
    flod_parameter = cgGetNamedParameter(fp, "flod");
    cgGLSetParameter1f(flod_parameter, 0.5f);

    /* Misc other */
    glBlendFunc(GL_ONE, GL_ONE);
}

/* Puts back the prior state, except for program bindings */
void RTVBFragVertexProcessorBlend::update_vertex_buffer(const Skeleton &skeleton)
{
    GLint old_viewport[4];

    glGetIntegerv(GL_VIEWPORT, old_viewport);
    update_transform_texture(transform_tex, skeleton);

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, middle_fbo);
    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

    glViewport(0, 0, buffer_width, buffer_height);
    glClear(GL_COLOR_BUFFER_BIT);
    glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glVertexPointer(2, GL_FLOAT, sizeof(Vertex), &chunks[0].v);
    glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), &chunks[0].tex);
    cgGLBindProgram(vp);
    cgGLBindProgram(fp);
    alodgl_check_framebuffer_status();

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, bone_tex);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, transform_tex);
    for (int i = 0; i < 3; i++)
    {
        glActiveTexture(GL_TEXTURE2 + i);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, vertex_tex[i]);
    }
    glActiveTexture(GL_TEXTURE0);

    begin_timing("generate1");
    glDrawArrays(GL_QUADS, 0, chunks.size());
    end_timing();

    glDisable(GL_BLEND);
    do_second_pass();

    do_readback();

    glViewport(old_viewport[0], old_viewport[1], old_viewport[2], old_viewport[3]);
    glPopClientAttrib();
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glEnable(GL_DEPTH_TEST);
}

RTVBFragVertexProcessorBlend::~RTVBFragVertexProcessorBlend()
{
    glDeleteTextures(3, vertex_tex);
    glDeleteTextures(1, &bone_tex);
    glDeleteTextures(1, &transform_tex);
    glBlendFunc(GL_ONE, GL_ZERO);
}

/**********************************************************************/

void RTVBFragVertexProcessorLoop::initialise(Mesh &m)
{
    GLint width, height;

    initialise_stats(m);
    get_size(GL_RGBA16F_ARB, width, height);
    initialise_gl(m, width, GL_STREAM_COPY,
                  "ashwtest/rtvbfrag-vp.cg", "ashwtest/rtvbfragloop-fp.cg");

    /* Create the chunk list */
    int prevh = 0, prev = 0;
    chunks.clear();
    for (size_t i = max_inf; i >= 1; i--)
    {
        Vertex c[4];
        int cur = (partial[i] + width - 1) / width;
        int curh = prevh + (cur - prev) * i;

        GLfloat starty = 2.0f * prev / buffer_height - 1.0f;
        GLfloat endy = 2.0f * cur / buffer_height - 1.0f;
        GLfloat startt = prevh - (i - 1) * 0.5f;
        GLfloat endt = curh - (i - 1) * 0.5f;

        c[0].tex[0] = 0.0f;         c[0].tex[1] = startt; c[0].v[0] = -1.0f; c[0].v[1] = starty;
        c[1].tex[0] = buffer_width; c[1].tex[1] = startt; c[1].v[0] =  1.0f; c[1].v[1] = starty;
        c[2].tex[0] = buffer_width; c[2].tex[1] = endt;   c[2].v[0] =  1.0f; c[2].v[1] = endy;
        c[3].tex[0] = 0.0f;         c[3].tex[1] = endt;   c[3].v[0] = -1.0f; c[3].v[1] = endy;

        if (cur != prev)
            for (int j = 0; j < 4; j++)
            {
                c[j].tex[2] = i;
                chunks.push_back(c[j]);
            }

        prev = cur;
        prevh = curh;
    }

    /* Create and populate the data arrays */
    GLfloat *vertex_data[3];
    GLubyte *bone_data;
    for (int i = 0; i < 3; i++)
    {
        vertex_data[i] = new GLfloat[width * height * 4];
        memset(vertex_data[i], 0, width * height * 4 * sizeof(GLfloat));
    }
    bone_data = new GLubyte[width * height];
    memset(bone_data, 0, width * height * sizeof(GLubyte));

    int r = 0, c = 0, dr = 0;
    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
    {
        const VectorPosition *fields[3] =
        {
            &v->position()->position,
            &v->data.tangent_plane.tangents[0],
            &v->data.tangent_plane.tangents[1]
        };
        VectorPosition test = *fields[0] + *fields[1] + *fields[2];
        int g = 0;
        for (VectorPosition::const_iterator i = test.begin(); i != test.end(); i++, g++)
        {
            bone_data[(r + g) * width + c] = i->index / 4;
            for (int j = 0; j < 3; j++)
                for (int k = 0; k < 4; k++)
                    vertex_data[j][((r + g) * width + c) * 4 + k]
                        = (*fields[j])[i->index + k];
        }
        if (c == 0) dr = g;
        if (++c == width)
        {
            r += dr;
            dr = 0;
            c = 0;
        }
    }
    g_assert(r + dr == height);

    /* Allocate and fill the textures */
    for (int i = 0; i < 3; i++)
    {
        vertex_tex[i] = get_texture_rectangle();
        glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA32F_ARB,
                     width, height, 0, GL_RGBA, GL_FLOAT, vertex_data[i]);
        delete[] vertex_data[i];
    }
    bone_tex = get_texture_rectangle();
    glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_LUMINANCE8,
                 width, height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, bone_data);
    delete[] bone_data;
    transform_tex = get_texture_rectangle();
    glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA16F_ARB,
                 3, m.skeleton.get_read().size(), 0, GL_RGBA, GL_FLOAT, NULL);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);

    /* Find the uniforms */
    flod_parameter = cgGetNamedParameter(fp, "flod");
    cgGLSetParameter1f(flod_parameter, 0.5f);
}

/* Puts back the prior state, except for program bindings */
void RTVBFragVertexProcessorLoop::update_vertex_buffer(const Skeleton &skeleton)
{
    GLint old_viewport[4];

    glGetIntegerv(GL_VIEWPORT, old_viewport);
    update_transform_texture(transform_tex, skeleton);

    glDisable(GL_DEPTH_TEST);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

    glViewport(0, 0, buffer_width, buffer_height);
    glClear(GL_COLOR_BUFFER_BIT);
    glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glVertexPointer(2, GL_FLOAT, sizeof(Vertex), &chunks[0].v);
    glTexCoordPointer(3, GL_FLOAT, sizeof(Vertex), &chunks[0].tex);
    cgGLBindProgram(vp);
    cgGLBindProgram(fp);
    alodgl_check_framebuffer_status();

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, bone_tex);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_RECTANGLE_ARB, transform_tex);
    for (int i = 0; i < 3; i++)
    {
        glActiveTexture(GL_TEXTURE2 + i);
        glBindTexture(GL_TEXTURE_RECTANGLE_ARB, vertex_tex[i]);
    }
    glActiveTexture(GL_TEXTURE0);

    begin_timing("generate");
    glDrawArrays(GL_QUADS, 0, chunks.size());
    end_timing();

    do_readback();

    glViewport(old_viewport[0], old_viewport[1], old_viewport[2], old_viewport[3]);
    glPopClientAttrib();
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glEnable(GL_DEPTH_TEST);
}

RTVBFragVertexProcessorLoop::~RTVBFragVertexProcessorLoop()
{
    glDeleteTextures(3, vertex_tex);
    glDeleteTextures(1, &bone_tex);
    glDeleteTextures(1, &transform_tex);
}

/*************************************************************/

void RTVBSpriteVertexProcessor::initialise(Mesh &m)
{
    initialise_stats(m);
    total_influences = accumulate(partial.begin(), partial.end(), 0);

    int width = 1;
    while ((size_t) width * width * 3 < num_vertices) width <<= 1;
    initialise_gl(m, width, GL_STREAM_COPY,
                  "ashwtest/rtvbsprite-vp.cg", "ashwtest/rtvbsprite-fp.cg");
    initialise_two_pass();
    int height = buffer_height;

    SpriteVertex *vertices = new SpriteVertex[total_influences];

    int index = 0;
    int r = 0, c = 0;
    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
    {
        const VectorPosition *fields[3] =
        {
            &v->position()->position,
            &v->data.tangent_plane.tangents[0],
            &v->data.tangent_plane.tangents[1]
        };
        VectorPosition test = *fields[0] + *fields[1] + *fields[2];
        for (VectorPosition::const_iterator t = test.begin(); t != test.end(); t++)
        {
            vertices[index].pos_bone[0] = 2.0f * (c + 0.5f) / width - 1.0f;
            vertices[index].pos_bone[1] = 2.0f * (r + 0.5f) / height - 1.0f;
            vertices[index].pos_bone[2] = t->index / 4;
            for (int j = 0; j < 3; j++)
                for (int k = 0; k < 4; k++)
                {
                    vertices[index].data[j][0][k] = (*fields[j])[t->index + k];
                    vertices[index].data[j][1][k] = (*fields[j])[t->index + k];
                }
            index++;
        }
        if (++c == width)
        {
            c = 0;
            r++;
        }
    }

    glGenBuffers(1, &vertex_vbo);
    glBindBuffer(GL_ARRAY_BUFFER, vertex_vbo);
    glBufferData(GL_ARRAY_BUFFER, total_influences * sizeof(SpriteVertex), vertices, GL_STATIC_DRAW);
    delete[] vertices;

    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, sizeof(SpriteVertex),
                          GSIZE_TO_POINTER(offsetof(SpriteVertex, pos_bone)));
    for (int i = 0; i < 3; i++)
        for (int j = 0; j < 2; j++)
            glVertexAttribPointer(i * 2 + j + 1, 4, GL_FLOAT, GL_FALSE, sizeof(SpriteVertex),
                                  (GLvoid *) &((SpriteVertex *) NULL)->data[i][j][0]);

    flod_parameter = cgGetNamedParameter(vp, "flod");
    bones_parameter = cgGetNamedParameter(vp, "bones");

    /* Misc other */
    glBlendFunc(GL_ONE, GL_ONE);
}

void RTVBSpriteVertexProcessor::update_uniforms(const Skeleton &skeleton)
{
    int bones = skeleton.size();

    GLfloat bone_data[12 * bones];
    int index = 0;
    for (int i = 0; i < bones; i++)
    {
        Matrix4 m = skeleton.get_absolute(i);
        for (int j = 0; j < 3; j++)
            for (int k = 0; k < 4; k++)
                bone_data[index++] = m[j][k];
    }
    cgGLSetMatrixParameterArrayfr(bones_parameter, 0, bones, bone_data);
    cgGLSetParameter1f(flod_parameter, 0.5);
}

void RTVBSpriteVertexProcessor::update_vertex_buffer(const Skeleton &skeleton)
{
    GLint old_viewport[4];

    glGetIntegerv(GL_VIEWPORT, old_viewport);

    glDisable(GL_DEPTH_TEST);
    glEnable(GL_BLEND);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, middle_fbo);
    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);

    glViewport(0, 0, buffer_width, buffer_height);
    glClear(GL_COLOR_BUFFER_BIT);
    glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);

    for (int i = 0; i < 7; i++)
        glEnableVertexAttribArray(i);
    cgGLBindProgram(vp);
    cgGLBindProgram(fp);
    update_uniforms(skeleton);
    glDrawArrays(GL_POINTS, 0, total_influences);

    glDisable(GL_BLEND);
    do_second_pass();

    do_readback();

    glViewport(old_viewport[0], old_viewport[1], old_viewport[2], old_viewport[3]);
    glPopClientAttrib();
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glEnable(GL_DEPTH_TEST);
}

RTVBSpriteVertexProcessor::~RTVBSpriteVertexProcessor()
{
    glDeleteBuffers(1, &vertex_vbo);
}

void CVertexProcessor::initialise(Mesh &m)
{
    const Skeleton &skeleton = m.skeleton.get_read();
    size_t bones = skeleton.size();

    initialise_stats(m);
    vertex_data.resize(bones);
    GLuint index = 0;
    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++, index++)
    {
        VectorPosition test =
            v->position()->position
            + v->data.tangent_plane.tangents[0]
            + v->data.tangent_plane.tangents[1];
        for (VectorPosition::const_iterator i = test.begin(); i != test.end(); i++)
        {
            VertexPair cur;
            int bone = i->index / 4;
            for (int j = 0; j < 4; j++)
            {
                cur.a.v[j] = v->position()->position[i->index + j];
                cur.a.t0[j] = v->data.tangent_plane.tangents[0][i->index + j];
                cur.a.t1[j] = v->data.tangent_plane.tangents[1][i->index + j];
                cur.b = cur.a;
            }
            cur.index = index;
            vertex_data[bone].push_back(cur);
        }
    }

    glGenBuffers(1, &vbo);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferData(GL_ARRAY_BUFFER, num_vertices * sizeof(Vertex), NULL, GL_STREAM_DRAW);
    glVertexPointer(3, GL_FLOAT, sizeof(Vertex), GSIZE_TO_POINTER(offsetof(Vertex, v)));
    glEnableClientState(GL_VERTEX_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glTexCoordPointer(3, GL_FLOAT, sizeof(Vertex), GSIZE_TO_POINTER(offsetof(Vertex, t0)));
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE1);
    glTexCoordPointer(3, GL_FLOAT, sizeof(Vertex), GSIZE_TO_POINTER(offsetof(Vertex, t1)));
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    buffer = new Vertex[num_vertices];
}

static inline void transform_pair(GLfloat out[], const GLfloat a[],
                                  const GLfloat b[], const GLfloat m[][4],
                                  const GLfloat flod)
{
    for (int k = 0; k < 3; k++)
        out[k] +=
            m[k][0] * (a[0] + flod * b[0])
            + m[k][1] * (a[1] + flod * b[1])
            + m[k][2] * (a[2] + flod * b[2])
            + m[k][3] * (a[3] + flod * b[3]);
}

void CVertexProcessor::update_vertex_buffer(const Skeleton &skeleton)
{
    size_t bones = skeleton.size();
    memset(buffer, 0, num_vertices * sizeof(Vertex));

    for (size_t i = 0; i < bones; i++)
    {
        GLfloat m[3][4];
        for (int j = 0; j < 3; j++)
            for (int k = 0; k < 4; k++)
                m[j][k] = skeleton.get_absolute(i)[j][k];
        for (size_t j = 0; j < vertex_data[i].size(); j++)
        {
            const VertexPair &cur = vertex_data[i][j];
            Vertex &out = buffer[cur.index];
            transform_pair(out.v, cur.a.v, cur.b.v, m, 0.5);
            transform_pair(out.t0, cur.a.t0, cur.b.t0, m, 0.5);
            transform_pair(out.t1, cur.a.t1, cur.b.t1, m, 0.5);
        }
    }
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferSubData(GL_ARRAY_BUFFER, 0, num_vertices * sizeof(Vertex), buffer);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

CVertexProcessor::~CVertexProcessor()
{
    delete[] buffer;
    glDisableClientState(GL_VERTEX_ARRAY);
    glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDeleteBuffers(1, &vbo);
}


#ifdef __SSE__

/* Takes a 4x4 matrix in column-major form (one column per __m128) and
 * transforms v.
 */
static __m128 inline matrix_transform_sse(const __m128 m[],
                                          const __m128 &v)
{
    __m128 out, vx;
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(0, 0, 0, 0));
    out = _mm_mul_ps(m[0], vx);
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(1, 1, 1, 1));
    out = _mm_add_ps(out, _mm_mul_ps(m[1], vx));
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(2, 2, 2, 2));
    out = _mm_add_ps(out, _mm_mul_ps(m[2], vx));
    vx = _mm_shuffle_ps(v, v, _MM_SHUFFLE(3, 3, 3, 3));
    out = _mm_add_ps(out, _mm_mul_ps(m[3], vx));
    return out;
}

/* Transforms a pair, applying a suitable flod blend */
static inline __m128 transform_pair_sse(const __m128 &a,
                                        const __m128 &b, const __m128 m[],
                                        const __m128 &flod)
{
    __m128 v = _mm_add_ps(a, _mm_mul_ps(b, flod));
    return matrix_transform_sse(m, v);
}

void SSEVertexProcessor::initialise(Mesh &m)
{
    const Skeleton &skeleton = m.skeleton.get_read();
    size_t bones = skeleton.size();

    initialise_stats(m);
    counts.resize(bones);
    vertex_data.resize(bones);
    index_data.resize(bones);
    GLuint index = 0;

    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
    {
        VectorPosition test =
            v->position()->position
            + v->data.tangent_plane.tangents[0]
            + v->data.tangent_plane.tangents[1];
        for (VectorPosition::const_iterator i = test.begin(); i != test.end(); i++)
            counts[i->index / 4]++;
    }
    for (size_t i = 0; i < bones; i++)
    {
        posix_memalign((void **) &vertex_data[i], 16, counts[i] * sizeof(SSEVertexPair));
        index_data[i] = (GLuint *) malloc(counts[i] * sizeof(GLuint));
        counts[i] = 0;
    }

    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++, index++)
    {
        VectorPosition test =
            v->position()->position
            + v->data.tangent_plane.tangents[0]
            + v->data.tangent_plane.tangents[1];
        for (VectorPosition::const_iterator i = test.begin(); i != test.end(); i++)
        {
            union
            {
                SSEVertexPair sse;
                float scalar[2][3][4];
            } cur;
            int bone = i->index / 4;
            for (int j = 0; j < 4; j++)
            {
                cur.scalar[0][0][j] = v->position()->position[i->index + j];
                cur.scalar[0][1][j] = v->data.tangent_plane.tangents[0][i->index + j];
                cur.scalar[0][2][j] = v->data.tangent_plane.tangents[1][i->index + j];
            }
            cur.sse.b = cur.sse.a;
            vertex_data[bone][counts[bone]] = cur.sse;
            index_data[bone][counts[bone]] = index;
            counts[bone]++;
        }
    }

    glGenBuffers(1, &vbo);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferData(GL_ARRAY_BUFFER, num_vertices * sizeof(SSEVertex), NULL, GL_STREAM_DRAW);
    glVertexPointer(3, GL_FLOAT, sizeof(SSEVertex), GSIZE_TO_POINTER(offsetof(SSEVertex, v)));
    glEnableClientState(GL_VERTEX_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glTexCoordPointer(3, GL_FLOAT, sizeof(SSEVertex), GSIZE_TO_POINTER(offsetof(SSEVertex, t0)));
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE1);
    glTexCoordPointer(3, GL_FLOAT, sizeof(SSEVertex), GSIZE_TO_POINTER(offsetof(SSEVertex, t1)));
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);

    posix_memalign((void **) &buffer, 16, num_vertices * sizeof(SSEVertex));
}

void SSEVertexProcessor::update_vertex_buffer(const Skeleton &skeleton)
{
    __m128 flod;
    size_t bones = skeleton.size();
    memset(buffer, 0, num_vertices * sizeof(SSEVertex));

    flod = _mm_set1_ps(0.5);
    for (size_t i = 0; i < bones; i++)
    {
        __m128 m[4];
        const Matrix4 &cm = skeleton.get_absolute(i);
        for (int j = 0; j < 4; j++)
            m[j] = _mm_setr_ps(cm[0][j], cm[1][j], cm[2][j], cm[3][j]);
        for (size_t j = 0; j < counts[i]; j++)
        {
            const SSEVertexPair &cur = vertex_data[i][j];
            SSEVertex &out = buffer[index_data[i][j]];
            out.v = _mm_add_ps(out.v, transform_pair_sse(cur.a.v, cur.b.v, m, flod));
            out.t0 = _mm_add_ps(out.t0, transform_pair_sse(cur.a.t0, cur.b.t0, m, flod));
            out.t1 = _mm_add_ps(out.t1, transform_pair_sse(cur.a.t1, cur.b.t1, m, flod));
        }
    }
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferSubData(GL_ARRAY_BUFFER, 0, num_vertices * sizeof(SSEVertex), buffer);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

SSEVertexProcessor::~SSEVertexProcessor()
{
    free(buffer);
    for (size_t i = 0; i < vertex_data.size(); i++)
    {
        free(vertex_data[i]);
        free(index_data[i]);
    }
    glDisableClientState(GL_VERTEX_ARRAY);
    glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDeleteBuffers(1, &vbo);
}

/*************************************************************/

void SSERowVertexProcessor::initialise(Mesh &m)
{
    int total_influences;

    initialise_stats(m);
    total_influences = accumulate(partial.begin(), partial.end(), 0);

    counts = new GLubyte[num_vertices];
    bone_data = new GLubyte[total_influences];
    posix_memalign((void **) &vertex_data, 16, total_influences * sizeof(SSEVertexPair));

    int pos = 0, index = 0;
    for (MeshVertexConstIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++, index++)
    {
        VectorPosition test =
            v->position()->position
            + v->data.tangent_plane.tangents[0]
            + v->data.tangent_plane.tangents[1];

        counts[index] = 0;
        for (VectorPosition::const_iterator i = test.begin(); i != test.end(); i++)
        {
            union
            {
                SSEVertexPair sse;
                float scalar[2][3][4];
            } cur;

            for (int j = 0; j < 4; j++)
            {
                cur.scalar[0][0][j] = v->position()->position[i->index + j];
                cur.scalar[0][1][j] = v->data.tangent_plane.tangents[0][i->index + j];
                cur.scalar[0][2][j] = v->data.tangent_plane.tangents[1][i->index + j];
            }
            cur.sse.b = cur.sse.a;
            vertex_data[pos] = cur.sse;
            bone_data[pos] = i->index / 4;

            counts[index]++;
            pos++;
        }
    }

    glGenBuffers(1, &vbo);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glBufferData(GL_ARRAY_BUFFER, num_vertices * sizeof(SSEVertex), NULL, GL_STREAM_DRAW);
    glVertexPointer(3, GL_FLOAT, sizeof(SSEVertex), GSIZE_TO_POINTER(offsetof(SSEVertex, v)));
    glEnableClientState(GL_VERTEX_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glTexCoordPointer(3, GL_FLOAT, sizeof(SSEVertex), GSIZE_TO_POINTER(offsetof(SSEVertex, t0)));
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE1);
    glTexCoordPointer(3, GL_FLOAT, sizeof(SSEVertex), GSIZE_TO_POINTER(offsetof(SSEVertex, t1)));
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void SSERowVertexProcessor::update_vertex_buffer(const Skeleton &skeleton)
{
    int bones = skeleton.size();
    SSEVertex *ptr;
    SSEVertex cur;
    __m128 transforms[bones][4];
    __m128 flod;

    for (int i = 0; i < bones; i++)
    {
        const Matrix4 &cm = skeleton.get_absolute(i);
        for (int j = 0; j < 4; j++)
            transforms[i][j] = _mm_setr_ps(cm[0][j], cm[1][j], cm[2][j], cm[3][j]);
    }

    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    ptr = (SSEVertex *) glMapBuffer(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
    g_assert(ptr);
    g_assert(!(GPOINTER_TO_SIZE(ptr) & 0xf));

    flod = _mm_set1_ps(0.5);
    int pos = 0;
    for (size_t i = 0; i < num_vertices; i++)
    {
        cur.v = _mm_setzero_ps();
        cur.t0 = _mm_setzero_ps();
        cur.t1 = _mm_setzero_ps();
        for (int j = 0; j < counts[i]; j++)
        {
            cur.v = _mm_add_ps(cur.v, transform_pair_sse(vertex_data[pos].a.v,
                                                         vertex_data[pos].b.v,
                                                         transforms[bone_data[pos]],
                                                         flod));
            cur.t0 = _mm_add_ps(cur.t0, transform_pair_sse(vertex_data[pos].a.t0,
                                                           vertex_data[pos].b.t0,
                                                           transforms[bone_data[pos]],
                                                           flod));
            cur.t1 = _mm_add_ps(cur.t1, transform_pair_sse(vertex_data[pos].a.t1,
                                                           vertex_data[pos].b.t1,
                                                           transforms[bone_data[pos]],
                                                           flod));
            pos++;
        }
        ptr[i] = cur;
    }

    glUnmapBuffer(GL_ARRAY_BUFFER);
}

SSERowVertexProcessor::~SSERowVertexProcessor()
{
    free(vertex_data);
    delete[] bone_data;
    delete[] counts;
    glDisableClientState(GL_VERTEX_ARRAY);
    glClientActiveTexture(GL_TEXTURE1);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glClientActiveTexture(GL_TEXTURE0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDeleteBuffers(1, &vbo);
}

#endif /* __SSE__ */
