#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include "common/paths.h"
#include "glee/GLee.h"
#include "geom/geom.h"
#include "alodgl/shaders.h"
#include "alodgl/glmisc.h"
#include <algorithm>
#include <utility>
#include <GL/gl.h>
#include <GL/glext.h>
#include <gtk/gtk.h>
#include <gdk/gdkgl.h>

using namespace std;

static AlodProgramOption resolution, in_format, out_format, no_scale, graph;
static AlodProgramOption verbose, debug;
static double scale = 1.0;
static string file1, file2;

static int process_options(int argc, char **argv)
{
    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('o', "output", "output format", 1, PROGRAM_ARGUMENT_STRING, &out_format),
        AlodProgramOptionInfo('r', "resolution", "size of image", 1, PROGRAM_ARGUMENT_INT, &resolution),
        AlodProgramOptionInfo('g', "graph", "step size for decay graph", 1, PROGRAM_ARGUMENT_INT, &graph),
        AlodProgramOptionInfo('n', "no-scale", "do not normalise for a bounding cube of length 1", 0, PROGRAM_ARGUMENT_STRING, &no_scale),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo()
    };

    int args;
    args = alod_process_options(argc, argv, opts);

    if (!in_format.used) in_format.string_value = "auto";
    if (!out_format.used) out_format.string_value = "auto";
    if (resolution.used)
    {
        if (resolution.int_value <= 0
            || (resolution.int_value & (resolution.int_value - 1)))
        {
            cerr << "resolution must be a power of 2\n";
            exit(1);
        }
    }
    else
        resolution.int_value = 2048;
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);

    if (argc - args != 1 && argc - args != 2)
    {
        cerr << "Error: must specify one or two files to compare\n";
        alod_show_help(opts);
        exit(1);
    }
    return args;
}

static void initialise_gl()
{
    const int attribs[] = {
        GDK_GL_USE_GL, TRUE,
        GDK_GL_RGBA, TRUE,
        GDK_GL_RED_SIZE, 8,
        GDK_GL_GREEN_SIZE, 8,
        GDK_GL_BLUE_SIZE, 8,
        GDK_GL_ALPHA_SIZE, 8,
        GDK_GL_ATTRIB_LIST_NONE
    };

    GdkGLConfig *config;
    GdkGLContext *context;
    GdkWindowAttr window_attr = {NULL};
    GdkWindow *window;
    GdkGLWindow *glwindow;

    config = gdk_gl_config_new(attribs);

    window_attr.width = 4;
    window_attr.height = 4;
    window_attr.wclass = GDK_INPUT_OUTPUT;
    window_attr.visual = gdk_gl_config_get_visual(config);
    window_attr.colormap = gdk_gl_config_get_colormap(config);
    window_attr.window_type = GDK_WINDOW_TOPLEVEL;
    window = gdk_window_new(NULL, &window_attr, GDK_WA_COLORMAP | GDK_WA_VISUAL);
    glwindow = gdk_gl_window_new(config, window, NULL);

    context = gdk_gl_context_new(GDK_GL_DRAWABLE(glwindow), NULL, true, GDK_GL_RGBA_TYPE);
    gdk_gl_drawable_make_current(GDK_GL_DRAWABLE(glwindow), context);

    GLeeInit();
    glClampColorARB(GL_CLAMP_VERTEX_COLOR_ARB, GL_FALSE);
    glClampColorARB(GL_CLAMP_FRAGMENT_COLOR_ARB, GL_FALSE);
    glClampColorARB(GL_CLAMP_READ_COLOR_ARB, GL_FALSE);
}

static GLuint create_texture(int width, int height)
{
    GLuint tex;

    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Creating %d by %d texture",
          width, height);
    glGenTextures(1, &tex);
    glBindTexture(GL_TEXTURE_2D, tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F_ARB, width, height, 0, GL_RGBA, GL_FLOAT, NULL);
    glBindTexture(GL_TEXTURE_2D, 0);
    return tex;
}

static void geometry_image(const Mesh &m, int w, int h, double alpha)
{
    const Skeleton &skeleton = m.skeleton.get_read();

    glBegin(GL_TRIANGLES);
    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
    {
        MeshHalfEdgeConstIterator e = f->edges();
        for (int i = 0; i < 3; i++, e = e->next())
        {
            MeshVertexConstIterator v = e->vertex();
            MeshPositionConstIterator p = v->position();
            Vector3 pos = skeleton.ssd_to_3d(p->position);
            glColor4d(pos[0], pos[1], pos[2], alpha);
            glVertex2d(v->data.texcoords[0], v->data.texcoords[1]);
        }
    }
    glEnd();
}

/* Returns mean and maximum squared error */
static pair<double, double> deltas(int w, int h, GLfloat (*img)[4])
{
    int a = 0;
    int c = 0;
    double worst = 0.0;
    double total = 0.0;

    for (int y = 0; y < h; y++)
        for (int x = 0; x < w; x++, a++)
            if (img[a][3] == 1.0)
            {
                double d = img[a][0];
                total += d;
                worst = max(worst, d);
                c++;
            }
    return make_pair(total / c, worst);
}

static GLuint tex[3], fbo[3];
static int width, height;
static GLhandleARB plain_prog, compute_prog, combine_prog;

static void setup_programs()
{
    static const GLcharARB plain_f[] =
        "#version 110\n"
        "void main()\n"
        "{\n"
        "    gl_FragColor = gl_Color;\n"
        "}\n";
    static const GLcharARB compute_f[] =
        "#version 110\n"
        "uniform sampler2D tex0, tex1;\n"
        "void main()\n"
        "{\n"
        "    vec4 c0 = texture2D(tex0, gl_TexCoord[0].xy);\n"
        "    vec4 c1 = texture2D(tex1, gl_TexCoord[0].xy);\n"
        "    vec3 offset = c0.xyz - c1.xyz;\n"
        "    float w = c0.w * c1.w;\n"
        "    gl_FragColor.rgb = vec3(dot(offset, offset)) * w;\n"
        "    gl_FragColor.a = w;\n"
        "}\n";
    static const GLcharARB combine_f[] =
        "#version 110\n"
        "uniform sampler2D tex;\n"
        "void main()\n"
        "{\n"
        "    vec4 ans = vec4(0.0);\n"
        "    for (int i = 0; i < 4; i++)\n"
        "    {\n"
        "        vec4 pull = texture2D(tex, gl_TexCoord[i].xy);\n"
        "        ans.ra += pull.ra;\n"
        "        ans.gb = max(ans.gb, pull.gb);\n"
        "    }\n"
        "    gl_FragColor = ans;\n"
        "}\n";

    GLhandleARB plain_f_s = alodgl_load_hl_shader(GL_FRAGMENT_SHADER_ARB, plain_f, "plain");
    plain_prog = glCreateProgramObjectARB();
    glAttachObjectARB(plain_prog, plain_f_s);
    glDeleteObjectARB(plain_f_s);
    alodgl_link_program(plain_prog);

    GLhandleARB compute_f_s = alodgl_load_hl_shader(GL_FRAGMENT_SHADER_ARB, compute_f, "compute");
    compute_prog = glCreateProgramObjectARB();
    glAttachObjectARB(compute_prog, compute_f_s);
    glDeleteObjectARB(compute_f_s);
    alodgl_link_program(compute_prog);

    GLhandleARB combine_f_s = alodgl_load_hl_shader(GL_FRAGMENT_SHADER_ARB, combine_f, "combine");
    combine_prog = glCreateProgramObjectARB();
    glAttachObjectARB(combine_prog, combine_f_s);
    glDeleteObjectARB(combine_f_s);
    alodgl_link_program(combine_prog);

    glUseProgramObjectARB(compute_prog);
    glUniform1iARB(glGetUniformLocation(compute_prog, "tex0"), 0);
    glUniform1iARB(glGetUniformLocation(compute_prog, "tex1"), 1);
    glUseProgramObjectARB(0);

    glUseProgramObjectARB(combine_prog);
    glUniform1iARB(glGetUniformLocation(combine_prog, "tex"), 0);
    glUseProgramObjectARB(0);
}

static void setup()
{
    width = resolution.int_value;
    height = resolution.int_value;

    tex[0] = create_texture(width, height);
    tex[1] = create_texture(width, height);
    tex[2] = create_texture(width, height);
    setup_programs();

    glGenFramebuffersEXT(3, fbo);
    for (int i = 0; i < 3; i++)
    {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo[i]);
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
                                  GL_TEXTURE_2D, tex[i], 0);
    }

    glViewport(0, 0, resolution.int_value, resolution.int_value);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
    glMatrixMode(GL_MODELVIEW);
}

static vector<double> compare_meshes(Mesh &m1, Mesh &m2)
{
    GLfloat x[4] = {0.0, 0.0, 1.0, 1.0};
    GLfloat y[4] = {0.0, 1.0, 1.0, 0.0};
    GLfloat image[4];

    Skeleton &s1 = m1.skeleton.get_write();
    Skeleton &s2 = m2.skeleton.get_write();

    double total_mean = 0.0, total_max = 0.0;
    double max_mean = 0.0, max_max = 0.0;
    for (size_t i = 0; i < s1.num_frames(); i++)
    {
        s1.set_to_frame(i);
        s2.set_to_frame(i);

        glUseProgramObjectARB(plain_prog);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo[0]);
        glClear(GL_COLOR_BUFFER_BIT);
        geometry_image(m1, width, height, 1.0);

        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo[1]);
        glClear(GL_COLOR_BUFFER_BIT);
        geometry_image(m2, width, height, 1.0);

        glUseProgramObjectARB(compute_prog);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo[2]);
        glActiveTexture(GL_TEXTURE1);
        glBindTexture(GL_TEXTURE_2D, tex[1]);
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, tex[0]);
        glBegin(GL_QUADS);
        for (int i = 0; i < 4; i++)
        {
            glMultiTexCoord2f(GL_TEXTURE0, x[i], y[i]);
            glVertex2f(x[i], y[i]);
        }
        glEnd();

        glUseProgramObjectARB(combine_prog);
        int src = 2;
        int trg = 0;
        for (int s = resolution.int_value; s > 1; s /= 2)
        {
            glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo[trg]);
            glBindTexture(GL_TEXTURE_2D, tex[src]);
            glViewport(0, 0, s / 2, s / 2);
            double hpxl = 0.5 / resolution.int_value;
            double res = (double) s / resolution.int_value;
            glBegin(GL_QUADS);
            for (int i = 0; i < 4; i++)
            {
                for (int j = 0; j < 4; j++)
                    glMultiTexCoord2f(GL_TEXTURE0 + j,
                                      x[i] * res + (x[j] * 2.0 - 1.0) * hpxl,
                                      y[i] * res + (y[j] * 2.0 - 1.0) * hpxl);
                glVertex2f(x[i], y[i]);
            }
            glEnd();

            src = trg;
            trg = !trg;
        }
        glViewport(0, 0, width, height);

        glReadPixels(0, 0, 1, 1, GL_RGBA, GL_FLOAT, (GLvoid *) image);

        pair<double, double> d(image[0] / image[3], image[1]);
        total_mean += d.first;
        max_mean = max(max_mean, d.first);
        total_max += d.second;
        max_max = max(max_max, d.second);
    }

    vector<double> values;
    values.push_back(sqrt(total_mean / s1.num_frames()) * scale);
    values.push_back(sqrt(max_mean) * scale);
    values.push_back(sqrt(total_max / s1.num_frames()) * scale);
    values.push_back(sqrt(max_max) * scale);
    return values;
}

static void cleanup()
{
    glDeleteTextures(3, tex);
    glDeleteFramebuffersEXT(3, fbo);
}

int main(int argc, char **argv)
{
    string comment = alod_generation_string(argc, argv);

    gtk_init(&argc, &argv);
    initialise_gl();

    int args;
    args = process_options(argc, argv);
    Mesh m1, m2;

    try
    {
        format_load_mesh_file(in_format.string_value, argv[args], m1,
                              MESH_LEVEL_PATCHES);
        format_load_mesh_file(in_format.string_value, argv[argc - 1], m2,
                              MESH_LEVEL_PATCHES);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to load mesh: " << e.what() << endl;
        return 1;
    }

    if (m1.skeleton.get_read().num_frames() != m2.skeleton.get_read().num_frames())
    {
        cerr << "Models have different numbers of frames\n";
        return 1;
    }
    if (m1.skeleton.get_read().num_frames() == 0)
    {
        cerr << "Models do not have any frames\n";
        return 1;
    }

    if (!no_scale.used)
    {
        pair<Vector3, Vector3> bbox = mesh_bounding_box_3d(m1);
        Vector3 size = bbox.second - bbox.first;
        scale = 1.0 / max(size[0], max(size[1], size[2]));
    }

    setup();
    if (graph.used)
    {
        size_t inf = 0;
        for (MeshPositionConstIterator p = m2.positions.real_begin(); p != m2.positions.real_end(); p++)
            inf += distance(p->position.begin(), p->position.end());
        int step = graph.int_value;
        vector<double> values;
        double err = 0.0;

        cout << "# " << comment << "\n";
        cout << "# Influences\tRMS\tMax(t) RMS(s)\tRMS(t) Max(s)\tMax\tError\n";
        for (;;)
        {
            values = compare_meshes(m1, m2);
            cout << inf << "\t";
            for (size_t i = 0; i < values.size(); i++)
                cout << values[i] << "\t";
            cout << err * scale << endl;

            size_t nxt = mesh_reduce(m2, (inf - 1) / step * step,
                                     false, err, true);
            if (nxt == inf) break;
            inf = nxt;
        }
    }
    else
    {
        vector<double> values = compare_meshes(m1, m2);
        cout << "RMS:           " << values[0] << "\n";
        cout << "Max(t) RMS(s): " << values[1] << "\n";
        cout << "RMS(t) Max(s): " << values[2] << "\n";
        cout << "Max:           " << values[3] << "\n";
    }
    cleanup();

    return 0;
}
