#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "geom/coords.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <cstdlib>
#include <cstdio>
#include <glib.h>
#include <set>

using namespace std;

static AlodProgramOption in_format, out_format;
static AlodProgramOption verbose, debug;
static AlodProgramOption material;
static AlodProgramOption texture, vertex_normals, align_normals;
static AlodProgramOption expect, unexpect, bound, unbound, flags, flip;
static AlodProgramOption scale, rotation, translation, rest_frame;
static AlodProgramOption sphere, prune;
static string out_file;

static int process_options(int argc, char **argv)
{
    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('o', "output", "output format", 1, PROGRAM_ARGUMENT_STRING, &out_format),
        AlodProgramOptionInfo('m', "material", "only operate on particular material", 1, PROGRAM_ARGUMENT_INT, &material),
        AlodProgramOptionInfo('t', "texture", "set texture map", 1, PROGRAM_ARGUMENT_STRING, &texture),
        AlodProgramOptionInfo('f', "flip", "reflect texture coordinates", 0, PROGRAM_ARGUMENT_STRING, &flip),
        AlodProgramOptionInfo('N', "vertex-normals", "recompute vertex normals", 0, PROGRAM_ARGUMENT_STRING, &vertex_normals),
        AlodProgramOptionInfo('A', "align-normals", "align vertex normals (base level only)", 0, PROGRAM_ARGUMENT_STRING, &align_normals),
        AlodProgramOptionInfo('e', "expect", "extract expectations from animation", 0, PROGRAM_ARGUMENT_STRING, &expect),
        AlodProgramOptionInfo('E', "clear-expect", "clear expectations from mesh", 0, PROGRAM_ARGUMENT_STRING, &unexpect),
        AlodProgramOptionInfo('b', "bounds", "extract bounds from animation", 0, PROGRAM_ARGUMENT_STRING, &bound),
        AlodProgramOptionInfo('B', "clear-bounds", "clear bounds from mesh", 0, PROGRAM_ARGUMENT_STRING, &unbound),
        AlodProgramOptionInfo('F', "flags", "extract transformation flags from animation", 0, PROGRAM_ARGUMENT_STRING, &flags),
        AlodProgramOptionInfo('s', "scale", "rescale the model (0 for autoscale)", 1, PROGRAM_ARGUMENT_DOUBLE, &scale),
        AlodProgramOptionInfo('\0',"rotate", "rotate the model (x y z a)", 1, PROGRAM_ARGUMENT_STRING, &rotation),
        AlodProgramOptionInfo('\0',"translate", "translate the model (x y z)", 1, PROGRAM_ARGUMENT_STRING, &translation),
        AlodProgramOptionInfo('\0', "rest-frame", "copy the given frame to the rest position", 1, PROGRAM_ARGUMENT_INT, &rest_frame),
        AlodProgramOptionInfo('p', "prune", "prune influences", 1, PROGRAM_ARGUMENT_INT, &prune),
        AlodProgramOptionInfo('r', "bounding-sphere", "adding bounding sphere as a comment", 0, PROGRAM_ARGUMENT_STRING, &sphere),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo()
    };

    int args;
    args = alod_process_options(argc, argv, opts);
    if (!in_format.used) in_format.string_value = "auto";
    if (!out_format.used) out_format.string_value = "auto";
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);


    if (argc - args < 1 || argc - args > 2)
    {
        cerr << "Error: must specify input (and optionally output) file\n";
        alod_show_help(opts);
        exit(1);
    }
    if (argc - args == 1)
        out_file = argv[args];
    else
        out_file = argv[args + 1];
    return args;
}

static void flip_texcoords(Mesh &m)
{
    for (MeshVertexIterator v = m.vertices.total_begin(); v != m.vertices.total_end(); v++)
        v->data.texcoords[1] = 1.0 - v->data.texcoords[1];
}

static void prune_positions(Mesh &m, int cutoff)
{
    const Skeleton &skeleton = m.skeleton.get_read();
    InfluenceSimplificationCache cache(skeleton, 32);

    double maxd = 0.0;
    for (MeshPositionIterator p = m.positions.total_begin(); p != m.positions.total_end(); p++)
    {
        VectorPosition q = p->position;
        int b = distance(q.begin(), q.end());
        while (b > cutoff)
        {
            q = simplify_position_cache(cache, q);
            b--;
        }
        double d = (p->position - q) * (p->position - q);
        maxd = max(maxd, d);
        p->position = q;
    }
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Max L2,2 error due to pruning: %f", sqrt(maxd));
}

static void prune_normals(Mesh &m)
{
    const Skeleton &skeleton = m.skeleton.get_read();
    InfluenceSimplificationCache cache(skeleton, 32);

    for (MeshVertexIterator v = m.vertices.total_begin(); v != m.vertices.total_end(); v++)
    {
        const VectorPosition &pos = v->position()->position;
        TangentPlane &tp = v->data.tangent_plane;

        vector<int> pinf;
        VectorPosition::const_iterator j;
        for (j = pos.begin(); j != pos.end(); j++)
            pinf.push_back(j->index);

        for (int i = 0; i < 2; i++)
        {
            vector<int> tinf, dinf;
            for (j = tp.tangents[i].begin(); j != tp.tangents[i].end(); j++)
                tinf.push_back(j->index);
            set_difference(tinf.begin(), tinf.end(), pinf.begin(), pinf.end(),
                           back_inserter(dinf));
            tp.tangents[i] = simplify_position_cache_remove(cache, tp.tangents[i], dinf);
            tp.tangents[i] = normalise(tp.tangents[i]);
        }
        tp.tangents[1] -= (tp.tangents[0] * tp.tangents[1]) * tp.tangents[1];
        tp.tangents[1] = normalise(tp.tangents[1]);
    }
}

static void recompute_vertex_normals(Mesh &m)
{
    mesh_complete(m, MESH_LEVEL_MANIFOLD);
    for (MeshVertexIterator v = m.vertices.all_begin(); v != m.vertices.all_end(); v++)
    {
        v->data.tangent_plane.tangents[0].zero();
        v->data.tangent_plane.tangents[1].zero();
    }
    mesh_compute_normals(m);
}

static void align_vertex_normals(Mesh &m)
{
    /* Note: only works for base-level normals */
    mesh_complete(m, MESH_LEVEL_MANIFOLD);
    for (MeshVertexIterator v = m.vertices.real_begin(); v != m.vertices.real_end(); v++)
    {
        const VectorPosition p = v->position()->position;
        MeshHalfEdgeConstIterator e = v->first_edge();
        Vector2 P(0.0, 0.0);
        do
        {
            MeshVertexConstIterator v2 = e->prev()->vertex();
            VectorPosition p2 = v2->position()->position;
            VectorPosition delta = p2 - p;
            Vector2 proj(delta * v->data.tangent_plane.tangents[0],
                         delta * v->data.tangent_plane.tangents[1]);
            Vector2 param = v2->data.texcoords - v->data.texcoords;
            if (!!proj && !!param)
            {
                proj = normalise(proj);
                param = normalise(param);
                Vector2 rot(proj[0] * param[0] + proj[1] * param[1],
                            proj[1] * param[0] - proj[0] * param[1]);
                P += rot;
            }
            e = e->prev()->dual();
        } while (e->next()->vertex() == v && e != v->first_edge());

        if (!!P)
        {
            TangentPlane &tp = v->data.tangent_plane;
            P = normalise(P);
            const VectorPosition tmp = tp.tangents[0] * P[0] + tp.tangents[1] * P[1];
            tp.tangents[1] = tp.tangents[0] * -P[1] + tp.tangents[1] * P[0];
            tp.tangents[0] = tmp;
        }
    }
}

static void bounding_sphere(Mesh &m, vector<string> &comments)
{
    pair<VectorPosition, double> sphere_as = mesh_bounding_sphere_as(m);
    BoundingSphere sphere_3d = mesh_bounding_sphere_3d(m);

    size_t i = 0;
    while (i < comments.size())
    {
        if (comments[i].substr(0, 15) == "%BoundingSphere")
            comments.erase(comments.begin() + i);
        else
            i++;
    }

    ostringstream sN, s;
    sN << "%BoundingSphereN: " << sphere_as.first << ", " << sphere_as.second;
    s << "%BoundingSphere: " << sphere_3d.centre << ", " << sphere_3d.radius;
    comments.push_back(sN.str());
    comments.push_back(s.str());
}

int main(int argc, char **argv)
{
    Mesh m;
    int args;
    vector<string> comments;

    args = process_options(argc, argv);
    try
    {
        format_load_mesh_file_comments(in_format.string_value, argv[args], m,
                                       MESH_LEVEL_NONE, comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to load mesh: " << e.what() << endl;
        return 1;
    }
    comments.push_back(alod_generation_string(argc, argv));

    if (scale.used || rotation.used || translation.used)
    {
        Matrix4 transform;

        if (translation.used)
        {
            Vector3 offset;
            istringstream in(translation.string_value);
            in >> offset[0] >> offset[1] >> offset[2];
            if (!in || !in.eof())
            {
                cerr << "Bad translation specification\n";
                return 1;
            }
            transform *= matrix_translate(offset);
        }
        if (rotation.used)
        {
            Vector3 axis;
            double angle;
            istringstream in(rotation.string_value);
            in >> axis[0] >> axis[1] >> axis[2] >> angle;
            if (!in || !in.eof() || !axis)
            {
                cerr << "Bad rotation specification\n";
                return 1;
            }
            transform *= matrix_rotate(angle * G_PI / 180.0, axis);
        }
        if (scale.used)
        {
            if (scale.double_value == 0.0)
                scale.double_value = 1.0 / mesh_bounding_sphere_3d(m).radius;
            transform *= matrix_scale(scale.double_value);
        }
        mesh_transform(m, transform);
    }
    if (flip.used) flip_texcoords(m);
    if (flags.used || unexpect.used || expect.used || unbound.used || bound.used)
    {
        Skeleton &skeleton = m.skeleton.get_write();
        if (unexpect.used) skeleton.clear_expect();
        if (unbound.used) skeleton.clear_bounds();
        if (flags.used)
        {
            if (!skeleton.num_frames())
                g_warning("No frames, so flags not updated");
            else
                skeleton.update_flags();
        }
        if (expect.used)
        {
            if (!skeleton.num_frames())
                g_warning("No frames, so expectations not updated");
            else
                skeleton.update_expect();
        }
        if (bound.used)
        {
            if (!skeleton.num_frames())
                g_warning("No frames, so bounds not updated");
            else
                skeleton.update_bounds();
        }
    }
    if (prune.used)
    {
        prune_positions(m, prune.int_value);
        prune_normals(m);
    }
    if (vertex_normals.used) recompute_vertex_normals(m);
    if (align_normals.used) align_vertex_normals(m);
    if (sphere.used) bounding_sphere(m, comments);

    if (material.used)
    {
        if (material.int_value < 0
            || material.int_value >= (int) m.materials.size())
        {
            cerr << "Material id is out of range. Valid range is 0 to "
                << m.materials.size() - 1 << "\n";
            return 1;
        }
    }
    else
        material.int_value = -1;

    if (texture.used)
    {
        m.textures.push_front(MeshTexture());
        mesh_load_texture(texture.string_value, "", m.textures.front());
    }

    int id = 0;
    for (MeshMaterialIterator mat = m.materials.begin(); mat != m.materials.end(); mat++, id++)
        if (material.int_value == -1 || material.int_value == id)
        {
            if (texture.used)
                mat->texture() = m.textures.begin();
        }

    /* Garbage collection for textures, tangent maps and bones */
    mesh_garbage_collect(m);

    if (rest_frame.used)
    {
        if (rest_frame.int_value < 0 || rest_frame.int_value >= (int) m.skeleton.get_read().num_frames())
            g_error("rest_frame is out of range");
        m.skeleton.get_write().set_rest_frame(rest_frame.int_value);
    }

    try
    {
        format_save_mesh_file(out_format.string_value, out_file, m,
                              comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to save mesh: " << e.what() << endl;
        return 1;
    }

    return 0;
}
