#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <gtk/gtk.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <sys/wait.h>
#include <stdarg.h>
#include <mysql/mysql.h>
#include <mysql/mysqld_error.h>
#include <cstdlib>
#include <vector>
#include <utility>
#include "common/paths.h"

using namespace std;

enum
{
    COLUMN_ID = 0,
    COLUMN_STUDENT_NUMBER = 1
};

static MYSQL *db;
static int subject_id = 1, sample_id = -1;
static GtkWidget *window, *run_button, *statusbar;
static guint statusbar_id;

static string video_files[2];
static bool swapped;
static signed char swp;
static my_bool null_swap;

static const gchar *instructions =
    "When you click the Play button, you will be shown two similar\n"
    "animations, one after the other. At the end, you will be asked\n"
    "which of the animations you preferred. Please select one or the\n"
    "other, even if you found it difficult to distinguish between them.\n";


static void popup_error(const char *primary, const char *secondary)
{
    GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(window),
                                               GTK_DIALOG_MODAL,
                                               GTK_MESSAGE_ERROR,
                                               GTK_BUTTONS_OK,
                                               "%s", primary);
    if (secondary && secondary[0])
        gtk_message_dialog_format_secondary_text(GTK_MESSAGE_DIALOG(dialog),
                                                 "%s", secondary);
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_destroy(dialog);
}

/**************************************************************
 * Database code
 */

static bool popup_mysql_error()
{
    popup_error("There was a problem accessing the database. Please ask for assistance.",
                mysql_error(db));
    exit(1);
    return true; // Allows for the && pattern
}

static MYSQL_STMT *prepare_statement(const char *str)
{
    MYSQL_STMT *stmt;
    stmt = mysql_stmt_init(db);
    if (!stmt) popup_mysql_error();
    mysql_stmt_prepare(stmt, str, strlen(str)) && popup_mysql_error();
    return stmt;
}

/* Executes a query, buffers the results and returns the number of rows */
static int do_query_stmt(MYSQL_STMT *stmt)
{
    mysql_stmt_execute(stmt) && popup_mysql_error();
    mysql_stmt_store_result(stmt) && popup_mysql_error();
    return mysql_stmt_num_rows(stmt);
}

/* Similar to do_query_stmt, but for INSERT etc, and returns affected rows */
static int do_exec_stmt(MYSQL_STMT *stmt)
{
    mysql_stmt_execute(stmt) && popup_mysql_error();
    return mysql_stmt_affected_rows(stmt);
}

static void bind_int(MYSQL_BIND *bind, int *i)
{
    memset(bind, 0, sizeof(*bind));
    bind->buffer_type = MYSQL_TYPE_LONG;
    bind->buffer = i;
    bind->buffer_length = sizeof(*i);
}

static void bind_char(MYSQL_BIND *bind, signed char *c)
{
    memset(bind, 0, sizeof(*bind));
    bind->buffer_type = MYSQL_TYPE_TINY;
    bind->buffer = c;
    bind->buffer_length = sizeof(*c);
}

static void bind_string(MYSQL_BIND *bind, char *s, unsigned long size, unsigned long *length, bool in)
{
    memset(bind, 0, sizeof(*bind));
    bind->buffer_type = in ? MYSQL_TYPE_STRING : MYSQL_TYPE_VAR_STRING;
    bind->buffer = s;
    bind->buffer_length = size;
    bind->length = length;
    if (in) *length = strlen(s);
}

static void log_event(const char *type, int subject = -1)
{
    MYSQL_BIND params[3];
    MYSQL_STMT *stmt;
    unsigned long length;
    my_bool is_null = sample_id < 0;

    if (subject == -1) subject = subject_id;
    bind_string(&params[0], (char *) type, strlen(type) + 1, &length, true);
    bind_int(&params[1], &subject);
    bind_int(&params[2], &sample_id);
    params[2].is_null = &is_null;
    stmt = prepare_statement("INSERT INTO eventlog SET event=?, subject=?, sample=?");
    mysql_stmt_bind_param(stmt, params);
    do_exec_stmt(stmt);
    mysql_stmt_close(stmt);
}

static void update_status()
{
    int done, total;
    MYSQL_BIND params[1];
    MYSQL_BIND results[1];
    MYSQL_STMT *stmt;
    gchar *status;

    bind_int(&params[0], &subject_id);
    bind_int(&results[0], &done);
    stmt = prepare_statement("SELECT COUNT(*) FROM results JOIN samples ON results.sample = samples.id WHERE subject=? AND experiment IN (SELECT id FROM experiments)");
    mysql_stmt_bind_param(stmt, params) && popup_mysql_error();
    do_query_stmt(stmt);
    mysql_stmt_bind_result(stmt, results) && popup_mysql_error();
    mysql_stmt_fetch(stmt) && popup_mysql_error();
    mysql_stmt_close(stmt);

    bind_int(&results[0], &total);
    stmt = prepare_statement("SELECT COUNT(*) FROM samples WHERE experiment IN (SELECT id FROM experiments)");
    do_query_stmt(stmt);
    mysql_stmt_bind_result(stmt, results) && popup_mysql_error();
    mysql_stmt_fetch(stmt) && popup_mysql_error();
    mysql_stmt_close(stmt);

    status = g_strdup_printf("Tests completed: %d/%d", done, total);
    gtk_statusbar_pop(GTK_STATUSBAR(statusbar), statusbar_id);
    gtk_statusbar_push(GTK_STATUSBAR(statusbar), statusbar_id, status);
    g_free(status);
}

static void record_result(bool swapped, bool preferred1)
{
    MYSQL_STMT *stmt;
    MYSQL_BIND params[4];
    signed char s = swapped ? 1 : 0, p = preferred1 ? 1 : 0;

    bind_int(&params[0], &subject_id);
    bind_int(&params[1], &sample_id);
    bind_char(&params[2], &s);
    bind_char(&params[3], &p);
    stmt = prepare_statement("INSERT INTO results VALUES (?, ?, ?, ?, CURRENT_TIMESTAMP)");
    mysql_stmt_bind_param(stmt, params) && popup_mysql_error();
    do_exec_stmt(stmt);
    mysql_stmt_close(stmt);
    sample_id = -1;
}

/* True on success, false if there are no more samples to take */
static bool db_choose_videos()
{
    char files[2][512]; unsigned long lengths[2];
    int priority;
    MYSQL_STMT *stmt;
    MYSQL_BIND params[1];
    MYSQL_BIND results[5];

    bind_int(&params[0], &subject_id);
    bind_int(&results[0], &sample_id);
    bind_string(&results[1], files[0], sizeof(files[0]), &lengths[0], false);
    bind_string(&results[2], files[1], sizeof(files[1]), &lengths[1], false);
    bind_char(&results[3], &swp);
    bind_int(&results[4], &priority);
    results[3].is_null = &null_swap;
    stmt = prepare_statement("SELECT samples.id, tests1.filename, tests2.filename, swap, priority FROM samples "
                             "JOIN tests AS tests1 ON samples.scene = tests1.scene"
                             " AND test1_memoryless = tests1.memoryless"
                             " AND test1_as_metric = tests1.as_metric"
                             " AND test1_static_normals = tests1.static_normals"
                             " AND test1_influences = tests1.influences"
                             " AND test1_tolerance = tests1.tolerance "
                             "JOIN tests AS tests2 ON samples.scene = tests2.scene"
                             " AND test2_memoryless = tests2.memoryless"
                             " AND test2_as_metric = tests2.as_metric"
                             " AND test2_static_normals = tests2.static_normals"
                             " AND test2_influences = tests2.influences"
                             " AND test2_tolerance = tests2.tolerance "
                             "WHERE ? NOT IN (SELECT subject FROM results WHERE sample=samples.id)"
                             " AND experiment IN (SELECT id FROM experiments) "
                             "ORDER BY priority DESC");
    mysql_stmt_bind_param(stmt, params) && popup_mysql_error();

    int nrow = do_query_stmt(stmt);
    if (!nrow)
    {
        GtkWidget *dialog = gtk_message_dialog_new(GTK_WINDOW(window),
                                                   GTK_DIALOG_MODAL,
                                                   GTK_MESSAGE_INFO,
                                                   GTK_BUTTONS_OK,
                                                   "All tests are complete. Thank you for your time.");
        gtk_dialog_run(GTK_DIALOG(dialog));
        gtk_widget_destroy(dialog);
        return false;
    }
    mysql_stmt_bind_result(stmt, results) && popup_mysql_error();

    int last = 0;
    mysql_stmt_fetch(stmt) && popup_mysql_error();
    int max_priority = priority;
    for (int i = 1; i < nrow; i++)
    {
        mysql_stmt_fetch(stmt) && popup_mysql_error();
        if (priority == max_priority)
            last = i;
        else
            break;
    }
    mysql_stmt_data_seek(stmt, rand() % (last + 1));
    mysql_stmt_fetch(stmt) && popup_mysql_error();

    video_files[0] = resource_path("ALOD_VIDEOS", ".", string(files[0], lengths[0]));
    video_files[1] = resource_path("ALOD_VIDEOS", ".", string(files[1], lengths[1]));
    mysql_stmt_close(stmt);
    return true;
}

static vector<pair<int, string> > db_get_subjects()
{
    MYSQL_STMT *stmt;
    MYSQL_BIND results[2];
    int id;
    char stdnum[512];
    vector<pair<int, string> > ans;
    int nrow;
    unsigned long length;

    bind_int(&results[0], &id);
    bind_string(&results[1], stdnum, sizeof(stdnum), &length, false);
    stmt = prepare_statement("SELECT id, student_number FROM subjects");
    nrow = do_query_stmt(stmt);
    mysql_stmt_bind_result(stmt, results) && popup_mysql_error();
    ans.resize(nrow);
    for (int i = 0; i < nrow; i++)
    {
        mysql_stmt_fetch(stmt) && popup_mysql_error();
        ans[i].first = id;
        ans[i].second = string(stdnum, length);
    }
    return ans;
}

/* Returns true if added successfully */
static bool db_add_subject(const char *name)
{
    MYSQL_STMT *stmt;
    MYSQL_BIND params[1];
    unsigned long length;

    bind_string(&params[0], (char *) name, strlen(name) + 1, &length, true);
    stmt = prepare_statement("INSERT INTO subjects SET student_number=?");
    mysql_stmt_bind_param(stmt, params) && popup_mysql_error();
    if (mysql_stmt_execute(stmt) != 0)
    {
        if (mysql_stmt_errno(stmt) == ER_DUP_ENTRY)
        {
            mysql_stmt_close(stmt);
            return false;
        }
        else
            popup_mysql_error();
    }
    log_event("newid", mysql_stmt_insert_id(stmt));
    mysql_stmt_close(stmt);
    return true;
}

static void db_init()
{
    my_bool reconnect = true;
    db = mysql_init(NULL);
    if (!db)
    {
        popup_error("Could not initialise the database. Please ask for assistance.", NULL);
        exit(1);
    }

    mysql_options(db, MYSQL_INIT_COMMAND, "SET sql_mode='TRADITIONAL'");
    mysql_options(db, MYSQL_READ_DEFAULT_GROUP, "meshexp");
    mysql_options(db, MYSQL_OPT_RECONNECT, (const char *) &reconnect);

    mysql_real_connect(db,
                       NULL,              /* Host (localhost) */
                       NULL,              /* User */
                       NULL,              /* Password */
                       NULL,              /* Database */
                       0,                 /* Port */
                       0,                 /* UNIX socket */
                       0)                 /* client flag */
        || popup_mysql_error();
}

/**************************************************************
 * GUI code
 */

static gboolean prevent_delete(GtkWidget *widget,
                               GdkEvent *event,
                               gpointer data)
{
    return TRUE;
}

static bool show_video(const string &filename, const string &subfile)
{
    g_setenv("MPLAYER_VERBOSE", "-4", TRUE);
    gchar *argv[] =
    {
        "/usr/bin/mplayer",
        "-fps", "20",
        "-fs",
        "-ao", "null",
        "-sub", (gchar *) subfile.c_str(),
        "-subpos", "0", "-subalign", "0", "-subfont-text-scale", "1",
        (gchar *) filename.c_str(),
        NULL
    };

    GTimer *timer;
    gint status;
    gchar *stderr = NULL;
    GError *error = NULL;

    timer = g_timer_new();
    g_spawn_sync(NULL,                 /* Working dir */
                 argv,                 /* Argument list */
                 NULL,                 /* Environment */
                 G_SPAWN_STDOUT_TO_DEV_NULL,
                 NULL,                 /* Child setup */
                 NULL,                 /* User data */
                 NULL,                 /* Standard output */
                 &stderr,              /* Standard error */
                 &status,              /* Exit status */
                 &error);
    if (error)
    {
        popup_error("Failed to play video. Please ask for assistance.",
                    error->message);
        g_error_free(error);
        g_timer_destroy(timer);
        return false;
    }
    else if (!WIFEXITED(status) || WEXITSTATUS(status) != 0
             || g_timer_elapsed(timer, NULL) < 0.5)
    {
        popup_error("Failed to play video. Please ask for assistance.",
                    stderr);
        g_free(stderr);
        g_timer_destroy(timer);
        return false;
    }
    else
    {
        g_timer_destroy(timer);
        g_free(stderr);
        return true;
    }
}

static void populate_subject_list(GtkListStore *store)
{
    vector<pair<int, string> > subjects;
    GtkTreeIter iter;

    subjects = db_get_subjects();
    gtk_list_store_clear(store);
    for (size_t i = 0; i < subjects.size(); i++)
    {
        gtk_list_store_append(store, &iter);
        gtk_list_store_set(store, &iter,
                           COLUMN_ID, subjects[i].first,
                           COLUMN_STUDENT_NUMBER, subjects[i].second.c_str(),
                           -1);
    }
}

static void add_subject(GtkWindow *parent)
{
    GtkWidget *dialog;
    GtkWidget *stdnum, *stdnum_label, *hbox;

    dialog = gtk_dialog_new_with_buttons("Add subject", parent,
                                         GTK_DIALOG_MODAL,
                                         GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
                                         NULL);
    window = dialog;

    stdnum_label = gtk_label_new("Student number");
    stdnum = gtk_entry_new_with_max_length(127);
    gtk_entry_set_width_chars(GTK_ENTRY(stdnum), 10);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_box_pack_start(GTK_BOX(hbox), stdnum_label, FALSE, FALSE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), stdnum, TRUE, TRUE, 0);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox), hbox);
    gtk_widget_show_all(hbox);

    while (1)
    {
        int response = gtk_dialog_run(GTK_DIALOG(dialog));
        if (response == GTK_RESPONSE_ACCEPT)
        {
            const gchar *sn = gtk_entry_get_text(GTK_ENTRY(stdnum));
            if (!*sn)
            {
                popup_error("Student number must be non-empty", NULL);
                continue;
            }
            if (!db_add_subject(sn))
            {
                popup_error("Student number already exists", NULL);
                continue;
            }
            break; /* Success! */
        }
        else
            break;
    }
    gtk_widget_destroy(dialog);
    window = GTK_WIDGET(parent);
}

/* Enable or disable the OK button, depending on whether anything is selected */
static void selection_changed(GtkTreeSelection *selection, GtkWidget *dialog)
{
    gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog),
                                      GTK_RESPONSE_ACCEPT,
                                      gtk_tree_selection_get_selected(selection, NULL, NULL));
}

static bool get_subject()
{
    GtkWidget *dialog;

    GtkListStore *store;
    GtkWidget *view;
    GtkTreeSelection *selection;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;
    GtkTreeIter iter;

    int response;

    store = gtk_list_store_new(2, G_TYPE_INT, G_TYPE_STRING);
    view = gtk_tree_view_new_with_model(GTK_TREE_MODEL(store));
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(view));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new_with_attributes("Student number",
                                                      renderer,
                                                      "text", COLUMN_STUDENT_NUMBER,
                                                      NULL);
    gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);
    g_object_unref(G_OBJECT(store)); /* Make it die with the view */

    dialog = gtk_dialog_new_with_buttons("Select subject", NULL,
                                         GTK_DIALOG_MODAL,
                                         GTK_STOCK_NEW, 1,
                                         GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
                                         GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
                                         NULL);
    window = dialog;
    gtk_widget_set_size_request(view, 300, 200);
    gtk_container_add(GTK_CONTAINER(GTK_DIALOG(dialog)->vbox),
                      view);
    gtk_widget_show(view);
    gtk_dialog_set_response_sensitive(GTK_DIALOG(dialog), GTK_RESPONSE_ACCEPT, FALSE);
    g_signal_connect(G_OBJECT(selection), "changed",
                     G_CALLBACK(selection_changed), dialog);

    while (1)
    {
        populate_subject_list(store);
        gtk_tree_selection_unselect_all(selection);
        response = gtk_dialog_run(GTK_DIALOG(dialog));
        switch (response)
        {
        case GTK_RESPONSE_ACCEPT:
            if (!gtk_tree_selection_get_selected(selection, NULL, &iter))
                continue;
            gtk_tree_model_get(GTK_TREE_MODEL(store), &iter,
                               COLUMN_ID, &subject_id,
                               -1);
            gtk_widget_destroy(dialog);
            window = NULL;
            log_event("pickid");
            return true;
            break;
        case 1:
            add_subject(GTK_WINDOW(dialog));
            continue;
        default:
            /* Various cancellations */
            gtk_widget_destroy(dialog);
            window = NULL;
            return false;
        }
    }
}

static void run_clicked(GtkWidget *button, gpointer user_data)
{
    GtkWidget *dialog;
    string sub1 = resource_path("ALOD_VIDEOS", ".", "scene1.srt");
    string sub2 = resource_path("ALOD_VIDEOS", ".", "scene2.srt");

    if (null_swap)
        swapped = rand() < RAND_MAX / 2;
    else
        swapped = swp;
    log_event("play");
    if (!show_video(video_files[swapped ? 1 : 0], sub1)
        || !show_video(video_files[swapped ? 0 : 1], sub2))
    {
        log_event("badvideo");
        return;
    }
    log_event("stop");

    dialog = gtk_message_dialog_new(GTK_WINDOW(window),
                                    GTK_DIALOG_MODAL,
                                    GTK_MESSAGE_QUESTION,
                                    GTK_BUTTONS_NONE,
                                    "Which animation did you prefer?");
    gtk_dialog_add_buttons(GTK_DIALOG(dialog),
                           "Animation 1", 1,
                           "Animation 2", 2,
                           NULL);
    gtk_window_set_deletable(GTK_WINDOW(dialog), FALSE);
    g_signal_connect(G_OBJECT(dialog), "delete_event",
                     G_CALLBACK(prevent_delete), NULL);

    gtk_widget_show(dialog);

    int result = 0;
    do
    {
        gtk_window_set_focus(GTK_WINDOW(dialog), NULL);
        result = gtk_dialog_run(GTK_DIALOG(dialog));
    } while (result <= 0);
    gtk_widget_destroy(dialog);

    bool choose1 = (result == 1) ^ swapped;
    log_event(choose1 ? "choose1" : "choose2");
    record_result(swapped, choose1);
    update_status();
    if (!db_choose_videos())
        gtk_main_quit();
}

static void create_window()
{
    GtkWidget *box, *event_box, *ebbox, *align, *inst;

    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(window), "User experiments");
    gtk_window_set_default_size(GTK_WINDOW(window), 800, 600);
    if (1)
    {
        gtk_window_set_deletable(GTK_WINDOW(window), FALSE);
        gtk_window_set_decorated(GTK_WINDOW(window), FALSE);
        gtk_window_fullscreen(GTK_WINDOW(window));
        g_signal_connect(G_OBJECT(window), "delete_event",
                         G_CALLBACK(prevent_delete), NULL);
    }
    g_signal_connect(G_OBJECT(window), "destroy",
                     G_CALLBACK(gtk_main_quit), NULL);

    box = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(window), box);
    event_box = gtk_event_box_new();
    ebbox = gtk_vbox_new(FALSE, 0);
    gtk_container_add(GTK_CONTAINER(event_box), ebbox);
    gtk_box_pack_start(GTK_BOX(box), event_box, TRUE, TRUE, 0);
    gtk_widget_set_name(event_box, "meshexp_bg");

    inst = gtk_label_new(instructions);
    gtk_box_pack_start(GTK_BOX(ebbox), inst, FALSE, FALSE, 10);
    gtk_widget_set_name(inst, "meshexp_inst");

    align = gtk_alignment_new(0.5, 0.25, 0.0, 0.0);
    gtk_box_pack_start(GTK_BOX(ebbox), align, TRUE, TRUE, 0);

    statusbar = gtk_statusbar_new();
    gtk_statusbar_set_has_resize_grip(GTK_STATUSBAR(statusbar), FALSE);
    statusbar_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
                                                "Tests done");
    gtk_statusbar_push(GTK_STATUSBAR(statusbar), statusbar_id, "");
    gtk_box_pack_start(GTK_BOX(box), statusbar, FALSE, FALSE, 0);

    run_button = gtk_button_new_from_stock(GTK_STOCK_MEDIA_PLAY);
    g_signal_connect(G_OBJECT(run_button), "clicked",
                     G_CALLBACK(run_clicked), NULL);
    gtk_container_add(GTK_CONTAINER(align), run_button);

    gtk_widget_show_all(window);
}

/* Does the query and pops up a dialog if there is an error. It is still
 * up to the caller to take action on an error.
 */
int main(int argc, char **argv)
{
    srand(time(NULL));
    gtk_init(&argc, &argv);
    db_init();
    log_event("open");
    if (get_subject() && db_choose_videos())
    {
        create_window();
        update_status();
        gtk_main();
    }
    log_event("close");
    mysql_close(db);
    return 0;
}
