/* TODO:
 * - more verbose status messages
 * - option to take influences from another file?
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "geom/coords.h"
#include "geom/sparse.h"
#include "geom/skeleton.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include "common/maputils.h"
#include "common/mesch.h"
#include "common/timing.h"
#include <cstdio>
#include <vector>
#include <map>
#include <queue>
#include <utility>
#include <cstddef>
#include <glib.h>
#include <string>
#include <sstream>

using namespace std;

static AlodProgramOption in_format, out_format;
static AlodProgramOption blur, influence_file;
static AlodProgramOption verbose, debug;
static AlodProgramOption rigid_bones, skeleton_only, skeleton_singular, adjust_skeleton, linear_bones;
static AlodProgramOption iterations;
static AlodProgramOption edge_reg, vertex_reg, skeleton_reg;

static vector<vector<Vector4> > ref_positions;
static vector<vector<Matrix4> > ref_skeletons;

static int process_options(int argc, char **argv)
{
    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('o', "output", "output format", 1, PROGRAM_ARGUMENT_STRING, &out_format),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo('b', "blur", "filter radius for influence sets", 1, PROGRAM_ARGUMENT_DOUBLE, &blur),
        AlodProgramOptionInfo('B', "influences", "load influence sets from separate file", 1, PROGRAM_ARGUMENT_STRING, &influence_file),
        AlodProgramOptionInfo('r', "rigid", "force bones to be rigid", 0, PROGRAM_ARGUMENT_STRING, &rigid_bones),
        AlodProgramOptionInfo('l', "linear", "use existing translations", 0, PROGRAM_ARGUMENT_STRING, &linear_bones),
        AlodProgramOptionInfo('a', "adjust-skeleton", "keep translations as constant as possible", 0, PROGRAM_ARGUMENT_STRING, &adjust_skeleton),
        AlodProgramOptionInfo('s', "skeleton-only", "fit skeleton but not points", 0, PROGRAM_ARGUMENT_STRING, &skeleton_only),
        AlodProgramOptionInfo('S', "singular", "fit skeletons to singular regions", 0, PROGRAM_ARGUMENT_STRING, &skeleton_singular),
        AlodProgramOptionInfo('I', "iterations", "number of LSQR iterations [10000]", 1, PROGRAM_ARGUMENT_INT, &iterations),
        AlodProgramOptionInfo('e', "edge-reg", "weight for edge regularisation", 1, PROGRAM_ARGUMENT_DOUBLE, &edge_reg),
        AlodProgramOptionInfo('w', "vertex-reg", "weight for vertex regularisation", 1, PROGRAM_ARGUMENT_DOUBLE, &vertex_reg),
        AlodProgramOptionInfo('R', "skeleton-reg", "weight for skeleton regularisation", 1, PROGRAM_ARGUMENT_DOUBLE, &skeleton_reg),
        AlodProgramOptionInfo()
    };

    int args;
    args = alod_process_options(argc, argv, opts);
    if (!in_format.used) in_format.string_value = "auto";
    if (!out_format.used) out_format.string_value = "auto";
    if (!iterations.used) iterations.int_value = 10000;
    if (!skeleton_reg.used) skeleton_reg.double_value = 1e-3;
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);

    if (argc - args < 2)
    {
        cerr << "Error: must specify input files and output file\n";
        alod_show_help(opts);
        exit(1);
    }
    return args;
}

static void show_info(ITER *ip, double res, VEC *, VEC *)
{
    if (ip->steps == 0)
        fprintf(stderr, "Iter: ");
    else
    {
        fprintf(stderr, "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    }
    fprintf(stderr, "%-6d  % 15.6f", ip->steps, res);
    fflush(stderr);
}

static void sp_add_val(SPMAT *m, int r, int c, double d)
{
    sp_set_val(m, r, c, d + sp_get_val(m, r, c));
}

static MeshPositionIterator get_correspond(const map<pair<double, double>, MeshPositionIterator> &correspond,
                                           MeshPositionConstIterator p)
{
    /* first_vertex might be a hole vertex, so take the second one. */
    MeshHalfEdgeConstIterator e = p->first_vertex()->first_edge();
    const Vector2 &texcoords = e->dual()->prev()->vertex()->data.texcoords;
    pair<double, double> key(texcoords[0], texcoords[1]);
    map<pair<double, double>, MeshPositionIterator>::const_iterator item;
    item = correspond.find(key);
    if (item == correspond.end())
        g_error("Correspondence failed");
    return item->second;
}

static void make_correspond(map<pair<double, double>, MeshPositionIterator> &correspond,
                            Mesh &m)
{
    correspond.clear();
    for (MeshVertexIterator v = m.vertices.all_begin();
         v != m.vertices.all_end(); v++)
    {
        pair<double, double> key(v->data.texcoords[0],
                                 v->data.texcoords[1]);
        if (correspond.count(key)
            && correspond[key] != v->position())
            g_error("Duplicate texcoords %f %f", key.first, key.second);
        correspond[key] = v->position();
    }
}

static VEC *my_iter_cg(SPMAT *A, VEC *b, double tol, VEC *x, int limit, int *steps)
{
    ITER *ip;

    x = v_resize(x, A->n);
    v_zero(x);
    ip = iter_get(0, 0);
    ip->Ax = (Fun_Ax) sp_mv_mlt;
    ip->A_par = (void *) A;
    ip->Bx = (Fun_Ax) NULL;
    ip->B_par = NULL;
    if (verbose.used || debug.used)
        ip->info = (void (*)()) (Fun_info) show_info;
    else
        ip->info = NULL;
    ip->b = b;
    ip->eps = tol;
    ip->limit = limit;
    ip->x = x;
    iter_cg(ip);
    x = ip->x;
    if (steps) *steps = ip->steps;
    ip->shared_x = ip->shared_b = TRUE;
    iter_free(ip);
    if (verbose.used || debug.used)
        fputs("\n", stderr);
    return x;
}

static void rigidify(vector<vector<Matrix4> > &T)
{
    size_t n_mesh = T.size();
    size_t sz = T[0].size();

    for (size_t m = 0; m < n_mesh; m++)
        for (size_t i = 1; i < sz; i++)
            T[m][i] = matrix3_translate3_to_matrix4(polar_rotation(matrix4_to_matrix3(T[m][i])),
                                                    matrix4_to_translate3(T[m][i]));
}

static void fit_transforms_full(vector<vector<Matrix4> > &T, const Mesh &control)
{
    const Skeleton &control_skeleton = control.skeleton.get_read();
    size_t sz = control_skeleton.size();
    size_t n_mesh = ref_positions.size();
    SPMAT *coeff;
    VEC *target = VNULL;
    VEC *x = VNULL;

    coeff = sp_get(12 * sz, 12 * sz, 12 * sz);
    for (size_t m = 0; m < n_mesh; m++)
    {
        const vector<Vector4> &cur = ref_positions[m];
        target = v_resize(target, 12 * sz);

        sp_zero(coeff);
        v_zero(target);
        /* Regularise the transformations */
        for (size_t i = 0; i < sz; i++)
        {
            for (size_t j = 0; j < 12; j++)
                sp_set_val(coeff, i * 12 + j, i * 12 + j, skeleton_reg.double_value * skeleton_reg.double_value);
            for (size_t j = 0; j < 12; j += 5)
                v_set_val(target, i * 12 + j, skeleton_reg.double_value * skeleton_reg.double_value);
        }

        size_t pindex = 0;
        for (MeshPositionConstIterator p = control.positions.all_begin();
             p != control.positions.all_end(); p++, pindex++)
        {
            const Vector4 &posc = cur[pindex];
            const VectorPosition &pos = p->position;
            if (skeleton_singular.used
                && distance(pos.begin(), pos.end()) != 1) continue;

            for (int r = 0; r < 3; r++)
            {
                for (VectorPosition::const_iterator b = pos.begin(); b != pos.end(); b++)
                    for (int j = 0; j < 4; j++)
                    {
                        int bone = b->index / 4;
                        int row = 12 * bone + 4 * r + j;
                        Vector4 v(b->values[0], b->values[1], b->values[2], b->values[3]);
                        v = control_skeleton.get_absolute(bone) * v;
                        for (VectorPosition::const_iterator b2 = pos.begin(); b2 != pos.end(); b2++)
                            for (int j2 = 0; j2 < 4; j2++)
                            {
                                int bone2 = b2->index / 4;
                                int col = 12 * bone2 + 4 * r + j2;
                                Vector4 v2(b2->values[0], b2->values[1], b2->values[2], b2->values[3]);
                                v2 = control_skeleton.get_absolute(bone2) * v2;
                                sp_add_val(coeff, row, col, v[j] * v2[j2]);
                            }
                        v_add_val(target, row, v[j] * posc[r]);
                    }
            }
        }

        x = my_iter_cg(coeff, target, 1e-8, x, 1000, NULL);
        for (size_t i = 1; i < sz; i++)
        {
            for (int j = 0; j < 3; j++)
                for (int k = 0; k < 4; k++)
                    T[m][i][j][k] = v_entry(x, 12 * i + 4 * j + k);
        }
    }
    SP_FREE(coeff);
    V_FREE(target);
    V_FREE(x);
}

/* Solves for translations of just the components directly below the root. */
static void fit_transforms_translate_subroot(vector<vector<Matrix4> > &T, const Mesh &control)
{
    SPMAT *coeff;
    VEC *target, *x;
    const Skeleton &control_skeleton = control.skeleton.get_read();
    size_t n_mesh = T.size();
    size_t sz = T[0].size();
    size_t top;
    vector<int> tops;

    for (size_t i = 1; i < sz; i++)
        if (control_skeleton.get_bone(i).parent == 0)
            tops.push_back(i);
    top = tops.size();

    coeff = sp_get(3 * top, 3 * top, 3 * top);
    target = v_get(3 * top);
    x = v_get(3 * top);
    for (size_t m = 0; m < n_mesh; m++)
    {
        const vector<Vector4> &cur = ref_positions[m];
        sp_zero(coeff);
        /* Regularisation */
        for (size_t i = 0; i < 3 * top; i++)
            sp_set_val(coeff, i, i, skeleton_reg.double_value);
        v_zero(target);

        size_t pindex = 0;
        for (MeshPositionConstIterator p = control.positions.all_begin();
             p != control.positions.all_end(); p++, pindex++)
        {
            const VectorPosition &pos = p->position;
            if (skeleton_singular.used
                && distance(pos.begin(), pos.end()) != 1) continue;
            Vector4 posc = cur[pindex];

            // w coordinate for each subroot
            vector<double> weights(top, 0.0);
            for (VectorPosition::const_iterator b = pos.begin(); b != pos.end(); b++)
            {
                int bone = b->index / 4;
                Vector4 v(b->values[0], b->values[1], b->values[2], b->values[3]);
                posc -= T[m][bone] * (control_skeleton.get_absolute(bone) * v);

                while (control_skeleton.get_bone(bone).parent > 0)
                    bone = control_skeleton.get_bone(bone).parent;
                size_t top_index = find(tops.begin(), tops.end(), bone) - tops.begin();
                weights[top_index] += b->values[3];
            }
            for (int r = 0; r < 3; r++)
                for (size_t i = 0; i < top; i++)
                    if (weights[i])
                    {
                        int row = 3 * i + r;
                        for (size_t j = 0; j < top; j++)
                            if (weights[j])
                            {
                                int col = 3 * j + r;
                                sp_add_val(coeff, row, col, weights[i] * weights[j]);
                            }
                        v_add_val(target, row, weights[i] * posc[r]);
                    }
        }

        x = my_iter_cg(coeff, target, 1e-8, x, 1000, NULL);

        for (size_t i = 1; i < sz; i++)
        {
            int bone = i;
            while (control_skeleton.get_bone(bone).parent > 0)
                bone = control_skeleton.get_bone(bone).parent;
            size_t top_index = find(tops.begin(), tops.end(), bone) - tops.begin();

            for (int j = 0; j < 3; j++)
                T[m][i][j][3] += v_entry(x, 3 * top_index + j);
        }
    }
}

/* Fits the transforms on the basis that control skeleton is rigid i.e.,
 * the bones cannot change length (translation part of L_i). The exception
 * is that for bones directly under the root, the translation may change
 * (this is predicated on the root being a pseudo-root). The root is
 * considered to be fixed.
 */
static void fit_transforms_linear(vector<vector<Matrix4> > &T, const Mesh &control)
{
    const Skeleton &control_skeleton = control.skeleton.get_read();
    size_t n_mesh = ref_positions.size();
    vector<int> tops;
    size_t sz = control_skeleton.size();
    size_t top, vars;
    SPMAT *coeff;
    VEC *target, *x;
    vector<pair<int, double> > eqn;

    for (size_t i = 1; i < sz; i++)
        if (control_skeleton.get_bone(i).parent == 0)
            tops.push_back(i);
    top = tops.size();
    vars = 9 * (sz - 1) + 3 * top;
    coeff = sp_get(vars, vars, vars);
    target = v_get(vars);
    x = v_get(vars);
    for (size_t m = 0; m < n_mesh; m++)
    {
        const vector<Vector4> &cur = ref_positions[m];
        Matrix3 Gl[sz];     /* Original global transforms - linear part */
        Vector3 Lnewt[sz];  /* New offset component */

        for (size_t i = 1; i < sz; i++)
        {
            Lnewt[i] = matrix4_to_translate3(control_skeleton.get_frame(m)[i]);
            Gl[i] = matrix4_to_matrix3(control_skeleton.get_absolute(i));
        }

        sp_zero(coeff);
        v_zero(target);
        /* Regularise the transformations */
        for (size_t i = 1; i < sz; i++)
        {
            for (size_t j = 0; j < 9; j++)
                sp_set_val(coeff, (i - 1) * 9 + j, (i - 1) * 9 + j, skeleton_reg.double_value * skeleton_reg.double_value);
            for (size_t j = 0; j < 9; j += 4)
                v_set_val(target, (i - 1) * 9 + j, skeleton_reg.double_value * skeleton_reg.double_value);
        }
        for (size_t i = 0; i < top; i++)
            for (size_t j = 0; j < 3; j++)
            {
                sp_set_val(coeff, (sz - 1) * 9 + 3 * i + j, (sz - 1) * 9 + 3 * i + j,
                           skeleton_reg.double_value * skeleton_reg.double_value);
                v_set_val(target, (sz - 1) * 9 + 3 * i + j,
                          skeleton_reg.double_value * skeleton_reg.double_value);
            }

        size_t pindex = 0;
        for (MeshPositionConstIterator p = control.positions.all_begin();
             p != control.positions.all_end(); p++, pindex++)
        {
            const Vector4 &posc = cur[pindex];
            const VectorPosition &pos = p->position;
            if (skeleton_singular.used
                && distance(pos.begin(), pos.end()) != 1) continue;

            /* Vector coefficients for each of the T_i rotation components. */
            vector<Vector3> block(sz);
            /* Scalar coefficients for each of the subroot offsets */
            vector<double> scalar(top);

            Vector3 trg = vector4_to_vector3(posc);
            for (VectorPosition::const_iterator i = pos.begin(); i != pos.end(); i++)
            {
                Vector3 l(i->values[0], i->values[1], i->values[2]);
                double w = i->values[3];
                int bone = i->index / 4;
                int pbone = control_skeleton.get_bone(bone).parent;
                if (pbone == -1)
                {
                    /* Assumptions don't cater for this to happen. Exclude it
                     * from the equations and hope it doesn't affect things.
                     */
                    g_assert_not_reached(); 
                }
                else
                {
                    block[bone] += Gl[bone] * l;
                    while (pbone != 0)
                    {
                        block[pbone] += w * Gl[pbone] * Lnewt[bone];
                        bone = pbone;
                        pbone = control_skeleton.get_bone(bone).parent;
                    }
                    trg -= w * Lnewt[bone]; /* Since R[0] == identity */
                    size_t top_index = find(tops.begin(), tops.end(), bone) - tops.begin();
                    g_assert(top_index < top);
                    scalar[top_index] += w;
                }
            }

            for (int eq = 0; eq < 3; eq++)
            {
                /* Build array of coefficients. Each element is a (index, value) pair */
                eqn.clear();
                for (size_t i = 1; i < sz; i++)
                    for (int i2 = 0; i2 < 3; i2++)
                        if (block[i][i2])
                            eqn.push_back(make_pair((i - 1) * 9 + eq * 3 + i2, block[i][i2]));
                for (size_t i = 0; i < top; i++)
                    if (scalar[i])
                        eqn.push_back(make_pair((sz - 1) * 9 + i * 3 + eq, scalar[i]));

                /* Add to the normal equations */
                for (size_t i = 0; i < eqn.size(); i++)
                {
                    for (size_t j = 0; j < eqn.size(); j++)
                        sp_add_val(coeff, eqn[i].first, eqn[j].first,
                                   eqn[i].second * eqn[j].second);
                    v_add_val(target, eqn[i].first, eqn[i].second * trg[eq]);
                }
            }
        }

        x = my_iter_cg(coeff, target, 1e-8, x, 1000, NULL);
        /* Extract the linear parts */
        for (size_t i = 1; i < sz; i++)
            for (int j = 0; j < 3; j++)
                for (int k = 0; k < 3; k++)
                    T[m][i][j][k] = v_entry(x, 9 * (i - 1) + 3 * j + k);
        if (rigid_bones.used)
            rigidify(T);

        /* Sort out the translation parts */
        vector<Matrix4> Gold(sz), Gnew(sz);
        for (size_t i = 1; i < sz; i++)
        {
            Matrix4 Lold, Lnew;
            /* Compute old and new global and local spaces (only rotation is
             * correct for new spaces) */
            int p = control_skeleton.get_bone(i).parent;
            Gold[i] = control_skeleton.get_absolute(i);
            Lold = control_skeleton.get_bone(i).rest_transform;
            Gnew[i] = T[m][i] * Gold[i];
            Lnew = inverse(Gnew[p]) * Gnew[i];

            /* Fill in translation part and reverse the process */
            for (int j = 0; j < 3; j++)
                Lnew[j][3] = Lnewt[i][j];
            if (p == 0)
            {
                size_t top_index = find(tops.begin(), tops.end(), (int) i) - tops.begin();
                g_assert(top_index < top);
                for (int j = 0; j < 3; j++)
                    Lnew[j][3] += v_entry(x, 9 * (sz - 1) + 3 * top_index + j);
            }
            Gnew[i] = Gnew[p] * Lnew;
            T[m][i] = Gnew[i] * inverse(Gold[i]);
        }

        /* Fix up the translations, which may have been invalidated by
         * the rigidification.
         */
        if (rigid_bones.used)
            fit_transforms_translate_subroot(T, control);
    }
    V_FREE(x);
    V_FREE(target);
    SP_FREE(coeff);
}

static void fit_transforms_translate(vector<vector<Matrix4> > &T, const Mesh &control)
{
    SPMAT *coeff;
    VEC *target, *x;
    const Skeleton &control_skeleton = control.skeleton.get_read();
    size_t n_mesh = T.size();
    size_t sz = T[0].size();

    coeff = sp_get(3 * sz, 3 * sz, 3 * sz);
    target = v_get(3 * sz);
    x = v_get(3 * sz);
    for (size_t m = 0; m < n_mesh; m++)
    {
        const vector<Vector4> &cur = ref_positions[m];
        sp_zero(coeff);
        /* Regularisation */
        for (size_t i = 0; i < 3 * sz; i++)
            sp_set_val(coeff, i, i, skeleton_reg.double_value);
        v_zero(target);

        size_t pindex = 0;
        for (MeshPositionConstIterator p = control.positions.all_begin();
             p != control.positions.all_end(); p++, pindex++)
        {
            const Vector4 &posc = cur[pindex];
            const VectorPosition &pos = p->position;
            if (skeleton_singular.used
                && distance(pos.begin(), pos.end()) != 1) continue;

            for (int r = 0; r < 3; r++)
            {
                double val = posc[r];
                for (VectorPosition::const_iterator b = pos.begin(); b != pos.end(); b++)
                {
                    int bone = b->index / 4;
                    Vector4 v(b->values[0], b->values[1], b->values[2], b->values[3]);
                    v = control_skeleton.get_absolute(bone) * v;
                    val -= T[m][bone][r][0] * v[0] + T[m][bone][r][1] * v[1] + T[m][bone][r][2] * v[2];
                }

                for (VectorPosition::const_iterator b = pos.begin(); b != pos.end(); b++)
                {
                    int bone = b->index / 4;
                    int row = 3 * bone + r;
                    double v = b->values[3];
                    for (VectorPosition::const_iterator b2 = pos.begin(); b2 != pos.end(); b2++)
                    {
                        int bone2 = b2->index / 4;
                        int col = 3 * bone2 + r;
                        double v2 = b2->values[3];
                        sp_add_val(coeff, row, col, v * v2);
                    }
                    v_add_val(target, row, v * val);
                }
            }
        }

        x = my_iter_cg(coeff, target, 1e-8, x, 1000, NULL);

        for (size_t i = 1; i < sz; i++)
            for (int j = 0; j < 3; j++)
                T[m][i][j][3] = v_entry(x, 3 * i + j);
    }
}

/* Adjust the control skeleton to keep the translations as constant
 * as possible. We use the current value as a regularisation.
 */
static void do_adjust_skeleton(const vector<vector<Matrix4> > &T, Mesh &control)
{
    Skeleton &control_skeleton = control.skeleton.get_write();
    size_t n_mesh = T.size();
    size_t sz = T[0].size();
    Matrix4 G[sz];
    vector<SkeletonBoneInfo> bones(sz);
    MAT *A;
    VEC *d, *target, *x;

    for (size_t i = 0; i < sz; i++)
    {
        bones[i] = control_skeleton.get_bone(i);
        G[i] = control_skeleton.get_absolute(i);
    }
    A = m_get(3 * n_mesh + 3, 6);
    target = v_get(3 * n_mesh + 3);
    d = v_get(6);
    x = v_get(6);

    for (size_t i = 1; i < sz; i++)
    {
        /* We're going to cancel the orientation on this bone, so we
         * fix the tail position to this orientation
         */
        bones[i].tail = matrix4_to_matrix3(G[i]) * bones[i].tail;

        size_t p = bones[i].parent;
        g_assert(p < i);
        m_zero(A);

        /* We solve for the global position of the head. See notebook
         * for details.
         */
        for (size_t m = 0; m < n_mesh; m++)
        {
            Matrix4 C = inverse(T[m][p]) * T[m][i];
            for (int j = 0; j < 3; j++)
            {
                for (int k = 0; k < 3; k++)
                    m_set_val(A, 3 * m + j, k, C[j][k]);
                m_set_val(A, 3 * m + j, 3 + j, -1.0); // Subtract m_j
            }
            for (int j = 0; j < 3; j++)
                v_set_val(target, 3 * m + j, -C[j][3]);
        }
        for (size_t j = 0; j < 3; j++)
        {
            m_set_val(A, 3 * n_mesh + j, j, skeleton_reg.double_value);
            v_set_val(target, 3 * n_mesh + j, G[i][j][3] * skeleton_reg.double_value);
        }
        QRfactor(A, d);
        x = QRsolve(A, d, target, x);
        bones[i].rest_transform.identity();
        for (int j = 0; j < 3; j++)
            bones[i].rest_transform[j][3] = v_entry(x, j) - G[p][j][3];
        G[i] = G[p] * bones[i].rest_transform;

        /* Mark tails that can be recomputed */
        bones[p].tail.zero();
    }
    control_skeleton.set_bones(bones);
    M_FREE(A);
    V_FREE(d);
    V_FREE(target);
    V_FREE(x);
}

static void fit_skeletons(Mesh &control)
{
    Skeleton &control_skeleton = control.skeleton.get_write();
    control_skeleton.set_to_rest();
    size_t sz = control_skeleton.size();
    size_t n_mesh = ref_positions.size();
    vector<vector<Matrix4> > T;

    T.resize(n_mesh);
    for (size_t m = 0; m < n_mesh; m++)
        T[m].resize(sz);

    if (!linear_bones.used)
    {
        fit_transforms_full(T, control);
        if (rigid_bones.used)
        {
            rigidify(T);
            fit_transforms_translate(T, control);
        }
    }
    else
        fit_transforms_linear(T, control);

    if (adjust_skeleton.used)
        do_adjust_skeleton(T, control);

    /* Compute frames */
    control_skeleton.clear_frames();
    ref_skeletons.clear();
    ref_skeletons.resize(n_mesh, vector<Matrix4>(sz));
    vector<Matrix4> frame(sz);
    for (size_t m = 0; m < n_mesh; m++)
    {
        for (size_t i = 1; i < sz; i++)
        {
            int p = control_skeleton.get_bone(i).parent;
            Matrix4 Gp = control_skeleton.get_absolute(p);
            Matrix4 Gi = control_skeleton.get_absolute(i);
            ref_skeletons[m][i] = T[m][i] * Gi;
            frame[i] = inverse(ref_skeletons[m][p]) * ref_skeletons[m][i];
        }
        control_skeleton.add_frame(frame);
    }
    if (adjust_skeleton.used)
        control_skeleton.fix_tails();
}

struct lsqr_point_data
{
    size_t var_offset;
    vector<int> bones;
};

struct lsqr_edge_data
{
    size_t p1, p2;
    double weight;
};

struct lsqr_data
{
    size_t n_point;
    size_t n_mesh;
    size_t n_edge;
    size_t n_eq;
    size_t n_var;

    vector<lsqr_point_data> points;
    vector<lsqr_edge_data> edges;
};

/* Handles the elimination of the last w coordinate. Note that we return
 * -w not 1.0 - w, because properly the 1.0 should be incorporated into
 * b and not A.
 */
static inline double lsqr_coordinate(const lsqr_point_data &p, size_t b, int j, VEC *x)
{
    if (j != 3 || b + 1 != p.bones.size())
        return v_entry(x, p.var_offset + 4 * b + j);
    else
    {
        double w = 0.0;
        for (size_t i = 0; i < b; i++)
            w += v_entry(x, p.var_offset + 4 * i + 3);
        return -w;
    }
}

VEC *lsqr_Ax(void *arg, VEC *x, VEC *out)
{
    lsqr_data *data = (lsqr_data *) arg;
    size_t eq = 0;

    out = v_resize(out, data->n_eq);
    for (size_t i = 0; i < data->n_point; i++)
    {
        const lsqr_point_data &p = data->points[i];

        vector<Vector4> coords(p.bones.size());
        for (size_t j = 0; j < p.bones.size(); j++)
            for (int k = 0; k < 4; k++)
                coords[j][k] = lsqr_coordinate(p, j, k, x);
        for (size_t j = 0; j < data->n_mesh; j++)
        {
            Vector4 pos(0.0, 0.0, 0.0, 0.0);
            for (size_t k = 0; k < p.bones.size(); k++)
                pos += ref_skeletons[j][p.bones[k]] * coords[k];
            for (int k = 0; k < 3; k++)
                v_set_val(out, eq++, pos[k]);
        }
    }

    if (edge_reg.used)
    {
        for (size_t i = 0; i < data->n_edge; i++)
        {
            const lsqr_point_data &p1 = data->points[data->edges[i].p1];
            const lsqr_point_data &p2 = data->points[data->edges[i].p2];
            double w = data->edges[i].weight;
            size_t b1 = 0;
            size_t b2 = 0;
            while (b1 < p1.bones.size() || b2 < p2.bones.size())
            {
                bool do1 = false, do2 = false;
                if (b1 == p1.bones.size()) do2 = true;
                else if (b2 == p2.bones.size()) do1 = true;
                else if (p1.bones[b1] < p2.bones[b2]) do1 = true;
                else if (p1.bones[b1] > p2.bones[b2]) do2 = true;
                else do1 = true, do2 = true;
                for (int j = 0; j < 4; j++)
                {
                    v_set_val(out, eq, 0.0);
                    if (do1)
                        v_add_val(out, eq, lsqr_coordinate(p1, b1, j, x) * w);
                    if (do2)
                        v_sub_val(out, eq, lsqr_coordinate(p2, b2, j, x) * w);
                    eq++;
                }
                if (do1) b1++;
                if (do2) b2++;
            }
        }
    }

    if (vertex_reg.used)
        for (size_t i = 0; i < data->n_point; i++)
        {
            const lsqr_point_data &p = data->points[i];
            for (size_t j = 0; j < p.bones.size(); j++)
                for (int k = 0; k < 4; k++)
                    v_set_val(out, eq++, lsqr_coordinate(p, j, k, x) * vertex_reg.double_value);
        }
    g_assert(eq == data->n_eq);

    return out;
}

static inline void lsqr_accumulate(const lsqr_point_data &p, size_t b, int j, double v, VEC *out)
{
    if (j != 3 || b + 1 != p.bones.size())
        v_add_val(out, p.var_offset + 4 * b + j, v);
    else
    {
        for (size_t i = 0; i < b; i++)
            v_sub_val(out, p.var_offset + 4 * i + 3, v);
    }
}

VEC *lsqr_ATx(void *arg, VEC *x, VEC *out)
{
    lsqr_data *data = (lsqr_data *) arg;
    size_t eq = 0;

    out = v_resize(out, data->n_var);
    v_zero(out);
    for (size_t i = 0; i < data->n_point; i++)
    {
        const lsqr_point_data &p = data->points[i];
        for (size_t j = 0; j < data->n_mesh; j++)
        {
            Vector4 weights(0.0, 0.0, 0.0, 0.0);
            for (int k = 0; k < 3; k++)
                weights[k] = v_entry(x, eq++);
            for (size_t k = 0; k < p.bones.size(); k++)
            {
                Vector4 pos = transpose(ref_skeletons[j][p.bones[k]]) * weights;
                for (int l = 0; l < 4; l++)
                    lsqr_accumulate(p, k, l, pos[l], out);
            }
        }
    }

    if (edge_reg.used)
    {
        for (size_t i = 0; i < data->n_edge; i++)
        {
            const lsqr_point_data &p1 = data->points[data->edges[i].p1];
            const lsqr_point_data &p2 = data->points[data->edges[i].p2];
            size_t b1 = 0;
            size_t b2 = 0;
            double w = data->edges[i].weight;
            while (b1 < p1.bones.size() || b2 < p2.bones.size())
            {
                bool do1 = false, do2 = false;
                if (b1 == p1.bones.size()) do2 = true;
                else if (b2 == p2.bones.size()) do1 = true;
                else if (p1.bones[b1] < p2.bones[b2]) do1 = true;
                else if (p1.bones[b1] > p2.bones[b2]) do2 = true;
                else do1 = true, do2 = true;

                for (int j = 0; j < 4; j++)
                {
                    double weight = v_entry(x, eq) * w;
                    if (do1)
                        lsqr_accumulate(p1, b1, j, weight, out);
                    if (do2)
                        lsqr_accumulate(p2, b2, j, -weight, out);
                    eq++;
                }
                if (do1) b1++;
                if (do2) b2++;
            }
        }
    }

    if (vertex_reg.used)
        for (size_t i = 0; i < data->n_point; i++)
        {
            const lsqr_point_data &p = data->points[i];
            for (size_t j = 0; j < p.bones.size(); j++)
                for (int k = 0; k < 4; k++)
                    lsqr_accumulate(p, j, k, v_entry(x, eq++) * vertex_reg.double_value, out);
        }
    g_assert(eq == data->n_eq);

    return out;
}

static VEC *my_lsqr(lsqr_data *data, VEC *b, double tol, VEC *x, int limit, int *steps)
{
    /* This code adapted from iternsym.c in meschach, but with custom
     * callbacks. */
    ITER *ip;

    ip = iter_get(b->dim, x->dim);
    ip->Ax = (Fun_Ax) lsqr_Ax;
    ip->A_par = (void *) data;
    ip->ATx = (Fun_Ax) lsqr_ATx;
    ip->AT_par = (void *) data;
    ip->Bx = (Fun_Ax) NULL;
    ip->B_par = NULL;

    if (verbose.used || debug.used)
        ip->info = (void (*)()) (Fun_info) show_info;
    else
        ip->info = NULL;
    ip->limit = limit;
    ip->b = b;
    ip->eps = tol;
    ip->x = x;
    ip->steps = -1;
    iter_lsqr(ip);
    x = ip->x;
    if (steps) *steps = ip->steps;
    ip->shared_x = ip->shared_b = TRUE;
    iter_free(ip);   /* release only ITER structure */
    if (verbose.used || debug.used)
        fputs("\n", stderr);
    return x;
}

static int fit_positions_big(Mesh &control)
{
    const Skeleton &control_skeleton = control.skeleton.get_read();
    size_t n_mesh = ref_positions.size();
    size_t n_point = control.positions.all_size();
    size_t n_edge = 0;
    size_t n_edge_eq = 0;
    size_t n_var = 0;
    size_t n_eq = 0;
    VEC *target;
    lsqr_data data;

    size_t pindex = 0;
    for (MeshPositionIterator p = control.positions.all_begin();
         p != control.positions.all_end(); p++, pindex++)
    {
        p->internal_data = GSIZE_TO_POINTER(pindex);
        const VectorPosition &pos = p->position;
        n_var += 4 * distance(pos.begin(), pos.end()) - 1;
    }
    for (MeshHalfEdgeConstIterator e = control.edges.all_begin();
         e != control.edges.all_end(); e++)
    {
        MeshPositionConstIterator p1 = e->next()->vertex()->position();
        MeshPositionConstIterator p2 = e->prev()->vertex()->position();
        if (!mesh_true_face(e->dual()->face())
            || p1->internal_data < p2->internal_data)
        {
            n_edge++;
            VectorPosition delta = p1->position - p2->position;
            n_edge_eq += 4 * distance(delta.begin(), delta.end());
        }
    }

    n_eq = 3 * n_point * n_mesh;
    if (edge_reg.used) n_eq += n_edge_eq;
    if (vertex_reg.used) n_eq += n_var + n_point;

    data.n_eq = n_eq;
    data.n_point = n_point;
    data.n_mesh = n_mesh;
    data.n_edge = n_edge;
    data.n_var = n_var;
    data.points.resize(n_point);
    data.edges.resize(n_edge);
    target = v_get(n_eq);

    size_t eq = 0;
    size_t var = 0;
    size_t edge = 0;
    pindex = 0;
    for (MeshPositionConstIterator p = control.positions.all_begin();
         p != control.positions.all_end(); p++, pindex++)
    {
        lsqr_point_data &pd = data.points[pindex];
        const VectorPosition &pos = p->position;
        pd.var_offset = var;
        size_t last_bone = 0;
        for (VectorPosition::const_iterator i = pos.begin(); i != pos.end(); i++)
        {
            last_bone = i->index / 4;
            pd.bones.push_back(last_bone);
            var += 4;
        }
        var--; /* Eliminated variable */

        for (size_t m = 0; m < n_mesh; m++)
        {
            const Vector4 &pos = ref_positions[m][pindex];
            for (int j = 0; j < 3; j++)
            {
                v_set_val(target, eq, pos[j]);
                /* Account for the eliminated variable */
                v_sub_val(target, eq, ref_skeletons[m][last_bone][j][3]);
                eq++;
            }
        }
    }

    if (edge_reg.used)
    {
        for (MeshHalfEdgeConstIterator e = control.edges.all_begin();
             e != control.edges.all_end(); e++)
        {
            MeshPositionConstIterator p1 = e->next()->vertex()->position();
            MeshPositionConstIterator p2 = e->prev()->vertex()->position();
            if (!mesh_true_face(e->dual()->face())
                || p1->internal_data < p2->internal_data)
            {
                data.edges[edge].p1 = GPOINTER_TO_SIZE(p1->internal_data);
                data.edges[edge].p2 = GPOINTER_TO_SIZE(p2->internal_data);

                double l = 0.0;
                for (size_t i = 0; i < n_mesh; i++)
                {
                    Vector4 delta = ref_positions[i][data.edges[edge].p1]
                        - ref_positions[i][data.edges[edge].p2];
                    l += length(delta);
                }
                g_assert(l > 1e-10);
                data.edges[edge].weight = edge_reg.double_value / l * n_mesh;

                VectorPosition delta = p1->position - p2->position;

                size_t last1 = 0, last2 = 0;
                for (VectorPosition::const_iterator b = p1->position.begin(); b != p1->position.end(); b++)
                    last1 = b->index / 4;
                for (VectorPosition::const_iterator b = p2->position.begin(); b != p2->position.end(); b++)
                    last2 = b->index / 4;
                for (VectorPosition::const_iterator b = delta.begin(); b != delta.end(); b++)
                    for (int j = 0; j < 4; j++)
                    {
                        size_t bone = b->index / 4;
                        v_set_val(target, eq, 0.0);
                        if (j == 3 && bone == last1) v_sub_val(target, eq, data.edges[edge].weight);
                        if (j == 3 && bone == last2) v_add_val(target, eq, data.edges[edge].weight);
                        eq++;
                    }
                edge++;
            }
        }
    }
    if (vertex_reg.used)
    {
        for (MeshPositionConstIterator p = control.positions.all_begin(); p != control.positions.all_end(); p++)
        {
            size_t bones = distance(p->position.begin(), p->position.end());
            for (size_t i = 0; i < 4 * bones - 1; i++)
                v_set_val(target, eq++, 0.0);
            v_set_val(target, eq++, -vertex_reg.double_value);
        }
    }

    g_assert(eq == n_eq);
    VEC *x = v_get(n_var);
    v_zero(x);
    int steps;
    my_lsqr(&data, target, 1e-7, x, iterations.int_value, &steps);

    var = 0;
    for (MeshPositionIterator p = control.positions.all_begin();
         p != control.positions.all_end(); p++)
    {
        VectorPosition &pos = p->position;
        double w = 0.0;
        for (VectorPosition::const_iterator i = pos.begin(); i != pos.end(); i++)
        {
            VectorPosition::const_iterator nxt = i;
            nxt++;

            if (nxt != pos.end())
            {
                for (int j = 0; j < 4; j++)
                    pos[i->index + j] = v_entry(x, var + j);
                w += pos[i->index + 3];
                var += 4;
            }
            else
            {
                for (int j = 0; j < 3; j++)
                    pos[i->index + j] = v_entry(x, var + j);
                pos[i->index + 3] = 1.0 - w;
                var += 3;
            }
        }
        g_log(G_LOG_DOMAIN, G_LOG_LEVEL_DEBUG, "w = % 12.6f", w);
    }

    V_FREE(target);
    V_FREE(x);

    /* Axe any existing normal information, which is most likely bogus */
    for (MeshVertexIterator v = control.vertices.total_begin(); v != control.vertices.total_end(); v++)
    {
        v->data.tangent_plane.tangents[0] = control_skeleton.ssd_vector();
        v->data.tangent_plane.tangents[1] = control_skeleton.ssd_vector();
    }

    return steps;
}

struct queue_entry
{
    double priority;
    MeshPositionIterator position;

    queue_entry() {}
    queue_entry(double priority, MeshPositionIterator position) : priority(priority), position(position) {}
    bool operator <(const queue_entry &b) const
    { return b.priority < priority; }
};

static void blur_influences(Mesh &control, double radius)
{
    const Skeleton &skeleton = control.skeleton.get_read();
    priority_queue<queue_entry> q;
    size_t n_mesh = ref_positions.size();

    /* First force all influence values to 1. Also label positions. */
    size_t pindex = 0;
    for (MeshPositionIterator j = control.positions.all_begin(); j != control.positions.all_end(); j++, pindex++)
    {
        for (VectorPosition::const_iterator k = j->position.begin(); k != j->position.end(); k++)
            if (k->values[3]) j->position[k->index + 3] = 1.0;
        j->internal_data = GSIZE_TO_POINTER(pindex);
    }

    if (radius == 0.0) return;
    for (size_t i = 0; i < skeleton.size(); i++)
    {
        for (MeshPositionIterator j = control.positions.all_begin(); j != control.positions.all_end(); j++)
        {
            /* Indicates not seen. We use first_vertex's internal_data for
             * this because our own is used for enumeration. */
            j->first_vertex()->internal_data = NULL;
            if (const_cast<const VectorPosition &>(j->position)[4 * i + 3])
                q.push(queue_entry(0.0, j));
        }
        while (!q.empty())
        {
            double p = q.top().priority;
            MeshPositionIterator cur = q.top().position;
            q.pop();
            if (cur->first_vertex()->internal_data) continue; /* already seen */
            cur->first_vertex()->internal_data = &q;          /* arbitrary non-NULL */
            cur->position[4 * i + 3] = 1.0;
            MeshHalfEdgeIterator e = cur->first_vertex()->first_edge();
            do
            {
                MeshPositionIterator nxt = e->prev()->vertex()->position();
                e = e->dual()->next();
                if (!mesh_true_position(nxt)) continue;
                double np = 0.0;
                for (size_t i = 0; i < n_mesh; i++)
                {
                    Vector4 pos_cur = ref_positions[i][GPOINTER_TO_SIZE(cur->internal_data)];
                    Vector4 pos_nxt = ref_positions[i][GPOINTER_TO_SIZE(nxt->internal_data)];
                    np += length(pos_cur - pos_nxt) / radius;
                }
                np = p + np / n_mesh;
                if (np < 1.0 && !nxt->first_vertex()->internal_data)
                    q.push(queue_entry(np, nxt));
            } while (e != cur->first_vertex()->first_edge());
        }
    }
}

static void add_mesh(const Mesh &control, Mesh &m)
{
    map<pair<double, double>, MeshPositionIterator> correspond;
    make_correspond(correspond, m);

    size_t index = 0;
    const Skeleton &skeleton = m.skeleton.get_read();
    ref_positions.push_back(vector<Vector4>());
    ref_positions.back().resize(control.positions.all_size());
    for (MeshPositionConstIterator p = control.positions.all_begin(); p != control.positions.all_end(); p++)
    {
        MeshPositionConstIterator p2 = get_correspond(correspond, p);
        Vector3 pos = skeleton.ssd_to_3d(p2->position);
        ref_positions.back()[index++] = vector3_to_vector4(pos, 1.0);
    }
}

/* We just pray that skeletons are defined in the same order... */
static void load_influences(Mesh &control, Mesh &influences)
{
    map<pair<double, double>, MeshPositionIterator> correspond;
    make_correspond(correspond, influences);

    const Skeleton &control_skeleton = control.skeleton.get_read();
    for (MeshPositionIterator p = control.positions.all_begin(); p != control.positions.all_end(); p++)
    {
        MeshPositionConstIterator p2 = get_correspond(correspond, p);
        VectorPosition &pos = p->position;
        const VectorPosition &pos2 = p2->position;
        pos = control_skeleton.ssd_vector();
        for (VectorPosition::const_iterator i = pos2.begin(); i != pos2.end(); i++)
            pos[i->index + 3] = i->values[3];
    }
}

static Matrix4 normalise_meshes(Mesh &control)
{
    pair<Vector3, Vector3> bounds = mesh_bounding_box_3d(control);

    Vector3 range = bounds.second - bounds.first;
    Vector3 mid = (bounds.first + bounds.second) * 0.5;
    double s = 1.0 / max(range[0], max(range[1], range[2]));
    Matrix4 transform = matrix_scale(s) * matrix_translate(-mid);

    mesh_transform(control, transform);
    for (size_t i = 0; i < ref_positions.size(); i++)
        for (size_t j = 0; j < ref_positions[i].size(); j++)
            ref_positions[i][j] = transform * ref_positions[i][j];

    return transform;
}

static double rms(Mesh &control)
{
    Skeleton skeleton = control.skeleton.get_write();
    double sum = 0.0;
    int cnt = 0;

    for (size_t i = 0; i < ref_positions.size(); i++)
    {
        skeleton.set_to_frame(i);
        size_t j = 0;
        for (MeshPositionConstIterator p = control.positions.all_begin();
             p != control.positions.all_end(); p++, j++)
        {
            Vector3 pos_fit = skeleton.ssd_to_3d(p->position);
            Vector3 ref_fit = vector4_to_vector3(ref_positions[i][j]);
            sum += (pos_fit - ref_fit) * (pos_fit - ref_fit);
            cnt++;
        }
    }
    return sqrt(sum / cnt);
}

int main(int argc, char **argv)
{
    int args = process_options(argc, argv);
    Mesh control;
    vector<string> comments;
    Matrix4 transform;

    for (int i = args; i < argc - 1; i++)
    {
        try
        {
            if (i == args)
            {
                format_load_mesh_file_comments(in_format.string_value, argv[i], control,
                                               MESH_LEVEL_PATCHES, comments);
            }
            else
            {
                Mesh fit;
                format_load_mesh_file(in_format.string_value, argv[i], fit,
                                      MESH_LEVEL_PATCHES);
                add_mesh(control, fit);
            }
        }
        catch (bad_format &e)
        {
            cerr << "Failed to load " << argv[i] << ": " << e.what() << endl;
            return 1;
        }
    }
    comments.push_back(alod_generation_string(argc, argv));

    Timer timer;

    transform = normalise_meshes(control);
    fit_skeletons(control);
    if (!skeleton_only.used)
    {
        if (influence_file.used)
        {
            Mesh influences;
            format_load_mesh_file(in_format.string_value, influence_file.string_value, influences,
                                  MESH_LEVEL_PATCHES);
            load_influences(control, influences);
        }
        if (blur.used) blur_influences(control, blur.double_value);
        else blur_influences(control, 0.0); // sets influences to 1.0
        int steps = fit_positions_big(control);

        ostringstream step_comment;
        step_comment << "LSQR steps: " << steps;
        comments.push_back(step_comment.str());
        control.skeleton.get_write().set_frame_rate(25.0);
    }
    ostringstream rms_comment, time_comment;;
    rms_comment << "RMS error: " << rms(control);
    time_comment << "Fitting time: " << timer.elapsed() << " seconds";
    comments.push_back(rms_comment.str());
    comments.push_back(time_comment.str());

    mesh_transform(control, inverse(transform));

    try
    {
        format_save_mesh_file(out_format.string_value, argv[argc - 1], control,
                              comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to save mesh: " << e.what() << endl;
        return 1;
    }

    return 0;
}
