#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "geom/coords.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include "common/maputils.h"
#include <iostream>
#include <fstream>
#include <iomanip>
#include <cstdlib>
#include <cstdio>
#include <glib.h>
#include <vector>
#include <string>
#include <sstream>
#include <set>
#include <map>

using namespace std;

static AlodProgramOption in_format;
static AlodProgramOption verbose, debug, param, validate, compact;
static AlodProgramOption bounds, expect, show_c, flags, textures, uv2eps;
static AlodProgramOption query;
static set<string> queries;

template<class T> void report_stat(const char *name, T value)
{
    if (query.used)
    {
        if (queries.count(name))
            cout << value << "\n";
    }
    else
        cout << name << ": " << value << "\n";
}

static int process_options(int argc, char **argv)
{
    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo('p', "param", "show parametrisation stats", 0, PROGRAM_ARGUMENT_STRING, &param),
        AlodProgramOptionInfo('P', "validate", "validate parametrisation (implies -p)", 0, PROGRAM_ARGUMENT_STRING, &validate),
        AlodProgramOptionInfo('c', "compact", "print segmentation metric with this compactness weight", 1, PROGRAM_ARGUMENT_DOUBLE, &compact),
        AlodProgramOptionInfo('b', "bounds", "show bone bounds", 0, PROGRAM_ARGUMENT_STRING, &bounds),
        AlodProgramOptionInfo('e', "expect", "show P", 0, PROGRAM_ARGUMENT_STRING, &expect),
        AlodProgramOptionInfo('C', "show-c", "show the transformation to normalised animation space", 0, PROGRAM_ARGUMENT_STRING, &show_c),
        AlodProgramOptionInfo('F', "flags", "show flags", 0, PROGRAM_ARGUMENT_STRING, &flags),
        AlodProgramOptionInfo('t', "textures", "extract textures to file", 0, PROGRAM_ARGUMENT_STRING, &textures),
        AlodProgramOptionInfo('\0',"uv2eps", "extract parametrisation to EPS file", 1, PROGRAM_ARGUMENT_STRING, &uv2eps),
        AlodProgramOptionInfo('q', "query", "show only these values (comma-separated)", 1, PROGRAM_ARGUMENT_STRING, &query),
        AlodProgramOptionInfo()
    };

    int args;
    args = alod_process_options(argc, argv, opts);
    if (!in_format.used) in_format.string_value = "auto";
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    if (validate.used) param.used = true;
    if (query.used)
    {
        string s = query.string_value;
        string::size_type pos = 0, nxt;
        for (;;)
        {
            nxt = s.find(',', pos);
            queries.insert(s.substr(pos, nxt - pos));
            if (nxt == string::npos) break;
            pos = nxt + 1;
        }
    }
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);


    if (argc - args != 1)
    {
        cerr << "Error: must specify input file\n";
        alod_show_help(opts);
        exit(1);
    }
    return args;
}

static void dump_counts(const Mesh &m)
{
    int F, TF, P, TP, V, TV;

    TF = m.faces.real_size();
    F = m.faces.whole_size();
    TP = m.positions.real_size();
    P = m.positions.whole_size();
    TV = m.vertices.real_size();
    V = m.vertices.whole_size();

    report_stat("positions", TP);
    report_stat("vertices", TV);
    report_stat("faces", TF);
    report_stat("holes", P - TP);
    report_stat("textures", m.textures.size());
    report_stat("materials", m.materials.size());
    report_stat("collapses", mesh_collapses(m));
    report_stat("bones", m.skeleton.get_read().size());
    report_stat("frames", m.skeleton.get_read().num_frames());
}

static void dump_volume(const Mesh &m)
{
    double vol = 0.0;
    const Skeleton &skeleton = m.skeleton.get_read();
    for (MeshFaceConstIterator f = m.faces.whole_begin(); f != m.faces.whole_end(); f++)
    {
        Vector3 p[3];
        MeshHalfEdgeConstIterator e = f->edges();

        for (int i = 0; i < 3; i++, e = e->next())
            p[i] = skeleton.ssd_to_3d(e->vertex()->position()->position);

        vol += cross(p[0], p[1]) * p[2] / 6.0;
    }
    report_stat("volume", fabs(vol));
    report_stat("winding", (vol >= 0.0) ? "ccw" : "cw");

    pair<Vector3, Vector3> bb = mesh_bounding_box_3d(m);
    ostringstream value;
    value << bb.first << " " << bb.second;
    report_stat("bbox", value.str());
}

static void dump_density(const Mesh &m)
{
    size_t total = 0;
    size_t mx = 0;
    size_t count = 0;
    size_t pmax = 0, mmax = 0;
    fast_map<int, fast_set<int>::type>::type patch_counts;
    id_map<MeshMaterialConstIterator, fast_set<int>::type>::type material_counts;
    for (MeshPositionConstIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
    {
        size_t cur = distance(p->position.begin(), p->position.end());
        mx = max(cur, mx);
        total += cur;
        count++;
    }

    for (MeshVertexConstIterator v = m.vertices.all_begin(); v != m.vertices.all_end(); v++)
    {
        for (int i = 0; i < 2; i++)
        {
            const VectorPosition &p = v->data.tangent_plane.tangents[i];
            for (VectorPosition::const_iterator j = p.begin(); j != p.end(); j++)
            {
                patch_counts[v->first_edge()->face()->patch].insert(j->index / 4);
                material_counts[v->first_edge()->face()->material()].insert(j->index / 4);
            }
        }
        pmax = max(pmax, patch_counts[v->first_edge()->face()->patch].size());
        mmax = max(mmax, material_counts[v->first_edge()->face()->material()].size());
    }

    report_stat("max bones", mx);
    report_stat("avg bones", total / (double) count);
    report_stat("influences", total);
    if (param.used)
    {
        report_stat("max bones in patch", pmax);
        report_stat("max bones in material", mmax);
    }
}

static void dump_bounds(const Mesh &m)
{
    const Skeleton &skeleton = m.skeleton.get_read();
    for (size_t i = 0; i < skeleton.size(); i++)
    {
        const SkeletonBone &b = skeleton.get_bone(i);
        cout << "bone " << i << ": B(" << b.max_singular << ", "
            << b.translate_bounds.centre << ", " << b.translate_bounds.radius << ")\n";
    }
}

static void dump_expect(const Mesh &m)
{
    const Skeleton &skeleton = m.skeleton.get_read();
    const MAT *P = skeleton.get_P();

    for (size_t i = 0; i < P->m; i++)
    {
        for (size_t j = 0; j < P->n; j++)
            printf("%8.4f ", m_entry(P, i, j));
        printf("\n");
    }
}

static void dump_c(const Mesh &m)
{
    const Skeleton &skeleton = m.skeleton.get_read();
    const MAT *C = skeleton.get_C();

    for (size_t i = 0; i < C->m; i++)
    {
        for (size_t j = 0; j < C->n; j++)
            printf("%8.4f ", m_entry(C, i, j));
        printf("\n");
    }
}

static void dump_flags(const Mesh &m)
{
    map<int, string> names;
    names[BONE_FLAG_FIXED_TRANSLATE] = "FIXED_TRANSLATE";
    names[BONE_FLAG_FIXED_ROTATE] = "FIXED_ROTATE";
    names[BONE_FLAG_FIXED_SCALE] = "FIXED_SCALE";
    names[BONE_FLAG_UNIFORM_SCALE] = "UNIFORM_SCALE";

    names[BONE_FLAG_EXPECT_L_LINEAR] = "EXPECT_L_LINEAR";
    names[BONE_FLAG_EXPECT_L_TRANSLATE] = "EXPECT_L_TRANSLATE";
    names[BONE_FLAG_EXPECT_GTG_LINEAR] = "EXPECT_GTG_LINEAR";
    names[BONE_FLAG_EXPECT_GTG_TRANSLATE] = "EXPECT_GTG_TRANSLATE";
    names[BONE_FLAG_EXPECT_GTG_WEIGHT] = "EXPECT_GTG_WEIGHT";
    names[BONE_FLAG_EXPECT_LTGTGL] = "EXPECT_L_LTGTGL";

    const Skeleton &skeleton = m.skeleton.get_read();
    for (size_t i = 0; i < skeleton.size(); i++)
    {
        const SkeletonBone &b = skeleton.get_bone(i);
        cout << "bone " << i << ":";
        for (map<int, string>::const_iterator j = names.begin(); j != names.end(); j++)
            if (b.flags & j->first)
                cout << " " << j->second;
        cout << "\n";
    }
}

static void dump_textures(const Mesh &m)
{
    int i = 1;
    for (MeshTextureConstIterator t = m.textures.begin(); t != m.textures.end(); t++, i++)
    {
        ostringstream name;
        name << "texture" << i << ".png";
        mesh_save_texture(name.str(), "", *t);
    }
}

static void dump_uv2eps(const Mesh &m)
{
    double size = 128.0;
    try
    {
        set<int> seen; // seen patches

        ofstream out(uv2eps.string_value.c_str());
        out << "%!PS-Adobe-2.0 EPSF-1.2\n"
            << "%%BoundingBox: 0.0 0.0 " << size << " " << size << "\n"
            << "%%EndComments\n"
            << "gsave\n"
            << "0.25 setlinewidth\n"
            << "/drawpatch { gsave 0.7 setgray fill grestore stroke } bind def\n"
            << "/drawedge { gsave 0.125 setlinewidth newpath moveto lineto closepath stroke grestore } bind def\n";

        // Boundaries
        for (MeshHalfEdgeConstIterator e = m.edges.real_begin(); e != m.edges.real_end(); e++)
        {
            if (!seen.count(e->face()->patch)
                && e->dual()->face()->patch != e->face()->patch)
            {
                seen.insert(e->face()->patch);
                Vector2 uv = e->next()->vertex()->data.texcoords;
                out << "newpath " << uv[0] * size << " " << uv[1] * size << " moveto";
                MeshHalfEdgeConstIterator e2 = e->prev()->vertex()->first_edge();
                while (e2 != e)
                {
                    uv = e2->next()->vertex()->data.texcoords;
                    out << " " << uv[0] * size << " " << uv[1] * size << " lineto";
                    e2 = e2->prev()->vertex()->first_edge();
                }
                out << " closepath drawpatch\n";
            }
        }

        // Interiors
        for (MeshHalfEdgeConstIterator e = m.edges.real_begin(); e != m.edges.real_end(); e++)
        {
            if (e->dual()->face()->patch == e->face()->patch)
            {
                Vector2 uv1 = e->next()->vertex()->data.texcoords;
                Vector2 uv2 = e->prev()->vertex()->data.texcoords;
                out << uv1[0] * size << " " << uv1[1] * size << " "
                    << uv2[0] * size << " " << uv2[1] * size
                    << " drawedge\n";
            }
        }

        out << "newpath 0 0 moveto "
            << "0 " << size << " lineto "
            << size << " " << size << " lineto "
            << size << " 0 lineto "
            << "closepath stroke\n";
        out << "grestore\n";
    }
    catch (ios::failure &e)
    {
        cerr << "Failed to write " << uv2eps.string_value << ": " << e.what();
    }
}

static void dump_param_stats(const Mesh &m)
{
    double l2, linf, pack, stretch;
    int patches;
    fast_set<int>::type seen;
    size_t bound = 0, corner = 0;

    for (MeshPositionConstIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
    {
        if (mesh_position_patch_valence(p) >= 2)
        {
            bound++;
            if (p->patch_corner) corner++;
        }
    }
    report_stat("boundary vertices", bound);
    report_stat("boundary corners", corner);

    patches = 0;
    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
        if (!seen.count(f->patch))
        {
            patches++;
            seen.insert(f->patch);
        }

    report_stat("charts", patches);

    l2 = mesh_texture_stretch(m, false);
    linf = mesh_texture_stretch(m, true);
    pack = 0.0;
    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
        pack += mesh_texture_double_area(f);
    pack *= 0.5;
    stretch = 1.0 / (l2 * l2);

    report_stat("L2 stretch", l2);
    report_stat("Linf stretch", linf);
    report_stat("packing efficiency", pack);
    report_stat("stretch efficiency", stretch);
    report_stat("texture efficiency", pack * stretch);
}

int main(int argc, char **argv)
{
    Mesh m;
    int args;
    vector<string> comments;

    args = process_options(argc, argv);
    try
    {
        bool patches = param.used || uv2eps.used || compact.used;
        format_load_mesh_file_comments(in_format.string_value, argv[args], m,
                                       patches ? MESH_LEVEL_PATCHES : MESH_LEVEL_MANIFOLD,
                                       comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to load mesh: " << e.what() << endl;
        return 1;
    }

    for (size_t i = 0; i < comments.size(); i++)
        report_stat("Comment", comments[i]);
    dump_counts(m);
    dump_volume(m);
    dump_density(m);
    if (param.used)
    {
        dump_param_stats(m);
        if (validate.used && !mesh_valid_parametrisation(m))
        {
            cerr << "invalid parametrisation" << endl;
            return 1;
        }
    }
    if (compact.used)
        report_stat("segmentation metric",
                    mesh_segmentation_metric(m, compact.double_value));
    if (bounds.used) dump_bounds(m);
    if (expect.used) dump_expect(m);
    if (show_c.used) dump_c(m);
    if (flags.used) dump_flags(m);
    if (textures.used) dump_textures(m);
    if (uv2eps.used) dump_uv2eps(m);

    return 0;
}
