#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include <iostream>
#include <sstream>
#include <vector>
#include <string>
#include <cstdlib>
#include <cstdio>
#include <glib.h>

using namespace std;

static AlodProgramOption in_format, out_format;
static AlodProgramOption patches, compact_weight, limit, straighten_tolerance;
static AlodProgramOption constrain_tolerance, no_split;
static string out_file;

static int process_options(int argc, char **argv)
{
    AlodProgramOption verbose, debug, threed;
    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('3', "3d", "ignore skeleton in computing metrics", 0, PROGRAM_ARGUMENT_STRING, &threed),
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('o', "output", "output format", 1, PROGRAM_ARGUMENT_STRING, &out_format),
        AlodProgramOptionInfo('p', "patches", "minimum number of output patches [1]", 1, PROGRAM_ARGUMENT_INT, &patches),
        AlodProgramOptionInfo('l', "limit", "limit on segmentation metric [none]", 1, PROGRAM_ARGUMENT_DOUBLE, &limit),
        AlodProgramOptionInfo('c', "compact", "weight to give to compacting term [0.2]", 1, PROGRAM_ARGUMENT_DOUBLE, &compact_weight),
        AlodProgramOptionInfo('s', "straighten", "tolerance for UV edge straightening [4.0]", 1, PROGRAM_ARGUMENT_DOUBLE, &straighten_tolerance),
        AlodProgramOptionInfo('C', "constrain", "tolerance for mesh edge straightening [1.2]", 1, PROGRAM_ARGUMENT_DOUBLE, &constrain_tolerance),
        AlodProgramOptionInfo('n', "no-split", "do not split patches with boundary self-intersections", 0, PROGRAM_ARGUMENT_STRING, &no_split),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo()
    };

    int args;
    args = alod_process_options(argc, argv, opts);
    if (!in_format.used) in_format.string_value = "auto";
    if (!out_format.used) out_format.string_value = "auto";
    if (!patches.used) patches.int_value = 1;
    if (!limit.used) limit.double_value = HUGE_VAL;
    if (!compact_weight.used) compact_weight.double_value = 0.2;
    if (!straighten_tolerance.used) straighten_tolerance.double_value = 4.0;
    if (!constrain_tolerance.used) constrain_tolerance.double_value = 1.2;
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    if (threed.used) Skeleton::set_metric_mode(Skeleton::MODE_3D);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);

    if (argc - args < 1 || argc - args > 2)
    {
        cerr << "Error: must specify input (and optionally output) file\n";
        alod_show_help(opts);
        exit(1);
    }
    if (argc - args == 1)
        out_file = argv[args];
    else
        out_file = argv[args + 1];
    return args;
}

static void add_stat(vector<string> &comments, const char *name, double value)
{
    ostringstream out;
    out << name << ": " << value;
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "%s", out.str().c_str());
    comments.push_back(out.str());
}

int main(int argc, char **argv)
{
    Mesh m;
    int args;
    vector<string> comments;

    args = process_options(argc, argv);
    try
    {
        format_load_mesh_file_comments(in_format.string_value, argv[args], m,
                                       MESH_LEVEL_MANIFOLD,
                                       comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to load mesh: " << e.what() << endl;
        return 1;
    }
    comments.push_back(alod_generation_string(argc, argv));

    ParametrisationStatistics stats;
    stats = mesh_parametrise(m,
                             patches.int_value,
                             limit.double_value,
                             compact_weight.double_value,
                             straighten_tolerance.double_value,
                             constrain_tolerance.double_value,
                             no_split.used);
    add_stat(comments, "Intra-rectangle efficiency", stats.texture_area / stats.rectangle_area);
    add_stat(comments, "Gutter efficiency", stats.rectangle_area / stats.gutter_area);
    add_stat(comments, "Inter-rectangle efficiency", stats.gutter_area);
    add_stat(comments, "Packing efficiency", stats.texture_area);
    add_stat(comments, "Sander packing efficiency", stats.texture_area / stats.rectangle_area * stats.gutter_area);

    try
    {
        format_save_mesh_file(out_format.string_value, out_file, m,
                              comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to save mesh: " << e.what() << endl;
        return 1;
    }
    return 0;
}
