#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include "common/paths.h"
#include "glee/GLee.h"
#include "alodgl/glmisc.h"
#include "geom/geom.h"
#include <algorithm>
#include <cstring>
#include <map>
#include <queue>
#include <vector>
#include <GL/gl.h>
#include <GL/glext.h>
#include <gtk/gtk.h>
#include <gdk/gdkgl.h>

using namespace std;

static AlodProgramOption oversample, resolution, in_format, out_format;
static AlodProgramOption do_textures, do_tangents;
static AlodProgramOption verbose, debug, blank;
static AlodProgramOption oldmeshfile;
static string out_file;

static int process_options(int argc, char **argv)
{
    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('o', "output", "output format", 1, PROGRAM_ARGUMENT_STRING, &out_format),
        AlodProgramOptionInfo('s', "oversample", "oversampling ratio", 1, PROGRAM_ARGUMENT_INT, &oversample),
        AlodProgramOptionInfo('r', "resolution", "size of image", 1, PROGRAM_ARGUMENT_INT, &resolution),
        AlodProgramOptionInfo('n', "tangents", "sample tangent maps", 0, PROGRAM_ARGUMENT_STRING, &do_tangents),
        AlodProgramOptionInfo('t', "texture", "sample texture map", 0, PROGRAM_ARGUMENT_STRING, &do_textures),
        AlodProgramOptionInfo('M', "old-mesh", "mesh file for old texture", 1, PROGRAM_ARGUMENT_STRING, &oldmeshfile),
        AlodProgramOptionInfo('b', "blank", "leave holes blank", 0, PROGRAM_ARGUMENT_STRING, &blank),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo()
    };

    int args;
    args = alod_process_options(argc, argv, opts);

    if (!in_format.used) in_format.string_value = "auto";
    if (!out_format.used) out_format.string_value = "auto";
    if (oversample.used)
    {
        if (oversample.int_value <= 0
            || (oversample.int_value & (oversample.int_value - 1)))
        {
            cerr << "oversample factor must be a power of 2\n";
            exit(1);
        }
    }
    else
        oversample.int_value = 4;
    if (resolution.used)
    {
        if (resolution.int_value <= 0
            || (resolution.int_value & (resolution.int_value - 1)))
        {
            cerr << "resolution must be a power of 2\n";
            exit(1);
        }
    }
    else
        resolution.int_value = -1; // auto
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);

    if (argc - args < 1 || argc - args > 2)
    {
        cerr << "Error: must specify input (and optionally output) file\n";
        alod_show_help(opts);
        exit(1);
    }
    if (argc - args == 1)
        out_file = argv[args];
    else
        out_file = argv[args + 1];
    return args;
}

static double get_resolution_scale(const Mesh &m)
{
    double stretch = mesh_texture_stretch(m, false);
    double pack = 0.0;
    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
        pack += fabs(mesh_texture_double_area(f));
    pack *= 0.5;

    return stretch / sqrt(pack);
}

static int get_resolution(const Mesh &m)
{
    double scale = get_resolution_scale(m);
    double sqrt_samples = sqrt((double) m.positions.real_size()) * scale;

    int r = 4;
    while (r < sqrt_samples) r *= 2;
    return r;
}

static void initialise_gl()
{
    const int attribs[] = {
        GDK_GL_USE_GL, TRUE,
        GDK_GL_RGBA, TRUE,
        GDK_GL_RED_SIZE, 8,
        GDK_GL_GREEN_SIZE, 8,
        GDK_GL_BLUE_SIZE, 8,
        GDK_GL_ALPHA_SIZE, 8,
        GDK_GL_STENCIL_SIZE, 1,
        GDK_GL_ATTRIB_LIST_NONE
    };

    GdkGLConfig *config;
    GdkGLContext *context;
    GdkWindowAttr window_attr = {NULL};
    GdkWindow *window;
    GdkGLWindow *glwindow;
    GLuint fbo;

    config = gdk_gl_config_new(attribs);

    window_attr.width = 4;
    window_attr.height = 4;
    window_attr.wclass = GDK_INPUT_OUTPUT;
    window_attr.visual = gdk_gl_config_get_visual(config);
    window_attr.colormap = gdk_gl_config_get_colormap(config);
    window_attr.window_type = GDK_WINDOW_TOPLEVEL;
    window = gdk_window_new(NULL, &window_attr, GDK_WA_COLORMAP | GDK_WA_VISUAL);
    glwindow = gdk_gl_window_new(config, window, NULL);

    context = gdk_gl_context_new(GDK_GL_DRAWABLE(glwindow), NULL, true, GDK_GL_RGBA_TYPE);
    gdk_gl_drawable_make_current(GDK_GL_DRAWABLE(glwindow), context);

    GLeeInit();
    glGenFramebuffersEXT(1, &fbo);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    if (GLEE_EXT_framebuffer_blit)
        glBindFramebufferEXT(GL_READ_FRAMEBUFFER_EXT, fbo);
}

static GLuint create_renderbuffer(int width, int height)
{
    GLuint rbo;
    
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Creating %d by %d renderbuffer",
          width, height);
    glGenRenderbuffersEXT(1, &rbo);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, rbo);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_RGBA8, width, height);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
            GL_RENDERBUFFER_EXT, rbo);
    return rbo;
}

static void destroy_renderbuffer(GLuint rbo)
{
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
            GL_RENDERBUFFER_EXT, 0);
    glDeleteRenderbuffersEXT(1, &rbo);
}

#define DATA(x, y, c) (data[((y) * width + (x)) * channels + (c)])
#define STENCIL(x, y) (stencil[(y) * width + (x)])

/* Do a multi-source breadth first search to fill in the holes. */
static void fill_blanks(int width, int height, int channels,
                        GLubyte *data, GLubyte *stencil)
{
    queue<pair<int, int> > q;
    for (int j = 0; j < height; j++)
        for (int i = 0; i < width; i++)
            if (STENCIL(i, j) != 0)
                q.push(make_pair(i, j));
    while (!q.empty())
    {
        pair<int, int> cur = q.front();
        q.pop();
        for (int dy = -1; dy <= 1; dy++)
            for (int dx = -1; dx <= 1; dx++)
            {
                int nx = cur.first + dx;
                int ny = cur.second + dy;
                if (nx >= 0 && nx < width && ny >= 0 && ny < height
                    && STENCIL(nx, ny) == 0)
                {
                    for (int k = 0; k < 4; k++)
                        DATA(nx, ny, k) = DATA(cur.first, cur.second, k);
                    STENCIL(nx, ny) = 1;
                    q.push(make_pair(nx, ny));
                }
            }
    }
}

static void buffer_to_texture(int width, int height, int oversample,
                              MeshTextureIterator t, bool keep_alpha)
{
    GLuint texid;
    GLubyte *data, *stencil;
    int xbits;
    int xs = width * oversample;
    int ys = height * oversample;

    /* We allocate this dynamically because the stack may not be
     * large enough.
     */
    data = new GLubyte[xs * ys * 4];
    stencil = new GLubyte[xs * ys];
    xbits = ffs(xs) - 1;

    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glReadPixels(0, 0, xs, ys, GL_RGBA, GL_UNSIGNED_BYTE, data);
    glReadPixels(0, 0, xs, ys, GL_STENCIL_INDEX, GL_UNSIGNED_BYTE, stencil);

    if (!blank.used) fill_blanks(xs, ys, 4, data, stencil);

    /* Downsample. We do it even if there is no oversampling, since it
     * also gives a quick-and-dirty way to strip the alpha.
     */
    glHint(GL_GENERATE_MIPMAP_HINT, GL_NICEST);
    glGenTextures(1, &texid);
    glBindTexture(GL_TEXTURE_2D, texid);
    glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, xs, ys, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, data);
    delete[] stencil;
    delete[] data;

    int lvl = ffs(oversample) - 1;
    t->set_size(width, height, keep_alpha ? 4 : 3);
    glGetTexImage(GL_TEXTURE_2D, lvl, keep_alpha ? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE, t->data());
    glDeleteTextures(1, &texid);
    glBindTexture(GL_TEXTURE_2D, 0);
}

static inline MeshTangentMapEntry &tm_pixel(MeshTangentMapIterator tm,
                                            int x, int y)
{
    return tm->data()[y * tm->width + x];
}

static void fill_blanks_tangents(MeshTangentMapIterator tm)
{
    const int width = tm->width;
    const int height = tm->height;

    queue<pair<int, int> > q;
    for (int j = 0; j < height; j++)
        for (int i = 0; i < width; i++)
            if (tm_pixel(tm, i, j).patch != -1)
                q.push(make_pair(i, j));
    while (!q.empty())
    {
        pair<int, int> cur = q.front();
        q.pop();
        for (int dy = -1; dy <= 1; dy++)
            for (int dx = -1; dx <= 1; dx++)
            {
                int nx = cur.first + dx;
                int ny = cur.second + dy;
                if (nx >= 0 && nx < width && ny >= 0 && ny < height
                    && tm_pixel(tm, nx, ny).patch == -1)
                {
                    tm_pixel(tm, nx, ny) = tm_pixel(tm, cur.first, cur.second);
                    q.push(make_pair(nx, ny));
                }
            }
    }
}

static void sample_tangents(Mesh &m)
{
    int size = resolution.int_value;

    const Skeleton &skeleton = m.skeleton.get_read();
    const size_t fnum = m.faces.real_size();

    MeshTangentMapIterator tm;

    guint32 (*ids)[3] = new guint32[fnum][3];
    guint32 *out_ids = new guint32[size * size];
    GLfloat (*coords)[3][2] = new GLfloat[fnum][3][2];
    MeshFaceConstIterator faces[fnum];

    size_t fi = 0;
    for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++, fi++)
    {
        MeshHalfEdgeConstIterator e = f->edges();
        for (int j = 0; j < 3; j++, e = e->next())
        {
            ids[fi][j] = fi;
            coords[fi][j][0] = e->vertex()->data.texcoords[0];
            coords[fi][j][1] = e->vertex()->data.texcoords[1];
        }
        faces[fi] = f;
    }

    /* Identify texel/triangle relationships */
    GLuint rbo = create_renderbuffer(size, size);
    glClearColor(1.0, 1.0, 1.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT);
    glViewport(-size, -size, 2 * size, 2 * size);
    glVertexPointer(2, GL_FLOAT, 0, coords);
    glColorPointer(4, GL_UNSIGNED_BYTE, 0, ids);
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
    glDrawArrays(GL_TRIANGLES, 0, 3 * fnum);

    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glReadPixels(0, 0, size, size, GL_RGBA, GL_UNSIGNED_BYTE, out_ids);
    destroy_renderbuffer(rbo);

    /* Create empty map */
    tm = m.tangent_maps.insert(m.tangent_maps.end(), MeshTangentMap());
    tm->set_size(size, size);

    for (int y = 0; y < size; y++)
        for (int x = 0; x < size; x++)
        {
            if (out_ids[y * size + x] == (guint32) -1) // blank space
            {
                tm_pixel(tm, x, y).patch = -1;
                tm_pixel(tm, x, y).tangent_plane.tangents[0] = VectorPosition(skeleton);
                tm_pixel(tm, x, y).tangent_plane.tangents[1] = VectorPosition(skeleton);
                continue;
            }

            MeshTangentMapEntry &cur = tm_pixel(tm, x, y);
            double bary[3];
            MeshFaceConstIterator f = faces[out_ids[y * size + x]];
            MeshMaterialConstIterator mat = f->material();
            Vector2 texcoords[3];
            VectorPosition positions[3];
            Vector2 mycoords((x + 0.5) / size, (y + 0.5) / size);

            vector<TangentPlane> planes;

            MeshHalfEdgeConstIterator e = f->edges();
            for (int i = 0; i < 3; i++, e = e->next())
            {
                positions[i] = e->vertex()->position()->position;
                texcoords[i] = e->vertex()->data.texcoords;
                planes.push_back(e->vertex()->data.tangent_plane);
            }
            for (int i = 0; i < 3; i++)
            {
                bary[i] = cross(texcoords[(i + 1) % 3] - mycoords,
                                texcoords[(i + 2) % 3] - mycoords);
                planes[i].tangents[1] *= bary[i];
            }

            cur.patch = f->patch;
            cur.tangent_plane = fit_tangent_plane(skeleton, planes);

            /* Reorient the tangent plane so that the triangle vertices
             * projected onto the tangent plane roughly line up with the
             * parametric space
             */
            VectorPosition mypos = (bary[0] * positions[0] + bary[1] * positions[1] + bary[2] * positions[2]) / (bary[0] + bary[1] + bary[2]);
            Vector2 P(0.0, 0.0);
            for (int i = 0; i < 3; i++)
            {
                VectorPosition delta = positions[i] - mypos;
                Vector2 param = texcoords[i] - mycoords;
                Vector2 proj(delta * cur.tangent_plane.tangents[0],
                             delta * cur.tangent_plane.tangents[1]);
                if (!!proj && !!param)
                {
                    proj = normalise(proj);
                    param = normalise(param);
                    Vector2 rot(proj[0] * param[0] + proj[1] * param[1],
                                proj[1] * param[0] - proj[0] * param[1]);
                    P += rot;
                }
            }
            if (!!P)
            {
                TangentPlane &tp = cur.tangent_plane;
                P = normalise(P);
                const VectorPosition tmp = tp.tangents[0] * P[0] + tp.tangents[1] * P[1];
                tp.tangents[1] = tp.tangents[0] * -P[1] + tp.tangents[1] * P[0];
                tp.tangents[0] = tmp;
            }
        }

    if (!blank.used)
        fill_blanks_tangents(tm);

    /* Links tangent maps to materials */
    for (MeshMaterialIterator mat = m.materials.begin(); mat != m.materials.end(); mat++)
        mat->tangent_map() = tm;

    delete[] ids;
    delete[] out_ids;
    delete[] coords;
}

/* We first need to establish the correspondence between the position and
 * also the faces. We have to assume that the coordinates are identical,
 * otherwise it becomes a complex problem.
 */
static void sample_texture(Mesh &m, const Mesh &oldm)
{
    if (m.materials.size() != 1)
    {
        cerr << "Output mesh must have exactly one material";
        m.materials.begin()->texture() = mesh_null_texture;
    }

    /* position_table maps locations in space to old positions.
     * edge_table maps pairs of old positions to old positions.
     * position_map and edge_map map new to old.
     */
    map<VectorPosition, MeshPositionConstIterator, mesh_compare_vector_position_float> position_table;
    id_map<pair<MeshPositionConstIterator, MeshPositionConstIterator>, MeshHalfEdgeConstIterator>::type edge_table;
    id_map<MeshPositionConstIterator, MeshPositionConstIterator>::type position_map;
    id_map<MeshHalfEdgeConstIterator, MeshHalfEdgeConstIterator>::type edge_map;

    for (MeshPositionConstIterator p = oldm.positions.real_begin(); p != oldm.positions.real_end(); p++)
        position_table[p->position] = p;
    for (MeshPositionConstIterator p = m.positions.real_begin(); p != m.positions.real_end(); p++)
    {
        if (!position_table.count(p->position))
        {
            cerr << "Meshes do not exactly correspond\n";
            exit(1);
        }
        position_map[p] = position_table[p->position];
    }
    position_table.clear();

    for (MeshHalfEdgeConstIterator e = oldm.edges.real_begin(); e != oldm.edges.real_end(); e++)
        edge_table[make_pair(e->next()->vertex()->position(),
                             e->prev()->vertex()->position())] = e;
    for (MeshHalfEdgeConstIterator e = m.edges.real_begin(); e != m.edges.real_end(); e++)
    {
        pair<MeshPositionConstIterator, MeshPositionConstIterator> ei;
        ei = make_pair(position_map[e->next()->vertex()->position()],
                       position_map[e->prev()->vertex()->position()]);
        if (!edge_table.count(ei))
        {
            cerr << "Meshes have different connectivity\n";
            exit(1);
        }
        edge_map[e] = edge_table[ei];
    }
    edge_table.clear();

    int res = resolution.int_value;
    int size = res * oversample.int_value;
    GLuint rbo = create_renderbuffer(size, size);

    /* We create a separate vertex for every edge, to avoid dealing with
     * extra splitting due to having two sets of patches
     */
    size_t num_faces = m.faces.real_size();
    GLfloat positions[num_faces * 6];
    GLfloat texcoords[num_faces * 6];

    glHint(GL_GENERATE_MIPMAP_HINT, GL_NICEST);

    typedef id_map<MeshTextureConstIterator, GLuint>::type texture_ids_type;
    texture_ids_type texture_ids;
    for (MeshMaterialConstIterator s = oldm.materials.begin(); s != oldm.materials.end(); s++)
    {
        if (s->texture() == mesh_null_texture) continue;
        if (texture_ids.count(s->texture())) continue;
        GError *err = NULL;
        texture_ids[s->texture()] = alodgl_load_texture(*s->texture(), GL_RGB8, true, &err);
        if (!texture_ids[s->texture()])
        {
            cerr << "Failed to load texture: " << err->message << "\n";
            g_error_free(err);
            exit(1);
        }
    }

    glClearColor(0.0, 0.0, 1.0, 0.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glEnable(GL_VERTEX_ARRAY);
    glEnable(GL_TEXTURE_COORD_ARRAY);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_STENCIL_TEST);
    glStencilOp(GL_INCR, GL_INCR, GL_INCR);
    glViewport(-size, -size, 2 * size, 2 * size);
    glVertexPointer(2, GL_FLOAT, 0, positions);
    glTexCoordPointer(2, GL_FLOAT, 0, texcoords);

    for (MeshMaterialConstIterator s = oldm.materials.begin(); s != oldm.materials.end(); s++)
    {
        int p = 0;
        for (MeshFaceConstIterator f = m.faces.real_begin(); f != m.faces.real_end(); f++)
        {
            MeshHalfEdgeConstIterator e = f->edges();
            MeshFaceConstIterator oldf = edge_map[e]->face();
            if (oldf->material() != s) continue;
            for (int i = 0; i < 3; i++, e = e->next())
            {
                MeshVertexConstIterator v = e->vertex();
                MeshVertexConstIterator oldv = edge_map[e]->vertex();
                positions[p * 2 + 0] = v->data.texcoords[0];
                positions[p * 2 + 1] = v->data.texcoords[1];
                texcoords[p * 2 + 0] = oldv->data.texcoords[0];
                texcoords[p * 2 + 1] = oldv->data.texcoords[1];
                p++;
            }
        }
        glBindTexture(GL_TEXTURE_2D, texture_ids[s->texture()]);
        glDrawArrays(GL_TRIANGLES, 0, p);
    }
    glBindTexture(GL_TEXTURE_2D, 0);
    for (texture_ids_type::const_iterator t = texture_ids.begin(); t != texture_ids.end(); t++)
        glDeleteTextures(1, &t->second);

    m.materials.begin()->texture() = m.textures.insert(m.textures.end(), MeshTexture());
    buffer_to_texture(res, res, oversample.int_value, m.materials.begin()->texture(), false);
    destroy_renderbuffer(rbo);
}

int main(int argc, char **argv)
{
    gtk_init(&argc, &argv);
    initialise_gl();

    vector<string> comments;
    int args;
    args = process_options(argc, argv);
    Mesh m;

    try
    {
        format_load_mesh_file_comments(in_format.string_value, argv[args], m,
                                       MESH_LEVEL_PATCHES, comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to load mesh: " << e.what() << endl;
        return 1;
    }
    comments.push_back(alod_generation_string(argc, argv));

    if (resolution.int_value < 0)
        resolution.int_value = get_resolution(m);

    if (do_tangents.used)
        sample_tangents(m);
    if (do_textures.used)
    {
        if (!oldmeshfile.used)
        {
            cerr << "Error: --texture must be accompanied by --old-mesh\n";
            return 1;
        }
        Mesh oldm;
        try
        {
            format_load_mesh_file(in_format.string_value,
                                  oldmeshfile.string_value.c_str(),
                                  oldm,
                                  MESH_LEVEL_PATCHES);
        }
        catch (bad_format &e)
        {
            cerr << "Failed to load old mesh: " << e.what() << endl;
            return 1;
        }
        sample_texture(m, oldm);
    }
    mesh_garbage_collect(m);

    try
    {
        format_save_mesh_file(out_format.string_value, out_file, m,
                              comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to save mesh: " << e.what() << endl;
        return 1;
    }

    return 0;
}
