#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "mesh/mesh.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include <iostream>
#include <vector>
#include <string>
#include <cstdlib>
#include <cstdio>
#include <glib.h>

using namespace std;

static AlodProgramOption in_format, out_format;
static AlodProgramOption method, no_influences, no_edges;
static string out_file;

static int process_options(int argc, char **argv)
{
    static AlodProgramOption verbose, debug, threed;

    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('3', "3d", "ignore skeleton in computing metrics",
                              0, PROGRAM_ARGUMENT_STRING, &threed),
        AlodProgramOptionInfo('i', "input", "input format",
                              1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('o', "output", "output format",
                              1, PROGRAM_ARGUMENT_STRING, &out_format),
        AlodProgramOptionInfo('m', "method", "simplification method",
                              1, PROGRAM_ARGUMENT_STRING, &method),
        AlodProgramOptionInfo('n', "no-influences", "don't do influence simplification",
                              0, PROGRAM_ARGUMENT_STRING, &no_influences),
        AlodProgramOptionInfo('N', "no-edges", "don't do edge collapses",
                              0, PROGRAM_ARGUMENT_STRING, &no_edges),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging",
                              0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information",
                              0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo()
    };

    int args;
    args = alod_process_options(argc, argv, opts);
    if (!in_format.used) in_format.string_value = "auto";
    if (!out_format.used) out_format.string_value = "auto";
    if (!method.used) method.string_value = "simple";
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    if (threed.used) Skeleton::set_metric_mode(Skeleton::MODE_3D);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);


    if (argc - args < 1 || argc - args > 2)
    {
        cerr << "Error: must specify input (and optionally output) file\n";
        alod_show_help(opts);
        exit(1);
    }
    if (argc - args == 1)
        out_file = argv[args];
    else
        out_file = argv[args + 1];
    return args;
}

int main(int argc, char **argv)
{
    Mesh m;
    int args;
    vector<string> comments;

    args = process_options(argc, argv);
    try
    {
        format_load_mesh_file_comments(in_format.string_value, argv[args], m,
                                       MESH_LEVEL_PATCHES, comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to load mesh: " << e.what() << endl;
        return 1;
    }
    comments.push_back(alod_generation_string(argc, argv));

    double metric = mesh_simplify(m, method.string_value.c_str(),
                                  !no_influences.used,
                                  !no_edges.used);
    g_log(G_LOG_DOMAIN, G_LOG_LEVEL_INFO, "Maximum metric: %g", metric);
    g_assert(mesh_manifold(m, MESH_LEVEL_PATCHES));

    try
    {
        format_save_mesh_file(out_format.string_value, out_file, m,
                              comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to save mesh: " << e.what() << endl;
        return 1;
    }
    return 0;
}
