#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "glee/GLee.h"
#include "mesh/mesh.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include "common/paths.h"
#include "alodgl/glmisc.h"
#include "alodgl/shaders.h"
#include "mesh/iheap.h"
#include <GL/glut.h>
#include <glib.h>
#include <string>
#include <utility>
#include <iostream>
#include <sstream>
#include <iterator>
#include <functional>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

using namespace std;

enum
{
    SHADER_INIT,
    SHADER_VERTEX,
    SHADER_PIXEL,

    SHADER_COUNT
};
#define NUM_SHADERS ((size_t) SHADER_COUNT)

enum
{
    PROGRAM_CORRECT,
    PROGRAM_INVERT,
    PROGRAM_COMBINE,
    PROGRAM_INVERT_ERROR,
    PROGRAM_COMBINE_ERROR,

    PROGRAM_COUNT
};
#define NUM_PROGRAMS ((size_t) PROGRAM_COUNT)

enum
{
    UNIFORM_UM,
    UNIFORM_UIM,

    UNIFORM_COUNT
};
#define NUM_UNIFORMS ((size_t) UNIFORM_COUNT)

#define BIND_OFFSET 3

static AlodProgramOption in_format;
static AlodProgramOption verbose, debug;
static Mesh mesh;
static GLuint range_end;
static GLsizei range_count;
static GLhandleARB shaders[NUM_SHADERS];
static GLhandleARB programs[NUM_PROGRAMS];
static GLint locations[NUM_PROGRAMS][NUM_UNIFORMS];

static bool anim = false;
static int mode = 0, old_mode = -1;

static int process_args(int argc, char **argv)
{
    AlodProgramOption mode_option;

    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo()
    };

    int args = alod_process_options(argc, argv, opts);
    if (!in_format.used) in_format.string_value = "auto";
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);

    if (argc - args != 1)
    {
        cerr << "Error: must specify input file\n";
        alod_show_help(opts);
        exit(1);
    }
    return args;
}

static size_t get_max_influences(Mesh &m)
{
    size_t max_inf = 0;
    for (MeshPositionIterator p = mesh.positions.real_begin(); p != mesh.positions.real_end(); p++)
    {
        VectorPosition influences = p->position;
        MeshHalfEdgeConstIterator e = p->first_vertex()->first_edge();
        do
        {
            MeshFaceConstIterator f = e->face();
            if (mesh_true_face(f))
                influences += e->prev()->vertex()->position()->position;
            e = e->dual()->next();
        } while (e != p->first_vertex()->first_edge());
        max_inf = max(max_inf, (size_t) distance(influences.begin(), influences.end()));
    }
    return max_inf;
}

static string int2str(int x)
{
    ostringstream out;
    out << x;
    return out.str();
}

const char *shader_files[NUM_SHADERS] =
{
    "ssdnormals/init.glsl",
    "ssdnormals/vertex.glsl",
    "ssdnormals/fragment.glsl"
};
const GLenum shader_types[NUM_SHADERS] =
{
    GL_VERTEX_SHADER_ARB,
    GL_VERTEX_SHADER_ARB,
    GL_FRAGMENT_SHADER_ARB
};

static void initialise_shaders()
{
    string source;
    for (size_t i = 0; i < NUM_SHADERS; i++)
    {
        source = alodgl_get_shader_file_m4(shader_files[i]);
        shaders[i] = alodgl_load_hl_shader(shader_types[i], source, shader_files[i]);
    }
}

static void initialise_programs(Mesh &m, size_t max_inf)
{
    size_t max_bones, max_groups;
    max_bones = m.skeleton.get_read().size();
    max_groups = (max_inf + 3) / 4;

    alodgl_set_shader_flag("MAX_BONES", max_bones);
    alodgl_set_shader_flag("MAX_INFLUENCES", max_inf);
    alodgl_set_shader_flag("MAX_GROUPS", max_groups);

    for (size_t i = 0; i < NUM_PROGRAMS; i++)
    {
        programs[i] = glCreateProgramObjectARB();
        alodgl_set_shader_flag("MODE", i);
        initialise_shaders();
        string name;

        for (size_t j = 0; j < max_inf; j++)
            glBindAttribLocationARB(programs[i], BIND_OFFSET + j, (name = "aW" + int2str(j)).c_str());
        for (size_t j = 0; j < max_groups; j++)
            glBindAttribLocationARB(programs[i], BIND_OFFSET + max_inf + j, (name = "aB" + int2str(j)).c_str());
        for (size_t j = 0; j < NUM_SHADERS; j++)
        {
            glAttachObjectARB(programs[i], shaders[j]);
            glDeleteObjectARB(shaders[j]);
        }
        alodgl_link_program(programs[i]);
    }
}

static void initialise_locations()
{
    const char *names[NUM_UNIFORMS] =
    {
        "uM",
        "uiM"
    };
    for (size_t i = 0; i < NUM_PROGRAMS; i++)
        for (size_t j = 0; j < NUM_UNIFORMS; j++)
            locations[i][j] = glGetUniformLocationARB(programs[i], names[j]);
}

/* Determines a matrix to rescale and shift a mesh to unit size. */
static void normalise_mesh(Mesh &mesh)
{
    /* Apply a normalisation to the geometry */
    pair<Vector3, Vector3> bounds;
    Vector3 range, mid;

    bounds = mesh_bounding_box_3d(mesh);
    range = bounds.second - bounds.first;
    mid = (bounds.first + bounds.second) * 0.5;
    double s = 1.0 / max(max(range[0], range[1]), range[2]);
    Matrix4 T = matrix_scale(s) * matrix_translate(-mid);
    mesh_transform(mesh, T);
}

static void initialise_geometry(Mesh &mesh, size_t max_inf)
{
    size_t num_verts = mesh.positions.real_size();
    size_t num_faces = mesh.faces.real_size();
    size_t max_groups = (max_inf + 3) / 4;
    GLuint buffer;
    GLfloat *vertices, *normals;
    GLuint *indices;
    vector<GLfloat *> weights(max_inf);
    vector<GLshort *> bones(max_groups);
    int index, bone;
    const Skeleton &skeleton = mesh.skeleton.get_read();

    vertices = new GLfloat[num_verts * 3];
    normals = new GLfloat[num_verts * 3];
    fill(normals, normals + num_verts * 3, 0.0f);
    for (size_t i = 0; i < max_inf; i++)
    {
        weights[i] = new GLfloat[num_verts * 4];
        fill(weights[i], weights[i] + num_verts * 4, 0.0f);
    }
    for (size_t i = 0; i < max_groups; i++)
    {
        bones[i] = new GLshort[num_verts * 4];
        fill(bones[i], bones[i] + num_verts * 4, 0);
    }
    indices = new GLuint[num_faces * 3];

    /* Construct rest coordinates, scalar weights and bones */
    index = 0;
    for (MeshPositionIterator p = mesh.positions.real_begin(); p != mesh.positions.real_end(); p++, index++)
    {
        const VectorPosition &pos = p->position;
        VectorPosition influences = pos;

        Vector3 pos3 = skeleton.ssd_to_3d(pos);
        vertices[index * 3 + 0] = pos3[0];
        vertices[index * 3 + 1] = pos3[1];
        vertices[index * 3 + 2] = pos3[2];
        p->internal_data = GSIZE_TO_POINTER(index);

        MeshHalfEdgeConstIterator e = p->first_vertex()->first_edge();
        do
        {
            MeshFaceConstIterator f = e->face();
            if (mesh_true_face(f))
                influences += e->prev()->vertex()->position()->position;
            e = e->dual()->next();
        } while (e != p->first_vertex()->first_edge());
        bone = 0;
        for (VectorPosition::const_iterator i = influences.begin(); i != influences.end(); i++, bone++)
        {
            weights[bone][index * 4 + 3] = pos[i->index + 3];
            bones[bone / 4][index * 4 + bone % 4] = i->index / 4;
        }
    }

    /* Construct normals and weight gradients, prior to normalisation */
    index = 0;
    for (MeshFaceConstIterator f = mesh.faces.real_begin(); f != mesh.faces.real_end(); f++, index++)
    {
        VectorPosition pos[3];
        Vector3 pos3[3];
        VectorPosition influences(skeleton);

        MeshHalfEdgeConstIterator e = f->edges();
        for (int j = 0; j < 3; j++, e = e->next())
        {
            pos[j] = e->vertex()->position()->position;
            pos3[j] = skeleton.ssd_to_3d(pos[j]);
            indices[index * 3 + j] = GPOINTER_TO_SIZE(e->vertex()->position()->internal_data);
            influences += e->vertex()->position()->position;
        }
        Vector3 s = pos3[1] - pos3[0];
        Vector3 t = pos3[2] - pos3[0];
        Vector3 n = normalise(cross(s, t));
        for (int j = 0; j < 3; j++, e = e->next())
            for (int k = 0; k < 3; k++)
                normals[indices[index * 3 + j] * 3 + k] += n[k];
        for (VectorPosition::const_iterator i = influences.begin(); i != influences.end(); i++)
        {
            /* grad = a*s + b*t
             * grad.s = ds = a(s.s) + b(s.t)
             * grad.t = dt = a(s.t) + b(t.t)
             */
            int bone = i->index / 4;
            Matrix2 eq(s * s, s * t, s * t, t * t);
            Vector2 delta(pos[1][i->index + 3] - pos[0][i->index + 3],
                          pos[2][i->index + 3] - pos[0][i->index + 3]);
            Vector2 ab = inverse(eq) * delta;
            Vector3 grad = ab[0] * s + ab[1] * t;
            for (int j = 0; j < 3; j++, e = e->next())
            {
                for (unsigned int k = 0; k < max_inf; k++)
                    if (bones[k / 4][4 * indices[index * 3 + j] + k % 4] == bone)
                    {
                        for (int l = 0; l < 3; l++)
                            weights[k][4 * indices[index * 3 + j] + l] += grad[l];
                        break;
                    }
            }
        }
    }

    /* Normalise normals and rescale gradients */
    index = 0;
    for (MeshPositionIterator p = mesh.positions.real_begin(); p != mesh.positions.real_end(); p++, index++)
    {
        Vector3 n(normals[index * 3 + 0],
                  normals[index * 3 + 1],
                  normals[index * 3 + 2]);
        n = normalise(n);
        normals[index * 3 + 0] = n[0];
        normals[index * 3 + 1] = n[1];
        normals[index * 3 + 2] = n[2];

        int incident = 0;
        MeshHalfEdgeConstIterator e = p->first_vertex()->first_edge();
        do
        {
            MeshFaceConstIterator f = e->face();
            if (mesh_true_face(f))
                incident++;
            e = e->dual()->next();
        } while (e != p->first_vertex()->first_edge());
        if (incident)
        {
            double s = 1.0 / incident;
            for (size_t i = 0; i < max_inf; i++)
            {
                Vector3 g(weights[i][4 * index + 0],
                          weights[i][4 * index + 1],
                          weights[i][4 * index + 2]);
                g *= s;
                g -= (g * n) * n;
                for (int j = 0; j < 3; j++)
                    weights[i][4 * index + j] = g[j];
            }
        }
    }

    glGenBuffersARB(1, &buffer);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, buffer);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, 3 * num_verts * sizeof(GLfloat),
                    vertices, GL_STATIC_DRAW_ARB);
    glVertexPointer(3, GL_FLOAT, 0, NULL);
    glEnableClientState(GL_VERTEX_ARRAY);
    delete[] vertices;

    glGenBuffersARB(1, &buffer);
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, buffer);
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, 3 * num_verts * sizeof(GLfloat),
                    normals, GL_STATIC_DRAW_ARB);
    glNormalPointer(GL_FLOAT, 0, NULL);
    glEnableClientState(GL_NORMAL_ARRAY);
    delete[] normals;

    glGenBuffersARB(1, &buffer);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, buffer);
    glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 3 * num_faces * sizeof(GLuint),
                    indices, GL_STATIC_DRAW_ARB);
    delete[] indices;

    for (size_t i = 0; i < max_inf; i++)
    {
        glGenBuffersARB(1, &buffer);
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, buffer);
        glBufferDataARB(GL_ARRAY_BUFFER_ARB, 4 * num_verts * sizeof(GLfloat),
                        weights[i], GL_STATIC_DRAW_ARB);
        glVertexAttribPointerARB(BIND_OFFSET + i, 4, GL_FLOAT, GL_FALSE, 0, NULL);
        glEnableVertexAttribArrayARB(BIND_OFFSET + i);
        delete[] weights[i];
    }
    for (size_t i = 0; i < max_groups; i++)
    {
        glGenBuffersARB(1, &buffer);
        glBindBufferARB(GL_ARRAY_BUFFER_ARB, buffer);
        glBufferDataARB(GL_ARRAY_BUFFER_ARB, 4 * num_verts * sizeof(GLshort),
                        bones[i], GL_STATIC_DRAW_ARB);
        glVertexAttribPointer(BIND_OFFSET + max_inf + i, 4, GL_SHORT, GL_FALSE, 0, NULL);
        glEnableVertexAttribArrayARB(BIND_OFFSET + max_inf + i);
        delete[] bones[i];
    }

    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    range_end = num_verts - 1;
    range_count = 3 * num_faces;
}

static void initialise_gl(Mesh &mesh)
{
    GLeeInit();
    if (!GLEE_VERSION_1_3)
    {
        cerr << "OpenGL 1.3 is required\n";
        exit(1);
    }
    if (!GLEE_ARB_shader_objects
        || !GLEE_ARB_vertex_shader)
    {
        cerr << "GLSL is required\n";
        exit(1);
    }
    if (!GLEE_ARB_vertex_buffer_object)
    {
        cerr << "Buffer objects are required\n";
        exit(1);
    }

    size_t max_inf = get_max_influences(mesh);
    initialise_programs(mesh, max_inf);
    initialise_locations();
    normalise_mesh(mesh);
    initialise_geometry(mesh, max_inf);
    glLoadIdentity();
    gluLookAt(1.0, 1.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
    glClearColor(0.0, 0.0, 0.0, 0.0);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
}

static void load_uniforms(Mesh &m)
{
    static double frame = 0.0;
    Skeleton &skeleton = m.skeleton.get_write();
    size_t bones = skeleton.size();
    Matrix4 rest[bones];
    GLfloat matrices[16 * bones];
    GLint location;

    skeleton.set_to_rest();
    for (size_t i = 0; i < bones; i++)
        rest[i] = skeleton.get_absolute(i);

    // if (anim) frame += 0.05;
    if (anim) frame += 1.0;
    if ((size_t) frame >= skeleton.num_frames()) frame -= skeleton.num_frames();
    skeleton.set_to_frame((size_t) frame);

    location = locations[mode][UNIFORM_UM];
    if (location >= 0)
    {
        GLfloat *cur = matrices;
        for (size_t i = 0; i < bones; i++)
        {
            Matrix4 warp = skeleton.get_model_transform() * skeleton.get_absolute(i) * inverse(rest[i]);
            for (int j = 0; j < 4; j++)
                for (int k = 0; k < 4; k++)
                    *cur++ = warp[k][j];
        }
        glUniformMatrix4fvARB(location, bones, GL_FALSE, matrices);
    }
    location = locations[mode][UNIFORM_UIM];
    if (location >= 0)
    {
        GLfloat *cur = matrices;
        for (size_t i = 0; i < bones; i++)
        {
            Matrix4 warp = skeleton.get_model_transform() * skeleton.get_absolute(i) * inverse(rest[i]);
            Matrix3 iwarp = inverse(matrix4_to_matrix3(warp));
            for (int j = 0; j < 3; j++)
                for (int k = 0; k < 3; k++)
                    *cur++ = iwarp[j][k]; /* Note: this is the transpose */
        }
        glUniformMatrix3fvARB(location, bones, GL_FALSE, matrices);
    }
}

static void display()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    if (mode != old_mode)
    {
        glUseProgramObjectARB(programs[mode]);
        old_mode = mode;
    }
    load_uniforms(mesh);
    glDrawRangeElements(GL_TRIANGLES, 0, range_end, range_count, GL_UNSIGNED_INT, NULL);
    glutSwapBuffers();
}

static void reshape(int w, int h)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0, (double) w / h, 0.01, 300.0);
    glMatrixMode(GL_MODELVIEW);
    glViewport(0, 0, w, h);
}

static void keyboard(unsigned char key, int x, int y)
{
    if (key >= '1' && key <= '1' + NUM_PROGRAMS)
        mode = key - '1';
    else
    {
        switch (key)
        {
        case 's':
            anim = !anim;
            break;
        }
    }
}

static void idle()
{
    glutPostRedisplay();
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);

    int args = process_args(argc, argv);
    try
    {
        format_load_mesh_file(in_format.string_value, argv[args], mesh, MESH_LEVEL_MANIFOLD);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to load mesh: " << e.what() << endl;
        exit(1);
    }

    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA);
    glutInitWindowSize(1024, 768);
    glutCreateWindow("SSDNormals");
    initialise_gl(mesh);
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(keyboard);
    glutIdleFunc(idle);
    glutMainLoop();

}
