#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "glee/GLee.h"
#include "mesh/mesh.h"
#include "formats/formats.h"
#include "common/options.h"
#include "common/logging.h"
#include "common/paths.h"
#include "alodgl/glmisc.h"
#include "alodgl/shaders.h"
#include "mesh/iheap.h"
#include <GL/glut.h>
#include <glib.h>
#include <string>
#include <utility>
#include <sstream>
#include <iterator>
#include <functional>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>

// #define SSDTEST_DEBUG

#define FRAME_COUNT 50000
#define FRAME_OFFSET 5000

using namespace std;

enum RenderMode
{
    MODE_STATIC,
    MODE_SSD,
    MODE_AS,
    MODE_MWE
};

struct GLVector3
{
    GLfloat v[3];
};

struct GLVector4
{
    GLfloat v[4];
};

struct GLShortVector4
{
    GLshort v[4];
};

static AlodProgramOption in_format;
static AlodProgramOption verbose, debug;
static RenderMode mode;

static Mesh mesh;
static vector<GLuint> shaders;
static GLuint num_vertices, num_faces;
static vector<GLuint> group_sizes;
static int frame = 0;
static vector<Matrix4> rest_inv;

static double frame_times[FRAME_COUNT];
static int frame_number = -FRAME_OFFSET;
static struct timeval frame_stamp;

static int face_metric(const MeshFaceConstIterator &f)
{
    int m = 1;
    MeshHalfEdgeConstIterator e;

    switch (mode)
    {
    case MODE_STATIC: return 1;
    case MODE_SSD:
    case MODE_AS:
    case MODE_MWE:
        e = f->edges();
        for (int i = 0; i < 3; i++, e = e->next())
        {
            const VectorPosition &p = e->vertex()->position()->position;
            m = max(m, (int) distance(p.begin(), p.end()));
        }
    }
    switch (mode)
    {
    case MODE_STATIC: break;
    case MODE_SSD:
    case MODE_AS:
        if (m > 8) g_error("Too many influences");
        break;
    case MODE_MWE:
        if (m > 4) g_error("Too many influences");
        break;
    }
    return 1;
}

static int process_args(int argc, char **argv)
{
    AlodProgramOption mode_option;
    struct { const char * label; RenderMode value; } modes[] =
    {
        { "static", MODE_STATIC },
        { "ssd", MODE_SSD },
        { "as", MODE_AS },
        { "mwe", MODE_MWE }
    };

    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('i', "input", "input format", 1, PROGRAM_ARGUMENT_STRING, &in_format),
        AlodProgramOptionInfo('m', "mode", "Specify animation mode", 1, PROGRAM_ARGUMENT_STRING, &mode_option),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo()
    };

    int args = alod_process_options(argc, argv, opts);
    if (!in_format.used) in_format.string_value = "auto";
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);

    if (!mode_option.used)
    {
        cerr << "Error: -m <mode> must be specified\n";
        alod_show_help(opts);
        exit(1);
    }
    for (gsize i = 0; i < G_N_ELEMENTS(modes); i++)
        if (mode_option.string_value == modes[i].label)
        {
            mode_option.used = false; // indicates that it is found
            mode = modes[i].value;
            break;
        }
    if (mode_option.used)
    {
        cerr << "Error: invalid mode " << mode_option.string_value << "\n";
        cerr << "Valid modes are:";
        for (gsize i = 0; i < G_N_ELEMENTS(modes); i++)
            cerr << " " << modes[i].label;
        cerr << "\n";
        exit(1);
    }

    if (argc - args != 1)
    {
        cerr << "Error: must specify input file\n";
        alod_show_help(opts);
        exit(1);
    }
    return args;
}

static void display()
{
    int G = group_sizes.size() - 1;
    Skeleton &skeleton = mesh.skeleton.get_write();
    skeleton.set_to_frame(frame);
    if (++frame == (int) skeleton.num_frames()) frame = 0;

    switch (mode)
    {
    case MODE_STATIC:
        break;
    case MODE_SSD:
    case MODE_AS:
    case MODE_MWE:
        for (size_t i = 0; i < skeleton.size(); i++)
        {
            Matrix4 m = skeleton.get_model_transform() * skeleton.get_absolute(i);
            if (mode == MODE_SSD || mode == MODE_MWE) m *= rest_inv[i];
            for (int r = 0; r < 3; r++)
                glProgramEnvParameter4dARB(GL_VERTEX_PROGRAM_ARB, 3 * i + r,
                                           m[r][0], m[r][1], m[r][2], m[r][3]);
        }
        break;
    }
    glLoadIdentity();
#ifdef SSDTEST_DEBUG
    gluLookAt(50.0, 50.0, 50.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
#else
    gluLookAt(50.0, 50.0, 50.0, 100.0, 100.0, 100.0, 0.0, 1.0, 0.0);
#endif

    for (int i = 1; i < 16; i++)
        glVertexAttrib4dARB(i, 0.0, 0.0, 0.0, 0.0);
    GLuint *start = NULL;

    for (int i = 1; i <= G; i++)
    {
        switch (mode)
        {
        case MODE_STATIC:
            break;
        case MODE_SSD:
            if (i == 5)
            {
                glEnableVertexAttribArrayARB(3);
                glEnableVertexAttribArrayARB(4);
            }
            break;
        case MODE_AS:
            if (i == 5) glEnableVertexAttribArrayARB(5);
            if (i > 1) glEnableVertexAttribArrayARB(i + (i - 1) / 4);
            break;
        case MODE_MWE:
            if (i > 1)
            {
                glEnableVertexAttribArrayARB(i * 3 - 2);
                glEnableVertexAttribArrayARB(i * 3 - 1);
                glEnableVertexAttribArrayARB(i * 3 - 0);
            }
            break;
        }
        if (group_sizes[i])
        {
            glBindProgramARB(GL_VERTEX_PROGRAM_ARB, shaders[i]);
            glDrawRangeElements(GL_TRIANGLES, 0, num_vertices - 1, 3 * group_sizes[i], GL_UNSIGNED_INT, start);
        }
        start += 3 * group_sizes[i];
    }
    glutSwapBuffers();
    switch (mode)
    {
    case MODE_STATIC:
        break;
    case MODE_SSD:
        glDisableVertexAttribArrayARB(3);
        glDisableVertexAttribArrayARB(4);
        break;
    case MODE_AS:
        for (int i = 2; i < G + (G - 1) / 4; i++)
            glDisableVertexAttribArrayARB(i);
        break;
    case MODE_MWE:
        for (int i = 4; i <= G * 3; i++)
            glDisableVertexAttribArrayARB(i);
        break;
    }

    struct timeval now;
    gettimeofday(&now, NULL);
    double tm = (now.tv_sec - frame_stamp.tv_sec) + 1e-6 * (now.tv_usec - frame_stamp.tv_usec);
    frame_stamp = now;

    if (frame_number >= 0) frame_times[frame_number] = tm;
    frame_number++;
    if (frame_number == FRAME_COUNT)
    {
        double mean = 0.0;
        double std = 0.0;
        double mean_std;
        for (int i = 0; i < FRAME_COUNT; i++)
            mean += frame_times[i];
        mean /= FRAME_COUNT;
        for (int i = 0; i < FRAME_COUNT; i++)
            std += (mean - frame_times[i]) * (mean - frame_times[i]);
        mean_std = sqrt(std) / FRAME_COUNT;
        std = sqrt(std / FRAME_COUNT);

        printf("Mean: %9.6f +/- %9.6f (ms)\n", mean * 1000.0, mean_std * 1000.0);
        exit(0);
    }
}

static void reshape(int w, int h)
{
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(90.0, (double) w / h, 0.01, 300.0);
    glMatrixMode(GL_MODELVIEW);
    glViewport(0, 0, w, h);
}

static void idle()
{
    glutPostRedisplay();
}

struct VertexReorder
{
    int last_used;
    int seq;

    VertexReorder() : last_used(-1), seq(-1) {}
};

struct FaceReorder
{
    int sum_last_used;
    int seq;

    FaceReorder() : sum_last_used(-3), seq(-1) {}
};

struct CompareReorder : public binary_function<MeshFaceIterator, MeshFaceIterator, bool>
{
    bool operator()(const MeshFaceIterator a, const MeshFaceIterator b) const
    {
        int ma = face_metric(a);
        int mb = face_metric(b);
        if (ma != mb) return ma > mb;
        ma = static_cast<FaceReorder *>(a->internal_data)->sum_last_used;
        mb = static_cast<FaceReorder *>(b->internal_data)->sum_last_used;
        return ma < mb;
    }
};

static void reorder_mesh(Mesh &mesh)
{
    VertexReorder *vertex_data = new VertexReorder[mesh.vertices.real_size()];
    FaceReorder *face_data = new FaceReorder[mesh.faces.real_size()];
    indexed_priority_queue<void *, MeshFaceIterator, vector<pair<void *, MeshFaceIterator> >, CompareReorder> pq;

    size_t index = 0;
    for (MeshVertexIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++, index++)
        v->internal_data = &vertex_data[index];
    index = 0;
    for (MeshFaceIterator f = mesh.faces.real_begin(); f != mesh.faces.real_end(); f++, index++)
    {
        f->internal_data = &face_data[index];
        pq.push(make_pair(f->internal_data, f));
    }

    int findex = 0;
    int vindex = 0;
    while (!pq.empty())
    {
        MeshFaceIterator f = pq.top().second;
        FaceReorder *rf = static_cast<FaceReorder *>(f->internal_data);
        pq.pop();
        rf->seq = findex++;
        MeshHalfEdgeIterator e = f->edges();
        for (int i = 0; i < 3; i++, e = e->next())
        {
            MeshVertexIterator v = e->vertex();
            VertexReorder *rv = static_cast<VertexReorder *>(v->internal_data);
            if (rv->seq == -1) rv->seq = vindex++;
            int delta = findex - rv->last_used;
            rv->last_used = findex;

            MeshHalfEdgeIterator e2 = v->first_edge();
            do
            {
                MeshFaceIterator f2 = e2->face();
                FaceReorder *rf2 = static_cast<FaceReorder *>(f2->internal_data);
                if (rf2->seq == -1)
                {
                    rf2->sum_last_used += delta;
                    // No change in the value, but forces a bubble-up
                    pq[f2->internal_data] = f2;
                }
                e2 = e2->prev()->dual();
            } while (e2 != v->first_edge() && e2->next()->vertex() == v);
        }
    }

    for (MeshVertexIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++)
    {
        int seq = static_cast<VertexReorder *>(v->internal_data)->seq;
        g_assert(seq != -1);
        v->internal_data = GSIZE_TO_POINTER((gsize) seq);
    }
    for (MeshFaceIterator f = mesh.faces.real_begin(); f != mesh.faces.real_end(); f++)
    {
        int seq = static_cast<FaceReorder *>(f->internal_data)->seq;
        g_assert(seq != -1);
        f->internal_data = GSIZE_TO_POINTER((gsize) seq);
    }

    delete[] vertex_data;
    delete[] face_data;
}

static void initialise_gl(Mesh &mesh)
{
    Skeleton &skeleton = mesh.skeleton.get_write();
    string shader_name, source;
    GLuint vb, ib;
    int maxb = 1;

    GLeeInit();
    if (!GLEE_ARB_vertex_program)
    {
        cerr << "Vertex programs not supported\n";
        exit(1);
    }
    if (!GLEE_ARB_vertex_buffer_object)
    {
        cerr << "Vertex buffer objects not supported\n";
        exit(1);
    }

    reorder_mesh(mesh);

    num_vertices = mesh.vertices.real_size();
    num_faces = mesh.faces.real_size();

    GLuint *faces = new GLuint[3 * num_faces];
    for (MeshFaceConstIterator f = mesh.faces.real_begin(); f != mesh.faces.real_end(); f++)
    {
        int seq = GPOINTER_TO_SIZE(f->internal_data);
        MeshHalfEdgeConstIterator e = f->edges();
        for (int j = 0; j < 3; j++, e = e->next())
            faces[seq * 3 + j] = GPOINTER_TO_SIZE(e->vertex()->internal_data);
        int m = face_metric(f);
        if ((int) group_sizes.size() <= m) group_sizes.resize(m + 1, 0);
        group_sizes[m]++;
    }

    glGenBuffersARB(1, &ib);
    glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, ib);
    glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 3 * num_faces * sizeof(GLuint),
                    faces, GL_STATIC_DRAW_ARB);
    delete[] faces;

    switch (mode)
    {
    case MODE_STATIC:
        {
            skeleton.set_to_rest();
            GLVector3 *vertices = new GLVector3[num_vertices];
            for (MeshVertexConstIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++)
            {
                int index = GPOINTER_TO_SIZE(v->internal_data);
                VectorPosition pos = v->position()->position;
                Vector3 rest = skeleton.ssd_to_3d(pos);
                vertices[index].v[0] = rest[0];
                vertices[index].v[1] = rest[1];
                vertices[index].v[2] = rest[2];
            }

            glGenBuffersARB(1, &vb);
            glBindBufferARB(GL_ARRAY_BUFFER_ARB, vb);
            glBufferDataARB(GL_ARRAY_BUFFER_ARB, num_vertices * sizeof(GLVector3),
                            vertices, GL_STATIC_DRAW_ARB);
            delete[] vertices;
            glVertexAttribPointerARB(0, 3, GL_FLOAT, GL_FALSE, sizeof(GLVector3), NULL);
            glEnableVertexAttribArrayARB(0);
        }
        break;
    case MODE_SSD:
        {
            skeleton.set_to_rest();
            GLVector3 *vertices = new GLVector3[num_vertices];
            GLVector4 *weights[2] = { new GLVector4[num_vertices], new GLVector4[num_vertices] };
            GLShortVector4 *bones[2] = { new GLShortVector4[num_vertices], new GLShortVector4[num_vertices] };
            memset(weights[0], 0, num_vertices * sizeof(GLVector4));
            memset(weights[1], 0, num_vertices * sizeof(GLVector4));
            memset(bones[0], 0, num_vertices * sizeof(GLShortVector4));
            memset(bones[1], 0, num_vertices * sizeof(GLShortVector4));
            for (MeshVertexConstIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++)
            {
                int index = GPOINTER_TO_SIZE(v->internal_data);
                const VectorPosition &pos = v->position()->position;
                int b = 0;
                double tweight = 0.0;
                for (VectorPosition::const_iterator i = pos.begin(); i != pos.end(); i++, b++)
                {
                    if (b == 8) g_error("Too many influences");
                    bones[b / 4][index].v[b % 4] = i->index / 4 * 3;
                    weights[b / 4][index].v[b % 4] = i->values[3];
                    tweight += i->values[3];
                }
                for (int i = 0; i < b; i++)
                    weights[i / 4][index].v[i % 4] /= tweight;
                Vector3 rest = skeleton.ssd_to_3d(pos);
                vertices[index].v[0] = rest[0];
                vertices[index].v[1] = rest[1];
                vertices[index].v[2] = rest[2];
                maxb = max(maxb, b);
            }

            glGenBuffersARB(1, &vb);
            glBindBufferARB(GL_ARRAY_BUFFER_ARB, vb);
            glBufferDataARB(GL_ARRAY_BUFFER_ARB, num_vertices * sizeof(GLVector3),
                            vertices, GL_STATIC_DRAW_ARB);
            glVertexAttribPointerARB(0, 3, GL_FLOAT, GL_FALSE, sizeof(GLVector3), NULL);
            delete[] vertices;

            for (int i = 0; i < 2; i++)
            {
                glGenBuffersARB(1, &vb);
                glBindBufferARB(GL_ARRAY_BUFFER_ARB, vb);
                glBufferDataARB(GL_ARRAY_BUFFER_ARB, num_vertices * sizeof(GLShortVector4),
                                bones[i], GL_STATIC_DRAW_ARB);
                glVertexAttribPointerARB(2 * i + 1, 4, GL_SHORT, GL_FALSE, sizeof(GLShortVector4), NULL);
                delete[] bones[i];

                glGenBuffersARB(1, &vb);
                glBindBufferARB(GL_ARRAY_BUFFER_ARB, vb);
                glBufferDataARB(GL_ARRAY_BUFFER_ARB, num_vertices * sizeof(GLVector4),
                                weights[i], GL_STATIC_DRAW_ARB);
                glVertexAttribPointerARB(2 * i + 2, 4, GL_FLOAT, GL_FALSE, sizeof(GLVector4), NULL);
                delete[] weights[i];
            }
            glEnableVertexAttribArrayARB(0);
            glEnableVertexAttribArrayARB(1);
            glEnableVertexAttribArrayARB(2);
        }
        break;
    case MODE_AS:
        {
            GLShortVector4 *bones[2] = { new GLShortVector4[num_vertices], new GLShortVector4[num_vertices] };
            GLVector4 *coords[8];
            for (int i = 0; i < 8; i++)
            {
                coords[i] = new GLVector4[num_vertices];
                memset(coords[i], 0, num_vertices * sizeof(GLVector4));
            }
            memset(bones[0], 0, num_vertices * sizeof(GLShortVector4));
            memset(bones[1], 0, num_vertices * sizeof(GLShortVector4));
            for (MeshVertexConstIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++)
            {
                int index = GPOINTER_TO_SIZE(v->internal_data);
                VectorPosition pos = v->position()->position;
                int b = 0;
                for (VectorPosition::const_iterator i = pos.begin(); i != pos.end(); i++, b++)
                {
                    if (b == 8) g_error("Too many influences");
                    bones[b / 4][index].v[b % 4] = i->index / 4 * 3;
                    for (int j = 0; j < 4; j++)
                        coords[b][index].v[j] = i->values[j];
                }
                maxb = max(maxb, b);
            }

            for (int i = 0; i < 2; i++)
            {
                glGenBuffersARB(1, &vb);
                glBindBufferARB(GL_ARRAY_BUFFER_ARB, vb);
                glBufferDataARB(GL_ARRAY_BUFFER_ARB, num_vertices * sizeof(GLShortVector4),
                                bones[i], GL_STATIC_DRAW_ARB);
                glVertexAttribPointerARB(5 * i, 4, GL_SHORT, GL_FALSE, sizeof(GLShortVector4), NULL);
                delete[] bones[i];

                for (int j = 0; j < 4; j++)
                {
                    glGenBuffersARB(1, &vb);
                    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vb);
                    glBufferDataARB(GL_ARRAY_BUFFER_ARB, num_vertices * sizeof(GLVector4),
                                    coords[4 * i + j], GL_STATIC_DRAW_ARB);
                    glVertexAttribPointerARB(5 * i + j + 1, 4, GL_FLOAT, GL_FALSE, sizeof(GLVector4), NULL);
                    delete[] coords[4 * i + j];
                }
            }
            glEnableVertexAttribArrayARB(0);
            glEnableVertexAttribArrayARB(1);
        }
        break;
    case MODE_MWE:
        {
            skeleton.set_to_rest();
            GLShortVector4 *bones = new GLShortVector4[num_vertices];
            GLVector4 *weights[4][3];
            for (int i = 0; i < 4; i++)
                for (int j = 0; j < 3; j++)
                {
                    weights[i][j] = new GLVector4[num_vertices];
                    memset(weights[i][j], 0, num_vertices * sizeof(GLVector4));
                }
            memset(bones, 0, num_vertices * sizeof(GLShortVector4));
            for (MeshVertexConstIterator v = mesh.vertices.real_begin(); v != mesh.vertices.real_end(); v++)
            {
                int index = GPOINTER_TO_SIZE(v->internal_data);
                VectorPosition pos = v->position()->position;
                int b = 0;
                for (VectorPosition::const_iterator i = pos.begin(); i != pos.end(); i++, b++)
                {
                    if (b == 4) g_error("Too many influences");
                    Vector4 world(i->values[0], i->values[1], i->values[2], i->values[3]);
                    world = skeleton.get_absolute(i->index / 4) * world;
                    bones[index].v[b] = i->index / 4 * 3;
                    for (int j = 0; j < 4; j++)
                        for (int k = 0; k < 3; k++)
                            weights[b][k][index].v[j] = world[j];
                }
                maxb = max(maxb, b);
            }

            glGenBuffersARB(1, &vb);
            glBindBufferARB(GL_ARRAY_BUFFER_ARB, vb);
            glBufferDataARB(GL_ARRAY_BUFFER_ARB, num_vertices * sizeof(GLShortVector4),
                            bones, GL_STATIC_DRAW_ARB);
            glVertexAttribPointerARB(0, 4, GL_SHORT, GL_FALSE, sizeof(GLShortVector4), NULL);
            delete[] bones;

            for (int j = 0; j < 4; j++)
                for (int k = 0; k < 3; k++)
                {
                    glGenBuffersARB(1, &vb);
                    glBindBufferARB(GL_ARRAY_BUFFER_ARB, vb);
                    glBufferDataARB(GL_ARRAY_BUFFER_ARB, num_vertices * sizeof(GLVector4),
                                    weights[j][k], GL_STATIC_DRAW_ARB);
                    glVertexAttribPointerARB(3 * j + k + 1, 4, GL_FLOAT, GL_FALSE, sizeof(GLVector4), NULL);
                    delete[] weights[j][k];
                }
            glEnableVertexAttribArrayARB(0);
            glEnableVertexAttribArrayARB(1);
            glEnableVertexAttribArrayARB(2);
            glEnableVertexAttribArrayARB(3);
        }
        break;
    }

    switch (mode)
    {
    case MODE_SSD:
    case MODE_MWE:
        skeleton.set_to_rest();
        for (size_t i = 0; i < skeleton.size(); i++)
        {
            Matrix4 m = skeleton.get_absolute(i);
            rest_inv.push_back(inverse(m));
        }
        break;
    case MODE_STATIC:
    case MODE_AS:
        break;
    }

    switch (mode)
    {
    case MODE_STATIC: shader_name = "ssdtest/static.vpm"; break;
    case MODE_SSD: shader_name = "ssdtest/ssd.vpm"; break;
    case MODE_AS: shader_name = "ssdtest/as.vpm"; break;
    case MODE_MWE: shader_name = "ssdtest/mwe.vpm"; break;
    }
    shaders.resize(maxb + 1);
    for (int i = 1; i <= maxb; i++)
    {
        alodgl_set_shader_flag("BONES", i);
        alodgl_set_shader_flag("TOTALBONES", skeleton.size());
        source = alodgl_get_shader_file_m4(shader_name);
        glGenProgramsARB(1, &shaders[i]);
        glBindProgramARB(GL_VERTEX_PROGRAM_ARB, shaders[i]);
        alodgl_load_ll_shader(GL_VERTEX_PROGRAM_ARB, source, shader_name);
    }
    glEnable(GL_VERTEX_PROGRAM_ARB);
    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
}

int main(int argc, char **argv)
{
    glutInit(&argc, argv);

    int args = process_args(argc, argv);
    try
    {
        format_load_mesh_file(in_format.string_value, argv[args], mesh, MESH_LEVEL_MANIFOLD);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to load mesh: " << e.what() << endl;
        exit(1);
    }

    glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH | GLUT_ALPHA);
    glutCreateWindow("SSDTest");
    initialise_gl(mesh);
    glutDisplayFunc(display);
    glutReshapeFunc(reshape);
    glutIdleFunc(idle);
    glutMainLoop();

    return 0;
}
