#if HAVE_CONFIG_H
# include <config.h>
#endif
#include <string>
#include <vector>
#include <algorithm>
#include "common/options.h"
#include "common/logging.h"
#include "common/mesch.h"
#include "mesh/mesh.h"
#include "formats/formats.h"
#include "geom/skeleton.h"
#include "geom/coords.h"
#include "geom/sparse.h"
using namespace std;

#define POSES 41
#define BLENDAREA 0.3

static AlodProgramOption out_format, ssd;
static AlodProgramOption rows, slices, tube_length, twist;
static AlodProgramOption low_mark, high_mark;
static string out_file;

static void process_options(int argc, char **argv)
{
    AlodProgramOption verbose, debug;

    AlodProgramOptionInfo opts[] =
    {
        AlodProgramOptionInfo('b', "basic", "use simple skinning", 0, PROGRAM_ARGUMENT_STRING, &ssd),
        AlodProgramOptionInfo('o', "output", "output format", 1, PROGRAM_ARGUMENT_STRING, &out_format),
        AlodProgramOptionInfo('s', "slices", "number of slices around circumference", 1, PROGRAM_ARGUMENT_INT, &slices),
        AlodProgramOptionInfo('r', "rows", "number of divisions along length", 1, PROGRAM_ARGUMENT_INT, &rows),
        AlodProgramOptionInfo('l', "length", "length of the tube (radius is 1)", 1, PROGRAM_ARGUMENT_DOUBLE, &tube_length),
        AlodProgramOptionInfo('v', "verbose", "more verbose logging", 0, PROGRAM_ARGUMENT_STRING, &verbose),
        AlodProgramOptionInfo('t', "twist", "twist rather than rotate", 0, PROGRAM_ARGUMENT_STRING, &twist),
        AlodProgramOptionInfo('L', "low", "angle low-mark", 1, PROGRAM_ARGUMENT_DOUBLE, &low_mark),
        AlodProgramOptionInfo('H', "high", "angle high-mark", 1, PROGRAM_ARGUMENT_DOUBLE, &high_mark),
        AlodProgramOptionInfo('d', "debug", "show debug information", 0, PROGRAM_ARGUMENT_STRING, &debug),
        AlodProgramOptionInfo()
    };

    int args;
    args = alod_process_options(argc, argv, opts);
    if (verbose.used) alod_log_set_level(G_LOG_LEVEL_INFO);
    if (debug.used) alod_log_set_level(G_LOG_LEVEL_DEBUG);
    if (!out_format.used) out_format.string_value = "auto";
    if (!rows.used) rows.int_value = 20;
    if (!slices.used) slices.int_value = 50;
    if (!tube_length.used) tube_length.double_value = 10.0;
    if (!low_mark.used) low_mark.double_value = 0.0;
    if (!high_mark.used) high_mark.double_value = 90.0;
    low_mark.double_value *= G_PI / 180.0;
    high_mark.double_value *= G_PI / 180.0;
    g_log_set_handler(NULL,
                      (GLogLevelFlags) (G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
                      alod_log_func,
                      NULL);

    if (argc - args != 1)
    {
        cerr << "Error: must output file\n";
        alod_show_help(opts);
        exit(1);
    }
    out_file = argv[args];
}

static void generate_skeleton(Mesh &m, double len)
{
    const size_t nf = 60;
    vector<SkeletonBoneInfo> bones(2);
    vector<Matrix4> frame(2);

    bones[0].parent = -1;
    bones[0].flags = BONE_FLAGS_FIXED_UNIFORM;
    bones[0].tail = Vector3(0.5 * len, 0.0, 0.0);
    bones[0].max_singular = 1.0;
    bones[0].translate_bounds.centre.assign(0.0, 0.0, 0.0);
    bones[0].translate_bounds.radius = 0.0;

    bones[1].parent = 0;
    bones[1].flags = BONE_FLAG_FIXED_TRANSLATE | BONE_FLAG_FIXED_SCALE | BONE_FLAG_UNIFORM_SCALE;
    bones[1].tail = Vector3(0.5 * len, 0.0, 0.0);
    bones[1].max_singular = 1.0;
    bones[1].translate_bounds.centre.assign(0.5 * len, 0.0, 0.0);
    bones[1].translate_bounds.radius = 0.0;
    bones[1].rest_transform.identity();
    bones[1].rest_transform[0][3] = len * 0.5;
    m.skeleton.get_write().set_bones(bones);

    for (size_t i = 0; i < nf; i++)
    {
        double h = high_mark.double_value;
        double l = low_mark.double_value;
        double ang = (1.0 - cos(i * G_PI * 2.0 / (nf - 1))) * (h - l) * 0.5 + l;
        Matrix4 r;
        if (twist.used)
            r = matrix_rotate(ang, Vector3(1.0, 0.0, 0.0));
        else
            r = matrix_rotate(ang, Vector3(0.0, 1.0, 0.0));
        Matrix4 t = matrix_translate(bones[1].translate_bounds.centre);
        frame[1] = t * r;
        m.skeleton.get_write().add_frame(frame);
    }
}

static Vector4 get_point_internal(const Vector4 &base, const Matrix4 &rot,
                                  double param, double len)
{
    return rot * Vector4(param * len, base[1], base[2], base[3]);
}

/* Cubic slow-in-slow-out curve, from [0,1] to [0,1] */
static inline double siso(double v)
{
    return v * v * (3 - 2 * v);
}

static Matrix4 bendpipe(double angle, double len)
{
    const Matrix4 t1 = matrix_translate(Vector3(len * 0.5, 0.0, 0.0));
    const Matrix4 t2 = matrix_translate(Vector3(len * -0.5, 0.0, 0.0));
    if (twist.used)
        return t1 * matrix_rotate(angle, Vector3(1.0, 0.0, 0.0)) * t2;
    else
        return t1 * matrix_rotate(angle, Vector3(0.0, 1.0, 0.0)) * t2;
}

static double pose_to_angle(int pose)
{
    double h = high_mark.double_value;
    double l = low_mark.double_value;
    return pose * (h - l) / (POSES - 1) + l;
}

static double pose_to_weight(int pose)
{
    if (pose == 0) return POSES;
    else return 1.0;
}

static Vector4 get_point(double ang, double param, double pose_ang, double len)
{
    double param2, scale;
    int i;

    param2 = (param - 0.5) / BLENDAREA + 0.5;
    if (param2 < 0.0) param2 = 0.0;
    if (param2 > 1.0) param2 = 1.0;

    Vector4 base(param * len, sin(ang), cos(ang), 1.0);

    if (twist.used)
        return bendpipe(pose_ang * siso(param2), len) * base;
    Matrix4 rot = bendpipe(pose_ang, len);
    Vector4 turned = rot * base;
    if (fabs(pose_ang) < 0.001 || param2 == 0.0 || param2 == 1.0)
        return base * (1.0 - param2) + turned * param2;
    else
    {
        Vector4 left = base;
        left[0] = 0.5 * (1.0 - BLENDAREA) * len;
        Vector4 right = get_point_internal(base, rot, 0.5 * (1.0 + BLENDAREA), len);
        Vector4 end = get_point_internal(base, rot, 1.0, len);
        Vector4 cent = left;
        for (i = 2; i < 3; i++)
        {
            scale = (left[0] - right[0]) * (end[0] - right[0]) / (end[i] - right[i]);
            cent[i] = right[i] - scale;
        }
        // turned is position relative to centre
        turned = base * (1.0 - param2) + turned * param2 - cent;
        return cent + length(left - cent) * normalise(turned);
    }
}

MAT *solve(MAT *A, MAT *B, MAT *X)
{
    MAT *QR = NULL;
    VEC *diag = NULL;
    VEC *b = NULL;
    VEC *x = NULL;
    size_t i;

    X = m_resize(X, A->n, B->n);
    QR = m_copy(A, MNULL);
    diag = v_get(A->m);

    QRfactor(QR, diag);
    for (i = 0; i < B->n; i++)
    {
        b = get_col(B, i, b);
        x = QRsolve(QR, diag, b, x);
        set_col(X, i, x);
    }

    M_FREE(QR);
    V_FREE(b);
    V_FREE(x);
    V_FREE(diag);
    return X;
}

static void generate_tube_ssd(vector<VectorPosition> &positions,
                              int rows, int slices, double len)
{
    vector<VectorPosition>::iterator cur = positions.begin();

    for (int i = 0; i <= rows; i++)
    {
        for (int j = 0; j < slices; j++, cur++)
        {
            double ang = j * M_PI * 2 / slices;
            double w[2];
            w[1] = (i / (double) rows - 0.5) / BLENDAREA + 0.5;
            w[1] = max(0.0, min(1.0, w[1]));
            w[1] = siso(w[1]);
            w[0] = 1.0 - w[1];
            for (int k = 0; k < 2; k++)
                if (w[k])
                {
                    (*cur)[k * 4 + 0] = (i * len / rows - k * 0.5 * len) * w[k];
                    (*cur)[k * 4 + 1] = sin(ang) * w[k];
                    (*cur)[k * 4 + 2] = cos(ang) * w[k];
                    (*cur)[k * 4 + 3] = w[k];
                }
        }
    }
}

static void generate_tube_anim(vector<VectorPosition> &positions,
                               int rows, int slices, double len)
{
    int coords = (rows + 1) * slices;
    int p, pi, i, j, v;
    MAT *A = NULL, *B = NULL, *X = NULL;
    double ang, pang, pw, param;

    A = m_get(3 * POSES + 7, 7);
    B = m_get(3 * POSES + 7, coords);
    A = m_zero(A);
    B = m_zero(B);
    v = 0;
    for (i = 0; i <= rows; i++)
        for (j = 0; j < slices; j++)
        {
            pi = 0;
            ang = j * M_PI * 2 / slices;
            param = i / (double) rows;
            for (p = 0; p < POSES; p++)
            {
                pang = pose_to_angle(p);
                pw = pose_to_weight(p);
                Vector4 pos = get_point(ang, param, pang, len);
                m_set_val(B, pi++, v, pw * pos[0]);
                m_set_val(B, pi++, v, pw * pos[1]);
                m_set_val(B, pi++, v, pw * pos[2]);
            }
            v++;
        }
    for (p = 0; p < POSES; p++)
    {
        pang = pose_to_angle(p);
        pw = pose_to_weight(p);
        for (i = 0; i < 2; i++)
        {
            Matrix4 t;
            if (i != 0) t = bendpipe(pang, len) * matrix_translate(Vector3(0.5 * len, 0.0, 0.0));
            for (j = 0; j < 3; j++)
                for (v = 0; v < 4; v++)
                    m_set_val(A, 3 * p + j, 3 * i + v, pw * t[j][v]);
        }
    }
    for (i = 0; i < 7; i++)
    {
        m_set_val(A, 3 * POSES + i, i, 0.00001);
        m_set_val(B, 3 * POSES + i, i, 0.0);
    }
    X = solve(A, B, X);

    for (i = 0; i < coords; i++)
    {
        for (v = 0; v < 2; v++)
            for (j = 0; j < 4; j++)
                positions[i][v * 4 + j] = m_entry(X, 3 * v + j, i);
        positions[i][3] = 1.0 - positions[i][7];
    }
    M_FREE(A);
    M_FREE(B);
    M_FREE(X);
}

static void generate_tube(Mesh &m, bool ssd, int rows, int slices, double len)
{
    generate_skeleton(m, len);
    const Skeleton &skeleton = m.skeleton.get_read();
    int coords = (rows + 1) * slices;

    vector<VectorPosition> positions(coords);
    for (int i = 0; i < coords; i++)
        positions[i] = skeleton.ssd_vector();
    if (ssd) generate_tube_ssd(positions, rows, slices, len);
    else generate_tube_anim(positions, rows, slices, len);

    vector<MeshRawVertex> vertices;
    vector<vector<int> > faces;
    /* Generate 4 patches around tube. We give every point in a patch
     * the same texture coordinates so that they end up on the same
     * patch, and leave parametrisation to the parametriser. */
    for (int patch = 0; patch < 4; patch++)
    {
        int start = (slices * (patch * 2 + 1)) / 8;
        int end = (slices * (patch * 2 + 3)) / 8;
        int range = end - start + 1;
        int base = vertices.size();
        MeshRawVertex cur;
        cur.texcoords[0] = patch;
        cur.texcoords[1] = 0.0;
        for (int r = 0; r <= rows; r++)
            for (int s = start; s <= end; s++)
            {
                cur.position = positions[r * slices + (s % slices)];
                vertices.push_back(cur);
            }
        for (int r = 0; r < rows; r++)
            for (int s = start; s < end; s++)
            {
                faces.push_back(vector<int>(4));
                vector<int> &f = faces.back();
                f[0] = base + r * range + s - start;
                f[3] = base + r * range + s - start + 1;
                f[2] = f[3] + range;
                f[1] = f[0] + range;
            }
    }
    for (int patch = 0; patch < 2; patch++)
    {
        int offset = patch ? rows * slices : 0;
        VectorPosition centre = positions[offset];
        for (int s = 1; s < slices; s++)
            centre += positions[offset + s];
        centre /= slices;

        MeshRawVertex cur;
        cur.texcoords[0] = patch + 4;
        cur.texcoords[1] = 0.0;
        cur.position = centre;
        vertices.push_back(cur);
        int base = vertices.size();
        for (int s = 0; s < slices; s++)
        {
            cur.position = positions[offset + s];
            vertices.push_back(cur);
            faces.push_back(vector<int>(3));
            vector<int> &f = faces.back();
            f[0] = base + s;
            f[1] = base + (s + 1) % slices;
            f[2] = base - 1;
            if (patch) reverse(f.begin(), f.end());
        }
    }

    vector<int> materials(faces.size(), 1);
    mesh_create(m, vertices, faces, materials);
}

int main(int argc, char **argv)
{
    Mesh m;
    vector<string> comments;
    comments.push_back(alod_generation_string(argc, argv));

    process_options(argc, argv);
    generate_tube(m, ssd.used, rows.int_value, slices.int_value,
                  tube_length.double_value);
    try
    {
        format_save_mesh_file(out_format.string_value, out_file, m,
                              comments);
    }
    catch (bad_format &e)
    {
        cerr << "Failed to save mesh: " << e.what() << endl;
        return 1;
    }

    return 0;
}
