#if HAVE_CONFIG_H
# include <config.h>
#endif
#include "glee/GLee.h"
#include <GL/gl.h>
#include "render/interfaces.h"
#include "render/meshrender.h"
#include "render/normal_static.h"
#include "render/geometry_sse.h"
#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <string>
#include <vector>
#include <complex>
#include <queue>

using namespace std;

class RenderHelperNormalStatic : public RenderHelper
{
private:
    std::vector<GLuint> normal_maps;
public:
    RenderHelperNormalStatic(RenderHelperClassNormalStatic &parent, const RenderMesh &m);
    virtual ~RenderHelperNormalStatic();

    virtual void begin_object(const Skeleton &skeleton, double lod, double z);
};


RenderHelper *RenderHelperClassNormalStatic::new_instance(const RenderMesh &m)
{
    return new RenderHelperNormalStatic(*this, m);
}

void RenderHelperClassNormalStatic::set_parameters(CGparameter vparameter, CGparameter fparameter)
{
    normal_map_parameters[0] = cgGetNamedStructParameter(fparameter, "normal_map_a");
    check_cg_error();
    g_assert(normal_map_parameters[0]);

    normal_map_parameters[1] = cgGetNamedStructParameter(fparameter, "normal_map_b");
    check_cg_error();
    g_assert(normal_map_parameters[1]);
}

/* Flood-fills the blank areas of the image */
static void flood_static(int width, int height, GLubyte *pixels)
{
    typedef complex<int> coord;
    static const coord dir[4] =
    {
        coord(1, 0),
        coord(0, 1),
        coord(-1, 0),
        coord(0, -1)
    };
    GLubyte (*groups)[4] = (GLubyte (*)[4]) pixels;
    queue<coord> q;

    for (int y = 0; y < height; y++)
        for (int x = 0; x < width; x++)
            if (groups[y * width + x][3])
                q.push(coord(x, y));
    while (!q.empty())
    {
        coord cur = q.front();
        q.pop();
        int index = imag(cur) * width + real(cur);
        for (int d = 0; d < 4; d++)
        {
            coord nxt = cur + dir[d];
            int x = real(nxt);
            int y = imag(nxt);
            int index2 = y * width + x;
            if (x >= 0 && x < width && y >= 0 && y < height
                && !groups[index2][3])
            {
                copy(groups[index], groups[index] + 4, groups[index2]);
                q.push(nxt);
            }
        }
    }
}

RenderHelperNormalStatic::RenderHelperNormalStatic(RenderHelperClassNormalStatic &parent, const RenderMesh &m)
: RenderHelper(parent, m)
{
    GLint old;
    glGetIntegerv(GL_TEXTURE_BINDING_2D, &old);
    /* We do render-to-texture by exploiting an existing geometry helper to
     * do the coordinate conversions. We start by constructing an
     * object-space normal map from the provided animation-space tangent map,
     * on the CPU. Then for each LOD level, we process this texture on the
     * GPU to convert it to the local tangent space.
     */
    g_assert(m.mesh.tangent_maps.size() <= 1);
    if (m.mesh.tangent_maps.empty())
    {
        /* Quick-n-dirty fallback code: use vertex normal. */
        for (size_t i = 0; i < m.levels.size(); i++)
        {
            GLubyte normal[3] = {255, 128, 128};
            GLuint tex_id;
            glGenTextures(1, &tex_id);
            glBindTexture(GL_TEXTURE_2D, tex_id);
            glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, 1, 1, 0,
                         GL_RGB, GL_UNSIGNED_BYTE, normal);
            normal_maps.resize(m.levels.size(), tex_id);
        }
        glBindTexture(GL_TEXTURE_2D, old);
        return;
    }
    const MeshTangentMap &tm = m.mesh.tangent_maps.front();
    Skeleton skeleton = m.mesh.skeleton.get_read();
    skeleton.set_to_rest();
    int width = tm.width;
    int height = tm.height;
    GLuint normal_texture;
    GLfloat *obj_data = new GLfloat[width * height * 4];

    int addr = 0;
    for (int y = 0; y < height; y++)
        for (int x = 0; x < width; x++, addr++)
        {
            Vector3 a = skeleton.ssd_to_3d(tm.data()[addr].tangent_plane.tangents[0]);
            Vector3 b = skeleton.ssd_to_3d(tm.data()[addr].tangent_plane.tangents[1]);
            Vector3 n = normalise(cross(a, b));
            for (int i = 0; i < 3; i++)
                obj_data[addr * 4 + i] = n[i];
        }
    glGenTextures(1, &normal_texture);
    glBindTexture(GL_TEXTURE_2D, normal_texture);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F_ARB, width, height, 0,
                 GL_RGBA, GL_FLOAT, obj_data);
    delete[] obj_data;
    glBindTexture(GL_TEXTURE_2D, old);

    /**********************************************************/

    RenderHelperClass *geometry_class;
    RenderHelper *geometry_helper;
    CGprogram vertex_program, fragment_program;
    CGparameter geometry_cg_class, geometry_cg_interface;
    CGparameter normal_parameter;
    GLuint fbo, rbo;
    GLubyte *pixels;
    GLint max_attribs;
    TextureManager manager;

    const char *args[] = {"-DVERTEX_TANGENTS", "-DVERTEX_NORMALS", "-DTEXCOORDS", NULL};
    vertex_program = load_cg_program("rtt_vp.cg", global_cg_vertex_profile(), args);
    fragment_program = load_cg_program("snm_fp.cg", global_cg_fragment_profile(), args);

    glPushMatrix();
    glLoadIdentity();  /* Prevent model transformations from being applied to normal */
    /* Prevent previously activated arrays from spoiling the party */
    glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);
    glGetIntegerv(GL_MAX_VERTEX_ATTRIBS_ARB, &max_attribs);
    for (int i = 1; i < max_attribs; i++)
        glDisableVertexAttribArrayARB(i);

    geometry_class = new RenderHelperClassGeometrySSE(true);
    geometry_cg_class = geometry_class->begin_vertex_program(vertex_program);
    geometry_cg_interface = cgGetNamedParameter(vertex_program, "geometry");
    cgConnectParameter(geometry_cg_class, geometry_cg_interface);
    check_cg_error();

    cgCompileProgram(vertex_program); check_cg_error();
    cgCompileProgram(fragment_program); check_cg_error();
    cgGLLoadProgram(vertex_program); check_cg_error();
    cgGLLoadProgram(fragment_program); check_cg_error();

    normal_parameter = cgGetNamedParameter(fragment_program, "obj_normal_texture");
    manager.bind_texture(normal_parameter, normal_texture);
    geometry_class->begin_state(vertex_program, fragment_program,
                                RENDER_FLAGS_VERTEX_DERIVATIVES | RENDER_FLAG_TEXCOORDS);
    geometry_helper = geometry_class->new_instance(m);

    normal_maps.resize(m.levels.size());
    glGenFramebuffersEXT(1, &fbo);
    glGenRenderbuffersEXT(1, &rbo);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, rbo);
    glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_RGBA8, width, height);
    glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT,
                                 GL_RENDERBUFFER_EXT, rbo);
    pixels = new GLubyte[width * height * 4];

    glDisable(GL_DEPTH_TEST);
    GLint old_viewport[4];
    glGetIntegerv(GL_VIEWPORT, old_viewport);
    glViewport(0, 0, width, height);
    glGetIntegerv(GL_TEXTURE_2D, &old);
    for (size_t i = 0; i < m.levels.size(); i++)
    {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo);

        geometry_helper->begin_object(skeleton, i, 0.0);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, m.levels[i].index_buffer);
        glClear(GL_COLOR_BUFFER_BIT);
        glDrawRangeElements(GL_TRIANGLES, 0, m.levels[i].vertices.size() - 1,
                            3 * m.levels[i].faces.size(), GL_UNSIGNED_INT, NULL);
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
        glReadPixels(0, 0, width, height, GL_RGBA, GL_UNSIGNED_BYTE, pixels);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
        geometry_helper->end_object();

        flood_static(width, height, pixels);
        glGetIntegerv(GL_TEXTURE_BINDING_2D, &old);
        glGenTextures(1, &normal_maps[i]);
        glBindTexture(GL_TEXTURE_2D, normal_maps[i]);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP, GL_TRUE);
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, width, height, 0,
                     GL_RGBA, GL_UNSIGNED_BYTE, pixels);
        glBindTexture(GL_TEXTURE_2D, old);
    }

    glViewport(old_viewport[0], old_viewport[1], old_viewport[2], old_viewport[3]);
    glEnable(GL_DEPTH_TEST);
    glDeleteFramebuffersEXT(1, &fbo);
    glDeleteRenderbuffersEXT(1, &rbo);
    delete[] pixels;
    delete geometry_helper;
    geometry_class->end_state();
    delete geometry_class;
    cgGLBindProgram(get_class().get_vertex_program());
    cgGLBindProgram(get_class().get_fragment_program());
    cgDestroyProgram(vertex_program);
    cgDestroyProgram(fragment_program);
    cgDestroyParameter(geometry_cg_class);
    glDeleteTextures(1, &normal_texture);
    glPopClientAttrib();
    glPopMatrix();
}

RenderHelperNormalStatic::~RenderHelperNormalStatic()
{
    for (size_t i = 0; i < normal_maps.size(); i++)
        glDeleteTextures(1, &normal_maps[i]);
}

void RenderHelperNormalStatic::begin_object(const Skeleton &skeleton, double lod, double z)
{
    RenderHelperClassNormalStatic &klass = static_cast<RenderHelperClassNormalStatic &>(get_class());
    int ilod0 = rmesh.int_lod(lod);
    int ilod1 = rmesh.int_lod(lod + 1.0);

    RenderHelper::begin_object(skeleton, lod, z);
    cgGLSetTextureParameter(klass.normal_map_parameters[0], normal_maps[ilod0]);
    cgGLSetTextureParameter(klass.normal_map_parameters[1], normal_maps[ilod1]);
    cgGLEnableTextureParameter(klass.normal_map_parameters[0]);
    cgGLEnableTextureParameter(klass.normal_map_parameters[1]);
}
